/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include <trompeloeil.hpp>
#include "ds/catch/catch.h"

#include <boost/smart_ptr/make_shared.hpp>

#include "cache/apartment-proxy.h"
#include "device-cache.h"
#include "ds/conf-mock.h"
#include "ds/log.h"
#include "ds485/dsactionrequest.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "hal/hardware-abstraction-mock.h"
#include "model/apartment.h"
#include "model/group.h"
#include "model/set.h"
#include "model/zone.h"

static ds::LogChannel dsLogChannel("deviceCacheTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][cache]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};
static const Dsuid dsuid4{"0000000000000000000000000000000004"};
static const Dsuid dsuid5{"0000000000000000000000000000000005"};

SCENARIO("SaveScene invocation", TAGS) {
  GIVEN("The cache contains 3 devices") {
    ds::sqlite3::Database d(":memory:");
    Apartment apt(nullptr);
    HardwareAbstraction hal{&apt, &d};
    auto& cacheDb = hal.getCache();
    ds::ConfMock conf;
    auto proxy = makeApartmentProxy(&apt, &hal, &d, nullptr, conf);
    DeviceCache A;
    A.scene.sceneTable[1].channels[ChannelType::brightness].value = 11;
    A.scene.sceneTable[2].channels[ChannelType::brightness].value = 21;
    DeviceCache B;
    B.scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value = 12;
    B.scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value = 22;
    DeviceCache C;
    C.scene.sceneTable[1].channels[ChannelType::brightness].value = 13;
    C.scene.sceneTable[1].channels[ChannelType::shadeOpeningAngleOutside].value = 14;
    C.scene.sceneTable[2].channels[ChannelType::brightness].value = 23;
    C.scene.sceneTable[2].channels[ChannelType::shadeOpeningAngleOutside].value = 24;
    DeviceCache D;
    D.scene.sceneTable[1].channels[ChannelType::brightness].value = 13;
    D.scene.sceneTable[1].flags = SceneFlags{.dontCare = false};
    D.scene.sceneTable[6].channels[ChannelType::brightness].value = 13;
    D.scene.sceneTable[6].flags = SceneFlags{.dontCare = false};
    cacheDb.update(dsuid1, A);
    cacheDb.update(dsuid2, B);
    cacheDb.update(dsuid3, C);
    cacheDb.update(dsuid4, D);

    WHEN("the scene 1 is invalidated for one device") {
      proxy->saveScene(dsuid1, 1);
      proxy->completeProcessing();

      THEN("values for scene 1 are deleted") {
        REQUIRE(!cacheDb.tryLookup(dsuid1).scene.sceneTable[1].channels[ChannelType::brightness].value);
      }
      AND_THEN("values for the other scenes are still intact") {
        REQUIRE(*cacheDb.tryLookup(dsuid1).scene.sceneTable[2].channels[ChannelType::brightness].value == 21);
      }
      AND_THEN("values for the other devices are still intact") {
        REQUIRE(*cacheDb.tryLookup(dsuid2).scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value == 12);
        REQUIRE(*cacheDb.tryLookup(dsuid2).scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value == 22);
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[1].channels[ChannelType::brightness].value == 13);
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[1].channels[ChannelType::shadeOpeningAngleOutside].value ==
                14);
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[2].channels[ChannelType::brightness].value == 23);
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[2].channels[ChannelType::shadeOpeningAngleOutside].value ==
                24);
      }
    }

    WHEN("the scene 6 (area) is invalidated") {
      proxy->saveScene(dsuid4, 6);
      proxy->completeProcessing();

      THEN("scene 1 and the flags for 1 and 6 are removed") {
        REQUIRE(!cacheDb.tryLookup(dsuid4).scene.sceneTable[1].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dsuid4).scene.sceneTable[6].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dsuid4).scene.sceneTable[1].flags.dontCare);
        REQUIRE(!cacheDb.tryLookup(dsuid4).scene.sceneTable[6].flags.dontCare);
      }
    }

    WHEN("the scene 1 is invalidated on a device with multiple channels") {
      proxy->saveScene(dsuid3, 1);
      proxy->completeProcessing();

      THEN("values for scene 1 are deleted") {
        REQUIRE(!cacheDb.tryLookup(dsuid3).scene.sceneTable[1].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dsuid3).scene.sceneTable[1].channels[ChannelType::shadeOpeningAngleOutside].value);
      }
      AND_THEN("values for the other scenes are still intact") {
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[2].channels[ChannelType::brightness].value == 23);
        REQUIRE(*cacheDb.tryLookup(dsuid3).scene.sceneTable[2].channels[ChannelType::shadeOpeningAngleOutside].value ==
                24);
      }
      AND_THEN("values for the other device are still intact") {
        REQUIRE(*cacheDb.tryLookup(dsuid1).scene.sceneTable[1].channels[ChannelType::brightness].value == 11);
        REQUIRE(*cacheDb.tryLookup(dsuid1).scene.sceneTable[2].channels[ChannelType::brightness].value == 21);
        REQUIRE(*cacheDb.tryLookup(dsuid2).scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value == 12);
        REQUIRE(*cacheDb.tryLookup(dsuid2).scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value == 22);
      }
    }

    WHEN("an yet unknown but valid scene is invalidated") {
      THEN("the call succeeds") { REQUIRE_NOTHROW(proxy->saveScene(dsuid3, 127)); }
    }

    WHEN("an invalid scene 129 is invalidated") {
      THEN("the call fails with an exception") { REQUIRE_THROWS(proxy->saveScene(dsuid3, 129)); }
    }
  }
}

SCENARIO("Scenario state", TAGS) {
  ds::sqlite3::Database db(":memory:");
  Apartment apt(nullptr);
  EventSinkMock eventSink;
  apt.setEventSink(eventSink);
  MockHal hal(&apt, &db);
  ds::DateTime now = std::chrono::high_resolution_clock::now();
  ds::ConfMock conf;

  DeviceSpec spec;
  spec.FunctionID = 0x1010;

  auto zone1 = apt.allocateZone(1);
  auto zone2 = apt.allocateZone(2);
  auto dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->addToGroup(1);
  dev1->setIsPresent(true);
  dev1->setIsConnected(true);
  auto dev2 = apt.allocateDevice(dsuid2);
  dev2->setPartiallyFromSpec(spec);
  dev2->addToGroup(1);
  dev2->setIsPresent(true);
  dev2->setIsConnected(true);
  auto dev3 = apt.allocateDevice(dsuid3);
  dev3->setPartiallyFromSpec(spec);
  dev3->addToGroup(1);
  dev3->setIsPresent(true);
  dev3->setIsConnected(true);
  auto dev4 = apt.allocateDevice(dsuid4);
  dev4->setPartiallyFromSpec(spec);
  dev4->addToGroup(2);
  dev4->setIsPresent(true);
  dev4->setIsConnected(true);
  auto dev5 = apt.allocateDevice(dsuid5);
  dev5->setPartiallyFromSpec(spec);
  dev5->addToGroup(2);
  dev5->setIsPresent(false);
  dev5->setIsConnected(false);

  DeviceReference d1(dev1, &apt);
  DeviceReference d2(dev2, &apt);
  DeviceReference d3(dev3, &apt);
  DeviceReference d4(dev4, &apt);
  DeviceReference d5(dev5, &apt);
  zone1->addDevice(d1);
  zone1->addDevice(d2);
  zone2->addDevice(d3);
  zone2->addDevice(d4);
  zone2->addDevice(d5);

  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);
  CacheDb<ScenarioCache> sc{&db, false};
  auto& dc = hal.getCache();
  {
    auto cache = dc.tryLookup(dsuid1);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = false;
    cache.scene.sceneTable[17].flags.dontCare = true;
    cache.scene.sceneTable[18].flags.dontCare = false;
    dc.update(dsuid1, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid2);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = true;
    cache.scene.sceneTable[17].flags.dontCare = false;
    cache.scene.sceneTable[18].flags.dontCare = false;
    dc.update(dsuid2, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid3);
    cache.scene.sceneTable[0].flags.dontCare = false;
    dc.update(dsuid3, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid4);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = false;
    dc.update(dsuid4, cache);
  }

  GIVEN("A scene is called in a zone/group") {
    REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
    REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);
    REQUIRE_CALL(eventSink, pushEvent(_));

    REQUIRE_NOTHROW(proxy->callScene(1, 1, 5, false, now));
    proxy->completeProcessing();
    {
      auto scenario = sc.tryLookup(DSUID_NULL);
      REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
    }

    WHEN("A device changes disjunct state") {
      REQUIRE_CALL(hal, callScene(dsuid1, 17, false, _)).RETURN(false);
      REQUIRE_CALL(hal, callScene(dsuid2, 17, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, 17, false, now));
      proxy->completeProcessing();
      THEN("the scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s17"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("A device changes shared state") {
      REQUIRE_CALL(hal, callScene(dsuid1, 18, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 18, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, 18, false, now));
      proxy->completeProcessing();
      THEN("the scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s18"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("callSceneMin is called") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, callSceneMin(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callSceneMin(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("undoScene is called") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, undoScene(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->undoScene(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setOutputValue is called") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, _, _)).RETURN(true);
      REQUIRE_CALL(hal, setOutputValue(dsuid2, _, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setOutputValue(1, 1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("A device callScene is called") {
      REQUIRE_CALL(hal, callScene(dsuid1, 17, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(dsuid1, 17, false, now));
      proxy->completeProcessing();
      THEN("the scenario becomes active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s17"].status == dsModel::ScenarioStatusStatus::none);
      }
    }
    WHEN("callSceneMin is called on device") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callSceneMin(dsuid1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("undoScene is called on device") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->undoScene(dsuid1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setOutputValue is called on device") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, 58, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setOutputValue(dsuid1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setOutputValue is called on device") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, ChannelType::brightness, 58.25, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setOutputValue(dsuid1, ChannelType::brightness, 58.25, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setSceneLocalPrio is called on device") {
      REQUIRE_CALL(hal, setSceneLocalPrio(dsuid1, 6)).RETURN(true);
      REQUIRE_CALL(hal, setSceneLocalPrio(dsuid2, 6)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setSceneLocalPrio(1, 1, 6));
      proxy->completeProcessing();
      THEN("the scenario does not change") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("setOperationLock is called on device") {
      REQUIRE_CALL(hal, setOperationLock(dsuid1, true));

      REQUIRE_NOTHROW(proxy->setOperationLock(dsuid1, 5));
      proxy->completeProcessing();
      THEN("the scenario does not change") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("Querying scenarios work") {
      auto scenarioStatus = proxy->zoneScenarioStatus(dsModel::ScenarioId{"applicationZone-z1-g1-s5"});
      REQUIRE(scenarioStatus.id == dsModel::ScenarioId{"applicationZone-z1-g1-s5"});
      REQUIRE(scenarioStatus.status == dsModel::ScenarioStatusStatus::active);
      auto fakeScenarioStatus = proxy->zoneScenarioStatus(dsModel::ScenarioId{"applicationZone-z13-g4-s1"});
      REQUIRE(fakeScenarioStatus.id == dsModel::ScenarioId{"applicationZone-z13-g4-s1"});
      REQUIRE(fakeScenarioStatus.status == dsModel::ScenarioStatusStatus{});
    }
  }

  GIVEN("A scene is called in the whole apartment") {
    REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
    REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);
    REQUIRE_CALL(hal, callScene(dsuid3, 5, false, _)).RETURN(true);
    REQUIRE_CALL(eventSink, pushEvent(_));

    REQUIRE_NOTHROW(proxy->callScene(0, 1, 5, false, now));
    proxy->completeProcessing();
    {
      auto scenario = sc.tryLookup(DSUID_NULL);
      REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
    }

    WHEN("A device changes in a zone state") {
      REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, 5, false, now));
      proxy->completeProcessing();
      THEN("the apartment scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("A device changes in another group") {
      REQUIRE_CALL(hal, callScene(dsuid4, 5, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(2, 2, 5, false, now));
      proxy->completeProcessing();
      THEN("the apartment scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z2-g2-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("An broadcast scene is called") {
      REQUIRE_CALL(hal, callScene(dsuid1, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid3, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid4, 0, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(0, 0, 0, false, now));
      proxy->completeProcessing();
      THEN("all scenarios are inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z0-g0-s0"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("callSceneMin is called") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, callSceneMin(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callSceneMin(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("undoScene is called") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, undoScene(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->undoScene(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setOutputValue is called") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, _, _)).RETURN(true);
      REQUIRE_CALL(hal, setOutputValue(dsuid2, _, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setOutputValue(1, 1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
  }
  GIVEN("a certain scenario on a group") {
    THEN("get a list of participating devices for scene 5") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 5);
      REQUIRE(result.size() == 1);
      REQUIRE(result[0] == ds::str(dsuid1));
    }
    THEN("get a list of participating devices for scene 18") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 18);
      REQUIRE(result.size() == 2);
      REQUIRE(result[0] == ds::str(dsuid1));
      REQUIRE(result[1] == ds::str(dsuid2));
    }
    THEN("get an empty list of participating devices for scene 6") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 6);
      REQUIRE(result.size() == 0);
    }
  }
}

SCENARIO("Read cache status") {
  ds::sqlite3::Database db(":memory:");
  Apartment apt(nullptr);
  EventSinkMock eventSink;
  apt.setEventSink(eventSink);
  MockHal hal(&apt, &db);
  ds::ConfMock conf;

  DeviceSpec spec;
  spec.FunctionID = 0x1010;

  auto zone1 = apt.allocateZone(1);
  auto zone2 = apt.allocateZone(2);
  auto dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->addToGroup(1);
  dev1->setIsPresent(true);
  dev1->setIsConnected(true);
  auto dev2 = apt.allocateDevice(dsuid2);
  dev2->setPartiallyFromSpec(spec);
  dev2->addToGroup(1);
  dev2->setIsPresent(true);
  dev2->setIsConnected(true);
  auto dev3 = apt.allocateDevice(dsuid3);
  dev3->setPartiallyFromSpec(spec);
  dev3->addToGroup(1);
  dev3->setIsPresent(true);
  dev3->setIsConnected(true);
  auto dev4 = apt.allocateDevice(dsuid4);
  dev4->setPartiallyFromSpec(spec);
  dev4->addToGroup(2);
  dev4->setIsPresent(true);
  dev4->setIsConnected(true);
  auto dev5 = apt.allocateDevice(dsuid5);
  dev5->setPartiallyFromSpec(spec);
  dev5->addToGroup(2);
  dev5->setIsPresent(false);
  dev5->setIsConnected(false);

  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  WHEN("Reading complete cache") {
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid1), .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid2), .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid3), .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid4), .total = 149, .completed = 149, .completion = 1});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 4);
    REQUIRE(*status.cache.completion == 100);
    REQUIRE(status.devices.size() == 4);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::complete);
  }

  WHEN("Reading partially complete cache") {
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid1), .total = 149, .completed = 75, .completion = 0.5});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid2), .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid3), .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid4), .total = 149, .completed = 149, .completion = 1});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 3);
    REQUIRE(roundDigits(*status.cache.completion, 1) == 87.6);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::incomplete);
    REQUIRE(status.devices.size() == 4);
  }

  WHEN("Reading empty cache") {
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid1), .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid2), .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid3), .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = ds::str(dsuid4), .total = 149, .completed = 0, .completion = 0});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 0);
    REQUIRE(roundDigits(*status.cache.completion, 3) == 0);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::incomplete);
    REQUIRE(status.devices.size() == 4);
  }
}

} // namespace
