/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "cache/cache-updater.h"

#include <ds/asio/thread-pool.h>
#include <event.h>
#include <boost/smart_ptr/make_shared.hpp>

#include <cmath>
#include <future>
#include <map>
#include <string>
#include <thread>
#include <tuple>
#include "cache/apartment-proxy.h"
#include "ds-model/model.h"
#include "ds-scenarios.h"
#include "ds/log.h"
#include "dss.h"
#include "event.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/modelconst.h"
#include "model/scenehelper.h"
#include "model/set.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "propertysystem_common_paths.h"

using ds::asio::ThreadPool;
using dss::Cluster;
using namespace std::chrono_literals;

DS_JSON_DEFINE_STRUCT((dss), SceneFlags);
DS_JSON_DEFINE_STRUCT((dss), SceneChannel);
DS_JSON_DEFINE_STRUCT((dss), SceneEntry);
DS_JSON_DEFINE_STRUCT((dss), DeviceSceneCache);
DS_JSON_DEFINE_STRUCT((dss), DeviceCache);
DS_JSON_DEFINE_STRUCT((dss), ScenarioCache);
DS_JSON_DEFINE_STRUCT((dss), CacheScenarioState);

namespace dss {

static ds::LogChannel dsLogChannel("Cache", ds::LogSeverity::NOTICE);

static constexpr ds::Clock::duration ReadoutTimePerParameter = 5s;

class ApartmentProxyEventPlugin : public EventInterpreterPlugin {
private:
  ApartmentProxy* m_proxy;

public:
  ApartmentProxyEventPlugin(ApartmentProxy* proxy, EventInterpreter* pInterpreter)
      : EventInterpreterPlugin("ApartmentProxy", pInterpreter), m_proxy(proxy) {}

  void subscribe() {
    getEventInterpreter().subscribe(
        boost::make_shared<EventSubscription>(*this, EventName::ApartmentProxyDeviceTimeout));
    getEventInterpreter().subscribe(
        boost::make_shared<EventSubscription>(*this, EventName::ApartmentProxyDeviceDimmingTimeout));
  }

  void handleEvent(Event& event, const EventSubscription& subscription) {
    if (!DSS::hasInstance()) {
      return;
    }
    auto deviceRef = event.getRaisedAtDevice();
    if (!deviceRef) {
      return;
    }
    auto device = deviceRef->getDevice();
    if (!device) {
      return;
    }
    if (event.getName() == EventName::ApartmentProxyDeviceTimeout) {
      m_proxy->completeMoving(device->getDSID());
    } else if (event.getName() == EventName::ApartmentProxyDeviceDimmingTimeout) {
      m_proxy->stopDimming(device->getDSID());
    }
  }
};

class ApartmentProxyImpl : public ApartmentProxy {
public:
  ApartmentProxyImpl(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                     EventInterpreter* pInterpreter, const ds::Conf& conf)
      : m_hal(hal),
        m_scenarioCache(db),
        m_deviceCache(hal->getCache()),
        m_apartment(apt),
        m_pool(1),
        m_cacheReady(true),
        m_emptyClickPreventionActive(true) {
    m_emptyClickPreventionActive = conf.readOr<bool>("emptyClickPrevention", true);
    if (pInterpreter) {
      EventInterpreterPlugin* plugin = new ApartmentProxyEventPlugin(this, pInterpreter);
      pInterpreter->addPlugin(plugin);
    }
    if (auto m_pPropSystem = m_apartment->getPropertySystem()) {
      auto cacheReadyProperty = m_pPropSystem->createProperty(pp_device_cache_ready);
      cacheReadyProperty->linkToProxy(PropertyProxyReference<bool>(m_cacheReady, false));
    }
  }

  void completeMoving(const Dsuid& device) const override {
    auto now = std::chrono::high_resolution_clock::now();
    m_pool.post([this, device, now]() {
      DS_DEBUG_ENTER(device, now);
      try {
        if (m_hal->completeMoving(device, now)) {
          auto devRef = DeviceReference(device, m_apartment);
          m_apartment->pushEvent(boost::make_shared<Event>(EventName::ApartmentProxyStateChanged, devRef));
        }
      } catch (const std::exception& e) {
        DS_NOTICE("completeMoving", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void saveScene(const Dsuid& device, int sceneId) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId]() {
      DS_DEBUG_ENTER(device, sceneId);
      try {
        auto cacheEntry = m_hal->getCache().tryLookup(device);
        for (auto& channel : cacheEntry.scene.sceneTable[sceneId].channels) {
          channel.second.value.reset();
        }
        if (SceneHelper::isAreaScene(sceneId) && sceneId >= SceneA11) {
          // saving an areaOn scene also changes the related areaOff scene
          for (auto& channel : cacheEntry.scene.sceneTable[sceneId - SceneA11 + SceneOffA1].channels) {
            channel.second.value.reset();
          }
          // and the flags for both scenes
          cacheEntry.scene.sceneTable[sceneId].flags = {};
          cacheEntry.scene.sceneTable[sceneId - SceneA11 + SceneOffA1].flags = {};
        }
        m_hal->getCache().update(device, cacheEntry);
        m_cacheReady = false;
        m_apartment->pushEvent(boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent));
      } catch (const std::exception& e) {
        DS_NOTICE("saveScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void callScene(const Dsuid& device, int sceneId, bool forced, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, forced, time]() {
      DS_DEBUG_ENTER(device, sceneId, forced, time);
      try {
        if (m_hal->callScene(device, sceneId, forced, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("callScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void callScene(int zoneId, int groupId, int sceneId, bool forced, ds::DateTime time) const override {
    m_pool.post([this, zoneId, groupId, sceneId, forced, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, forced, time);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->callScene(s.get(i).getDSID(), sceneId, forced, time)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("callScene", e.what());
        }
      }

      updateScenarios(changedDevices, zoneId, groupId, time);

      if (s.length() > 0 && SceneHelper::rememberScene(sceneId) && groupId != GroupIDControlTemperature) {
        // mark current scenario active
        auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
        scenarios.scenarios[encodeId(ApplicationZoneScenarioId{.zone = zoneId, .group = groupId, .scene = sceneId})] = {
            .sceneId = sceneId,
            .zoneId = zoneId,
            .groupId = groupId,
            .status = dsModel::ScenarioStatusStatus::active,
            .lastChanged = time};
        m_scenarioCache.update(DSUID_NULL, scenarios);
      }
      DS_DEBUG_LEAVE();
    });
  }

  void callSceneMin(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->callSceneMin(s.get(i).getDSID(), sceneId, time)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("callSceneMin", e.what());
        }
      }

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void callSceneMin(const Dsuid& device, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, time]() {
      DS_DEBUG_ENTER(device, sceneId, time);
      try {
        if (m_hal->callSceneMin(device, sceneId, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("callSceneMin", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void undoScene(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->undoScene(s.get(i).getDSID(), sceneId, time)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("undoScene", e.what());
        }
      }

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void undoScene(const Dsuid& device, int sceneId, ds::DateTime time) const override {
    DS_REQUIRE(sceneId <= MaxSceneNumber, "SceneId need to be valid");
    m_pool.post([this, device, sceneId, time]() {
      DS_DEBUG_ENTER(device, sceneId, time);
      try {
        if (m_hal->undoScene(device, sceneId, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("undoScene", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(int zoneId, int groupId, uint8_t value, ds::DateTime time) const override {
    m_pool.post([this, zoneId, groupId, value, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, time);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->setOutputValue(s.get(i).getDSID(), value, time)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("setOutputValue", e.what());
        }
      }

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(const Dsuid& device, uint8_t value, ds::DateTime time) const override {
    m_pool.post([this, device, value, time]() {
      DS_DEBUG_ENTER(device, time);
      try {
        if (m_hal->setOutputValue(device, value, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("setOutputValue", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOutputValue(const Dsuid& device, ChannelType channel, double value, ds::DateTime time) const override {
    m_pool.post([this, device, channel, value, time]() {
      DS_DEBUG_ENTER(device, channel, time);
      try {
        if (m_hal->setOutputValue(device, channel, value, time)) {
          updateScenarios(device, time);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("setOutputValue", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setSceneLocalPrio(int zoneId, int groupId, int sceneId) const override {
    m_pool.post([this, zoneId, groupId, sceneId]() {
      DS_DEBUG_ENTER(zoneId, groupId);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->setSceneLocalPrio(s.get(i).getDSID(), sceneId)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("setSceneLocalPrio", e.what());
        }
      }
      if (changedDevices.size() > 0) {
        sendChangeEvent();
      }
      DS_DEBUG_LEAVE();
    });
  }

  void setOperationLock(const Dsuid& device, bool lock) const override {
    m_pool.post([this, device, lock]() {
      DS_DEBUG_ENTER(device);
      try {
        m_hal->setOperationLock(device, lock);
      } catch (const std::exception& e) {
        DS_NOTICE("setOperationLock", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  void stopDimming(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {
    m_pool.post([this, zoneId, groupId, sceneId, time]() {
      DS_DEBUG_ENTER(zoneId, groupId, sceneId, time);
      boost::shared_ptr<Zone> zone = m_apartment->getZone(zoneId);
      Set s = zone->getDevices().getByGroup(groupId).getByPresence(true);

      std::vector<Dsuid> changedDevices{};

      for (int i = 0; i < s.length(); i++) {
        try {
          if (m_hal->stopDimming(s.get(i).getDSID(), sceneId, time)) {
            changedDevices.push_back(s.get(i).getDSID());
          }
        } catch (const std::exception& e) {
          DS_NOTICE("stopDimming", e.what());
        }
      }

      updateScenarios(changedDevices, zoneId, groupId, time);
      DS_DEBUG_LEAVE();
    });
  }

  void stopDimming(const Dsuid& device) const override {
    auto now = std::chrono::high_resolution_clock::now();
    m_pool.post([this, device, now]() {
      DS_DEBUG_ENTER(device);

      try {
        if (m_hal->stopDimming(device, 0, now)) {
          updateScenarios(device, now);
        }
      } catch (const std::exception& e) {
        DS_NOTICE("stopDimming", e.what());
      }
      DS_DEBUG_LEAVE();
    });
  }

  dsModel::ScenarioStatus zoneScenarioStatus(const dsModel::ScenarioId& scenarioId) const override {
    DS_DEBUG_ENTER(scenarioId);
    try {
      auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
      for (const auto& scenario : scenarios.scenarios) {
        if (scenarioId == dsModel::ScenarioId{scenario.first}) {
          return {.id = dsModel::ScenarioId{scenario.first},
                  .status = scenario.second.status,
                  .completed = {},
                  .scheduledAt = scenario.second.lastChanged,
                  .terminatesAt = scenario.second.lastChanged};
        }
      }
      DS_INFO("Zone scenario could not be found in cache", scenarioId);
    } catch (std::exception& e) {
      DS_ERROR("Error looking up zone scenario in cache", e);
    }
    DS_DEBUG_LEAVE();
    return {.id = scenarioId};
  }

  dsModel::DSDeviceStatus dsDeviceStatus(const dsModel::DeviceId& id) const override {
    DS_DEBUG_ENTER(id);
    auto state = m_hal->getState(Dsuid{id});

    auto dsOutputs = std::vector<dsModel::OutputStatus>{};
    auto cache = m_hal->getCache().tryLookup(Dsuid{id});
    if (!(cache.property[DeviceProperty::outputMode] &&
          *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED)) {
      for (const auto& output : state.outputs) {
        dsOutputs.push_back({
            .id = ds::str(output.first),
            .value = (state.status == dsModel::OutputStatusStatus::moving || state.status == dsModel::OutputStatusStatus::dimming) ? std::optional<double>{} : output.second.value,
            .initialValue = (state.status == dsModel::OutputStatusStatus::moving || state.status == dsModel::OutputStatusStatus::dimming) ? output.second.value : std::optional<double>{},
            .status = state.status,
            .targetValue = output.second.targetValue,
            .startedAt = (state.status == dsModel::OutputStatusStatus::moving || state.status == dsModel::OutputStatusStatus::dimming) ? state.lastChanged : ds::DateTime{},
            .terminatesAt = (state.terminatesAt) ? *state.terminatesAt : ds::DateTime{},
            .level = output.second.level,
        });
      }
    }

    DS_DEBUG_LEAVE();
    return dsModel::DSDeviceStatus{.id = id,
                                   .type = "dsDeviceStatus",
                                   .lastChanged = {},
                                   .attributes = {.functionBlocks = {{
                                                      .id = dsModel::FunctionBlockId{id},
                                                      .outputs = std::move(dsOutputs),
                                                  }}}};
  }

  dsModel::LightsStatus lightsStatus(const dsModel::ZoneId& id) const override {
    DS_DEBUG_ENTER(id);
    auto result = dsModel::LightsStatus{};
    auto zone = m_apartment->zone(id);

    auto devices = zone->getDevices().getByPresence(true).getByGroup(DEVICE_CLASS_GE);
    if (devices.size() > 0) {
      dsModel::LightsStatusStatus zoneStateDsm{};
      std::tie(result, zoneStateDsm) = getLightApplicationStatus(devices);
      auto lightStatusToDsmApi = [&](const dsModel::LightsStatusStatus& status) {
        if (status == dsModel::LightsStatusStatus::none) {
          return -1;
        } else if (status == dsModel::LightsStatusStatus::off) {
          return 0;
        } else {
          return 1;
        }
      };
      if (m_emptyClickPreventionActive && m_cacheReady && m_apartment->getBusInterface() &&
          m_apartment->getBusInterface()->getActionRequestInterface()) {
        m_apartment->getBusInterface()->getActionRequestInterface()->setState(
            zone->getGroup(GroupIDYellow).get(), lightStatusToDsmApi(zoneStateDsm), lightStatusToDsmApi(result.area1),
            lightStatusToDsmApi(result.area2), lightStatusToDsmApi(result.area3), lightStatusToDsmApi(result.area4));
      }
    }
    return result;
  }
  dsModel::ShadesStatus shadesStatus(const dsModel::ZoneId& id) const override {
    DS_DEBUG_ENTER(id);
    auto result = dsModel::ShadesStatus{};
    auto zone = m_apartment->zone(id);

    auto devices = zone->getDevices().getByPresence(true).getByGroup(DEVICE_CLASS_GR);
    if (devices.size() > 0) {
      dsModel::ShadesStatusStatus zoneStateDsm{};
      std::tie(result, zoneStateDsm) = getShadeApplicationStatus(devices);
      auto shadeAreaStatusToDsmApi = [&](const dsModel::ShadesStatusStatus& status) {
        if (status == dsModel::ShadesStatusStatus::none) {
          return -1;
        } else if (status == dsModel::ShadesStatusStatus::closed) {
          return 0;
        } else {
          return 1;
        }
      };
      auto shadeZoneStatusToDsmApi = [&](const dsModel::ShadesStatusStatus& status) {
        if (status == dsModel::ShadesStatusStatus::none || status == dsModel::ShadesStatusStatus::closed) {
          return 0;
        } else {
          return 1;
        }
      };
      auto shadeMovingStatus = [&](const dsModel::ShadesStatusStatus& status) {
        if (status == dsModel::ShadesStatusStatus::moving || status == dsModel::ShadesStatusStatus::none) {
          return true;
        } else {
          return false;
        }
      };
      if (m_emptyClickPreventionActive && m_cacheReady && m_apartment->getBusInterface() &&
          m_apartment->getBusInterface()->getActionRequestInterface()) {
        m_apartment->getBusInterface()->getActionRequestInterface()->setState(
            zone->getGroup(GroupIDGray).get(), shadeZoneStatusToDsmApi(zoneStateDsm),
            shadeAreaStatusToDsmApi(result.area1), shadeAreaStatusToDsmApi(result.area2),
            shadeAreaStatusToDsmApi(result.area3), shadeAreaStatusToDsmApi(result.area4));
        if (!shadeMovingStatus(zoneStateDsm)) {
          m_apartment->getBusInterface()->getActionRequestInterface()->stopStopTimer(zone->getGroup(GroupIDGray).get(),
                                                                                     0);
        }
        for (int area = 1; area <= 4; ++area) {
          if (!shadeMovingStatus([&]() {
                switch (area) {
                  case 1: return result.area1;
                  case 2: return result.area2;
                  case 3: return result.area3;
                  case 4: return result.area4;
                }
                return dsModel::ShadesStatusStatus::none;
              }())) {
            m_apartment->getBusInterface()->getActionRequestInterface()->stopStopTimer(
                zone->getGroup(GroupIDGray).get(), area);
          }
        }
      }
    }
    return result;
  }

  std::vector<dsModel::GroupStatus> groupsStatus() const override {
    DS_DEBUG_ENTER();
    std::vector<boost::shared_ptr<Cluster>> clusters = m_apartment->getClusters();
    auto result = std::vector<dsModel::GroupStatus>{};

    for (size_t i = 0; i < clusters.size(); i++) {
      boost::shared_ptr<Cluster> cluster = clusters.at(i);
      if ((cluster->getColor() != GroupIDYellow) && (cluster->getColor() != GroupIDGray)) {
        continue;
      }

      Set devices = cluster->getDevices().getByPresence(true);
      if (devices.size() == 0) {
        continue;
      }

      dsModel::GroupStatus gs{};
      gs.id = intToString(cluster->getID());

      if (cluster->getColor() == GroupIDYellow) {
        dsModel::LightsStatusStatus zoneStateDsm{};
        std::tie(gs.application, zoneStateDsm) = getLightGroupApplicationStatus(devices);
        result.push_back(gs);
        auto lightStatusToDsmApi = [&](const dsModel::LightsStatusStatus& status) {
          if (status == dsModel::LightsStatusStatus::none) {
            return -1;
          } else if (status == dsModel::LightsStatusStatus::off) {
            return 0;
          } else {
            return 1;
          }
        };
        if (m_emptyClickPreventionActive && m_cacheReady && m_apartment->getBusInterface() &&
            m_apartment->getBusInterface()->getActionRequestInterface() &&
            zoneStateDsm != dsModel::LightsStatusStatus::none) {
          m_apartment->getBusInterface()->getActionRequestInterface()->setState(cluster.get(),
                                                                                lightStatusToDsmApi(zoneStateDsm));
        }
      } else if (cluster->getColor() == GroupIDGray) {
        dsModel::ShadesStatusStatus zoneStateDsm{};
        std::tie(gs.application, zoneStateDsm) = getShadeGroupApplicationStatus(devices);
        result.push_back(gs);
        auto shadeStatusToDsmApi = [&](const dsModel::ShadesStatusStatus& status) {
          if (status == dsModel::ShadesStatusStatus::none) {
            return -1;
          } else if (status == dsModel::ShadesStatusStatus::closed) {
            return 0;
          } else {
            return 1;
          }
        };
        auto shadeMovingStatus = [&](const dsModel::ShadesStatusStatus& status) {
          if (status == dsModel::ShadesStatusStatus::moving || status == dsModel::ShadesStatusStatus::none) {
            return true;
          } else {
            return false;
          }
        };
        if (m_emptyClickPreventionActive && m_cacheReady && m_apartment->getBusInterface() &&
            m_apartment->getBusInterface()->getActionRequestInterface() &&
            zoneStateDsm != dsModel::ShadesStatusStatus::none) {
          m_apartment->getBusInterface()->getActionRequestInterface()->setState(cluster.get(),
                                                                                shadeStatusToDsmApi(zoneStateDsm));
          if (!shadeMovingStatus(zoneStateDsm)) {
            m_apartment->getBusInterface()->getActionRequestInterface()->stopStopTimer(cluster.get());
          }
        }
      }
    }
    DS_DEBUG_LEAVE();
    return result;
  }

  void write(const Dsuid& dsuid, std::string parameter, std::string value) const override {
    m_pool.post([this, dsuid, parameter, value]() { m_hal->write(dsuid, parameter, value); });
  }

  void write(const Dsuid& dsuid, std::string parameter, double value) const override {
    m_pool.post([this, dsuid, parameter, value]() { m_hal->write(dsuid, parameter, value); });
  }

  void write(const Dsuid& dsuid, std::string parameter, bool value) const override {
    m_pool.post([this, dsuid, parameter, value]() { m_hal->write(dsuid, parameter, value); });
  }

  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const override {
    m_pool.post([this, dsuid, bank, offset, value]() { m_hal->updateDeviceConfig(dsuid, bank, offset, value); });
  }

  std::optional<std::vector<double>> getDeviceThresholds(const Dsuid& dsuid) const override {
    std::optional<std::vector<double>> levelThresholds;
    try {
      if (m_hal->deviceCacheable(dsuid)) {
        levelThresholds = m_hal->getDeviceThresholds(dsuid);
      }
    } catch (const ds::Exception& e) {
      DS_NOTICE("Could not scan DSUID: " + dsuid2str(dsuid), e.what());
    }
    return levelThresholds;
  }

  void completeProcessing() const override {
    auto task = std::packaged_task<void(void)>([] {});
    m_pool.post([&] { task(); });
    task.get_future().get();
  }

  std::vector<std::string> scenarioDeviceList(Group& group, int scene) const override {
    DS_DEBUG_ENTER(group, scene);
    auto devices = group.getDevices();
    auto result = std::vector<std::string>{};
    for (int i = 0; i < devices.length(); i++) {
      auto dsid = devices.get(i).getDSID();
      auto cache = m_deviceCache.tryLookup(dsid);
      if (cache.scene.sceneTable[scene].flags.dontCare.value_or(true) == false) {
        result.push_back(ds::str(dsid));
      }
    }
    std::sort(result.begin(), result.end());
    DS_DEBUG_LEAVE();
    return result;
  }

  ParameterCacheStatus getParameterCacheStatus() const override {
    DS_DEBUG_ENTER();
    ParameterCacheStatus result{};

    auto devices = m_apartment->getDevices().getByPresence(true);
    for (size_t i = 0; i < devices.size(); ++i) {
      try {
        if (m_hal->deviceCacheable(devices[i].getDSID())) {
          result.devices.push_back(m_hal->cacheStatus(devices[i].getDSID()));
        } else {
          DS_DEBUG("Skip non-cacheable", devices[i].getDSID());
        }
      } catch (const ds::Exception& e) {
        DS_DEBUG("Skip", devices[i].getDSID(), e.what());
        continue;
      }
    }
    auto total = 0;
    auto completed = 0;
    result.cache.total = 0;
    result.cache.completed = 0;
    for (auto& d : result.devices) {
      (*result.cache.total)++;
      if (d.completed == d.total) {
        (*result.cache.completed)++;
      }
      total += d.total;
      completed += d.completed;
    }
    result.cache.completion = (total > 0) ? (static_cast<double>(completed) / total) * 100 : 100;
    result.cache.remainingTime = (total - completed) * ReadoutTimePerParameter;

    if (m_apartment->getPropertySystem() &&
        m_apartment->getPropertySystem()->getBoolValue(pp_device_cache_update_status)) {
      result.cache.status = dsModel::ParameterCacheStatus::updating;
    } else {
      if (result.cache.completion < 100) {
        result.cache.status = dsModel::ParameterCacheStatus::incomplete;
      } else {
        result.cache.status = dsModel::ParameterCacheStatus::complete;
      }
    }
    m_cacheReady = (result.cache.status == dsModel::ParameterCacheStatus::complete);

    DS_DEBUG_LEAVE();
    return result;
  }

  void updateParameterCacheStatus() const override {
    getParameterCacheStatus();
  }

private:
  void updateScenarios(const Dsuid& changedDevice, ds::DateTime now) const {
    auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);
    auto device = m_apartment->getDeviceByDSID(changedDevice);

    for (auto& scenario : scenarios.scenarios) {
      if (scenario.second.status == dsModel::ScenarioStatusStatus::active &&
          (scenario.second.zoneId == 0 || (scenario.second.zoneId == device->getZoneID())) &&
          (scenario.second.groupId == 0 || (device->isInGroup(scenario.second.groupId)))) {
        DS_DEBUG("Checking scenario", scenario);
        auto cache = m_deviceCache.tryLookup(changedDevice);
        if (cache.scene.sceneTable[scenario.second.sceneId].flags.dontCare.value_or(true) == false) {
          scenario.second.status = dsModel::ScenarioStatusStatus::inactive;
          scenario.second.lastChanged = now;
          break;
        }
      }
    }
    m_scenarioCache.update(DSUID_NULL, scenarios);
    sendChangeEvent();
  }

  void updateScenarios(const std::vector<Dsuid>& changedDevices, int zoneId, int groupId, ds::DateTime now) const {
    DS_DEBUG_ENTER();
    auto scenarios = m_scenarioCache.tryLookup(DSUID_NULL);

    for (auto& scenario : scenarios.scenarios) {
      if (scenario.second.status == dsModel::ScenarioStatusStatus::active &&
          (scenario.second.zoneId == 0 || zoneId == 0 || (scenario.second.zoneId == zoneId)) &&
          (scenario.second.groupId == 0 || groupId == 0 || (scenario.second.groupId == groupId))) {
        DS_DEBUG("Checking scenario", scenario);
        for (auto& device : changedDevices) {
          auto cache = m_deviceCache.tryLookup(device);
          if (cache.scene.sceneTable[scenario.second.sceneId].flags.dontCare.value_or(true) == false) {
            scenario.second.status = dsModel::ScenarioStatusStatus::inactive;
            scenario.second.lastChanged = now;
            break;
          }
        }
      }
    }
    m_scenarioCache.update(DSUID_NULL, scenarios);

    if (changedDevices.size() > 0) {
      sendChangeEvent();
    }
    DS_DEBUG_LEAVE();
  }

  dsModel::LightsStatusStatus getLightStateFor(dsModel::LightsStatusStatus roomState,
                                               dsModel::OutputStatusStatus deviceState, DeviceChannelValues& channel,
                                               bool localPriority = false) const {
    if (localPriority) {
      return roomState;
    }
    if (deviceState == dsModel::OutputStatusStatus::dimming) {
      return dsModel::LightsStatusStatus::dimming;
    }
    if (deviceState == dsModel::OutputStatusStatus::moving) {
      return dsModel::LightsStatusStatus::dimming;
    }
    if (roomState == dsModel::LightsStatusStatus::dimming) {
      return roomState;
    }
    if (roomState == dsModel::LightsStatusStatus::on) {
      return roomState;
    }
    if (deviceState == dsModel::OutputStatusStatus::ok) {
      if ((!channel.level && channel.value > 0) || (channel.level > 0)) {
        return dsModel::LightsStatusStatus::on;
      } else {
        return dsModel::LightsStatusStatus::off;
      }
    }
    return roomState;
  }

  std::pair<dsModel::LightsStatus, dsModel::LightsStatusStatus> getLightApplicationStatus(Set devices) const {
    dsModel::LightsStatus aps{};

    /* If at least one device is "dimming", then the whole zone is
     * dimming, otherwise, if the device returns "ok", then check
     * its brightness channel output. If output of at least one
     * device is >0 then the zone is "on", else it is "off".  */
    dsModel::LightsStatusStatus zoneState{};
    dsModel::LightsStatusStatus zoneStateDsm{};
    dsModel::LightsStatusStatus area1{};
    dsModel::LightsStatusStatus area2{};
    dsModel::LightsStatusStatus area3{};
    dsModel::LightsStatusStatus area4{};

    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::exception& ex) {
        DS_WARNING(std::string("Ignoring state for device ") + dsuid2str(devRef.getDSID()) + " " + ex.what());
        continue;
      } catch (...) {
        /* I was not able to figure out which exception was thrown, perhaps
         * something custom originating from the testing framework? */
        DS_WARNING(std::string("Ignoring state for device ") + dsuid2str(devRef.getDSID()) + " unknown exception.");

        continue;
      }

      std::vector<int> areas = m_hal->areaMember(devRef.getDSID());
      for (size_t a = 0; a < areas.size(); a++) {
        switch (areas.at(a)) {
          case SceneA11: // area 1
            area1 = getLightStateFor(area1, state.status, state.outputs[ChannelType::brightness]);
            break;
          case SceneA21: // area 2
            area2 = getLightStateFor(area2, state.status, state.outputs[ChannelType::brightness]);
            break;
          case SceneA31: // area 3
            area3 = getLightStateFor(area3, state.status, state.outputs[ChannelType::brightness]);
            break;
          case SceneA41: // area 4
            area4 = getLightStateFor(area4, state.status, state.outputs[ChannelType::brightness]);
            break;
          default: break;
        }
      }

      zoneState = getLightStateFor(zoneState, state.status, state.outputs[ChannelType::brightness]);
      zoneStateDsm =
          getLightStateFor(zoneStateDsm, state.status, state.outputs[ChannelType::brightness], state.localPriority);
    }

    aps.status = zoneState;
    aps.area1 = area1;
    aps.area2 = area2;
    aps.area3 = area3;
    aps.area4 = area4;

    return {aps, zoneStateDsm};
  }

  dsModel::ShadesStatusStatus getShadeStateFor(dsModel::ShadesStatusStatus roomStatus,
                                               dsModel::OutputStatusStatus deviceStatus, double value,
                                               bool localPriority = false) const {
    if (roomStatus == dsModel::ShadesStatusStatus::moving) {
      return dsModel::ShadesStatusStatus::moving;
    }
    if (deviceStatus == dsModel::OutputStatusStatus::moving || deviceStatus == dsModel::OutputStatusStatus::dimming) {
      return dsModel::ShadesStatusStatus::moving;
    }
    if (roomStatus == dsModel::ShadesStatusStatus::mixed) {
      return dsModel::ShadesStatusStatus::mixed;
    }
    if (localPriority) {
      return roomStatus;
    }
    if (deviceStatus == dsModel::OutputStatusStatus::ok) {
      value = round(value);
      if (value == 0) {
        if (roomStatus == dsModel::ShadesStatusStatus::open) {
          return dsModel::ShadesStatusStatus::mixed;
        } else {
          return dsModel::ShadesStatusStatus::closed;
        }
      } else if (value == 100) {
        if (roomStatus == dsModel::ShadesStatusStatus::closed) {
          return dsModel::ShadesStatusStatus::mixed;
        } else {
          return dsModel::ShadesStatusStatus::open;
        }
      } else {
        return dsModel::ShadesStatusStatus::mixed;
      }
    }

    return roomStatus;
  }

  std::pair<dsModel::ShadesStatus, dsModel::ShadesStatusStatus> getShadeApplicationStatus(Set devices) const {
    dsModel::ShadesStatus aps{};

    /* When all are closed, the state is closed, when all are open, the
     * the state is open, if some are open and some closed, then the
     * state is mixed, but if at least one is moving, then the state is
     * moving. */
    dsModel::ShadesStatusStatus zoneState{};
    dsModel::ShadesStatusStatus zoneStateDsm{};
    dsModel::ShadesStatusStatus area1{};
    dsModel::ShadesStatusStatus area2{};
    dsModel::ShadesStatusStatus area3{};
    dsModel::ShadesStatusStatus area4{};

    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::runtime_error& ex) {
        DS_WARNING(std::string("Ignoring state for device ") + dsuid2str(devRef.getDSID()) + " " + ex.what());
        continue;
      }

      std::vector<int> areas = m_hal->areaMember(devRef.getDSID());
      for (size_t a = 0; a < areas.size(); a++) {
        switch (areas.at(a)) {
          case SceneA11: // area 1
            area1 = getShadeStateFor(area1, state.status, state.outputs[ChannelType::shadePositionOutside].value);
            break;
          case SceneA21: // area 2
            area2 = getShadeStateFor(area2, state.status, state.outputs[ChannelType::shadePositionOutside].value);
            break;
          case SceneA31: // area 3
            area3 = getShadeStateFor(area3, state.status, state.outputs[ChannelType::shadePositionOutside].value);
            break;
          case SceneA41: // area 4
            area4 = getShadeStateFor(area4, state.status, state.outputs[ChannelType::shadePositionOutside].value);
            break;
          default: break;
        }
      }

      zoneState = getShadeStateFor(zoneState, state.status, state.outputs[ChannelType::shadePositionOutside].value);
      zoneStateDsm = getShadeStateFor(zoneStateDsm, state.status,
                                      state.outputs[ChannelType::shadePositionOutside].value, state.localPriority);
    }

    aps.status = zoneState;
    aps.area1 = area1;
    aps.area2 = area2;
    aps.area3 = area3;
    aps.area4 = area4;

    return {aps, zoneStateDsm};
  }

  std::pair<dsModel::ZoneApplicationsStatus, dsModel::ShadesStatusStatus> getShadeGroupApplicationStatus(
      Set devices) const {
    dsModel::ZoneApplicationsStatus aps{};

    dsModel::ShadesStatusStatus zoneState{};
    dsModel::ShadesStatusStatus zoneStateDsm{};
    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::runtime_error& ex) {
        DS_WARNING(std::string("Ignoring state for device ") + dsuid2str(devRef.getDSID()) + " " + ex.what());
        continue;
      }

      zoneState = getShadeStateFor(zoneState, state.status, state.outputs[ChannelType::shadePositionOutside].value);
      zoneStateDsm = getShadeStateFor(zoneStateDsm, state.status,
                                      state.outputs[ChannelType::shadePositionOutside].value, state.localPriority);
    }

    aps.shades.status = zoneState;
    return {aps, zoneStateDsm};
  }

  std::pair<dsModel::ZoneApplicationsStatus, dsModel::LightsStatusStatus> getLightGroupApplicationStatus(
      Set devices) const {
    dsModel::ZoneApplicationsStatus aps{};

    dsModel::LightsStatusStatus zoneState{};
    dsModel::LightsStatusStatus zoneStateDsm{};
    for (size_t d = 0; d < devices.size(); d++) {
      const DeviceReference& devRef = devices.get(d);
      DeviceState state;
      try {
        state = m_hal->getState(devRef.getDSID());
      } catch (std::runtime_error& ex) {
        DS_WARNING(std::string("Ignoring state for device ") + dsuid2str(devRef.getDSID()) + " " + ex.what());
        continue;
      }
      zoneState = getLightStateFor(zoneState, state.status, state.outputs[ChannelType::brightness]);
      zoneStateDsm =
          getLightStateFor(zoneStateDsm, state.status, state.outputs[ChannelType::brightness], state.localPriority);
    }

    aps.lights.status = zoneState;

    return {aps, zoneStateDsm};
  }

  void sendChangeEvent() const {
    m_apartment->pushEvent(boost::make_shared<Event>(EventName::ApartmentProxyStateChanged));
  }

private:
  HardwareAbstraction* m_hal;
  mutable CacheDb<ScenarioCache> m_scenarioCache;
  CacheDb<DeviceCache>& m_deviceCache;
  Apartment* m_apartment;
  mutable ThreadPool m_pool;
  mutable bool m_cacheReady;
  bool m_emptyClickPreventionActive;
}; // namespace dss

std::unique_ptr<ApartmentProxy> makeApartmentProxy(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db,
                                                   EventInterpreter* pInterpreter, const ds::Conf& conf) {
  return std::make_unique<ApartmentProxyImpl>(apt, hal, db, pInterpreter, conf);
}

class NullApartmentProxy : public ApartmentProxy {
public:
  void callScene(const Dsuid& device, int sceneId, bool forced, ds::DateTime time) const override {}
  void callScene(int zoneId, int groupId, int sceneId, bool forced, ds::DateTime time) const override {}
  void saveScene(const Dsuid& device, int sceneId) const override {}
  void callSceneMin(const Dsuid& device, int sceneId, ds::DateTime time) const override {}
  void callSceneMin(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {}
  void undoScene(const Dsuid& device, int sceneId, ds::DateTime time) const override {}
  void undoScene(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {}
  void setOutputValue(const Dsuid& device, uint8_t value, ds::DateTime time) const override {}
  void setOutputValue(int zoneId, int groupId, uint8_t value, ds::DateTime time) const override {}
  void setOutputValue(const Dsuid& device, ChannelType channel, double value, ds::DateTime time) const override {}
  void setSceneLocalPrio(int zoneId, int groupId, int sceneId) const override {}
  void setOperationLock(const Dsuid& device, bool lock) const override {}
  void stopDimming(int zoneId, int groupId, int sceneId, ds::DateTime time) const override {}
  void stopDimming(const Dsuid& device) const override {}
  void completeMoving(const Dsuid& device) const override {}

  dsModel::ScenarioStatus zoneScenarioStatus(const dsModel::ScenarioId&) const override { return {}; }
  dsModel::DSDeviceStatus dsDeviceStatus(const dsModel::DeviceId& id) const override { return {}; }
  dsModel::LightsStatus lightsStatus(const dsModel::ZoneId&) const override { return {}; }
  dsModel::ShadesStatus shadesStatus(const dsModel::ZoneId&) const override { return {}; }
  std::vector<dsModel::GroupStatus> groupsStatus() const override { return {}; }
  std::vector<std::string> scenarioDeviceList(Group& group, int scene) const override { return {}; }

  void write(const Dsuid& dsuid, std::string parameter, std::string value) const override {}
  void write(const Dsuid& dsuid, std::string parameter, double value) const override {}
  void write(const Dsuid& dsuid, std::string parameter, bool value) const override {}

  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const override {}
  std::optional<std::vector<double>> getDeviceThresholds(const Dsuid& dsuid) const override { return {}; }

  void completeProcessing() const override {}
  ParameterCacheStatus getParameterCacheStatus() const override { return {}; }
  void updateParameterCacheStatus() const override {}
};

std::unique_ptr<ApartmentProxy> makeNullApartmentProxy() {
  return std::make_unique<NullApartmentProxy>();
}

} /* namespace dss */
