/*
 *  Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/smart_ptr/make_shared.hpp>
#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/apartment-proxy.h"
#include "ds/catch/catch.h"
#include "ds/conf-mock.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "hal/hardware-abstraction-mock.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/modelconst.h"
#include "model/modelmaintenance-mock.h"
#include "model/zone.h"
#include "propertysystem_common_paths.h"
#include "util.h"

static ds::LogChannel dsLogChannel("apartmentStatus", ds::LogSeverity::DEBUG);
static const char* TAGS = "[apartmentStatus]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid l1{"0000000000000000000000000000000001"};
static const Dsuid l2{"0000000000000000000000000000000002"};
static const Dsuid l3{"0000000000000000000000000000000003"};
static const Dsuid l4{"0000000000000000000000000000000004"};
static const Dsuid l5{"0000000000000000000000000000000005"};
static const Dsuid l6{"0000000000000000000000000000000006"};
static const Dsuid l7{"0000000000000000000000000000000007"};

static const Dsuid s1{"0000000000000000000000000000000008"};
static const Dsuid s2{"0000000000000000000000000000000009"};
static const Dsuid s3{"000000000000000000000000000000000a"};
static const Dsuid s4{"000000000000000000000000000000000b"};
static const Dsuid s5{"000000000000000000000000000000000c"};
static const Dsuid s6{"000000000000000000000000000000000d"};
static const Dsuid s7{"000000000000000000000000000000000e"};
static const Dsuid s8{"000000000000000000000000000000000f"};

void createDevice(Apartment& apt, const Dsuid dsid, bool yellow = true) {
  DeviceSpec spec;
  if (yellow) {
    spec.FunctionID = 0x1010;
    spec.ProductID = (1 << 10) | 200;
  } else {
    spec.FunctionID = 0x2000;
    spec.ProductID = (3 << 10) | 200;
  }
  boost::shared_ptr<Device> dev = apt.allocateDevice(dsid);
  dev->setPartiallyFromSpec(spec);
  dev->setIsPresent(true);
  dev->setIsValid(true);
  dev->setIsConnected(true);
  if (yellow) {
    dev->addToGroup(GroupIDYellow);
  } else {
    dev->addToGroup(GroupIDGray);
  }
}

void addDeviceToZone(Apartment& apt, Zone& zone, const Dsuid dsid) {
  DeviceReference lightref(dsid, &apt);
  zone.addDevice(lightref);
}

SCENARIO("Apartment status light calculation", TAGS) {
  Apartment apt(nullptr);
  PropertySystem propSys;
  apt.setPropertySystem(&propSys);
  ds::ConfMock conf;

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  auto& zone1 = *apt.allocateZone(1);
  auto& zone2 = *apt.allocateZone(2);
  apt.getZone(0)->addGroup(boost::make_shared<Cluster>(16, apt));
  apt.getCluster(16)->setApplicationType(ApplicationType::Lights);
  apt.getZone(0)->addGroup(boost::make_shared<Cluster>(17, apt));
  apt.getCluster(17)->setApplicationType(ApplicationType::Lights);

  createDevice(apt, l1);
  createDevice(apt, l2);
  createDevice(apt, l3);
  createDevice(apt, l4);
  createDevice(apt, l5);
  createDevice(apt, l6);
  createDevice(apt, l7);

  addDeviceToZone(apt, zone1, l1);
  addDeviceToZone(apt, zone1, l2);
  addDeviceToZone(apt, zone1, l3);
  addDeviceToZone(apt, zone1, l4);
  addDeviceToZone(apt, zone2, l5);
  addDeviceToZone(apt, zone2, l6);
  addDeviceToZone(apt, zone2, l7);

  apt.getDeviceByDSID(l1)->addToGroup(16);
  apt.getDeviceByDSID(l5)->addToGroup(16);
  apt.getDeviceByDSID(l5)->addToGroup(17);
  apt.getDeviceByDSID(l6)->addToGroup(17);
  apt.getDeviceByDSID(l7)->addToGroup(17);

  ds::sqlite3::Database db(":memory:");
  MockHal hal(&apt, &db);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  ALLOW_CALL(hal, areaMember(l1)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(l2)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(l3)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(l4)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(l5)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(l6)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(l7)).RETURN(std::vector<int>{});

  WHEN("l1 is on") {
    THEN("zone 1, area 1 and cluster 1 must be on") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      dss::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 30.0, .targetValue = 30.0, .level = 0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds2);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDYellow).get(), 1, 1, 0, 255, 255));

      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG();

      REQUIRE(lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDYellow).get(), 0, 255, 255, 255, 255));
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(16).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(17).get(), 0));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
    THEN("but locally priorized") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}},
                           .localPriority = true};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      dss::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 30.0, .targetValue = 30.0, .level = 0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds2);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDYellow).get(), 0, 1, 0, 255, 255));

      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG();

      REQUIRE(lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDYellow).get(), 0, 255, 255, 255, 255));
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(16).get(), 0));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(17).get(), 0));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }

  WHEN("l1 is dimming") {
    THEN("zone 1, area 1 and cluster 1 must be dimming") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::dimming,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds0);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDYellow).get(), 1, 1, 0, 255, 255));
      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG(lights);

      REQUIRE(lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDYellow).get(), 0, 255, 255, 255, 255));
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(16).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(17).get(), 0));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }

  WHEN("l1 is moving") {
    THEN("zone 1, area 1 and cluster 1 must be dimming") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds0);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDYellow).get(), 1, 1, 0, 255, 255));
      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG(lights);

      REQUIRE(lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDYellow).get(), 0, 255, 255, 255, 255));
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(16).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(17).get(), 0));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }
}

SCENARIO("Apartment status shade calculation", TAGS) {
  Apartment apt(nullptr);
  PropertySystem propSys;
  apt.setPropertySystem(&propSys);
  ds::ConfMock conf;

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  auto& zone1 = *apt.allocateZone(1);
  auto& zone2 = *apt.allocateZone(2);
  auto& zone3 = *apt.allocateZone(3);

  apt.getZone(0)->addGroup(boost::make_shared<Cluster>(18, apt));
  apt.getCluster(18)->setApplicationType(ApplicationType::Blinds);
  apt.getZone(0)->addGroup(boost::make_shared<Cluster>(19, apt));
  apt.getCluster(19)->setApplicationType(ApplicationType::Blinds);

  createDevice(apt, s1, false);
  createDevice(apt, s2, false);
  createDevice(apt, s3, false);
  createDevice(apt, s4, false);
  createDevice(apt, s5, false);
  createDevice(apt, s6, false);
  createDevice(apt, s7, false);
  createDevice(apt, s8, false);

  addDeviceToZone(apt, zone1, s1);
  addDeviceToZone(apt, zone1, s2);
  addDeviceToZone(apt, zone1, s3);
  addDeviceToZone(apt, zone1, s4);
  addDeviceToZone(apt, zone2, s5);
  addDeviceToZone(apt, zone2, s6);
  addDeviceToZone(apt, zone2, s7);
  addDeviceToZone(apt, zone3, s8);

  apt.getDeviceByDSID(s1)->addToGroup(18);
  apt.getDeviceByDSID(s5)->addToGroup(18);
  apt.getDeviceByDSID(s5)->addToGroup(19);
  apt.getDeviceByDSID(s6)->addToGroup(19);
  apt.getDeviceByDSID(s7)->addToGroup(19);

  ds::sqlite3::Database db(":memory:");
  MockHal hal(&apt, &db);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  ALLOW_CALL(hal, areaMember(s1)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(s2)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(s3)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(s4)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(s5)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s6)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s7)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s8)).RETURN(std::vector<int>{});

  WHEN("shade status is checked") {
    THEN("it must be correct") {
      ds::DateTime timeNow;
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}}};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      dss::DeviceState ds2{.lastChanged = timeNow,
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s8)).TIMES(2).RETURN(ds2);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 1, 1, 0, 255, 255));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 0));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 1));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 2));
      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone2.getGroup(GroupIDGray).get(), 0));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(dsModel::ZoneId{"3"}));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(18).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 1));
      REQUIRE_CALL(busInterface.action, stopStopTimer(apt.getCluster(18).get()));
      REQUIRE_CALL(busInterface.action, stopStopTimer(apt.getCluster(19).get()));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::open);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::mixed);

      dsModel::DSDeviceStatus status;
      REQUIRE_NOTHROW(status = proxy->dsDeviceStatus(dsModel::DeviceId{ds::str(s8)}));
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].initialValue == 50);
      REQUIRE(!status.attributes.functionBlocks[0].outputs[0].value);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].startedAt == timeNow);
      
      REQUIRE_NOTHROW(status = proxy->dsDeviceStatus(dsModel::DeviceId{ds::str(s1)}));
      REQUIRE(!status.attributes.functionBlocks[0].outputs[0].initialValue);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].value == 100);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].startedAt == ds::DateTime());
    }
    THEN("it must be correct also with local prio") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}},
                           .localPriority = true};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      dss::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s8)).TIMES(1).RETURN(ds2);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 0, 1, 0, 255, 255));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 0));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 1));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 2));
      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 0, 255, 255, 255, 255));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone2.getGroup(GroupIDGray).get(), 0));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(dsModel::ZoneId{"3"}));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 0));
      REQUIRE_CALL(busInterface.action, stopStopTimer(apt.getCluster(19).get()));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::open);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::mixed);
    }
  }

  WHEN("shade status with dimming") {
    THEN("it must be correct") {
      dss::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::dimming,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}}};
      dss::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      dss::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s8)).TIMES(1).RETURN(ds2);

      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 1, 1, 0, 255, 255));
      REQUIRE_CALL(busInterface.action, stopStopTimer(zone1.getGroup(GroupIDGray).get(), 2));
      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone3.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(18).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 1));
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::moving);
    }
  }
}

} // namespace
