/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include <ds/catch/catch.h>
#include <ds/log.h>
#include "cache-db.h"
#include "ds/reflect.h"

using namespace dss;

static ds::LogChannel dsLogChannel("cacheDbTest");
static const char* TAGS = "[dss][cacheDb]";

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};

namespace {
struct TestEntry {
  std::optional<int> a;
  std::optional<std::string> b;
  bool operator==(const TestEntry& x) {
    return (a == x.a) && (b == x.b);
  }
};
struct TestEntry2 {
  int a;
  bool operator==(const TestEntry2& x) {
    return a == x.a;
  }
};
} // namespace
DS_REFLECT_ANONYMOUS_STRUCT(TestEntry, (a)(b));
DS_REFLECT_ANONYMOUS_STRUCT(TestEntry2, (a));

namespace {

SCENARIO("Normal operation", TAGS) {
  ds::sqlite3::Database d("cache.db");
  CacheDb<TestEntry> db(&d);
  TestEntry X{42, std::string("device-caching ?/)")};
  TestEntry Empty{};
  GIVEN("an empty cache") {
    REQUIRE_NOTHROW(db.empty());
    WHEN("an entry is queried") {
      THEN("the cache returns an empty entry") {
        TestEntry Y = db.tryLookup(dsuid1);
        REQUIRE(Y == Empty);
      }
    }
    WHEN("a new entry is added") {
      REQUIRE_NOTHROW(db.update(dsuid1, X));
      THEN("it can be read back correctly") {
        TestEntry Y = db.tryLookup(dsuid1);
        REQUIRE(Y == X);
      }
    }
    WHEN("an existing entry is updated") {
      REQUIRE_NOTHROW(db.update(dsuid1, X));
      X.a = 43;
      REQUIRE_NOTHROW(db.update(dsuid1, X));
      THEN("it can be read back correctly") {
        TestEntry Y = db.tryLookup(dsuid1);
        REQUIRE(Y == X);
      }
    }
    WHEN("an existing entry is deleted") {
      REQUIRE_NOTHROW(db.update(dsuid1, X));
      REQUIRE_NOTHROW(db.remove(dsuid1));
      THEN("the cache returns an empty entry") {
        TestEntry Y = db.tryLookup(dsuid1);
        REQUIRE(Y == Empty);
      }
    }
    WHEN("a non-existing entry is deleted") {
      THEN("the result is still true") { REQUIRE_NOTHROW(db.remove(dsuid1)); }
    }
  }
  GIVEN("undefined entries") {
    TestEntry Y{{}, std::string("q29474tzq93zylusfh$üö$ü")};
    WHEN("added to the cache") {
      REQUIRE_NOTHROW(db.update(dsuid2, Y));
      THEN("the value stays not-initialized") {
        TestEntry y = db.tryLookup(dsuid2);
        REQUIRE(!y.a);
        REQUIRE(y.b);
      }
    }
  }
  GIVEN("a cache with 3 entries") {
    TestEntry Y{{}, std::string("q29474tzq93zylusfh$üö$ü")};
    TestEntry Z{81579785, {}};
    REQUIRE_NOTHROW(db.update(dsuid1, X));
    REQUIRE_NOTHROW(db.update(dsuid2, Y));
    REQUIRE_NOTHROW(db.update(dsuid3, Z));
    THEN("each entry can be queried") {
      REQUIRE(db.tryLookup(dsuid1) == X);
      REQUIRE(db.tryLookup(dsuid2) == Y);
      REQUIRE(db.tryLookup(dsuid3) == Z);
    }
    WHEN("one entry is removed") {
      REQUIRE_NOTHROW(db.remove(dsuid2));
      THEN("the others are still available") {
        REQUIRE(db.tryLookup(dsuid1) == X);
        REQUIRE(db.tryLookup(dsuid3) == Z);
      }
    }
  }
}

SCENARIO("Multiple cache instances", TAGS) {
  ds::sqlite3::Database d("cache.db");
  GIVEN("There are 2 cache instances for different types") {
    CacheDb<TestEntry> db1(&d);
    db1.empty();
    CacheDb<TestEntry2> db2(&d);
    db2.empty();
    TestEntry X1{42, std::string("device-caching ?/)")};
    TestEntry2 X2{99};
    WHEN("adding an entry to one instance") {
      REQUIRE_NOTHROW(db1.update(dsuid1, X1));
      THEN("there is no entry in the other instance") {
        REQUIRE(db2.tryLookup(dsuid1) == TestEntry2{});
      }
    }
    WHEN("adding an entry with the same key to both instances") {
      REQUIRE_NOTHROW(db1.update(dsuid1, X1));
      REQUIRE_NOTHROW(db2.update(dsuid1, X2));
      THEN("both instances are independent") {
        REQUIRE(db1.tryLookup(dsuid1) == X1);
        REQUIRE(db2.tryLookup(dsuid1) == X2);
      }
    }
  }
}

} // namespace
