
/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "cache-updater.h"
#include <stdio.h>
#include <regex>
#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "device-cache.h"
#include "ds-api/route-segments.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "hal/hardware-abstraction-mock.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/modelmaintenance-mock.h"
#include "propertysystem_common_paths.h"

static ds::LogChannel dsLogChannel("cacheUpdaterTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][cacheUpdater]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;
namespace {

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};

SCENARIO("A device gets added to cache", TAGS) {
  Apartment apt(nullptr);
  ds::sqlite3::Database d(":memory:");
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;
  boost::shared_ptr<PropertySystem> propSystem(new PropertySystem);

  EventInterpreter interpreter(NULL);
  boost::shared_ptr<EventQueue> queue(new EventQueue(&interpreter));
  boost::shared_ptr<EventRunner> runner(new EventRunner(&interpreter));

  interpreter.initialize();
  interpreter.setEventQueue(queue.get());
  interpreter.setEventRunner(runner.get());
  queue->setEventRunner(runner.get());
  runner->setEventQueue(queue.get());

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  MockHal hal(&apt, &d);
  auto& db = hal.getCache();
  CacheUpdatePlugin interface(&apt, &hal, interpreter, propSystem, queue, runner);

  GIVEN("A new device with no scene cache") {
    DeviceSpec spec;
    spec.FunctionID = 0x1010;

    boost::shared_ptr<Device> dev1 = apt.allocateDevice(dsuid1);
    dev1->setPartiallyFromSpec(spec);
    dev1->setIsPresent(true);
    dev1->setIsConnected(true);

    DeviceCache empty;
    db.update(dsuid1, empty);

    WHEN("We start the periodic updater") {
      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/brightness)")))
          .TIMES(69)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            std::regex pathRegex(R"(([0-9]+))");

            auto words_begin = std::sregex_iterator(_2.begin(), _2.end(), pathRegex);
            auto words_end = std::sregex_iterator();

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].channels[ChannelType::brightness].value = 2.1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/brightness/dontCare)")))
          .TIMES(69)
          .RETURN(true)
          .LR_SIDE_EFFECT({
            std::regex pathRegex(R"(([0-9]+))");

            auto words_begin = std::sregex_iterator(_2.begin(), _2.end(), pathRegex);
            auto words_end = std::sregex_iterator();

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].channels[ChannelType::brightness].dontCare =
                true;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/dontCare)")))
          .TIMES(69)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            std::regex pathRegex(R"(([0-9]+))");

            auto words_begin = std::sregex_iterator(_2.begin(), _2.end(), pathRegex);
            auto words_end = std::sregex_iterator();

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.dontCare = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.ignoreLocalPrio = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.specialHandling = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.blink = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.ledConfig = LedConfig::ledCon0;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.dimTime = DimTime::dimTime2;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readString(dsuid1, trompeloeil::re(R"(/property/outputMode)"))).RETURN("1").LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::outputMode] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/switchedThreshold)")))
          .RETURN(1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.property[DeviceProperty::switchedThreshold] = 1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepUp] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepDown)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepDown] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/minDim)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::minDim] = 1;
        db.update(dsuid1, devCache);
      });

      boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
      boost::shared_ptr<Event> pResumeEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      boost::shared_ptr<SubscriptionOptions> opts = boost::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEvent);
      auto subscription = boost::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);

      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all scences fetched from device") { REQUIRE(!interface.isUpdateRunning()); }
    }
  }

  GIVEN("A multi channel light device with no scene cache") {
    DeviceSpec spec;
    spec.FunctionID = 0x1010;

    boost::shared_ptr<Device> dev1 = apt.allocateDevice(dsuid1);
    dev1->setPartiallyFromSpec(spec);
    dev1->setIsPresent(true);
    dev1->setIsConnected(true);
    dev1->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation, ChannelType::colortemp,
                             ChannelType::x, ChannelType::y});

    DeviceCache empty;
    db.update(dsuid1, empty);

    WHEN("We start the periodic updater") {
      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/[a-z]*)")))
          .TIMES(6 * 69)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channel}", _2);

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[strToInt(route.args["sceneNr"])]
                .channels[ds::toEnum<ChannelType>(route.args["channel"])]
                .value = 2.1;
            db.update(dsuid1, devCache);
          });
      REQUIRE_CALL(hal, readBool(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/[a-z]*/dontCare)")))
          .TIMES(6 * 69)
          .RETURN(true)
          .LR_SIDE_EFFECT({
            auto&& route = dsApi::matchRoute("/scene/{sceneNr}/{channel}/dontCare", _2);

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[strToInt(route.args["sceneNr"])]
                .channels[ds::toEnum<ChannelType>(route.args["channel"])]
                .dontCare = true;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid1, trompeloeil::re(R"(/scene/[0-9]*/dontCare)")))
          .TIMES(69)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            std::regex pathRegex(R"(([0-9]+))");

            auto words_begin = std::sregex_iterator(_2.begin(), _2.end(), pathRegex);
            auto words_end = std::sregex_iterator();

            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.dontCare = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.ignoreLocalPrio = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.specialHandling = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.blink = false;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.ledConfig = LedConfig::ledCon0;
            devCache.scene.sceneTable[std::stoi((*words_begin).str())].flags.dimTime = DimTime::dimTime2;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readString(dsuid1, trompeloeil::re(R"(/property/outputMode)"))).RETURN("1").LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::outputMode] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/switchedThreshold)")))
          .RETURN(1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.property[DeviceProperty::switchedThreshold] = 1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime0Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime0Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime1Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime1Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Up)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Up] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimTime2Down)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimTime2Down] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepUp] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/dimStepDown)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::dimStepDown] = 1;
        db.update(dsuid1, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/property/minDim)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid1);
        devCache.property[DeviceProperty::minDim] = 1;
        db.update(dsuid1, devCache);
      });

      boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
      boost::shared_ptr<Event> pResumeEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      boost::shared_ptr<SubscriptionOptions> opts = boost::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEvent);
      auto subscription = boost::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);

      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all scences fetched from device") { REQUIRE(!interface.isUpdateRunning()); }
    }
  }

  GIVEN("An existing device in cache") {
    DeviceSpec spec;

    boost::shared_ptr<Device> dev2 = apt.allocateDevice(dsuid2);
    spec.FunctionID = 0x2000;
    spec.ProductID = 0xCC8;
    dev2->setPartiallyFromSpec(spec);
    dev2->setIsPresent(true);
    dev2->setIsConnected(true);

    hal.populateCache(dsuid2);
    auto mostlyThere = db.tryLookup(dsuid2);

    mostlyThere.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));
    mostlyThere.property[DeviceProperty::dimStepLen] = 2;

    SceneEntry entry;

    entry.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    entry.flags.dontCare = true;
    entry.flags.ignoreLocalPrio = true;
    entry.flags.specialHandling = true;
    entry.flags.blink = true;
    entry.flags.ledConfig = LedConfig::ledCon0;
    entry.flags.dimTime = DimTime::dimTime0;

    entry.action = "";

    entry.channels[ChannelType::shadePositionOutside].value = 50.0;
    entry.channels[ChannelType::shadePositionOutside].dontCare = false;

    for (auto& scene : mostlyThere.scene.sceneTable) {
      scene.second = entry;
    }

    // Now delete some part and check if they get fetched

    mostlyThere.scene.sceneTable[2].flags.specialHandling = {};
    mostlyThere.scene.sceneTable[2].flags.blink = {};
    mostlyThere.scene.sceneTable[56].flags.ignoreLocalPrio = {};
    mostlyThere.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = {};

    mostlyThere.scene.sceneTable[6].channels.clear();

    mostlyThere.property[DeviceProperty::posTimeDown] = 10;

    db.update(dsuid2, mostlyThere);

    WHEN("Periodic cache update runs") {
      REQUIRE_CALL(hal, readBool(dsuid2, trompeloeil::re(R"(/scene/2/dontCare)")))
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[2].flags.blink = false;
            devCache.scene.sceneTable[2].flags.specialHandling = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/scene/5/shadePositionOutside)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = 2.1;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid2, trompeloeil::re(R"(/scene/56/dontCare)")))
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[56].flags.ignoreLocalPrio = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepNum)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepNum] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepPause)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepPause] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/posTimeUp)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::posTimeUp] = 1;
        db.update(dsuid2, devCache);
      });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/scene/6/shadePositionOutside)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[6].channels[ChannelType::shadePositionOutside].value = 2.1;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid2, trompeloeil::re(R"(/scene/6/shadePositionOutside/dontCare)")))
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[6].channels[ChannelType::shadePositionOutside].dontCare = false;
            db.update(dsuid2, devCache);
          });

      boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast);
      boost::shared_ptr<Event> pResumeEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      boost::shared_ptr<SubscriptionOptions> opts = boost::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEventFast);
      auto subscription = boost::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);
      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      THEN("The device should have all missing scences fetched from device") {}
    }
  }

  GIVEN("3 Devices in apartmentwith 2 existing devices in cache (1 device not supported by HAL)") {
    DeviceSpec spec;

    boost::shared_ptr<Device> devShade = apt.allocateDevice(dsuid2);
    spec.FunctionID = 0x2000;
    spec.ProductID = 0xCC8;
    devShade->setPartiallyFromSpec(spec);
    devShade->setIsPresent(true);
    devShade->setIsConnected(true);

    hal.populateCache(dsuid2);
    auto completeShade = db.tryLookup(dsuid2);

    completeShade.property[DeviceProperty::dimStepNum] = 4;
    completeShade.property[DeviceProperty::dimStepLen] = 3;
    completeShade.property[DeviceProperty::dimStepPause] = 4;
    completeShade.property[DeviceProperty::posTimeUp] = 4;
    completeShade.property[DeviceProperty::posTimeDown] = 4;

    completeShade.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    SceneEntry entry;

    entry.lastChanged = ds::DateTime(std::chrono::milliseconds(1505731033777));

    entry.flags.dontCare = true;
    entry.flags.ignoreLocalPrio = true;
    entry.flags.specialHandling = true;
    entry.flags.blink = true;
    entry.flags.ledConfig = LedConfig::ledCon0;
    entry.flags.dimTime = DimTime::dimTime0;

    entry.action = "";

    entry.channels[ChannelType::shadePositionOutside].value = 50.0;
    entry.channels[ChannelType::shadePositionOutside].dontCare = false;

    for (auto& scene : completeShade.scene.sceneTable) {
      scene.second = entry;
    }
    db.update(dsuid2, completeShade);

    boost::shared_ptr<Device> devLight = apt.allocateDevice(dsuid1);
    spec.FunctionID = 0x1010;

    devLight->setPartiallyFromSpec(spec);
    devLight->setIsPresent(true);
    devLight->setIsConnected(true);

    hal.populateCache(dsuid1);
    auto completeLight = db.tryLookup(dsuid1);

    completeLight.property[DeviceProperty::outputMode] = 1;
    completeLight.property[DeviceProperty::switchedThreshold] = 1;
    completeLight.property[DeviceProperty::dimTime0Up] = 1;
    completeLight.property[DeviceProperty::dimTime0Down] = 1;
    completeLight.property[DeviceProperty::dimTime1Up] = 1;
    completeLight.property[DeviceProperty::dimTime1Down] = 1;
    completeLight.property[DeviceProperty::dimTime2Up] = 1;
    completeLight.property[DeviceProperty::dimTime2Down] = 1;
    completeLight.property[DeviceProperty::dimStepUp] = 1;
    completeLight.property[DeviceProperty::dimStepDown] = 1;
    completeLight.property[DeviceProperty::minDim] = 1;
    completeLight.property[DeviceProperty::dimStepPause] = 1; // superfluous entry

    // Modify entry fore reuse

    entry.channels.clear();
    entry.channels[ChannelType::brightness].value = 100.0;
    entry.channels[ChannelType::brightness].dontCare = false;

    for (auto& scene : completeLight.scene.sceneTable) {
      scene.second = entry;
    }
    completeLight.scene.sceneTable[127] = entry;
    completeLight.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value = 42;
    db.update(dsuid1, completeLight);

    boost::shared_ptr<Device> devUnsupported = apt.allocateDevice(dsuid3);
    spec.FunctionID = 0x3000;

    devUnsupported->setPartiallyFromSpec(spec);
    devUnsupported->setIsPresent(true);
    devUnsupported->setIsConnected(true);

    WHEN("Periodic cache update runs and something changes") {
      // First no calls are allowed
      boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast);
      boost::shared_ptr<Event> pResumeEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
      pEvent->setProperty(EventProperty::ICalStartTime, DateTime().toRFC2445IcalDataTime());
      pEvent->setProperty(EventProperty::ICalRRule, "FREQ=SECONDLY;INTERVAL=" + intToString(5));

      boost::shared_ptr<SubscriptionOptions> opts = boost::make_shared<SubscriptionOptions>();
      opts->setParameter("event_name", CacheUpdatePlugin::CacheUpdateEventFast);
      auto subscription = boost::make_shared<EventSubscription>("my_event", "raise_event", interpreter, opts);

      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }

      // Then things get invalidated and calls are expected
      DeviceCache lightCache = db.tryLookup(dsuid1);
      lightCache.scene.sceneTable[5].channels[ChannelType::brightness].value = {};
      lightCache.scene.sceneTable[2].channels[ChannelType::brightness].dontCare = {};
      db.update(dsuid1, lightCache);

      DeviceCache shadeCache = db.tryLookup(dsuid2);
      shadeCache.scene.sceneTable[2].flags.specialHandling = {};
      shadeCache.property[DeviceProperty::dimStepLen] = {};
      db.update(dsuid2, shadeCache);

      REQUIRE_CALL(hal, readBool(dsuid2, trompeloeil::re(R"(/scene/2/dontCare)")))
          .TIMES(1)
          .RETURN(false)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid2);
            devCache.scene.sceneTable[2].flags.specialHandling = false;
            db.update(dsuid2, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid1, trompeloeil::re(R"(/scene/5/brightness)")))
          .TIMES(1)
          .RETURN(2.1)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[5].channels[ChannelType::brightness].value = 2.1;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readBool(dsuid1, trompeloeil::re(R"(/scene/2/brightness/dontCare)")))
          .TIMES(1)
          .RETURN(true)
          .LR_SIDE_EFFECT({
            auto devCache = db.tryLookup(dsuid1);
            devCache.scene.sceneTable[2].channels[ChannelType::brightness].dontCare = true;
            db.update(dsuid1, devCache);
          });

      REQUIRE_CALL(hal, readDouble(dsuid2, trompeloeil::re(R"(/property/dimStepLen)"))).RETURN(1).LR_SIDE_EFFECT({
        auto devCache = db.tryLookup(dsuid2);
        devCache.property[DeviceProperty::dimStepLen] = 1;
        db.update(dsuid2, devCache);
      });

      interface.handleEvent(*pEvent, *subscription);
      while (interface.isUpdateRunning()) {
        interface.handleEvent(*pResumeEvent, *subscription);
      }
      {
        auto completeLight = db.tryLookup(dsuid1);
        REQUIRE_THROWS(completeLight.property.at(DeviceProperty::dimStepPause));
        REQUIRE_THROWS(completeLight.scene.sceneTable.at(127));
        REQUIRE_THROWS(completeLight.scene.sceneTable[5].channels.at(ChannelType::shadePositionOutside));
      }
    }
  }
}

} // namespace
