/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "cache-updater.h"
#include <algorithm>
#include <map>
#include <sstream>
#include <thread>
#include "apartment-proxy.h"
#include "cache-db.h"
#include "device-cache.h"
#include "event.h"
#include "hal/hardware-abstraction.h"
#include "model/apartment.h"
#include "model/set.h"
#include "model/zone.h"
#include "propertysystem_common_paths.h"
#include "src/dsa-http-client.h"
#include "test-device-presence.h"

namespace dss {

constexpr const char* CacheUpdatePlugin::CacheUpdateEvent;
constexpr const char* CacheUpdatePlugin::CacheUpdateEventFast;
constexpr const char* CacheUpdatePlugin::CacheUpdateEventResume;
constexpr const int CacheUpdatePlugin::CacheUpdateDefaultBackoffInSec;

static ds::LogChannel dsLogChannel("CacheUpdatePlugin");

class CacheMissException : public std::runtime_error {
  using std::runtime_error::runtime_error;
};

CacheUpdatePlugin::CacheUpdatePlugin(const Apartment* apt, HardwareAbstraction* hal, EventInterpreter& interpreter,
                                     boost::shared_ptr<PropertySystem> propSystem,
                                     boost::shared_ptr<EventQueue> eventQueue,
                                     boost::shared_ptr<EventRunner> eventRunner)
    : EventInterpreterPlugin("device_scene_cache_updater", &interpreter),
      m_apartment(apt),
      m_hal(hal),
      m_pPropSystem(propSystem),
      m_pEventQueue(eventQueue),
      m_pEventRunner(eventRunner),
      m_updateInProgress(false),
      m_hitBackoffInSec(CacheUpdateDefaultBackoffInSec) {
  if (m_pPropSystem) {
    auto statusProperty = m_pPropSystem->createProperty(pp_device_cache_update_status);
    statusProperty->linkToProxy(PropertyProxyReference<bool>(m_updateInProgress, false));
  }
}

CacheUpdatePlugin::~CacheUpdatePlugin() {}

bool CacheUpdatePlugin::updateIfNotCached(const Dsuid& dsuid, SceneFlags& flags, std::string path) {
  if (!flags.dontCare || !flags.ignoreLocalPrio || !flags.specialHandling || !flags.blink || !flags.ledConfig ||
      !flags.dimTime) {
    m_hal->readBool(dsuid, path + "/dontCare");
    return false;
  }

  return true;
}

bool CacheUpdatePlugin::updateIfNotCached(const Dsuid& dsuid, SceneEntry& entry, std::string path) {
  if (!updateIfNotCached(dsuid, entry.flags, path)) {
    return false;
  }
  if (!entry.action) {
    // Not used
  }
  // Figure out if this scene pertains to device
  for (auto& channelPair : entry.channels) {
    if (!channelPair.second.value) {
      m_hal->readDouble(dsuid, path + "/" + ds::fromEnum<std::string>(channelPair.first));
      return false;
    } else if (!channelPair.second.dontCare) {
      m_hal->readBool(dsuid, path + "/" + ds::fromEnum<std::string>(channelPair.first) + "/dontCare");
      return false;
    }
  }

  return true;
}

bool CacheUpdatePlugin::updateIfNotCached(const Dsuid& dsuid, DeviceSceneCache& sceneCache, std::string path) {
  // TODO: Optimize the number of times we run through
  if (!sceneCache.changeCounter) {
    // Not implemented yet
  }

  DS_DEBUG("Scene count", sceneCache.sceneTable.size());
  for (auto& scenePair : sceneCache.sceneTable) {
    if (!updateIfNotCached(dsuid, scenePair.second, path + "/" + std::to_string(scenePair.first))) {
      return false;
    }
  }

  return true;
}

bool CacheUpdatePlugin::updateIfNotCached(const Dsuid& dsuid, DeviceCache& cache, std::string path) {
  if (!updateIfNotCached(dsuid, cache.scene, path + "/scene")) {
    return false;
  }

  for (auto& property : cache.property) {
    if (!property.second) {
      std::stringstream propertyStr;
      propertyStr << property.first;
      if (property.first == DeviceProperty::outputMode) {
        m_hal->readString(dsuid, path + "/property/" + propertyStr.str());
      } else {
        m_hal->readDouble(dsuid, path + "/property/" + propertyStr.str());
      }
      return false;
    }
  }
  return true;
}

void CacheUpdatePlugin::reschedule() {
  boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventResume);
  pEvent->setProperty(EventProperty::Time, "+" + std::to_string(m_hitBackoffInSec));
  m_pEventQueue->pushEvent(pEvent);

  DS_INFO("Cache miss detected, rescheduling in " + std::to_string(m_hitBackoffInSec) + " seconds.");
}

void CacheUpdatePlugin::scanCache() {
  DS_DEBUG("Scanning Device Scene Cache");
  Set devices = m_apartment->getDevices().getByPresence(true);

  DS_DEBUG("Set size", devices.length());

  int continueIndex = 0;
  if (m_currentDev) {
    for (int d = 0; d < devices.length(); d++) {
      if (devices[d].getDevice()->getDSID() == *m_currentDev) {
        continueIndex = d;
        break;
      }
    }
  }

  for (int d = continueIndex; d < devices.length(); d++) {
    boost::shared_ptr<Device> device = devices[d].getDevice();
    if (!m_hal->deviceCacheable(device->getDSID())) {
      continue;
    }

    int retriesLeft = 2;
    bool deviceFailHard = false;
    bool success = false;

    while ((retriesLeft > 0 && !deviceFailHard) && !success) {
      try {
        if (!m_currentDev || *m_currentDev != device->getDSID()) {
          m_currentDev = device->getDSID();
          m_hal->populateCache(*m_currentDev);
        }
        DeviceCache cacheEntry = m_hal->getCache().tryLookup(device->getDSID());
        DS_INFO("Device to scan ", dsuid2str(device->getDSID()), ds::toJson(cacheEntry));

        std::string path = "";
        if (!updateIfNotCached(device->getDSID(), cacheEntry, path)) {
          reschedule();
          return;
        }
        success = true;
      } catch (const BusApiError& e) {
        if (std::string(e.what()).find("[-20]") != std::string::npos) {
          retriesLeft--;
          if (retriesLeft == 0) {
            auto&& dss = *DSS::getInstance();
            auto dsa = makeDsaHttpClient();
            auto& backendVdcs = dss.backendVdcs();
            DS_WARNING("Device failed to react during cache scan, pinging device", dsuid2str(device->getDSID()));
            testDevicePresence(std::locale(""), *device, *dsa, backendVdcs);
          } else {
            DS_WARNING("Device failed to react during cache scan, retries Left: " + std::to_string(retriesLeft));
          }
        } else {
          deviceFailHard = true;
        }
      } catch (const ds::Exception& e) {
        DS_NOTICE("Could not scan DSUID: " + dsuid2str(device->getDSID()), e.what());
        deviceFailHard = true;
      }
    }
  }
  DS_DEBUG("Cache update process completed");
  m_hitBackoffInSec = CacheUpdateDefaultBackoffInSec;
  m_updateInProgress = false;
  m_currentDev = {};
  m_apartment->proxy().updateParameterCacheStatus();
  boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(EventName::CacheUpdateFinished);
  m_pEventQueue->pushEvent(pEvent);
}

void CacheUpdatePlugin::handleEvent(Event& _event, const EventSubscription& _subscription) {
  try {
    if (_event.getName() == CacheUpdateEvent) {
      if (!m_updateInProgress) {
        m_apartment->proxy().updateParameterCacheStatus();
        m_updateInProgress = true;
        scanCache();
      }
    } else if (_event.getName() == CacheUpdateEventFast) {
      if (!m_updateInProgress) {
        m_updateInProgress = true;
        m_hitBackoffInSec = 0;
        m_apartment->proxy().updateParameterCacheStatus();
        scanCache();
      }
    } else if (_event.getName() == CacheUpdateEventResume) {
      scanCache();
    }
  } catch (const std::exception& e) {
    DS_WARNING("Exception occured during cache update process", e.what());
  } catch (...) {
    DS_WARNING("Unexpected error encountered in cache updater.");
    reschedule();
  }
}

void CacheUpdatePlugin::subscribe() {
  auto&& interpreter = getEventInterpreter();
  interpreter.subscribe(boost::make_shared<EventSubscription>(*this, CacheUpdateEvent));
  interpreter.subscribe(boost::make_shared<EventSubscription>(*this, CacheUpdateEventFast));
  interpreter.subscribe(boost::make_shared<EventSubscription>(*this, CacheUpdateEventResume));

  boost::shared_ptr<Event> pEvent = boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent);
  pEvent->setProperty(EventProperty::ICalStartTime, DateTime().addMinute(10).toRFC2445IcalDataTime());
  PropertyNodePtr scanIntervalNode = m_pPropSystem->getProperty(pp_device_cache_update_interval);
  int scanDelay = 1440;
  if (scanIntervalNode != nullptr) {
    scanDelay = m_pPropSystem->getProperty(pp_device_cache_update_interval)->getIntegerValue();
  } else {
    DS_WARNING("Could not find interval for cache update in property tree, assuming 1 day.");
  }
  pEvent->setProperty(EventProperty::ICalRRule, "FREQ=MINUTELY;INTERVAL=" + intToString(scanDelay));
  m_pEventQueue->pushEvent(pEvent);
}

} // namespace dss
