/*
    Copyright (c) 2009,2011 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
            Michael Tross, aizo GmbH <michael.tross@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef _DATE_TOOLS_H_INCLUDED
#define _DATE_TOOLS_H_INCLUDED

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "logger.h"

#include <ds/date-time.h>
#include <libical/ical.h>
#include <sys/time.h>
#include <ostream>
#include <vector>

#if defined(__APPLE__)
#include <mach/mach_init.h>
#include <mach/thread_act.h>
#include <mach/mach_port.h>
#endif

namespace dss {

  typedef enum {
    Sunday = 0, Monday, Tuesday, Wednesday, Thursday, Friday, Saturday
  } Weekday;

  typedef enum {
    January = 0, February, March, April, May, June, July, August, September, October, November, December
  } Month;

  class TZSwitcher {
  public:
    TZSwitcher(const char *tz) {
      m_old_tz = getenv("TZ");
      (void)setenv("TZ", tz, true);
      tzset();
    }
    ~TZSwitcher() {
      if (m_old_tz) {
        setenv("TZ", m_old_tz, true);
      } else {
        unsetenv("TZ");
      }
      tzset();
    }
  private:
    const char *m_old_tz;
  };

  /** Class that supports various date and time calculations */
  class DateTime {
  private:
    struct timeval m_timeval;
  public:
    /** Current date-time */
    DateTime();
    /** Construct from sec.usec since epoch (UTC) */
    DateTime(time_t secs, suseconds_t usecs = 0);
    DateTime(struct timeval tv) : m_timeval(tv) {}
    explicit DateTime(ds::DateTime);
    /** Copy constuctor */
    DateTime(const DateTime& _copy) = default;

    /** Adds \a _hours hours to the time and normalizes the DateTime */
    DateTime addHour(const int _hours) const;
    /** Adds \a _minutes minutes to the time and normalizes the DateTime */
    DateTime addMinute(const int _minutes) const;
    /** Adds \a _seconds seconds to the time and normalizes the DateTime */
    DateTime addSeconds(const int _seconds) const;
    /** Adds \a _milliseconds seconds to the time and normalizes the DateTime */
    DateTime addMilliSeconds(const int _milliseconds) const;
    /** Adds \a _month months to the date and normalizes the DateTime */
    DateTime addMonth(const int _month) const;
    /** Adds \a _years years to the date and normalizes the DateTime */
    DateTime addYear(const int _years) const;
    /** Adds \a _days days to the date and normalizes the DateTime */
    DateTime addDay(const int _days) const;

    /** Returns the day of month */
    int getDay() const;
    /** Returns the month */
    int getMonth() const;
    /** Returns the year */
    int getYear() const;

    /** Returns the hour */
    int getHour() const;
    /** Returns the minute */
    int getMinute() const;
    /** Returns the second */
    int getSecond() const;

    /** Returns the day of year */
    int getDayOfYear() const;
    /** Returns the weekday */
    Weekday getWeekday() const;

    /** Returns true if the instance and \a _other represent the same date and time */
    bool operator==(const DateTime& _other) const;
    /** Returns true if the instance and \a _other do not represent the same time and date */
    bool operator!=(const DateTime& _other) const;
    /** Returns true if the instance is before _other.
      * @see before */
    bool operator<(const DateTime& _other) const;
    /** Returns true if the instance is after _other.
      * @see after */
    bool operator>(const DateTime& _other) const;
    /** Returns true if the instance is before or equal _other.
      * @see before */
    bool operator<=(const DateTime& _other) const;
    /** Returns true if the instance is after or equal _other.
      * @see after */
    bool operator>=(const DateTime& _other) const;

    /** Returns the difference in seconds */
    int difference(const DateTime& _other) const;

    /** Returns the seconds since epoch */
    time_t secondsSinceEpoch() const;

    /** Returns the offset in seconds from GMT */
    long int getTimezoneOffset() const;

    operator std::string() const;
    std::string toString() const;

    std::string toRFC2822String() const;

    /**
     * Parses RFC2445 date-time
     * @param _isoStr DateTime string formatted as "yyyymmddThhmmss[Z]"
     * @throw invalid_argument if a malformatted \a _isoStr is provided
     */
    static DateTime parseRFC2445(const std::string& _isoStr);

    /**
     * Emit RFC2445 date-time format
     */
    std::string toRFC2445IcalDataTime() const;

    /**
     * parseISO8601 -- ISO8601 or similar RFC3339
     * http://www.cl.cam.ac.uk/~mgk25/iso-time.html
     * http://www.cs.tut.fi/~jkorpela/iso8601.html
     * http://www.ietf.org/rfc/rfc3339.txt
     */
    static DateTime parseISO8601(std::string in);

    /**
     * Emit ISO8601 or RFC3339 format
     * http://www.cl.cam.ac.uk/~mgk25/iso-time.html
     * http://www.cs.tut.fi/~jkorpela/iso8601.html
     * http://www.ietf.org/rfc/rfc3339.txt
     */
    std::string toISO8601() const;

    /**
     * Will emit ISO8601 format with timezone appended
     *
     * Various versions are floating around for the timezone
     * format: e.g. 02:00, 0200, 02 and not all entities
     * accept all formats
     * We emit it as '+hh:mm' since that is the format
     * mentioned in the public rfc3339 specification[2]
     * It's recomended to emit the time in UTC where no
     * timezone needs to be appended
     *
     * @see toISO8601
     * [1] http://www.cs.tut.fi/~jkorpela/iso8601.html
     * [2] http://www.ietf.org/rfc/rfc3339.txt
     */
    std::string toISO8601_local() const;

    /**
     * Emit ISO8601 or RFC3339 with ms precision
     * http://www.cl.cam.ac.uk/~mgk25/iso-time.html
     * http://www.cs.tut.fi/~jkorpela/iso8601.html
     * http://www.ietf.org/rfc/rfc3339.txt
     * TODO probably non-standard:
     * - rfc3339 has 100 ms digit precision
     * - ISO8601 seems seconds only precision(spec is 140$)
     *
     * Various versions are floating around for the timezone
     * format: e.g. 02:00, 0200, 02 and not all entities
     * accept all formats
     * We emit it as '+hh:mm' since that is the format
     * mentioned in the public rfc3339 specification[2]
     * It's recomended to emit the time in UTC where no
     * timezone needs to be appended
     */
    std::string toISO8601_ms_local() const;

    /**
     * Emit ISO8601 or RFC3339 with ms precision
     * http://www.cl.cam.ac.uk/~mgk25/iso-time.html
     * http://www.cs.tut.fi/~jkorpela/iso8601.html
     * http://www.ietf.org/rfc/rfc3339.txt
     * TODO probably non-standard:
     * - rfc3339 has 100 ms digit precision
     * - ISO8601 seems seconds only precision(spec is 140$)
     */
    std::string toISO8601_ms() const;

    /**
     * Parses human readable "2014-08-07 23:33:30" format
     * @throw invalid_argument if parsing fails
     */
    static DateTime parsePrettyString(const std::string& strTime);

    /**
     * Emits human readable "2014-08-07 23:33:30" format
     */
    std::string toPrettyString() const;

    explicit operator ds::DateTime() const;

    /** The NullDate has it's date and time parts set to 0. It should
      * be used for default values. */
    static DateTime NullDate;
  }; // DateTime

  std::ostream& operator<<(std::ostream& out, const DateTime& _dt);

  /** A Schedule is an abstract construct that's able to determine the
    * next occurence or even the next occurences. */
  class Schedule {
  public:
     virtual ~Schedule() {}
    /** Returns the date of the next occurence.
      * @return \a DateTime::NullDate or a \a DateTime value after \a _from
      */
    virtual DateTime getNextOccurence(const DateTime& _from) = 0;
    virtual bool hasRecurrence() const = 0;
    virtual size_t leapAdjust(const time_t endTime) = 0;
  };

  /** Schedule that's scheduled on a specific DateTime */
  class StaticSchedule : public Schedule {
  private:
    DateTime m_When;
  public:
    StaticSchedule(const DateTime& _when) : m_When(_when) {}
    virtual ~StaticSchedule() {}
    virtual DateTime getNextOccurence(const DateTime& _from);
    virtual bool hasRecurrence() const { return false; }
    virtual size_t leapAdjust(const time_t endTime) { return 0; }
  };

  /** Schedule that gets it's schedule from an iCal's RRULE */
  class ICalSchedule : public Schedule {
    __DECL_LOG_CHANNEL__
  private:
    struct icalrecurrencetype m_Recurrence;
    struct icaltimetype m_StartDate;
    struct icaltimetype m_NextSchedule;
    icalrecur_iterator* m_ICalIterator;
  public:
    ICalSchedule(const std::string& _rrule, const std::string _startDateISO);
    virtual ~ICalSchedule();

    virtual DateTime getNextOccurence(const DateTime& _from);
    virtual bool hasRecurrence() const { return true; }
    virtual size_t leapAdjust(const time_t endTime);
}; // ICalSchedule

  class ICalEvent {
    __DECL_LOG_CHANNEL__
  private:
    struct icalrecurrencetype m_Recurrence;
    struct icaltimetype m_StartDate;
    struct icaldurationtype m_Duration;
  public:
    ICalEvent(const std::string& _rrule, const std::string& _startDate, const std::string& _endDate);
    ~ICalEvent();
    bool isDateInside(const DateTime& _date);
  }; // ICalEvent

  class TimeStamp {
  public:
    TimeStamp() {
      m_stamp.tv_sec = 0;
      m_stamp.tv_nsec = 0;
    };

    void timestamp();
    //unsigned toNanoSec() const; /* mind the range is only 32bit */
    unsigned toMicroSec() const;

    TimeStamp operator+(const TimeStamp &o) const;
    TimeStamp operator-(const TimeStamp &o) const;
    TimeStamp operator+=(const TimeStamp &o);
    TimeStamp operator-=(const TimeStamp &o);
    bool operator<(const TimeStamp &other) const;

  private:
    struct timespec m_stamp;
  };

#ifndef __APPLE__
  /* inline to minimize overhead */
  inline void TimeStamp::timestamp() {
    (void)clock_gettime(CLOCK_THREAD_CPUTIME_ID, &m_stamp);
  }
#else
  // gets the CPU time used by the current thread (both system and user), in
  // microseconds
  // http://stackoverflow.com/questions/17372110/clock-thread-cputime-id-on-macosx
  inline void TimeStamp::timestamp() {
    thread_port_t thread = mach_thread_self();
    mach_msg_type_number_t count = THREAD_BASIC_INFO_COUNT;
    thread_basic_info_data_t info;

    int kr = thread_info(thread, THREAD_BASIC_INFO, (thread_info_t) &info, &count);
    if (kr != KERN_SUCCESS) {
        return;
    }
    mach_port_deallocate(mach_task_self(), thread);
    m_stamp.tv_sec = info.user_time.seconds + info.system_time.seconds;
    m_stamp.tv_nsec = (info.user_time.microseconds + info.system_time.microseconds) * 1e3;
  }
#endif
}
#endif
