/*
    Copyright (c) 2016 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/
#include "device-info.h"
#include <ds-model/scenarios.h>
#include <ds/catch/catch.h>
#include <ds/locale/ds-locale.h>
#include <ds/locale/test.h>
#include <ds/log.h>
#include <ds/system-mock.h>
#include <boost/locale/generator.hpp>
#include "businterface-mock.h"
#include "ds-scenarios.h"
#include "dsm-api-mock.h"
#include "messages/vdc-messages.pb.h"
#include "model/device.h"
#include "protobufjson.h"
#include "vdc-db-mock.h"

static ds::LogChannel dsLogChannel("deviceInfoTest");

using namespace dss;

namespace {

static const char* TAGS = "[dssDeviceInfo][dss]";

DeviceInfo::Filter ALL_SET_FILTER = [] {
  DeviceInfo::Filter x = {};
  x.spec = 1;
  x.stateDesc = 1;
  x.eventDesc = 1;
  x.propertyDesc = 1;
  x.sensorDesc = 1;
  x.actionDesc = 1;
  x.standardActions = 1;
  x.customActions = 1;
  x.dynamicDeviceActions = 1;
  x.operational = 1;
  x.categories = 1;
  x.outputDesc = 1;
  return x;
}();

TEST_CASE("dssDeviceInfoParseFilter", TAGS) {
  SECTION("Every field is parsed on its own") {
    DeviceInfo::Filter filter = {};
    SECTION("spec") {
      filter.spec = 1;
      CHECK(DeviceInfo::parseFilter("spec") == filter);
    }
    SECTION("stateDesc") {
      filter.stateDesc = 1;
      CHECK(DeviceInfo::parseFilter("stateDesc") == filter);
    }
    SECTION("eventDesc") {
      filter.eventDesc = 1;
      CHECK(DeviceInfo::parseFilter("eventDesc") == filter);
    }
    SECTION("propertyDesc") {
      filter.propertyDesc = 1;
      CHECK(DeviceInfo::parseFilter("propertyDesc") == filter);
    }
    SECTION("sensorDesc") {
      filter.sensorDesc = 1;
      CHECK(DeviceInfo::parseFilter("sensorDesc") == filter);
    }
    SECTION("actionDesc") {
      filter.actionDesc = 1;
      CHECK(DeviceInfo::parseFilter("actionDesc") == filter);
    }
    SECTION("stdActions") {
      filter.standardActions = 1;
      CHECK(DeviceInfo::parseFilter("standardActions") == filter);
    }
    SECTION("customActions") {
      filter.customActions = 1;
      CHECK(DeviceInfo::parseFilter("customActions") == filter);
    }
    SECTION("dynamicDeviceActions") {
      filter.dynamicDeviceActions = 1;
      CHECK(DeviceInfo::parseFilter("dynamicDeviceActions") == filter);
    }
    SECTION("operational") {
      filter.operational = 1;
      CHECK(DeviceInfo::parseFilter("operational") == filter);
    }
    SECTION("categories") {
      filter.categories = 1;
      CHECK(DeviceInfo::parseFilter("categories") == filter);
    }
    SECTION("outputDesc") {
      filter.outputDesc = 1;
      CHECK(DeviceInfo::parseFilter("outputDesc") == filter);
    }
  }

  SECTION("Empty filter is parsed as all set") { CHECK(DeviceInfo::parseFilter("") == ALL_SET_FILTER); }

  SECTION("All fields set are parsed") {
    CHECK(DeviceInfo::parseFilter("spec,"
                                  "stateDesc,"
                                  "eventDesc,"
                                  "propertyDesc,"
                                  "sensorDesc,"
                                  "actionDesc,"
                                  "standardActions,"
                                  "customActions,"
                                  "dynamicDeviceActions,"
                                  "operational,"
                                  "categories,"
                                  "outputDesc") == ALL_SET_FILTER);
  }
}

vdcapi::PropertyElement& addField(vdcapi::PropertyElement& e, const std::string& key) {
  auto&& x = *e.add_elements();
  x.set_name(key);
  return x;
}

void addField(vdcapi::PropertyElement& e, const std::string& key, bool value) {
  addField(e, key).mutable_value()->set_v_bool(value);
}

void addField(vdcapi::PropertyElement& e, const std::string& key, int value) {
  addField(e, key).mutable_value()->set_v_int64(value);
};

// void addField(vdcapi::PropertyElement& e, const std::string& key, unsigned value) {
//   addField(e, key).mutable_value()->set_v_uint64(value);
// };
//
// void addField(vdcapi::PropertyElement& e, const std::string& key, int64_t value) {
//   addField(e, key).mutable_value()->set_v_int64(value);
// };
//
// void addField(vdcapi::PropertyElement& e, const std::string& key, uint64_t value) {
//   addField(e, key).mutable_value()->set_v_uint64(value);
// };

void addField(vdcapi::PropertyElement& e, const std::string& key, double value) {
  addField(e, key).mutable_value()->set_v_double(value);
};

// void addField(vdcapi::PropertyElement& e, const std::string& key, const std::string& value) {
//   addField(e, key).mutable_value()->set_v_string(value);
// };

// otherwise const char *  maps to bool
void addField(vdcapi::PropertyElement& e, const std::string& key, const char* value) {
  addField(e, key).mutable_value()->set_v_string(value);
}

TEST_CASE("dssDeviceInfo", TAGS) {
  JSONWriter json;
  auto&& gtin = "1234567890123";
  auto&& locale = ds::locale::test::makeLocale();

  ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(trompeloeil::_, trompeloeil::_, trompeloeil::_, trompeloeil::_, trompeloeil::_, trompeloeil::_, trompeloeil::_)).RETURN(0);

  StructureQueryBusInterfaceMock missingQuery;
  ALLOW_CALL(missingQuery, getVdcProperty(trompeloeil::_, trompeloeil::_, trompeloeil::_)).RETURN(vdcapi::Message{});

  VdcDbMock missingDb;
  ALLOW_CALL(missingDb, getSpec(gtin)).RETURN(VdcDb::Specs{});
  ALLOW_CALL(missingDb, getStates(gtin)).RETURN(VdcDb::States{});
  ALLOW_CALL(missingDb, getEvents(gtin)).RETURN(VdcDb::Events{});
  ALLOW_CALL(missingDb, getProperties(gtin)).RETURN(VdcDb::Properties{});
  ALLOW_CALL(missingDb, getSensors(gtin)).RETURN(VdcDb::Sensors{});
  ALLOW_CALL(missingDb, getActions(gtin)).RETURN(VdcDb::Actions{});
  ALLOW_CALL(missingDb, getStandardActions(gtin)).RETURN(VdcDb::StandardActions{});
  ALLOW_CALL(missingDb, getCategories(gtin)).RETURN(VdcDb::Categories{});
  ALLOW_CALL(missingDb, getOutputs(gtin)).RETURN(VdcDb::Outputs{});

  auto system = ds::SystemMock{};

  Device vdcDevice(DSUID_NULL, NULL);
  vdcDevice.setOemInfo(1234567890123, 0, 0, DEVICE_OEM_EAN_NO_INTERNET_ACCESS, 0);
  vdcDevice.setName("dev::name");
  // isConnected iff isPresent && isConnected
  vdcDevice.setIsPresent(true);
  vdcDevice.setIsConnected(true);
  vdcDevice.setVdcDevice(true);
  VdsdSpec_t vdsdSpec;
  vdsdSpec.oemGuid = "x-oemGuid";
  vdsdSpec.oemModelGuid = "x-oemModelGuid";
  vdsdSpec.vendorGuid = "x-vendorGuid";
  vdsdSpec.vendorId = "x-vendorId";
  vdsdSpec.hardwareGuid = "x-hardwareGuid";
  vdsdSpec.hardwareModelGuid = "x-hardwareModelGuid";
  vdsdSpec.modelUID = "x-modelUID";
  vdsdSpec.descriptionsGroup = "x-descriptionsGroup";
  vdsdSpec.descriptionsClass = "x-descriptionsClass";
  vdsdSpec.name = "x-name";
  vdsdSpec.model = "x-model";
  vdsdSpec.hardwareVersion = "x-hardwareVersion";
  vdsdSpec.modelVersion = "x-modelVersion";
  vdsdSpec.vendorName = "x-vendorName";
  vdcDevice.setVdsdSpec(std::move(vdsdSpec));

  auto presentQueryFunction = [&](const dsuid_t& vdcId, const dsuid_t& targetId, const std::vector<std::string>& keys) {
    auto addStates = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& states = *getProperty.add_properties();
      states.set_name("deviceStateDescriptions");
      { // state present in db
        auto& state = addField(states, "state0");
        auto& values = addField(addField(state, "value"), "values");
        addField(values, "a");
        addField(values, "c");
      }
      { // state missing in db
        auto& state = addField(states, "stateX");
        auto& values = addField(addField(state, "value"), "values");
        addField(values, "a");
      }
    };
    auto addEvents = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& events = *getProperty.add_properties();
      events.set_name("deviceEventDescriptions");
      { addField(events, "event0"); }
      { addField(events, "eventx"); }
    };
    auto addProperties = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& properties = *getProperty.add_properties();
      properties.set_name("devicePropertyDescriptions");
      {
        auto& property = addField(properties, "integer0");
        addField(property, "type", "integer");
        addField(property, "default", "8");
        addField(property, "min", "-888");
        addField(property, "max", "888");
        addField(property, "resolution", "2");
        addField(property, "siunit", "s");
      }
      {
        auto& property = addField(properties, "integer9");
        addField(property, "type", "integer");
        addField(property, "default", "6");
        addField(property, "min", "-666");
        addField(property, "max", "666");
        addField(property, "resolution", "2");
        addField(property, "readonly", true);
      }
      {
        auto& property = addField(properties, "numeric0");
        addField(property, "type", "numeric");
        addField(property, "default", "6");
        addField(property, "min", "-666");
        addField(property, "max", "666");
        addField(property, "resolution", "2");
        addField(property, "siunit", "s");
        addField(property, "readonly", false); // false is default
      }
      {
        auto& property = addField(properties, "numeric1");
        addField(property, "type", "numeric");
        addField(property, "default", "6.6");
        addField(property, "min", "-666.6");
        addField(property, "max", "666.6");
        addField(property, "resolution", "0.2");
        addField(property, "siunit", "s");
      }
      {
        auto& property = addField(properties, "numeric2");
        addField(property, "type", "string"); // does not match db record
        addField(property, "default", "a");
        addField(property, "readonly", true);
      }
      {
        auto& property = addField(properties, "string0");
        addField(property, "type", "string");
        addField(property, "default", "hi");
      }
      {
        auto& property = addField(properties, "string1");
        addField(property, "type", "string");
        addField(property, "default", "hi");
      }
      {
        auto& property = addField(properties, "string2");
        addField(property, "type", "numeric"); // does not match db record
        addField(property, "default", "6");
      }
      {
        auto& property = addField(properties, "enum0");
        addField(property, "type", "enumeration");
        addField(property, "default", "c");
        auto& values = addField(property, "values");
        addField(values, "a");
        addField(values, "c");
      }
    };
    auto addSensors = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& sensors = *getProperty.add_properties();
      sensors.set_name("sensorDescriptions");
      {
        // vdc-api v3 format: "techn. name" : { dsIndex: 7 }
        auto& sensor = addField(sensors, "sensor0");
        addField(sensor, "name", "Sensor 0");
        addField(sensor, "min", -6.6);
        addField(sensor, "max", 6.6);
        addField(sensor, "resolution", 0.1);
        addField(sensor, "sensorType", 4);
        addField(sensor, "dsIndex", 0);
      }
      {
        // legacy format: "1" : { name: "techn. name" }
        auto& sensor = addField(sensors, "1");
        addField(sensor, "name", "sensor1x"); // no match in db
        addField(sensor, "min", -6.6);
        addField(sensor, "max", 6.6);
        addField(sensor, "resolution", 0.1);
        addField(sensor, "sensorType", 5);
      }
    };
    auto addActions = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& actions = *getProperty.add_properties();
      actions.set_name("deviceActionDescriptions");
      {
        auto& action = addField(actions, "action0");
        auto& params = addField(action, "params");
        {
          auto& param = addField(params, "param0");
          addField(param, "type", "enumeration");
          addField(param, "optional", true);
          auto& values = addField(param, "values");
          addField(values, "a");
          addField(values, "c");
        }
        {
          auto& param = addField(params, "param9");
          addField(param, "type", "string");
          addField(param, "optional", false); // false is default
        }
      }
      { addField(actions, "action9"); }
    };
    auto addStandardActions = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& actions = *getProperty.add_properties();
      actions.set_name("standardActions");
      {
        auto& action = addField(actions, "std.action0");
        addField(action, "action", "action0");
        addField(action, "title", "title0");
        auto& params = addField(action, "params");
        {
          addField(params, "a", "A");
          addField(params, "b", "B");
        }
      }
    };
    auto addCustomActions = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& actions = *getProperty.add_properties();
      actions.set_name("customActions");
      {
        auto& action = addField(actions, "custom.action1");
        addField(action, "action", "action1");
        addField(action, "title", "title1");
        auto& params = addField(action, "params");
        {
          addField(params, "a", "A");
          addField(params, "b", "B");
        }
      }
    };
    auto addDynamicActions = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& actions = *getProperty.add_properties();
      actions.set_name("dynamicActionDescriptions");
      {
        auto& action = addField(actions, "dynamic.action2");
        addField(action, "action", "action2");
        addField(action, "title", "title2");
        auto& params = addField(action, "params");
        {
          addField(params, "a", "A");
          addField(params, "b", "B");
        }
      }
    };
    auto addDeviceStates = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& states = *getProperty.add_properties();
      states.set_name("deviceStates");
      {
        // extracts only age/value
        auto& operation = addField(states, "operation");
        addField(operation, "age", 1.011742);
        addField(operation, "value", "ready");
        addField(operation, "changed", 22070.542712); // ignored
      }
    };
    auto addDeviceProperties = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& states = *getProperty.add_properties();
      states.set_name("deviceProperties");
      {
        addField(states, "currentTemperature", 25.0);
        addField(states, "defaultkeepwarmtime", 0.0);
      }
    };
    auto addSensorStates = [&](::vdcapi::vdc_ResponseGetProperty& getProperty) {
      auto& states = *getProperty.add_properties();
      states.set_name("sensorStates");
      {
        auto& consumption = addField(states, "consumption");
        addField(consumption, "value", 0.2);
        addField(consumption, "age", 13.077); // will be rounded int(seconds)
        addField(consumption, "error", 0);
        auto& windowOpen = addField(states, "windowOpen");
        addField(windowOpen, "value", true);
        addField(windowOpen, "age", 13.077);
        addField(windowOpen, "error", 0);
      }
    };

    vdcapi::Message message;
    message.set_type(vdcapi::VDC_RESPONSE_GET_PROPERTY);
    auto& getProperty = *message.mutable_vdc_response_get_property();
    for (auto&& key : keys) {
      if (key == "deviceStateDescriptions") {
        addStates(getProperty);
      } else if (key == "deviceEventDescriptions") {
        addEvents(getProperty);
      } else if (key == "devicePropertyDescriptions") {
        addProperties(getProperty);
      } else if (key == "sensorDescriptions") {
        addSensors(getProperty);
      } else if (key == "deviceActionDescriptions") {
        addActions(getProperty);
      } else if (key == "standardActions") {
        addStandardActions(getProperty);
      } else if (key == "deviceStates") {
        addDeviceStates(getProperty);
      } else if (key == "deviceProperties") {
        addDeviceProperties(getProperty);
      } else if (key == "sensorStates") {
        addSensorStates(getProperty);
      } else if (key == "customActions") {
        addCustomActions(getProperty);
      } else if (key == "dynamicActionDescriptions") {
        addDynamicActions(getProperty);
      }
    }
    {
      JSONWriter jsonMessage;
      ProtobufToJSon::protoPropertyToJson(message, jsonMessage);
      DS_DEBUG_LEAVE(keys, jsonMessage.successJSON());
    }
    return message;
  };
  StructureQueryBusInterfaceMock presentQuery;
  ALLOW_CALL(presentQuery, getVdcProperty(trompeloeil::_, trompeloeil::_, trompeloeil::_))
      .RETURN(presentQueryFunction(_1, _2, _3));

  VdcDbMock presentDb;
  ALLOW_CALL(presentDb, getSpec(gtin))
      .RETURN(VdcDb::Specs{
          {"dsDeviceGTIN", "settings:5", ""},
          // [dS-Article Data]
          {"vendorId", "invisible", ""},
          // [Hardware Endpoint description]
          {"hardwareGuid", "settings:4", ""}, {"hardwareModelGuid", "invisible", ""},
          // [System ID's]
          {"descriptionsGroup", "invisible", ""}, {"descriptionsClass", "invisible", ""},
          // [End-User Info]
          {"name", "overview:1;settings:1", ""}, {"model", "overview:2;settings:2", ""},
          {"modelVersion", "invisible", ""}, {"vendorName", "overview:3;settings:3", ""},
          // [Added by db]
          {"noteId", "foo", "note"},
      });
  ALLOW_CALL(presentDb, getStates(gtin))
      .RETURN(VdcDb::States{{"state0", {"a", "b"}, "tag0"}, {"state1", {"a"}, "tag1"}});
  ALLOW_CALL(presentDb, getEvents(gtin))
      .RETURN(VdcDb::Events{
          {"event0", "category0", VdcDb::AccessLevel::SERVICE}, //
          {"event1", "category1", {}}                           //

      });
  ALLOW_CALL(presentDb, getProperties(gtin))
      .RETURN(VdcDb::Properties{
          {"integer0", VdcDb::integer, "", "1", "-100", "100", "1", "m", "tag0", false, false, {"-"}},
          {"numeric0", VdcDb::numeric, "", "1", "-100", "100", "1", "m", "tag0", false, false, {"-"}},
          {"numeric1", VdcDb::numeric, "numeric.duration", "1.1", "-100.0", "100.0", "0.1", "s", "tag1", false, false,
              {}},
          {"numeric2", VdcDb::numeric, "numeric.duration", "1.1", "-100.0", "100.0", "0.1", "s", "tag2", false, false,
              {}},
          {"string0", VdcDb::string, "", "hello", "-", "-", "-", "-", "tag0", false, false,
              {
                  "-",
              }},
          {"string1", VdcDb::string, "string.foo", "hello2", "-", "-", "-", "-", "tag1", false, false, {}},
          {"string2", VdcDb::string, "string.foo", "hello3", "-", "-", "-", "-", "tag2", false, false, {}},
          {"enum0", VdcDb::enumeration, "-", "a", "-", "-", "-", "-", "tag0", false, false, {"a", "b"}},
          {"bool0", VdcDb::numeric, "numeric.boolean", "true", "-", "-", "-", "-", "tag0", false, false, {}},
      });
  ALLOW_CALL(presentDb, getSensors(gtin))
      .RETURN(VdcDb::Sensors{
          {{"sensor0", VdcDb::integer, "", "2", "-10", "10", "1", "m", "tag0", false, false, {}}, "0",
              SensorType::ActivePower}, //
          {{"sensor1", VdcDb::integer, "", "2", "-10", "10", "1", "m", "tag1", false, false, {}}, "1",
              SensorType::ActivePower}, //
      });
  ALLOW_CALL(presentDb, getActions(gtin))
      .RETURN(
          VdcDb::Actions{{"action0", "category0", "invalidForCustomActions",
                             {{"param0", VdcDb::enumeration, "", "", "", "", "", "", "tag0", true, false, {"a", "b"}}}},
              {"action1", "category1", "", {}}});
  ALLOW_CALL(presentDb, getStandardActions(gtin))
      .RETURN(VdcDb::StandardActions{{"std.action0", "action0", {{"a", "A"}, {"b", "B"}}, "invalidForActivities"}});
  ALLOW_CALL(presentDb, getCategories(gtin))
      .RETURN(VdcDb::Categories{{"name0", "category0", false}, {"name1", "category1", false}});
  ALLOW_CALL(presentDb, getOutputs(gtin))
      .RETURN(VdcDb::Outputs{{"name0", VdcDb::numeric, "timeOfDay", "defaultValue0", "minValue0", "maxValue0",
          "resolution0", "siUnit0", "outputType0", "outputIndex0", "tags0", "categoryId0"}});

  SECTION("db is queried in constructor and passes correct gtin") {
    StructureQueryBusInterfaceMock query;
    VdcDbMock db;
    REQUIRE_CALL(db, getSpec(gtin)).RETURN(VdcDb::Specs{});
    REQUIRE_CALL(db, getStates(gtin)).RETURN(VdcDb::States{});
    REQUIRE_CALL(db, getEvents(gtin)).RETURN(VdcDb::Events{});
    REQUIRE_CALL(db, getProperties(gtin)).RETURN(VdcDb::Properties{});
    REQUIRE_CALL(db, getSensors(gtin)).RETURN(VdcDb::Sensors{});
    REQUIRE_CALL(db, getActions(gtin)).RETURN(VdcDb::Actions{});
    REQUIRE_CALL(db, getStandardActions(gtin)).RETURN(VdcDb::StandardActions{});
    REQUIRE_CALL(db, getCategories(gtin)).RETURN(VdcDb::Categories{});
    REQUIRE_CALL(db, getOutputs(gtin)).RETURN(VdcDb::Outputs{});

    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, query, db, nullptr, gtin, locale, json, system);
    deviceInfo.addByFilter(ALL_SET_FILTER);
  }

  SECTION("query is not invoked without device") {
    StructureQueryBusInterfaceMock query;
    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, query, missingDb, nullptr, gtin, locale, json, system);
    deviceInfo.addByFilter(ALL_SET_FILTER);
  }

  SECTION("device:missing, db:missing") {
    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, missingQuery, missingDb, nullptr, gtin, locale, json, system);
    deviceInfo.addByFilter(ALL_SET_FILTER);
    CHECK(
        json.successJSON() ==
        R"({"result":{)"
        R"("spec":{"dsDeviceGTIN":{"title":"vdc-common,vdc-gtin-1234567890123|spec|dsDeviceGTIN","tags":"","value":"1234567890123"}},)"
        R"("stateDescriptions":{},)"
        R"("eventDescriptions":{},)"
        R"("propertyDescriptions":{},)"
        R"("sensorDescriptions":{},)"
        R"("actionDescriptions":{},)"
        R"("standardActions":{},)"
        R"("operational":{"outputs":{},"states":{},"properties":{},"sensors":{}},)"
        R"("categories":{},)"
        R"("outputDescriptions":{})"
        R"(},"ok":true})");
  }

#define TEST_DOMAIN "vdc-common,vdc-gtin-1234567890123"
  SECTION("device:present, db:missing") {
    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, presentQuery, missingDb, &vdcDevice, gtin, locale, json, system);
    SECTION("addSpec") {
      deviceInfo.addSpec();
      CHECK(
          // no translations, no tags, but all values
          json.successJSON() ==
          R"({"result":{"spec":{)"
          R"("dsDeviceGTIN":{"title":")" TEST_DOMAIN R"(|spec|dsDeviceGTIN","tags":"","value":"1234567890123"},)"
          R"("vendorId":{"title":")" TEST_DOMAIN R"(|spec|vendorId","tags":"","value":"x-vendorId"},)"
          R"("hardwareGuid":{"title":")" TEST_DOMAIN R"(|spec|hardwareGuid","tags":"","value":"x-hardwareGuid"},)"
          R"("hardwareModelGuid":{"title":")" TEST_DOMAIN
          R"(|spec|hardwareModelGuid","tags":"","value":"x-hardwareModelGuid"},)"
          R"("descriptionsGroup":{"title":")" TEST_DOMAIN
          R"(|spec|descriptionsGroup","tags":"","value":"x-descriptionsGroup"},)"
          R"("descriptionsClass":{"title":")" TEST_DOMAIN
          R"(|spec|descriptionsClass","tags":"","value":"x-descriptionsClass"},)"
          R"("name":{"title":")" TEST_DOMAIN R"(|spec|name","tags":"","value":"dev::name"},)" // dev.getName()
          R"("model":{"title":")" TEST_DOMAIN R"(|spec|model","tags":"","value":"x-model"},)"
          R"("modelVersion":{"title":")" TEST_DOMAIN R"(|spec|modelVersion","tags":"","value":"x-modelVersion"},)"
          R"("vendorName":{"title":")" TEST_DOMAIN R"(|spec|vendorName","tags":"","value":"x-vendorName"})"
          R"(}},"ok":true})");
    }

    SECTION("addStateDescriptions") {
      deviceInfo.addStateDescriptions();
      CHECK(json.successJSON() ==
            R"({"result":{"stateDescriptions":{)"
            R"("state0":{"title":")" TEST_DOMAIN R"(|state|state0","tags":"","options":{"a":")" TEST_DOMAIN
            R"(|state/state0/enum|a","c":")" TEST_DOMAIN R"(|state/state0/enum|c"}},)"
            R"("stateX":{"title":")" TEST_DOMAIN R"(|state|stateX","tags":"","options":{"a":")" TEST_DOMAIN
            R"(|state/stateX/enum|a"})"
            R"(}}},"ok":true})");
    }

    SECTION("addEventDescriptions") {
      deviceInfo.addEventDescriptions();
      CHECK(json.successJSON() == R"({"result":{"eventDescriptions":{)"
                                  R"("event0":{"title":")" TEST_DOMAIN R"(|event|event0"},)"
                                  R"("eventx":{"title":")" TEST_DOMAIN R"(|event|eventx"})"
                                  R"(}},"ok":true})");
    }

    SECTION("addPropertyDescriptions") {
      deviceInfo.addPropertyDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"propertyDescriptions":{)"
          R"("integer0":{"title":")" TEST_DOMAIN
          R"(|property|integer0","tags":"","type":"numeric","min":"-888","max":"888","resolution":"2","siunit":"s","default":"8"},)"
          R"("integer9":{"title":")" TEST_DOMAIN
          R"(|property|integer9","tags":"","type":"numeric","min":"-666","max":"666","resolution":"2","siunit":"","default":"6","readonly":true},)"
          R"("numeric0":{"title":")" TEST_DOMAIN
          R"(|property|numeric0","tags":"","type":"numeric","min":"-666","max":"666","resolution":"2","siunit":"s","default":"6"},)"
          R"("numeric1":{"title":")" TEST_DOMAIN
          R"(|property|numeric1","tags":"","type":"numeric","min":"-666.6","max":"666.6","resolution":"0.2","siunit":"s","default":"6.6"},)"
          R"("numeric2":{"title":")" TEST_DOMAIN
          R"(|property|numeric2","tags":"","type":"string","default":"a","readonly":true},)"
          R"("string0":{"title":")" TEST_DOMAIN R"(|property|string0","tags":"","type":"string","default":"hi"},)"
          R"("string1":{"title":")" TEST_DOMAIN R"(|property|string1","tags":"","type":"string","default":"hi"},)"
          R"("string2":{"title":")" TEST_DOMAIN
          R"(|property|string2","tags":"","type":"numeric","min":"","max":"","resolution":"","siunit":"","default":"6"},)"
          R"("enum0":{"title":")" TEST_DOMAIN
          R"(|property|enum0","tags":"","type":"enumeration","default":"c","options":{"a":")" TEST_DOMAIN
          R"(|property/enum0/enum|a","c":")" TEST_DOMAIN R"(|property/enum0/enum|c"}})"
          R"(}},"ok":true})");
    }

    SECTION("addSensorDescriptions") {
      deviceInfo.addSensorDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"sensorDescriptions":{)"
          R"("sensor0":{"title":")" TEST_DOMAIN
          R"(|sensor|sensor0","tags":"","type":"numeric","min":"-6.6","max":"6.6","resolution":"0.1","siunit":"","default":"","dsIndex":0,"dsType":4},)"
          R"("sensor1x":{"title":")" TEST_DOMAIN
          R"(|sensor|sensor1x","tags":"","type":"numeric","min":"-6.6","max":"6.6","resolution":"0.1","siunit":"","default":"","dsIndex":1,"dsType":5})"
          R"(}},"ok":true})");
    }

    SECTION("addActionDescriptions") {
      deviceInfo.addActionDescriptions();
      CHECK(json.successJSON() ==
            R"({"result":{"actionDescriptions":{)"
            R"("action0":{"title":")" TEST_DOMAIN R"(|action|action0","params":{)"
            R"("param0":{"title":")" TEST_DOMAIN
            R"(|action/action0/param|param0","tags":"","type":"enumeration","default":"",)"
            R"("options":{"a":")" TEST_DOMAIN R"(|action/action0/param/param0/enum|a",)"
            R"("c":")" TEST_DOMAIN R"(|action/action0/param/param0/enum|c"},"optional":true},)"
            R"("param9":{"title":")" TEST_DOMAIN
            R"(|action/action0/param|param9","tags":"","type":"string","default":""})"
            R"(}},)"
            R"("action9":{"title":")" TEST_DOMAIN R"(|action|action9","params":{}})"
            R"(}},"ok":true})");
    }

    SECTION("addStandardAction") {
      deviceInfo.addStandardActions();
      CHECK(json.successJSON() ==
            R"({"result":{"standardActions":{)"
            R"("std.action0":{"title":")" TEST_DOMAIN
            R"(|standardAction|action0","action":"action0","params":{"a":"A","b":"B"}})"
            R"(}},"ok":true})");
    }

    // device:present db:missing
    SECTION("addOperational") {
      deviceInfo.addOperational();
      CHECK(json.successJSON() ==
            R"({"result":{"outputs":{},"states":{)"
            R"("operation":{"title":")" TEST_DOMAIN
            R"(|state|operation","value":"ready","age":1.011742,"displayValue":")" TEST_DOMAIN
            R"(|state/operation/enum|ready"})"
            R"(},"properties":{)"
            R"("currentTemperature":{"title":")" TEST_DOMAIN R"(|property|currentTemperature","value":25.0},)"
            R"("defaultkeepwarmtime":{"title":")" TEST_DOMAIN R"(|property|defaultkeepwarmtime","value":0.0})"
            R"(},"sensors":{)"
            R"("consumption":{"title":")" TEST_DOMAIN
            R"(|sensor|consumption","value":0.2,"timestamp":"2017-09-18T10:37:00.777Z"},)"
            R"("windowOpen":{"title":")" TEST_DOMAIN
            R"(|sensor|windowOpen","value":true,"timestamp":"2017-09-18T10:37:00.777Z"})"
            R"(}},"ok":true})");
    }

    // device:present db:missing
    SECTION("addCustomActions") {
      deviceInfo.addCustomActions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"customActions":{"custom.action1":{"action":"action1","title":"title1","params":{"a":"A","b":"B"}}}},"ok":true})");
    }

    SECTION("addDynamicDeviceActions") {
      deviceInfo.addDynamicDeviceActions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"deviceActions":{"dynamic.action2":{"action":"action2","title":"title2","params":{"a":"A","b":"B"}}}},"ok":true})");
    }

    SECTION("scenarios()") {
      const auto domain = "vdc-gtin-1234567890123";
      CHECK(deviceInfo.scenarios() ==
            std::vector<dsModel::Scenario>{
                {.id = encodeId(DeviceScenarioId{.device = vdcDevice.id(),
                                                 .action = dsModel::DeviceScenarioActionId{"custom.action1"}}),
                 .attributes =
                     dsModel::DeviceScenarioAttributes{.name = {.domain = {}, .message = "title1"},
                                                       .dsDevice = vdcDevice.id(),
                                                       .actionId = dsModel::DeviceScenarioActionId{"custom.action1"}}},
                {.id = encodeId(DeviceScenarioId{.device = vdcDevice.id(),
                                                 .action = dsModel::DeviceScenarioActionId{"dynamic.action2"}}),
                 .attributes =
                     dsModel::DeviceScenarioAttributes{.name = {.domain = {}, .message = "title2"},
                                                       .dsDevice = vdcDevice.id(),
                                                       .actionId = dsModel::DeviceScenarioActionId{"dynamic.action2"}}},
                {.id = encodeId(DeviceScenarioId{.device = vdcDevice.id(),
                                                 .action = dsModel::DeviceScenarioActionId{"std.action0"}}),
                 .attributes =
                     dsModel::DeviceScenarioAttributes{
                         .name = {.domain = domain, .message = "action0", .context = "standardAction"},
                         .dsDevice = vdcDevice.id(),
                         .actionId = dsModel::DeviceScenarioActionId{"std.action0"}}},
            });
    }

    SECTION("deviceStates") {
      CHECK(
          deviceInfo.deviceStates() ==
          std::vector<dsModel::DeviceState>{
              dsModel::DeviceState{
                  .id = dsModel::DeviceStateId{"state0"},
                  .options =
                      std::vector<dsModel::DeviceStateOption>{
                          dsModel::DeviceStateOption{.id = "a",
                                                     .name = {.domain = "vdc-gtin-1234567890123",
                                                              .message = "a",
                                                              .context = "state|state0|option"}},
                          dsModel::DeviceStateOption{.id = "c",
                                                     .name = {.domain = "vdc-gtin-1234567890123",
                                                              .message = "c",
                                                              .context = "state|state0|option"}}},
                  .name = {.domain = "vdc-gtin-1234567890123", .message = "state0", .context = "state"}},
              dsModel::DeviceState{
                  .id = dsModel::DeviceStateId{"stateX"},
                  .options = std::vector<dsModel::DeviceStateOption>{dsModel::DeviceStateOption{
                      .id = "a",
                      .name = {.domain = "vdc-gtin-1234567890123", .message = "a", .context = "state|stateX|option"}}},
                  .name = {.domain = "vdc-gtin-1234567890123", .message = "stateX", .context = "state"}}});
    }

    SECTION("deviceStatesStatus") {
      CHECK(deviceInfo.deviceStatesStatus() == std::vector<dsModel::DeviceStateStatus>{dsModel::DeviceStateStatus{
                                                   .id = dsModel::DeviceStateId{"operation"}, .value = "ready"}});
    }
  }

  SECTION("device:missing, db:present") {
    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, missingQuery, presentDb, nullptr, gtin, locale, json, system);

    SECTION("addSpec") {
      deviceInfo.addSpec();
      CHECK(
          // no values except gtin
          json.successJSON() ==
          R"({"result":{"spec":{)"
          R"("dsDeviceGTIN":{"title":"vdc-common,vdc-gtin-1234567890123|spec|dsDeviceGTIN","tags":"settings:5","value":"1234567890123"},)"
          R"("vendorId":{"title":"vdc-common,vdc-gtin-1234567890123|spec|vendorId","tags":"invisible","value":""},)"
          R"("hardwareGuid":{"title":"vdc-common,vdc-gtin-1234567890123|spec|hardwareGuid","tags":"settings:4","value":""},)"
          R"("hardwareModelGuid":{"title":"vdc-common,vdc-gtin-1234567890123|spec|hardwareModelGuid","tags":"invisible","value":""},)"
          R"("descriptionsGroup":{"title":"vdc-common,vdc-gtin-1234567890123|spec|descriptionsGroup","tags":"invisible","value":""},)"
          R"("descriptionsClass":{"title":"vdc-common,vdc-gtin-1234567890123|spec|descriptionsClass","tags":"invisible","value":""},)"
          R"("name":{"title":"vdc-common,vdc-gtin-1234567890123|spec|name","tags":"overview:1;settings:1","value":""},)"
          R"("model":{"title":"vdc-common,vdc-gtin-1234567890123|spec|model","tags":"overview:2;settings:2","value":""},)"
          R"("modelVersion":{"title":"vdc-common,vdc-gtin-1234567890123|spec|modelVersion","tags":"invisible","value":""},)"
          R"("vendorName":{"title":"vdc-common,vdc-gtin-1234567890123|spec|vendorName","tags":"overview:3;settings:3","value":""},)"
          R"("noteId":{"title":"vdc-common,vdc-gtin-1234567890123|spec|noteId","tags":"foo","value":"vdc-common,vdc-gtin-1234567890123|spec/noteId/value|note"})"
          R"(}},"ok":true})");
    }

    SECTION("addStateDescriptions") {
      deviceInfo.addStateDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"stateDescriptions":{)"
          R"("state0":{"title":"vdc-common,vdc-gtin-1234567890123|state|state0","tags":"tag0","options":{"a":"vdc-common,vdc-gtin-1234567890123|state/state0/enum|a","b":"vdc-common,vdc-gtin-1234567890123|state/state0/enum|b"}},)"
          R"("state1":{"title":"vdc-common,vdc-gtin-1234567890123|state|state1","tags":"tag1","options":{"a":"vdc-common,vdc-gtin-1234567890123|state/state1/enum|a"}})"
          R"(}},"ok":true})");
    }

    SECTION("addEventDescriptions") {
      deviceInfo.addEventDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"eventDescriptions":{)"
          R"("event0":{"title":"vdc-common,vdc-gtin-1234567890123|event|event0","category":"category0","accessLevel":"service"},)"
          R"("event1":{"title":"vdc-common,vdc-gtin-1234567890123|event|event1","category":"category1"})"
          R"(}},"ok":true})");
    }

    SECTION("addPropertyDescriptions") {
      deviceInfo.addPropertyDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"propertyDescriptions":{)"
          R"("integer0":{"title":"vdc-common,vdc-gtin-1234567890123|property|integer0","tags":"tag0","type":"numeric","min":"-100","max":"100","resolution":"1","siunit":"m","default":"1"},)"
          R"("numeric0":{"title":"vdc-common,vdc-gtin-1234567890123|property|numeric0","tags":"tag0","type":"numeric","min":"-100","max":"100","resolution":"1","siunit":"m","default":"1"},)"
          R"("numeric1":{"title":"vdc-common,vdc-gtin-1234567890123|property|numeric1","tags":"tag1","type":"numeric.duration","min":"-100.0","max":"100.0","resolution":"0.1","siunit":"s","default":"1.1"},)"
          R"("numeric2":{"title":"vdc-common,vdc-gtin-1234567890123|property|numeric2","tags":"tag2","type":"numeric.duration","min":"-100.0","max":"100.0","resolution":"0.1","siunit":"s","default":"1.1"},)"
          R"("string0":{"title":"vdc-common,vdc-gtin-1234567890123|property|string0","tags":"tag0","type":"string","default":"hello"},)"
          R"("string1":{"title":"vdc-common,vdc-gtin-1234567890123|property|string1","tags":"tag1","type":"string.foo","default":"hello2"},)"
          R"("string2":{"title":"vdc-common,vdc-gtin-1234567890123|property|string2","tags":"tag2","type":"string.foo","default":"hello3"},)"
          R"("enum0":{"title":"vdc-common,vdc-gtin-1234567890123|property|enum0","tags":"tag0","type":"enumeration","default":"a",)"
          R"("options":{"a":"vdc-common,vdc-gtin-1234567890123|property/enum0/enum|a","b":"vdc-common,vdc-gtin-1234567890123|property/enum0/enum|b"}},)"
          R"("bool0":{"title":"vdc-common,vdc-gtin-1234567890123|property|bool0","tags":"tag0","type":"numeric.boolean","min":"-","max":"-","resolution":"-","siunit":"-","default":"true"})"
          R"(}},"ok":true})");
    }

    SECTION("addSensorDescriptions") {
      deviceInfo.addSensorDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"sensorDescriptions":{)"
          R"("sensor0":{"title":"vdc-common,vdc-gtin-1234567890123|sensor|sensor0","tags":"tag0","type":"numeric","min":"-10","max":"10","resolution":"1","siunit":"m","default":"2","dsIndex":0,"dsType":4},)"
          R"("sensor1":{"title":"vdc-common,vdc-gtin-1234567890123|sensor|sensor1","tags":"tag1","type":"numeric","min":"-10","max":"10","resolution":"1","siunit":"m","default":"2","dsIndex":1,"dsType":4})"
          R"(}},"ok":true})");
    }

    SECTION("addActionDescriptions") {
      deviceInfo.addActionDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"actionDescriptions":{)"
          R"("action0":{"title":"vdc-common,vdc-gtin-1234567890123|action|action0","category":"category0","tags":"invalidForCustomActions","params":{)"
          R"("param0":{"title":"vdc-common,vdc-gtin-1234567890123|action/action0/param|param0","tags":"tag0","type":"enumeration","default":"",)"
          R"("options":{"a":"vdc-common,vdc-gtin-1234567890123|action/action0/param/param0/enum|a",)"
          R"("b":"vdc-common,vdc-gtin-1234567890123|action/action0/param/param0/enum|b"},"optional":true}}},)"
          R"("action1":{"title":"vdc-common,vdc-gtin-1234567890123|action|action1","category":"category1","params":{}})"
          R"(}},"ok":true})");
    }

    SECTION("addStandardAction") {
      deviceInfo.addStandardActions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"standardActions":{)"
          R"("std.action0":{"title":"vdc-common,vdc-gtin-1234567890123|standardAction|action0","action":"action0","params":{"a":"A","b":"B"}})"
          R"(}},"ok":true})");
    }

    // device:missing db:present
    SECTION("addOperational") {
      deviceInfo.addOperational();
      CHECK(json.successJSON() ==
            R"({"result":{"outputs":{},"states":{},"properties":{},"sensors":{}},"ok":true})");
    }

    SECTION("addCustomActions") {
      deviceInfo.addCustomActions();
      CHECK(json.successJSON() ==
            R"({"result":{},"ok":true})");
    }

    SECTION("addDynamicDeviceActions") {
      deviceInfo.addDynamicDeviceActions();
      CHECK(json.successJSON() ==
            R"({"result":{},"ok":true})");
    }
  }

  SECTION("device:present, db:present") {
    DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, presentQuery, presentDb, &vdcDevice, gtin, locale, json, system);
    SECTION("addSpec") {
      deviceInfo.addSpec();
      CHECK(
          // translations, tags, values
          json.successJSON() ==
          R"({"result":{"spec":{)"
          R"("dsDeviceGTIN":{"title":")" TEST_DOMAIN
          R"(|spec|dsDeviceGTIN","tags":"settings:5","value":"1234567890123"},)"
          R"("vendorId":{"title":")" TEST_DOMAIN R"(|spec|vendorId","tags":"invisible","value":"x-vendorId"},)"
          R"("hardwareGuid":{"title":")" TEST_DOMAIN
          R"(|spec|hardwareGuid","tags":"settings:4","value":"x-hardwareGuid"},)"
          R"("hardwareModelGuid":{"title":")" TEST_DOMAIN
          R"(|spec|hardwareModelGuid","tags":"invisible","value":"x-hardwareModelGuid"},)"
          R"("descriptionsGroup":{"title":")" TEST_DOMAIN
          R"(|spec|descriptionsGroup","tags":"invisible","value":"x-descriptionsGroup"},)"
          R"("descriptionsClass":{"title":")" TEST_DOMAIN
          R"(|spec|descriptionsClass","tags":"invisible","value":"x-descriptionsClass"},)"
          R"("name":{"title":")" TEST_DOMAIN R"(|spec|name","tags":"overview:1;settings:1","value":"dev::name"},)"
          R"("model":{"title":")" TEST_DOMAIN R"(|spec|model","tags":"overview:2;settings:2","value":"x-model"},)"
          R"("modelVersion":{"title":")" TEST_DOMAIN
          R"(|spec|modelVersion","tags":"invisible","value":"x-modelVersion"},)"
          R"("vendorName":{"title":")" TEST_DOMAIN
          R"(|spec|vendorName","tags":"overview:3;settings:3","value":"x-vendorName"},)"
          R"("noteId":{"title":")" TEST_DOMAIN R"(|spec|noteId","tags":"foo","value":")" TEST_DOMAIN
          R"(|spec/noteId/value|note"})"
          R"(}},"ok":true})");
    }

    SECTION("addStateDescriptions") {
      deviceInfo.addStateDescriptions();
      CHECK(json.successJSON() ==
            R"({"result":{"stateDescriptions":{)"
            R"("state0":{"title":")" TEST_DOMAIN R"(|state|state0","tags":"tag0","options":{"a":")" TEST_DOMAIN
            R"(|state/state0/enum|a","c":")" TEST_DOMAIN R"(|state/state0/enum|c"}},)"
            R"("stateX":{"title":")" TEST_DOMAIN R"(|state|stateX","tags":"","options":{"a":")" TEST_DOMAIN
            R"(|state/stateX/enum|a"}})"
            R"(}},"ok":true})");
    }

    SECTION("addEventDescriptions") {
      deviceInfo.addEventDescriptions();
      CHECK(json.successJSON() == R"({"result":{"eventDescriptions":{)"
                                  R"("event0":{"title":")" TEST_DOMAIN
                                  R"(|event|event0","category":"category0","accessLevel":"service"},)"
                                  R"("eventx":{"title":")" TEST_DOMAIN R"(|event|eventx"})"
                                  R"(}},"ok":true})");
    }

    SECTION("addPropertyDescriptions") {
      deviceInfo.addPropertyDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"propertyDescriptions":{)"
          R"("integer0":{"title":")" TEST_DOMAIN
          R"(|property|integer0","tags":"tag0","type":"numeric","min":"-888","max":"888","resolution":"2","siunit":"s","default":"8"},)"
          R"("integer9":{"title":")" TEST_DOMAIN
          R"(|property|integer9","tags":"","type":"numeric","min":"-666","max":"666","resolution":"2","siunit":"","default":"6","readonly":true},)"
          R"("numeric0":{"title":")" TEST_DOMAIN
          R"(|property|numeric0","tags":"tag0","type":"numeric","min":"-666","max":"666","resolution":"2","siunit":"s","default":"6"},)"
          R"("numeric1":{"title":")" TEST_DOMAIN
          R"(|property|numeric1","tags":"tag1","type":"numeric.duration","min":"-666.6","max":"666.6","resolution":"0.2","siunit":"s","default":"6.6"},)"
          R"("numeric2":{"title":")" TEST_DOMAIN
          R"(|property|numeric2","tags":"tag2","type":"string","default":"a","readonly":true},)"
          R"("string0":{"title":")" TEST_DOMAIN R"(|property|string0","tags":"tag0","type":"string","default":"hi"},)"
          R"("string1":{"title":")" TEST_DOMAIN
          R"(|property|string1","tags":"tag1","type":"string.foo","default":"hi"},)"
          R"("string2":{"title":")" TEST_DOMAIN
          R"(|property|string2","tags":"tag2","type":"numeric","min":"","max":"","resolution":"","siunit":"","default":"6"},)"
          R"("enum0":{"title":")" TEST_DOMAIN
          R"(|property|enum0","tags":"tag0","type":"enumeration","default":"c","options":{"a":")" TEST_DOMAIN
          R"(|property/enum0/enum|a","c":")" TEST_DOMAIN R"(|property/enum0/enum|c"}})"
          R"(}},"ok":true})");
    }

    SECTION("addSensorDescriptions") {
      deviceInfo.addSensorDescriptions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"sensorDescriptions":{)"
          R"("sensor0":{"title":")" TEST_DOMAIN
          R"(|sensor|sensor0","tags":"tag0","type":"numeric","min":"-6.6","max":"6.6","resolution":"0.1","siunit":"m","default":"2","dsIndex":0,"dsType":4},)"
          R"("sensor1x":{"title":")" TEST_DOMAIN
          R"(|sensor|sensor1x","tags":"","type":"numeric","min":"-6.6","max":"6.6","resolution":"0.1","siunit":"","default":"","dsIndex":1,"dsType":5})"
          R"(}},"ok":true})");
    }

    SECTION("addActionDescriptions") {
      deviceInfo.addActionDescriptions();
      CHECK(json.successJSON() ==
            R"({"result":{"actionDescriptions":{)"
            R"("action0":{"title":")" TEST_DOMAIN R"(|action|action0","category":"category0",)"
            R"("tags":"invalidForCustomActions",)"
            R"("params":{"param0":{"title":")" TEST_DOMAIN
            R"(|action/action0/param|param0","tags":"tag0","type":"enumeration","default":"",)"
            R"("options":{"a":")" TEST_DOMAIN R"(|action/action0/param/param0/enum|a",)"
            R"("c":")" TEST_DOMAIN R"(|action/action0/param/param0/enum|c"},"optional":true},)"
            R"("param9":{"title":")" TEST_DOMAIN
            R"(|action/action0/param|param9","tags":"","type":"string","default":""})"
            R"(}},)"
            R"("action9":{"title":")" TEST_DOMAIN R"(|action|action9","params":{}})"
            R"(}},"ok":true})");
    }

    SECTION("addStandardAction") {
      deviceInfo.addStandardActions();
      CHECK(json.successJSON() ==
            R"({"result":{"standardActions":{)"
            R"("std.action0":{"title":")" TEST_DOMAIN
            R"(|standardAction|action0","action":"action0","params":{"a":"A","b":"B"}})"
            R"(}},"ok":true})");
    }

    SECTION("addOperational") {
      deviceInfo.addOperational();
      CHECK(json.successJSON() ==
            R"({"result":{)"
            R"("outputs":{},"states":{)"
            R"("operation":{"title":")" TEST_DOMAIN
            R"(|state|operation","value":"ready","age":1.011742,"displayValue":")" TEST_DOMAIN
            R"(|state/operation/enum|ready"})"
            R"(},"properties":{)"
            R"("currentTemperature":{"title":")" TEST_DOMAIN R"(|property|currentTemperature","value":25.0},)"
            R"("defaultkeepwarmtime":{"title":")" TEST_DOMAIN R"(|property|defaultkeepwarmtime","value":0.0})"
            R"(},"sensors":{)"
            R"("consumption":{"title":")" TEST_DOMAIN
            R"(|sensor|consumption","value":0.2,"timestamp":"2017-09-18T10:37:00.777Z"},)"
            R"("windowOpen":{"title":")" TEST_DOMAIN
            R"(|sensor|windowOpen","value":true,"timestamp":"2017-09-18T10:37:00.777Z"})"
            R"(}},"ok":true})");
    }

    // device:present, db:present
    SECTION("addCustomActions") {
      deviceInfo.addCustomActions();
      CHECK(
          json.successJSON() ==
          R"({"result":{"customActions":{"custom.action1":{"action":"action1","title":"title1","params":{"a":"A","b":"B"}}}},"ok":true})");
    }

    SECTION("addDynamicDeviceActions") {
      deviceInfo.addDynamicDeviceActions();
      CHECK(json.successJSON() ==
            R"({"result":{"deviceActions":{"dynamic.action2":{"action":"action2","title":"title2","params":{"a":"A","b":"B"}}}},"ok":true})");
    }
  }
#undef TEST_DOMAIN
}

TEST_CASE("dssDeviceInfoPowerLineDevice", TAGS) {
  std::string gtin = "4290046000010";
  ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(trompeloeil::_, trompeloeil::_, trompeloeil::_, trompeloeil::_,
                               trompeloeil::_, trompeloeil::_, trompeloeil::_))
      .LR_SIDE_EFFECT(*_7 = (char *)gtin.c_str())
      .RETURN(0);

  JSONWriter json;
  auto&& locale = ds::locale::test::makeLocale();
  dsuid_t dsuid;
  dsuid_from_string("302ed89f43f0000000000ec00009d22600", &dsuid);
  Device device(dsuid, NULL);
  device.setName("geKm200-x");
  device.setIsPresent(true);
  device.setIsConnected(true);
  device.setPartiallyFromSpec([&] {
    DeviceSpec result{};
    result.FunctionID = 0x1000;
    result.ProductID = 0x00C8;
    result.VendorID = 1;
    result.revisionId = 0x123;
    return result;
  }());

  StructureQueryBusInterfaceMock query;
  VdcDbMock db;
  auto system = ds::SystemMock{};

  DeviceInfo deviceInfo(DeviceInfo::Prefer::DEVICE, query, db, &device, "", locale, json, system);
  SECTION("addSpec") {
#define TEST_DOMAIN "vdc-common,vdc-gtin-0"
    deviceInfo.addSpec();
    CHECK(
        // no translations, no tags, but all values
        json.successJSON() ==
        R"({"result":{"spec":{)"
        R"("dsDeviceGTIN":{"title":")" TEST_DOMAIN R"(|spec|dsDeviceGTIN","tags":"","value":"0"},)"        //
        R"("vendorId":{"title":")" TEST_DOMAIN R"(|spec|vendorId","tags":"","value":"1"},)"                //
        R"("vendorName":{"title":")" TEST_DOMAIN R"(|spec|vendorName","tags":"","value":"aizo GmbH"},)"    //
        R"("hardwareGuid":{"title":")" TEST_DOMAIN R"(|spec|hardwareGuid","tags":"","value":"0009d226"},)" //
        R"("hardwareModelGuid":{"title":")" TEST_DOMAIN
        R"(|spec|hardwareModelGuid","tags":"","value":"4290046000010"},)"                                    //
        R"("descriptionsGroup":{"title":")" TEST_DOMAIN R"(|spec|descriptionsGroup","tags":"","value":""},)" //
        R"("descriptionsClass":{"title":")" TEST_DOMAIN R"(|spec|descriptionsClass","tags":"","value":""},)" //
        R"("name":{"title":")" TEST_DOMAIN R"(|spec|name","tags":"","value":"geKm200-x"},)"                  //
        R"("model":{"title":")" TEST_DOMAIN R"(|spec|model","tags":"","value":"GE-KM200"},)"                 //
        R"("modelVersion":{"title":")" TEST_DOMAIN R"(|spec|modelVersion","tags":"","value":"1.2.3"})"       //
        R"(}},"ok":true})");
#undef TEST_DOMAIN
  }

  SECTION("addStandardAction") {
    deviceInfo.addStandardActions();
    CHECK(json.successJSON() ==
          R"({"result":{"standardActions":{)"                                                     //
          R"("std.turnOff":{"title":"vdc-common,vdc-gtin-0|standardAction|turnOff"},)"            //
          R"("std.turnOn":{"title":"vdc-common,vdc-gtin-0|standardAction|turnOn"},)"              //
          R"("std.turnOffSlowly":{"title":"vdc-common,vdc-gtin-0|standardAction|turnOffSlowly"})" //
          R"(}},"ok":true})");
  }
  SECTION("addStandardAction") {
    CHECK(deviceInfo.scenarios() ==
          std::vector<dsModel::Scenario>{
              {.id = encodeId(
                   DeviceScenarioId{.device = device.id(), .action = dsModel::DeviceScenarioActionId{"std.turnOff"}}),
               .attributes =
                   dsModel::DeviceScenarioAttributes{
                       .name = {.domain = "dss-mainline", .message = "Turn off", .context = "deviceAction"},
                       .dsDevice = device.id(),
                       .actionId = dsModel::DeviceScenarioActionId{"std.turnOff"}}},
              {.id = encodeId(DeviceScenarioId{.device = device.id(),
                                               .action = dsModel::DeviceScenarioActionId{"std.turnOffSlowly"}}),
               .attributes =
                   dsModel::DeviceScenarioAttributes{
                       .name = {.domain = "dss-mainline", .message = "Turn off slowly", .context = "deviceAction"},
                       .dsDevice = device.id(),
                       .actionId = dsModel::DeviceScenarioActionId{"std.turnOffSlowly"}}},
              {.id = encodeId(
                   DeviceScenarioId{.device = device.id(), .action = dsModel::DeviceScenarioActionId{"std.turnOn"}}),
               .attributes =
                   dsModel::DeviceScenarioAttributes{
                       .name = {.domain = "dss-mainline", .message = "Turn on", .context = "deviceAction"},
                       .dsDevice = device.id(),
                       .actionId = dsModel::DeviceScenarioActionId{"std.turnOn"}}},
          });
  }
}

} // namespace
