#include "rest.h"
#include <boost/algorithm/string/predicate.hpp>

DS_JSON_DEFINE_STRUCT((dsApi), HttpArgs);
DS_JSON_DEFINE_STRUCT((dsApi), HttpRequest);
DS_JSON_DEFINE_STRUCT((dsApi), HttpResult);
DS_JSON_DEFINE_STRUCT((dsApi), Error);
DS_JSON_DEFINE_STRUCT((dsApi), Document);
DS_JSON_DEFINE_STRUCT((dsApi), PatchOperation);
DS_JSON_DEFINE_STRUCT((dsApi), PatchRequest);

namespace dsApi {

HttpResult noContentResult() {
  return {.status = 204};
}

HttpResult errorResult(int status, std::string&& message) {
  return {
      .status = status,
      .content = ds::toJsonFragment(dsApi::Document{
          .data = {}, .error = {.type = "error", .title = "error", .status = status, .detail = std::move(message)}})};
}

HttpResult errorResult(const std::exception& e) {
  return errorResult(500, e.what());
}

static std::string eraseTrailingSpace(std::string&& x) {
  // ds::context() tends to produce trailing space
  if (boost::ends_with(x, " ")) {
    x.erase(x.size() - 1, 1);
  }
  return std::move(x);
}

HttpResult resourceNotFoundResult() {
  return errorResult(404, eraseTrailingSpace(ds::str("Resource not found. ", ds::hideArg(ds::logContext()))));
}

HttpResult methodNotAllowedResult(HttpMethod method) {
  return errorResult(
      406, eraseTrailingSpace(ds::str("Method not allowed. method:", method, " ", ds::hideArg(ds::logContext()))));
}

HttpResult notImplementedResult() {
  return errorResult(501, eraseTrailingSpace(ds::str("Not implemented. ", ds::hideArg(ds::logContext()))));
}

HttpResult cancelledResult() {
  return errorResult(500, eraseTrailingSpace(ds::str("Cancelled. ", ds::hideArg(ds::logContext()))));
}

Error error(const HttpResult& result) {
  const auto status = result.status;
  DS_REQUIRE(result.status / 100 != 2, "Failed to access error due to success http status.", status);
  return ds::fromJson<dsApi::Document>(result.content).error;
}

} // namespace dsApi
