#pragma once

#include "rest-fwd.h"

#include <ds/json.h>
#include <ds/promise-fwd.h>

// Upper case to match http specification (intentionally violates ds style guideline)
DS_REFLECTED_ENUM((dsApi), HttpMethod, (UNKNOWN)(GET)(POST)(PUT)(DELETE)(PATCH))

DS_REFLECTED_ENUM((dsApi), PatchOperationType, (unknown)(remove)(add)(replace)(test))

namespace dsApi {

using HttpHeaders = std::map<std::string, std::vector<std::string>>;

struct HttpArgs {
  HttpMethod method{};
  ds::JsonFragment content{};
  std::string query{};
  HttpHeaders headers{};
};

using RouteHttpArgs = std::map<std::string, std::string>;
struct HttpRequest {
  std::string path{};
  HttpArgs httpArgs{};
  RouteHttpArgs routeArgs{};
};

struct HttpResult {
  int status{};
  ds::JsonFragment content{};
  HttpHeaders headers{};

  bool isSuccess() const { return status / 100 == 2; }
};

/// Http callee invoking requests
class AsyncCallee {
public:
  virtual ~AsyncCallee() = default;
  virtual void asyncInvoke(HttpRequest&&, std::unique_ptr<ds::Promise<HttpResult>>&&) = 0;
};

struct Error {
  std::string type{};
  std::string title{};
  int status{};
  std::string detail{};
};

struct Document {
  ds::JsonFragment data{};
  Error error{};
};

template <class Data>
Document document(Data&& data) {
  return Document{.data = ds::toJsonFragment(std::forward<Data>(data))};
}

inline HttpResult okResult() {
  return {.status = 200};
}

template <class Data>
HttpResult okResult(Data&& data) {
  return {.status = 200, .content = ds::toJsonFragment(document(data))};
}

HttpResult noContentResult();

HttpResult errorResult(int status, std::string&& message);
HttpResult errorResult(const std::exception& e);
HttpResult resourceNotFoundResult();
HttpResult methodNotAllowedResult(HttpMethod);
HttpResult notImplementedResult();
HttpResult cancelledResult();

/// return casted data from Document
template <class Data>
Data data(const Document& document) {
  return ds::fromJson<Data>(document.data);
}

/// return casted data from HttpResult
template <class Data>
Data data(const HttpResult& result) {
  const auto status = result.status;
  DS_REQUIRE(result.status / 100 == 2, "Failed to access data due to error http status.", status);
  return data<Data>(ds::fromJson<dsApi::Document>(result.content));
}

/// return casted data from HttpResult
template <class Data>
Data data(const HttpArgs& httpArgs) {
  return data<Data>(ds::fromJson<dsApi::Document>(httpArgs.content));
}

/// return error from HttpResult
Error error(const HttpResult& result);

/// Rest resource template according to
/// https://git.digitalstrom.org/digitalstrom-documentation/public-api-documentation/blob/master/guidelines/RESTApiGuidelines.md#document-structure
template <class Id, class Attributes>
struct ResourceData {
  Id id{};
  std::string type{};
  ds::DateTime lastChanged{};
  Attributes attributes{};
};

template <class Id, class Attributes, class Included>
struct ResourceDataWithIncluded {
  Id id{};
  std::string type{};
  Attributes attributes{};
  Included included{};
};

#define DS_JSON_DECLARE_API_RESOURCE_DATA(NAMESPACES, TYPE) \
  DS_JSON_DECLARE_STRUCT(NAMESPACES, TYPE, (id)(type)(lastChanged)(attributes));
#define DS_JSON_DECLARE_API_RESOURCE_DATA_WITH_INCLUDED(NAMESPACES, TYPE) \
  DS_JSON_DECLARE_STRUCT(NAMESPACES, TYPE, (id)(type)(attributes)(included));

template <class Id, class... Alternatives>
struct VariantResourceData {
  Id id{};
  std::variant<std::monostate, Alternatives...> attributes{};
};

#define DS_JSON_DECLARE_API_VARIANT_RESOURCE_DATA(NAMESPACES, TYPE, NAME_VISITOR) \
  DS_JSON_DECLARE_STRUCT(NAMESPACES, TYPE, (id)(attributes));                     \
  template <>                                                                     \
  struct DSJsonVariant<&DS_REFLECT_FULL_TYPE(NAMESPACES, TYPE)::attributes> {     \
    static constexpr auto mode = ds::JsonVariantMode::alternativeInSiblingMember; \
    static constexpr auto alternative = "type";                                   \
    static constexpr auto& nameVisitor = NAME_VISITOR;                            \
  };

/// Patch operation according to rfc6902
struct PatchOperation {
  PatchOperationType op{};
  std::string path{};
  ds::JsonFragment value{};
};

using PatchOperations = std::vector<PatchOperation>;

struct PatchRequest {
  PatchOperation operation{};
  RouteHttpArgs routeArgs{};
};

} // namespace dsApi

DS_JSON_DECLARE_STRUCT((dsApi), HttpArgs, (method)(content)(query)(headers));
DS_JSON_DECLARE_STRUCT((dsApi), HttpRequest, (path)(httpArgs)(routeArgs));
DS_JSON_DECLARE_STRUCT((dsApi), HttpResult, (status)(content)(headers));
DS_JSON_DECLARE_STRUCT((dsApi), Error, (type)(title)(status)(detail));
DS_JSON_DECLARE_STRUCT((dsApi), Document, (data)(error));
DS_JSON_DECLARE_STRUCT((dsApi), PatchOperation, (op)(path)(value));
DS_JSON_DECLARE_STRUCT((dsApi), PatchRequest, (operation)(routeArgs));
