/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "route-segments.h"
#include <ds/catch/catch.h>
#include <ds/log.h>

namespace {

static const char *TAGS = "[dsApiRouteSegments][dsApi]";

SCENARIO("dsApiRouteSegments", TAGS) {

  const auto literalA = dsApi::RouteSegment{dsApi::LiteralRouteSegment{"a"}};
  const auto literalB = dsApi::RouteSegment{dsApi::LiteralRouteSegment{"b"}};
  const auto literalBb = dsApi::RouteSegment{dsApi::LiteralRouteSegment{"bb"}};
  const auto literalC = dsApi::RouteSegment{dsApi::LiteralRouteSegment{"c"}};
  const auto parameterA = dsApi::RouteSegment{dsApi::ArgRouteSegment{"a"}};
  const auto parameterB = dsApi::RouteSegment{dsApi::ArgRouteSegment{"b"}};
  const auto wildcardC = dsApi::RouteSegment{dsApi::WildcardArgRouteSegment{"c"}};

  // CHECK's in following section are using double braces in order to compile with Catch2 on gcc 7.4.0
  using Segments = std::vector<dsApi::RouteSegment>;
  CHECK((dsApi::RouteSegmentsView("").segments == Segments()));
  CHECK((dsApi::RouteSegmentsView("/a").segments == (Segments{literalA})));
  CHECK((dsApi::RouteSegmentsView("/a/bb").segments == (Segments{literalA, literalBb})));
  CHECK((dsApi::RouteSegmentsView("/{a}").segments == (Segments{parameterA})));
  CHECK((dsApi::RouteSegmentsView("/{a}/b").segments == (Segments{parameterA, literalB})));
  CHECK((dsApi::RouteSegmentsView("/a/{b}").segments == (Segments{literalA, parameterB})));
  CHECK((dsApi::RouteSegmentsView("/{a}/{b}").segments == (Segments{parameterA, parameterB})));
  CHECK((dsApi::RouteSegmentsView("/a/{b}/c").segments == (Segments{literalA, parameterB, literalC})));
  CHECK((dsApi::RouteSegmentsView("/a/{*c}").segments == (Segments{literalA, wildcardC})));

  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/"), "Route segment must not be empty.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("//"), "Route segment must not be empty.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/a/"), "Route segment must not be empty.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/a//"), "Route segment must not be empty.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/{a}/"), "Route segment must not be empty.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/{*a}/"), "Wildcard route parameter must be the last route segment.");
  REQUIRE_THROWS_FIND(dsApi::RouteSegmentsView("/{*a}//"), "Wildcard route parameter must be the last route segment.");
}

SCENARIO("dsApiMatchRoute", TAGS) {
  GIVEN("empty route") {
    auto route = "";
    CHECK(dsApi::matchRoute(route, "/"));
    CHECK(dsApi::matchRoute(route, "/") == (dsApi::MatchRouteResult{.matched = true}));
    CHECK(dsApi::matchRoute(route, "") == (dsApi::MatchRouteResult{.matched = true}));

    CHECK(!dsApi::matchRoute(route, "/a"));
    CHECK(!dsApi::matchRoute(route, "//"));
  }

  GIVEN("/a route") {
    auto route = "/a";
    CHECK(dsApi::matchRoute(route, "/a") == (dsApi::MatchRouteResult{.matched = true}));
    CHECK(dsApi::matchRoute(route, "/a/") == (dsApi::MatchRouteResult{.matched = true}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "a"));
    CHECK(!dsApi::matchRoute(route, "a/"));
    CHECK(!dsApi::matchRoute(route, "/"));
    CHECK(!dsApi::matchRoute(route, "/b"));
    CHECK(!dsApi::matchRoute(route, "/aa"));
    CHECK(!dsApi::matchRoute(route, "/a/b"));
  }

  GIVEN("/a/bb route") {
    auto route = "/a/bb";
    CHECK(dsApi::matchRoute(route, "/a/bb") == (dsApi::MatchRouteResult{.matched = true}));
    CHECK(dsApi::matchRoute(route, "/a/bb/") == (dsApi::MatchRouteResult{.matched = true}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "a/bb"));
    CHECK(!dsApi::matchRoute(route, "a/bb/"));
    CHECK(!dsApi::matchRoute(route, "/a/b"));
    CHECK(!dsApi::matchRoute(route, "/a/bbb"));
    CHECK(!dsApi::matchRoute(route, "/a/bb/c"));
  }

  GIVEN("/{a} route") {
    auto route = "/{a}";
    CHECK(dsApi::matchRoute(route, "/") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}}}));
    CHECK(dsApi::matchRoute(route, "/a") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}}}));
    CHECK(dsApi::matchRoute(route, "/A") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "A"}}}));
    CHECK(dsApi::matchRoute(route, "/hi") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "hi"}}}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "/A/bbb"));
    CHECK(!dsApi::matchRoute(route, "/A/a/bb"));
  }

  GIVEN("/{a}/bb route") {
    auto route = "/{a}/bb";
    CHECK(dsApi::matchRoute(route, "/a/bb") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}}}));
    CHECK(dsApi::matchRoute(route, "/A/bb") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "A"}}}));
    CHECK(dsApi::matchRoute(route, "/hi/bb/") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "hi"}}}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "/"));
    CHECK(!dsApi::matchRoute(route, "/A"));
    CHECK(!dsApi::matchRoute(route, "/A/"));
    CHECK(!dsApi::matchRoute(route, "/A/bbb"));
    CHECK(!dsApi::matchRoute(route, "/A/a/bb"));
  }

  GIVEN("/{b}/{a} route") {
    auto route = "/{b}/{a}";
    CHECK(dsApi::matchRoute(route, "//") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}, {"b", ""}}}));
    CHECK(dsApi::matchRoute(route, "/b/") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}, {"b", "b"}}}));
    CHECK(dsApi::matchRoute(route, "/b/a") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}, {"b", "b"}}}));
    CHECK(dsApi::matchRoute(route, "/BB/AA") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "AA"}, {"b", "BB"}}}));
    CHECK(dsApi::matchRoute(route, "/hi/all/") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "all"}, {"b", "hi"}}}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "/"));
    CHECK(!dsApi::matchRoute(route, "/B"));
    CHECK(!dsApi::matchRoute(route, "/B/A/c"));
  }

  GIVEN("/{*a} route") {
    auto route = "/{*a}";
    CHECK(dsApi::matchRoute(route, "") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}}}));
    CHECK(dsApi::matchRoute(route, "/") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "/"}}}));
    CHECK(dsApi::matchRoute(route, "/a") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "/a"}}}));
    CHECK(dsApi::matchRoute(route, "/a/b/cc") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "/a/b/cc"}}}));
  }

  GIVEN("/a/{*b} route") {
    auto route = "/a/{*b}";
    CHECK(dsApi::matchRoute(route, "/a") == (dsApi::MatchRouteResult{.matched = true, .args = {{"b", ""}}}));
    CHECK(dsApi::matchRoute(route, "/a/") == (dsApi::MatchRouteResult{.matched = true, .args = {{"b", "/"}}}));
    CHECK(dsApi::matchRoute(route, "/a/b/cc") == (dsApi::MatchRouteResult{.matched = true, .args = {{"b", "/b/cc"}}}));

    CHECK(!dsApi::matchRoute(route, ""));
    CHECK(!dsApi::matchRoute(route, "/"));
    CHECK(!dsApi::matchRoute(route, "//"));
  }

  GIVEN("/{a}/{*b} route") {
    auto route = "/{a}/{*b}";
    CHECK(dsApi::matchRoute(route, "/") == (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}, {"b", ""}}}));
    CHECK(dsApi::matchRoute(route, "/a") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}, {"b", ""}}}));
    CHECK(dsApi::matchRoute(route, "//") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", ""}, {"b", "/"}}}));
    CHECK(dsApi::matchRoute(route, "/a/") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}, {"b", "/"}}}));
    CHECK(dsApi::matchRoute(route, "/a/b/cc") ==
          (dsApi::MatchRouteResult{.matched = true, .args = {{"a", "a"}, {"b", "/b/cc"}}}));

    CHECK(!dsApi::matchRoute(route, ""));
  }
}

} // namespace
