/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "service.h"
#include <ds/catch/catch.h>
#include <ds/log.h>
#include <boost/algorithm/string/predicate.hpp>
#include <trompeloeil.hpp>

using namespace trompeloeil;

namespace {
static const char *TAGS = "[dsApiService][dsApi]";

class Method {
public:
  MAKE_MOCK0(call, void());
};

struct Args {
  int a;
  int b;
};

class ArgsMethod {
public:
  MAKE_MOCK1(call, void(Args &&));
};

SCENARIO("dsApiService", TAGS) {
  dsApi::Service service;
  auto requireResourceNotFound = [&](const dsApi::HttpResult &x) {
    INFO(x);
    CHECK(x.status == 404);
    CHECK(boost::starts_with(dsApi::error(x).detail, "Resource not found."));
  };
  auto requireReplyMethodNotAllowed = [&](const dsApi::HttpResult &x) {
    INFO(x);
    CHECK(x.status == 406);
    CHECK(boost::starts_with(dsApi::error(x).detail, "Method not allowed."));
  };
  auto requireError = [&](const dsApi::HttpResult &x, auto &&message) {
    INFO(x);
    CHECK(x.status == 500);
    CHECK(boost::starts_with(dsApi::error(x).detail, message));
  };

  GIVEN("/echo route") {
    service.routes.push_back({.pattern = "/echo", .invoke = [&](auto &&context) {
                                CHECK(context.httpArgs.method == dsApi::HttpMethod::POST);
                                auto data = dsApi::data<std::string>(context.httpArgs);
                                CHECK(data == "test");
                                return dsApi::okResult(data);
                              }});

    THEN("request to /echo succeeds") {
      CHECK(dsApi::data<std::string>(service.invoke(
                {.path = "/echo",
                 {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(dsApi::document("test"))}})) ==
            "test");
    }

    THEN("request to non-matching fails") {
      SECTION("non-matching prefix") {
        requireResourceNotFound(service.invoke({.path = "", {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/", {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/ech", {.method = dsApi::HttpMethod::POST}}));
      }

      SECTION("matching prefix with suffix") {
        requireResourceNotFound(service.invoke({.path = "/echo/foo", {.method = dsApi::HttpMethod::POST}}));
        requireResourceNotFound(service.invoke({.path = "/echo2", {.method = dsApi::HttpMethod::POST}}));
      }
    }
  }

  GIVEN("/wildcard route returning remaining path") {
    auto makeArgs = [] {
      return dsApi::HttpArgs{.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment("test")};
    };
    service.routes.push_back({.pattern = "/wildcard/{*path}", .invoke = [&](auto &&context) {
                                CHECK(context.httpArgs.method == dsApi::HttpMethod::POST);
                                CHECK(ds::fromJson<std::string>(context.httpArgs.content) == "test");
                                return dsApi::okResult(context.routeArgs["path"]);
                              }});

    THEN("request to /wildcard") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard", .httpArgs = makeArgs()})) == "");
      ;
    }

    THEN("request to /wildcard/ succeeds") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard/", .httpArgs = makeArgs()})) == "/");
      ;
    }

    THEN("request to /wildcard/* succeeds") {
      CHECK(dsApi::data<std::string>(service.invoke({.path = "/wildcard/1", .httpArgs = makeArgs()})) == "/1");
    }

    THEN("request to non-matching prefix fails") {
      requireResourceNotFound(service.invoke({.path = "", .httpArgs = makeArgs()}));
      requireResourceNotFound(service.invoke({.path = "/", .httpArgs = makeArgs()}));
      requireResourceNotFound(service.invoke({.path = "/wildcar", .httpArgs = makeArgs()}));
    }

    THEN("request to matching prefix without '/'") {
      requireResourceNotFound(service.invoke({.path = "/wildcards", .httpArgs = makeArgs()}));
    }
  }

  GIVEN("/fail route throwing exception") {
    service.routes.push_back(
        {.pattern = "/fail", .invoke = [&](auto &&...) -> dsApi::HttpResult { DS_FAIL_REQUIRE("Test failure."); }});

    THEN("request to /fail fails") {
      requireError(service.invoke({.path = "/fail", .httpArgs = {.method = dsApi::HttpMethod::POST}}), "Test failure.");
    }
  }

  GIVEN("method /test()") {
    Method method;
    service.routes.push_back(dsApi::makeMethodRoute("/test", [&]() { method.call(); }));
    auto call = [&](dsApi::HttpMethod method = dsApi::HttpMethod::POST) {
      return service.invoke({.path = "/test", .httpArgs = {.method = method}});
    };
    THEN("POST succeeds with 204 and on content") {
      REQUIRE_CALL(method, call());
      CHECK(call() == dsApi::noContentResult());
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call()).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure"));
      requireError(call(), "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(call(method));
      }
    }
  }

  GIVEN("method /{b}/test(args)") {
    ArgsMethod method;
    service.routes.push_back(
        dsApi::makeMethodRoute<Args>("/{b}/test", [&](auto &&args) { method.call(std::move(args)); }));

    THEN("POST passes arguments from path and content") {
      REQUIRE_CALL(method, call(Args{.a = 7, .b = 8}));
      service.invoke({.path = "/8/test",
                      .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{.a = 7})}});
    }

    THEN("POST route arg has higher priority than content arg") {
      REQUIRE_CALL(method, call(Args{.a = 7, .b = 8}));
      service.invoke(
          {.path = "/8/test",
           .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{.a = 7, .b = 7})}});
    }

    THEN("POST passes default arguments for missing argument") {
      REQUIRE_CALL(method, call(Args{.a = {}, .b = 8}));
      service.invoke(
          {.path = "/8/test", .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{})}});
    }

    THEN("POST accepts empty content as empty struct") {
      REQUIRE_CALL(method, call(Args{.a = {}, .b = 8}));
      service.invoke({.path = "/8/test", .httpArgs = {.method = dsApi::HttpMethod::POST}});
    }

    THEN("POST succeeds with 204 and on content") {
      REQUIRE_CALL(method, call(_));
      auto result = service.invoke(
          {.path = "/8/test", .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{})}});
      CHECK(result == dsApi::noContentResult());
    }

    THEN("POST returns error on exception") {
      REQUIRE_CALL(method, call(_)).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("Test failure"));
      requireError(
          service.invoke({.path = "/7/test",
                          .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(Args{})}}),
          "Test failure");
    }

    THEN("Other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/7/test", dsApi::HttpArgs{.method = method}}));
      }
    }
  }

  GIVEN("document /test with GET") {
    auto called = 0;
    service.routes.push_back(dsApi::makeDocumentRoute("/test", {.methods = {dsApi::makeGet([&]() {
                                                                  called++;
                                                                  return 7;
                                                                })}}));

    THEN("GET method returns document") {
      CHECK(dsApi::data<int>(service.invoke({.path = "/test", dsApi::HttpArgs{.method = dsApi::HttpMethod::GET}})) ==
            7);
      CHECK(called == 1);
    }

    THEN("other methods fail") {
      for (auto method :
           {dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::PATCH, dsApi::HttpMethod::POST}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/test", dsApi::HttpArgs{.method = method}}));
      }
      CHECK(called == 0);
    }
  }

  GIVEN("document /test with GET with argument") {
    service.routes.push_back(
        dsApi::makeDocumentRoute("/test/{a}/{b}", {.methods = {dsApi::makeGet<Args>([&](auto &&args) {
                                                     CHECK(args.a == 1);
                                                     CHECK(args.b == 2);
                                                     return 7;
                                                   })}}));

    THEN("GET method returns document and parses arguments") {
      CHECK(dsApi::data<int>(
                service.invoke({.path = "/test/1/2", dsApi::HttpArgs{.method = dsApi::HttpMethod::GET}})) == 7);
    }
  }

  GIVEN("document /{a} with PATCH routes") {
    auto called = 0;
    service.routes.push_back(dsApi::makeDocumentRoute(
        "/{a}", {.methods = {},
                 .patchRoutes = {{.pattern = "/{b}",
                                  .operations = {dsApi::makeReplacePatch<Args, double>([&](auto &&args, auto &&value) {
                                    CHECK(args.a == 1);
                                    CHECK(args.b == 2);
                                    CHECK(value == 3);
                                    called++;
                                  })}},
                                  {.pattern = "/x/{b}",
                                  .operations = {dsApi::makeReplacePatch<Args, double>([&](auto &&args, auto &&value) {
                                    CHECK(args.a == 1);
                                    CHECK(args.b == 3);
                                    CHECK(value == 5);
                                    called++;
                                  })}}}
                }));

    THEN("replace PATCH is routed") {
      dsApi::PatchOperations patchOperations;
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/x/3", .value = ds::toJsonFragment(5.0)});
      CHECK(service.invoke({.path = "/1",
                            dsApi::HttpArgs{.method = dsApi::HttpMethod::PATCH,
                                            .content = ds::toJsonFragment(patchOperations)}}) ==
            dsApi::noContentResult());
      CHECK(called == 2);
    }

    THEN("other PATCH operations fail") {
      for (auto op :
           {dsApi::PatchOperationType::add, dsApi::PatchOperationType::remove, dsApi::PatchOperationType::test}) {
        dsApi::PatchOperations patchOperations;
        patchOperations.push_back({.op = op, .path = "/2", .value = ds::toJsonFragment(3.0)});
        requireError(service.invoke({.path = "/1",
                                     dsApi::HttpArgs{.method = dsApi::HttpMethod::PATCH,
                                                     .content = ds::toJsonFragment(patchOperations)}}),
                     "Path operation not supported.");
      }
      CHECK(!called);
    }

    THEN("replace PATCH is routed more times till first error") {
      dsApi::PatchOperations patchOperations;
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back({.op = dsApi::PatchOperationType::add, .path = "/2", .value = ds::toJsonFragment(3.0)});
      patchOperations.push_back(
          {.op = dsApi::PatchOperationType::replace, .path = "/2", .value = ds::toJsonFragment(3.0)});
      requireError(service.invoke({.path = "/1",
                                   dsApi::HttpArgs{.method = dsApi::HttpMethod::PATCH,
                                                   .content = ds::toJsonFragment(patchOperations)}}),
                   "Path operation not supported.");
      CHECK(called == 2);
    }

    THEN("other http methods fail") {
      for (auto method :
           {dsApi::HttpMethod::GET, dsApi::HttpMethod::PUT, dsApi::HttpMethod::DELETE, dsApi::HttpMethod::POST}) {
        requireReplyMethodNotAllowed(service.invoke({.path = "/7", dsApi::HttpArgs{.method = method}}));
        CHECK(!called);
      }
    }
  }
}

} // namespace

DS_REFLECT_ANONYMOUS_STRUCT(Args, (a)(b));
