/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include "service-fwd.h"

#include <functional>
#include "rest.h"

namespace dsApi {

/// Route describing one resource
struct Route {
  /// Each route pattern segment is delimited by '/' character.
  /// When a segment is enclosed in braces ('{' and '}'), the segment is an argument.
  ///
  /// Example route patterns:
  /// - "/apartments/{apartmentId}
  /// - "/apartments/{apartmentId}/status
  /// - "/apartments/{apartmentId}/zones
  /// - "/apartments/{apartmentId}/zones/{zoneId}
  /// - "/apartments/{apartmentId}/zones/{zoneId}/status
  /// - "/apartments/{apartmentId}/{*path}
  std::string pattern{};

  /// @param args arguments specified by pattern parsed from url path
  /// @param HttpArgs http payload
  std::function<HttpResult(HttpRequest&&)> invoke{};
};

/// make route from function `void ()`
template <class F, class = decltype(std::declval<F>()())>
Route makeMethodRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern),
               .invoke = [f = std::forward<F>(f)](auto&& context){auto method = context.httpArgs.method;
    if (method != HttpMethod::POST) {
      return methodNotAllowedResult(method);
    }

f();
return noContentResult();
}
}
;
}

/// make route from function `void (Args &&args)` with automatic conversion from route and http content to Args
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
Route makeMethodRoute(std::string&& pattern, F&& f) {
  return Route{.pattern = std::move(pattern),
               .invoke = [f = std::forward<F>(f)](auto&& context){auto method = context.httpArgs.method;
    if (method != HttpMethod::POST) {
      return methodNotAllowedResult(method);
    }
// TODO(someday): this could be more effective without conversion to std::map<string, json>
auto args = ds::fromJson<std::map<std::string, ds::JsonFragment>>(context.httpArgs.content);
for (const auto& routeArg : context.routeArgs) {
  args[routeArg.first] = ds::toJsonFragment(routeArg.second);
}
f(ds::fromJson<Args>(ds::toJsonFragment(args)));
return noContentResult();
}
}
;
}

/// Service routing requests according to available routes
struct Service {
  Service() = default;
  Service(std::vector<Route>&& routes) : routes{std::move(routes)} {}
  virtual ~Service() = default; // be usable as base class in std::unique_ptr

  std::vector<Route> routes{};

  // const because it does not modify Service.
  // `const Service&` is for invocation and reflection purposes.
  // `Service&` allows to also edit routes.
  HttpResult invoke(HttpRequest&&) const;
};

/// make wildcard route forwarding all subpaths to service
Route makeServiceRoute(std::string&& pattern, dsApi::Service& service);

/// make GET function returning document
template <class F, class = decltype(std::declval<F>()())>
auto makeGet(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET, [f = std::forward<F>(f)](auto&&) {
        return HttpResult{.status = 200,
                          .content = ds::toJsonFragment(dsApi::Document{.data = ds::toJsonFragment(f())})};
      }};
}

/// make GET function with arguments returning document
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
auto makeGet(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::GET, [f = std::forward<F>(f)](auto&& context) {
        auto args = ds::fromJson<std::map<std::string, ds::JsonFragment>>(context.httpArgs.content);
        for (const auto& routeArg : context.routeArgs) {
          args[routeArg.first] = ds::toJsonFragment(routeArg.second);
        }
        return HttpResult{.status = 200,
                          .content = ds::toJsonFragment(dsApi::Document{
                              .data = ds::toJsonFragment(f(ds::fromJson<Args>(ds::toJsonFragment(args))))})};
      }};
}

/// make DELETE function deleting document
template <class Args, class F, class = decltype(std::declval<F>()(std::declval<Args>()))>
auto makeDelete(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::DELETE, [f = std::forward<F>(f)](auto&& context) {
        // TODO(someday): this could be more effective without conversion to std::map<string, json>
        auto args = ds::fromJson<std::map<std::string, ds::JsonFragment>>(context.httpArgs.content);
        for (const auto& routeArg : context.routeArgs) {
          args[routeArg.first] = ds::toJsonFragment(routeArg.second);
        }
        f(ds::fromJson<Args>(ds::toJsonFragment(args)));
        return noContentResult();
      }};
}

/// make POST function creating document and returning id of created resource as std::string.
template <class Data, class F>
auto makeCreate(F&& f) {
  return std::pair<HttpMethod, std::function<HttpResult(HttpRequest &&)>>{
      HttpMethod::POST, [f = std::forward<F>(f)](auto&& context) {
        auto data = dsApi::data<Data>(context.httpArgs);
        DS_REQUIRE(static_cast<std::string>(data.id).empty(), "Resource id must not be passed to POST method");
        auto id = static_cast<std::string>(f(std::move(data)));
        return HttpResult{.status = 201, .content = {}, .headers = {{"Location", {std::move(id)}}}};
      }};
}

/// Route describing one path operation.
///
/// Functions `makeReplacePatch`, etc
/// are intended to make instances of PatchRoute in application code.
struct PatchRoute {
  std::string pattern{};

  std::map<PatchOperationType, std::function<void(PatchRequest&&)>> operations{};
};

/// make replace route from function `void (Args &&args)`
/// with automatic conversion from route and http content to Args
template <class Args, class Value, class F,
          class = decltype(std::declval<F>()(std::declval<Args>(), std::declval<Value>()))>
auto makeReplacePatch(F&& f) {
  return std::pair<PatchOperationType, std::function<void(PatchRequest &&)>>{
      dsApi::PatchOperationType::replace, [f = std::forward<F>(f)](auto&& context) {
        f(ds::fromJson<Args>(ds::toJsonFragment(context.routeArgs)), ds::fromJson<Value>(context.operation.value));
      }};
}

struct DocumentMethods {
  std::map<HttpMethod, std::function<HttpResult(HttpRequest&&)>> methods{};

  /// default PATCH method implementation
  /// interates over all patch operations and
  /// invokes matching patch route
  std::vector<PatchRoute> patchRoutes{};
};

Route makeDocumentRoute(std::string&& pattern, DocumentMethods&& document);

/// Makes test service suitable for dsApi integration tests
std::unique_ptr<Service> makeTestService();

} // namespace dsApi
