/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

///
/// Command line utility to like connecting to ds connector
/// similar to netcat, wscat.
///

#include <ds-api/rest.h>
#include <ds-api/service.h>
#include <ds-connector/protocol.h>
#include <ds/conf.h>
#include <ds/promise.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/asio/posix/stream_descriptor.hpp>
#include <boost/asio/read_until.hpp>
#include <boost/asio/streambuf.hpp>
#include <functional>
#include <iostream>
#include "client-bus.h"
#include "ds-connector/backend-stream-client.h"
#include "http-server.h"
#include "subscriptions.h"
#include "wss-client.h"

static ds::LogChannel dsLogChannel("dsConnectorCat");

namespace {

class ClientBusHook final : public dsConnector::ClientBusHook {
  dsApi::Service& m_service;
  const dsConnector::Authenticate m_authenticate;

public:
  ClientBusHook(dsApi::Service& service, dsConnector::Authenticate&& authenticate)
      : m_service{service}, m_authenticate{std::move(authenticate)} {}

  dsConnector::Authenticate authenticate() override { return {m_authenticate}; }

  void asyncInvoke(dsApi::HttpRequest&& request, std::unique_ptr<ds::Promise<dsApi::HttpResult>>&& promise) override {
    const auto path = request.path;
    DS_CONTEXT(path);
    promise->value = m_service.invoke(std::move(request));
  }
};

class StdinReader {
public:
  StdinReader(boost::asio::io_service& io, dsConnector::ClientBus& bus) : m_io{io}, m_bus{bus} { asyncRead(); }

private:
  boost::asio::io_service& m_io;
  dsConnector::ClientBus& m_bus;
  boost::asio::posix::stream_descriptor m_stream{m_io, ::dup(STDIN_FILENO)};
  boost::asio::streambuf m_buffer{};

  void asyncRead() {
    boost::asio::async_read_until(m_stream, m_buffer, '\n', [this](boost::system::error_code ec, std::size_t n) {
      DS_REQUIRE(!ec, "Failed to read from stdin.", ec);

      std::string line;
      std::istream(&m_buffer) >> line;
      m_buffer.consume(n);

      try {
        if (!line.empty()) {
          DS_INFO(line);
          m_bus.event(ds::fromJson<dsApi::Event>(line));
        }
      } catch (const std::exception& e) {
        DS_ERROR("Failed to send message.", e);
      }

      asyncRead();
    });
  }
};
} // namespace

int main(int argc, char** argv) {
  try {
    puts("DSS - backend connector client");
    puts("");
    puts("Client will automatically connect (and reconnect) to server after 10s.");
    puts("Messages typed to stdin are sent to server.");
    puts("");

    auto& system = ds::system();
    auto& conf = system.conf();
    const auto clientId = [&]() -> std::string {
      if (const auto x = getenv("CLIENT")) {
        return x;
      }
      if (const auto x = conf.readOr<std::string>("dssId", {}); !x.empty()) {
        return x;
      }
      DS_FAIL_REQUIRE("Required CLIENT variable not passed.");
    }();
    const auto secret = [&]() -> std::string {
      if (const auto x = getenv("SECRET")) {
        return x;
      }
      if (const auto x = conf.readOr<std::string>("ospToken", {}); !x.empty()) {
        return x;
      }
      DS_FAIL_REQUIRE("Required SECRET variable not passed.");
    }();
    const auto apartmentId = [&]() -> std::string {
      if (const auto x = getenv("APARTMENT")) {
        return x;
      }
      if (const auto x = conf.readOr<std::string>("apartmentId", {}); !x.empty()) {
        return x;
      }
      DS_FAIL_REQUIRE("Required APARTMENT variable not passed.");
    }();

    boost::asio::io_service io;
    boost::asio::ssl::context sslContext{boost::asio::ssl::context::sslv23_client};
    auto streamClient = dsConnector::makeBackendStreamClient(io, system, sslContext);
    DS_NOTICE(clientId, apartmentId, streamClient);

    auto testService = dsApi::makeTestService();
    dsApi::Service service;
    service.routes.push_back(dsApi::makeServiceRoute(ds::str("/v1/apartment/test"), *testService));
    puts("RPC enpoints:");
    for (const auto& route : service.routes) {
      for (const auto& testRoute : testService->routes) {
        std::cout << route.pattern << testRoute.pattern << "\n";
      }
    }

    ClientBusHook hook{service, {.client = {clientId}, .secret = {secret}, .apartmentId = {apartmentId}}};
    auto subscriptions = dsConnector::makeSubscriptions(io);
    auto clientBus = dsConnector::makeClientBus(io, conf, *streamClient, hook, *subscriptions);
    clientBus->setActive(true);
    StdinReader stdinReader(io, *clientBus);

    auto httpServer = dsConnector::makeHttpServer(io, {boost::asio::ip::make_address("127.0.0.1"), 8089}, *clientBus);
    DS_NOTICE("Http reverse proxy to backend at http://localhost:8089");

    io.run();
    DS_FAIL_REQUIRE("Event loop shall never exit.");
  } catch (const std::exception& e) {
    DS_ERROR(e);
    puts("");
    puts("Usage:");
    puts("  CLIENT={clientId} \\");
    puts("  SECRET={secret} \\");
    puts("  APARTMENT={apartmentId} \\");
    puts("  ds-connector-cat");
  }
}
