/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "client-bus.h"
#include <ds-api/rest.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/ds-signal-mock.h>
#include <ds/log.h>
#include <ds/promise.h>
#include <boost/algorithm/string/predicate.hpp>
#include "client-bus-mock.h"
#include "stream-mock.h"
#include "subscriptions-mock.h"

using namespace trompeloeil;
using namespace std::literals;

namespace {

static const char* TAGS = "[dsConnectorClientBus][dsConnector]";

struct HttpResultFulfilled {
public:
  MAKE_MOCK1(call, void(std::optional<dsApi::HttpResult>));
};

SCENARIO("dsConnectorClientBus", TAGS) {
  auto clock = ds::ClockMocker{};
  auto loop = ds::asio::TestLoop{};
  ds::ConfMock dsConf{};
  dsConnector::StreamClientMock streamClient{};
  const auto reconnectSchedule = std::chrono::seconds{10};
  const auto keepAlivePeriod = std::chrono::seconds{120};
  const auto deadPeriod = keepAlivePeriod * 2;
  dsConnector::ClientBusHookMock busHook;
  ALLOW_CALL(busHook, authenticate())
      .RETURN(dsConnector::Authenticate{.client = "dss", .secret = "dssSecret", .apartmentId = "dssApartmentId"});

  dsConnector::SubscriptionsMock subscriptions;
  // stream must be destroyed when clientBus is out of scope
  std::unique_ptr<trompeloeil::lifetime_monitor> streamDestruction;
  auto clientBus = dsConnector::makeClientBus(loop.io, dsConf, streamClient, busHook, subscriptions);
  auto isConnectedChanged = dsConnector::BusIsConnectedChangedMock{loop.io, clientBus->isConnectedChanged()};
  clientBus->setActive(true);

  const auto error = boost::system::errc::make_error_code(boost::system::errc::operation_canceled);
  WHEN("just created") {
    THEN("is disconnected") { REQUIRE(!clientBus->isConnected()); }
    THEN("discards published events") { clientBus->event({}); }
  }

  THEN("after reconnectSchedule, connects") {
    clock.advance(reconnectSchedule);

    auto uniqueStream = std::make_unique<deathwatched<dsConnector::StreamMock>>();
    auto& stream = *uniqueStream;
    streamDestruction = NAMED_REQUIRE_DESTRUCTION(stream);

    // when old stream is deleted, we connect 2nd one
    auto uniqueStream2 = std::make_unique<dsConnector::StreamMock>();
    auto& stream2 = *uniqueStream2;

    dsConnector::StreamClient::ConnectHandler pendingConnectHandler;
    REQUIRE_CALL(streamClient, asyncConnect(_))
        .LR_SIDE_EFFECT(loop.stop(); pendingConnectHandler = std::move(_1))
        .LR_RETURN(std::move(uniqueStream));
    loop.run();

    WHEN("connect fails") {
      pendingConnectHandler(boost::system::errc::make_error_code(boost::system::errc::operation_not_supported));

      THEN("after reconnectSchedule, connects") {
        clock.advance(reconnectSchedule);
        REQUIRE_CALL(streamClient, asyncConnect(_)).LR_SIDE_EFFECT(loop.stop()).LR_RETURN(std::move(uniqueStream2));
        loop.run();
      }
    }
    WHEN("connect succeeds") {
      // unhandled reads stops io
      auto pendingReadHandler = dsConnector::Stream::ReadHandler{};
      ALLOW_CALL(stream, asyncRead(_)).LR_SIDE_EFFECT(loop.stop(); pendingReadHandler = std::move(_1));
      REQUIRE_CALL(
          stream,
          asyncWrite(R"({"authenticate":{"client":"dss","secret":"dssSecret","apartmentId":"dssApartmentId"}})", _))
          .LR_SIDE_EFFECT(_2({}));

      pendingConnectHandler({});

      auto disconnectsAndReconnects = [&] {
        auto streamDestructionNow = std::move(streamDestruction);
        if (!streamDestructionNow->is_satisfied()) { // disconnect can be asynchronous
          stream.setOnDestroyed([&] { loop.stop(); });
          loop.run();
        }
        REQUIRE(!clientBus->isConnected());
        clock.advance(reconnectSchedule);
        REQUIRE_CALL(streamClient, asyncConnect(_))
            .LR_SIDE_EFFECT(pendingConnectHandler = std::move(_1); loop.stop())
            .LR_RETURN(std::move(uniqueStream2));
        loop.run();
      };

      WHEN("read fails") {
        pendingReadHandler(error, {});
        THEN("disconnects and reconnects") { disconnectsAndReconnects(); }
      }

      WHEN("authenticated times out") {
        clock.advance(deadPeriod);
        THEN("disconnectsAndReconnects") { disconnectsAndReconnects(); }
      }

      WHEN("receives unknown message") {
        pendingReadHandler({}, R"({"unknown":{}})");
        THEN("disconnectsAndReconnects") { disconnectsAndReconnects(); }
      }

      WHEN("authenticated") {
        {
          REQUIRE_CALL(isConnectedChanged, called(true)).LR_SIDE_EFFECT(loop.stop());
          pendingReadHandler({}, R"({"authenticated":{}})");
          loop.run();
        }

        THEN("is connected") { REQUIRE(clientBus->isConnected()); }

        WHEN("read fails") {
          {
            REQUIRE_CALL(isConnectedChanged, called(false)).LR_SIDE_EFFECT(loop.stop());
            pendingReadHandler(error, {});
            loop.run();
          }
          THEN("disconnects and reconnects") { disconnectsAndReconnects(); }
        }

        WHEN("ignores unknown messages") {
          pendingReadHandler({}, R"({"unknown":{}})");
          REQUIRE(clientBus->isConnected());
        }

        THEN("responds to pings") {
          REQUIRE_CALL(stream, asyncWrite(R"({"pong":{"id":"4"}})", _)).LR_SIDE_EFFECT(_2({}));
          pendingReadHandler({}, R"({"ping":{"id":"4"}})");
        }

        THEN("sends ping within keepAlivePeriod") {
          clock.advance(keepAlivePeriod);
          REQUIRE_CALL(stream, asyncWrite(R"({"ping":{"id":"1"}})", _)).LR_SIDE_EFFECT(loop.stop(); _2({}));
          loop.run();

          WHEN("pong is not received within keepAlivePeriod") {
            {
              REQUIRE_CALL(isConnectedChanged, called(false)).LR_SIDE_EFFECT(loop.stop());
              clock.advance(keepAlivePeriod);
              loop.run();
            }
            THEN("disconnects and reconnects") { disconnectsAndReconnects(); }
          }

          WHEN("after keepAlivePeriod / 2") {
            clock.advance(keepAlivePeriod / 2);
            loop.postStopRun();

            THEN("is connected") { REQUIRE(clientBus->isConnected()); }

            WHEN("wrong pong is received") {
              pendingReadHandler({}, R"({"pong":{"id":"7"}})");
              THEN("is connected") { REQUIRE(clientBus->isConnected()); }
            }

            WHEN("pong is received") {
              pendingReadHandler({}, ds::str(R"({"pong":{"id":"1"}})"));
              THEN("sends another ping within keepAlivePeriod") {
                clock.advance(keepAlivePeriod * 2);
                REQUIRE_CALL(stream, asyncWrite(R"({"ping":{"id":"2"}})", _)).LR_SIDE_EFFECT(loop.stop(); _2({}));
                loop.run();
                REQUIRE(clientBus->isConnected());
              }
            }
          }
        }

        THEN("writes events") {
          REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a"}})", _)).LR_SIDE_EFFECT(_2({}); loop.stop(););
          clientBus->event({.type = dsApi::EventType{"a"}});
          loop.run();
        }

        WHEN("write fails") {
          {
            REQUIRE_CALL(stream, asyncWrite(_, _)).LR_SIDE_EFFECT(_2(error));
            REQUIRE_CALL(isConnectedChanged, called(false)).LR_SIDE_EFFECT(loop.stop());
            clientBus->event({.type = dsApi::EventType{"a"}});
            loop.run();
          }
          THEN("disconnects and reconnects") { disconnectsAndReconnects(); }
        }

        THEN("writes are queued when stream write is blocking") {
          dsConnector::Stream::WriteHandler pendingWriteHandler{};
          REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"a"}})", _))
              .LR_SIDE_EFFECT(loop.postStop(); pendingWriteHandler = std::move(_2));
          clientBus->event({.type = dsApi::EventType{"a"}});
          clientBus->event({.type = dsApi::EventType{"b"}});
          clientBus->event({.type = dsApi::EventType{"c"}});
          loop.run();

          THEN("queued writes are written when write is unblocked") {
            REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"c"}})", _)).LR_SIDE_EFFECT(_2({}););
            REQUIRE_CALL(stream, asyncWrite(R"({"event":{"type":"b"}})", _)).LR_SIDE_EFFECT(_2({}););
            pendingWriteHandler({});
          }

          THEN("queued writes are discarded on disconnect and not written to new connection") {
            {
              REQUIRE_CALL(isConnectedChanged, called(false));
              pendingWriteHandler(error);
              disconnectsAndReconnects();
            }

            ALLOW_CALL(stream2, asyncRead(_)).LR_SIDE_EFFECT(loop.stop(); pendingReadHandler = std::move(_1));
            {
              REQUIRE_CALL(
                  stream2,
                  asyncWrite(R"({"authenticate":{"client":"dss","secret":"dssSecret","apartmentId":"dssApartmentId"}})",
                             _))
                  .LR_SIDE_EFFECT(_2({}));
              pendingConnectHandler({});
            }
            {
              REQUIRE_CALL(isConnectedChanged, called(true)).LR_SIDE_EFFECT(loop.stop());
              pendingReadHandler({}, R"({"authenticated":{}})");
              loop.run();
            }
            {
              REQUIRE_CALL(stream2, asyncWrite(R"({"event":{"type":"z"}})", _)).LR_SIDE_EFFECT(_2({}); loop.stop());
              clientBus->event({.type = dsApi::EventType{"z"}});
              loop.run();
            }
          }
        }

        WHEN("request is read") {
          std::unique_ptr<ds::Promise<dsApi::HttpResult>> promise;
          REQUIRE_CALL(busHook, asyncInvoke(_, _))
              .LR_SIDE_EFFECT(
                  CHECK(_1 == dsApi::HttpRequest{.path = "/a", .httpArgs = {.method = dsApi::HttpMethod::GET}});
                  promise = std::move(_2););
          pendingReadHandler({}, R"({"request":{"id":"id-1","path":"/a","args":{"method":"GET"}}})");

          THEN("reply is written") {
            REQUIRE_CALL(stream, asyncWrite(R"({"response":{"id":"id-1","result":{"status":200}}})", _))
                .LR_SIDE_EFFECT(_2({}); loop.stop());
            promise->value = {.status = 200};
            promise.reset();
            loop.run();
          }
        }

        THEN("subscribe is propagated to subscriptions") {
          REQUIRE_CALL(subscriptions, subscribe(_))
              .LR_SIDE_EFFECT(CHECK(_1 == dsConnector::Subscription{.type = dsApi::EventType{"type1"},
                                                                    .expiresAt = ds::Clock::now() + 60s,
                                                                    clientBus.get()});
                              loop.stop());
          pendingReadHandler({}, R"({"subscribe":{"type":"type1", "timeout":60}})");
        }

        THEN("callee invoke with /backend prefix is sent as request") {
          auto httpResultFulfilled = HttpResultFulfilled{};
          auto message = dsConnector::Message{};
          REQUIRE_CALL(stream, asyncWrite(_, _))
              .LR_SIDE_EFFECT(message = ds::fromJson<dsConnector::Message>(std::string{_1}); _2({}); loop.stop(););
          clientBus->asyncInvoke({.path = "/backend/test", .httpArgs = {.method = dsApi::HttpMethod::GET}},
                                 ds::makePromise<dsApi::HttpResult>([&](std::optional<dsApi::HttpResult>&& x) {
                                   httpResultFulfilled.call(std::move(x));
                                 }));
          loop.run();
          REQUIRE(message.request);
          const auto& request = *message.request;
          CHECK(request.id != dsConnector::RequestId{});
          CHECK(request.path == "/test");
          CHECK(request.args == dsApi::HttpArgs{.method = dsApi::HttpMethod::GET});

          THEN("callee is replied on response") {
            REQUIRE_CALL(httpResultFulfilled, call(dsApi::HttpResult{.status = 200}));
            pendingReadHandler(
                {}, ds::str(R"({"response":{"id":")", message.request->id, R"(", "result":{"status":200}}})"));
            SUCCEED();
          }

          THEN("called is replied with cancel on timeout") {
            clock.advance(2min);
            REQUIRE_CALL(httpResultFulfilled, call(std::nullopt)).LR_SIDE_EFFECT(loop.stop(););
            loop.run();
            SUCCEED();
          }

          THEN("called is replied with cancel on disconnect") {
            REQUIRE_CALL(httpResultFulfilled, call(std::nullopt));
            pendingReadHandler(error, {});
            SUCCEED();
          }
        }

        THEN("callee invoke without /backend prefix fails") {
          auto httpResultFulfilled = HttpResultFulfilled{};
          REQUIRE_CALL(httpResultFulfilled, call(_))
              .LR_SIDE_EFFECT(CHECK(_1.value().status == 500);
                              CHECK(boost::contains(ds::fromJson<dsApi::Document>(_1.value().content).error.detail,
                                                    "Paths in backend must start with backendPrefix")));
          clientBus->asyncInvoke({.path = "/test", .httpArgs = {.method = dsApi::HttpMethod::GET}},
                                 ds::makePromise<dsApi::HttpResult>([&](std::optional<dsApi::HttpResult>&& x) {
                                   httpResultFulfilled.call(std::move(x));
                                 }));
          SUCCEED();
        }

        WHEN("setActive(false)") {
          clientBus->setActive(false);
          THEN("is !connected") {
            REQUIRE_CALL(isConnectedChanged, called(false)).LR_SIDE_EFFECT(loop.stop());
            loop.run();
            REQUIRE(!clientBus->isConnected());
          }
        }
      }
    }
  }
}

} // namespace
