/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "client-bus.h"
#include <ds-api/rest.h>
#include <ds/asio/move-only.h>
#include <ds/asio/reconnect-timer.h>
#include <ds/asio/timer.h>
#include <ds/elapsed.h>
#include <ds/json.h>
#include <ds/latency-statistics.h>
#include <ds/promise.h>
#include <ds/random.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/atomic.hpp>
#include <boost/range/numeric.hpp>
#include <deque>
#include "protocol.h"
#include "subscriptions.h"

static ds::LogChannel dsLogChannel("dsConnectorClientBus");
// log channel to support investigation of dss-cloud communication
static ds::LogChannel messages("dsConnectorMessages");
using namespace std::literals;

namespace dsConnector {

namespace {
struct Conf {
  const ds::Conf &dsConf;

  const std::chrono::seconds keepAlivePeriod =
      std::chrono::seconds{dsConf.readOr<int>("dsConnector/clientBus/keepAlivePeriod", 60)};
  const std::chrono::seconds latencyStatisticsPeriod =
      std::chrono::seconds{dsConf.readOr<int>("dsConnector/clientBus/latencyStatisticsPeriod", 3600)};
};

class ClientBusImpl final : public ClientBus {
public:
  ClientBusImpl(boost::asio::io_service &io, ds::Conf &dsConf, StreamClient &streamClient, ClientBusHook &hook,
                Subscriptions &subscriptions)
      : m_io{io}, m_conf{dsConf}, m_streamClient{streamClient}, m_hook{hook}, m_subscriptions{subscriptions} {
    asyncReconnect();
    asyncCalleeInvocationsDeadlineLoop();
  }

  void event(dsApi::Event &&event) override {
    DS_INFO_ENTER(event);
    postEnqueue(Message{
        .authenticate = {},
        .authenticated = {},
        .ping = {},
        .pong = {},
        .request = {},
        .response = {},
        .event = Event{.type = std::move(event.type), .path = std::move(event.path), .data = std::move(event.data)}});
  }

  bool isConnected() const override { return m_isConnected; }
  void setActive(bool isActive) override {
    m_io.dispatch([this, isActive] {
      if (m_isActive == isActive) {
        return;
      }
      DS_NOTICE("Set active.", isActive);
      m_isActive = isActive;
      m_reconnectTimer.resetSchedule();
      asyncReconnect();
    });
  }

  const StreamClient &streamClient() const { return m_streamClient; }

private:
  boost::asio::io_service &m_io;
  Conf m_conf;
  StreamClient &m_streamClient;
  ClientBusHook &m_hook;
  Subscriptions &m_subscriptions;
  std::thread::id m_threadId{std::this_thread::get_id()};
  boost::atomic_bool m_isActive{};
  boost::atomic_bool m_isConnected{};
  ds::asio::ReconnectTimer m_reconnectTimer{m_io};
  std::unique_ptr<Stream> m_stream{};
  ds::asio::Timer m_keepAliveTimer{m_io};
  int m_pingId{};
  ds::Elapsed m_pingElapsed{};
  std::string m_pongId{};
  ds::Elapsed m_uptime{};
  ds::LatencyStatistics m_statisticsMs{};
  ds::Elapsed m_statisticsNotice{};
  std::deque<std::string> m_writeDeque{};
  std::size_t m_writeQueueMaxSize{};
  struct CalleeInvocation {
    std::unique_ptr<ds::Promise<dsApi::HttpResult>> promise;
    ds::Clock::time_point startedAt;
    std::string path;
  };
  std::map<RequestId, CalleeInvocation> m_calleeInvocations;
  ds::asio::Timer m_calleeInvocationsDeadline{m_io};
  const ds::Clock::duration calleeInvocationDeadline = 60s;

  void asyncReconnect() {
    DS_INFO_ENTER("");
    try {
      if (m_isConnected) {
        DS_NOTICE("Disconnected. latency[ms]:", ds::hideArg(m_statisticsMs), "uptime:", ds::hideArg(m_uptime),
                  "maxQueue:", ds::hideArg(m_writeQueueMaxSize));
        m_isConnected = false;
        m_isConnectedChanged.emit(false);
      }
      m_keepAliveTimer.cancel();
      m_stream.reset();
      m_pingId = {};
      m_pongId = ds::str(m_pingId);
      m_statisticsMs.drop();
      m_writeQueueMaxSize = {};
      m_writeDeque.clear();
      m_calleeInvocations.clear();

      if (m_isActive) {
        m_reconnectTimer.asyncDelayedReconnect([this] {
          asyncConnect();
          asyncKeepAlive(); // keep alive timer guards also initial connection setup
        });
      }
    } catch (const std::exception &e) {
      DS_ASSERT("Asynchronous reconnect loop must not fail.");
    }
  }

  void asyncConnect() {
    DS_INFO_ENTER();
    try {
      m_stream = m_streamClient.asyncConnect([this](boost::system::error_code ec) {
        if (ec) {
          DS_ERROR("Connect failed.", ec, *this);
          asyncReconnect();
          return;
        }
        asyncWriteAuthenticate();
      });
    } catch (const std::exception &e) {
      DS_ERROR("Async resolve failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncWriteAuthenticate() {
    DS_INFO_ENTER("");
    try {
      auto message = std::make_shared<std::string>(ds::toJson(Message{.authenticate = m_hook.authenticate()}));
      DS_INFO("Write authenticate", ds::hideArg(*message));
      DS_CHANNEL_INFO(messages, "Write", ds::hideArg(*message));
      m_stream->asyncWrite(*message, [this, message](boost::system::error_code ec) {
        if (ec) {
          DS_ERROR("Authenticate write failed.", ec, *this);
          asyncReconnect();
          return;
        }
        asyncReadAuthenticated();
      });
    } catch (const std::exception &e) {
      DS_ERROR("Async write authenticate failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncReadAuthenticated() {
    DS_INFO_ENTER("");
    try {
      m_stream->asyncRead([this](boost::system::error_code ec, std::string &&message) {
        if (ec) {
          DS_ERROR("Authenticated read failed.", ec.message(), *this);
          asyncReconnect();
          return;
        }
        asyncReadAuthenticatedConsume(std::move(message));
      });
    } catch (const std::exception &e) {
      DS_ERROR("Async authenticated read failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncReadAuthenticatedConsume(std::string &&rawMessage) {
    DS_INFO("Read authenticated", ds::hideArg(rawMessage));
    DS_CHANNEL_INFO(messages, "Read", ds::hideArg(rawMessage));
    try {
      const auto message = ds::fromJson<Message>(rawMessage);
      DS_REQUIRE(message.authenticated, "Authenticated message not received.", rawMessage);
      asyncConnected();
    } catch (const std::exception &e) {
      DS_ERROR("Async authenticated read failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncConnected() {
    DS_NOTICE("Connected.", *this);
    try {
      DS_ASSERT(!m_isConnected, "Invariant failed");
      m_isConnected = true;
      m_uptime.restart();
      m_statisticsNotice.restart();
      m_reconnectTimer.connected();
      m_isConnectedChanged.emit(true);
      asyncRead();
    } catch (const std::exception &e) {
      DS_ERROR("Async connected failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncKeepAlive() {
    DS_DEBUG_ENTER("");
    try {
      m_keepAliveTimer.randomlyExpiresFromNowPercentDown(m_conf.keepAlivePeriod, 10, [this]() { asyncAliveCheck(); });
    } catch (const std::exception &e) {
      DS_ERROR("Async keep alive failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncAliveCheck() {
    DS_INFO_ENTER("");
    try {
      DS_REQUIRE(m_isConnected, "Not connected.");
      DS_REQUIRE(ds::str(m_pingId) == m_pongId, "Pong timeout.", m_pingId, m_pongId);
      enqueue({.authenticate = {}, .authenticated = {}, .ping = Ping{.id = ds::str(++m_pingId)}});
      m_pingElapsed.restart();
      asyncKeepAlive();
    } catch (const std::exception &e) {
      DS_ERROR("Async alive check failed.", e, *this);
      asyncReconnect();
    }
  }

  size_t writeQueueBytes() const {
    return boost::accumulate(m_writeDeque, size_t{}, [](auto sum, const auto &x) { return sum + x.size(); });
  }

  void enqueue(const Message &message) {
    DS_DEBUG_ENTER(m_isConnected, m_writeDeque.size(), writeQueueBytes());
    DS_ASSERT(std::this_thread::get_id() == m_threadId, "Enqueue must be called from m_io thread.");
    if (!isConnected()) {
      DS_INFO("Discarding message in disconnected state.");
      return;
    }
    auto wasEmpty = m_writeDeque.empty();
    m_writeDeque.push_back(ds::toJson(message));
    m_writeQueueMaxSize = std::max(m_writeQueueMaxSize, m_writeDeque.size());
    DS_INFO("Enqueued", ds::hideArg(m_writeDeque.back()));
    if (wasEmpty) {
      DS_DEBUG("Starting write loop.");
      asyncWrite();
    }
  }

  void postEnqueue(Message &&message) {
    m_io.post(ds::asio::MoveOnly{[this, message = std::move(message)]() mutable { enqueue(std::move(message)); }});
  }

  void asyncWrite() {
    DS_DEBUG_ENTER(m_writeDeque.size(), writeQueueBytes());
    try {
      if (m_writeDeque.empty()) {
        DS_DEBUG("Stopping write loop.");
        return;
      }
      DS_DEBUG("Write", ds::hideArg(m_writeDeque.front()));
      DS_CHANNEL_INFO(messages, "Write", ds::hideArg(m_writeDeque.front()));
      m_stream->asyncWrite(m_writeDeque.front(), [this](boost::system::error_code ec) {
        if (ec) {
          DS_ERROR("Write failed.", ec, *this);
          asyncReconnect();
          return;
        }
        asyncWritten();
      });
    } catch (const std::exception &e) {
      DS_ERROR("Write failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncWritten() {
    DS_DEBUG_ENTER(m_writeDeque.size(), writeQueueBytes());
    try {
      m_writeDeque.pop_front();
      asyncWrite();
    } catch (const std::exception &e) {
      DS_ERROR("Write failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncRead() {
    DS_INFO_ENTER("");
    try {
      m_stream->asyncRead([this](boost::system::error_code ec, std::string &&message) {
        if (ec) {
          DS_ERROR("Read failed.", ec, *this);
          asyncReconnect();
          return;
        }
        asyncReadConsume(std::move(message));
      });
    } catch (const std::exception &e) {
      DS_ERROR("Read failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncReadConsume(std::string &&rawMessage) {
    DS_DEBUG_ENTER();
    try {
      DS_INFO("Read", ds::hideArg(rawMessage));
      DS_CHANNEL_INFO(messages, "Read", ds::hideArg(rawMessage));
      auto message = ds::fromJson<Message>(rawMessage);
      if (message.ping) {
        enqueue({.authenticate = {}, .authenticated = {}, .ping = {}, .pong = Pong{.id = message.ping->id}});
      } else if (message.pong) {
        m_pongId = message.pong->id;
        auto latencyMs = std::chrono::duration_cast<std::chrono::milliseconds>(m_pingElapsed.elapsed()).count();
        m_statisticsMs(latencyMs);
        DS_INFO(latencyMs, m_statisticsMs, m_uptime);
        if (m_statisticsNotice.elapsed() > m_conf.latencyStatisticsPeriod) {
          DS_NOTICE("latency[ms]:", ds::hideArg(m_statisticsMs), "uptime:", ds::hideArg(m_uptime),
                    "maxQueue:", ds::hideArg(m_writeQueueMaxSize));
          m_statisticsNotice.restart();
          m_statisticsMs.drop();
          m_writeQueueMaxSize = {};
        }
      } else if (message.request) {
        auto &request = *message.request;
        const auto id = std::move(request.id);
        try {
          m_hook.asyncInvoke(
              dsApi::HttpRequest{.path = std::move(request.path), .httpArgs = std::move(request.args)},
              ds::makePromise<dsApi::HttpResult>([this, id = std::move(id)](std::optional<dsApi::HttpResult> &&result) {
                postEnqueue(
                    Message{.authenticate = {},
                            .authenticated = {},
                            .ping = {},
                            .pong = {},
                            .request = {},
                            .response = Response{.id = std::move(id),
                                                 .result = result ? std::move(*result) : dsApi::cancelledResult()}});
              }));
        } catch (const std::exception &e) {
          DS_ERROR("Failed to process request", e, id, rawMessage);
          // Response (likely a cancelledResult) is sent when promise is destroyed.
          // We could propagate this error into the response instead of cancelledResult
        }
      } else if (message.response) {
        auto &response = *message.response;
        auto it = m_calleeInvocations.find(response.id);
        if (it != m_calleeInvocations.end()) {
          it->second.promise->value = std::move(response.result);
          m_calleeInvocations.erase(it);
        } else {
          DS_WARNING("Unexpected response ", response.id);
          DS_INFO("Unexpected response ", response);
        }
      } else if (message.subscribe) {
        auto &subscribe = *message.subscribe;
        m_subscriptions.subscribe(
            {.type = std::move(subscribe.type), .expiresAt = ds::Clock::now() + subscribe.timeout, .sink = this});
      } else {
        // ignore unknown message
        DS_WARNING("Unknown message.", ds::hideArg(rawMessage));
      }
      asyncRead();
    } catch (const std::exception &e) {
      DS_ERROR("Read consume failed.", e, *this);
      asyncReconnect();
    }
  }

  void asyncInvoke(dsApi::HttpRequest &&request, std::unique_ptr<ds::Promise<dsApi::HttpResult>> &&promise) override {
    try {
      DS_INFO("Async invoke", request);

      // Require and strip /backend prefix in path.
      // This routing may be done in some router later
      auto path = std::move(request.path);
      constexpr auto backendPrefix = "/backend"sv;
      DS_REQUIRE(boost::starts_with(path, backendPrefix), "Paths in backend must start with backendPrefix.", path,
                 backendPrefix);
      path.erase(0, backendPrefix.size());

      const auto id = RequestId{ds::randomId()};
      postEnqueue(Message{.authenticate = {},
                          .authenticated = {},
                          .ping = {},
                          .pong = {},
                          .request = Request{.id = id, .path = path, .args = std::move(request.httpArgs)}});
      m_calleeInvocations[id] = {.promise = std::move(promise), .startedAt = ds::Clock::now(), .path = std::move(path)};
    } catch (const std::exception &e) {
      DS_NOTICE("Async invoke failed", e);
      promise->value = dsApi::errorResult(e);
    }
  }

  void asyncCalleeInvocationsDeadlineLoop() {
    DS_DEBUG_ENTER("");
    m_calleeInvocationsDeadline.randomlyExpiresFromNow(calleeInvocationDeadline / 2, [this]() mutable {
      DS_INFO_ENTER("");
      asyncCalleeInvocationsDeadlineLoop();
      try {
        const auto now = ds::Clock::now();
        for (auto it = m_calleeInvocations.begin(); it != m_calleeInvocations.end();) {
          if (now - it->second.startedAt > calleeInvocationDeadline) {
            DS_WARNING("Cancelling old invocation.", it->first, it->second.path);
            it = m_calleeInvocations.erase(it);
          } else {
            ++it;
          }
        }
      } catch (const std::exception &e) {
        DS_ERROR("Cancelling old invacations failed.", e);
      }
    });
  }
};

std::ostream &operator<<(std::ostream &ostream, const ClientBusImpl &x) {
  return ostream << x.streamClient();
}

} // namespace

std::unique_ptr<ClientBus> makeClientBus(boost::asio::io_service &io, ds::Conf &dsConf, StreamClient &streamClient,
                                         ClientBusHook &hook, Subscriptions &subscriptions) {
  return std::make_unique<ClientBusImpl>(io, dsConf, streamClient, hook, subscriptions);
}

} // namespace dsConnector
