#include "http-server.h"
#include <ds-api/rest.h>
#include <ds/asio/move-only.h>
#include <ds/log.h>
#include <ds/promise.h>
#include <boost/asio.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/version.hpp>
#include <chrono>
#include <cstdlib>
#include <memory>
#include <string>

static ds::LogChannel dsLogChannel("dsConnectorHttpServer");

using namespace std::literals;
namespace ip = boost::asio::ip;      // from <boost/asio.hpp>
using tcp = boost::asio::ip::tcp;    // from <boost/asio.hpp>
namespace http = boost::beast::http; // from <boost/beast/http.hpp>
using error_code = boost::beast::error_code;

namespace dsConnector {

class HttpConnectionShared : public std::enable_shared_from_this<HttpConnectionShared> {
private:
  tcp::socket m_socket;
  dsApi::AsyncCallee& m_router;
  bool m_replied{false};

public:
  HttpConnectionShared(tcp::socket&& socket, dsApi::AsyncCallee& router)
      : m_socket(std::move(socket)), m_router(router) {
    DS_INFO_ENTER("");
  }

  void asyncRun() {
    DS_INFO_ENTER("");
    asyncRead();
  }

private:
  // Asynchronously receive a complete request message.
  void asyncRead() {
    DS_DEBUG_ENTER("");
    auto buffer = std::make_shared<boost::beast::flat_buffer>(4096);     // TODO(newer asio): make_unique
    auto request = std::make_shared<http::request<http::string_body>>(); // TODO(newer asio): make_unique

    http::async_read(
        m_socket, *buffer, *request,
        [self = shared_from_this(), buffer = std::move(buffer), request = std::move(request)](error_code ec,
                                                                                              std::size_t) {
          DS_INFO_ENTER(request->target());
          DS_DEBUG_ENTER(request->target(), request->method_string(), request->body());
          if (ec) {
            DS_ERROR(ec);
            return;
          }
          const auto version = request->version();
          try {
            auto path = request->target();
            auto query = decltype(path){};
            {
              const auto pos = path.find('?');
              if (pos != std::string_view::npos) {
                query = path.substr(pos + 1);
                path = path.substr(0, pos);
              }
            }
            self->m_router.asyncInvoke(
                dsApi::HttpRequest{
                    .path = std::string{path},
                    .httpArgs = {.method = ds::requireEnum<dsApi::HttpMethod>(request->method_string()),
                                 .content = request->body().size() ? ds::fromJson<ds::JsonFragment>(request->body())
                                                                   : ds::JsonFragment{},
                                 .query = std::string{query},
                                 .headers = {}}},
                ds::makePromise<dsApi::HttpResult>([self, version](std::optional<dsApi::HttpResult>&& result) {
                  self->m_socket.get_io_context().post(
                      ds::asio::MoveOnly{[self = std::move(self), result = std::move(result), version]() mutable {
                        self->asyncReply(result ? std::move(*result) : dsApi::cancelledResult(), version);
                      }});
                }));
          } catch (const std::exception& e) {
            DS_ERROR(e);
            self->asyncReply(dsApi::errorResult(e), version);
          }
        });
  }

  void asyncReply(dsApi::HttpResult&& result, unsigned version) {
    DS_INFO_ENTER(result);
    if (m_replied) {
      return;
    }
    m_replied = true;

    auto response = std::make_shared<http::response<http::string_body>>(); // TODO(newer asio): make_unique
    response->version(version);
    response->keep_alive(false);
    response->result(result.status);
    response->set(http::field::server, "beast");
    response->set(http::field::content_type, "application/json");
    response->body() = ds::toJson(result.content);
    response->set(http::field::content_length, response->body().size());
    http::async_write(m_socket, *response,
                      [self = shared_from_this(), response = std::move(response)](error_code ec, std::size_t) {
                        self->m_socket.shutdown(tcp::socket::shutdown_send, ec);
                      });
  }
};

class HttpServerShared {
  boost::asio::io_context& m_io;
  tcp::acceptor m_acceptor;
  dsApi::AsyncCallee& m_router;
  tcp::socket m_socket{m_io};

public:
  HttpServerShared(boost::asio::io_context& io, const tcp::endpoint& endpoint, dsApi::AsyncCallee& router)
      : m_io{io}, m_acceptor{m_io, endpoint}, m_router{router} {
    DS_INFO_ENTER(endpoint);
  }

  void asyncRun() {
    DS_INFO_ENTER("");
    // "Loop" forever accepting new connections.
    asynAcceptLoop();
  }

  void close() {
    DS_INFO_ENTER("");
    error_code ec;
    m_acceptor.close(ec);
    m_socket.close(ec);
  }

private:
  void asynAcceptLoop() {
    DS_INFO_ENTER("");
    m_acceptor.async_accept(m_socket, [&](error_code ec) {
      DS_INFO_ENTER(ec);
      if (!ec) {
        auto connection = std::make_shared<HttpConnectionShared>(std::move(m_socket), m_router);
        connection->asyncRun();
      } else {
        if (ec != boost::system::errc::operation_canceled) {
          DS_ERROR("Accept failed", ec);
        }
      }
      asynAcceptLoop();
    });
  }

  friend class HttpServerImpl;
};

class HttpServerImpl : public HttpServer {
public:
  HttpServerImpl(boost::asio::io_context& io, const tcp::endpoint& endpoint, dsApi::AsyncCallee& router)
      : m_impl{std::make_shared<HttpServerShared>(io, endpoint, router)} {
    DS_DEBUG_ENTER("");
    m_impl->asyncRun();
  }

  ~HttpServerImpl() override {
    DS_DEBUG_ENTER("");
    m_impl->close();
  }

  boost::asio::ip::tcp::endpoint localEndpoint() const override { return m_impl->m_acceptor.local_endpoint(); }

private:
  const std::shared_ptr<HttpServerShared> m_impl;
};

std::unique_ptr<HttpServer> makeHttpServer(boost::asio::io_context& io, const tcp::endpoint& endpoint,
                                           dsApi::AsyncCallee& router) {
  return std::make_unique<HttpServerImpl>(io, endpoint, router);
}

} // namespace dsConnector
