/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "subscriptions.h"
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <trompeloeil.hpp>
#include "subscriptions-mock.h"

using namespace trompeloeil;
using namespace std::literals;

namespace {

static const char* TAGS = "[dsConnectorSubscriptions][dsConnector]";

struct DiscardedEvent {
public:
  MAKE_MOCK1(call, void(dsApi::Event&&));
};

SCENARIO("dsConnectorSubscriptions", TAGS) {
  auto clock = ds::ClockMocker{};
  auto loop = ds::asio::TestLoop{};

  dsConnector::SubscriberMock sink;

  auto discardedEvent = DiscardedEvent{};
  auto subscriptions = dsConnector::makeSubscriptions(
      loop.io, {.discardedEvent = [&](auto&& x) { return discardedEvent.call(std::move(x)); }});

  auto eventType = [&] { return dsApi::EventType{"eventType0"}; };
  auto event = [&] { return dsApi::Event{.type = eventType(), .path = "path0", .data = ds::toJsonFragment(1)}; };

  auto eventTypeNeverSubscribed = [&] { return dsApi::EventType{"neverSubscribed"}; };
  auto eventNeverSubscribed = [&] { return dsApi::Event{.type = eventTypeNeverSubscribed()}; };

  auto eventIsPublished = [&] {
    {
      REQUIRE_CALL(sink, event(event())).LR_SIDE_EFFECT(loop.stop());
      subscriptions->event(event());
      loop.run();
    }
    {
      // sanity check that neverevent b is not poblished
      REQUIRE_CALL(discardedEvent, call(eventNeverSubscribed())).LR_SIDE_EFFECT(loop.stop());
      subscriptions->event(eventNeverSubscribed());
      loop.run();
    }
  };
  auto eventIsNotPublished = [&] {
    {
      REQUIRE_CALL(discardedEvent, call(event())).LR_SIDE_EFFECT(loop.stop());
      subscriptions->event(event());
      loop.run();
    }
    {
      // sanity check that neverevent b is not poblished
      REQUIRE_CALL(discardedEvent, call(eventNeverSubscribed())).LR_SIDE_EFFECT(loop.stop());
      subscriptions->event(eventNeverSubscribed());
      loop.run();
    }
  };

  WHEN("can be created and deleted") {}

  WHEN("no subscription") {
    THEN("event is not published") { eventIsNotPublished(); }
  }

  WHEN("permanent subscription to a") {
    subscriptions->subscribe({.type = eventType(), .expiresAt = ds::Clock::time_point::max(), .sink = &sink});
    THEN("event is published") { eventIsPublished(); }

    WHEN("1h passed") {
      clock.advance(1h);
      THEN("event is published") { eventIsPublished(); }
    }
  }

  WHEN("60s subscription to a") {
    subscriptions->subscribe({.type = eventType(), .expiresAt = ds::Clock::now() + 60s, .sink = &sink});
    THEN("event is published") { eventIsPublished(); }

    WHEN("45s passed") {
      clock.advance(45s);
      THEN("event is published") { eventIsPublished(); }

      WHEN("15s passed") {
        clock.advance(15s);
        THEN("event is not published") { eventIsNotPublished(); }
      }

      WHEN("60s subscription to a") {
        subscriptions->subscribe({.type = eventType(), .expiresAt = ds::Clock::now() + 60s, .sink = &sink});
        THEN("event is published") { eventIsPublished(); }

        WHEN("45s passed") {
          clock.advance(45s);
          THEN("event is published") { eventIsPublished(); }
        }
        WHEN("60s passed") {
          clock.advance(60s);
          THEN("event is not published") { eventIsNotPublished(); }
        }
      }

      WHEN("5s subscription to a") {
        subscriptions->subscribe({.type = eventType(), .expiresAt = ds::Clock::now() + 5s, .sink = &sink});
        THEN("event is published") { eventIsPublished(); }

        WHEN("5s passed") {
          clock.advance(5s);
          THEN("event is published") { eventIsPublished(); }
        }

        WHEN("10s passed") {
          clock.advance(10s);
          THEN("event is published") { eventIsPublished(); }
        }

        WHEN("15s passed") {
          clock.advance(15s);
          THEN("event is not published") { eventIsNotPublished(); }
        }
      }
    }
  }
}

} // namespace
