/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "subscriptions.h"
#include <ds-api/rest.h>
#include <ds/asio/move-only.h>
#include <ds/clock.h>
#include <ds/json.h>
#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index_container.hpp>

namespace dsConnector {

static ds::LogChannel dsLogChannel("dsConnectorSubscriptions");

namespace {

struct TypeEqual {
  template <class A, class B>
  std::size_t operator()(const A& a, const B& b) const {
    return a.type == b.type;
  }
};
struct TypeHash {
  template <class X>
  std::size_t operator()(const X& x) const {
    return boost::hash_value(static_cast<std::string>(x.type));
  }
};

using namespace boost::multi_index;
struct SubscriptionsImpl final : public Subscriptions {
  boost::asio::io_service& m_io;
  SubscriptionsHooks m_hooks;

  SubscriptionsImpl(boost::asio::io_service& io, SubscriptionsHooks&& hooks) : m_io(io), m_hooks(std::move(hooks)) {}

  struct ByType {};

  using Container =
      boost::multi_index_container<Subscription,
                                   indexed_by<hashed_unique<tag<ByType>, identity<Subscription>, TypeHash, TypeEqual>>>;
  Container m_container;

  void subscribe(Subscription&& subscription) override {
    DS_INFO_ENTER(subscription);
    auto& byType = m_container.get<ByType>();
    auto it = byType.find(subscription);
    if (it != byType.end()) {
      if (subscription.expiresAt > it->expiresAt) {
        byType.replace(it, std::move(subscription));
      }
    } else {
      m_container.insert(std::move(subscription));
    }
  }

  void event(dsApi::Event&& event) override {
    DS_INFO_ENTER(event);
    m_io.dispatch(ds::asio::MoveOnly{[this, event = std::move(event)]() mutable {
      auto now = ds::Clock::now();
      auto& byType = m_container.get<ByType>();
      auto it = byType.find(event);
      if (it != byType.end() && now < it->expiresAt) {
        it->sink->event(std::move(event));
      } else {
        DS_INFO("Dropping not subscribed event.", event.type);
        m_hooks.discardedEvent(std::move(event));
      }
    }});
  }
};

} // namespace

std::unique_ptr<Subscriptions> makeSubscriptions(boost::asio::io_service& io, SubscriptionsHooks&& hooks) {
  return std::make_unique<SubscriptionsImpl>(io, std::move(hooks));
}

} // namespace dsConnector

DS_REFLECT_DEFINE_STRUCT((dsConnector), Subscription);
