/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "wss-client.h"
#include <ds/log.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/asio/connect.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/websocket.hpp>
#include <boost/beast/websocket/ssl.hpp>
#include <boost/range/algorithm/find_first_of.hpp>

static ds::LogChannel dsLogChannel("dsConnectorWSSClient");

using boost::system::error_code;

namespace dsConnector {

class WSSStreamImpl final : public std::enable_shared_from_this<WSSStreamImpl> {
public:
  WSSStreamImpl(boost::asio::io_service &io, boost::asio::ssl::context &sslContext, WSSUrl &&url)
      : m_io{io}, m_sslContext(sslContext), m_url(std::move(url)) {
    DS_INFO_ENTER();
  }

  ~WSSStreamImpl() { DS_INFO_ENTER(); }

  template <typename T>
  void asyncConnect(T &&handler) {
    m_connectHandler = std::move(handler);
    asyncResolve();
  }

  void asyncWrite(gsl::cstring_span buffer, std::function<void(error_code)> &&handler) {
    DS_INFO_ENTER(buffer);
    m_ws.async_write(
        boost::asio::buffer(buffer.data(), buffer.size()),
        [ this, handler = std::move(handler), sharedThis = shared_from_this() ](error_code ec, std::size_t) {
          DS_INFO_ENTER(ec, m_closed);
          if (m_closed) {
            return;
          }
          return handler(ec);
        });
  }

  void asyncRead(std::function<void(error_code, std::string &&)> &&handler) {
    DS_INFO_ENTER("");
    m_ws.async_read(m_readBuffer, [ this, handler = std::move(handler), sharedThis = shared_from_this() ](error_code ec,
                                                                                                          std::size_t) {
      DS_INFO_ENTER(ec, m_closed);
      if (m_closed) {
        return;
      }
      if (ec) {
        return handler(ec, {});
      }
      auto message = boost::beast::buffers_to_string(m_readBuffer.data());
      m_readBuffer.consume(message.size());
      DS_INFO_LEAVE(message);
      return handler({}, std::move(message));
    });
  }

  std::string str() const { return ds::str(m_url); }

  void close() {
    DS_INFO_ENTER(m_closed, bool(m_connectHandler));
    if (m_closed) {
      return;
    }
    m_closed = true;
    if (m_connectHandler) {
      // websocket fully connected yet, nice async close
      m_ws.async_close(boost::beast::websocket::close_code::none, [sharedThis = shared_from_this()](error_code){});
    } else {
      // websocket not fully connected yet, close at tcp level.
      error_code ec;
      m_ws.next_layer().next_layer().close(ec);
      connected(boost::system::errc::make_error_code(boost::system::errc::operation_canceled));
    }
  }

private:
  boost::asio::io_service &m_io;
  boost::asio::ssl::context &m_sslContext;
  const WSSUrl m_url;
  std::function<void(error_code)> m_connectHandler;
  boost::asio::ip::tcp::resolver m_resolver{m_io};
  boost::beast::websocket::stream<boost::asio::ssl::stream<boost::asio::ip::tcp::socket>> m_ws{m_io, m_sslContext};
  boost::beast::multi_buffer m_readBuffer;
  bool m_closed{false};

  void connected(error_code ec) {
    if (!m_connectHandler) {
      return;
    }
    auto connectHandler = std::move(m_connectHandler);
    m_connectHandler = nullptr;
    DS_INFO_LEAVE(ec);
    connectHandler(ec);
  }

  void asyncResolve() {
    DS_INFO_ENTER(m_url);
    m_resolver.async_resolve(
        {m_url.host, m_url.port},
        [ this, sharedThis = shared_from_this() ](error_code ec, boost::asio::ip::tcp::resolver::iterator result) {
          DS_INFO_ENTER(ec, m_closed);
          if (m_closed) {
            return;
          }
          if (ec) {
            DS_ERROR("Resolve failed.", ec, m_url.host, m_url.port);
            return connected(ec);
          }
          return asyncConnect(std::move(result));
        });
  }

  void asyncConnect(boost::asio::ip::tcp::resolver::iterator &&result) {
    DS_INFO_ENTER();
    boost::asio::async_connect(
        m_ws.next_layer().next_layer(), result,
        [ this, result, sharedThis = shared_from_this() ](error_code ec, boost::asio::ip::tcp::resolver::iterator) {
          DS_INFO_ENTER(ec, m_closed);
          if (m_closed) {
            return;
          }
          if (ec) {
            DS_ERROR("Connect failed.", ec, m_url);
            return connected(ec);
          }
          asyncSSLHandshake();
        });
  }

  void asyncSSLHandshake() {
    DS_INFO_ENTER();
    m_ws.next_layer().async_handshake(boost::asio::ssl::stream_base::client,
                                      [ this, sharedThis = shared_from_this() ](error_code ec) {
                                        DS_INFO_ENTER(ec, m_closed);
                                        if (m_closed) {
                                          return;
                                        }
                                        if (ec) {
                                          DS_ERROR("SSL handshake failed.", ec, m_url);
                                          return connected(ec);
                                        }
                                        asyncWSHandshake();
                                      });
  }

  void asyncWSHandshake() {
    DS_INFO_ENTER(m_url);
    m_ws.async_handshake(m_url.host, m_url.target, [ this, sharedThis = shared_from_this() ](error_code ec) {
      DS_INFO_ENTER(ec, m_closed);
      if (m_closed) {
        return;
      }
      if (ec) {
        DS_ERROR("Ws handshake failed.", ec.message(), m_url);
        return connected(ec);
      }
      connected({});
    });
  }
};

class WSSStream : public Stream {
public:
  WSSStream(std::shared_ptr<WSSStreamImpl> impl) : m_impl(std::move(impl)) { DS_DEBUG_ENTER(); }

  ~WSSStream() override {
    DS_DEBUG_ENTER();
    m_impl->close();
  }

  void asyncWrite(std::string_view buffer, std::function<void(error_code)> &&handler) override {
    m_impl->asyncWrite(buffer, std::move(handler));
  }

  void asyncRead(std::function<void(error_code, std::string &&)> &&handler) override {
    m_impl->asyncRead(std::move(handler));
  }

  std::string str() const override { return m_impl->str(); }

private:
  const std::shared_ptr<WSSStreamImpl> m_impl;
};

class WSSStreamClientImpl final : public std::enable_shared_from_this<WSSStreamClientImpl> {
public:
  WSSStreamClientImpl(boost::asio::io_service &io, boost::asio::ssl::context &sslContext, WSSUrl &&url)
      : m_io{io}, m_sslContext(sslContext), m_url(std::move(url)) {
    DS_INFO_ENTER(m_url);
  }

  template <class Handler>
  std::unique_ptr<Stream> asyncConnect(Handler &&handler) {
    DS_DEBUG_ENTER();

    auto stream = std::make_shared<WSSStreamImpl>(m_io, m_sslContext, WSSUrl{m_url});
    stream->asyncConnect(
        [this, sharedThis = shared_from_this(), handler = std::forward<Handler>(handler)](error_code ec) {
          if (m_closed) {
            return;
          }
          return handler(ec);
        });
    return std::make_unique<WSSStream>(std::move(stream));
  }

  void close() {
    m_closed = true;
    if (m_connectingStream) {
      m_connectingStream.reset();
    }
  }

  std::string str() const { return ds::str(m_url); }

private:
  boost::asio::io_service &m_io;
  boost::asio::ssl::context &m_sslContext;
  const WSSUrl m_url{};
  std::shared_ptr<WSSStreamImpl> m_connectingStream;
  bool m_closed{false};
};

class WSSStreamClient final : public StreamClient {
public:
  WSSStreamClient(boost::asio::io_service &io, boost::asio::ssl::context &sslContext, WSSUrl &&url)
      : m_impl{std::make_shared<WSSStreamClientImpl>(io, sslContext, std::move(url))} {
    DS_DEBUG_ENTER();
  }

  ~WSSStreamClient() override {
    DS_DEBUG_ENTER();
    m_impl->close();
  }

  std::unique_ptr<Stream> asyncConnect(StreamClient::ConnectHandler &&handler) override {
    return m_impl->asyncConnect(std::move(handler));
  }

  std::string str() const override { return m_impl->str(); }

private:
  const std::shared_ptr<WSSStreamClientImpl> m_impl;
};

std::unique_ptr<StreamClient> makeWSSClient(boost::asio::io_service &io, boost::asio::ssl::context &sslContext,
                                            WSSUrl &&url) {
  return std::make_unique<WSSStreamClient>(io, sslContext, std::move(url));
}

WSSUrl parseWSSUrl(gsl::cstring_span url) {
  WSSUrl result;
  auto s = url;

  const auto protocol = gsl::cstring_span{"wss://"};
  DS_REQUIRE(boost::starts_with(s, protocol), "Failed to parse wss url (missing wss://)", url);
  s = {s.begin() + protocol.size(), s.end()};
  const auto hostEnd = boost::find_first_of(s, ":/");
  DS_REQUIRE(hostEnd != s.end(), "Failed to parse wss url (missing ':'  or '/')", url);
  result.host = {s.begin(), hostEnd};

  if (*hostEnd == ':') {
    auto portEnd = boost::find_first_of(s, "/");
    DS_REQUIRE(portEnd != s.end(), "Failed to parse wss url (missing `/` after `:`)", url);
    result.port = {hostEnd + 1, portEnd};
    s = {portEnd, s.end()};
  } else {
    result.port = "443";
    s = {hostEnd, s.end()};
  }

  result.target = gsl::to_string(s);
  return result;
}

std::unique_ptr<StreamClient> makeWSSClient(boost::asio::io_service &io, boost::asio::ssl::context &sslContext,
                                            gsl::cstring_span url) {
  return makeWSSClient(io, sslContext, parseWSSUrl(url));
}

} // namespace dsConnector
