/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-model.h"
#include <ds-connector/client-bus-mock.h>
#include <ds-connector/subscriptions-mock.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/json.h>
#include <ds/log.h>
#include <boost/asio/io_service.hpp>
#include "businterface-mock.h"
#include "cache/apartment-proxy.h"
#include "event-mock.h"
#include "event/event_create.h"
#include "model/apartment-mock.h"
#include "model/cluster.h"
#include "model/dsmeter.h"
#include "model/floor.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/state.h"
#include "model/zone.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsModelTest");
static const char* TAGS = "[dss][dssDsModel]";

using namespace dss;
using namespace trompeloeil;

namespace {

struct DsModelHooksMock {
public:
  MAKE_MOCK0(apartmentStructurePolled, void());
  MAKE_MOCK0(apartmentStatusPolled, void());

  dss::DsModelHooks dsModelHooks() {
    return {
        .apartmentStructurePolled = [&] { return apartmentStructurePolled(); },
        .apartmentStatusPolled = [&] { return apartmentStatusPolled(); },
    };
  }
};

SCENARIO("dssDsModel", TAGS) {
  auto clock = ds::ClockMocker{};
  auto propertySystem = PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  EventSinkMock eventSink;

  ApartmentMock apartment{};
  const auto apartmentId = "apartmentId-0"s;
  ALLOW_CALL(apartment, id()).RETURN(apartmentId);
  apartment.setName("TestName");
  apartment.setPropertySystem(&propertySystem);
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::ConfMock conf;
  ds::asio::TestLoopResource loop;
  dsConnector::ClientBusMock connectorBus;
  ALLOW_CALL(connectorBus, setActive(_));
  EventInterpreter eventInterpreter{nullptr};
  eventInterpreter.initialize();
  ds::sqlite3::Database d(":memory:");
  HardwareAbstraction hal{&apartment, &d};
  auto proxy = makeApartmentProxy(&apartment, &hal, &d, &eventInterpreter, conf);
  auto subscriptions = dsConnector::SubscriptionsMock{};
  auto hooks = DsModelHooksMock();
  ALLOW_CALL(hooks, apartmentStructurePolled());
  ALLOW_CALL(hooks, apartmentStatusPolled());
  auto dsModel = makeDsModel(loop.io, conf, connectorBus, apartment, eventInterpreter, subscriptions, loop.context,
                             hooks.dsModelHooks());

  auto requireAsyncStatusChanged = [&](auto&& f) {
    f();
    REQUIRE_CALL(hooks, apartmentStatusPolled()).LR_SIDE_EFFECT(loop.stop());
    loop.run();
  };

  SECTION("apartment empty included collections are serialized") {
    apartment.removeZone(NEW_DEVICES_ZONE_ID);
    apartment.removeFloor(APARTMENT_FLOOR_ID);
    CHECK(dsModel->apartment().included ==
          dsModel::ApartmentIncluded{.installation = dsModel->installation(),
                                     .dsDevices = std::vector<dsModel::DSDevice>{},
                                     .submodules = std::vector<dsModel::Submodule>{},
                                     .functionBlocks = std::vector<dsModel::FunctionBlock>{},
                                     .zones = std::vector<dsModel::Zone>{},
                                     .scenarios = std::vector<dsModel::Scenario>{},
                                     .userDefinedStates = std::vector<dsModel::UserDefinedState>{},
                                     .floors = std::vector<dsModel::Floor>{},
                                     .clusters = std::vector<dsModel::Cluster>{},
                                     .applications = std::vector<dsModel::Application>{},
                                     .dsServer = dsModel->dsServer()});
    CHECK(dsModel->apartmentStatus().included ==
          dsModel::ApartmentStatusIncluded{.dsDevices = std::vector<dsModel::DSDeviceStatus>{},
                                           .zones = std::vector<dsModel::ZoneStatus>{},
                                           .clusters = std::vector<dsModel::ClusterStatus>{},
                                           .userDefinedStates = std::vector<dsModel::UserDefinedStateStatus>{}});
  }

  SECTION("setZoneName") {
    auto& zone1 = *apartment.allocateZone(1);
    zone1.setName("zone1");

    THEN("success") {
      dsModel->setZoneName({.zoneId = zone1.id()}, "zone1x");
      CHECK(zone1.getName() == "zone1x");
    }

    THEN("Empty name fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = zone1.id()}, ""), "New name cannot be empty");
    }

    THEN("Invalid zoneId fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{"455"}}, "xyz"), "Zone not found");
    }

    THEN("Zone 0 fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{"0"}}, "xyz"), "Zone not found");
    }
    THEN("New devices zone fails") {
      CHECK_THROWS_FIND(dsModel->setZoneName({.zoneId = dsModel::ZoneId{ds::str(NEW_DEVICES_ZONE_ID)}}, "xyz"),
                        "New devices zone cannot be renamed");
    }
  }

  SECTION("basic apartment information") {
    const auto dsApartment = dsModel->apartment();
    CHECK(dsApartment.id == apartmentId);
    CHECK(dsApartment.type == "apartment");
    CHECK(dsApartment.attributes.name == apartment.getName());

    SECTION("included are present") {
      const auto& included = dsApartment.included;
      CHECK(included.installation == dsModel->installation());
      CHECK(included.scenarios == dsModel->scenarios().scenarios());
      CHECK(included.userDefinedStates == dsModel->userDefinedStates());
      CHECK(included.floors == dsModel->floors().floors);
      CHECK(included.applications == dsModel->applications());
    }
  }

  SECTION("shadeDeviceType check") {
    const auto id = "1111111111111111111111111111111100";
    const auto deviceId = dsModel::DeviceId{id};
    auto& device = *apartment.allocateDevice(Dsuid{id});

    SECTION("device of shade type") {
      GIVEN("GR_KL200") {
        device.setHardwareType(HardwareType::GR_KL200);
        THEN("rollerShutter") {
          CHECK(dsModel->device(deviceId) ==
                (dsModel::DSDevice{
                    .id = dsModel::DeviceId{"1111111111111111111111111111111100"},
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = {},
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = dsModel::ShadeDeviceType::rollerShutter,
                                                       .protectionClass = {},
                                                       .mountOrientation = {}},
                                   .submodules = {dsModel::SubmoduleId{"1111111111111111111111111111111100"}},
                                   .zone = "65534"}}));
        };
      }
      GIVEN("GR_KL210") {
        device.setHardwareType(HardwareType::GR_KL210);
        THEN("awningTerrace") {
          CHECK(dsModel->device(deviceId) ==
                (dsModel::DSDevice{
                    .id = dsModel::DeviceId{"1111111111111111111111111111111100"},
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = {},
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = dsModel::ShadeDeviceType::awningTerrace,
                                                       .protectionClass = {},
                                                       .mountOrientation = {}},
                                   .submodules = {dsModel::SubmoduleId{"1111111111111111111111111111111100"}},
                                   .zone = "65534"}}));
        };
      }
      GIVEN("GR_KL220") {
        device.setHardwareType(HardwareType::GR_KL220);
        THEN("windowBlind") {
          CHECK(dsModel->device(deviceId) ==
                (dsModel::DSDevice{
                    .id = dsModel::DeviceId{"1111111111111111111111111111111100"},
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = {},
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = dsModel::ShadeDeviceType::windowBlind,
                                                       .protectionClass = {},
                                                       .mountOrientation = {}},
                                   .submodules = {dsModel::SubmoduleId{"1111111111111111111111111111111100"}},
                                   .zone = "65534"}}));
        };
      }
      GIVEN("GR_HKL230") {
        device.setHardwareType(HardwareType::GR_HKL230);
        THEN("windowBlind") {
          CHECK(dsModel->device(deviceId) ==
                (dsModel::DSDevice{
                    .id = dsModel::DeviceId{"1111111111111111111111111111111100"},
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = {},
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = dsModel::ShadeDeviceType::windowBlind,
                                                       .protectionClass = {},
                                                       .mountOrientation = {}},
                                   .submodules = {dsModel::SubmoduleId{"1111111111111111111111111111111100"}},
                                   .zone = "65534"}}));
        };
      }
    }

    SECTION("non-shade device") {
      GIVEN("GE_KM200") {
        device.setHardwareType(HardwareType::GE_KM200);
        THEN("brightness") {
          CHECK(dsModel->device(deviceId) ==
                (dsModel::DSDevice{
                    .id = dsModel::DeviceId{"1111111111111111111111111111111100"},
                    .type = "dsDevice",
                    .lastChanged = {},
                    .attributes = {.name = {},
                                   .model = {},
                                   .gtin = {},
                                   .dsid = {},
                                   .displayId = device.getDisplayID(),
                                   .present = 0,
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = {.shadeType = {}, .protectionClass = {}, .mountOrientation = {}},
                                   .submodules = {dsModel::SubmoduleId{"1111111111111111111111111111111100"}},
                                   .zone = "65534"}}));
        };
      }
    }
  }

  SECTION("apartment") {
    auto& zone0 = *apartment.allocateZone(0);
    GIVEN("access") {
      THEN("Missing state") {
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }

      auto presence = apartment.allocateState(StateType_Apartment, dss::StateName::Presence, "");
      State::ValueRange_t presenceValues;
      presenceValues.push_back("unknown");
      presenceValues.push_back("present");
      presenceValues.push_back("absent");
      presenceValues.push_back("invalid");
      presence->setValueRange(presenceValues);

      THEN("State invalid") {
        presence->setState(coUnknown, "invalid");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }
      THEN("State unknown") {
        presence->setState(coUnknown, "unknown");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = {}}));
      }
      THEN("value == true") {
        presence->setState(coUnknown, "absent");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = true}));
      }
      THEN("value == false") {
        presence->setState(coUnknown, "present");
        CHECK(dsModel->apartmentAccessStatus() == (dsModel::ApartmentAccessStatus{.absent = false}));
      }
    }

    GIVEN("user") {
      THEN("Missing state") { CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}})); }

      auto hibernation = apartment.allocateState(StateType_Apartment, dss::StateName::Hibernation, "");
      State::ValueRange_t sleepmodeValues;
      sleepmodeValues.push_back("unknown");
      sleepmodeValues.push_back("awake");
      sleepmodeValues.push_back("sleeping");
      sleepmodeValues.push_back("invalid");
      hibernation->setValueRange(sleepmodeValues);

      THEN("State invalid") {
        hibernation->setState(coUnknown, "invalid");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}}));
      }
      THEN("State unknown") {
        hibernation->setState(coUnknown, "unknown");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = {}}));
      }
      THEN("value == true") {
        hibernation->setState(coUnknown, "awake");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = false}));
      }
      THEN("value == false") {
        hibernation->setState(coUnknown, "sleeping");
        CHECK(dsModel->apartmentUserStatus() == (dsModel::ApartmentUserStatus{.sleeping = true}));
      }
    }

    GIVEN("security") {
      auto& group00 = *zone0.getGroup(0);
      SECTION("set alarm 1 name") {
        dsModel->setApartmentAlarm1Name("testName");
        CHECK(group00.getSceneName(SceneAlarm) == "testName");
      }
      SECTION("set alarm 2 name") {
        dsModel->setApartmentAlarm2Name("testName");
        CHECK(group00.getSceneName(SceneAlarm2) == "testName");
      }
      SECTION("set alarm 3 name") {
        dsModel->setApartmentAlarm3Name("testName");
        CHECK(group00.getSceneName(SceneAlarm3) == "testName");
      }
      SECTION("set alarm 4 name") {
        dsModel->setApartmentAlarm4Name("testName");
        CHECK(group00.getSceneName(SceneAlarm4) == "testName");
      }
    }

    GIVEN("securityStatus") {
      THEN("Missing state") {
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }

      auto panic = apartment.allocateState(StateType_Apartment, dss::StateName::Panic, "");
      auto fire = apartment.allocateState(StateType_Apartment, dss::StateName::Fire, "");
      auto alarm = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm, "");
      auto alarm2 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm2, "");
      auto alarm3 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm3, "");
      auto alarm4 = apartment.allocateState(StateType_Apartment, dss::StateName::Alarm4, "");
      THEN("State invalid") {
        panic->setState(coUnknown, State_Invalid);
        fire->setState(coUnknown, State_Invalid);
        alarm->setState(coUnknown, State_Invalid);
        alarm2->setState(coUnknown, State_Invalid);
        alarm3->setState(coUnknown, State_Invalid);
        alarm4->setState(coUnknown, State_Invalid);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("State unknown") {
        panic->setState(coUnknown, State_Unknown);
        fire->setState(coUnknown, State_Unknown);
        alarm->setState(coUnknown, State_Unknown);
        alarm2->setState(coUnknown, State_Unknown);
        alarm3->setState(coUnknown, State_Unknown);
        alarm4->setState(coUnknown, State_Unknown);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = {},
                                                                                      .fire = {},
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = {},
                                                                                      .alarm2 = {},
                                                                                      .alarm3 = {},
                                                                                      .alarm4 = {},
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("value == true") {
        panic->setState(coUnknown, State_Active);
        fire->setState(coUnknown, State_Active);
        alarm->setState(coUnknown, State_Active);
        alarm2->setState(coUnknown, State_Active);
        alarm3->setState(coUnknown, State_Active);
        alarm4->setState(coUnknown, State_Active);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = true,
                                                                                      .fire = true,
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = true,
                                                                                      .alarm2 = true,
                                                                                      .alarm3 = true,
                                                                                      .alarm4 = true,
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
      THEN("value == false") {
        panic->setState(coUnknown, State_Inactive);
        fire->setState(coUnknown, State_Inactive);
        alarm->setState(coUnknown, State_Inactive);
        alarm2->setState(coUnknown, State_Inactive);
        alarm3->setState(coUnknown, State_Inactive);
        alarm4->setState(coUnknown, State_Inactive);
        CHECK(dsModel->apartmentSecurityStatus() == (dsModel::ApartmentSecurityStatus{.panic = false,
                                                                                      .fire = false,
                                                                                      .smoke = {},
                                                                                      .gas = {},
                                                                                      .pollution = {},
                                                                                      .alarm1 = false,
                                                                                      .alarm2 = false,
                                                                                      .alarm3 = false,
                                                                                      .alarm4 = false,
                                                                                      .burglaryAlarm = {},
                                                                                      .burglaryAttempt = {}}));
      }
    }

    SECTION("temperature") {
      auto settingsNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings");
      auto setSettingsChildValue = [&](auto&& path, auto&& value) {
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        settingsNode->setChildValue(std::move(path), std::move(value));
      };
      auto ecoNode =
          propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller/operationmode.0.Eco");
      auto setEcoValue = [&](auto&& value) {
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        ecoNode->setChildValue("value", std::move(value));
      };

      GIVEN("no config") {
        THEN("heating not available") {
          CHECK(dsModel->apartmentTemperatureStatus() ==
                (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::off, .ecoPreferred = false}));
        }
      }

      GIVEN("heating enabled") {
        setSettingsChildValue("heatingEnabled", true);

        GIVEN("no forced settings") {
          THEN("mode = automatic") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                                       .ecoPreferred = false}));
          }
          GIVEN("eco preferred") {
            setEcoValue(1);

            THEN("eco = true") {
              CHECK(dsModel->apartmentTemperatureStatus() ==
                    (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                                         .ecoPreferred = true}));
            }
          }
        }

        GIVEN("forced heating") {
          setSettingsChildValue("overrideSystemModeAsHeating", true);

          THEN("mode = heating") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::heating,
                                                       .ecoPreferred = false}));
          }
        }

        GIVEN("forced cooling") {
          setSettingsChildValue("overrideSystemModeAsCooling", true);

          THEN("mode = cooling") {
            CHECK(dsModel->apartmentTemperatureStatus() ==
                  (dsModel::ApartmentTemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                                       .ecoPreferred = false}));
          }
        }
      }
    }

    GIVEN("weather") {
      THEN("Missing state") {
        CHECK(dsModel->apartmentWeatherStatus() ==
              (dsModel::ApartmentWeatherStatus{.wind = {}, .rain = {}, .hail = {}, .frost = {}}));
      }

      auto wind = apartment.allocateState(StateType_Apartment, dss::StateName::Wind, "");
      auto rain = apartment.allocateState(StateType_Apartment, dss::StateName::Rain, "");
      auto hail = apartment.allocateState(StateType_Apartment, dss::StateName::Hail, "");
      auto frost = apartment.allocateState(StateType_Apartment, dss::StateName::Frost, "");
      THEN("State invalid") {
        wind->setState(coUnknown, State_Invalid);
        rain->setState(coUnknown, State_Invalid);
        hail->setState(coUnknown, State_Invalid);
        frost->setState(coUnknown, State_Invalid);
        CHECK(dsModel->apartmentWeatherStatus() ==
              (dsModel::ApartmentWeatherStatus{.wind = {}, .rain = {}, .hail = {}, .frost = {}}));
      }
      THEN("State unknown") {
        wind->setState(coUnknown, State_Unknown);
        rain->setState(coUnknown, State_Unknown);
        hail->setState(coUnknown, State_Unknown);
        frost->setState(coUnknown, State_Unknown);
        CHECK(dsModel->apartmentWeatherStatus() ==
              (dsModel::ApartmentWeatherStatus{.wind = {}, .rain = {}, .hail = {}, .frost = {}}));
      }
      THEN("value == true") {
        wind->setState(coUnknown, State_Active);
        rain->setState(coUnknown, State_Active);
        hail->setState(coUnknown, State_Active);
        frost->setState(coUnknown, State_Active);
        CHECK(dsModel->apartmentWeatherStatus() ==
              (dsModel::ApartmentWeatherStatus{.wind = true, .rain = true, .hail = true, .frost = true}));
      }
      THEN("value == false") {
        wind->setState(coUnknown, State_Inactive);
        rain->setState(coUnknown, State_Inactive);
        hail->setState(coUnknown, State_Inactive);
        frost->setState(coUnknown, State_Inactive);
        CHECK(dsModel->apartmentWeatherStatus() ==
              (dsModel::ApartmentWeatherStatus{.wind = false, .rain = false, .hail = false, .frost = false}));
      }
    }

    GIVEN("environment") {
      THEN("No values set") {
        CHECK(dsModel->apartmentMeasurementStatus() ==
              (dsModel::ApartmentMeasurementStatus{
                  .temperature = {}, .humidity = {}, .brightness = {}, .airPressure = {}, .precipitation = {}}));
      }

      THEN("Values set") {
        DateTime now;
        apartment.setTemperature(15.5, now);
        apartment.setHumidityValue(50, now);
        apartment.setBrightnessValue(25, now);
        apartment.setAirPressure(900, now);
        apartment.setPrecipitation(5, now);
        // collecting gustSpeed + gustDirection internally
        apartment.setWindSpeed(6, now);
        apartment.setGustSpeed(13, now);

        // publishing gustSpeed as windGust, gustDirection not planned to be published
        CHECK(dsModel->apartmentMeasurementStatus() == (dsModel::ApartmentMeasurementStatus{.temperature = 15.5,
                                                                                            .humidity = 50,
                                                                                            .brightness = 25,
                                                                                            .airPressure = 900,
                                                                                            .precipitation = 5,
                                                                                            .windSpeed = 6,
                                                                                            .windGust = 13}));
      }
    }
  }

  SECTION("temperatureStatus") {
    auto& zone1 = *apartment.allocateZone(1);
    DateTime now;

    auto settingsNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings");
    auto setSettingsChildValue = [&](auto&& path, auto&& value) {
      REQUIRE_CALL(hooks, apartmentStatusPolled());
      settingsNode->setChildValue(std::move(path), std::move(value));
    };
    auto configNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig/1");
    auto setConfigChildValue = [&](auto&& path, auto&& value) {
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      REQUIRE_CALL(hooks, apartmentStatusPolled());
      configNode->setChildValue(std::move(path), std::move(value));
    };

    GIVEN("no temperature device") {
      THEN("temperatureStatus throws") {
        CHECK_THROWS_FIND(dsModel->temperatureStatus(zone1.id()), "Temperature application does not exist.");
      }
      THEN("applications zoneApplicationsStatus() is empty") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) == dsModel::ZoneApplicationsStatus{});
      }
    }

    auto& device = *apartment.allocateDevice(Dsuid{"5111111111111111111111111111111100"});
    device.setActiveGroup(ApplicationType::ControlTemperature);
    zone1.addDevice({device.sharedFromThis(), &apartment});

    GIVEN("no PID controller") {
      THEN("temperatureStatus throws") {
        CHECK_THROWS_FIND(dsModel->temperatureStatus(zone1.id()), "Temperature application does not exist.");
      }
      THEN("applications zoneApplicationsStatus() is empty") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) == dsModel::ZoneApplicationsStatus{});
      }
    }

    zone1.setTemperatureConfig([] {
      TemperatureConfig result{};
      result.m_mode = HeatingControlMode::PID; // PID mode required for temperature application to exist
      return result;
    }());

    GIVEN("heating disabled") {
      setSettingsChildValue("heatingEnabled", false);
      THEN("all is off") {
        CHECK(dsModel->temperatureStatus(zone1.id()) ==
              dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                         .status = dsModel::TemperatureStatusStatus::off,
                                         .emergency = {},
                                         .setpoint = {},
                                         .controlValue = {},
                                         .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                         .setbackRequested = false,
                                         .ecoPreferred = false});
      }
      THEN("applications zoneApplicationsStatus() containes temperature status") {
        CHECK(dsModel->zoneApplicationsStatus(zone1.id()) ==
              dsModel::ZoneApplicationsStatus{.lights = {},
                                              .shades = {},
                                              .awnings = {},
                                              .audio = {},
                                              .ventilation = {},
                                              .recirculation = {},
                                              .temperature = dsModel->temperatureStatus(zone1.id())});
      }
    }

    GIVEN("apartment heating enabled") {
      setSettingsChildValue("heatingEnabled", true);

      THEN("zone heating/cooling not possible") {
        setConfigChildValue("heatingPossible", false);
        setConfigChildValue("coolingPossible", false);
        setConfigChildValue("passiveCoolingPossible", false);

        CHECK(dsModel->temperatureStatus(zone1.id()) ==
              dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                         .status = dsModel::TemperatureStatusStatus::off,
                                         .emergency = {},
                                         .setpoint = {},
                                         .controlValue = {},
                                         .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                         .setbackRequested = false,
                                         .ecoPreferred = false});
      }

      GIVEN("zone heating, cooling, passiveCooling possible") {
        setConfigChildValue("heatingPossible", true);
        setConfigChildValue("coolingPossible", true);
        setConfigChildValue("passiveCoolingPossible", true);

        THEN("heating") {
          zone1.setControlValue(25, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::heating,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = 25,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }

        THEN("cooling") {
          zone1.setControlValue(-50, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::cooling,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = -50,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }

        THEN("neutral") {
          zone1.setControlValue(0, now);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::neutral,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = 0,
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }

        SECTION("emergency") {
          zone1.setHeatingControlState(HeatingControlStateIDEmergency);

          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::automatic,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = true,
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }
      }

      GIVEN("heating forced") {
        setSettingsChildValue("overrideSystemModeAsHeating", true);

        THEN("heating not possible") {
          setConfigChildValue("heatingPossible", false);
          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = {},
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }

        THEN("heating possible") {
          setConfigChildValue("heatingPossible", true);
          CHECK(dsModel->temperatureStatus(zone1.id()) ==
                dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::heating,
                                           .status = dsModel::TemperatureStatusStatus::off,
                                           .emergency = false,
                                           .setpoint = {},
                                           .controlValue = {},
                                           .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                           .setbackRequested = false,
                                           .ecoPreferred = false});
        }

        GIVEN("cooling forced") {
          setSettingsChildValue("overrideSystemModeAsHeating", false);
          setSettingsChildValue("overrideSystemModeAsCooling", true);

          THEN("cooling and passive cooling not possible") {
            setConfigChildValue("coolingPossible", false);
            setConfigChildValue("passiveCoolingPossible", false);
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false});
          }

          THEN("cooling possible") {
            setConfigChildValue("coolingPossible", true);
            setConfigChildValue("passiveCoolingPossible", false);
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = false,
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false});
          }

          THEN("passive cooling possible") {
            setConfigChildValue("coolingPossible", false);
            setConfigChildValue("passiveCoolingPossible", true);
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::cooling,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = false,
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = false,
                                             .ecoPreferred = false});
          }
        }

        GIVEN("lastCalledScene and setpoint") {
          requireAsyncStatusChanged([&] { zone1.setNominalValue(20, DateTime()); });
          requireAsyncStatusChanged([&] { zone1.setHeatingOperationMode(HeatingOperationModeIDEconomy); });

          THEN("check") {
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = 20,
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingEconomy,
                                             .setbackRequested = false,
                                             .ecoPreferred = false});
          }
        }

        GIVEN("eco and setback set") {
          auto ecoNode =
              propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller/operationmode.1.Eco");
          auto setEcoValue = [&](auto&& value) {
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            ecoNode->setChildValue("value", std::move(value));
          };
          auto setbackNode = propertySystem.createProperty(
              "/usr/addon-states/system-addon-heating-controller/operationmode.1.NotUsed");
          auto setSetbackValue = [&](auto&& value) {
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            setbackNode->setChildValue("value", std::move(value));
          };

          setEcoValue(1);
          setSetbackValue(1);

          THEN("check") {
            CHECK(dsModel->temperatureStatus(zone1.id()) ==
                  dsModel::TemperatureStatus{.mode = dsModel::TemperatureMode::off,
                                             .status = dsModel::TemperatureStatusStatus::off,
                                             .emergency = {},
                                             .setpoint = {},
                                             .controlValue = {},
                                             .lastCalledScene = dsModel::TemperatureScene::heatingOff,
                                             .setbackRequested = true,
                                             .ecoPreferred = true});
          }
        }
      }
    }
  }

  SECTION("Set apartment temperature status eco preferred") {
    const auto& value = GENERATE(true, false);
    REQUIRE_CALL(eventSink, pushEvent(_))
        .LR_SIDE_EFFECT(CHECK(
            *_1 ==
            Event{"heating-controller.operation-mode",
                  Properties{{"actions", "setOperationMode"}, {"zoneID", "0"}, {"operationMode", value ? "2" : "1"}}}));
    dsModel->setApartmentTemperatureStatusEcoPreferred({}, value);
  }

  SECTION("Set temperature status setpoint") {
    auto& zone1 = *apartment.allocateZone(1);
    const auto value = 25.4;
    REQUIRE_CALL(busInterface.action, pushSensor(zone1.getGroup(GroupIDControlTemperature).get(), coJSON, SAC_MANUAL,
                                                 Dsuid{}, SensorType::RoomTemperatureSetpoint, value, ""));
    dsModel->setTemperatureStatusSetpoint({.zoneId = zone1.id()}, value);
  }

  SECTION("Set temperature status setbackRequested") {
    THEN("setbackRequested true") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "3"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusSetbackRequested(args, true);
    }
    THEN("setbackRequested false") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "resetOperationMode"}, {"zoneID", "1"}, {"operationMode", "3"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusSetbackRequested(args, false);
    }
  }

  SECTION("Set temperature status ecoPreferred") {
    THEN("ecoPreferred true") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "2"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusEcoPreferred(args, true);
    }
    THEN("ecoPreferred false") {
      REQUIRE_CALL(eventSink, pushEvent(_))
          .LR_SIDE_EFFECT(CHECK(
              *_1 == Event{"heating-controller.operation-mode",
                           Properties{{"actions", "setOperationMode"}, {"zoneID", "1"}, {"operationMode", "1"}}}));
      const dsModel::TemperatureStatusArgs& args = {.zoneId = dsModel::ZoneId{"1"}};
      dsModel->setTemperatureStatusEcoPreferred(args, false);
    }
  }

  SECTION("zoneStatus") {
    auto& zone1 = *apartment.allocateZone(1);
    SECTION("zoneMeasurementsStatus") {
      THEN("no measurements") {
        CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) == dsModel::ZoneMeasurementsStatus{});
      }

      GIVEN("states") {
        auto& motion = *apartment.allocateState(StateType_Service, zoneStateName(StateName::Motion, zone1.getID()), {});
        auto& presence =
            *apartment.allocateState(StateType_Service, zoneStateName(StateName::Presence, zone1.getID()), {});
        motion.setState(coUnknown, State_Invalid);
        presence.setState(coUnknown, State_Invalid);

        SECTION("no measurements") {
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) == dsModel::ZoneMeasurementsStatus{});
        }

        SECTION("zone stored measurements") {
          zone1.setTemperature(11, {});
          zone1.setHumidityValue(12, {});
          zone1.setBrightnessValue(13, {});
          zone1.setCO2ConcentrationValue(14, {});
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) ==
                dsModel::ZoneMeasurementsStatus{
                    .temperature = 11, .humidity = 12, .brightness = 13, .carbonDioxide = 14});
        }

        SECTION("state represented measurements") {
          struct X {
            eState dssValue;
            std::optional<bool> dsValue;
          };
          const auto& x = GENERATE(values(std::initializer_list<X>{
              {State_Invalid, {}}, {State_Active, true}, {State_Inactive, false}, {State_Unknown, {}}}));
          INFO(x.dssValue);
          motion.setState(coUnknown, x.dssValue);
          presence.setState(coUnknown, x.dssValue);
          CHECK(dsModel->zoneMeasurementsStatus(dsModel::ZoneId{"1"}) ==
                dsModel::ZoneMeasurementsStatus{.temperature = {},
                                                .humidity = {},
                                                .brightness = {},
                                                .carbonDioxide = {},
                                                .motion = x.dsValue,
                                                .presence = x.dsValue});
        }
      }
    }
  }

  SECTION("Energy and power meterings") {
    auto& meter = *apartment.allocateDSMeter(Dsuid{"1234511111111111111111111111111100"});
    meter.setIsPresent(true);
    meter.setCapability_HasMetering(true);
    meter.setIsConnected(true);
    meter.setPowerConsumption(100);
    meter.updateEnergyMeterValue(50);
    meter.setHardwareName("dsmTest");

    SECTION("meter does not have metering capability") {
      meter.setCapability_HasMetering(false);
      CHECK(
          dsModel->meterings() ==
          dsModel::Meterings{
              .meterings = {
                  {.id = dsModel::MeteringId{"apartment-energy"},
                   .type = "energyMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = {.type = "apartment"}}},
                  {.id = dsModel::MeteringId{"apartment-power"},
                   .type = "powerMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "apartment power", .origin = {.type = "apartment"}}},
              }});
      CHECK(dsModel->meteringValues() == dsModel::MeteringValues{.values = {
                                                                     {.id = dsModel::MeteringId{"apartment-energy"},
                                                                      .type = "meteringValue",
                                                                      .lastChanged = {},
                                                                      .attributes = {.value = 0}},
                                                                     {.id = dsModel::MeteringId{"apartment-power"},
                                                                      .type = "meteringValue",
                                                                      .lastChanged = {},
                                                                      .attributes = {.value = 0}},
                                                                 }});
    }

    SECTION("meter is not present") {
      meter.setIsPresent(false);
      CHECK(
          dsModel->meterings() ==
          dsModel::Meterings{
              .meterings = {
                  {.id = dsModel::MeteringId{"apartment-energy"},
                   .type = "energyMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = {.type = "apartment"}}},
                  {.id = dsModel::MeteringId{"apartment-power"},
                   .type = "powerMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "apartment power", .origin = {.type = "apartment"}}},
              }});
      CHECK(dsModel->meteringValues() == dsModel::MeteringValues{.values = {
                                                                     {.id = dsModel::MeteringId{"apartment-energy"},
                                                                      .type = "meteringValue",
                                                                      .lastChanged = {},
                                                                      .attributes = {.value = 0}},
                                                                     {.id = dsModel::MeteringId{"apartment-power"},
                                                                      .type = "meteringValue",
                                                                      .lastChanged = {},
                                                                      .attributes = {.value = 0}},
                                                                 }});
    }

    SECTION("meterings") {
      CHECK(
          dsModel->meterings() ==
          dsModel::Meterings{
              .meterings = {
                  {.id = dsModel::MeteringId{"apartment-energy"},
                   .type = "energyMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = {.type = "apartment"}}},
                  {.id = dsModel::MeteringId{"apartment-power"},
                   .type = "powerMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "W", .technicalName = "apartment power", .origin = {.type = "apartment"}}},

                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                   .type = "energyMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "Wh",
                                  .technicalName = "dsmTest energy",
                                  .origin = {.type = "dsMeter", .id = "1234511111111111111111111111111100"}}},
                  {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                   .type = "powerMetering",
                   .lastChanged = {},
                   .attributes = {.unit = "W",
                                  .technicalName = "dsmTest power",
                                  .origin = {.type = "dsMeter", .id = "1234511111111111111111111111111100"}}},

              }});
    }
    SECTION("metering") {
      CHECK(
          dsModel->metering(dsModel::MeteringArgs{
              .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"}}) ==
          dsModel::Metering{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                            .type = "energyMetering",
                            .lastChanged = {},
                            .attributes = {.unit = "Wh",
                                           .technicalName = "dsmTest energy",
                                           .origin = {.type = "dsMeter", .id = "1234511111111111111111111111111100"}}});
      CHECK(
          dsModel->metering(dsModel::MeteringArgs{
              .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"}}) ==
          dsModel::Metering{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                            .type = "powerMetering",
                            .lastChanged = {},
                            .attributes = {.unit = "W",
                                           .technicalName = "dsmTest power",
                                           .origin = {.type = "dsMeter", .id = "1234511111111111111111111111111100"}}});
      CHECK(dsModel->metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}) ==
            dsModel::Metering{
                .id = dsModel::MeteringId{"apartment-energy"},
                .type = "energyMetering",
                .lastChanged = {},
                .attributes = {.unit = "Wh", .technicalName = "apartment energy", .origin = {.type = "apartment"}}});
      CHECK(dsModel->metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}) ==
            dsModel::Metering{
                .id = dsModel::MeteringId{"apartment-power"},
                .type = "powerMetering",
                .lastChanged = {},
                .attributes = {.unit = "W", .technicalName = "apartment power", .origin = {.type = "apartment"}}});
    }
    SECTION("meteringValues") {
      auto values = [&] {
        return dsModel::MeteringValues{.values = {
                                           {.id = dsModel::MeteringId{"apartment-energy"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{meter.getCachedEnergyMeterTimeStamp()},
                                            .attributes = {.value = 50}},
                                           {.id = dsModel::MeteringId{"apartment-power"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{meter.getCachedPowerConsumptionTimeStamp()},
                                            .attributes = {.value = 100}},
                                           {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{meter.getCachedEnergyMeterTimeStamp()},
                                            .attributes = {.value = 50}},
                                           {.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                                            .type = "meteringValue",
                                            .lastChanged = ds::DateTime{meter.getCachedPowerConsumptionTimeStamp()},
                                            .attributes = {.value = 100}},
                                       }};
      };
      SECTION("polled") { CHECK(dsModel->meteringValues() == values()); }

      SECTION("pushed every second") {
        auto event = [&] {
          return dsApi::Event{.type = dsModel::meteringValues, .path = {}, .data = ds::toJsonFragment(values())};
        };
        REQUIRE_CALL(subscriptions, event(_)).LR_SIDE_EFFECT(CHECK(_1 == event()); loop.stop());
        clock.advance(2s);
        loop.run();
        REQUIRE_CALL(subscriptions, event(_)).LR_SIDE_EFFECT(CHECK(_1 == event()); loop.stop());
        clock.advance(2s);
        loop.run();
      }
    }
    SECTION("meteringValue") {
      CHECK(dsModel->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-energy"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{meter.getCachedEnergyMeterTimeStamp()},
                                   .attributes = {.value = 50}});
      CHECK(dsModel->meteringValue(dsModel::MeteringArgs{
                .meteringId = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"dsm-1234511111111111111111111111111100-power"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{meter.getCachedPowerConsumptionTimeStamp()},
                                   .attributes = {.value = 100}});
      CHECK(dsModel->meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-energy"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{meter.getCachedEnergyMeterTimeStamp()},
                                   .attributes = {.value = 50}});
      CHECK(dsModel->meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}) ==
            dsModel::MeteringValue{.id = dsModel::MeteringId{"apartment-power"},
                                   .type = "meteringValue",
                                   .lastChanged = ds::DateTime{meter.getCachedPowerConsumptionTimeStamp()},
                                   .attributes = {.value = 100}});
    }
  }

  SECTION("decode metering id") {
    CHECK(decodeId(dsModel::MeteringId{"dsm-x-power"}) ==
          DecodedMeteringId{.origin = {.type = "dsm", .id = "x"}, .meteringType = "power"});
    CHECK(decodeId(dsModel::MeteringId{"dsm-y-energy"}) ==
          DecodedMeteringId{.origin = {.type = "dsm", .id = "y"}, .meteringType = "energy"});

    CHECK_THROWS_FIND(decodeId(dsModel::MeteringId{"meter-x-power"}), "Unsupported measuring device.");
    CHECK_THROWS_FIND(decodeId(dsModel::MeteringId{"dsm-x-phase"}), "Unsupported measuring type.");
  }

  SECTION("applications") {
    auto& zone1 = *apartment.allocateZone(1);
    auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
    DeviceReference devRef(device.sharedFromThis(), &apartment);
    zone1.addDevice(devRef);

    SECTION("applications") {
      GIVEN("No device") {
        THEN("No applications") { CHECK(dsModel->applications() == std::vector<dsModel::Application>{}); }
      }
    }

    SECTION("temperatureApplication") {
      SECTION("read settings") {
        GIVEN("No temperature application device in zone") {
          THEN("No temperatureApplication returned") {
            CHECK_THROWS_FIND(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}),
                              "Temperature application does not exist.");
          }
        }

        GIVEN("Temperature application device present in zone") {
          GIVEN("No PID controller ") {
            THEN("No temperatureApplication returned") {
              CHECK_THROWS_FIND(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}),
                                "Temperature application does not exist.");
            }
          }

          zone1.setTemperatureConfig([] {
            TemperatureConfig result{};
            result.m_mode = HeatingControlMode::PID; // PID mode required for temperature application to exist
            return result;
          }());

          device.setActiveGroup(ApplicationType::ControlTemperature);
          auto configNode =
              propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig/1");
          auto setConfigChildValue = [&](auto&& path, auto&& value) {
            REQUIRE_CALL(hooks, apartmentStructurePolled());
            REQUIRE_CALL(hooks, apartmentStatusPolled());
            configNode->setChildValue(std::move(path), std::move(value));
          };
          TemperatureConfig config{};
          config.targetTemperatures = {10, 11, 12, 13, 14, 15, 0, 0, 0, 20, 21, 22, 23, 24, 25};
          zone1.setHeatingSetpoints(config);

          THEN("Default temperatureApplication") {
            CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                  dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                       .attributes = dsModel::TemperatureApplicationAttributes{
                                           .applicationType = dsModel::ZoneApplicationType::temperature,
                                           .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                           .availableLevels = {},
                                           .setpoints = {}}});
          }

          GIVEN("Heating possible") {
            setConfigChildValue("heatingPossible", true);
            THEN("Heating levels available") {
              setConfigChildValue("heatingPossible", true);
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::heatingComfort, dsModel::TemperatureLevel::heatingEconomy,
                  dsModel::TemperatureLevel::heatingSetback, dsModel::TemperatureLevel::heatingNight,
                  dsModel::TemperatureLevel::heatingVacation};
              const std::vector<dsModel::TemperatureSetpoint> setpoints = {
                  {.id = dsModel::TemperatureLevel::heatingComfort, .value = 11},
                  {.id = dsModel::TemperatureLevel::heatingEconomy, .value = 12},
                  {.id = dsModel::TemperatureLevel::heatingSetback, .value = 13},
                  {.id = dsModel::TemperatureLevel::heatingNight, .value = 14},
                  {.id = dsModel::TemperatureLevel::heatingVacation, .value = 15}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = setpoints}});
            }
          }

          GIVEN("Cooling possible") {
            setConfigChildValue("coolingPossible", true);
            THEN("cooling levels available") {
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::coolingComfort, dsModel::TemperatureLevel::coolingEconomy,
                  dsModel::TemperatureLevel::coolingSetback, dsModel::TemperatureLevel::coolingNight,
                  dsModel::TemperatureLevel::coolingVacation};
              const std::vector<dsModel::TemperatureSetpoint> setpoints = {
                  {.id = dsModel::TemperatureLevel::coolingComfort, .value = 21},
                  {.id = dsModel::TemperatureLevel::coolingEconomy, .value = 22},
                  {.id = dsModel::TemperatureLevel::coolingSetback, .value = 23},
                  {.id = dsModel::TemperatureLevel::coolingNight, .value = 24},
                  {.id = dsModel::TemperatureLevel::coolingVacation, .value = 25}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = setpoints}});
            }
          }

          GIVEN("Passive cooling possible") {
            auto modesNode = propertySystem.createProperty("/scripts/system-addon-heating-controller/modes/1");
            auto setPassiveCoolingThreshold = [&](auto&& value) {
              REQUIRE_CALL(hooks, apartmentStructurePolled());
              modesNode->setChildValue("minimalTemperatur", ds::str(value));
            };

            setPassiveCoolingThreshold(8.0);
            setConfigChildValue("passiveCoolingPossible", true);
            THEN("Passive cooling levels available") {
              const std::vector<dsModel::TemperatureLevel> levels = {
                  dsModel::TemperatureLevel::passiveCoolingThreshold};
              const std::vector<dsModel::TemperatureSetpoint> threshold = {
                  {.id = dsModel::TemperatureLevel::passiveCoolingThreshold, .value = 8}};
              CHECK(dsModel->temperatureApplication(dsModel::ZoneId{ds::str(zone1.getID())}) ==
                    dsModel::Application{.id = dsModel::ApplicationId{"temperature-z1"},
                                         .attributes = dsModel::TemperatureApplicationAttributes{
                                             .applicationType = dsModel::ZoneApplicationType::temperature,
                                             .zone = dsModel::ZoneId{ds::str(zone1.getID())},
                                             .availableLevels = levels,
                                             .setpoints = threshold}});
            }
          }
        }
      }
      SECTION("write settings") {
        GIVEN("Temperature application device present in zone") {
          device.setActiveGroup(ApplicationType::ControlTemperature);
          THEN("set new setpoint value") {
            struct X {
              dsModel::TemperatureLevel level;
              double value;
              std::string levelName;
            };
            const auto& x = GENERATE(values(std::initializer_list<X>{
                {.level = dsModel::TemperatureLevel::heatingComfort, .value = 22.3, .levelName = "Comfort"},
                {.level = dsModel::TemperatureLevel::heatingEconomy, .value = 23.4, .levelName = "Economy"},
                {.level = dsModel::TemperatureLevel::heatingSetback, .value = 24.5, .levelName = "NotUsed"},
                {.level = dsModel::TemperatureLevel::heatingNight, .value = 25, .levelName = "Night"},
                {.level = dsModel::TemperatureLevel::heatingVacation, .value = 26, .levelName = "Holiday"},
                {.level = dsModel::TemperatureLevel::coolingComfort, .value = 28.2, .levelName = "Cooling"},
                {.level = dsModel::TemperatureLevel::coolingEconomy, .value = 29.3, .levelName = "CoolingEconomy"},
                {.level = dsModel::TemperatureLevel::coolingSetback, .value = 30, .levelName = "CoolingNotUsed"},
                {.level = dsModel::TemperatureLevel::coolingNight, .value = 31.7, .levelName = "CoolingNight"},
                {.level = dsModel::TemperatureLevel::coolingVacation, .value = 32, .levelName = "CoolingHoliday"},
            }));
            REQUIRE_CALL(eventSink, pushEvent(_))
                .LR_SIDE_EFFECT(CHECK(*_1 == Event{"heating-controller.operation-mode",
                                                   Properties{{"actions", "setTemperatureValues"},
                                                              {"zoneID", "1"},
                                                              {x.levelName, ds::str(x.value)}}}));
            const dsModel::TemperatureSetpointArgs& args = {.applicationId = dsModel::ApplicationId{"temperature-z1"},
                                                            .temperatureLevel = x.level};
            dsModel->setTemperatureSetpoint(args, x.value);
          }
          THEN("set new passive cooling threshold value") {
            REQUIRE_CALL(eventSink, pushEvent(_))
                .LR_SIDE_EFFECT(CHECK(
                    *_1 == Event{"heating-controller.passive-cooling",
                                 Properties{{"actions", "setMinimalTemperature"}, {"zoneID", "1"}, {"value", "25.3"}}}));
            const dsModel::TemperatureSetpointArgs& args = {
                .applicationId = dsModel::ApplicationId{"temperature-z1"},
                .temperatureLevel = dsModel::TemperatureLevel::passiveCoolingThreshold};
            dsModel->setTemperatureSetpoint(args, 25.3);
          }
        }
      }
    }
  }

  GIVEN("device in zone") {
    auto& zone = *apartment.allocateZone(1);
    const auto zoneId = zone.id();

    const auto deviceId = dsModel::DeviceId{"1111111111111111111111111111111100"s};
    auto& device = *apartment.allocateDevice(Dsuid{deviceId});
    device.setButtonInputMode(ButtonInputMode::ONE_WAY);
    device.setButtonInputCount(1);
    zone.addDevice({device.sharedFromThis(), &apartment});

    /// hidden TNY device
    auto hiddenDeviceId = dsModel::DeviceId{"3333333333333333333333333333333300"};
    auto& hiddenDevice = *apartment.allocateDevice(Dsuid{hiddenDeviceId});
    hiddenDevice.setPartiallyFromSpec({.ShortAddress = {}, .FunctionID = {}, .ProductID = (10 << 10)}); // TNY
    hiddenDevice.setVisibility(false);
    zone.addDevice({hiddenDevice.sharedFromThis(), &apartment});

    THEN("check device()") {
      CHECK(dsModel->device(deviceId) == dsModel::DSDevice{.id = deviceId,
                                                           .type = "dsDevice",
                                                           .lastChanged = {},
                                                           .attributes = {.name = {},
                                                                          .model = {},
                                                                          .gtin = {},
                                                                          .dsid = {},
                                                                          .displayId = "11111111…",
                                                                          .present = false,
                                                                          .firstSeen = {},
                                                                          .lastDiscovered = {},
                                                                          .inactiveSince = {},
                                                                          .shadeAttributes = {},
                                                                          .submodules = {device.submoduleId()},
                                                                          .zone = zoneId}});
    }

    THEN("check devices()") {
      CHECK(dsModel->devices() == std::vector<dsModel::DSDevice>{dsModel->device(device.id())});
    }

    THEN("apartment links devices") {
      const auto dsApartment = dsModel->apartment();
      CHECK(dsApartment.attributes.dsDevices == std::vector<dsModel::DeviceId>{deviceId});
      CHECK(dsApartment.included.dsDevices == dsModel->devices());
    }

    THEN("check zone()") {
      CHECK(dsModel->zone(zoneId) == (dsModel::Zone{.id = zoneId,
                                                    .type = "zone",
                                                    .lastChanged = {},
                                                    .attributes = {.name = {},
                                                                   .floor = "0",
                                                                   .orderId = 0,
                                                                   .submodules = {device.submoduleId()},
                                                                   .applications = {},
                                                                   .applicationTypes = {}}}));
    }

    THEN("check zones()") {
      CHECK(dsModel->zones() ==
            std::vector<dsModel::Zone>{dsModel->zone(zoneId), dsModel->zone(dsModel::ZoneId{"65534"})});
    }

    THEN("apartment links zones") {
      const auto dsApartment = dsModel->apartment();
      CHECK(dsApartment.attributes.zones == std::vector<dsModel::ZoneId>{zoneId, dsModel::ZoneId{"65534"}});
      CHECK(dsApartment.included.zones == dsModel->zones());
    }

    WHEN("set device parameters") {
      device.setName("test device");
      device.setActiveGroup(ApplicationType::Lights);

      THEN("check submodule()") {
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{
                  .id = device.submoduleId(),
                  .type = "submodule",
                  .lastChanged = {},
                  .attributes = {.name = "test device",
                                 .technicalName = {},
                                 .dsDevice = {device.id()},
                                 .functionBlocks = {device.functionBlockId()},
                                 .zone = zoneId,
                                 .application = ds::fromEnum<std::string>(dsModel::ZoneApplicationType::lights)}}));
      }

      THEN("check submodules()") {
        CHECK(dsModel->submodules() == std::vector<dsModel::Submodule>{dsModel->submodule(device.submoduleId())});
      }

      THEN("apartment includes submodules") {
        const auto dsApartment = dsModel->apartment();
        CHECK(dsApartment.included.submodules == dsModel->submodules());
      }

      THEN("check functionBlock()") {
        auto button = getButtons(device);
        CHECK(dsModel->functionBlock(device.functionBlockId()) ==
              (dsModel::FunctionBlock{.id = device.functionBlockId(),
                                      .type = "functionBlock",
                                      .lastChanged = {},
                                      .attributes = {
                                          .name = "test device",
                                          .technicalName = "",
                                          .active = 1,
                                          .outputs = {},
                                          .buttonInputs = button,
                                          .sensorInputs = {},
                                          .submodule = device.submoduleId(),
                                          .deviceAdapter = "1111111111111111111111111111111100",
                                      }}));
      }

      THEN("check functionBlocks") {
        CHECK(dsModel->functionBlocks() ==
              std::vector<dsModel::FunctionBlock>{dsModel->functionBlock(device.functionBlockId())});
      }

      THEN("apartment links functionBlocks") {
        const auto dsApartment = dsModel->apartment();
        CHECK(dsApartment.included.functionBlocks == dsModel->functionBlocks());
      }

      THEN("legacy awnings") {
        device.setActiveGroup(ApplicationType::Blinds);
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{
                  .id = device.submoduleId(),
                  .type = "submodule",
                  .lastChanged = {},
                  .attributes = {.name = "test device",
                                 .technicalName = {},
                                 .dsDevice = {device.id()},
                                 .functionBlocks = {device.functionBlockId()},
                                 .zone = zoneId,
                                 .application = ds::fromEnum<std::string>(dsModel::ZoneApplicationType::shades)}}));

        device.setHardwareType(HardwareType::GR_KL210);
        CHECK(dsModel->submodule(device.submoduleId()) ==
              (dsModel::Submodule{
                  .id = device.submoduleId(),
                  .type = "submodule",
                  .lastChanged = {},
                  .attributes = {.name = "test device",
                                 .technicalName = {},
                                 .dsDevice = {device.id()},
                                 .functionBlocks = {device.functionBlockId()},
                                 .zone = zoneId,
                                 .application = ds::fromEnum<std::string>(dsModel::ZoneApplicationType::awnings)}}));
      }
    }

    WHEN("device buttonID set") {
      device.setButtonID(1);
      device.setButtonUsage(ButtonUsage::used);
      THEN("check buttons") {
        CHECK(getButtons(device)[0] == (dsModel::Button{.id = "button0",
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Pushbutton#0",
                                                            .type = dsModel::ButtonType::area1,
                                                            .mode = dsModel::ButtonMode::button1way,
                                                            .channel = {},
                                                        }}));
      }
    }

    WHEN("device buttonID set but unused") {
      device.setButtonID(1);
      device.setButtonUsage(ButtonUsage::auto_unused);
      THEN("check buttons") {
        CHECK(getButtons(device)[0] == (dsModel::Button{.id = "button0",
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Pushbutton#0",
                                                            .type = dsModel::ButtonType::area1,
                                                            .mode = dsModel::ButtonMode::disabled,
                                                            .channel = {},
                                                        }}));
      }
    }

    WHEN("device activeGroup and buttonID set") {
      device.setButtonID(14);
      device.setButtonUsage(ButtonUsage::used);

      WHEN("activeGroup in range 16 .. 39") {
        device.setActiveGroup(20);

        THEN("check buttons") {
          CHECK(getButtons(device)[0] == (dsModel::Button{.id = "button0",
                                                          .attributes = {
                                                              .name = "",
                                                              .technicalName = "Pushbutton#0",
                                                              .type = dsModel::ButtonType::group,
                                                              .mode = dsModel::ButtonMode::button1way,
                                                              .channel = {},
                                                          }}));
        }
      }

      WHEN("activeGroup in range 64 .. 239") {
        device.setActiveGroup(100);

        THEN("check buttons") {
          CHECK(getButtons(device)[0] == (dsModel::Button{.id = "button0",
                                                          .attributes = {
                                                              .name = "",
                                                              .technicalName = "Pushbutton#0",
                                                              .type = dsModel::ButtonType::application,
                                                              .mode = dsModel::ButtonMode::button1way,
                                                              .channel = {},
                                                          }}));
        }
      }
    }

    WHEN("device.activeGroup=light") {
      device.setActiveGroup(ApplicationType::Lights);

      THEN("check dsModel") {
        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{.id = zoneId,
                             .type = "zone",
                             .lastChanged = {},
                             .attributes = {.name = {},
                                            .floor = "0",
                                            .orderId = 0,
                                            .submodules = {device.submoduleId()},
                                            .applications = {dsModel::ZoneApplicationType::lights},
                                            .applicationTypes = {dsModel::ZoneApplicationType::lights}}}));
      }
    }

    WHEN("Sonos device in zone") {
      zone.getGroup(ColorIDCyan)->addConnectedDevice();

      THEN("check dsModel") {
        CHECK(dsModel->zone(zoneId) ==
              (dsModel::Zone{.id = zoneId,
                             .type = "zone",
                             .lastChanged = {},
                             .attributes = {.name = {},
                                            .floor = "0",
                                            .orderId = 0,
                                            .submodules = {device.submoduleId()},
                                            .applications = {dsModel::ZoneApplicationType::audio},
                                            .applicationTypes = {dsModel::ZoneApplicationType::audio}}}));
      }
    }

    WHEN("sensor inputs set") {
      std::vector<DeviceSensorSpec_t> deviceSensorInputs;
      for (uint i = 0; i < 2; i++) {
        deviceSensorInputs.push_back(DeviceSensorSpec_t{});
        auto& x = deviceSensorInputs.back();
        x.sensorType = static_cast<SensorType>(i + 4);
      }
      device.setSensors(deviceSensorInputs);
      device.setSensorsInfo(0, "", "", 1);
      device.setSensorsInfo(1, "", "", 2);

      THEN("check sensors") {
        CHECK(getSensors(device)[0] == (dsModel::Sensor{.id = "0",
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Sensor#0",
                                                            .type = dsModel::MeasurementType::activePower,
                                                            .usage = dsModel::MeasurementUsage::zone,
                                                            .min = {},
                                                            .max = {},
                                                            .resolution = {},
                                                        }}));
        CHECK(getSensors(device)[1] == (dsModel::Sensor{.id = "1",
                                                        .attributes = {
                                                            .name = "",
                                                            .technicalName = "Sensor#1",
                                                            .type = dsModel::MeasurementType::outputCurrent,
                                                            .usage = dsModel::MeasurementUsage::outdoor,
                                                            .min = {},
                                                            .max = {},
                                                            .resolution = {},
                                                        }}));
      }
    }

    SECTION("device states") {
      std::vector<dsModel::DeviceState> states{};
      std::vector<dsModel::DeviceStateStatus> stateStatuses{};

      dsModel::DeviceState state{
          .id = dsModel::DeviceStateId{"stateId"}, .options = {}, .name = {.domain = {}, .message = "stateName"}};
      dsModel::DeviceStateOption options{.id = "optionId", .name = {.domain = {}, .message = "optionName"}};
      state.options.push_back(options);
      dsModel::DeviceStateStatus stateStatus{.id = dsModel::DeviceStateId{"stateStatusId"}, .value = "stateStatusValue"};

      states.push_back(state);
      stateStatuses.push_back(stateStatus);
      device.setStates(std::move(states));
      device.setStatesStatus(std::move(stateStatuses));

      CHECK(dsModel->device(deviceId) == dsModel::DSDevice{.id = deviceId,
                                                           .type = "dsDevice",
                                                           .lastChanged = {},
                                                           .attributes = {.name = {},
                                                                          .model = {},
                                                                          .gtin = {},
                                                                          .dsid = {},
                                                                          .displayId = "11111111…",
                                                                          .present = false,
                                                                          .firstSeen = {},
                                                                          .lastDiscovered = {},
                                                                          .inactiveSince = {},
                                                                          .shadeAttributes = {},
                                                                          .submodules = {device.submoduleId()},
                                                                          .zone = zoneId,
                                                                          .scenarios = {},
                                                                          .states = device.states()}});

      CHECK(dsModel->deviceStatus(deviceId) ==
            dsModel::DSDeviceStatus{.id = {},
                                    .attributes = {.functionBlocks = {}, .states = device.statesStatus()}});
    }

    SECTION("outputs smoke test") {}

    SECTION("output types") {
      device.setOutputChannels({
          ChannelType::brightness, ChannelType::hue,
          //             ChannelType::saturation,
          //             ChannelType::colortemp,
          //             ChannelType::x,
          //             ChannelType::y,
          //             ChannelType::shadePositionOutside,
          //             ChannelType::shadePositionIndoor,
          //             ChannelType::shadeOpeningAngleOutside,
          //             ChannelType::shadeOpeningAngleIndoor,
          //             ChannelType::transparency,
          //             ChannelType::airFlowIntensity,
          //             ChannelType::airFlowDirection,
          //             ChannelType::airFlapPosition,
          //             ChannelType::airLouverPosition,
          //             ChannelType::heatingPower,
          //             ChannelType::coolingCapacity,
          //             ChannelType::audioVolume,
          //             ChannelType::powerState,
          //             ChannelType::airLouverAuto,
      });
      device.setOutputMode(16);
      auto outputs = getOutputs(device, std::nullopt);
      CHECK(getOutputs(device, std::nullopt) ==
            std::vector<dsModel::Output>{{.id = "brightness",
                                          .attributes = {.name = {},
                                                         .technicalName = "brightness",
                                                         .type = dsModel::OutputType::lightBrightness,
                                                         .function = {},
                                                         .min = 0,
                                                         .max = 100,
                                                         .resolution = 0.1,
                                                         .mode = dsModel::OutputMode::switched,
                                                         .levels = 1}},
                                         {.id = "hue",
                                          .attributes = {.name = {},
                                                         .technicalName = "hue",
                                                         .type = dsModel::OutputType::lightHue,
                                                         .function = {},
                                                         .min = 0,
                                                         .max = 360,
                                                         .resolution = 0.1,
                                                         .mode = dsModel::OutputMode::switched,
                                                         .levels = 1}}});
    }

    SECTION("output modes") {
      device.setOutputChannels({ChannelType::brightness});
      struct X {
        std::vector<int> dssModes;
        dsModel::OutputMode dsMode;
      };
      for (const auto& x : std::initializer_list<X>{
               {.dssModes = {255, 1}, .dsMode = {}},
               {.dssModes = {16, 25, 26, 27, 32, 34, 35, 36, 37, 38, 39, 40, 41, 43, 65, 66, 67, 68},
               .dsMode = dsModel::OutputMode::switched},
               {.dssModes = {33, 42}, .dsMode = dsModel::OutputMode::positional},
               {.dssModes = {17, 18, 19, 20, 21, 22, 23, 24, 31, 48, 49, 50, 51, 64},
               .dsMode = dsModel::OutputMode::gradual},
               {.dssModes = {0}, .dsMode = dsModel::OutputMode::disabled},
           }) {
        for (const auto& dssMode : x.dssModes) {
          INFO(dssMode);
          INFO(x.dsMode);
          device.setOutputMode(dssMode);
          CHECK(getOutputs(device, std::nullopt).at(0).attributes.mode == x.dsMode);
        }
      }
    }

    SECTION("setOutputMode propagates to bus") {
      device.setOutputChannels(
          {ChannelType::brightness, ChannelType::hue, ChannelType::colortemp, ChannelType::shadePositionOutside});
      device.setOutputMode(16);
      SECTION("minimal brightness") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 0, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0);
      }
      SECTION("maximal brightness") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 0, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 100);
      }
      SECTION("minimal hue") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 1, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "hue"}, 0);
      }
      SECTION("maximal hue") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 1, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "hue"}, 360);
      }
      SECTION("minimal colortemp") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 2, 8, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "colortemp"}, 100);
      }
      SECTION("maximal colortemp") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 2, 8, 255, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "colortemp"}, 1000);
      }
      SECTION("minimal shadePositionOutside") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 3, 16, 0, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 0);
      }
      SECTION("maximal shadePositionOutside") {
        REQUIRE_CALL(busInterface.device, setDeviceOutputChannelValue(_, 3, 16, 65535, true)).LR_WITH(_1 == device);
        dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 100);
      }
    }

    SECTION("setOutputValue on joker configured as light device") {
      SECTION("joker as light") {
        device.setHardwareType(HardwareType::SW_KL200);
        device.setActiveGroup(ColorIDYellow);
        device.setOutputChannels({});

        SECTION("minimal brightness") {
          REQUIRE_CALL(busInterface.device, setValue(_, 0)).LR_WITH(_1 == device);
          dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0);
        }
        SECTION("maximal brightness") {
          REQUIRE_CALL(busInterface.device, setValue(_, 255)).LR_WITH(_1 == device);
          dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 100);
        }
      }
      SECTION("joker as shade") {
        device.setHardwareType(HardwareType::SW_KL200);
        device.setActiveGroup(ColorIDGray);
        device.setOutputChannels({});

        SECTION("try to set brightness channel type output") {
          CHECK_THROWS_FIND(dsModel->setOutputValue({.deviceId = deviceId, .outputId = "brightness"}, 0),
                            "Could not find item");
        }
        SECTION("try to set shadePositionOutside channel type output") {
          CHECK_THROWS_FIND(dsModel->setOutputValue({.deviceId = deviceId, .outputId = "shadePositionOutside"}, 0),
                            "Could not find item");
        }
      }
    }

    SECTION("powerline output devices have output channels") {
      GIVEN("device") {
        enum Ds485 {
          setOutputValue,
          setDeviceConfig64_3_2,
          setDeviceConfig64_4,
        };
        struct Output {
          dss::ChannelType dssType;
          dsModel::Output ds;
          Ds485 ds485;
        };
        static const Output lightBrightness{
            .dssType = ChannelType::brightness,
            .ds = {.id = "brightness",
                   .attributes = {.name = {},
                                  .technicalName = "brightness",
                                  .type = dsModel::OutputType::lightBrightness,
                                  .function = {},
                                  .min = 0,
                                  .max = 100,
                                  .resolution = 0.1,
                                  .mode = dsModel::OutputMode::switched,
                                  .levels = 1,
                                  .levelThresholds = std::optional<std::vector<double>>({50.0, 75.0})}},
            .ds485 = Ds485::setOutputValue};
        static const Output shadePositionOutside{
            .dssType = ChannelType::shadePositionOutside,
            .ds = {.id = "shadePositionOutside",
                   .attributes = {.name = {},
                                  .technicalName = "shadePositionOutside",
                                  .type = dsModel::OutputType::shadePositionOutside,
                                  .function = {},
                                  .min = 0,
                                  .max = 100,
                                  .resolution = 0.1,
                                  .mode = dsModel::OutputMode::switched,
                                  .levels = 1,
                                  .levelThresholds = std::optional<std::vector<double>>({50.0, 75.0})}},
            .ds485 = Ds485::setDeviceConfig64_3_2};
        static const Output shadeOpeningAngleOutside{
            .dssType = ChannelType::shadeOpeningAngleOutside,
            .ds = {.id = "shadeOpeningAngleOutside",
                   .attributes = {.name = {},
                                  .technicalName = "shadeOpeningAngleOutside",
                                  .type = dsModel::OutputType::shadeOpeningAngleOutside,
                                  .function = {},
                                  .min = 0,
                                  .max = 100,
                                  .resolution = 0.1,
                                  .mode = dsModel::OutputMode::switched,
                                  .levels = 1,
                                  .levelThresholds = std::optional<std::vector<double>>({50.0, 75.0})}},
            .ds485 = Ds485::setDeviceConfig64_4};

        struct X {
          HardwareType hardwareType;
          std::vector<Output> outputs;
        };
        const auto& x = GENERATE(values(std::initializer_list<X>{
            {.hardwareType = HardwareType::GE_KM200, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_TKM210, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_KM300, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_TKM300, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDM200, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDM200JS, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDM200CS, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDS200CW, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDS200CS, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_SDS220CT, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_KL200, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GE_UMV200, .outputs = {lightBrightness}},
            {.hardwareType = HardwareType::GR_KL200, .outputs = {shadePositionOutside}},
            {.hardwareType = HardwareType::GR_KL210, .outputs = {shadePositionOutside}},
            {.hardwareType = HardwareType::GR_KL220, .outputs = {shadePositionOutside, shadeOpeningAngleOutside}},
            {.hardwareType = HardwareType::GR_HKL230, .outputs = {shadePositionOutside, shadeOpeningAngleOutside}},

        }));
        INFO(x.hardwareType);
        device.setOutputMode(16);
        device.setHardwareType(x.hardwareType);
        device.setOutputChannels({});
        THEN("Outputs are matching") {
          auto dsOutputs = std::vector<dsModel::Output>{};
          for (const auto& output : x.outputs) {
            dsOutputs.push_back(output.ds);
          }
          CHECK(getOutputs(device, std::optional<std::vector<double>>({50.0, 75.0})) == dsOutputs);
        }
        THEN("setOutputValue propagates to ds485 bus") {
          for (const auto& output : x.outputs) {
            const auto& dsOutput = output.ds;
            INFO(dsOutput);
            auto value = (*dsOutput.attributes.min + *dsOutput.attributes.max * 2) / 3;
            switch (output.ds485) {
              case Ds485::setOutputValue: {
                REQUIRE_CALL(busInterface.device, setValue(_, 170)).LR_WITH(_1 == device);
                dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
                break;
              }
              case Ds485::setDeviceConfig64_3_2: {
                REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 3, 170)).LR_WITH(_1 == device);
                REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 2, 170)).LR_WITH(_1 == device);
                dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
                break;
              }
              case Ds485::setDeviceConfig64_4: {
                REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 64, 4, 170)).LR_WITH(_1 == device);
                dsModel->setOutputValue({.deviceId = deviceId, .outputId = dsOutput.id}, value);
                break;
              }
            }
          }
        }
      }
    }
  }

  GIVEN("userDefinedStates") {
    auto udssNode = propertySystem.createProperty("/usr/addon-states/system-addon-user-defined-states/");

    const auto id = "id0"s;
    const auto name = "name0"s;
    const auto setName = "set"s;
    const auto resetName = "reset"s;
    REQUIRE_CALL(hooks, apartmentStructurePolled());
    auto udsNode = udssNode->createProperty(id);
    udsNode->setChildValue("name", id);
    udsNode->setChildValue("displayName", name);
    udsNode->setChildValue("setName", setName);
    udsNode->setChildValue("resetName", resetName);
    udsNode->setChildValue("showOnPhone", true);
    auto& state = *apartment.allocateState(StateType_Script, id, "system-addon-user-defined-states");

    SECTION("structure") {
      const auto& [type, typeAPI] =
          GENERATE(values<std::pair<std::string, std::string>>({{"custom-states", "manualState"},
                                                                {"zone-sensor-states", "measurementState"},
                                                                {"consumption-states", "consumptionState"},
                                                                {"power-states", "circuitConsumptionState"},
                                                                {"triggered-states", "triggerState"},
                                                                {"combined-states", "combinedState"},
                                                                {"window-states", "windowState"}}));
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      udsNode->setChildValue("type", type);
      CHECK(dsModel->userDefinedStates() ==
            std::vector<dsModel::UserDefinedState>{
                {.id = id,
                 .type = typeAPI,
                 .lastChanged = {},
                 .attributes = dsModel::UserDefinedStateAttributes{
                     .name = name, .setName = setName, .resetName = resetName, .visibleForUsers = true}}});
    }

    SECTION("user defined state is removed") {
      REQUIRE_CALL(hooks, apartmentStructurePolled());
      udssNode->removeChild(udsNode);
      CHECK(dsModel->userDefinedStates() == std::vector<dsModel::UserDefinedState>{});
    }

    SECTION("state") {
      const auto& [active, activeEnum] = GENERATE(values<std::pair<std::string, dsModel::UserDefinedStateStatusStatus>>(
          {{"active", dsModel::UserDefinedStateStatusStatus::active},
           {"inactive", dsModel::UserDefinedStateStatusStatus::inactive},
           {"undefined", dsModel::UserDefinedStateStatusStatus::undefined}}));
      udsNode->setChildValue("state", active);
      CHECK(dsModel->userDefinedStateStatuses() ==
            std::vector<dsModel::UserDefinedStateStatus>{
                {.id = id, .type = "userDefinedStateStatus", .lastChanged = {}, .attributes = {.status = activeEnum}}});
    }

    SECTION("change state") {
      SECTION("success") {
        udsNode->setChildValue("type", "custom-states"s);
        dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "active");
        CHECK(state.getState() == State_Active);
        dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "inactive");
        CHECK(state.getState() == State_Inactive);
      }
      SECTION("wrong value") {
        udsNode->setChildValue("type", "custom-states"s);
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "balab"),
                          "Requested state value is illegal");
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = "wrongId"}, "active"),
                          "State is unknown");
      }
      SECTION("no manual state") {
        const auto& type = GENERATE(values<std::string>({"zone-sensor-states", "consumption-states", "power-states",
                                                         "triggered-states", "combined-states", "window-states"}));
        udsNode->setChildValue("type", type);
        CHECK_THROWS_FIND(dsModel->setUserDefinedStateStatus({.userDefinedStateId = id}, "active"),
                          "Only a manualState can be set");
      }
      SECTION("poll apartment status") {
        // State does not raise event in this mocked setup without DSS instance
        // We similate the state event directly
        REQUIRE_CALL(hooks, apartmentStatusPolled());
        dsModel->handleEvent(*createStateChangeEvent(state.shared_from_this(), {}, {}));
      }
    }
  }

  SECTION("floors") {
    auto& floor = *apartment.getFloor(0);
    floor.setName("floor0");
    auto& zone1 = *apartment.allocateZone(1);

    SECTION("floor") {
      CHECK(dsModel->floor({.floorId = floor.id()}) ==
            dsModel::Floor{.id = dsModel::FloorId{"0"},
                           .type = "floor",
                           .lastChanged = {},
                           .attributes = {.name = "floor0", .orderId = 0, .zones = {zone1.id()}}});
    }

    SECTION("floors") {
      CHECK(dsModel->floors() ==
            dsModel::Floors{.floors = {{.id = dsModel::FloorId{"0"},
                                        .type = "floor",
                                        .lastChanged = {},
                                        .attributes = {.name = "floor0", .orderId = 0, .zones = {zone1.id()}}}}});
    }

    SECTION("setFloorName") {
      THEN("success") {
        dsModel->setFloorName({.floorId = floor.id()}, "floor1x");
        CHECK(floor.getName() == "floor1x");
      }

      THEN("Empty name fails") {
        CHECK_THROWS_FIND(dsModel->setFloorName({.floorId = floor.id()}, ""), "New name cannot be empty");
      }

      THEN("Invalid zoneId fails") {
        CHECK_THROWS_FIND(dsModel->setFloorName({.floorId = dsModel::FloorId{"455"}}, "xyz"), "Floor not found");
      }
    }
  }

  SECTION("clusters") {
    auto cluster16 = apartment.getCluster(16);
    auto cluster17 = apartment.getCluster(17);
    cluster16->setName("cluster16");
    cluster17->setName("cluster17");
    cluster16->setApplicationType(ApplicationType::Blinds);
    cluster17->setApplicationType(ApplicationType::Lights);
    const auto id16 = dsModel::ClusterId{"16"};
    const auto id17 = dsModel::ClusterId{"17"};

    SECTION("get cluster") {
      const dsModel::Cluster cluster = {.id = id16,
                                        .type = "cluster",
                                        .attributes = {.name = "cluster16",
                                                       .orderId = {},
                                                       .applicationType = dsModel::ClusterApplicationType::shades,
                                                       .submodules = {},
                                                       .application = {}}};
      CHECK(dsModel->cluster(id16) == cluster);
    }

    SECTION("get cluster status") {
      THEN("no alarms in status") {
        CHECK(dsModel->clusterStatus(id16) ==
              dsModel::ClusterStatus{
                  .id = id16, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}}});
      }
      GIVEN("wind and hail alarm set") {
        auto wind = apartment.allocateState(StateType_Apartment, "wind.group16", "");
        auto hail = apartment.allocateState(StateType_Apartment, "hail.group16", "");
        wind->setState(coUnknown, State_Active);
        hail->setState(coUnknown, State_Active);
        THEN("alarms in status") {
          CHECK(dsModel->clusterStatus(id16) ==
                dsModel::ClusterStatus{.id = dsModel::ClusterId{"16"},
                                       .type = "clusterStatus",
                                       .lastChanged = {},
                                       .attributes = {.weather = {.wind = true, .hail = true}}});
        }
      }
    }

    SECTION("all clusters") {
      std::vector<dsModel::Cluster> clusters;
      clusters.push_back(dsModel::Cluster{.id = id16,
                                          .type = "cluster",
                                          .attributes = {.name = "cluster16",
                                                         .orderId = {},
                                                         .applicationType = dsModel::ClusterApplicationType::shades,
                                                         .submodules = {},
                                                         .application = {}}});
      clusters.push_back(dsModel::Cluster{.id = id17,
                                          .type = "cluster",
                                          .attributes = {.name = "cluster17",
                                                         .orderId = {},
                                                         .applicationType = dsModel::ClusterApplicationType::lights,
                                                         .submodules = {},
                                                         .application = {}}});
      CHECK(dsModel->clusters() == clusters);
    }

    SECTION("all cluster statuses") {
      std::vector<dsModel::ClusterStatus> clusterStatuses;
      clusterStatuses.push_back(dsModel::ClusterStatus{
          .id = id16, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}}});
      clusterStatuses.push_back(dsModel::ClusterStatus{
          .id = id17, .type = "clusterStatus", .lastChanged = {}, .attributes = {.weather = {}}});
      CHECK(dsModel->clusterStatuses() == clusterStatuses);
    }

    SECTION("set cluster name") {
      const auto args = dsModel::ClusterArgs{.clusterId = id16};
      const auto newName = "{newName}";
      dsModel->setClusterName(args, newName);
      CHECK(dsModel->cluster(id16) ==
            dsModel::Cluster{.id = id16,
                             .type = "cluster",
                             .attributes = {.name = "{newName}",
                                            .orderId = {},
                                            .applicationType = dsModel::ClusterApplicationType::shades,
                                            .submodules = {},
                                            .application = {}}});
    }
  }

  SECTION("installation") {
    const auto installationId = dsModel::InstallationId{apartmentId};
    SECTION("No properties") {
      CHECK(dsModel->installation() ==
            dsModel::Installation{.id = installationId, .type = "installation", .lastChanged = {}, .attributes = {}});
    }
    SECTION("Full properties") {
      auto geodataNode = configNode->createProperty("geodata");
      geodataNode->setChildValue("longitude", "11.5"s);
      geodataNode->setChildValue("latitude", "22.5"s);
      geodataNode->setChildValue("country", "DE"s);
      geodataNode->setChildValue("location", "Berlin"s);
      CHECK(dsModel->installation() ==
            dsModel::Installation{
                .id = installationId,
                .type = "installation",
                .lastChanged = {},
                .attributes = {
                    .location = {.longitude = 11.5, .latitude = 22.5}, .countryCode = "DE", .city = "Berlin"}});
    }
  }

  SECTION("dsServer") {
    const auto serverId = dsModel::ServerId{"dsuid"};

    auto systemNode = propertySystem.createProperty("/system/");
    systemNode->setChildValue("dSUID", "dsuid"s);
    systemNode->setChildValue("dSID", "dsid"s);
    auto versionNode = propertySystem.createProperty("/system/version/");
    versionNode->setChildValue("distroVersion", "1.0"s);
    versionNode->setChildValue("model", "model"s);    
    versionNode->setChildValue("gtin", "gtin"s);

    CHECK(dsModel->dsServer() == dsModel::DSServer{.id = serverId,
                                                   .type = "dsServer",
                                                   .lastChanged = {},
                                                   .attributes = {.name = "TestName",
                                                                  .model = "model",
                                                                  .gtin = "gtin",
                                                                  .dsid = "dsid",
                                                                  .version = "1.0",
                                                                  .environment = dsModel::Environment::development}});

    SECTION("environment") {
      struct X {
        eState dssValue;
        std::optional<bool> dsValue;
      };
      const auto& x = GENERATE(values(std::initializer_list<std::pair<ds::Release, dsModel::Environment>>{
          {ds::Release::devel, dsModel::Environment::development},
          {ds::Release::testing, dsModel::Environment::alpha},
          {ds::Release::fieldtest, dsModel::Environment::beta},
          {ds::Release::production, dsModel::Environment::production}}));
      REQUIRE_CALL(loop.system, release()).RETURN(x.first);
      CHECK(dsModel->dsServer().attributes.environment == x.second);
    }
  }
}

} // namespace
