/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-model.h"
#include <ds-api/service.h>
#include <ds-connector/client-bus.h>
#include <ds-connector/subscriptions.h>
#include <ds/asio/loop.h>
#include <ds/asio/state-poller.h>
#include <ds/asio/timer.h>
#include <ds/enum.h>
#include <ds/log.h>
#include <ds/promise.h>
#include <ds/system.h>
#include <model/cluster.h>
#include <model/device.h>
#include <model/dsmeter.h>
#include <model/floor.h>
#include <model/group.h>
#include <model/set.h>
#include <model/state.h>
#include <model/zone.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/make_shared.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/sort.hpp>
#include "cache/apartment-proxy.h"
#include "cache/cache-updater.h"
#include "ds-scenarios.h"
#include "event.h"
#include "propertysystem_common_paths.h"
#include "structuremanipulator.h"
#include "util.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsModel");

std::optional<dsModel::ZoneApplicationType> DSToEnum<dsModel::ZoneApplicationType, dss::ApplicationType, void>::tryTo(
    dss::ApplicationType x) {
  switch (x) {
    case dss::ApplicationType::None: return {};
    case dss::ApplicationType::Lights: return dsModel::ZoneApplicationType::lights;
    case dss::ApplicationType::Blinds: return dsModel::ZoneApplicationType::shades;
    case dss::ApplicationType::Heating: return dsModel::ZoneApplicationType::heating;
    case dss::ApplicationType::Audio: return dsModel::ZoneApplicationType::audio;
    case dss::ApplicationType::Video: return dsModel::ZoneApplicationType::video;
    case dss::ApplicationType::Security: return dsModel::ZoneApplicationType::security;
    case dss::ApplicationType::Access: return dsModel::ZoneApplicationType::access;
    case dss::ApplicationType::Joker: return dsModel::ZoneApplicationType::joker;
    case dss::ApplicationType::Cooling: return dsModel::ZoneApplicationType::cooling;
    case dss::ApplicationType::Ventilation: return dsModel::ZoneApplicationType::ventilation;
    case dss::ApplicationType::Window: return dsModel::ZoneApplicationType::window;
    case dss::ApplicationType::Recirculation: return dsModel::ZoneApplicationType::recirculation;
    case dss::ApplicationType::ControlTemperature: return dsModel::ZoneApplicationType::temperature;
    case dss::ApplicationType::ApartmentVentilation: return {};
    case dss::ApplicationType::Awnings: return {};
    case dss::ApplicationType::ApartmentRecirculation: return {};
  }
  return {};
}

std::optional<dsModel::StateType> DSToEnum<dsModel::StateType, dss::BinaryInputType, void>::tryTo(
    dss::BinaryInputType x) {
  switch (x) {
    case dss::BinaryInputType::AppMode: return dsModel::StateType::generic;
    case dss::BinaryInputType::Presence: return dsModel::StateType::presence;
    case dss::BinaryInputType::RoomBrightness: return dsModel::StateType::brightness;
    case dss::BinaryInputType::PresenceInDarkness: return dsModel::StateType::presenceInDarkness;
    case dss::BinaryInputType::TwilightExternal: return dsModel::StateType::twilight;
    case dss::BinaryInputType::Movement: return dsModel::StateType::motion;
    case dss::BinaryInputType::MovementInDarkness: return dsModel::StateType::motionInDarkness;
    case dss::BinaryInputType::SmokeDetector: return dsModel::StateType::smoke;
    case dss::BinaryInputType::WindDetector: return dsModel::StateType::windMonitor;
    case dss::BinaryInputType::RainDetector: return dsModel::StateType::rainMonitor;
    case dss::BinaryInputType::SunRadiation: return dsModel::StateType::sunRadiation;
    case dss::BinaryInputType::RoomThermostat: return dsModel::StateType::roomThermostat;
    case dss::BinaryInputType::BatteryLow: return dsModel::StateType::batteryLow;
    case dss::BinaryInputType::WindowContact: return dsModel::StateType::windowOpen;
    case dss::BinaryInputType::DoorContact: return dsModel::StateType::doorOpen;
    case dss::BinaryInputType::WindowTilt: return dsModel::StateType::windowHandle;
    case dss::BinaryInputType::GarageDoorContact: return dsModel::StateType::garageDoorOpen;
    case dss::BinaryInputType::SunProtection: return dsModel::StateType::sunProtection;
    case dss::BinaryInputType::FrostDetector: return dsModel::StateType::frost;
    case dss::BinaryInputType::HeatingSystem: return dsModel::StateType::heatingSystemEnabled;
    case dss::BinaryInputType::HeatingSystemMode: return dsModel::StateType::heatingChangeOver;
    case dss::BinaryInputType::PowerUp: return dsModel::StateType::initialization;
    case dss::BinaryInputType::Malfunction: return dsModel::StateType::malfunction;
    case dss::BinaryInputType::Service: return dsModel::StateType::service;
  }
  return {};
}

std::optional<dsModel::OutputType> DSToEnum<dsModel::OutputType, dss::ChannelType, void>::tryTo(dss::ChannelType x) {
  switch (x) {
    case dss::ChannelType::none: return {};
    case dss::ChannelType::brightness: return dsModel::OutputType::lightBrightness;
    case dss::ChannelType::hue: return dsModel::OutputType::lightHue;
    case dss::ChannelType::saturation: return dsModel::OutputType::lightSaturation;
    case dss::ChannelType::colortemp: return dsModel::OutputType::lightTemperature;
    case dss::ChannelType::x: return dsModel::OutputType::lightCieX;
    case dss::ChannelType::y: return dsModel::OutputType::lightCieY;
    case dss::ChannelType::shadePositionOutside: return dsModel::OutputType::shadePositionOutside;
    case dss::ChannelType::shadePositionIndoor: return dsModel::OutputType::shadePositionIndoor;
    case dss::ChannelType::shadeOpeningAngleOutside: return dsModel::OutputType::shadeOpeningAngleOutside;
    case dss::ChannelType::shadeOpeningAngleIndoor: return dsModel::OutputType::shadeOpeningAngleIndoor;
    case dss::ChannelType::transparency: return dsModel::OutputType::shadeTransparency;
    case dss::ChannelType::airFlowIntensity: return dsModel::OutputType::airFlowIntensity;
    case dss::ChannelType::airFlowDirection: return dsModel::OutputType::airFlowDirection;
    case dss::ChannelType::airFlapPosition: return dsModel::OutputType::airFlapOpeningAngle;
    case dss::ChannelType::airLouverPosition: return dsModel::OutputType::ventilationLouverPosition;
    case dss::ChannelType::heatingPower: return dsModel::OutputType::heatingPower;
    case dss::ChannelType::coolingCapacity: return dsModel::OutputType::coolingCapacity;
    case dss::ChannelType::audioVolume: return dsModel::OutputType::audioVolume;
    case dss::ChannelType::powerState: return dsModel::OutputType::powerState;
    case dss::ChannelType::airLouverAuto: return dsModel::OutputType::ventilationSwingMode;
  }
  return {};
}

std::optional<dsModel::MeasurementUsage> DSToEnum<dsModel::MeasurementUsage, int, void>::tryTo(int x) {
  switch (x) {
    case 1: return dsModel::MeasurementUsage::zone;
    case 2: return dsModel::MeasurementUsage::outdoor;
    case 3: return dsModel::MeasurementUsage::settings;
    case 4: return dsModel::MeasurementUsage::device;
    case 5: return dsModel::MeasurementUsage::deviceLastRun;
    case 6: return dsModel::MeasurementUsage::deviceAverage;
  }
  return {};
}

namespace dss {

dsModel::ZoneApplicationType zoneApplication(const int& application) {
  switch (application) {
    case 1: return dsModel::ZoneApplicationType::lights;
    case 2: return dsModel::ZoneApplicationType::shades;
    case 3: return dsModel::ZoneApplicationType::heating;
    case 4: return dsModel::ZoneApplicationType::audio;
    case 5: return dsModel::ZoneApplicationType::video;
    case 6: return dsModel::ZoneApplicationType::security;
    case 7: return dsModel::ZoneApplicationType::access;
    case 8: return dsModel::ZoneApplicationType::joker;
    case 9: return dsModel::ZoneApplicationType::cooling;
    case 10: return dsModel::ZoneApplicationType::ventilation;
    case 11: return dsModel::ZoneApplicationType::window;
    case 12: return dsModel::ZoneApplicationType::recirculation;
    case 48: return dsModel::ZoneApplicationType::temperature;
    default: DS_INFO("Incorrect zone application type.", application); break;
  }
  return {};
}

dsModel::MeasurementType measurementType(const SensorType& type) {
  switch (static_cast<int>(type)) {
    case 4: return dsModel::MeasurementType::activePower;
    case 5: return dsModel::MeasurementType::outputCurrent;
    case 6: return dsModel::MeasurementType::energyMeter;
    case 9: return dsModel::MeasurementType::temperatureIndoors;
    case 10: return dsModel::MeasurementType::temperatureOutdoors;
    case 11: return dsModel::MeasurementType::brightnessIndoors;
    case 12: return dsModel::MeasurementType::brightnessOutdoors;
    case 13: return dsModel::MeasurementType::humidityIndoors;
    case 14: return dsModel::MeasurementType::humidityOutdoors;
    case 15: return dsModel::MeasurementType::airPressure;
    case 16: return dsModel::MeasurementType::windGustSpeed;
    case 17: return dsModel::MeasurementType::windGustDirection;
    case 18: return dsModel::MeasurementType::windAverageSpeed;
    case 19: return dsModel::MeasurementType::windDirection;
    case 20: return dsModel::MeasurementType::precipitation;
    case 21: return dsModel::MeasurementType::carbonDioxide;
    case 22: return dsModel::MeasurementType::carbonMonoxide;
    case 25: return dsModel::MeasurementType::soundPressureLevel;
    case 50: return dsModel::MeasurementType::tControlSetPoint;
    case 51: return dsModel::MeasurementType::tControlVariable;
    case 60: return dsModel::MeasurementType::status;
    case 61: return dsModel::MeasurementType::reserved1;
    case 62: return dsModel::MeasurementType::reserved2;
    case 64: return dsModel::MeasurementType::outputCurrent16A;
    case 65: return dsModel::MeasurementType::apparentPower;
    case 66: return dsModel::MeasurementType::temperature;
    case 67: return dsModel::MeasurementType::brightness;
    case 68: return dsModel::MeasurementType::humidity;
    case 69: return dsModel::MeasurementType::generatedActivePower;
    case 70: return dsModel::MeasurementType::generatedEnergyMeter;
    case 71: return dsModel::MeasurementType::waterQuantity;
    case 72: return dsModel::MeasurementType::waterFlowRate;
    case 73: return dsModel::MeasurementType::length;
    case 74: return dsModel::MeasurementType::mass;
    case 75: return dsModel::MeasurementType::time;
    case 253: return dsModel::MeasurementType::notUsed;
    case 255: return dsModel::MeasurementType::unknownType;
    default: DS_INFO("Incorrect sensor type.", type);
  }
  return {};
}

dsModel::ButtonType buttonType(const int& type, const int& activeGroup) {
  switch (type) {
    case 0: return dsModel::ButtonType::device;
    case 1: return dsModel::ButtonType::area1;
    case 2: return dsModel::ButtonType::area2;
    case 3: return dsModel::ButtonType::area3;
    case 4: return dsModel::ButtonType::area4;
    case 5: return dsModel::ButtonType::zone;
    case 6: return dsModel::ButtonType::zone1;
    case 7: return dsModel::ButtonType::zone2;
    case 8: return dsModel::ButtonType::zone3;
    case 9: return dsModel::ButtonType::zone4;
    case 10: return dsModel::ButtonType::zonex1;
    case 11: return dsModel::ButtonType::zonex2;
    case 12: return dsModel::ButtonType::zonex3;
    case 13: return dsModel::ButtonType::zonex4;
    case 14:
      if (activeGroup >= 64 && activeGroup <= 239) {
        return dsModel::ButtonType::application;
      }
      if (activeGroup >= 16 && activeGroup <= 39) {
        return dsModel::ButtonType::group;
      }
      break;
    case 15: return dsModel::ButtonType::appmode;
  }
  DS_INFO("Incorrect button type.", type, activeGroup);
  return {};
}

// https://intranet.aizo.net/_layouts/15/start.aspx#/Wiki/Parametersatz.aspx#BANK3 - the MODE register at offset "0"
// https://confluence.digitalstrom.org/display/DSD/Outputs
dsModel::OutputMode outputMode(const int& mode) {
  if ((mode >= 17 && mode <= 24) || (mode == 31) || (mode == 64) || (mode >= 48 && mode <= 51)) {
    return dsModel::OutputMode::gradual;
  } else if ((mode == 33) || (mode == 42)) {
    return dsModel::OutputMode::positional;
  } else if ((mode == 16) || (mode >= 25 && mode <= 27) || (mode == 32) || (mode >= 34 && mode <= 41) || (mode == 43) ||
             (mode >= 65 && mode <= 69)) {
    return dsModel::OutputMode::switched;
  } else if (mode == 0) {
    return dsModel::OutputMode::disabled;
  }
  DS_INFO("Unrecognized output mode.", mode);
  return {};
}

std::optional<int> outputLevels(const int& mode) {
  if ((mode == 16) || (mode >= 25 && mode <= 27) || (mode == 32) || (mode == 35) || (mode >= 39 && mode <= 41) ||
      (mode == 43) || (mode == 65)) {
    return 1;
  } else if ((mode == 34) || (mode == 36) || (mode == 37) || (mode == 43) || (mode == 67) || (mode == 69)) {
    return 2;
  } else if ((mode == 38) || (mode == 68)) {
    return 3;
  }
  DS_DEBUG("Output mode has no switched levels", mode);
  return {};
}

// Basic information about button input modes:
// https://intranet.aizo.net/_layouts/15/start.aspx#/Wiki/Parametersatz.aspx#BANK3 - address 0x1e
dsModel::ButtonMode buttonMode(const ButtonInputMode& mode, const ButtonUsage& usage) {
  if (usage != ButtonUsage::used) {
    return dsModel::ButtonMode::disabled;
  }
  int id = static_cast<int>(mode);
  if (id == 0 || id == 1 || id == 14 || id == 65) {
    return dsModel::ButtonMode::button1way;
  } else if (id >= 5 && id <= 13) {
    return dsModel::ButtonMode::button2way;
  } else if (id == 255) {
    return dsModel::ButtonMode::disabled;
  }
  DS_INFO("Unrecognized button mode.", mode);
  return {};
}

dsModel::ShadeAttributes shadeAttributes(const Device& device) {
  return {.shadeType = shadeDeviceType(device), .protectionClass = {}, .mountOrientation = {}};
}

dsModel::ShadeDeviceType shadeDeviceType(const Device& device) {
  switch (device.hardwareType()) {
    case HardwareType::GR_KL200: return dsModel::ShadeDeviceType::rollerShutter;
    case HardwareType::GR_KL210: return dsModel::ShadeDeviceType::awningTerrace;
    case HardwareType::GR_KL220: return dsModel::ShadeDeviceType::windowBlind;
    case HardwareType::GR_HKL230: return dsModel::ShadeDeviceType::windowBlind;
    case HardwareType::GE_KM200:
    case HardwareType::GE_TKM210:
    case HardwareType::GE_SDM200:
    case HardwareType::GE_SDM200JS:
    case HardwareType::GE_SDM200CS:
    case HardwareType::GE_SDS200CW:
    case HardwareType::GE_SDS200CS:
    case HardwareType::GE_SDS220CT:
    case HardwareType::GE_TKM220:
    case HardwareType::GE_TKM230:
    case HardwareType::GE_KL200:
    case HardwareType::GE_UMV200:
    case HardwareType::GE_UMV210:
    case HardwareType::GE_KM300:
    case HardwareType::GE_TKM300:
    case HardwareType::GN_KM200:
    case HardwareType::GN_TKM200:
    case HardwareType::GN_TKM210:
    case HardwareType::GR_TKM200:
    case HardwareType::GR_TKM210:
    case HardwareType::BL_KM200:
    case HardwareType::BL_SDS200:
    case HardwareType::RT_TKM200:
    case HardwareType::RT_SDM200:
    case HardwareType::SW_TKM200:
    case HardwareType::SW_TKM210:
    case HardwareType::SW_AKM200:
    case HardwareType::SW_AKM210:
    case HardwareType::SW_AKM220:
    case HardwareType::SW_KL200:
    case HardwareType::SW_ZWS200:
    case HardwareType::SW_ZWS200F:
    case HardwareType::SW_ZWS200EF:
    case HardwareType::SW_SSL200JS:
    case HardwareType::SW_SSL200FS:
    case HardwareType::SW_SDS200CW:
    case HardwareType::SW_SDS200CS:
    case HardwareType::SW_SDS220CT:
    case HardwareType::SW_TNY200:
    case HardwareType::SW_UMR200:
    case HardwareType::SW_SK200:
    case HardwareType::SW_SK201:
    case HardwareType::SW_SK202:
    case HardwareType::BL_SK204:
    case HardwareType::SW_ZWS205:
    case HardwareType::UNKNOWN:
    default: break;
  }
  DS_DEBUG("Unrecognized shade type.", device.getProductID());
  return {};
}

std::vector<dsModel::Sensor> getSensors(const Device& device) {
  std::vector<dsModel::Sensor> result = {};
  int sensorCount = 0;
  for (const auto& sensor : device.getSensors()) {
    const auto sensorId = ds::str(sensorCount);
    DS_CONTEXT(sensorId);
    try {
      result.push_back({.id = sensorId,
                        .attributes = {
                            .name = "",
                            .technicalName = ds::str("Sensor#", sensorCount),
                            .type = measurementType(sensor->m_sensorType),
                            .usage = ds::toEnum<dsModel::MeasurementUsage>((int)sensor->m_sensorUsage),
                            .min = {},
                            .max = {},
                            .resolution = {}, // TODO: Find detailed information about those values
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Sensor failed.", e);
      result.push_back({.id = sensorId});
    }
    sensorCount++;
  };
  return result;
}

std::vector<dsModel::Button> getButtons(const Device& device) {
  std::vector<dsModel::Button> result = {};
  if (device.getButtonID() != 255 && device.getButtonInputCount() != 0) {
    const auto buttonId = ds::str("button", device.getButtonInputIndex());
    DS_CONTEXT(buttonId);
    try {
      result.push_back({.id = buttonId,
                        .attributes = {
                            .name = "",
                            .technicalName = ds::str("Pushbutton#0"),
                            .type = buttonType(device.getButtonID(), device.getActiveGroup()),
                            .mode = buttonMode(device.getButtonInputMode(), device.getButtonUsage()),
                            .channel = {},
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Button failed.", e);
      result.push_back({.id = buttonId});
    }
  }
  return result;
}

std::vector<dsModel::State> getStates(const Device& device) {
  std::vector<dsModel::State> result = {};
  int stateCount = 0;
  for (const auto& state : device.getBinaryInputs()) {
    const auto stateId = ds::str(stateCount);
    DS_CONTEXT(stateId);
    try {
      result.push_back({.id = stateId,
                        .attributes = {
                            .name = "",
                            .technicalName = ds::str("BinaryInput#", stateCount),
                            .type = ds::toEnum<dsModel::StateType>(state->m_inputType),
                            .mode = {},
                            .turnOnDelay = {},
                            .turnOffDelay = {},
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("State failed.", e);
      result.push_back({.id = stateId});
    }
    stateCount++;
  };
  return result;
}

std::vector<dsModel::Output> getOutputs(const Device& device, std::optional<std::vector<double>> levelThresholds) {
  auto result = std::vector<dsModel::Output>{};
  auto channels = device.getOutputChannels();
  for (const auto& channel : channels) {
    const auto& channelId = channel->id;
    DS_CONTEXT(channelId);
    try {
      result.push_back({.id = channelId,
                        .attributes = {
                            .name = "",
                            .technicalName = channel->technicalName,
                            .type = ds::toEnum<dsModel::OutputType>(channel->type),
                            .function = {},
                            .min = [&]() -> double {
                              if (channel->type == ChannelType::colortemp) {
                                return 100;
                              } else {
                                return 0;
                              }
                            }(),
                            .max = [&]() -> double {
                              if (channel->type == ChannelType::hue) {
                                return 360;
                              } else if (channel->type == ChannelType::colortemp) {
                                return 1000;
                              } else if (channel->type == ChannelType::x || channel->type == ChannelType::y) {
                                return 1;
                              } else {
                                return 100;
                              }
                            }(),
                            .resolution =
                                [&] {
                                  if (channel->type == ChannelType::colortemp) {
                                    return 1.0;
                                  } else if (channel->type == ChannelType::x || channel->type == ChannelType::y) {
                                    return 0.01;
                                  } else {
                                    return 0.1;
                                  }
                                }(),
                            .mode = outputMode(device.getOutputMode()),
                            .levels = outputLevels(device.getOutputMode()),
                            .levelThresholds = levelThresholds,
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Output failed.", e);
      result.push_back({.id = channelId});
    }
  };
  return result;
}


DecodedMeteringId decodeId(const dsModel::MeteringId& meteringId) {
  auto result = DecodedMeteringId{};
  std::stringstream ss(meteringId);

  std::string originType;
  std::getline(ss, originType, '-');
  if (originType != "dsm" && originType != "apartment") {
    DS_FAIL_REQUIRE("Unsupported measuring device.", originType);
  }
  result.origin.type = originType;

  if (originType == "dsm") {
    std::string meterId;
    std::getline(ss, meterId, '-');
    result.origin.id = meterId;
  }

  std::string meteringType;
  std::getline(ss, meteringType);
  DS_REQUIRE((meteringType == "power" || meteringType == "energy"), "Unsupported measuring type.", meteringType);
  result.meteringType = meteringType;

  return result;
}

DecodedApplicationId decodeId(const dsModel::ApplicationId& applicationId) {
  auto result = DecodedApplicationId{};
  std::stringstream ss(applicationId);

  std::string applicationType;
  std::getline(ss, applicationType, '-');
  if (applicationType != "temperature") {
    DS_FAIL_REQUIRE("Unsupported application type.", applicationType);
  }

  std::string zoneId;
  std::getline(ss, zoneId);
  zoneId = zoneId.substr(1);
  result.zoneId = strToInt(zoneId);

  return result;
}

static std::string heatingLevelName(const dsModel::TemperatureLevel& level) {
  switch (level) {
    case dsModel::TemperatureLevel::heatingComfort: return "Comfort";
    case dsModel::TemperatureLevel::heatingEconomy: return "Economy";
    case dsModel::TemperatureLevel::heatingSetback: return "NotUsed";
    case dsModel::TemperatureLevel::heatingNight: return "Night";
    case dsModel::TemperatureLevel::heatingVacation: return "Holiday";
    case dsModel::TemperatureLevel::coolingComfort: return "Cooling";
    case dsModel::TemperatureLevel::coolingEconomy: return "CoolingEconomy";
    case dsModel::TemperatureLevel::coolingSetback: return "CoolingNotUsed";
    case dsModel::TemperatureLevel::coolingNight: return "CoolingNight";
    case dsModel::TemperatureLevel::coolingVacation: return "CoolingHoliday";
    case dsModel::TemperatureLevel::none:
    case dsModel::TemperatureLevel::heatingOff:
    case dsModel::TemperatureLevel::coolingOff:
    case dsModel::TemperatureLevel::passiveCoolingThreshold: break;
  }
  DS_FAIL_REQUIRE("Unrecognized temperature level", level);
}

static int heatingOperationMode(const dsModel::TemperatureLevel& level) {
  switch (level) {
    case dsModel::TemperatureLevel::none: break;
    case dsModel::TemperatureLevel::heatingOff: return HeatingOperationModeIDOff;
    case dsModel::TemperatureLevel::heatingComfort: return HeatingOperationModeIDComfort;
    case dsModel::TemperatureLevel::heatingEconomy: return HeatingOperationModeIDEconomy;
    case dsModel::TemperatureLevel::heatingSetback: return HeatingOperationModeIDNotUsed;
    case dsModel::TemperatureLevel::heatingNight: return HeatingOperationModeIDNight;
    case dsModel::TemperatureLevel::heatingVacation: return HeatingOperationModeIDHoliday;
    case dsModel::TemperatureLevel::coolingOff: return HeatingOperationModeIDCoolOff;
    case dsModel::TemperatureLevel::coolingComfort: return HeatingOperationModeIDCoolComfort;
    case dsModel::TemperatureLevel::coolingEconomy: return HeatingOperationModeIDCoolEconomy;
    case dsModel::TemperatureLevel::coolingSetback: return HeatingOperationModeIDCoolNotUsed;
    case dsModel::TemperatureLevel::coolingNight: return HeatingOperationModeIDCoolNight;
    case dsModel::TemperatureLevel::coolingVacation: return HeatingOperationModeIDCoolHoliday;
    case dsModel::TemperatureLevel::passiveCoolingThreshold: break;
  }
  DS_FAIL_REQUIRE("Unrecognized temperature level", level);
}

namespace {

// Makes vector from a range. Until std::vector gains range constructor.
template <class Range>
auto makeVector(Range&& range) {
  auto result = std::vector<std::decay_t<decltype(*std::begin(range))>>{};
  result.reserve(std::size(range));
  std::move(range.begin(), range.end(), std::back_inserter(result));
  return result;
};

class DsModelImpl final : public DsModel, public RegisteredEventInterpreterPlugin, public PropertyListener {
  std::thread::id m_threadId{std::this_thread::get_id()};
  boost::asio::io_service& m_io;
  ds::Conf& m_conf;
  dsConnector::ClientBus& m_bus;
  Apartment& m_apartment;
  dsConnector::Subscriptions& m_subscriptions;
  ds::asio::ResourceContext& m_apartmentContext;
  ds::asio::Resource<> m_resource;

  DsModelHooks m_hooks;
  ds::Slot<bool> DS_UNIQUE_NAME(slot){m_io, m_bus.isConnectedChanged(), [this](bool x) { onBusIsConnectedChanged(x); }};
  ds::Slot<> DS_UNIQUE_NAME(slot){m_io, m_apartment.structureChanged(), [this]() { pollApartmentStructureImpl(); }};
  ds::Slot<> DS_UNIQUE_NAME(slot){m_io, m_apartment.statusChanged(), [this]() { pollApartmentStatusImpl(); }};
  StructureManipulator m_structureManipulator{m_apartment};
  PropertySystem& m_propertySystem = *m_apartment.getPropertySystem();
  PropertyNodePtr m_mshubActiveNode{m_propertySystem.getProperty(pp_websvc_mshub_active)};
  PropertyNodePtr m_apartmentNode{m_propertySystem.createProperty("/apartment")};
  PropertyNodePtr m_udasNode{m_propertySystem.createProperty("/scripts/system-addon-user-defined-actions/")};
  PropertyNodePtr m_udsNode{m_propertySystem.createProperty("/usr/addon-states/system-addon-user-defined-states/")};
  PropertyNodePtr m_configNode{m_propertySystem.getProperty("/config")};
  PropertyNodePtr m_heatingGeneralSettingsNode{
      m_propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings")};
  PropertyNodePtr m_heatingControllerConfigNode{
      m_propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig")};
  PropertyNodePtr m_heatingModesNode{m_propertySystem.createProperty("/scripts/system-addon-heating-controller/modes")};
  PropertyNodePtr m_heatingStatesNode{
      m_propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller")};
  PropertyNodePtr m_systemNode{m_propertySystem.createProperty("/system")};
  // event loop for with apartment level latency. io loop has higher latency requirements.
  ds::asio::StatePoller m_apartmentPoller{
      m_apartmentContext.io,
      "apartmentStructurePoller",
      m_conf.readOr<ds::asio::StatePollerConfig>("apartmentStructurePoller", {.moderate = 10s, .loop = 1h}),
      {.poll = [this] { return ds::toJsonFragment(apartment()); },
       .fire =
           [this](auto&& x) {
             onPollerFired(std::forward<decltype(x)>(x), "", dsModel::apartmentStructureChanged);
             pollApartmentStatusImpl();
           }}};
  ds::asio::StatePoller m_apartmentStatusPoller{
      m_apartmentContext.io,
      "apartmentStatusPoller",
      m_conf.readOr<ds::asio::StatePollerConfig>("apartmentStatusPoller", {.moderate = 1s, .loop = 1min}),
      {.poll = [this] { return ds::toJsonFragment(apartmentStatus()); },
       .fire =
           [this](auto&& x) {
             onPollerFired(std::forward<decltype(x)>(x), "/status", dsModel::apartmentStatusChanged);
           }}};
  ds::asio::Timer m_pushMeteringValuesTimer{m_io};
  std::unique_ptr<dsModel::ScenariosApi> m_scenarios{makeDsScenarios(m_resource, m_apartment)};

public:
  DsModelImpl(boost::asio::io_service& io, ds::Conf& conf, dsConnector::ClientBus& bus, Apartment& apartment,
              EventInterpreter& eventInterpreter, dsConnector::Subscriptions& subscriptions,
              ds::asio::ResourceContext& apartmentContext, DsModelHooks&& hooks)
      : RegisteredEventInterpreterPlugin("DSModel", eventInterpreter),
        m_io{io},
        m_conf{conf},
        m_bus{bus},
        m_apartment{apartment},
        m_subscriptions{subscriptions},
        m_apartmentContext{apartmentContext},
        m_resource{apartmentContext, "apartment"},
        m_hooks{std::move(hooks)} {
    DS_INFO_ENTER("");

    DS_REQUIRE(m_configNode, "Missing config property node");
    subscribe();

    if (m_mshubActiveNode) {
      m_mshubActiveNode->addListener(this);
    }
    m_udsNode->addListener(this);
    m_heatingGeneralSettingsNode->addListener(this);
    m_heatingControllerConfigNode->addListener(this);
    m_heatingModesNode->addListener(this);
    m_heatingStatesNode->addListener(this);
    updateBusActive();

    pushMeteringValuesAsyncLoop();
  }

  ~DsModelImpl() override { DS_INFO_ENTER(""); }

  [[nodiscard]] bool isApartmentIdMocked() const { return !m_conf.readOr<std::string>("apartmentId", {}).empty(); }

  void onPollerFired(ds::JsonFragment&& value, std::string&& path, const dsApi::EventType& eventType) {
    DS_INFO_ENTER(eventType);
    if (!m_bus.isConnected()) {
      DS_INFO("Not publishing apartment status. Not connected.");
      return;
    }
    const auto apartmentId = m_apartment.id();
    DS_REQUIRE(!apartmentId.empty());

    { // hint for signalR clients
      auto event = dsApi::Event{.type = eventType};
      DS_NOTICE("Publish event.", event);
      m_bus.event(std::move(event));
    }

    { // payload for backend apartment cache
      auto request =
          dsApi::HttpRequest{.path = ds::str("/backend/v1/apartment", path),
                             .httpArgs = {.method = dsApi::HttpMethod::PUT,
                                          .content = ds::toJsonFragment(dsApi::Document{.data = std::move(value)})}};
      DS_INFO("Publish", request); // too verbose for notice
      m_bus.asyncInvoke(std::move(request),
                        ds::makePromise<dsApi::HttpResult>([path](std::optional<dsApi::HttpResult>&& result) {
                          DS_INFO_ENTER(path, result);
                          if (!result || !result->isSuccess()) {
                            DS_ERROR("Failed to put to backend cache", path, result);
                          }
                        }));
    }

    // we rely on poller to recover from exceptions
  }

  void pollApartmentStructureImpl() {
    DS_DEBUG_ENTER("");
    m_hooks.apartmentStructurePolled();
    m_io.dispatch([this] { m_apartmentPoller.poll(); });
  }

  void pollApartmentStatusImpl() {
    DS_DEBUG_ENTER("");
    m_hooks.apartmentStatusPolled();
    m_io.dispatch([this] { m_apartmentStatusPoller.poll(); });
  }

  void onBusIsConnectedChanged(bool isBusConnected) {
    DS_INFO_ENTER(isBusConnected);

    m_apartmentPoller.setActive(isBusConnected);
    m_apartmentStatusPoller.setActive(isBusConnected);
  }

  void updateBusActive() {
    m_bus.setActive(isApartmentIdMocked() || (m_mshubActiveNode && m_mshubActiveNode->getBoolValue()));
  }

  void subscribe() override {
    getEventInterpreter().subscribe(boost::make_shared<EventSubscription>(*this, EventName::RetryRegistration));
    getEventInterpreter().subscribe(
        boost::make_shared<EventSubscription>(*this, EventName::ApartmentProxyStateChanged));
    getEventInterpreter().subscribe(boost::make_shared<EventSubscription>(*this, EventName::StateChange));
    getEventInterpreter().subscribe(boost::make_shared<EventSubscription>(*this, EventName::AddonStateChange));
  }

  void handleEvent(Event& event, const EventSubscription&) override {
    DS_INFO_ENTER(event);
    handleEvent(event);
  }

  void handleEvent(const Event& event) override {
    DS_DEBUG_ENTER(event);
    const auto& eventName = event.getName();
    if (eventName == EventName::RetryRegistration) {
      pollApartmentStructureImpl();
    }
    if (eventName == EventName::ApartmentProxyStateChanged) {
      pollApartmentStatusImpl();
    }
    if (eventName == EventName::StateChange) { // many states lead to changes in status.
      pollApartmentStatusImpl();
    }

    if (eventName == EventName::AddonStateChange) {
      const auto& scriptId = event.getPropertyByName("scriptID");
      if (scriptId == "system-addon-user-defined-states") {
        pollApartmentStatusImpl();
      }
    }
  }

  void propertyChanged(PropertyNodePtr node) {
    DS_DEBUG_ENTER(node->getName());
    if (node == m_mshubActiveNode) {
      updateBusActive();
    }
    if (node->isOnPathToRoot(*m_heatingControllerConfigNode) || node->isOnPathToRoot(*m_heatingModesNode) ||
        node->isOnPathToRoot(*m_udsNode)) {
      pollApartmentStructureImpl();
    }
    if (node->isOnPathToRoot(*m_heatingControllerConfigNode) || node->isOnPathToRoot(*m_heatingGeneralSettingsNode) ||
        node->isOnPathToRoot(*m_heatingStatesNode)) {
      pollApartmentStatusImpl();
    }
  }

  void propertyChanged(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(node)); }

  void propertyRemoved(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(parent)); }

  void propertyAdded(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(parent)); }

  [[nodiscard]] bool isTemperatureInDsModel(const dsModel::ZoneId& id) const {
    return m_apartment.zone(id)->temperatureConfig().m_mode == HeatingControlMode::PID;
  }

  dsModel::ScenariosApi& scenarios() override { return *m_scenarios; }
  [[nodiscard]] const dsModel::ScenariosApi& scenarios() const override { return *m_scenarios; }

  [[nodiscard]] dsModel::Zone zone(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    auto zonePtr = m_apartment.zone(zoneId);
    auto& zone = *zonePtr;
    try {
      return {.id = zoneId,
              .type = "zone",
              .lastChanged = {},
              .attributes = {
                  .name = zone.getName(),
                  .floor =
                      [&] {
                        if (zone.getFloorID() < 0) {
                          return std::string{};
                        }
                        return ds::str(zone.getFloorID());
                      }(),
                  .orderId = 0,
                  .submodules = zone.submoduleIds(),
                  .applications = zone.applications(),
                  .applicationTypes = zone.applications(),
                  .scenarios = {}, // TODO(someday): next story
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Zone failed.", e);
      return {.id = zoneId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Zone> zones() const override {
    return makeVector(m_apartment.zoneIds() | boost::adaptors::transformed([&](auto&& x) { return zone(x); }));
  }

  void setZoneName(dsModel::ZoneArgs&& args, const std::string& value) override {
    DS_NOTICE("Set zone name", args, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    auto zone = m_apartment.zone(args.zoneId);
    DS_REQUIRE(zone->getID() != 0, "Zone not found.", args);
    DS_REQUIRE(zone->getID() != NEW_DEVICES_ZONE_ID, "New devices zone cannot be renamed.", args);
    zone->setName(value);
  }

  [[nodiscard]] dsModel::Cluster cluster(const dsModel::ClusterId& clusterId) const override {
    DS_CONTEXT(clusterId);
    auto clusterPtr = m_apartment.cluster(clusterId);
    auto& cluster = *clusterPtr;
    try {
      return dsModel::Cluster{.id = clusterId,
                              .type = "cluster",
                              .attributes = {.name = cluster.getName(),
                                             .orderId = {},
                                             .applicationType = cluster.applicationType(),
                                             .submodules = {},
                                             .application = {}}};
    } catch (const std::exception& e) {
      DS_ERROR("Cluster failed.", e);
      return {.id = clusterId};
    }
  }

  [[nodiscard]] dsModel::ClusterStatus clusterStatus(const dsModel::ClusterId& clusterId) const override {
    DS_CONTEXT(clusterId);
    try {
      auto wind = m_apartment.tryGetNonScriptState("wind.group" + ds::str(clusterId));
      auto hail = m_apartment.tryGetNonScriptState("hail.group" + ds::str(clusterId));
      std::optional<bool> windState;
      std::optional<bool> hailState;
      if (wind) {
        windState = wind->getState() == State_Active;
      }
      if (hail) {
        hailState = hail->getState() == State_Active;
      }
      return {.id = clusterId,
              .type = "clusterStatus",
              .lastChanged = {},
              .attributes = {.weather = {.wind = windState, .hail = hailState}}};
    } catch (const std::exception& e) {
      DS_ERROR("Cluster status failed.", e);
      return {.id = clusterId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Cluster> clusters() const override {
    return makeVector(m_apartment.clusterIds() | boost::adaptors::transformed([&](auto&& x) { return cluster(x); }));
  }

  [[nodiscard]] std::vector<dsModel::ClusterStatus> clusterStatuses() const override {
    return makeVector(m_apartment.clusterIds() |
                      boost::adaptors::transformed([&](auto&& x) { return clusterStatus(x); }));
  }

  void setClusterName(const dsModel::ClusterArgs& args, std::string&& name) override {
    DS_NOTICE("Set cluster name.", args, name);
    DS_REQUIRE(!name.empty(), "New name cannot be empty.");
    m_apartment.cluster(args.clusterId)->setName(std::move(name));
  }

  [[nodiscard]] dsModel::DSDevice device(const dsModel::DeviceId& deviceId) const override {
    DS_CONTEXT(deviceId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{deviceId});
    auto& device = *devicePtr;
    try {
      return {.id = deviceId,
              .type = "dsDevice",
              .lastChanged = {},
              .attributes = {
                  .name = device.getName(),
                  .model = {},
                  .gtin = device.getGTIN(),
                  .dsid =
                      [&] {
                        dsid_t dsid;
                        std::string ret = "";
                        if (dsuid_to_dsid(device.getDSID(), &dsid)) {
                          ret = dsid2str(dsid);
                        }
                        return ret;
                      }(),
                  .displayId = device.getDisplayID(),
                  .present = device.isPresent(),
                  .firstSeen = {},
                  .lastDiscovered = {},
                  .inactiveSince = {},
                  .shadeAttributes = shadeAttributes(device),
                  .submodules = {device.submoduleId()},
                  .zone = ds::str(device.getZoneID()),
                  .scenarios = device.scenarioIds(),
                  .states = device.states(),
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Device failed.", e);
      return {.id = deviceId};
    }
  }

  [[nodiscard]] std::vector<dsModel::DSDevice> devices() const override {
    return makeVector(m_apartment.deviceIds() | boost::adaptors::transformed([&](auto&& x) { return device(x); }));
  }

  [[nodiscard]] dsModel::DSDeviceStatus deviceStatus(const dsModel::DeviceId& deviceId) const override {
    DS_CONTEXT(deviceId);
    try {
      auto result = m_apartment.proxy().dsDeviceStatus(deviceId);
      result.attributes.states = m_apartment.device(deviceId)->statesStatus();
      return result;
    } catch (const std::exception& e) {
      DS_ERROR("Device status failed.", e);
      return {.id = deviceId};
    }
  }

  [[nodiscard]] std::vector<dsModel::DSDeviceStatus> deviceStatuses() const override {
    return makeVector(m_apartment.deviceIds() |
                      boost::adaptors::transformed([&](auto&& x) { return deviceStatus(x); }));
  }

  void setOutputValue(dsModel::OutputStatusArgs&& args, double value) override {
    auto& device = *m_apartment.getDeviceByDSID(Dsuid{args.deviceId});
    auto& output = *device.getOutputChannelById(args.outputId);
    const auto size = getOutputChannelIdAndSize(output.id).second;
    const auto convertedValue = convertToOutputChannelValue(output.type, value);
    DS_NOTICE("Set output value", args, value, size, convertedValue);
    device.setDeviceOutputChannelValue(output.index, size, convertedValue);
  }

  [[nodiscard]] dsModel::Submodule submodule(const dsModel::SubmoduleId& submoduleId) const override {
    DS_CONTEXT(submoduleId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{submoduleId});
    auto& device = *devicePtr;
    auto hardwareType = device.hardwareType();
    try {
      return {.id = submoduleId,
              .type = "submodule",
              .lastChanged = {},
              .attributes = {
                  .name = device.getName(),
                  .technicalName = device.getHWInfo(),
                  .dsDevice = device.id(),
                  .functionBlocks = {device.functionBlockId()},
                  .zone = ds::str(device.getZoneID()),
                  .application = hardwareType == HardwareType::GR_KL210
                                     ? ds::fromEnum<std::string>(dsModel::ZoneApplicationType::awnings)
                                     : ds::fromEnum<std::string>(zoneApplication(device.getActiveGroup())),
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Submodule failed.", e);
      return {.id = submoduleId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Submodule> submodules() const override {
    return makeVector(m_apartment.submoduleIds() |
                      boost::adaptors::transformed([&](auto&& x) { return submodule(x); }));
  }

  [[nodiscard]] dsModel::FunctionBlock functionBlock(const dsModel::FunctionBlockId& functionBlockId) const override {
    DS_CONTEXT(functionBlockId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{functionBlockId});
    auto thresholds = m_apartment.proxy().getDeviceThresholds(Dsuid{functionBlockId});
    auto& device = *devicePtr;
    try {
      return {.id = functionBlockId,
              .type = "functionBlock",
              .lastChanged = {},
              .attributes = {
                  .name = device.getName(),
                  .technicalName = device.getHWInfo(),
                  .active = true,
                  .outputs = getOutputs(device, thresholds),
                  .buttonInputs = getButtons(device),
                  .sensorInputs = getSensors(device),
                  .submodule = ds::str(device.getDSID()),
                  .deviceAdapter = ds::str(device.getDSID()),
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Function block failed.", e);
      return {.id = functionBlockId};
    }
  }

  [[nodiscard]] std::vector<dsModel::FunctionBlock> functionBlocks() const override {
    return makeVector(m_apartment.functionBlockIds() |
                      boost::adaptors::transformed([&](auto&& x) { return functionBlock(x); }));
  }

  [[nodiscard]] std::optional<bool> stateStatus(const State& state) const {
    if (state.getState() == State_Active) {
      return true;
    } else if (state.getState() == State_Inactive) {
      return false;
    }
    return {};
  }

  [[nodiscard]] std::optional<bool> systemStateStatus(std::string name) const {
    auto state = m_apartment.tryGetNonScriptState(name);
    if (state) {
      if (name == dss::StateName::Presence) {
        if (state->getEnumState<dss::PresenceStates>() == dss::PresenceStates::present) {
          return false;
        } else if (state->getEnumState<dss::PresenceStates>() == dss::PresenceStates::absent) {
          return true;
        }
      } else if (name == dss::StateName::Hibernation) {
        if (state->getEnumState<dss::HibernationStates>() == dss::HibernationStates::sleeping) {
          return true;
        } else if (state->getEnumState<dss::HibernationStates>() == dss::HibernationStates::awake) {
          return false;
        }
      } else {
        return stateStatus(*state);
      }
    }
    // Some states are populated lazily. We cannot log error uncoditionally.
    DS_DEBUG("Could not get state status", name);
    return {};
  }

  [[nodiscard]] dsModel::ApartmentAccessStatus apartmentAccessStatus() const override {
    dsModel::ApartmentAccessStatus accessStatus{};

    accessStatus.absent = systemStateStatus(dss::StateName::Presence);

    return accessStatus;
  }

  [[nodiscard]] dsModel::ApartmentUserStatus apartmentUserStatus() const override {
    dsModel::ApartmentUserStatus userStatus{};

    userStatus.sleeping = systemStateStatus(dss::StateName::Hibernation);

    return userStatus;
  }

  [[nodiscard]] dsModel::ApartmentSecurity apartmentSecurity() const override {
    auto group00 = m_apartment.getZone(0)->getGroup(0);
    return {
        .alarm1 = {.name = group00->getSceneName(SceneAlarm)},
        .alarm2 = {.name = group00->getSceneName(SceneAlarm2)},
        .alarm3 = {.name = group00->getSceneName(SceneAlarm3)},
        .alarm4 = {.name = group00->getSceneName(SceneAlarm4)},
    };
  }

  void setApartmentAlarmName(int scene, std::string&& value) {
    m_apartment.getZone(0)->getGroup(0)->setSceneName(scene, std::move(value));
  }

  void setApartmentAlarm1Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm1 name", value);
    setApartmentAlarmName(SceneAlarm, std::move(value));
  }

  void setApartmentAlarm2Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm2 name", value);
    setApartmentAlarmName(SceneAlarm2, std::move(value));
  }

  void setApartmentAlarm3Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm3 name", value);
    setApartmentAlarmName(SceneAlarm3, std::move(value));
  }

  void setApartmentAlarm4Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm4 name", value);
    setApartmentAlarmName(SceneAlarm4, std::move(value));
  }

  [[nodiscard]] dsModel::ApartmentSecurityStatus apartmentSecurityStatus() const override {
    dsModel::ApartmentSecurityStatus securityStatus{};

    securityStatus.panic = systemStateStatus(dss::StateName::Panic);
    securityStatus.fire = systemStateStatus(dss::StateName::Fire);
    securityStatus.alarm1 = systemStateStatus(dss::StateName::Alarm);
    securityStatus.alarm2 = systemStateStatus(dss::StateName::Alarm2);
    securityStatus.alarm3 = systemStateStatus(dss::StateName::Alarm3);
    securityStatus.alarm4 = systemStateStatus(dss::StateName::Alarm4);

    return securityStatus;
  }

  [[nodiscard]] dsModel::ApartmentTemperatureStatus apartmentTemperatureStatus() const override {
    dsModel::ApartmentTemperatureStatus temperatureStatus{};

    auto mode = [&]() -> dsModel::TemperatureMode {
      const auto& heatingEnabled = m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
      if (!heatingEnabled) {
        return dsModel::TemperatureMode::off;
      }
      const auto& heatingForced =
          m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsHeating").value_or(false);
      const auto& coolingForced =
          m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsCooling").value_or(false);

      if(heatingForced) {
        return dsModel::TemperatureMode::heating;
      } else if (coolingForced) {
        return dsModel::TemperatureMode::cooling;
      } else {
        return dsModel::TemperatureMode::automatic;
      }
    };

    temperatureStatus.mode = mode();
    temperatureStatus.ecoPreferred =
        m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.0.Eco/value")) == 1 ? true : false;

    return temperatureStatus;
  }

  void setApartmentTemperatureStatusEcoPreferred(const dsModel::ApartmentArgs& args, bool value) override {
    DS_NOTICE("Set apartment temperature status eco preferred", args, value);
    m_apartment.pushEvent(
        boost::make_shared<Event>("heating-controller.operation-mode", Properties{
                                                                           {"actions", "setOperationMode"},
                                                                           {"zoneID", "0"},
                                                                           {"operationMode", value ? "2" : "1"},
                                                                       }));
  }

  [[nodiscard]] dsModel::ApartmentWeatherStatus apartmentWeatherStatus() const override {
    dsModel::ApartmentWeatherStatus weatherStatus{};

    weatherStatus.wind = systemStateStatus(dss::StateName::Wind);
    weatherStatus.rain = systemStateStatus(dss::StateName::Rain);
    weatherStatus.hail = systemStateStatus(dss::StateName::Hail);
    weatherStatus.frost = systemStateStatus(dss::StateName::Frost);

    return weatherStatus;
  }

  [[nodiscard]] dsModel::ApartmentEnvironmentStatus apartmentEnvironmentStatus() const override {
    dsModel::ApartmentEnvironmentStatus environmentStatus{};
    return environmentStatus; // Environment status is not used for now
  }

  [[nodiscard]] dsModel::ApartmentMeasurementStatus apartmentMeasurementStatus() const override {
    dsModel::ApartmentMeasurementStatus result{};

    ApartmentSensorStatus_t aSensorStatus = m_apartment.getSensorStatus();
    if (aSensorStatus.m_TemperatureTS != DateTime::NullDate) {
      result.temperature = aSensorStatus.m_TemperatureValue;
    }
    if (aSensorStatus.m_HumidityTS != DateTime::NullDate) {
      result.humidity = aSensorStatus.m_HumidityValue;
    }
    if (aSensorStatus.m_BrightnessTS != DateTime::NullDate) {
      result.brightness = aSensorStatus.m_BrightnessValue;
    }
    if (aSensorStatus.m_AirPressureTS != DateTime::NullDate) {
      result.airPressure = aSensorStatus.m_AirPressureValue;
    }
    if (aSensorStatus.m_PrecipitationTS != DateTime::NullDate) {
      result.precipitation = aSensorStatus.m_PrecipitationValue;
    }
    if (aSensorStatus.m_WindSpeedTS != DateTime::NullDate) {
      result.windSpeed = aSensorStatus.m_WindSpeedValue;
    }
    if (aSensorStatus.m_GustSpeedTS != DateTime::NullDate) {
      result.windGust = aSensorStatus.m_GustSpeedValue;
    }

    return result;
  }

  [[nodiscard]] dsModel::ZoneMeasurementsStatus zoneMeasurementsStatus(const dsModel::ZoneId& zoneId) const override {
    const auto& zone = *m_apartment.zone(zoneId);
    const auto sensorStatus = zone.getSensorStatus();
    return {.temperature = sensorStatus.temperatureValue,
            .humidity = sensorStatus.humidityValue,
            .brightness = sensorStatus.brightnessValue,
            .carbonDioxide = sensorStatus.co2ConcentrationValue,
            .motion = systemStateStatus(zoneStateName(StateName::Motion, zone.getID())),
            .presence = systemStateStatus(zoneStateName(StateName::Presence, zone.getID()))};
  }

  [[nodiscard]] dsModel::ZoneApplicationsStatus zoneApplicationsStatus(const dsModel::ZoneId& id) const override {
    const auto zone = m_apartment.zone(id);
    auto result = dsModel::ZoneApplicationsStatus{};
    for (const auto& application : zone->applications()) {
      switch (application) {
        case dsModel::ZoneApplicationType::none: break;
        case dsModel::ZoneApplicationType::lights: result.lights = m_apartment.proxy().lightsStatus(id); break;
        case dsModel::ZoneApplicationType::shades: result.shades = m_apartment.proxy().shadesStatus(id); break;
        case dsModel::ZoneApplicationType::audio: break;
        case dsModel::ZoneApplicationType::video: break;
        case dsModel::ZoneApplicationType::security: break;
        case dsModel::ZoneApplicationType::access: break;
        case dsModel::ZoneApplicationType::heating: break;
        case dsModel::ZoneApplicationType::cooling: break;
        case dsModel::ZoneApplicationType::temperature:
          if (isTemperatureInDsModel(id)) {
            result.temperature = temperatureStatus(id);
          }
          break;
        case dsModel::ZoneApplicationType::ventilation: break;
        case dsModel::ZoneApplicationType::recirculation: break;
        case dsModel::ZoneApplicationType::window: break;
        case dsModel::ZoneApplicationType::joker: break;
        case dsModel::ZoneApplicationType::awnings: break;
      }
    }
    return result;
  }

  [[nodiscard]] dsModel::ZoneStatus zoneStatus(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    std::vector<dsModel::ScenarioStatus> scenarios{};

    try {
      for (const auto& scenarioId : zoneScenarioIds(*m_apartment.zone(zoneId))) {
        scenarios.push_back(m_apartment.proxy().zoneScenarioStatus(scenarioId));
      }

      return {.id = zoneId,
              .type = "zoneStatus",
              .lastChanged = {},
              .attributes = {
                  .measurements = zoneMeasurementsStatus(zoneId),
                  .applications = zoneApplicationsStatus(zoneId),
                  .scenarios = std::move(scenarios),
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Zone status failed.", e);
      return {.id = zoneId};
    }
  }

  [[nodiscard]] std::vector<dsModel::ZoneStatus> zoneStatuses() const override {
    return makeVector(m_apartment.zoneIds() | boost::adaptors::transformed([&](auto&& x) { return zoneStatus(x); }));
  }

  [[nodiscard]] dsModel::TemperatureScene temperatureLastScene(const int& scene) const {
    switch (scene) {
      case HeatingOperationModeIDOff: return dsModel::TemperatureScene::heatingOff;
      case HeatingOperationModeIDComfort: return dsModel::TemperatureScene::heatingComfort;
      case HeatingOperationModeIDEconomy: return dsModel::TemperatureScene::heatingEconomy;
      case HeatingOperationModeIDNotUsed: return dsModel::TemperatureScene::heatingSetback;
      case HeatingOperationModeIDNight: return dsModel::TemperatureScene::heatingNight;
      case HeatingOperationModeIDHoliday: return dsModel::TemperatureScene::heatingVacation;
      case HeatingOperationModeIDPassiveCoolingOn: return dsModel::TemperatureScene::passiveCoolingActive;
      case HeatingOperationModeIDPassiveCoolingOff: return dsModel::TemperatureScene::passiveCoolingInactive;
      case HeatingOperationModeIDCoolOff: return dsModel::TemperatureScene::coolingOff;
      case HeatingOperationModeIDCoolComfort: return dsModel::TemperatureScene::coolingComfort;
      case HeatingOperationModeIDCoolEconomy: return dsModel::TemperatureScene::coolingEconomy;
      case HeatingOperationModeIDCoolNotUsed: return dsModel::TemperatureScene::coolingSetback;
      case HeatingOperationModeIDCoolNight: return dsModel::TemperatureScene::coolingNight;
      case HeatingOperationModeIDCoolHoliday: return dsModel::TemperatureScene::coolingVacation;
    }
    return dsModel::TemperatureScene::none;
  }

  [[nodiscard]] dsModel::TemperatureStatusStatus temperatureStatusStatus(const std::optional<double>& controlValue,
                                                                         const dsModel::ZoneId& id) const {
    const auto heatingEnabled = m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
    const auto heatingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/heatingPossible")).value_or(false);
    const auto coolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/coolingPossible")).value_or(false);
    const auto passiveCoolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/passiveCoolingPossible")).value_or(false);

    if (!heatingEnabled) {
      return dsModel::TemperatureStatusStatus::off;
    } else {
      if (controlValue) {
        if (controlValue > 0) {
          if (heatingPossible) {
            return dsModel::TemperatureStatusStatus::heating;
          }
        } else if (controlValue < 0) {
          if (coolingPossible || passiveCoolingPossible) {
            return dsModel::TemperatureStatusStatus::cooling;
          }
        } else if (heatingPossible || coolingPossible || passiveCoolingPossible) {
          return dsModel::TemperatureStatusStatus::neutral;
        }
      }
    }
    return dsModel::TemperatureStatusStatus::off;
  }

  [[nodiscard]] dsModel::TemperatureStatus temperatureStatus(const dsModel::ZoneId& id) const override {
    DS_CONTEXT(id);
    const auto& zone = *m_apartment.zone(id);
    DS_REQUIRE(isTemperatureInDsModel(id), "Temperature application does not exist.", id);
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.applications(), dsModel::ZoneApplicationType::temperature),
               "Temperature application does not exist.", id);
    try {
      const auto& config = zone.temperatureConfig();
      const auto& heatingStatus = zone.getHeatingStatus();
      const auto& controlValue = heatingStatus.controlValue;

      auto mode = [&]() -> dsModel::TemperatureMode {
        const auto& heatingEnabled =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
        if (!heatingEnabled) {
          return dsModel::TemperatureMode::off;
        }
        const auto& heatingForced =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsHeating").value_or(false);
        const auto& coolingForced =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsCooling").value_or(false);
        const auto heatingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/heatingPossible")).value_or(false);
        const auto coolingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/coolingPossible")).value_or(false);
        const auto passiveCoolingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/passiveCoolingPossible")).value_or(false);
        if (!heatingPossible && !coolingPossible && !passiveCoolingPossible) {
          return dsModel::TemperatureMode::off; 
        } else if (heatingForced) {
          if (heatingPossible) {
            return dsModel::TemperatureMode::heating; 
          } else {
            return dsModel::TemperatureMode::off; 
          }
        } else if (coolingForced) {
          if (coolingPossible || passiveCoolingPossible) {
            return dsModel::TemperatureMode::cooling; 
          } else {
            return dsModel::TemperatureMode::off; 
          }
        } else {
          return dsModel::TemperatureMode::automatic;
        }
      }();

      return dsModel::TemperatureStatus{
          .mode = mode,
          .status = temperatureStatusStatus(controlValue, id),
          .emergency = mode == dsModel::TemperatureMode::off
                           ? std::optional<bool>{}
                           : std::optional<bool>{config.m_HeatingControlState == HeatingControlStateIDEmergency},
          .setpoint = heatingStatus.nominalValue,
          .controlValue = controlValue,
          .lastCalledScene = temperatureLastScene(zone.getHeatingOperationMode()),
          .setbackRequested =
              m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.", zone.id(), ".NotUsed/value")) == 1
                  ? true
                  : false,
          .ecoPreferred =
              m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.", zone.id(), ".Eco/value")) == 1 ? true
                                                                                                               : false,
      };
    } catch (const std::exception& e) {
      DS_ERROR("Failed to get temperature status", e);
      return {};
    }
  }

  [[nodiscard]] dsModel::Application temperatureApplication(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    auto applicationId = dsModel::ApplicationId{ds::str("temperature-z", zoneId)};
    const auto& zone = *m_apartment.zone(zoneId);
    DS_REQUIRE(isTemperatureInDsModel(zoneId), "Temperature application does not exist.", zoneId);
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.applications(), dsModel::ZoneApplicationType::temperature),
               "Temperature application does not exist.", zoneId);
    try {
      const auto& config = zone.temperatureConfig();
      const auto& targetTemperatures = config.targetTemperatures;
      const auto heatingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/heatingPossible")).value_or(false);
      const auto coolingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/coolingPossible")).value_or(false);
      const auto passiveCoolingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/passiveCoolingPossible"))
              .value_or(false);

      std::vector<dsModel::TemperatureLevel> availableLevels{};
      std::vector<dsModel::TemperatureSetpoint> setpoints{};
      auto addAvailableLevel = [&](dsModel::TemperatureLevel level) {
        try {
          availableLevels.push_back(level);
          if (level == dsModel::TemperatureLevel::passiveCoolingThreshold) {
            setpoints.push_back(
                {.id = level,
                 .value = ds::tryToDouble(
                     m_heatingModesNode->tryChildValue<std::string>(ds::str(zoneId, "/minimalTemperatur"))
                         .value_or(std::string{}))});
          } else {
            setpoints.push_back({.id = level, .value = targetTemperatures[heatingOperationMode(level)]});
          }
        } catch (std::exception& e) {
          DS_ERROR("Failed to populate temperature level", level);
        }
      };

      if (heatingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::heatingComfort);
        addAvailableLevel(dsModel::TemperatureLevel::heatingEconomy);
        addAvailableLevel(dsModel::TemperatureLevel::heatingSetback);
        addAvailableLevel(dsModel::TemperatureLevel::heatingNight);
        addAvailableLevel(dsModel::TemperatureLevel::heatingVacation);
      }
      if (coolingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::coolingComfort);
        addAvailableLevel(dsModel::TemperatureLevel::coolingEconomy);
        addAvailableLevel(dsModel::TemperatureLevel::coolingSetback);
        addAvailableLevel(dsModel::TemperatureLevel::coolingNight);
        addAvailableLevel(dsModel::TemperatureLevel::coolingVacation);
      }
      if (passiveCoolingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::passiveCoolingThreshold);
      }

      return {.id = applicationId,
              .attributes = dsModel::TemperatureApplicationAttributes{
                  .applicationType = dsModel::ZoneApplicationType::temperature,
                  .zone = zoneId,
                  .availableLevels = std::move(availableLevels),
                  .setpoints = std::move(setpoints),
              }};
    } catch (const std::exception& e) {
      DS_ERROR("Temperature application failed.", e);
      return {.id = applicationId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Application> applications() const override {
    auto result = std::vector<dsModel::Application>{};
    for (const auto& zoneId : m_apartment.zoneIds()) {
      auto zone = m_apartment.zone(zoneId);

      for (const auto& application : zone->applications()) {
        switch (application) {
          case dsModel::ZoneApplicationType::none: break;
          case dsModel::ZoneApplicationType::lights: break;
          case dsModel::ZoneApplicationType::shades: break;
          case dsModel::ZoneApplicationType::audio: break;
          case dsModel::ZoneApplicationType::video: break;
          case dsModel::ZoneApplicationType::security: break;
          case dsModel::ZoneApplicationType::access: break;
          case dsModel::ZoneApplicationType::heating: break;
          case dsModel::ZoneApplicationType::cooling: break;
          case dsModel::ZoneApplicationType::temperature:
            if (isTemperatureInDsModel(zoneId)) {
              result.push_back(temperatureApplication(zone->id()));
            }
            break;
          case dsModel::ZoneApplicationType::ventilation: break;
          case dsModel::ZoneApplicationType::recirculation: break;
          case dsModel::ZoneApplicationType::window: break;
          case dsModel::ZoneApplicationType::joker: break;
          case dsModel::ZoneApplicationType::awnings: break;
        }
      }
    }
    return result;
  }

  [[nodiscard]] dsModel::Apartment apartment() const override {
    const auto apartmentId = m_apartment.id();
    try {
      return {.id = apartmentId,
              .type = "apartment",
              .attributes = {.name = m_apartment.getName(),
                             .deviceRoot = {}, // TODO(someday): not in dss model
                             .createdAt = {},  // TODO(someday): not in dss model
                             .security = apartmentSecurity(),
                             .installation = {}, // TODO(someday): not in dss model
                             .zones = m_apartment.zoneIds(),
                             .dsDevices = m_apartment.deviceIds(),
                             .clusters = m_apartment.clusterIds()},
              .included = {.installation = installation(),
                           .dsDevices = devices(),
                           .submodules = submodules(),
                           .functionBlocks = functionBlocks(),
                           .zones = zones(),
                           .scenarios = scenarios().scenarios(),
                           .userDefinedStates = userDefinedStates(),
                           .floors = floors().floors,
                           .clusters = clusters(),
                           .applications = applications(),
                           .dsServer = dsServer()}};
    } catch (const std::exception& e) {
      DS_ERROR("Apartment failed.", e);
      return {.id = apartmentId};
    }
  }

  void pollApartment() override {
    DS_NOTICE("Poll apartment");
    pollApartmentStructureImpl();
  }

  [[nodiscard]] dsModel::ApartmentStatus apartmentStatus() const override {
    const auto apartmentId = m_apartment.id();
    try {
      return {.id = apartmentId,
              .type = "apartmentStatus",
              .attributes = {.applications = {},
                             .access = apartmentAccessStatus(),
                             .user = apartmentUserStatus(),
                             .security = apartmentSecurityStatus(),
                             .temperature = apartmentTemperatureStatus(),
                             .weather = apartmentWeatherStatus(),
                             .environment = {},
                             .measurements = apartmentMeasurementStatus()},
              .included = {.dsDevices = deviceStatuses(),
                           .zones = zoneStatuses(),
                           .clusters = clusterStatuses(),
                           .userDefinedStates = userDefinedStateStatuses()}};
    } catch (const std::exception& e) {
      DS_ERROR("Apartment status failed.", e);
      return {.id = apartmentId};
    }
  }

  void pollApartmentStatus() override {
    DS_NOTICE("Poll apartment status");
    pollApartmentStatusImpl();
  }

  [[nodiscard]] std::vector<dsModel::UserDefinedState> userDefinedStates() const override {
    auto result = std::vector<dsModel::UserDefinedState>{};
    if (!m_udsNode) {
      return result;
    }
    for (auto&& udsNode : m_udsNode->getChildNodes()) {
      if (auto&& idNode = udsNode->getPropertyByName("name")) {
        if (auto&& nameNode = udsNode->getPropertyByName("displayName")) {
          if (auto&& setNameNode = udsNode->getPropertyByName("setName")) {
            if (auto&& resetNameNode = udsNode->getPropertyByName("resetName")) {
              if (auto&& phoneNode = udsNode->getPropertyByName("showOnPhone")) {
                if (auto&& typeNode = udsNode->getPropertyByName("type")) {
                  try {
                    result.push_back({.id = idNode->getValue<std::string>(),
                                      .type =
                                          [&] {
                                            auto name = typeNode->getValue<std::string>();
                                            if (name == "custom-states")
                                              return "manualState";
                                            if ((name == "zone-sensor-states") || (name == "device-sensor-states"))
                                              return "measurementState";
                                            if (name == "consumption-states")
                                              return "consumptionState";
                                            if (name == "power-states")
                                              return "circuitConsumptionState";
                                            if (name == "triggered-states")
                                              return "triggerState";
                                            if (name == "combined-states")
                                              return "combinedState";
                                            if (name == "window-states")
                                              return "windowState";
                                            DS_FAIL_REQUIRE("Unkown UDS type", name);
                                          }(),
                                      .lastChanged = {},
                                      .attributes = dsModel::UserDefinedStateAttributes{
                                          .name = nameNode->getValue<std::string>(),
                                          .setName = setNameNode->getValue<std::string>(),
                                          .resetName = resetNameNode->getValue<std::string>(),
                                          .visibleForUsers = phoneNode->getValue<bool>()}});
                  } catch (const std::exception& e) {
                    DS_ERROR("User defined state failed.", e);
                  }
                }
              }
            }
          }
        }
      }
    }
    boost::sort(result, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  [[nodiscard]] std::vector<dsModel::UserDefinedStateStatus> userDefinedStateStatuses() const override {
    auto result = std::vector<dsModel::UserDefinedStateStatus>{};
    if (!m_udsNode) {
      return result;
    }
    for (auto&& udsNode : m_udsNode->getChildNodes()) {
      if (auto&& idNode = udsNode->getPropertyByName("name")) {
        if (auto&& stateNode = udsNode->getPropertyByName("state")) {
          const auto userDefinedStateId = idNode->getValue<std::string>();
          DS_CONTEXT(userDefinedStateId);
          try {
            result.push_back({.id = userDefinedStateId,
                              .type = "userDefinedStateStatus",
                              .lastChanged = {},
                              .attributes = {.status = ds::toEnum<dsModel::UserDefinedStateStatusStatus>(
                                                 stateNode->getValue<std::string>())}});
          } catch (const std::exception& e) {
            DS_ERROR("User defined state status failed.", e);
            result.push_back({.id = userDefinedStateId});
          }
        }
      }
    }
    boost::sort(result, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  void setUserDefinedStateStatus(dsModel::UserDefinedStateStatusArgs&& args, const std::string& value) override {
    if (!m_udsNode) {
      return;
    }
    auto&& stateNode = m_udsNode->getPropertyByName(args.userDefinedStateId);
    DS_REQUIRE(stateNode, "State is unknown", args.userDefinedStateId);

    auto&& typeNode = stateNode->getPropertyByName("type");
    DS_REQUIRE(typeNode, "Type node unavailable");
    DS_REQUIRE(typeNode->getValue<std::string>() == "custom-states", "Only a manualState can be set");

    auto state = m_apartment.tryGetState(StateType_Script, args.userDefinedStateId);
    DS_REQUIRE(state, "State is unknown", args.userDefinedStateId);

    auto&& realValue = state->tryValueFromName(value);
    DS_REQUIRE(realValue, "Requested state value is illegal", value);
    state->setState(coJSON, *realValue);
    DS_NOTICE("Set UserDefinedState value", args, value);
  }

  [[nodiscard]] dsModel::Metering metering(const dsModel::MeteringArgs& args) const override {
    auto& meteringId = args.meteringId;
    DS_CONTEXT(meteringId);
    try {
      std::string type{};
      std::string unit{};
      std::string technicalName{};
      dsModel::Origin origin{};

      auto meteringDecoded = decodeId(args.meteringId);
      if (meteringDecoded.origin.type == "dsm") {
        auto meter = m_apartment.getDSMeterByDSID(str2dsuid(meteringDecoded.origin.id));
        DS_REQUIRE(meter->isPresent(), "Metering device not present.", meteringDecoded.origin.id);
        origin = {.type = "dsMeter", .id = ds::str(meter->getDSID())};
        if (meteringDecoded.meteringType == "power") {
          type = "powerMetering";
          unit = "W";
          technicalName = ds::str(meter->getHardwareName(), " power");
        } else if (meteringDecoded.meteringType == "energy") {
          type = "energyMetering";
          unit = "Wh";
          technicalName = ds::str(meter->getHardwareName(), " energy");
        }
      } else if (meteringDecoded.origin.type == "apartment") {
        origin = {.type = "apartment"};
        if (meteringDecoded.meteringType == "power") {
          type = "powerMetering";
          unit = "W";
          technicalName = "apartment power";
        } else if (meteringDecoded.meteringType == "energy") {
          type = "energyMetering";
          unit = "Wh";
          technicalName = "apartment energy";
        }
      }

      return {.id = args.meteringId,
              .type = type,
              .lastChanged = {},
              .attributes = {.unit = unit, .technicalName = technicalName, .origin = origin}};
    } catch (const std::exception& e) {
      DS_ERROR("Metering failed.", e);
      return {.id = meteringId};
    }
  }

  [[nodiscard]] dsModel::Meterings meterings() const override {
    DS_NOTICE("Get meterings.");
    auto result = dsModel::Meterings{};

    result.meterings.push_back(metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}));
    result.meterings.push_back(metering(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}));

    for (const auto& meter : m_apartment.getDSMeters()) {
      if (!meter->isPresent() || !meter->getCapability_HasMetering()) {
        continue;
      }
      result.meterings.push_back(metering(
          dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{ds::str("dsm-", meter->getDSID(), "-power")}}));
      result.meterings.push_back(metering(
          dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{ds::str("dsm-", meter->getDSID(), "-energy")}}));
    }
    boost::sort(result.meterings, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  [[nodiscard]] dsModel::MeteringValue meteringValue(const dsModel::MeteringArgs& args) const override {
    const auto& meteringId = args.meteringId;
    DS_CONTEXT(meteringId);
    try {
      double value{};
      ds::DateTime lastChanged{};

      auto meteringDecoded = decodeId(meteringId);
      if (meteringDecoded.origin.type == "dsm") {
        auto meter = m_apartment.getDSMeterByDSID(str2dsuid(meteringDecoded.origin.id));
        DS_REQUIRE(meter->isPresent(), "Metering device not present.", meteringDecoded.origin.id);

        if (meteringDecoded.meteringType == "power") {
          value = meter->getCachedPowerConsumption();
          lastChanged = ds::DateTime{meter->getCachedPowerConsumptionTimeStamp()};
        } else if (meteringDecoded.meteringType == "energy") {
          value = meter->getCachedEnergyMeterValue();
          lastChanged = ds::DateTime{meter->getCachedEnergyMeterTimeStamp()};
        }
      } else if (meteringDecoded.origin.type == "apartment") {
        if (meteringDecoded.meteringType == "power") {
          for (const auto& meter : m_apartment.getDSMeters()) {
            if (!meter->isPresent() || !meter->getCapability_HasMetering()) {
              continue;
            }
            value += meter->getCachedPowerConsumption();
            lastChanged = ds::max(lastChanged, ds::DateTime{meter->getCachedPowerConsumptionTimeStamp()});
          }
        } else if (meteringDecoded.meteringType == "energy") {
          for (const auto& meter : m_apartment.getDSMeters()) {
            if (!meter->isPresent() || !meter->getCapability_HasMetering()) {
              continue;
            }
            value += meter->getCachedEnergyMeterValue();
            lastChanged = ds::max(lastChanged, ds::DateTime{meter->getCachedEnergyMeterTimeStamp()});
          }
        }
      }

      return {.id = meteringId, .type = "meteringValue", .lastChanged = lastChanged, .attributes = {.value = value}};
    } catch (const std::exception& e) {
      DS_ERROR("Metering value failed.", e);
      return {.id = meteringId};
    }
  }

  [[nodiscard]] dsModel::MeteringValues meteringValues() const override {
    DS_INFO_ENTER("Get metering values.");
    return meteringValuesImpl();
  }

  [[nodiscard]] dsModel::MeteringValues meteringValuesImpl() const {
    DS_DEBUG_ENTER("Get metering values.");
    auto result = dsModel::MeteringValues{};

    result.values.push_back(meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-power"}}));
    result.values.push_back(
        meteringValue(dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"apartment-energy"}}));

    for (const auto& meter : m_apartment.getDSMeters()) {
      if (!meter->isPresent() || !meter->getCapability_HasMetering()) {
        continue;
      }
      result.values.push_back(meteringValue(
          dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{ds::str("dsm-", meter->getDSID(), "-power")}}));
      result.values.push_back(meteringValue(
          dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{ds::str("dsm-", meter->getDSID(), "-energy")}}));
    }
    boost::sort(result.values, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  void pushMeteringValuesAsyncLoop() {
    // TODO(someday): avoid polling when there is no subscription
    m_pushMeteringValuesTimer.randomlyExpiresFromNow(1s, [this] {
      DS_CONTEXT("pushMeteringValues");
      DS_DEFER(pushMeteringValuesAsyncLoop());
      try {
        m_subscriptions.event(
            {.type = dsModel::meteringValues, .path = {}, .data = ds::toJsonFragment(meteringValuesImpl())});
      } catch (std::exception& e) {
        DS_ERROR("Failed to push event", e);
      }
    });
  }

  [[nodiscard]] dsModel::ParameterCache parameterCacheStatus() const override {
    auto cacheStatus = m_apartment.proxy().getParameterCacheStatus();
    auto result =
        dsModel::ParameterCache{.id = m_apartment.id(), .type = "parameterCache", .attributes = cacheStatus.cache};

    auto fbVec = std::vector<dsModel::FunctionBlockParameterCache>{};
    for (auto& fb : cacheStatus.devices) {
      fbVec.push_back({.id = fb.id,
                       .type = "parameterCacheFunctionBlock",
                       .lastChanged = {},
                       .attributes = {.completion = fb.completion}});
    }
    result.included.functionBlocks = std::move(fbVec);
    return result;
  }

  void updateParameterCache(const dsModel::UpdateDeviceParameterCacheArgs& args) override {
    if (args.mode == dsModel::ParameterCacheUpdateMode::fast) {
      m_apartment.pushEvent(boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast));
    } else {
      m_apartment.pushEvent(boost::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent));
    }
  }

  [[nodiscard]] dsModel::Floor floor(const Floor& floor) const {
    const auto floorId = floor.id();
    try {
      return dsModel::Floor{
          .id = floorId,
          .type = "floor",
          .lastChanged = {},
          .attributes = {.name = floor.getName(), .orderId = floor.getOrder(), .zones = floor.zones()}};
    } catch (const std::exception& e) {
      DS_ERROR("Floor failed.", e);
      return {.id = floorId};
    }
  }

  [[nodiscard]] dsModel::Floor floor(const dsModel::FloorArgs& args) const override {
    DS_INFO_ENTER(args);
    return floor(*m_apartment.floor(args.floorId));
  }

  [[nodiscard]] dsModel::Floors floors() const override {
    DS_INFO_ENTER("");
    dsModel::Floors result{};
    for (const auto& floor : m_apartment.getFloors()) {
      result.floors.push_back(this->floor(*floor));
    }
    return result;
  }

  void setFloorName(const dsModel::FloorArgs& args, const std::string& value) override {
    DS_NOTICE("Set floor name.", args, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    auto floor = m_apartment.tryGetFloor(strToInt(args.floorId));
    if (auto floorPtr = floor.lock()) {
      floorPtr->setName(value);
      if (m_apartment.getModelMaintenance()) {
        m_apartment.getModelMaintenance()->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
      }
    } else {
      DS_FAIL_REQUIRE("Floor not found");
    }
  }

  void setTemperatureSetpoint(const dsModel::TemperatureSetpointArgs& args, double value) override {
    DS_NOTICE("Set temperature setpoint.", args, value);

    int zoneId = decodeId(args.applicationId).zoneId;

    if (args.temperatureLevel == dsModel::TemperatureLevel::passiveCoolingThreshold) {
      m_apartment.pushEvent(boost::make_shared<Event>(
          "heating-controller.passive-cooling",
          Properties{{"actions", "setMinimalTemperature"}, {"zoneID", ds::str(zoneId)}, {"value", ds::str(value)}}));
    } else {
      m_apartment.pushEvent(boost::make_shared<Event>(
          "heating-controller.operation-mode", Properties{{"actions", "setTemperatureValues"},
                                                          {"zoneID", ds::str(zoneId)},
                                                          {heatingLevelName(args.temperatureLevel), ds::str(value)}}));
    }
  }

  void setTemperatureStatusSetpoint(const dsModel::TemperatureStatusArgs& args, double value) override {
    DS_NOTICE("Set temperature status setpoint.", args, value);
    m_apartment.zone(args.zoneId)
        ->getGroup(GroupIDControlTemperature)
        ->pushSensor(coJSON, SAC_MANUAL, {}, SensorType::RoomTemperatureSetpoint, value, {});
  }

  void setTemperatureStatusSetbackRequested(const dsModel::TemperatureStatusArgs& args, bool value) override {
    DS_NOTICE("Set temperature status setback requested.", args, value);
    m_apartment.pushEvent(boost::make_shared<Event>(
        "heating-controller.operation-mode", Properties{{"actions", value ? "setOperationMode" : "resetOperationMode"},
                                                        {"zoneID", args.zoneId},
                                                        {"operationMode", "3"}}));
  }

  void setTemperatureStatusEcoPreferred(const dsModel::TemperatureStatusArgs& args, bool value) override {
   DS_NOTICE("Set temperature status eco preferred.", args, value);
    m_apartment.pushEvent(boost::make_shared<Event>(
        "heating-controller.operation-mode", Properties{{"actions", "setOperationMode"},
                                                        {"zoneID", args.zoneId},
                                                        {"operationMode", value ? "2" : "1"}}));
  }

  [[nodiscard]] dsModel::Installation installation() const override {
    DS_INFO_ENTER("");
    const auto installationId = dsModel::InstallationId{m_apartment.id()};
    try {
      return dsModel::Installation{
          .id = installationId, // we use apartment id for installation id for now
          .type = "installation",
          .lastChanged = {},
          .attributes = {
              .location = {.longitude = [&]() -> std::optional<double> {
                             if (auto x = m_configNode->tryChildValue<std::string>("geodata/longitude")) {
                               return stod(*x);
                             }
                             return {};
                           }(),
                           .latitude = [&]() -> std::optional<double> {
                             if (auto x = m_configNode->tryChildValue<std::string>("geodata/latitude")) {
                               return stod(*x);
                             }
                             return {};
                           }()},
              .countryCode = m_configNode->tryChildValue<std::string>("geodata/country").value_or(std::string{}),
              .city = m_configNode->tryChildValue<std::string>("geodata/location").value_or(std::string{})}};
    } catch (const std::exception& e) {
      DS_ERROR("Installation failed.", e);
      return {.id = installationId};
    }
  }

  [[nodiscard]] dsModel::DSServer dsServer() const override {
    DS_INFO_ENTER("");
    const auto serverId = dsModel::ServerId{m_systemNode->tryChildValue<std::string>("dSUID").value_or(std::string{})};
    try {
      return dsModel::DSServer{
          .id = serverId,
          .type = "dsServer",
          .lastChanged = {},
          .attributes = {
              .name = m_apartment.getName(),
              .model = m_systemNode->tryChildValue<std::string>("version/model").value_or(std::string{}),
              .gtin = m_systemNode->tryChildValue<std::string>("version/gtin").value_or(std::string{}),
              .dsid = m_systemNode->tryChildValue<std::string>("dSID").value_or(std::string{}),
              .version = m_systemNode->tryChildValue<std::string>("version/distroVersion").value_or(std::string{}),
              .environment = [&]() -> std::optional<dsModel::Environment> {
                switch (m_apartmentContext.system.release()) {
                  case ds::Release::devel: return dsModel::Environment::development;
                  case ds::Release::testing: return dsModel::Environment::alpha;
                  case ds::Release::fieldtest: return dsModel::Environment::beta;
                  case ds::Release::production: return dsModel::Environment::production;
                }
                return {};
              }()}};
    } catch (const std::exception& e) {
      DS_ERROR("Server failed.", e);
      return {.id = serverId};
    }
  }
};
} // namespace

std::unique_ptr<DsModel> makeDsModel(boost::asio::io_service& io, ds::Conf& conf, dsConnector::ClientBus& bus,
                                     Apartment& apartment, EventInterpreter& eventInterpreter,
                                     dsConnector::Subscriptions& subscriptions,
                                     ds::asio::ResourceContext& apartmentContext, DsModelHooks&& hooks) {
  return std::make_unique<DsModelImpl>(io, conf, bus, apartment, eventInterpreter, subscriptions, apartmentContext,
                                       std::move(hooks));
}

} // namespace dss
