/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include <ds-api/service.h>
#include <ds/catch/catch.h>
#include <ds/log.h>
#include "apartment-api-mock.h"
#include "scenarios-mock.h"

using namespace trompeloeil;

namespace {

static const char *TAGS = "[dsModelApartment][dsModel]";

SCENARIO("dsModelApartmentApiService", TAGS) {
  dsModel::ScenariosApiMock scenariosApi;
  dsModel::ApartmentApiMock apartmentApi;

  REQUIRE_CALL(apartmentApi, scenarios()).LR_RETURN(scenariosApi);
  auto service = dsModel::makeService(apartmentApi);

  SECTION("scenarios smoke test") {
    // full coverage is in scenarios-test
    const auto args = dsModel::ScenarioArgs{.scenarioId = dsModel::ScenarioId{"x"}};
    REQUIRE_CALL(scenariosApi, deleteScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args));
    CHECK(service->invoke(dsApi::HttpRequest{.path = ds::str("/scenarios/", args.scenarioId),
                                             .httpArgs = {.method = dsApi::HttpMethod::DELETE}}) ==
          dsApi::noContentResult());
  }

  auto checkGet = [&](const auto &path, const auto &data) {
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = path, .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<std::decay_t<decltype(data)>>(result) == data);
  };

  SECTION("zone") {
    const auto data = dsModel::Zone{.id = dsModel::ZoneId{"id0"}};
    REQUIRE_CALL(apartmentApi, zone(data.id)).LR_RETURN(data);
    checkGet("/zones/id0", data);
  }

  SECTION("zones") {
    const auto data = std::vector<dsModel::Zone>{{.id = dsModel::ZoneId{"id0"}}};
    REQUIRE_CALL(apartmentApi, zones()).LR_RETURN(data);
    checkGet("/zones", data);
  }

  SECTION("zoneStatus") {
    const auto data = dsModel::ZoneStatus{.id = dsModel::ZoneId{"id0"}};
    REQUIRE_CALL(apartmentApi, zoneStatus(data.id)).LR_RETURN(data);
    checkGet("/zones/id0/status", data);
  }

  SECTION("zoneStatuses") {
    const auto data = std::vector<dsModel::ZoneStatus>{{.id = dsModel::ZoneId{"id0"}}};
    REQUIRE_CALL(apartmentApi, zoneStatuses()).LR_RETURN(data);
    checkGet("/zones/status", data);
  }

  SECTION("setZoneName") {
    const auto args = dsModel::ZoneArgs{.zoneId = dsModel::ZoneId{"x"}};
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setZoneName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/zones/", args.zoneId),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setOutputValue") {
    const auto args = dsModel::OutputStatusArgs{.deviceId = dsModel::DeviceId{"dev"}, .outputId = "output"};
    const auto value = 3.0;
    REQUIRE_CALL(apartmentApi, setOutputValue(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/functionBlocks/", args.deviceId, "/outputs/", args.outputId, "/value"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/dsDevices/", args.deviceId, "/status"),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("apartment") {
    const auto apartment = [&] { return dsModel::Apartment{.id = "a"}; };
    REQUIRE_CALL(apartmentApi, apartment()).LR_RETURN(apartment());
    const auto result = service->invoke(dsApi::HttpRequest{.path = "", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Apartment>(result) == apartment());
  }

  SECTION("setApartmentAlarm1Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm1Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm1/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm2Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm2Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm2/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm3Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm3Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm3/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("setApartmentAlarm4Name") {
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setApartmentAlarm4Name(value));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/security/alarm4/name"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("pollApartment") {
    REQUIRE_CALL(apartmentApi, pollApartment());
    CHECK(service->invoke(dsApi::HttpRequest{.path = "/poll", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }

  SECTION("apartmentStatus") {
    const auto apartmentStatus = [&] { return dsModel::ApartmentStatus{.id = "a"}; };
    REQUIRE_CALL(apartmentApi, apartmentStatus()).LR_RETURN(apartmentStatus());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/status", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::ApartmentStatus>(result) == apartmentStatus());
  }

  SECTION("setApartmentTemperatureStatusEcoPreferred") {
    const auto args = dsModel::ApartmentArgs{};
    const auto value = true;
    REQUIRE_CALL(apartmentApi, setApartmentTemperatureStatusEcoPreferred(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = ds::str("/temperature/ecoPreferred"),
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/status",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("pollApartmentStatus") {
    REQUIRE_CALL(apartmentApi, pollApartmentStatus());
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/status/poll", .httpArgs = {.method = dsApi::HttpMethod::POST}}) == dsApi::noContentResult());
  }

  SECTION("setUserDefinedState") {
    const auto args = dsModel::UserDefinedStateStatusArgs{.userDefinedStateId = "dev"};
    const auto value = ds::str("active");
    REQUIRE_CALL(apartmentApi, setUserDefinedStateStatus(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/status"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/userDefinedStates/", args.userDefinedStateId, "/status"),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("floor") {
    const auto args = dsModel::FloorArgs{.floorId = dsModel::FloorId{"x"}};
    const auto floor = [&] { return dsModel::Floor{.id = args.floorId}; };
    REQUIRE_CALL(apartmentApi, floor(args)).LR_RETURN(floor());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = ds::str("/floors/", args.floorId), .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Floor>(result) == floor());
  }

  SECTION("floors") {
    const auto floors = [&] { return dsModel::Floors{}; };
    REQUIRE_CALL(apartmentApi, floors()).LR_RETURN(floors());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/floors", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Floors>(result) == floors());
  }

  SECTION("setFloorName") {
    const auto args = dsModel::FloorArgs{.floorId = dsModel::FloorId{"x"}};
    const auto value = "{newName}";
    REQUIRE_CALL(apartmentApi, setFloorName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = ds::str("/floors/", args.floorId),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("cluster") {
    const auto data = dsModel::Cluster{.id = dsModel::ClusterId{"id0"}};
    REQUIRE_CALL(apartmentApi, cluster(data.id)).LR_RETURN(data);
    checkGet("/clusters/id0", data);
  }

  SECTION("clusters") {
    const auto data = std::vector<dsModel::Cluster>{{.id = dsModel::ClusterId{"id0"}}};
    REQUIRE_CALL(apartmentApi, clusters()).LR_RETURN(data);
    checkGet("/clusters", data);
  }

  SECTION("clusterStatus") {
    const auto data = dsModel::ClusterStatus{.id = dsModel::ClusterId{"id0"}};
    REQUIRE_CALL(apartmentApi, clusterStatus(data.id)).LR_RETURN(data);
    checkGet("/clusters/id0/status", data);
  }

  SECTION("clusterStatuses") {
    const auto data = std::vector<dsModel::ClusterStatus>{{.id = dsModel::ClusterId{"id0"}}};
    REQUIRE_CALL(apartmentApi, clusterStatuses()).LR_RETURN(data);
    checkGet("/clusters/status", data);
  }

  SECTION("set cluster name") {
    const auto args = dsModel::ClusterArgs{.clusterId = dsModel::ClusterId{"{id}"}};
    const auto name = "{newName}";
    REQUIRE_CALL(apartmentApi, setClusterName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == name));
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(name)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/clusters/{id}",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("dsDevice") {
    const auto data = dsModel::DSDevice{.id = dsModel::DeviceId{"id0"}};
    REQUIRE_CALL(apartmentApi, device(data.id)).LR_RETURN(data);
    checkGet("/dsDevices/id0", data);
  }

  SECTION("dsDevices") {
    const auto data = std::vector<dsModel::DSDevice>{{.id = dsModel::DeviceId{"id0"}}};
    REQUIRE_CALL(apartmentApi, devices()).LR_RETURN(data);
    checkGet("/dsDevices", data);
  }

  SECTION("dsDeviceStatus") {
    const auto data = dsModel::DSDeviceStatus{.id = dsModel::DeviceId{"id0"}};
    REQUIRE_CALL(apartmentApi, deviceStatus(data.id)).LR_RETURN(data);
    checkGet("/dsDevices/id0/status", data);
  }

  SECTION("dsDeviceStatuses") {
    const auto data = std::vector<dsModel::DSDeviceStatus>{{.id = dsModel::DeviceId{"id0"}}};
    REQUIRE_CALL(apartmentApi, deviceStatuses()).LR_RETURN(data);
    checkGet("/dsDevices/status", data);
  }

  SECTION("submodule") {
    const auto data = dsModel::Submodule{.id = dsModel::SubmoduleId{"id0"}};
    REQUIRE_CALL(apartmentApi, submodule(data.id)).LR_RETURN(data);
    checkGet("/submodules/id0", data);
  }

  SECTION("submodules") {
    const auto data = std::vector<dsModel::Submodule>{{.id = dsModel::SubmoduleId{"id0"}}};
    REQUIRE_CALL(apartmentApi, submodules()).LR_RETURN(data);
    checkGet("/submodules", data);
  }

  SECTION("functionBlock") {
    const auto data = dsModel::FunctionBlock{.id = dsModel::FunctionBlockId{"id0"}};
    REQUIRE_CALL(apartmentApi, functionBlock(data.id)).LR_RETURN(data);
    checkGet("/functionBlocks/id0", data);
  }

  SECTION("functionBlocks") {
    const auto data = std::vector<dsModel::FunctionBlock>{{.id = dsModel::FunctionBlockId{"id0"}}};
    REQUIRE_CALL(apartmentApi, functionBlocks()).LR_RETURN(data);
    checkGet("/functionBlocks", data);
  }

  SECTION("setTemperatureSetpoint") {
    const auto args = dsModel::TemperatureSetpointArgs{.applicationId = dsModel::ApplicationId{"temperature"},
                                                       .temperatureLevel = dsModel::TemperatureLevel::heatingComfort};
    const auto value = 15.0;
    REQUIRE_CALL(apartmentApi, setTemperatureSetpoint(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = "/setpoints/heatingComfort/value",
                               .value = ds::toJsonFragment(value)});
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/applications/temperature",
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("installation") {
    // The code for /apartment/installation and /installation is shared.
    // Full coverage is in dsModelInstallationApiService
    SECTION("metering") {
      const auto metering = [&] { return dsModel::Metering{.id = dsModel::MeteringId{"a"}}; };
      const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
      REQUIRE_CALL(apartmentApi, metering(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(metering());
      const auto result = service->invoke(dsApi::HttpRequest{.path = "/installation/meterings/dsm-x-power",
                                                             .httpArgs = {.method = dsApi::HttpMethod::GET}});
      INFO(result);
      CHECK(dsApi::data<dsModel::Metering>(result) == metering());
    }

    SECTION("installation") {
      const auto installation = [&] { return dsModel::Installation{.id = dsModel::InstallationId{"a"}}; };
      REQUIRE_CALL(apartmentApi, installation()).LR_RETURN(installation());
      const auto result =
          service->invoke(dsApi::HttpRequest{.path = "/installation", .httpArgs = {.method = dsApi::HttpMethod::GET}});
      INFO(result);
      CHECK(dsApi::data<dsModel::Installation>(result) == installation());
    }
  }

  SECTION("dsServer") {
    const auto server = [&] { return dsModel::DSServer{.id = dsModel::ServerId{"a"}}; };
    REQUIRE_CALL(apartmentApi, dsServer()).LR_RETURN(server());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/dsServer", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::DSServer>(result) == server());
  }  
}

SCENARIO("dsModelInstallationApiService", TAGS) {
  dsModel::ApartmentApiMock apartmentApi;
  auto service = dsModel::makeInstallationService(apartmentApi);

  SECTION("metering") {
    const auto metering = [&] { return dsModel::Metering{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(apartmentApi, metering(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(metering());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Metering>(result) == metering());
  }

  SECTION("meterings") {
    const dsModel::Metering meteringA = {.id = dsModel::MeteringId{"a"}};
    const dsModel::Metering meteringB = {.id = dsModel::MeteringId{"b"}};
    const auto meterings = [&] { return dsModel::Meterings{.meterings = {meteringA, meteringB}}; };
    REQUIRE_CALL(apartmentApi, meterings()).LR_RETURN(meterings());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/meterings", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Meterings>(result) == meterings());
  }

  SECTION("meteringValue") {
    const auto meteringValue = [&] { return dsModel::MeteringValue{.id = dsModel::MeteringId{"a"}}; };
    const auto args = dsModel::MeteringArgs{.meteringId = dsModel::MeteringId{"dsm-x-power"}};
    REQUIRE_CALL(apartmentApi, meteringValue(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(meteringValue());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values/dsm-x-power", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValue>(result) == meteringValue());
  }

  SECTION("meteringValues") {
    const auto meteringValues = [&] { return dsModel::MeteringValues{.values = {{.id = dsModel::MeteringId{"a"}}}}; };
    REQUIRE_CALL(apartmentApi, meteringValues()).LR_RETURN(meteringValues());
    const auto result = service->invoke(
        dsApi::HttpRequest{.path = "/meterings/values", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::MeteringValues>(result) == meteringValues());
  }

  SECTION("updateParameterCache") {
    const auto args = dsModel::UpdateDeviceParameterCacheArgs{.mode = dsModel::ParameterCacheUpdateMode::slow};
    REQUIRE_CALL(apartmentApi, updateParameterCache(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service->invoke(dsApi::HttpRequest{
              .path = "/parameterCache/update",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::noContentResult());
  }

  SECTION("parameterCacheStatus") {
    const auto parameterCacheStatus = [&] { return dsModel::ParameterCache{.id = "a"}; };
    REQUIRE_CALL(apartmentApi, parameterCacheStatus()).LR_RETURN(parameterCacheStatus());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/parameterCache", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::ParameterCache>(result) == parameterCacheStatus());
  }

  SECTION("installation") {
    const auto installation = [&] { return dsModel::Installation{.id = dsModel::InstallationId{"a"}}; };
    REQUIRE_CALL(apartmentApi, installation()).LR_RETURN(installation());
    const auto result =
        service->invoke(dsApi::HttpRequest{.path = "/", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    INFO(result);
    CHECK(dsApi::data<dsModel::Installation>(result) == installation());
  }
}

} // namespace
