#pragma once

#include <ds-api/service-fwd.h>
#include <ds/reflect.h>
#include <boost/noncopyable.hpp>
#include "model.h"
#include "scenarios-fwd.h"

DS_REFLECTED_ENUM((dsModel), ParameterCacheUpdateMode, (none)(slow)(fast));

namespace dsModel {

struct ApartmentArgs {
  std::string dummy{}; // TODO(someday): remove once reflection supports empty structs
};

struct ZoneArgs {
  ZoneId zoneId{};
};

struct MeteringArgs {
  dsModel::MeteringId meteringId{};
};

struct DeviceArgs {
  DeviceId deviceId{};
};

struct SubmoduleArgs {
  SubmoduleId submoduleId{};
};

struct FunctionBlockArgs {
  FunctionBlockId functionBlockId{};
};

struct OutputStatusArgs {
  DeviceId deviceId{};
  std::string outputId{};
};

struct UserDefinedStateStatusArgs {
  std::string userDefinedStateId{};
};

struct UpdateDeviceParameterCacheArgs {
  ParameterCacheUpdateMode mode{};
};

struct FloorArgs {
  FloorId floorId{};
};

struct ClusterArgs {
  ClusterId clusterId{};
};

struct TemperatureSetpointArgs {
  ApplicationId applicationId{};
  TemperatureLevel temperatureLevel{};
};

struct TemperatureStatusArgs {
  ZoneId zoneId{};
};

class ApartmentApi : public boost::noncopyable {
public:
  virtual ~ApartmentApi() = default;

  virtual ScenariosApi &scenarios() = 0;
  [[nodiscard]] virtual const ScenariosApi &scenarios() const = 0;

  [[nodiscard]] virtual Zone zone(const ZoneId &id) const = 0;
  [[nodiscard]] virtual std::vector<Zone> zones() const = 0;
  virtual void setZoneName(ZoneArgs &&, const std::string &value) = 0;
  [[nodiscard]] virtual ZoneStatus zoneStatus(const ZoneId &id) const = 0;
  [[nodiscard]] virtual std::vector<ZoneStatus> zoneStatuses() const = 0;

  [[nodiscard]] virtual Cluster cluster(const ClusterId &id) const = 0;
  [[nodiscard]] virtual std::vector<Cluster> clusters() const = 0;
  [[nodiscard]] virtual ClusterStatus clusterStatus(const ClusterId &id) const = 0;
  [[nodiscard]] virtual std::vector<ClusterStatus> clusterStatuses() const = 0;
  virtual void setClusterName(const ClusterArgs &args, std::string &&name) = 0;

  [[nodiscard]] virtual DSDevice device(const DeviceId &) const = 0;
  [[nodiscard]] virtual std::vector<DSDevice> devices() const = 0;
  [[nodiscard]] virtual DSDeviceStatus deviceStatus(const DeviceId &) const = 0;
  [[nodiscard]] virtual std::vector<DSDeviceStatus> deviceStatuses() const = 0;
  virtual void setOutputValue(OutputStatusArgs &&, double value) = 0;

  [[nodiscard]] virtual Submodule submodule(const SubmoduleId &) const = 0;
  [[nodiscard]] virtual std::vector<Submodule> submodules() const = 0;

  [[nodiscard]] virtual FunctionBlock functionBlock(const FunctionBlockId &) const = 0;
  [[nodiscard]] virtual std::vector<FunctionBlock> functionBlocks() const = 0;

  [[nodiscard]] virtual ApartmentAccessStatus apartmentAccessStatus() const = 0;
  [[nodiscard]] virtual ApartmentUserStatus apartmentUserStatus() const = 0;
  [[nodiscard]] virtual ApartmentSecurity apartmentSecurity() const = 0;
  virtual void setApartmentAlarm1Name(std::string &&) = 0;
  virtual void setApartmentAlarm2Name(std::string &&) = 0;
  virtual void setApartmentAlarm3Name(std::string &&) = 0;
  virtual void setApartmentAlarm4Name(std::string &&) = 0;
  [[nodiscard]] virtual ApartmentSecurityStatus apartmentSecurityStatus() const = 0;
  [[nodiscard]] virtual ApartmentTemperatureStatus apartmentTemperatureStatus() const = 0;
  virtual void setApartmentTemperatureStatusEcoPreferred(const ApartmentArgs &, bool) = 0;
  [[nodiscard]] virtual ApartmentWeatherStatus apartmentWeatherStatus() const = 0;
  [[nodiscard]] virtual ApartmentEnvironmentStatus apartmentEnvironmentStatus() const = 0;
  [[nodiscard]] virtual ApartmentMeasurementStatus apartmentMeasurementStatus() const = 0;

  [[nodiscard]] virtual ZoneMeasurementsStatus zoneMeasurementsStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual TemperatureStatus temperatureStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual ZoneApplicationsStatus zoneApplicationsStatus(const ZoneId &) const = 0;
  [[nodiscard]] virtual Application temperatureApplication(const ZoneId &) const = 0;
  virtual void setTemperatureSetpoint(const TemperatureSetpointArgs&, double) = 0;
  virtual void setTemperatureStatusSetpoint(const TemperatureStatusArgs&, double) = 0;
  virtual void setTemperatureStatusSetbackRequested(const TemperatureStatusArgs&, bool) = 0;
  virtual void setTemperatureStatusEcoPreferred(const TemperatureStatusArgs&, bool) = 0;

  [[nodiscard]] virtual std::vector<Application> applications() const = 0;

  [[nodiscard]] virtual Apartment apartment() const = 0;
  virtual void pollApartment() = 0;

  [[nodiscard]] virtual ApartmentStatus apartmentStatus() const = 0;
  virtual void pollApartmentStatus() = 0;

  [[nodiscard]] virtual std::vector<UserDefinedState> userDefinedStates() const = 0;
  [[nodiscard]] virtual std::vector<UserDefinedStateStatus> userDefinedStateStatuses() const = 0;
  virtual void setUserDefinedStateStatus(UserDefinedStateStatusArgs &&, const std::string &value) = 0;

  [[nodiscard]] virtual Meterings meterings() const = 0;
  [[nodiscard]] virtual Metering metering(const MeteringArgs &) const = 0;

  [[nodiscard]] virtual MeteringValues meteringValues() const = 0;
  [[nodiscard]] virtual MeteringValue meteringValue(const MeteringArgs &) const = 0;

  virtual void updateParameterCache(const UpdateDeviceParameterCacheArgs &) = 0;
  [[nodiscard]] virtual ParameterCache parameterCacheStatus() const = 0;

  [[nodiscard]] virtual Floor floor(const FloorArgs &args) const = 0;
  [[nodiscard]] virtual Floors floors() const = 0;
  virtual void setFloorName(const FloorArgs &args, const std::string &value) = 0;

  [[nodiscard]] virtual Installation installation() const = 0;
  [[nodiscard]] virtual DSServer dsServer() const = 0;
};

std::unique_ptr<dsApi::Service> makeInstallationService(ApartmentApi &);
std::unique_ptr<dsApi::Service> makeService(ApartmentApi &);

} // namespace dsModel

DS_JSON_DECLARE_STRUCT((dsModel), ApartmentArgs, (dummy));
DS_JSON_DECLARE_STRUCT((dsModel), ZoneArgs, (zoneId));
DS_JSON_DECLARE_STRUCT((dsModel), MeteringArgs, (meteringId));
DS_JSON_DECLARE_STRUCT((dsModel), DeviceArgs, (deviceId));
DS_JSON_DECLARE_STRUCT((dsModel), SubmoduleArgs, (submoduleId));
DS_JSON_DECLARE_STRUCT((dsModel), FunctionBlockArgs, (functionBlockId));
DS_JSON_DECLARE_STRUCT((dsModel), OutputStatusArgs, (deviceId)(outputId));
DS_JSON_DECLARE_STRUCT((dsModel), UserDefinedStateStatusArgs, (userDefinedStateId));
DS_JSON_DECLARE_STRUCT((dsModel), UpdateDeviceParameterCacheArgs, (mode));
DS_JSON_DECLARE_STRUCT((dsModel), FloorArgs, (floorId));
DS_JSON_DECLARE_STRUCT((dsModel), ClusterArgs, (clusterId));
DS_JSON_DECLARE_STRUCT((dsModel), TemperatureSetpointArgs, (applicationId)(temperatureLevel));
DS_JSON_DECLARE_STRUCT((dsModel), TemperatureStatusArgs, (zoneId));
