/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include <ds-api/service.h>
#include <ds/catch/catch.h>
#include <ds/log.h>
#include "scenarios-mock.h"

using namespace trompeloeil;

namespace {

static const char *TAGS = "[dsModelScenarios][dsModel]";

SCENARIO("Scenario json serialization", TAGS) {
  // variants are somewhat tricky to serialize
  CHECK(ds::toJson(dsModel::Scenario{}) == R"({})");
  CHECK(ds::toJson(dsModel::Scenario{.id = dsModel::ScenarioId{"id0"},
                                     .attributes = dsModel::ApplicationZoneScenarioAttributes{.name = "name0"}}) ==
        R"({"id":"id0","type":"applicationZoneScenario","attributes":{"name":"name0"}})");
  CHECK(ds::toJson(dsModel::Scenario{.id = dsModel::ScenarioId{"id0"},
                                     .attributes = dsModel::UserDefinedActionAttributes{.name = "name0"}}) ==
        R"({"id":"id0","type":"userDefinedAction","attributes":{"name":"name0"}})");
  CHECK(
      ds::toJson(dsModel::Scenario{
          .id = dsModel::ScenarioId{"id0"},
          .attributes = dsModel::DeviceScenarioAttributes{.name = {.domain = {}, .message = "name0"},
                                                          .dsDevice = dsModel::DeviceId{"dev0"},
                                                          .actionId = dsModel::DeviceScenarioActionId{"action0"}}}) ==
      R"({"id":"id0","type":"deviceScenario","attributes":{"name":{"message":"name0"},"dsDevice":"dev0","actionId":"action0"}})");
}

SCENARIO("dsModelScenariosApartmentApiRoutes", TAGS) {
  dsModel::ScenariosApiMock mock;
  dsApi::Service service{dsModel::apartmentApiRoutes(mock)};

  SECTION("invokeStandardScenario") {
    const auto args = dsModel::InvokeStandardScenarioArgs{.context = "x"};
    REQUIRE_CALL(mock, invokeStandardScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service.invoke(dsApi::HttpRequest{
              .path = "/scenarios/invoke",
              .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(args)}}) ==
          dsApi::noContentResult());
  }

  SECTION("createScenario") {
    const auto data =
        dsModel::Scenario{.id = {}, .attributes = dsModel::ApplicationZoneScenarioAttributes{.name = "name0"}};
    REQUIRE_CALL(mock, createScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == data);).LR_RETURN(dsModel::ScenarioId{"id0"});
    const auto result = service.invoke(dsApi::HttpRequest{
        .path = "/scenarios",
        .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(dsApi::document(data))}});
    CHECK(result.status == 201);
    CHECK(result.headers == dsApi::HttpHeaders{{"Location", {"id0"}}});
  }

  SECTION("deleteScenario") {
    const auto args = dsModel::ScenarioArgs{.scenarioId = dsModel::ScenarioId{"x"}};
    REQUIRE_CALL(mock, deleteScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args));
    CHECK(service.invoke(dsApi::HttpRequest{.path = ds::str("/scenarios/", args.scenarioId),
                                            .httpArgs = {.method = dsApi::HttpMethod::DELETE}}) ==
          dsApi::noContentResult());
  }

  SECTION("setScenarioName") {
    const auto args = dsModel::ScenarioArgs{.scenarioId = dsModel::ScenarioId{"x"}};
    const auto value = "{newName}";
    REQUIRE_CALL(mock, setScenarioName(_, _)).LR_SIDE_EFFECT(CHECK(_1 == args); CHECK(_2 == value););
    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = ds::str("/name"), .value = ds::toJsonFragment(value)});
    CHECK(service.invoke(dsApi::HttpRequest{
              .path = ds::str("/scenarios/", args.scenarioId),
              .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}}) ==
          dsApi::noContentResult());
  }

  SECTION("invokeScenario") {
    const auto args = dsModel::ScenarioArgs{.scenarioId = dsModel::ScenarioId{"x"}};
    REQUIRE_CALL(mock, invokeScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service.invoke(
              dsApi::HttpRequest{.path = "/scenarios/x/invoke", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }

  SECTION("persistScenario") {
    const auto args = dsModel::ScenarioArgs{.scenarioId = dsModel::ScenarioId{"x"}};
    REQUIRE_CALL(mock, persistScenario(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    CHECK(service.invoke(
              dsApi::HttpRequest{.path = "/scenarios/x/persist", .httpArgs = {.method = dsApi::HttpMethod::POST}}) ==
          dsApi::noContentResult());
  }
}

} // namespace
