#pragma once

#include "model-ids.h"
#include "scenarios-fwd.h"
#include "scenarios-ids.h"

#include <ds-api/rest.h>
#include <ds-api/service-fwd.h>
#include <ds-api/text.h>
#include <ds/json.h>
#include <boost/serialization/strong_typedef.hpp>
#include "model-fwd.h"

DS_REFLECTED_ENUM((dsModel), ZoneActionId, (none)(deepOff)(standby));
DS_REFLECTED_ENUM((dsModel), LightsActionId, (none)(off)(on)(brighter)(darker)(forcedOff));
DS_REFLECTED_ENUM((dsModel), ShadesActionId, (none)(down)(up)(stepDown)(stepUp)(stop));
DS_REFLECTED_ENUM((dsModel), AwningsActionId, (none)(out)(in)(stepOut)(stepIn)(stop)(sunProtection));
DS_REFLECTED_ENUM((dsModel), AudioActionId, (none)(play)(pause)(volumeUp)(volumeDown));
DS_REFLECTED_ENUM((dsModel), DeviceActionId, (none)(off)(on)(stop)(inc)(dec));
DS_REFLECTED_ENUM((dsModel), CommonActionId,
                  (none)(preset0)(preset1)(preset2)(preset3)(preset4)(preset10)(preset11)(preset12)(preset13)(preset14)(
                      preset20)(preset21)(preset22)(preset23)(preset24)(preset30)(preset31)(preset32)(preset33)(
                      preset34)(preset40)(preset41)(preset42)(preset43)(preset44));
DS_REFLECTED_ENUM((dsModel), ScenarioStatusStatus, (none)(active)(inactive)(undefined));

namespace dsModel {

struct UserDefinedActionAttributes {
  std::string name{};
};

struct ApplicationZoneScenarioAttributes {
  std::string name;
  CommonActionId actionId{};
  std::string context{}; // deprecated in favour of `data.type`
  std::vector<std::string> submodules{};
  std::vector<std::string> dsDevices{};
  std::string zone{};
  ZoneApplicationType application{};
  std::string area{};
};

struct DeviceScenarioAttributes {
  dsApi::Text name{};
  DeviceId dsDevice{};
  DeviceScenarioActionId actionId{};
};

using Scenario =
    dsApi::VariantResourceData<dsModel::ScenarioId, dsModel::UserDefinedActionAttributes,
                               dsModel::ApplicationZoneScenarioAttributes, dsModel::DeviceScenarioAttributes>;

struct ScenarioStatus {
  ScenarioId id;
  ScenarioStatusStatus status{};
  std::optional<double> completed{}; // percents
  ds::DateTime scheduledAt{};
  ds::DateTime terminatesAt{};
};

//
// api
//
struct InvokeStandardScenarioArgs {
  std::string context{};
  std::string actionId{};
  std::string application{};
  std::string dsDevice{};
  std::string zone{};
  std::string area{};
};

struct ScenarioArgs {
  ScenarioId scenarioId{};
};

class ScenariosApi : public boost::noncopyable {
public:
  virtual ~ScenariosApi() = default;

  [[nodiscard]] virtual std::vector<Scenario> applicationZoneScenarios() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> userDefinedActions() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> deviceScenarios() const = 0;
  [[nodiscard]] virtual std::vector<Scenario> scenarios() const = 0;
  virtual void invokeStandardScenario(const InvokeStandardScenarioArgs &) = 0;
  virtual ScenarioId createScenario(const Scenario &) = 0;
  virtual void deleteScenario(const ScenarioArgs &) = 0;
  virtual void setScenarioName(const ScenarioArgs &, const std::string &value) = 0;
  virtual void invokeScenario(const ScenarioArgs &) = 0;
  virtual void persistScenario(const ScenarioArgs &) = 0;
};

std::vector<dsApi::Route> apartmentApiRoutes(ScenariosApi &api);

} // namespace dsModel

DS_JSON_DECLARE_STRUCT((dsModel), UserDefinedActionAttributes, (name));
DS_JSON_DECLARE_STRUCT((dsModel), ApplicationZoneScenarioAttributes,
                       (name)(actionId)(context)(submodules)(dsDevices)(zone)(application)(area));
DS_JSON_DECLARE_STRUCT((dsModel), DeviceScenarioAttributes, (name)(dsDevice)(actionId));
static constexpr auto dsModelScenarioNameVisitor = ::ds::Visitor{
    [](const std::monostate &) { return ""; },
    [](const dsModel::UserDefinedActionAttributes &) { return "userDefinedAction"; },
    [](const dsModel::ApplicationZoneScenarioAttributes &) { return "applicationZoneScenario"; },
    [](const dsModel::DeviceScenarioAttributes &) { return "deviceScenario"; },
};
DS_JSON_DECLARE_API_VARIANT_RESOURCE_DATA((dsModel), Scenario, dsModelScenarioNameVisitor);
DS_JSON_DECLARE_STRUCT((dsModel), ScenarioStatus, (id)(status)(completed)(scheduledAt)(terminatesAt));

DS_JSON_DECLARE_STRUCT((dsModel), InvokeStandardScenarioArgs, (context)(actionId)(application)(dsDevice)(zone)(area));
DS_JSON_DECLARE_STRUCT((dsModel), ScenarioArgs, (scenarioId));
