/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-scenarios.h"
#include <ds-model/scenarios.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include "businterface-mock.h"
#include "event-mock.h"
#include "model/apartment-mock.h"
#include "model/group.h"
#include "model/zone.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssDsScenariosTest");
static const char* TAGS = "[dss][dssDsScenarios]";

using namespace dss;
using namespace trompeloeil;

namespace {

TEST_CASE("Scenario id", TAGS) {
  auto checkDecodeEncode = [&](const dsModel::ScenarioId& id, const dss::DecodedScenarioId& decodedId) {
    INFO(id);
    INFO(decodedId);
    CHECK(decodeId(id) == decodedId);
    CHECK(encodeId(decodedId) == id);
  };
  SECTION("applicationZone") {
    checkDecodeEncode(dsModel::ScenarioId{"applicationZone-z1-g3-s5"},
                      ApplicationZoneScenarioId{.zone = 1, .group = 3, .scene = 5});

    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z0-g3-s5"}), "Cannot use scenario ID for zone 0");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z2-g0-s5"}), "Cannot use scenario ID for group 0");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone--z1-g2-s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1--g2-s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1-g2-/s3"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-1-32-23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-c1-g32-s23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-s1-z32-g23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1-g32-d23"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1-g2-s3-"}),
                      "strToInt: Could not parse value: '3-'");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1g2-s3"}),
                      "strToInt: Could not parse value: '1g2'");
    SECTION("Non-common scenario") {
      CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"applicationZone-z1-g1-s1"}),
                        "Wrong scenario ID - not a common scene");
    }
  }

  SECTION("userDefined") {
    checkDecodeEncode(dsModel::ScenarioId{"userDefined-32"}, UserDefinedActionId{"32"});
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"userDefined"}), "Wrong scenario ID format");
  }

  SECTION("device") {
    checkDecodeEncode(
        dsModel::ScenarioId{"device-123-action"},
        DeviceScenarioId{.device = dsModel::DeviceId{"123"}, .action = dsModel::DeviceScenarioActionId{"action"}});
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"dev-123-action"}), "Wrong scenario ID type");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"device-123"}), "Wrong scenario ID format");
    CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"device-action"}), "Wrong scenario ID format");
  }

  CHECK_THROWS_FIND(decodeId(dsModel::ScenarioId{"foo-3"}), "Wrong scenario ID type");
}

SCENARIO("dssDsScenarios", TAGS) {
  auto clock = ds::ClockMocker{};
  auto propertySystem = PropertySystem{};
  EventSinkMock eventSink;

  ApartmentMock apartment{};
  const auto apartmentId = "apartmentId-0"s;
  ALLOW_CALL(apartment, id()).RETURN(apartmentId);
  apartment.setName("TestName");
  apartment.setPropertySystem(&propertySystem);
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::asio::TestLoopResource loop;
  auto dsScenarios = makeDsScenarios(loop.resource, apartment);

  GIVEN("applicationZoneScenario") {
    auto& zone1 = *apartment.allocateZone(1);
    auto& group1 = *zone1.getGroup(1);
    const auto sceneId = 19;
    const auto id = ApplicationZoneScenarioId{.zone = zone1.getID(), .group = group1.getID(), .scene = sceneId};
    const auto encodedId = encodeId(id);
    const auto args = dsModel::ScenarioArgs{.scenarioId = encodedId};

    const auto nonExistingArgs =
        dsModel::ScenarioArgs{.scenarioId = encodeId(ApplicationZoneScenarioId{
                                  .zone = zone1.getID(), .group = group1.getID(), .scene = 18})}; // preset3

    group1.setSceneName(sceneId, "name1");

    SECTION("structure") {
      CHECK(dsScenarios->applicationZoneScenarios() ==
            std::vector<dsModel::Scenario>{{.id = encodedId,
                                            .attributes = dsModel::ApplicationZoneScenarioAttributes{
                                                .name = "name1",
                                                .actionId = dsModel::CommonActionId::preset4,
                                                .context = "applicationZone",
                                                .submodules = {},
                                                .dsDevices = {},
                                                .zone = ds::str(zone1.getID()),
                                                .application = dsModel::ZoneApplicationType::lights,
                                            }}});
    }

    SECTION("invoke") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.action, callScene(&group1, coJSON, SAC_MANUAL, sceneId, "", false));
        dsScenarios->invokeScenario(args);
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs), "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("rename") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.structureModify, sceneSetName(zone1.getID(), group1.getID(), sceneId, "name2"));
        dsScenarios->setScenarioName(args, "name2");
      }
      SECTION("zone application scenario") {
        CHECK(
            zoneApplicationScenarioIds(zone1) ==
            std::vector<ApplicationZoneScenarioId>{{.zone = zone1.getID(), .group = group1.getID(), .scene = sceneId}});
      }
      SECTION("zone scenario id") {
        CHECK(zoneScenarioIds(zone1) == std::vector<dsModel::ScenarioId>{{dsModel::ScenarioId{"applicationZone-z1-g1-s19"}}});
      }
      SECTION("fail on empty name") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(args, ""), "New name cannot be empty.");
      }
      SECTION("non existing") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(nonExistingArgs, "name2"), "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("persist") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.action, saveScene(&group1, coJSON, sceneId, ""));
        dsScenarios->persistScenario(args);
      }
      SECTION("non existing") {
        CHECK_THROWS_FIND(dsScenarios->persistScenario(nonExistingArgs), "Invalid ApplicationZoneScenarioId");
      }
    }

    SECTION("delete") {
      SECTION("success") {
        REQUIRE_CALL(busInterface.structureModify, sceneSetName(zone1.getID(), group1.getID(), sceneId, ""));
        dsScenarios->deleteScenario(args);
        CHECK(dsScenarios->applicationZoneScenarios() == std::vector<dsModel::Scenario>{});
      }
      SECTION("success on non existing") { dsScenarios->deleteScenario(nonExistingArgs); }
    }

    SECTION("create") {
      dsModel::Scenario args2 = (dsModel::Scenario{.id = {},
                                                   .attributes = dsModel::ApplicationZoneScenarioAttributes{
                                                       .name = "newName",
                                                       .actionId = dsModel::CommonActionId::preset2,
                                                       .context = {},
                                                       .submodules = {},
                                                       .dsDevices = {},
                                                       .zone = "1",
                                                       .application = dsModel::ZoneApplicationType::lights,
                                                   }});
      REQUIRE_CALL(busInterface.structureModify,
                   sceneSetName(zone1.getID(), group1.getID(), 17 /*preset2*/, "newName"));
      dsScenarios->createScenario(args2);

      SECTION("fail on empty name") {
        dsModel::Scenario args3 = (dsModel::Scenario{.id = {},
                                                     .attributes = dsModel::ApplicationZoneScenarioAttributes{
                                                         .name = "",
                                                         .actionId = dsModel::CommonActionId::preset2,
                                                         .context = {},
                                                         .submodules = {},
                                                         .dsDevices = {},
                                                         .zone = "1",
                                                         .application = dsModel::ZoneApplicationType::lights,
                                                     }});
        CHECK_THROWS_FIND(dsScenarios->createScenario(args3), "New name cannot be empty.");
      }
    }
  }

  GIVEN("userDefinedAction") {
    auto udasNode = propertySystem.createProperty("/scripts/system-addon-user-defined-actions");

    const auto id = "id0"s;
    const auto name = "name0"s;
    auto udaNode = udasNode->createProperty(id);
    udaNode->setChildValue("id", id);
    udaNode->setChildValue("name", name);

    const auto args = dsModel::ScenarioArgs{.scenarioId = encodeId(UserDefinedActionId{id})};
    const auto nonExistingArgs = dsModel::ScenarioArgs{.scenarioId = encodeId(UserDefinedActionId{"idx"})};

    SECTION("structure") {
      CHECK(dsScenarios->userDefinedActions() ==
            std::vector<dsModel::Scenario>{{.id = encodeId(UserDefinedActionId{id}),
                                            .attributes = dsModel::UserDefinedActionAttributes{.name = name}}});
    }

    SECTION("invoke") {
      SECTION("success") {
        REQUIRE_CALL(eventSink, pushEvent(_))
            .LR_SIDE_EFFECT(CHECK(*_1 == Event{EventName::HighLevelEvent, Properties{{"id", id}}}));
        dsScenarios->invokeScenario(args);
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs), "Invalid UserDefinedActionId");
      }
    }

    SECTION("rename") {
      SECTION("success") {
        SECTION("success TODO") { CHECK_THROWS_FIND(dsScenarios->setScenarioName(args, "name2"), "TODO"); }
      }
      SECTION("fail on empty name") {
        CHECK_THROWS_FIND(dsScenarios->setScenarioName(args, ""), "New name cannot be empty.");
      }
      SECTION("non-existing") {
        CHECK_THROWS_FIND(dsScenarios->invokeScenario(nonExistingArgs), "Invalid UserDefinedActionId");
      }
    }

    SECTION("persist") {
      SECTION("always fails") {
        CHECK_THROWS_FIND(dsScenarios->persistScenario(args), "Persist of user defined action is not possible.");
      }
    }

    SECTION("delete") {
      SECTION("success TODO") { CHECK_THROWS_FIND(dsScenarios->deleteScenario(args), "TODO"); }
      SECTION("success on non existing") { dsScenarios->deleteScenario(nonExistingArgs); }
    }
  }

  GIVEN("deviceScenario") {
    auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
    const auto actionId = "actionId"s;

    std::vector<dsModel::Scenario> scenarios{};
    scenarios.push_back({.id = dsModel::ScenarioId{"device-1111111111111111111111111111111100-actionId"},
                         .attributes = dsModel::DeviceScenarioAttributes{
                             .name = {.domain = {}, .message = "name"},
                             .dsDevice = dsModel::DeviceId{"1111111111111111111111111111111100"},
                             .actionId = dsModel::DeviceScenarioActionId{actionId}}});
    device.setScenarios(std::move(scenarios));

    SECTION("structure") {
      CHECK(dsScenarios->deviceScenarios() ==
            std::vector<dsModel::Scenario>{
                {.id = encodeId(DeviceScenarioId{dsModel::DeviceId{"1111111111111111111111111111111100"},
                                                 dsModel::DeviceScenarioActionId{actionId}}),
                 .attributes = dsModel::DeviceScenarioAttributes{
                     .name = {.domain = {}, .message = "name"},
                     .dsDevice = dsModel::DeviceId{"1111111111111111111111111111111100"},
                     .actionId = dsModel::DeviceScenarioActionId{actionId}}}});
    }
  }

  SECTION("invoke standard scenario") {
    auto& zone0 = *apartment.allocateZone(0);
    auto& zone3 = *apartment.allocateZone(3);
    auto& group00 = *zone0.getGroup(0);
    auto& group3Lights = *zone3.getGroup(GroupIDYellow);
    auto& group3Shades = *zone3.getGroup(GroupIDGray);

    SECTION("Mixed shades and legacy awnings in apartment context") {
      auto& zone5 = *apartment.allocateZone(5);
      auto& zoneNewDevices = *apartment.allocateZone(65534);
      auto& group5Shades = *zone5.getGroup(GroupIDGray);
      auto& group5NewDevices = *zoneNewDevices.getGroup(GroupIDGray);

      auto& deviceAwning = *apartment.allocateDevice(Dsuid{"5111111111111111111111111111111100"});
      deviceAwning.setHardwareType(HardwareType::GR_KL210);
      deviceAwning.addToGroup(GroupIDGray);
      DeviceReference devAwningRef(deviceAwning.sharedFromThis(), &apartment);
      DeviceSpec specAwning{};
      specAwning.FunctionID = 0x6001; // This defines that device has output
      specAwning.activeGroup = static_cast<int>(ApplicationType::Blinds);
      deviceAwning.setPartiallyFromSpec(specAwning);

      auto& deviceShade = *apartment.allocateDevice(Dsuid{"6111111111111111111111111111111100"});
      deviceShade.setHardwareType(HardwareType::GR_KL200);
      deviceShade.addToGroup(GroupIDGray);
      DeviceReference devShadeRef(deviceShade.sharedFromThis(), &apartment);
      DeviceSpec specShade{};
      specShade.FunctionID = 0x6001; // This defines that device has output
      specShade.activeGroup = static_cast<int>(ApplicationType::Blinds);
      deviceShade.setPartiallyFromSpec(specShade);

      auto checkRequire = [&](std::string actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
        dsScenarios->invokeStandardScenario({.context = "applicationApartment",
                                             .actionId = actionId,
                                             .application = "shades",
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkRequireAwnings = [&](std::string actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
        dsScenarios->invokeStandardScenario({.context = "applicationApartment",
                                             .actionId = actionId,
                                             .application = "awnings",
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkForbid = [&](std::string actionId, std::optional<int> scene) {
        REQUIRE_CALL(busInterface.action, callScene(&group5NewDevices, coJSON, SAC_MANUAL, *scene, "", false));
        FORBID_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
        dsScenarios->invokeStandardScenario({.context = "applicationApartment",
                                             .actionId = actionId,
                                             .application = "shades",
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };
      auto checkForbidAwnings = [&](std::string actionId, std::optional<int> scene) {
        FORBID_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
        dsScenarios->invokeStandardScenario({.context = "applicationApartment",
                                             .actionId = actionId,
                                             .application = "awnings",
                                             .dsDevice = {},
                                             .zone = {},
                                             .area = {}});
      };

      SECTION("Only shades in zone") {
        zone5.addDevice(devShadeRef);

        checkRequire("up", 5);
        checkRequire("down", 0);
        checkRequire("stepUp", 12);
        checkRequire("stepDown", 11);
        checkRequire("stop", 15);
      }
      SECTION("Both awnings and shades in zone") {
        zone5.addDevice(devAwningRef);
        zone5.addDevice(devShadeRef);

        checkRequire("up", 5);
        checkRequire("down", 0);
        checkRequire("stepUp", 12);
        checkRequire("stepDown", 11);
        checkRequire("stop", 15);
        checkForbidAwnings("out", 5);
        checkForbidAwnings("in", 0);
        checkForbidAwnings("stepOut", 12);
        checkForbidAwnings("stepIn", 11);
        checkForbidAwnings("stop", 15);
        checkForbidAwnings("sunProtection", 56);
      }
      SECTION("Only awnings in zone") {
        zone5.addDevice(devAwningRef);

        checkForbid("up", 5);
        checkForbid("down", 0);
        checkForbid("stepUp", 12);
        checkForbid("stepDown", 11);
        checkForbid("stop", 15);
        checkRequireAwnings("out", 5);
        checkRequireAwnings("in", 0);
        checkRequireAwnings("stepOut", 12);
        checkRequireAwnings("stepIn", 11);
        checkRequireAwnings("stop", 15);
        checkRequireAwnings("sunProtection", 56);
      }

      SECTION("Check for zoneApplication") {
        auto checkForbidAwningsZone = [&](std::string actionId, std::optional<int> scene) {
          FORBID_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
          dsScenarios->invokeStandardScenario({.context = "applicationZone",
                                               .actionId = actionId,
                                               .application = "awnings",
                                               .dsDevice = {},
                                               .zone = "5",
                                               .area = {}});
        };
        auto checkRequireAwningsZone = [&](std::string actionId, std::optional<int> scene) {
          REQUIRE_CALL(busInterface.action, callScene(&group5Shades, coJSON, SAC_MANUAL, *scene, "", false));
          dsScenarios->invokeStandardScenario({.context = "applicationZone",
                                               .actionId = actionId,
                                               .application = "awnings",
                                               .dsDevice = {},
                                               .zone = "5",
                                               .area = {}});
        };

        SECTION("Mixed shades and legacy awnings") {
          zone5.addDevice(devAwningRef);
          zone5.addDevice(devShadeRef);

          checkForbidAwningsZone("out", 5);
          checkForbidAwningsZone("in", 0);
          checkForbidAwningsZone("stepOut", 12);
          checkForbidAwningsZone("stepIn", 11);
          checkForbidAwningsZone("stop", 15);
          checkForbidAwningsZone("sunProtection", 56);
        }
        SECTION("Only awnings") {
          zone5.addDevice(devAwningRef);

          checkRequireAwningsZone("out", 5);
          checkRequireAwningsZone("in", 0);
          checkRequireAwningsZone("stepOut", 12);
          checkRequireAwningsZone("stepIn", 11);
          checkRequireAwningsZone("stop", 15);
          checkRequireAwningsZone("sunProtection", 56);
        }
      }
    }

    struct Fail {};
    struct CallScene {
      int scene{};
      bool force = false;
    };
    struct UndoScene {
      int scene{};
    };
    using Operation = std::variant<Fail, CallScene, UndoScene>;
    auto check = [&](AddressableModelItem& target, dsModel::InvokeStandardScenarioArgs args,
                     const Operation& operation) {
      INFO(args);
      std::visit(ds::Visitor{
                     [&](const Fail& x) { CHECK_THROWS(dsScenarios->invokeStandardScenario(std::move(args))); },
                     [&](const CallScene& x) {
                       INFO(x.scene);
                       INFO(x.force);
                       REQUIRE_CALL(busInterface.action, callScene(&target, coJSON, SAC_MANUAL, x.scene, "", x.force));
                       dsScenarios->invokeStandardScenario(std::move(args));
                     },
                     [&](const UndoScene& x) {
                       INFO(x.scene);
                       REQUIRE_CALL(busInterface.action, undoScene(&target, coJSON, SAC_MANUAL, x.scene, ""));
                       dsScenarios->invokeStandardScenario(std::move(args));
                     },
                 },
                 operation);
    };
    SECTION("applicationApartment") {
      struct X {
        dsModel::ApartmentActionId actionId{};
        std::string application;
        Operation operation{};
      };
      for (const auto& x : std::initializer_list<X>{
               {{}, {}, Fail{}},
               {dsModel::ApartmentActionId::absent, "", CallScene{SceneAbsent}},
               {dsModel::ApartmentActionId::absent, "access", CallScene{SceneAbsent}}, // "access" is deprecated
               {dsModel::ApartmentActionId::present, "", CallScene{ScenePresent}},
               {dsModel::ApartmentActionId::present, "access", CallScene{ScenePresent}}, // "access" is deprecated
               {dsModel::ApartmentActionId::sleeping, "", CallScene{SceneSleeping}},
               {dsModel::ApartmentActionId::wakeup, "", CallScene{SceneWakeUp}},
               {dsModel::ApartmentActionId::doorBell, "", CallScene{SceneBell}},
               {dsModel::ApartmentActionId::panic, "", CallScene{ScenePanic}},
               {dsModel::ApartmentActionId::panicEnd, "", UndoScene{ScenePanic}},
               {dsModel::ApartmentActionId::fire, "", CallScene{SceneFire}},
               {dsModel::ApartmentActionId::fireEnd, "", UndoScene{SceneFire}},
               {dsModel::ApartmentActionId::alarm1, "", CallScene{SceneAlarm}},
               {dsModel::ApartmentActionId::alarm2, "", CallScene{SceneAlarm2}},
               {dsModel::ApartmentActionId::alarm3, "", CallScene{SceneAlarm3}},
               {dsModel::ApartmentActionId::alarm4, "", CallScene{SceneAlarm4}},
               {dsModel::ApartmentActionId::alarm1End, "", UndoScene{SceneAlarm}},
               {dsModel::ApartmentActionId::alarm2End, "", UndoScene{SceneAlarm2}},
               {dsModel::ApartmentActionId::alarm3End, "", UndoScene{SceneAlarm3}},
               {dsModel::ApartmentActionId::alarm4End, "", UndoScene{SceneAlarm4}},
           }) {
        const auto& actionId = x.actionId;
        check(group00,
              {.context = "applicationApartment",
               .actionId = actionId != decltype(actionId){} ? ds::str(actionId) : std::string{},
               .application = x.application},
              x.operation);
      }
      SECTION("zone, area must not be passed") {
        check(group00,
              {.context = "applicationApartment",
               .actionId = "absent",
               .application = "access",
               .dsDevice = {},
               .zone = "0",
               .area = {}},
              {});
        check(group00,
              {.context = "applicationApartment",
               .actionId = "absent",
               .application = "access",
               .dsDevice = {},
               .zone = {},
               .area = "1"},
              {});
      }
    }
    SECTION("applicationZone") {
      // all zone applications can be invoked also in the context of the whole apartment
      auto checkZoneApplication = [&](int groupId, dsModel::ZoneApplicationType application, auto actionId,
                                      const Operation& operation) {
        auto impl = [&](auto&& context, auto&& zone) {
          check(*zone.getGroup(groupId),
                {.context = context,
                 .actionId = actionId != decltype(actionId){} ? ds::str(actionId) : std::string{},
                 .application = application != decltype(application){} ? ds::str(application) : std::string{},
                 .dsDevice = {},
                 .zone = zone.getID() ? ds::str(zone.getID()) : std::string{}},
                operation);
        };
        impl("applicationZone", zone3);
        impl("applicationApartment", zone0);
      };
      auto checkZoneApplicationCommonAction = [&](int groupId, dsModel::ZoneApplicationType application) {
        struct X {
          dsModel::CommonActionId actionId{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, Fail{}},
                 {dsModel::CommonActionId::preset0, CallScene{SceneOff}},
                 {dsModel::CommonActionId::preset1, CallScene{Scene1}},
                 {dsModel::CommonActionId::preset2, CallScene{Scene2}},
                 {dsModel::CommonActionId::preset3, CallScene{Scene3}},
                 {dsModel::CommonActionId::preset4, CallScene{Scene4}},
                 {dsModel::CommonActionId::preset10, CallScene{SceneOffE1}},
                 {dsModel::CommonActionId::preset11, CallScene{SceneOnE1}},
                 {dsModel::CommonActionId::preset12, CallScene{Scene12}},
                 {dsModel::CommonActionId::preset13, CallScene{Scene13}},
                 {dsModel::CommonActionId::preset14, CallScene{Scene14}},
                 {dsModel::CommonActionId::preset20, CallScene{SceneOffE2}},
                 {dsModel::CommonActionId::preset21, CallScene{SceneOnE2}},
                 {dsModel::CommonActionId::preset22, CallScene{Scene22}},
                 {dsModel::CommonActionId::preset23, CallScene{Scene23}},
                 {dsModel::CommonActionId::preset24, CallScene{Scene24}},
                 {dsModel::CommonActionId::preset30, CallScene{SceneOffE3}},
                 {dsModel::CommonActionId::preset31, CallScene{SceneOnE3}},
                 {dsModel::CommonActionId::preset32, CallScene{Scene32}},
                 {dsModel::CommonActionId::preset33, CallScene{Scene33}},
                 {dsModel::CommonActionId::preset34, CallScene{Scene34}},
                 {dsModel::CommonActionId::preset40, CallScene{SceneOffE4}},
                 {dsModel::CommonActionId::preset41, CallScene{SceneOnE4}},
                 {dsModel::CommonActionId::preset42, CallScene{Scene42}},
                 {dsModel::CommonActionId::preset43, CallScene{Scene43}},
                 {dsModel::CommonActionId::preset44, CallScene{Scene44}},
             }) {
          checkZoneApplication(groupId, application, x.actionId, x.operation);
        }
      };
      SECTION("broadcast") {
        struct X {
          dsModel::ZoneActionId actionId;
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, Fail{}},
                 {dsModel::ZoneActionId::deepOff, CallScene{SceneDeepOff}},
                 {dsModel::ZoneActionId::standby, CallScene{SceneStandBy}},
             }) {
          checkZoneApplication(GroupIDBroadcast, {}, x.actionId, x.operation);
        }
      }
      SECTION("lights application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDYellow, dsModel::ZoneApplicationType::lights);
        }
        SECTION("lights action") {
          struct X {
            dsModel::LightsActionId actionId{};
            Operation operation{};
          };
          for (const auto& x : std::initializer_list<X>{
                   {{}, Fail{}},
                   {dsModel::LightsActionId::off, CallScene{SceneOff}},
                   {dsModel::LightsActionId::on, CallScene{Scene1}},
                   {dsModel::LightsActionId::brighter, CallScene{SceneInc}},
                   {dsModel::LightsActionId::darker, CallScene{SceneDec}},
                   {dsModel::LightsActionId::forcedOff, CallScene{SceneOff, true}},
               }) {
            checkZoneApplication(GroupIDYellow, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
          }
        }
      }
      SECTION("shades application") {
        SECTION("common action") {
          checkZoneApplicationCommonAction(GroupIDGray, dsModel::ZoneApplicationType::shades);
        }
        SECTION("shades action") {
          struct X {
            dsModel::ShadesActionId actionId{};
            Operation operation{};
          };
          for (const auto& x : std::initializer_list<X>{
                   {{}, Fail{}},
                   {dsModel::ShadesActionId::down, CallScene{SceneOff}},
                   {dsModel::ShadesActionId::up, CallScene{Scene1}},
                   {dsModel::ShadesActionId::stepDown, CallScene{SceneDec}},
                   {dsModel::ShadesActionId::stepUp, CallScene{SceneInc}},
                   {dsModel::ShadesActionId::stop, CallScene{SceneStop}},
               }) {
            checkZoneApplication(GroupIDGray, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
          }
        }
      }
      SECTION("audio application") {
        SECTION("common action") { checkZoneApplicationCommonAction(GroupIDCyan, dsModel::ZoneApplicationType::audio); }
        SECTION("audio action") {
          struct X {
            dsModel::AudioActionId actionId{};
            Operation operation{};
          };
          for (const auto& x : std::initializer_list<X>{
                   {{}, Fail{}},
                   {dsModel::AudioActionId::play, CallScene{Scene1}},
                   {dsModel::AudioActionId::pause, CallScene{SceneOff}},
                   {dsModel::AudioActionId::volumeUp, CallScene{SceneInc}},
                   {dsModel::AudioActionId::volumeDown, CallScene{SceneDec}},
               }) {
            checkZoneApplication(GroupIDCyan, dsModel::ZoneApplicationType::audio, x.actionId, x.operation);
          }
        }
      }
      SECTION("area must not be passed") {
        check(group00,
              {.context = "applicationZone",
               .actionId = "absent",
               .application = "access",
               .dsDevice = {},
               .zone = {},
               .area = "1"},
              Fail{});
      }
    }
    SECTION("application area") {
      auto checkAreaApplication = [&](Group& group, int area, dsModel::ZoneApplicationType application, auto actionId,
                                      const Operation& operation) {
        check(group,
              {.context = "applicationArea",
               .actionId = actionId != decltype(actionId){} ? ds::str(actionId) : std::string{},
               .application = application != decltype(application){} ? ds::str(application) : std::string{},
               .dsDevice = {},
               .zone = ds::str(group.getZoneID()),
               .area = ds::str(area)},
              operation);
      };
      SECTION("lights application") {
        struct X {
          dsModel::LightsActionId actionId{};
          int area{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, {}, Fail{}},
                 {dsModel::LightsActionId::off, 0, Fail{}},
                 {dsModel::LightsActionId::off, 1, CallScene{SceneOffA1}},
                 {dsModel::LightsActionId::off, 2, CallScene{SceneOffA2}},
                 {dsModel::LightsActionId::off, 3, CallScene{SceneOffA3}},
                 {dsModel::LightsActionId::off, 4, CallScene{SceneOffA4}},
                 {dsModel::LightsActionId::on, 0, Fail{}},
                 {dsModel::LightsActionId::on, 1, CallScene{SceneA11}},
                 {dsModel::LightsActionId::on, 2, CallScene{SceneA21}},
                 {dsModel::LightsActionId::on, 3, CallScene{SceneA31}},
                 {dsModel::LightsActionId::on, 4, CallScene{SceneA41}},
                 {dsModel::LightsActionId::brighter, 0, Fail{}},
                 {dsModel::LightsActionId::brighter, 1, CallScene{SceneIncA1}},
                 {dsModel::LightsActionId::brighter, 2, CallScene{SceneIncA2}},
                 {dsModel::LightsActionId::brighter, 3, CallScene{SceneIncA3}},
                 {dsModel::LightsActionId::brighter, 4, CallScene{SceneIncA4}},
                 {dsModel::LightsActionId::darker, 0, Fail{}},
                 {dsModel::LightsActionId::darker, 1, CallScene{SceneDecA1}},
                 {dsModel::LightsActionId::darker, 2, CallScene{SceneDecA2}},
                 {dsModel::LightsActionId::darker, 3, CallScene{SceneDecA3}},
                 {dsModel::LightsActionId::darker, 4, CallScene{SceneDecA4}},
                 {dsModel::LightsActionId::forcedOff, 0, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 1, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 2, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 3, Fail{}},
                 {dsModel::LightsActionId::forcedOff, 4, Fail{}},
             }) {
          checkAreaApplication(group3Lights, x.area, dsModel::ZoneApplicationType::lights, x.actionId, x.operation);
        }
      }
      SECTION("shades application") {
        struct X {
          dsModel::ShadesActionId actionId{};
          int area{};
          Operation operation{};
        };
        for (const auto& x : std::initializer_list<X>{
                 {{}, {}, Fail{}},
                 {dsModel::ShadesActionId::down, 0, Fail{}},
                 {dsModel::ShadesActionId::down, 1, CallScene{SceneOffA1}},
                 {dsModel::ShadesActionId::down, 2, CallScene{SceneOffA2}},
                 {dsModel::ShadesActionId::down, 3, CallScene{SceneOffA3}},
                 {dsModel::ShadesActionId::down, 4, CallScene{SceneOffA4}},
                 {dsModel::ShadesActionId::up, 0, Fail{}},
                 {dsModel::ShadesActionId::up, 1, CallScene{SceneA11}},
                 {dsModel::ShadesActionId::up, 2, CallScene{SceneA21}},
                 {dsModel::ShadesActionId::up, 3, CallScene{SceneA31}},
                 {dsModel::ShadesActionId::up, 4, CallScene{SceneA41}},
                 {dsModel::ShadesActionId::stepDown, 0, Fail{}},
                 {dsModel::ShadesActionId::stepDown, 1, CallScene{SceneDecA1}},
                 {dsModel::ShadesActionId::stepDown, 2, CallScene{SceneDecA2}},
                 {dsModel::ShadesActionId::stepDown, 3, CallScene{SceneDecA3}},
                 {dsModel::ShadesActionId::stepDown, 4, CallScene{SceneDecA4}},
                 {dsModel::ShadesActionId::stepUp, 0, Fail{}},
                 {dsModel::ShadesActionId::stepUp, 1, CallScene{SceneIncA1}},
                 {dsModel::ShadesActionId::stepUp, 2, CallScene{SceneIncA2}},
                 {dsModel::ShadesActionId::stepUp, 3, CallScene{SceneIncA3}},
                 {dsModel::ShadesActionId::stepUp, 4, CallScene{SceneIncA4}},
                 {dsModel::ShadesActionId::stop, 0, Fail{}},
                 {dsModel::ShadesActionId::stop, 1, CallScene{SceneStopA1}},
                 {dsModel::ShadesActionId::stop, 2, CallScene{SceneStopA2}},
                 {dsModel::ShadesActionId::stop, 3, CallScene{SceneStopA3}},
                 {dsModel::ShadesActionId::stop, 4, CallScene{SceneStopA4}},
             }) {
          checkAreaApplication(group3Shades, x.area, dsModel::ZoneApplicationType::shades, x.actionId, x.operation);
        }
      }
      SECTION("common action fails") {
        for (const auto& application : {dsModel::AreaApplicationType::none, dsModel::AreaApplicationType::lights,
                                        dsModel::AreaApplicationType::shades}) {
          for (const auto& actionOption : dsEnum(dsModel::CommonActionId{}).options) {
            const auto actionId = actionOption.first;
            INFO(application);
            INFO(actionId);
            CHECK_THROWS(dsScenarios->invokeStandardScenario(
                {.context = "applicationArea",
                 .actionId = actionId != decltype(actionId){} ? ds::str(actionId) : std::string{},
                 .application = application != decltype(application){} ? ds::str(application) : std::string{},
                 .dsDevice = {},
                 .zone = "3",
                 .area = "2"}));
          }
        }
      }
    }
    SECTION("applicationDevice") {
      auto& device = *apartment.allocateDevice(Dsuid{"1111111111111111111111111111111100"});
      struct X {
        dsModel::DeviceActionId actionId{};
        Operation operation{};
      };
      for (const auto& x : std::initializer_list<X>{
               {{}, Fail{}},
               {dsModel::DeviceActionId::off, CallScene{SceneMin}},
               {dsModel::DeviceActionId::on, CallScene{SceneMax}},
               {dsModel::DeviceActionId::stop, CallScene{SceneStop}},
               {dsModel::DeviceActionId::inc, CallScene{SceneInc}},
               {dsModel::DeviceActionId::dec, CallScene{SceneDec}},
           }) {
        const auto& actionId = x.actionId;
        check(device,
              {.context = "applicationDevice",
               .actionId = actionId != decltype(actionId){} ? ds::str(actionId) : std::string{},
               .application = {},
               .dsDevice = ds::str(device.getDSID())},
              x.operation);
      }

      SECTION("application, zone, area must not be passed") {
        check(group00,
              {.context = "applicationDevice",
               .actionId = "off",
               .application = "access",
               .dsDevice = ds::str(device.getDSID()),
               .zone = {},
               .area = {}},
              Fail{});
        check(group00,
              {.context = "applicationDevice",
               .actionId = "off",
               .application = {},
               .dsDevice = ds::str(device.getDSID()),
               .zone = "0",
               .area = {}},
              Fail{});
        check(group00,
              {.context = "applicationDevice",
               .actionId = "off",
               .application = {},
               .dsDevice = ds::str(device.getDSID()),
               .zone = {},
               .area = "1"},
              Fail{});
      }
    }
  }
}

} // namespace
