/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "latency-watchdog.h"
#include <ds/asio/loop.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/log.h>
#include <trompeloeil.hpp>

static ds::LogChannel dsLogChannel("dsAsioLatencyWatchdogTest");

static const char* TAGS = "[dsAsioLatencyWatchdog][dsAsio][ds]";

using namespace std::literals;
using namespace trompeloeil;

struct Fire {
public:
  MAKE_MOCK0(call, void());
};

TEST_CASE("dsAsioLatencyWatchdog", TAGS) {
  {
    // extra logging to investigate CI failures
    auto& logger = ds::Logger::instance();
    logger.setChannelSeverity("dsClock", ds::LogSeverity::DEBUG);
    logger.setChannelSeverity("dsAsioLatencyWatchdog", ds::LogSeverity::DEBUG);
    logger.setChannelSeverity("dsAsioTimer", ds::LogSeverity::DEBUG);
  }

  auto clock = ds::ClockMocker{};
  auto watchedLoop = ds::asio::TestLoop{};
  auto watchdogLoop = ds::asio::TestLoop{};

  const auto latency = 1000ms;
  auto fire = Fire{};
  auto latencyWatchdog =
      ds::asio::LatencyWatchdog{watchedLoop.io, watchdogLoop.io, latency, [&](auto&&) { fire.call(); }};

  SECTION("can be created and destroyed") {}

  SECTION("does not fire when watchedIo is executing handlers") {
    for (auto i = 0; i < 30; i++) {
      clock.advance(latency / 5);
      watchedLoop.postStopRun();
      watchdogLoop.postStopRun();
    }

    SECTION("does not fire when watchedIo does not execute handlers for latency.") {
      for (auto i = 0; i < 4; i++) {
        clock.advance(latency / 5);
        watchdogLoop.postStopRun();
      }

      SECTION("fires when watchedIo does not execute handlers for 2*latency.") {
        REQUIRE_CALL(fire, call());
        for (auto i = 0; i < 10; i++) {
          clock.advance(latency / 5);
          watchdogLoop.postStopRun();
        }
      }
    }
  }
  {
    // extra logging to investigate CI failures
    auto& logger = ds::Logger::instance();
    logger.setChannelSeverity("dsClock", ds::LogSeverity::NOTICE);
    logger.setChannelSeverity("dsAsioLatencyWatchdog", ds::LogSeverity::NOTICE);
    logger.setChannelSeverity("dsAsioTimer", ds::LogSeverity::NOTICE);
  }
}

// real assertion. Not part of test suite
TEST_CASE("dsAsioLatencyWatchdog assert", "[!hide]") {
  auto conf = ds::ConfMock{};
  const auto config = ds::asio::LatencyWatchdogConfig{.warningLatency = 1s, .assertLatency = 3s};
  auto loop = ds::asio::Loop{conf, "testLoop", ds::asio::LatencyWatchdogConfig{config}};

  DS_NOTICE("Event loop running.");
  std::this_thread::sleep_for(config.warningLatency * 2);

  DS_NOTICE("Event loop blocked.");
  loop.io().post([&] { std::this_thread::sleep_for(config.assertLatency * 2); });
  std::this_thread::sleep_for(config.assertLatency * 3);
}
