/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "latency-watchdog.h"
#include <ds/asio/timer.h>
#include <ds/log.h>
#include <ds/stack-trace.h>
#include <atomic>
#include <future>
#include "loop-impl.h"

#include <execinfo.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

static ds::LogChannel dsLogChannel("dsAsioLatencyWatchdog");

using namespace std::literals;

namespace ds::asio {

static void assertSignalHandler(int) {
  DS_FAIL_ASSERT("This thread stalled.");
}

static void assertWithStackTrace(pthread_t threadId) {
  const auto signal = SIGUSR2;

  // setup signal handler
  struct sigaction sa {};
  struct sigaction oldSa {};
  sigfillset(&sa.sa_mask);
  sa.sa_handler = assertSignalHandler;
  sigaction(signal, &sa, &oldSa);

  // invoke signal handler in target thread
  pthread_kill(threadId, signal);
}

struct LatencyWatchdog::Impl {
  boost::asio::io_service &io;
  boost::asio::io_service &watchogIo;
  const ds::Clock::duration latency{};
  std::function<void(ds::Clock::duration)> fire{};
  ds::asio::Timer m_timer{watchogIo};
  ds::Clock::time_point m_pingedAt{ds::Clock::now()};
  std::atomic_bool m_ponged{true};
  pthread_t m_watchedThreadId{};

  void asyncPingLoop() {
    DS_DEBUG_ENTER(ds::Clock::now());

    if (latency == 0s) { // disabled
      return;
    }

    m_timer.randomlyExpiresFromNow(latency, [this] {
      DS_DEBUG_ENTER(m_ponged, m_pingedAt, m_watchedThreadId);
      if (!m_ponged) {
        auto fireLatency = ds::Clock::now() - m_pingedAt;
        DS_DEBUG("Fire", fireLatency, m_watchedThreadId);
        fire(fireLatency);
      } else {
        m_pingedAt = ds::Clock::now();
        m_ponged = false;
      }
      DS_INFO("Ping", m_pingedAt);
      io.post([this]() {
        DS_INFO("Pong", ds::Clock::now());
        m_ponged = true;
        m_watchedThreadId = pthread_self();
      });
      asyncPingLoop();
    });
  }

  void fireByConfig(const LatencyWatchdogConfig &config, ds::Clock::duration latency) {
    if (config.warningLatency != 0s && latency >= config.warningLatency) {
      DS_WARNING("LatencyWatchdog failed.", latency, m_watchedThreadId);
    }

    if (config.assertLatency != 0s && latency >= config.assertLatency) {
      DS_ERROR("LatencyWatchdog failed.", latency, m_watchedThreadId);
      if (m_watchedThreadId) {
        assertWithStackTrace(m_watchedThreadId);
        std::this_thread::sleep_for(2s);
      }
      // we shall not get here, the signal hander shall assert
      DS_FAIL_ASSERT("LatencyWatchdog failed without stalled stacktrace");
    }
  }
};

boost::asio::io_service &globalWatchdogIo() {
  static LoopImpl s_loop{};
  return s_loop.io;
}

LatencyWatchdog::LatencyWatchdog(boost::asio::io_service &io, boost::asio::io_service &watchdogIo,
                                 ds::Clock::duration latency, std::function<void(ds::Clock::duration)> &&fire)
    : m_impl(new Impl{.io = io, .watchogIo = watchdogIo, .latency = latency, .fire = std::move(fire)}) {
  DS_INFO_ENTER(latency);
  m_impl->asyncPingLoop();
}

LatencyWatchdog::LatencyWatchdog(boost::asio::io_service &io, ds::Clock::duration latency,
                                 std::function<void(ds::Clock::duration)> &&fire)
    : LatencyWatchdog(io, globalWatchdogIo(), latency, std::move(fire)) {}

LatencyWatchdog::LatencyWatchdog(boost::asio::io_service &io, LatencyWatchdogConfig config)
    : LatencyWatchdog(io, config.warningLatency != 0s ? config.warningLatency : config.assertLatency,
                      [this, config = std::move(config)](auto latency) { m_impl->fireByConfig(config, latency); }) {}

LatencyWatchdog::~LatencyWatchdog() {
  DS_INFO_ENTER("");
}

} // namespace ds::asio
