/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "moderation-timer.h"
#include <ds/asio/catch.h>
#include <ds/clock-mock.h>
#include <trompeloeil.hpp>

namespace {

static const char* TAGS = "[dsAsioModerationTimer][dsAsio][ds]";

using namespace trompeloeil;
using namespace std::literals;

struct Fire {
public:
  MAKE_MOCK0(call, void());
};

SCENARIO("dsAsioModerationTimer", TAGS) {
  auto clock = ds::ClockMocker{};
  auto io = boost::asio::io_service{};
  auto work = boost::asio::io_service::work{io};
  auto ioResetRun = [&] {
    io.reset();
    io.run();
  };
  auto postPostIoStop = [&] { io.post([&] { io.post([&] { io.stop(); }); }); };

  auto fire = Fire{};
  const auto period = 1000ms;
  auto moderationTimer = ds::asio::ModerationTimer{io, [&] { fire.call(), io.stop(); }, period};

  SECTION("can be created and deleted") {}

  WHEN("triggered") {
    moderationTimer.trigger();

    THEN("fires immediately") {
      {
        REQUIRE_CALL(fire, call());
        ioResetRun();
      }

      WHEN("triggered") {
        moderationTimer.trigger();

        THEN("not fire immediately") {
          postPostIoStop();
          ioResetRun();
        }

        THEN("fires after period") {
          clock.advance(0s);
          {
            clock.advance(period * 3 / 4);
            REQUIRE_CALL(fire, call());
            clock.advance(period / 2);
            ioResetRun();
          }

          WHEN("triggered") {
            moderationTimer.trigger();

            THEN("not fire immediately") {
              postPostIoStop();
              ioResetRun();
            }

            THEN("fires after period") {
              clock.advance(period * 3 / 4);
              REQUIRE_CALL(fire, call());
              clock.advance(period / 2);
              ioResetRun();
            }
          }
        }

        WHEN("triggered again after 3/4 period") {
          clock.advance(period * 3 / 4);
          moderationTimer.trigger();

          THEN("fires after original period once") {
            REQUIRE_CALL(fire, call());
            clock.advance(period / 2);
            ioResetRun();
          }
        }

        WHEN("triggered again after 5/4 period") {
          clock.advance(period * 5 / 4);
          moderationTimer.trigger();

          THEN("fires immediately") {
            REQUIRE_CALL(fire, call());
            ioResetRun();
          }
        }

        WHEN("cancelled") {
          moderationTimer.cancel();
          THEN("not fire") {
            postPostIoStop();
            ioResetRun();
          }
        }

        WHEN("cancelled after 3/4 period") {
          clock.advance(period * 3 / 4);
          moderationTimer.cancel();
          THEN("not fire") {
            postPostIoStop();
            ioResetRun();
          }
        }
      }
    }

    WHEN("cancelled") {
      moderationTimer.cancel();
      THEN("not fire") {
        postPostIoStop();
        ioResetRun();
      }
    }
  }
}

} // namespace
