#include "reconnect-timer.h"
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/log.h>
#include <boost/range/algorithm/adjacent_find.hpp>
#include <trompeloeil.hpp>

namespace {

static const char* TAGS = "[dsAsioReconnectTimer][dsAsio][ds]";

using namespace std::literals;

struct Connect {
public:
  MAKE_MOCK0(call, void());
};

SCENARIO("dsAsioReconnectTimer", TAGS) {
  auto clock = ds::ClockMocker{};
  auto io = boost::asio::io_service{};
  auto work = boost::asio::io_service::work{io};
  auto reconnectTimer = ds::asio::ReconnectTimer{io};
  auto connect = Connect{};

  auto ioResetRun = [&] {
    io.reset();
    io.run();
  };

  THEN("default schedule is sane") {
    auto schedule = reconnectTimer.schedule();
    CHECK(schedule.front() == 0s);
    CHECK(schedule.back() >= 10min);
    // exponential backoff
    CHECK(boost::adjacent_find(schedule, [](auto&& a, auto&& b) { return !(a < b && b - a > a); }) == schedule.end());
  }

  auto checkDelayedReconnect = [&](auto delay) {
    reconnectTimer.asyncDelayedReconnect([&] {
      io.stop();
      connect.call();
    });

    // not called too early
    if (delay > 0s) {
      clock.advance(delay * 3 / 4);
      ds::asio::Timer timer{io};
      timer.expiresNow([&] { io.stop(); });
      ioResetRun();
    }
    // called +- at time
    REQUIRE_CALL(connect, call());
    clock.advance(delay / 4);
    ioResetRun();

    CHECK(true);
  };

  THEN("asyncDelayedReconnect delayed according to schedule") {
    auto schedule = reconnectTimer.schedule();
    auto maxDelay = schedule.back();

    // default schedue with last value repeating
    schedule.push_back(maxDelay);
    schedule.push_back(maxDelay);

    for (const auto& delay : schedule) {
      INFO(delay.count());
      checkDelayedReconnect(delay);
    }

    WHEN("connected") {
      reconnectTimer.connected();
      THEN("reconnected with max delay if connected for short time") {
        clock.advance(reconnectTimer.resetScheduleAfterConnectedDuration() / 2);
        checkDelayedReconnect(schedule.back());
      }
      THEN("reconnected fast if connected long enough") {
        clock.advance(reconnectTimer.resetScheduleAfterConnectedDuration() * 3 / 2);
        checkDelayedReconnect(schedule.front());
      }
    }
  }
}

} // namespace
