/*
 *  Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <ds/log.h>
#include "timer.h"

namespace ds::asio {

using namespace std::literals;

/// Steady timer schedulling reconnects maintaining long lived connections.
///
/// The timer uses exponential backoff to schedule reconnects.
/// First reconnect is immediate.
///
/// Default reconnect schedule shall be suitable for internet use cases.
/// schedule: 0s, 2s, 5s, ... 20m, 20m, 20m, ...
/// resetScheduleAfterConnected: 1min
///
class ReconnectTimer {
public:
  ReconnectTimer(boost::asio::io_service& io);
  ~ReconnectTimer();

  template <class F>
  void asyncDelayedReconnect(F&& f) {
    m_timer.randomlyExpiresFromNowPercentDown(nextSchedule(), 20, [f = std::forward<F>(f)] { f(); });
  }

  /// Connection is connected. Schedule will be reset if the connection is connected long enough
  void connected();

  /// Reset the reconnection schedule. Next reconnect will happen fast.
  void resetSchedule();

  std::vector<Timer::duration> schedule() const { return m_schedule; }
  Timer::duration resetScheduleAfterConnectedDuration() const { return m_resetScheduleAfterConnectedDuration; }

private:
  Timer m_timer;
  const std::vector<Timer::duration> m_schedule{0s, 2s, 5s, 20s, 1min, 5min, 20min};
  const Timer::duration m_resetScheduleAfterConnectedDuration{1min};
  decltype(m_schedule)::const_iterator m_scheduleIt{m_schedule.cbegin()};
  std::optional<Timer::time_point> m_connectedAt{};

  Timer::duration nextSchedule();
};

} // namespace ds::asio
