/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include "resource-fwd.h"

#include <ds/conf.h>
#include <ds/system.h>
#include <boost/asio/io_service.hpp>
#include <string>
#include <string_view>

namespace ds::asio {

struct ResourceContext {
  boost::asio::io_service& io;
  ds::System& system;
};

class IResource {
public:
  IResource(const IResource&) = delete;
  IResource() = default;
  virtual ~IResource() = default;

  virtual ResourceContext& context() const = 0;
  virtual std::string_view id() const = 0;
  virtual boost::asio::io_service& io() = 0;
  const ds::Conf& conf() const { return context().system.conf(); }
};

std::ostream& operator<<(std::ostream&, const IResource&);

class ResourceBase : public IResource {
public:
  ResourceBase(ResourceBase& parent, std::string_view localId);
  ResourceBase(ResourceContext& context, std::string&& id);

  ResourceContext& context() const override final { return m_context; }
  std::string_view id() const override final { return m_id; }
  boost::asio::io_service& io() override final { return m_context.io; }

private:
  ResourceContext& m_context;
  const std::string m_id{};
};

std::ostream& operator<<(std::ostream&, const ResourceBase&);

struct NoResourceConfig {};

template <class Config>
class Resource : public ResourceBase {
public:
  Resource(ResourceBase& parent, std::string_view localId, Config&& defaultConfig = {})
      : ResourceBase(parent, std::move(localId)), m_config{readConfig(std::move(defaultConfig))} {}
  Resource(ResourceContext& context, std::string&& id, Config&& defaultConfig = {})
      : ResourceBase(context, std::move(id)), m_config{readConfig(std::move(defaultConfig))} {}

  const Config& config() const { return m_config; }

private:
  Config m_config{};

  Config readConfig(Config&& defaultConfig) {
    return context().system.conf().template readOr<Config>(id(), std::move(defaultConfig));
  }
};

template <>
class Resource<NoResourceConfig> : public ResourceBase {
public:
  using ResourceBase::ResourceBase;
};

} // namespace ds::asio
