/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "retryer.h"
#include <ds/asio/test.h>
#include <ds/clock-mock.h>

namespace {

static const char* TAGS = "[dsAsioRetryer][dsAsio][ds]";
using namespace std::chrono_literals;
using namespace trompeloeil;

struct HooksMock {
public:
  MAKE_MOCK0(attempt, void());
  MAKE_MOCK0(succeeded, void());
  MAKE_MOCK0(attemptFailed, void());
};

TEST_CASE("dsAsioRetryer", TAGS) {
  auto clock = ds::ClockMocker{};
  auto loop = ds::asio::TestLoopResource{};
  auto hooksMock = HooksMock{};
  auto retryer = ds::asio::Retryer{loop.resource, "retryer"};
  auto hooks = ds::asio::RetryerHooks{.attempt = [&] { hooksMock.attempt(); },
                                      .succeeded = [&] { hooksMock.succeeded(); },
                                      .attemptFailed = [&] { hooksMock.attemptFailed(); }};

  WHEN("asyncPerform") {
    retryer.asyncPerform(ds::asio::RetryerHooks{hooks});
    THEN("suceeded attempt is called immediately") {
      REQUIRE_CALL(hooksMock, attempt());
      REQUIRE_CALL(hooksMock, succeeded()).LR_SIDE_EFFECT(loop.stop());
      loop.run();

      WHEN("asyncPerform") {
        retryer.asyncPerform(ds::asio::RetryerHooks{hooks});
        THEN("suceeded attempt is called immediately") {
          REQUIRE_CALL(hooksMock, attempt());
          REQUIRE_CALL(hooksMock, succeeded()).LR_SIDE_EFFECT(loop.stop());
          loop.run();
        }
      }
    }

    THEN("failed attempt is called immediately") {
      REQUIRE_CALL(hooksMock, attempt()).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("test failure"));
      REQUIRE_CALL(hooksMock, attemptFailed()).LR_SIDE_EFFECT(loop.stop());
      loop.run();
      THEN("failed attempt is called after back off") {
        REQUIRE_CALL(hooksMock, attempt()).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("test failure"));
        REQUIRE_CALL(hooksMock, attemptFailed()).LR_SIDE_EFFECT(loop.stop());
        loop.runFor(clock, retryer.backOff().peek());

        THEN("failed attempt is called after back off") {
          REQUIRE_CALL(hooksMock, attempt()).LR_SIDE_EFFECT(DS_FAIL_REQUIRE("test failure"));
          REQUIRE_CALL(hooksMock, attemptFailed()).LR_SIDE_EFFECT(loop.stop());
          loop.runFor(clock, retryer.backOff().peek());

          THEN("suceeded attempt is called after back off") {
            REQUIRE_CALL(hooksMock, attempt());
            REQUIRE_CALL(hooksMock, succeeded()).LR_SIDE_EFFECT(loop.stop());
            loop.runFor(clock, retryer.backOff().peek());

            WHEN("asyncPerform") {
              retryer.asyncPerform(ds::asio::RetryerHooks{hooks});
              THEN("suceeded attempt is called immediately") {
                REQUIRE_CALL(hooksMock, attempt());
                REQUIRE_CALL(hooksMock, succeeded()).LR_SIDE_EFFECT(loop.stop());
                loop.run();
              }
            }
          }
        }
      }
    }
  }
}

} // namespace
