/*
 *  Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */
#include "state-poller.h"
#include <ds-crypt/digest.h>
#include <ds/log.h>
#include "moderation-timer.h"
#include "timer.h"

static ds::LogChannel dsLogChannel{"dsAsioStatePoller"};

namespace ds::asio {

struct StatePoller::Impl {
  boost::asio::io_service& m_io;
  const std::string m_id;
  StatePollerConfig m_config{};
  StatePollerHooks m_hooks{};
  bool m_isActive{};
  dsCrypt::Sha256 m_valueSum{};
  ds::asio::ModerationTimer m_moderationTimer{m_io, [this] { pollNow(); }, m_config.moderate};
  ds::asio::Timer m_loopTimer{m_io};

  void setActive(bool isActive) {
    DS_CONTEXT(m_id);
    DS_INFO_ENTER(isActive);
    m_io.dispatch([this, isActive] {
      if (m_isActive == isActive) {
        return;
      }
      DS_CONTEXT(m_id);
      DS_DEBUG_ENTER("");
      m_isActive = isActive;
      if (isActive) {
        m_moderationTimer.trigger();
      } else {
        m_moderationTimer.cancel();
        m_loopTimer.cancel();
        m_valueSum = {};
      }
    });
  }

  void poll() {
    DS_CONTEXT(m_id);
    DS_INFO_ENTER("");
    m_io.dispatch([this] {
      DS_CONTEXT(m_id);
      pollImpl();
    });
  }

private:
  void pollImpl() {
    DS_DEBUG_ENTER("");
    m_moderationTimer.trigger();
  }

  void pollNow() {
    DS_CONTEXT("poller:", ::ds::hideArg(m_id));
    DS_DEBUG_ENTER("");
    try {
      // Reset and continue async loop in any case.
      // But at the end to start timer after busy work is done
      DS_DEFER(asyncLoop());
      auto value = m_hooks.poll();
      {
        auto valueSum = dsCrypt::sha256Sum(as_bytes(gsl::make_span(ds::toJson(value))));
        if (valueSum == m_valueSum) {
          DS_DEBUG("The same");
          m_hooks.polledAndNoFire();
          return;
        }
        m_valueSum = valueSum;
      }
      m_hooks.fire(std::move(value));
    } catch (const std::exception& e) {
      DS_ERROR("Failed to poll.", e);
    }
  }

  void asyncLoop() {
    const auto& period = m_config.loop;
    DS_DEBUG_ENTER(period);
    if (period == 0s) {
      m_loopTimer.cancel();
    } else {
      m_loopTimer.randomlyExpiresFromNow(period, [&] {
        DS_DEBUG_ENTER(m_id);
        pollImpl();
      });
    }
  }
};

StatePollerHooks::~StatePollerHooks() {}

StatePoller::StatePoller(boost::asio::io_service& io, std::string&& id, StatePollerConfig&& config,
                         StatePollerHooks&& hooks)
    : m_impl{new Impl{.m_io = io, .m_id = std::move(id), .m_config = std::move(config), .m_hooks = std::move(hooks)}} {
  DS_NOTICE("StatePoller", ds::hideArg(m_impl->m_id), ds::hideArg(m_impl->m_config));
}

StatePoller::~StatePoller() {
  DS_INFO_ENTER(m_impl->m_id);
}

void StatePoller::poll() {
  m_impl->poll();
}

void StatePoller::setActive(bool x) {
  m_impl->setActive(x);
}

} // namespace ds::asio
