/*
 *  Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <ds/clock.h>
#include <ds/common.h>
#include <ds/log.h>
#include <boost/asio/steady_timer.hpp>
#include <chrono>
#include "io-service.h"

namespace ds::asio {

/// Steady mockable timer.
///
/// Scheduled callback is _not_ invoked after cancel() and after the timer is destroyed.
///
/// Nearly all code wants to use steady clock, so `steady` is skipped from the name.
///
class Timer {
public:
  using Component = ::boost::asio::basic_waitable_timer<ds::Clock, ds::ClockWaitTraits>;
  using duration = Component::duration;
  using time_point = Component::time_point;
  using clock_type = Component::clock_type;

  Timer(boost::asio::io_service &ioService);
  ~Timer();

  /// Convenient method calling `expires_from_now(x)` and `asyncWait(f)`.
  template <typename F>
  void expiresFromNow(duration d, F &&f) {
    expiresFromNow(d);
    asyncWait(std::forward<F>(f));
  }

  /// Convenient method calling `expiresFromNow(0, f)`.
  template <typename F>
  void expiresNow(F &&f) {
    expiresFromNow(duration(0), std::forward<F>(f));
  }

  /// Expires randomly in closed interval [a, b] relatively to now.
  template <typename F>
  void randomlyExpiresFromNow(duration a, duration b, F &&f) {
    randomlyExpiresFromNow(a, b);
    asyncWait(std::forward<F>(f));
  }

  /// Expires randomly in closed interval [d - d * p / 100, d] relatively to now.
  template <typename F>
  void randomlyExpiresFromNowPercentDown(duration d, int p, F &&f) {
    randomlyExpiresFromNowPercentDown(d, p);
    asyncWait(std::forward<F>(f));
  }

  /// Expires randomly in closed interval [d, d + 10%] relatively to now.
  template <typename F>
  void randomlyExpiresFromNow(duration d, F &&f) {
    randomlyExpiresFromNow(d);
    asyncWait(std::forward<F>(f));
  }

  /// Calls async_wait, but calls callback only on NO error.
  ///
  /// Callback can be sure that it is not called when timer is destroyed / cancelled.
  template <typename F>
  void asyncWait(F &&f) {
    static_assert(std::is_void<decltype(f())>::value, "required F type: void ()");

    auto seq = ++m_impl->seq;
    m_impl->component.async_wait(
        [sharedImpl = m_impl->shared_from_this(), f = std::move(f), seq](boost::system::error_code e) mutable {
          if (e || sharedImpl->seq != seq) {
            return; // timer was aborted
          }
          f();
        });
  }

  void cancel();

  boost::asio::io_service &ioService() { return m_impl->component.get_io_service(); }

private:
  struct Impl : std::enable_shared_from_this<Impl> {
    Component component;
    unsigned seq{};

    Impl(boost::asio::io_service &io);
    ~Impl();
  };
  const std::shared_ptr<Impl> m_impl;

  void expiresFromNow(duration d);
  void randomlyExpiresFromNow(duration a, duration b);
  void randomlyExpiresFromNow(duration a);
  void randomlyExpiresFromNowPercentDown(duration d, int p);
};

} // namespace ds::asio
