/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-signal.h"
#include <ds/catch/catch.h>
#include "ds-signal-mock.h"

using namespace std::literals;

static const char* TAGS = "[dsSignal][ds]";

using namespace trompeloeil;

namespace {
DS_SLOT_MOCK1(IntSlotMock, int);

TEST_CASE("dsSignal", TAGS) {
  auto io = boost::asio::io_service{};
  auto ioResetRun = [&] {
    io.reset();
    io.run();
  };

  SECTION("Invoke without slots does nothing.") {
    {
      ds::Signal<> signal;
      signal.emit();
      CHECK(true);
    }
    {
      ds::Signal<int> signal;
      signal.emit(5);
      CHECK(true);
    }
  }

  SECTION("Linked slot is invoked.") {
    {
      int x{};
      ds::Signal<> signal;
      auto slot = ds::Slot<>(io, signal, [&] { ++x; });
      signal.emit();
      ioResetRun();
      CHECK(x == 1);
    }
    {
      int sum{};
      ds::Signal<int> signal;
      auto slot = ds::Slot<int>(io, signal, [&](auto&& x) { sum += x; });
      signal.emit(5);
      ioResetRun();
      CHECK(sum == 5);
    }
  }

  SECTION("Linked mocked slot is invoked .") {
    {
      ds::Signal<> signal;
      auto slot = ds::VoidSlotMock(io, signal);
      REQUIRE_CALL(slot, called());
      signal.emit();
      ioResetRun();
    }
    {
      ds::Signal<int> signal;
      auto slot = IntSlotMock(io, signal);
      REQUIRE_CALL(slot, called(5));
      signal.emit(5);
      ioResetRun();
    }
  }

  SECTION("2 linked mocked slots are invoked .") {
    {
      ds::Signal<> signal;
      auto slot1 = ds::VoidSlotMock(io, signal);
      auto slot2 = ds::VoidSlotMock(io, signal);
      REQUIRE_CALL(slot1, called());
      REQUIRE_CALL(slot2, called());
      signal.emit();
      ioResetRun();
    }
    {
      ds::Signal<int> signal;
      auto slot1 = IntSlotMock(io, signal);
      auto slot2 = IntSlotMock(io, signal);
      REQUIRE_CALL(slot1, called(5));
      REQUIRE_CALL(slot2, called(5));
      signal.emit(5);
      ioResetRun();
    }
  }

  SECTION("Deleted slot is unlinked.") {
    int x = {};
    ds::Signal<> signal;
    {
      auto slot = ds::Slot<>(io, signal, [&] { ++x; });
    }
    signal.emit();
    ioResetRun();
    CHECK(x == 0);
  }

  SECTION("Linked slots are invoked in link order.") {
    int x = {};
    ds::Signal<> signal;
    auto slot1 = ds::Slot<>(io, signal, [&] {
      CHECK(x == 0);
      ++x;
    });
    auto slot2 = std::make_unique<ds::Slot<>>(io, signal, [&] { ++x; });
    auto slot3 = ds::Slot<>(io, signal, [&] {
      CHECK(x == 1);
      ++x;
    });
    slot2.reset();
    signal.emit();
    ioResetRun();
    CHECK(x == 2);
  }
}

} // namespace
