/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "json.h"
#include <ds/catch/catch.h>
#include <ds/str.h>
#include <array>
#include <boost/serialization/strong_typedef.hpp>
#include <cmath>
#include <tuple>
#include <variant>
static const char* TAGS = "[dsJson][ds]";

namespace {

using namespace std::literals;

enum class Transport { CAR, PLANE };

ds::Enum<Transport> dsEnum(Transport) {
  static constexpr std::pair<Transport, const char*> x[] = {
      {Transport::CAR, "car"}, {Transport::PLANE, "plane"},
  };
  return {gsl::make_span(x), "Transport"};
}
std::ostream& operator<<(std::ostream& stream, Transport x) { return ds::ostream(stream, x); }

struct X {
  int i;
  std::string s;
  Transport transport;
  ds::JsonFragment json = nullptr;
  std::optional<int> oi;
};

BOOST_STRONG_TYPEDEF(std::string, Id);

struct StructAsIdArray {
  X x;
};

struct A {
  int a;
};
struct B {
  int b;
};

struct AB {
  std::variant<std::monostate, A, B> ab{};
};

} // namespace

DS_REFLECT_ANONYMOUS_STRUCT(A, (a));
DS_REFLECT_ANONYMOUS_STRUCT(B, (b));
DS_REFLECT_ANONYMOUS_STRUCT(AB, (ab));
template <>
struct DSJsonVariant<&AB::ab> {
  static constexpr auto mode = ds::JsonVariantMode::alternativeInSiblingMember;
  static constexpr auto alternative = "type"; // name of sibling member carrying alternative
  static constexpr auto nameVisitor = ::ds::Visitor{
      // name for every alternative
      [](const std::monostate&) { return ""; },
      [](const A&) { return "a"; },
      [](const B&) { return "b"; },
  };
};

DS_REFLECT_ANONYMOUS_STRUCT(X, (i)(s)(transport)(json)(oi));
DS_REFLECT_ANONYMOUS_STRUCT(StructAsIdArray, (x));
template <>
struct DSJsonArrayId<StructAsIdArray> : public std::true_type {};

namespace {
TEST_CASE("fromJson", TAGS) {
  SECTION("bool") {
    CHECK(ds::fromJson<bool>("false") == false);
    CHECK(ds::fromJson<bool>("true") == true);

    SECTION("accepts only json bool") {
      CHECK_THROWS(ds::fromJson<std::string>("7"));
      CHECK_THROWS(ds::fromJson<std::string>("7.5"));
      CHECK_THROWS(ds::fromJson<std::string>("a"));
      CHECK_THROWS(ds::fromJson<std::string>("{}"));
      CHECK_THROWS(ds::fromJson<std::string>("[]"));
    }
  }

  SECTION("number") {
    CHECK(ds::fromJson<int>("7") == 7);
    CHECK(ds::fromJson<int>("5.4") == 5);
    CHECK(ds::fromJson<int>("5.5") == 5);
    CHECK(ds::fromJson<double>("7.5") == 7.5);
    CHECK(ds::fromJson<double>("7") == 7);
    CHECK(std::isnan(ds::fromJson<double>("\"nan\"")));
    CHECK(std::isnan(ds::fromJson<double>("\"NaN\"")));
    CHECK(ds::fromJson<double>("\"inf\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"Inf\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"infinity\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"Infinity\"") == std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-inf\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-Inf\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-infinity\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<double>("\"-Infinity\"") == -std::numeric_limits<double>::infinity());
    CHECK(ds::fromJson<unsigned char>("7") == 7);
    CHECK(ds::fromJson<signed char>("7") == 7);
    CHECK(ds::fromJson<int>("\"7\"") == 7);

    SECTION("fails on narrowing error") { CHECK_THROWS(ds::fromJson<unsigned char>("1000")); }

    SECTION("accepts only numbers in strings") {
      // TODO(someday): the code needs to be fixed
    }

    SECTION("accepts only json numbers || strings") {
      CHECK_THROWS(ds::fromJson<int>("false"));
      CHECK_THROWS(ds::fromJson<int>("true"));
      CHECK_THROWS(ds::fromJson<int>("{}"));
      CHECK_THROWS(ds::fromJson<int>("[]"));
    }
  }

  SECTION("std::string") {
    CHECK(ds::fromJson<std::string>(R"("")") == "");
    CHECK(ds::fromJson<std::string>(R"("a")") == "a");
    CHECK(ds::fromJson<std::string>(R"("hello")") == "hello");
    CHECK(ds::fromJson<std::string>("\"one\\ttwo\"") == "one\ttwo");

    SECTION("accepts only json strings") {
      CHECK_THROWS_FIND(ds::fromJson<std::string>("false"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<std::string>("true"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<std::string>("7"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<std::string>("7.5"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<std::string>("{}"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<std::string>("[]"), "Expecting json string");
    }
  }

  SECTION("Id") { CHECK(ds::fromJson<Id>(R"("hi")") == Id{"hi"}); }

  SECTION("ds::DateTime") {
    CHECK(ds::fromJson<ds::DateTime>("\"2017-09-18T10:37:13.777Z\"") == ds::toDateTime("2017-09-18T10:37:13.777Z"));
    CHECK(ds::fromJson<ds::DateTime>("\"2017-09-18T10:37:13Z\"") == ds::toDateTime("2017-09-18T10:37:13Z"));

    SECTION("accepts only correct DateTimes") {
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("false"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("true"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("7"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("7.5"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("{}"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("[]"), "Expecting json string");
      CHECK_THROWS_FIND(ds::fromJson<ds::DateTime>("\"2017-09-18\""), "Expecting DateTime");
    }
  }

  SECTION("std::duration") {
    CHECK(ds::fromJson<std::chrono::milliseconds>("0.0999") == 99ms);
    CHECK(ds::fromJson<std::chrono::seconds>("1.99") == 1s);
  }

  SECTION("enum") {
    CHECK(ds::fromJson<Transport>(R"("car")") == Transport::CAR);
    CHECK(ds::fromJson<Transport>(R"("plane")") == Transport::PLANE);
    CHECK(ds::fromJson<Transport>(R"("unknown")") == static_cast<Transport>(0));

    SECTION("accepts only json strings") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("{}"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }

  SECTION("std::optional") {
    // json library cannot parse "null" :(
    CHECK(ds::fromJson<std::optional<int>>(ds::JsonFragment()) == std::nullopt);
    CHECK(ds::fromJson<std::optional<int>>(R"(7)") == 7);
  }

  SECTION("std::vector") {
    CHECK(ds::fromJson<std::vector<int>>(R"([3,1.4,6])") == std::vector<int>({3, 1, 6}));
    CHECK(ds::fromJson<std::vector<std::string>>(R"(["a","b"])") == std::vector<std::string>({"a", "b"}));

    SECTION("accepts only json array") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("{}"));
    }
  }

  SECTION("std::vector<std::pair<std::string, T>>") {
    typedef std::vector<std::pair<std::string, int>> VectorStringInt;
    typedef std::vector<std::pair<std::string, std::string>> VectorStringString;
    CHECK(ds::fromJson<VectorStringInt>(R"({})") == VectorStringInt{});
    {
      VectorStringInt expected{{"b", 1}, {"a", -1}, {"", 7}};
      CHECK(ds::fromJson<VectorStringInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
    {
      VectorStringString expected{{"b", "B"}};
      CHECK(ds::fromJson<VectorStringString>(R"({"b":"B"})") == expected);
    }

    SECTION("accepts only json object") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }

  SECTION("std::vector<std::pair<Id, T>>") {
    typedef std::vector<std::pair<Id, int>> VectorIdInt;
    CHECK(ds::fromJson<VectorIdInt>(R"({})") == VectorIdInt{});
    {
      VectorIdInt expected{{Id{"b"}, 1}, {Id{"a"}, -1}, {Id{}, 7}};
      CHECK(ds::fromJson<VectorIdInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
  }

  SECTION("std::vector<std::pair<Enum, T>>") {
    typedef std::vector<std::pair<Transport, int>> VectorTransportInt;
    CHECK(ds::fromJson<VectorTransportInt>(R"({})") == VectorTransportInt{});
    {
      VectorTransportInt expected{{Transport::CAR, 1}, {Transport::PLANE, -1}};
      CHECK(ds::fromJson<VectorTransportInt>(R"({"car":1, "plane":-1, "unknown": 7})") == expected);
    }
  }

  SECTION("std::vector<std::pair<int, T>>") {
    typedef std::vector<std::pair<int, int>> VectorIntInt;
    CHECK(ds::fromJson<VectorIntInt>(R"({})") == VectorIntInt{});
    {
      VectorIntInt expected{{10, 1}, {20, -1}};
      CHECK(ds::fromJson<VectorIntInt>(R"({"10":1, "20":-1, "unknown": 7})") == expected);
    }
  }

  SECTION("std::map<std::string, T>>") {
    typedef std::map<std::string, int> MapStringInt;
    CHECK(ds::fromJson<MapStringInt>(R"({})") == MapStringInt{});
    {
      MapStringInt expected{{Id{"b"}, 1}, {Id{"a"}, -1}, {Id{}, 7}};
      CHECK(ds::fromJson<MapStringInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
  }

  SECTION("std::map<Id, T>>") {
    typedef std::map<Id, int> MapIdInt;
    CHECK(ds::fromJson<MapIdInt>(R"({})") == MapIdInt{});
    {
      MapIdInt expected{{Id{"b"}, 1}, {Id{"a"}, -1}, {Id{}, 7}};
      CHECK(ds::fromJson<MapIdInt>(R"({"b":1, "a":-1,"":7})") == expected);
    }
  }

  SECTION("std::map<Enum, T>>") {
    typedef std::map<Transport, int> MapTransportInt;
    CHECK(ds::fromJson<MapTransportInt>(R"({})") == MapTransportInt{});
    {
      MapTransportInt expected{{Transport::CAR, 1}, {Transport::PLANE, -1}};
      CHECK(ds::fromJson<MapTransportInt>(R"({"car":1, "plane":-1, "unknown": 7})") == expected);
    }
  }

  SECTION("std::map<int, T>>") {
    typedef std::map<int, int> MapIntInt;
    CHECK(ds::fromJson<MapIntInt>(R"({})") == MapIntInt{});
    {
      MapIntInt expected{{10, 1}, {20, -1}};
      CHECK(ds::fromJson<MapIntInt>(R"({"10":1, "20":-1, "unknown": 7})") == expected);
    }
  }

  SECTION("std::set<Enum>") {
    typedef std::set<Transport> SetEnum;
    CHECK(ds::fromJson<SetEnum>(R"({})") == SetEnum{});
    {
      SetEnum expected{Transport::CAR, Transport::PLANE};
      CHECK(ds::fromJson<SetEnum>(R"({"car":true, "plane":true, "":true})") == expected);
    }
  }

  SECTION("JsonFragment") {
    CHECK(ds::fromJson<int>(ds::fromJson<ds::JsonFragment>("7")) == 7);
    CHECK(ds::fromJson<bool>(ds::fromJson<ds::JsonFragment>("false")) == false);
    CHECK(ds::fromJson<std::string>(ds::fromJson<ds::JsonFragment>("\"hi\"")) == "hi");
  }

  SECTION("struct") {
    CHECK(ds::fromJson<X>(R"({})") == X{});
    CHECK(ds::fromJson<X>(R"({"i":4})") == X{.i = 4});
    CHECK(ds::fromJson<X>(R"({"s":"a"})") == X{.i = {}, .s = "a"});
    CHECK(ds::fromJson<X>(R"({"transport":"car"})") == X{.i = {}, .s = {}, .transport = Transport::CAR});
    CHECK(ds::fromJson<X>(R"({"transport":"plane"})") == X{.i = {}, .s = {}, .transport = Transport::PLANE});
    CHECK(ds::fromJson<X>(R"({"s":"a","i":7})") == X{.i = 7, .s = "a"});
    CHECK(ds::fromJson<X>(ds::fromJson<X>(R"({"json":{"s":"a","i":7}})").json) == X{.i = 7, .s = "a"});
    CHECK(ds::fromJson<X>(R"({"s":null})") == X{});
    SECTION("optional<T> can be used to distinguish missing field / null field and value default") {
      CHECK(ds::fromJson<X>(R"({})") == X{});
      CHECK(ds::fromJson<X>(R"({"oi":null})") == X{});
      CHECK(ds::fromJson<X>(R"({"oi":0})") == X{.i = {}, .s = {}, .transport = {}, .json = {}, 0});
      CHECK(ds::fromJson<X>(R"({"oi":7})") == X{.i = {}, .s = {}, .transport = {}, .json = {}, 7});
    }

    SECTION("accepts only json object") {
      CHECK_THROWS(ds::fromJson<Transport>("false"));
      CHECK_THROWS(ds::fromJson<Transport>("true"));
      CHECK_THROWS(ds::fromJson<Transport>("7"));
      CHECK_THROWS(ds::fromJson<Transport>("7.5"));
      CHECK_THROWS(ds::fromJson<Transport>("a"));
      CHECK_THROWS(ds::fromJson<Transport>("[]"));
    }
  }

  SECTION("struct as id array") {
    CHECK(ds::fromJson<StructAsIdArray>(R"([])") == StructAsIdArray{});
    CHECK(ds::fromJson<StructAsIdArray>(R"([{"id":"x","i":4}])") == StructAsIdArray{.x = {.i = 4}});
  }

  SECTION("Variant member") {
    CHECK(ds::fromJson<AB>(R"({})") == AB{});
    CHECK(ds::fromJson<AB>(R"({"type":"a","ab":{"a":7}})") == AB{.ab = A{.a = 7}});
    CHECK(ds::fromJson<AB>(R"({"type":"b","ab":{"b":6}})") == AB{.ab = B{.b = 6}});
  }
}

TEMPLATE_TEST_CASE("fromJson JsonFragment{} returns default instance", TAGS, bool, int, double, std::string,
                   std::vector<int>, X) {
  CHECK(ds::fromJson<TestType>(ds::JsonFragment{}) == TestType{});
}

TEST_CASE("toJson", TAGS) {
  SECTION("bool") {
    CHECK(ds::toJson(false) == "false");
    CHECK(ds::toJson(true) == "true");
  }

  SECTION("number") {
    CHECK(ds::toJson(7) == "7");
    CHECK(std::atof(ds::toJson(5.5).c_str()) == 5.5);
    CHECK(ds::toJson((unsigned char)7) == "7");
    CHECK(ds::toJson((signed char)7) == "7");
    CHECK(ds::toJson(7LL) == "\"7\"");
    CHECK(ds::toJson(7ULL) == "\"7\"");
    CHECK(ds::toJson(9007199254740991ULL) == "\"9007199254740991\"");
    CHECK(ds::toJson(-9223372036854775807LL) == "\"-9223372036854775807\"");
  }

  SECTION("string") {
    CHECK(ds::toJson("") == "\"\"");
    CHECK(ds::toJson("a") == "\"a\"");
    CHECK(ds::toJson("hello") == "\"hello\"");
    CHECK(ds::toJson("one\ttwo") == "\"one\\ttwo\"");
    CHECK(ds::toJson(std::string{}) == "\"\"");
    CHECK(ds::toJson(std::string{"hello"}) == "\"hello\"");
  }

  SECTION("Id") { CHECK(ds::toJson(Id{"hi"}) == "\"hi\""); }

  SECTION("ds::DateTime") {
    CHECK(ds::toJson(ds::toDateTime("2017-09-18T10:37:13.777Z")) == "\"2017-09-18T10:37:13.777Z\"");
  }

  SECTION("std::duration") {
    CHECK(ds::toJson(1500ms) == "1.5");
    CHECK(ds::toJson(3.5s) == "3.5");
  }

  SECTION("enum") {
    CHECK(ds::toJson(Transport::CAR) == "\"car\"");
    CHECK(ds::toJson(Transport::PLANE) == "\"plane\"");
  }

  SECTION("std::optional") {
    CHECK(ds::toJson(std::optional<int>()) == R"(null)");
    CHECK(ds::toJson(std::optional<int>(7)) == R"(7)");
    CHECK(ds::toJson(std::optional<std::string>("a")) == R"("a")");
  }

  SECTION("std::vector") {
    CHECK(ds::toJson(std::vector<int>({})) == R"([])");
    CHECK(ds::toJson(std::vector<int>({3, 1, 6})) == R"([3,1,6])");
    CHECK(ds::toJson(std::vector<std::string>{"a", "b"}) == R"(["a","b"])");
  }

  SECTION("std::vector<std::pair<std::string, T>>") {
    CHECK(ds::toJson(std::vector<std::pair<std::string, int>>({})) == R"({})");
    CHECK(ds::toJson(std::vector<std::pair<std::string, int>>({{"b", 3}, {"a", 7}})) == R"({"b":3,"a":7})");
    CHECK(ds::toJson(std::vector<std::pair<std::string, std::string>>({{"b", "B"}, {"", ""}})) == R"({"b":"B","":""})");
  }

  SECTION("std::vector<std::pair<Id, T>>") {
    CHECK(ds::toJson(std::vector<std::pair<Id, int>>({})) == R"({})");
    CHECK(ds::toJson(std::vector<std::pair<Id, int>>({{Id{"b"}, 3}, {Id{"a"}, 7}})) == R"({"b":3,"a":7})");
    CHECK(ds::toJson(std::vector<std::pair<Id, std::string>>({{Id{"b"}, "B"}, {Id{""}, ""}})) == R"({"b":"B","":""})");
  }

  SECTION("std::vector<std::pair<Enum, T>>") {
    CHECK(ds::toJson(std::vector<std::pair<Transport, int>>({})) == R"({})");
    CHECK(ds::toJson(std::vector<std::pair<Transport, int>>({{Transport::CAR, 3}, {Transport::PLANE, 7}})) == R"({"car":3,"plane":7})");
    CHECK(ds::toJson(std::vector<std::pair<Transport, std::string>>({{Transport::CAR, "B"}, {static_cast<Transport>(3), ""}})) == R"({"car":"B"})");
  }

  SECTION("std::vector<std::pair<int, T>>") {
    CHECK(ds::toJson(std::vector<std::pair<int, int>>({})) == R"({})");
    CHECK(ds::toJson(std::vector<std::pair<int, int>>({{10, 3}, {20, 7}})) == R"({"10":3,"20":7})");
    CHECK(ds::toJson(std::vector<std::pair<int, std::string>>({{10, "B"}, {20, ""}})) == R"({"10":"B","20":""})");
  }

  SECTION("std::map<<std::string, T>") {
    CHECK(ds::toJson(std::map<std::string, int>({})) == R"({})");
    CHECK(ds::toJson(std::map<std::string, int>({{"b", 3}, {"a", 7}})) == R"({"a":7,"b":3})");
    CHECK(ds::toJson(std::map<std::string, std::string>({{"b", "B"}, {"", ""}})) == R"({"":"","b":"B"})");
  }

  SECTION("std::map<Id, T>") {
    CHECK(ds::toJson(std::map<Id, int>({})) == R"({})");
    CHECK(ds::toJson(std::map<Id, int>({{Id{"b"}, 3}, {Id{"a"}, 7}})) == R"({"a":7,"b":3})");
    CHECK(ds::toJson(std::map<Id, std::string>({{Id{"b"}, "B"}, {Id{""}, ""}})) == R"({"":"","b":"B"})");
  }

  SECTION("std::map<Enum, T>") {
    CHECK(ds::toJson(std::map<Transport, int>({})) == R"({})");
    CHECK(ds::toJson(std::map<Transport, int>({{Transport::CAR, 3}, {Transport::PLANE, 7}})) == R"({"car":3,"plane":7})");
    CHECK(ds::toJson(std::map<Transport, std::string>({{Transport::CAR, "B"}, {static_cast<Transport>(3), ""}})) == R"({"car":"B"})");
  }

  SECTION("std::map<int, T>") {
    CHECK(ds::toJson(std::map<int, int>({})) == R"({})");
    CHECK(ds::toJson(std::map<int, int>({{10, 3}, {20, 7}})) == R"({"10":3,"20":7})");
    CHECK(ds::toJson(std::map<int, std::string>({{10, "B"}, {20, ""}})) == R"({"10":"B","20":""})");
  }

  SECTION("std::set<Enum>") {
    CHECK(ds::toJson(std::set<Transport>({})) == R"({})");
    CHECK(ds::toJson(std::set<Transport>({Transport::CAR, Transport::PLANE, static_cast<Transport>(-1)})) ==
          R"({"car":true,"plane":true})");
  }

  SECTION("jsonObject") {
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("7")) == "7");
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("false")) == "false");
    CHECK(ds::toJson(ds::fromJson<ds::JsonFragment>("\"hi\"")) == "\"hi\"");
  }

  SECTION("struct") {
    CHECK(ds::toJson(X{}) == R"({})");
    CHECK(ds::toJson(X{}) == R"({})");
    CHECK(ds::toJson(X{.i = 4, .s = "", .transport = {}}) == R"({"i":4})");
    CHECK(ds::toJson(X{.i = 0, .s = "a", .transport = {}}) == R"({"s":"a"})");
    CHECK(ds::toJson(X{.i = 0, .s = "", .transport = Transport::CAR}) == R"({})");
    CHECK(ds::toJson(X{.i = 0, .s = "", .transport = Transport::PLANE}) == R"({"transport":"plane"})");
    CHECK(ds::toJson(X{.i = 7, .s = "a", .transport = Transport{}}) == R"({"i":7,"s":"a"})");
    CHECK(ds::toJson(X{.i = 7,
                       .s = "a",
                       .transport = Transport{},
                       .json = ds::toJsonFragment(X{.i = {}, .s = "a", .transport = Transport::PLANE})}) ==
          R"({"i":7,"s":"a","json":{"s":"a","transport":"plane"}})");

    SECTION("optional<T> can be used to distinguish missing field / null field and value default") {
      CHECK(ds::toJson(X{}) == R"({})");
      CHECK(ds::toJson(X{.i = {}, .s = {}, .transport = {}, .json = {}, .oi = 0}) == R"({"oi":0})");
      CHECK(ds::toJson(X{.i = {}, .s = {}, .transport = {}, .json = {}, .oi = 7}) == R"({"oi":7})");
    }
  }

  SECTION("struct as id array") {
    CHECK(ds::toJson(StructAsIdArray{}) == R"([])");
    CHECK(ds::toJson(StructAsIdArray{.x = X{.i = 4}}) == R"([{"i":4,"id":"x"}])");
  }

  SECTION("Variant member") {
    CHECK(ds::toJson(AB{}) == R"({})");
    CHECK(ds::toJson(AB{.ab = A{.a = 7}}) == R"({"type":"a","ab":{"a":7}})");
    CHECK(ds::toJson(AB{.ab = B{.b = 6}}) == R"({"type":"b","ab":{"b":6}})");
  }
}

TEST_CASE("toJsonFragment", TAGS) {
  // toJson test covers all details.
  CHECK(ds::toJson(ds::toJsonFragment(X{.i = 0, .s = "a", .transport = Transport::PLANE})) ==
        R"({"s":"a","transport":"plane"})");
}

TEST_CASE("JsonFragment", TAGS) {
  SECTION("can be stringified", TAGS) {
    CHECK(ds::str(ds::toJsonFragment(X{.i = 0, .s = "a", .transport = Transport::PLANE})) ==
          R"({"s":"a","transport":"plane"})");
  }
  SECTION("can be compared by value", TAGS) {
    CHECK(ds::toJsonFragment(1) == ds::toJsonFragment(1));
    CHECK(ds::toJsonFragment("a") == ds::toJsonFragment("a"));

    CHECK(ds::toJsonFragment(1) != ds::toJsonFragment(2));
    CHECK(ds::toJsonFragment(1) != ds::toJsonFragment("1"));
    CHECK(ds::toJsonFragment("a") != ds::toJsonFragment("b"));
  }
}
} // namespace
