/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <gsl.h>
#include <json-c/json.h>
#include <chrono>
#include <map>
#include <set>
#include <variant>
#include "common.h"
#include "date-time.h"
#include "enum.h"
#include "log.h"
#include "reflect.h"

///
/// Json parsing and serialization for types with reflection.
///
/// Use:
///
///   ds::toJson(std::vector<int>({4,3})) -> "([4,3])"
///
///   ds::fromJson<std::vector<int>>("[4,3]") -> std::vector<int>({4,3})
///
/// Supported types:
///
/// bool
///   * bool: true, false
///
/// [u|]int[8|16|32]_t
///   * number: 1,-10,0,
///   * JSON value will be a decimal number. Either numbers or strings are accepted.
///
/// [u|]int64_t
///   * string: "1", "10""
///   * JSON value will be a decimal string. Either numbers or strings are accepted.
///   * Compatible with protobuf, capnproto.
///   * It may be worthwhile not use [u]int64_t in json apis. Consider string?
///
/// std::string
///   * string: "Hello world!"
///
/// ds::DateTime
///   * string: "2017-09-18T10:37:13.777Z"
///
/// std::chrono::duration
///   * number: 0.0344
///   * in seconds
///
/// enum with reflection (ds/enum.h)
///   * camelCase string or number: "fooBar", 4
///   * Unknown enum option is serialized as number.
///   * Only known string options are accepted.
///
/// vector<T>
///   * array: [v, ...]
///
/// vector<std::pair<std::string, T>>, vector<std::pair<Key, T>>
///   * object: {"key1" : value1, ...}
///   * Key can be any type convertible to and from std::string. E.g. `BOOST_STRONG_TYPEDEF(std::string, Id)`
///
/// std::map<std::string, T>, std::map<Key, T>
///   * object: {"key1" : value1, ...}
///   * Key can be any type convertible to and from std::string. E.g. `BOOST_STRONG_TYPEDEF(std::string, Id)`
///
/// std::set<Enum>
///   * object: {"key1" : true, ...}
///   * Unknown enum options are skipped. Keys with `false` value are skipped.
///
/// struct with reflection (ds/reflect.h)
///   * object: {"name": "John", "height": 1.82}
///   * Field names use "lowerCamelCase".
///   * Missing fields preserve default value
///   * Fields with default value are not serialized.
///   * Fields of std::optional type need DSJsonVariant type trait
///
/// struct of structs with reflection with ::DSJsonArrayId::value == true
///   * array: [{"id"="fieldName1", ...(fields of field1)...}, {"id"="fieldName2", ...(fields of field2)...}]
///   * Fields with default value are not serialized.
///   * motivation: ds-model likes to use json arrays indexed by id where struct is natural fit
///
/// JsonFragment
///   * any json fragment injected / accepted
///   * for places where type is not known at parsing / serialization time
///

/// specialize to std::true_type to serialize given struct of structs as array of structs indexed by id field.
template <class T>
struct DSJsonArrayId : public std::false_type {};

namespace ds {
enum JsonVariantMode { none, alternativeInSiblingMember };
} // namespace ds

/// Type trait for serialization of variant members.
/// See DSJsonVariant<&AB::ab> in json-test.cpp for example. Commented code bit-rots fast.
template <auto member>
struct DSJsonVariant;

namespace ds {

struct JsonFragmentDeleter {
  void operator()(json_object* x) { json_object_put(x); }
};

// Json object with ownership.
typedef std::unique_ptr<json_object, JsonFragmentDeleter> JsonFragment;

std::ostream& operator<<(std::ostream&, const JsonFragment&);
bool operator==(const ds::JsonFragment& a, const ds::JsonFragment& b);
bool operator!=(const ds::JsonFragment& a, const ds::JsonFragment& b);

template <typename T, class Enable = void>
struct FromJson;

template <typename T>
T fromJson(json_object* object) {
  if (!object) {
    return {}; // Interpret null / missing json fragment as empty value
  }
  return FromJson<T>::call(object);
}

template <typename T>
T fromJson(const JsonFragment& json) {
  return fromJson<T>(json.get());
}

template <typename T>
T fromJson(gsl::czstring text);

template <typename T>
T fromJson(const std::string& text) {
  return fromJson<T>(text.c_str());
}

template <typename T, class Enable = void>
struct ToJsonFragment;

template <typename T>
JsonFragment toJsonFragment(const T& x) {
  return ToJsonFragment<typename std::decay<T>::type>::call(x);
}

template <typename T>
std::string toJson(const T& x) {
  return json_object_to_json_string_ext(toJsonFragment(x).get(), 0);
}

// implementation details - from json

template <typename T>
T fromJson(gsl::czstring text) {
  json_tokener* tok = json_tokener_new();
  DS_REQUIRE(tok);
  DS_DEFER(json_tokener_free(tok));

  // we don't use gsl::string_span on input,
  // because json-c is not able to parse bool and numbers
  // with explicit len instead of -1
  json_object* object = json_tokener_parse_ex(tok, text, -1);
  DS_REQUIRE(object, "Failed to parse json.", ds::hideArg(json_tokener_error_desc(json_tokener_get_error(tok))));
  auto json = JsonFragment{object};

  return fromJson<T>(json);
}

template <>
struct FromJson<bool> {
  static bool call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_boolean, "Expecting json boolen.", type);
    return json_object_get_boolean(json);
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_integral<T>::value>::type> {
  static T call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string || type == json_type_int || type == json_type_double,
               "Expecting json number / string.", type);
    auto x = json_object_get_int64(json);
    return ds::narrow<T>(x);
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_floating_point<T>::value>::type> {
  static T call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string || type == json_type_int || type == json_type_double,
               "Expecting json number / string.", type);
    return json_object_get_double(json);
  }
};

template <>
struct FromJson<std::string> {
  static std::string call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string, "Expecting json string", type);
    return json_object_get_string(json);
  }
};

// support for BOOST_STRONG_TYPEDEF(std::string, T)
template <typename T>
struct FromJson<T, typename std::enable_if_t<std::is_convertible_v<T, std::string>>> {
  static T call(json_object* json) { return T{FromJson<std::string>::call(json)}; }
};

template <>
struct FromJson<ds::DateTime> {
  static ds::DateTime call(json_object* json) {
    const auto& type = json_object_get_type(json);
    DS_REQUIRE(type == json_type_string, "Expecting json string", type);
    return ds::toDateTime(json_object_get_string(json));
  }
};

template <class Rep, class Period>
struct FromJson<std::chrono::duration<Rep, Period>> {
  static std::chrono::duration<Rep, Period> call(json_object* json) {
    return std::chrono::duration_cast<std::chrono::duration<Rep, Period>>(
        std::chrono::duration<double>{FromJson<double>::call(json)});
  }
};

template <typename T>
struct FromJson<T, typename std::enable_if<std::is_enum<T>::value>::type> {
  static T call(json_object* json) {
    const auto& jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_string, "Expecting json string.", jsonType);
    return ds::toEnum<T>(json_object_get_string(json));
  }
};

template <typename T>
struct FromJson<std::optional<T>> {
  static std::optional<T> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    if (jsonType == json_type_null) {
      return {};
    }
    return FromJson<T>::call(json);
  }
};

template <>
struct FromJson<std::monostate> {
  static std::monostate call(json_object* json) { return {}; }
};

template <typename Item>
struct FromJson<std::vector<Item>> {
  static std::vector<Item> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_array, "Expecting json array.", jsonType);
    std::vector<Item> result{};
    int size = json_object_array_length(json);
    result.reserve(size);
    for (int i = 0; i < size; i++) {
      DS_CONTEXT("/", ds::hideArg(i));
      result.push_back(fromJson<Item>(json_object_array_get_idx(json, i)));
    }
    return result;
  }
};

template <typename T, typename Enable = void>
struct JsonKey {
  static std::optional<T> fromJson(const char* x) { return T{x}; }
  static std::optional<std::string> toJson(const T& x) { return static_cast<std::string>(x); }
};

template <typename T>
struct JsonKey<T, typename std::enable_if<std::is_enum<T>::value>::type> {
  static std::optional<T> fromJson(const char* x) { return ds::tryToEnum<T>(x); }
  static std::optional<std::string> toJson(const T& x) { return ds::tryFromEnum<std::string>(x); }
};

template <typename T>
struct JsonKey<T, typename std::enable_if<std::is_integral<T>::value>::type> {
  static std::optional<T> fromJson(const char* x) { return ds::tryToInt(x); }
  static std::optional<std::string> toJson(const T& x) { return ds::intToString(x); }
};

template <typename First, typename Second>
struct FromJson<std::vector<std::pair<First, Second>>> {
  static std::vector<std::pair<First, Second>> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    std::vector<std::pair<First, Second>> result{};
    json_object_object_foreach(json, key, value) {
      if (const auto parsedKey = JsonKey<First>::fromJson(key)) {
        DS_CONTEXT("/", ds::hideArg(key));
        result.push_back({*parsedKey, fromJson<Second>(value)});
      } else {
        // Unknown enum keys are skipped.
      }
    }
    return result;
  }
};

template <typename First, typename Second>
struct FromJson<std::map<First, Second>> {
  static std::map<First, Second> call(json_object* json) {
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    std::map<First, Second> result{};
    json_object_object_foreach(json, key, value) {
      if (const auto parsedKey = JsonKey<First>::fromJson(key)) {
        DS_CONTEXT("/", ds::hideArg(key));
        result[*parsedKey] = fromJson<Second>(value);
      } else {
        // Unknown enum keys are skipped.
      }
    }
    return result;
  }
};

template <typename T>
struct FromJson<std::set<T>> {
  static std::set<T> call(json_object* json) {
    static_assert(std::is_enum<T>::value, "Enum type required as key");
    auto jsonType = json_object_get_type(json);
    DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
    std::set<T> result{};
    json_object_object_foreach(json, key, value) {
      DS_CONTEXT("/", ds::hideArg(key));
      auto valueType = json_object_get_type(value);
      DS_REQUIRE(valueType == json_type_boolean, "Expecting json bool.", valueType);
      if (json_object_get_boolean(value)) {
        if (auto enumKey = ds::tryToEnum<T>(key)) {
          result.insert(*enumKey);
        } else {
          // Unknown enum keys are skipped.
        }
      }
    }
    return result;
  }
};

template <>
struct ToJsonFragment<std::monostate> {
  static JsonFragment call(const std::monostate& x) { return JsonFragment{}; }
};

template <>
struct FromJson<JsonFragment> {
  static JsonFragment call(json_object* json) { return JsonFragment{json_object_get(json)}; }
};

template <auto member, class = void>
struct JsonMember {
  template <class Field>
  static void fromJson(Field& field, json_object* json, const char* name) {
    json_object* jsonField{};
    if (json_object_object_get_ex(json, name, &jsonField)) {
      if (json_object_get_type(jsonField) != json_type_null) {
        field = ::ds::fromJson<std::decay_t<decltype(field)>>(jsonField);
      }
    }
  }
  template <class Field>
  static void toJson(const JsonFragment& json, const Field& field, const char* name) {
    json_object_object_add(json.get(), name, toJsonFragment(field).release());
  }
};

/// visit variant by alternative name
template <auto member, class Variant, class Visitor, int i = std::variant_size_v<Variant>>
auto visitJsonVariant(Visitor visitor, std::string_view name) {
  using Alternative = std::variant_alternative_t<size_t(i - 1), Variant>;
  if (::DSJsonVariant<member>::nameVisitor(Alternative{}) == name) {
    return visitor(Alternative{});
  }
  if constexpr (i > 1) {
    return visitJsonVariant<member, Variant, Visitor, i - 1>(visitor, name);
  }
  return visitor(Alternative{});
}

template <auto member>
struct JsonMember<member, std::void_t<decltype(::DSJsonVariant<member>::mode)>> {
  using Trait = ::DSJsonVariant<member>;
  template <class Variant>
  static void fromJson(Variant& variant, json_object* json, const char* name) {
    if constexpr (Trait::mode == ds::JsonVariantMode::alternativeInSiblingMember) {
      json_object* jsonAlternative{};
      if (json_object_object_get_ex(json, Trait::alternative, &jsonAlternative) && jsonAlternative) {
        const auto alternative = ::ds::fromJson<std::string>(jsonAlternative);
        json_object* jsonVariant{};
        if (json_object_object_get_ex(json, name, &jsonVariant) && jsonVariant) {
          visitJsonVariant<member, Variant>(
              ds::Visitor{
                  [&](auto&& x) { variant = ::ds::fromJson<std::decay_t<decltype(x)>>(jsonVariant); },
              },
              alternative);
        }
      }
    } else {
      static_assert(ds::DependentFalse<Variant>::value, "Unrecognized json variant mode");
    }
  }

  template <class Variant>
  static void toJson(const JsonFragment& json, const Variant& variant, const char* name) {
    if constexpr (Trait::mode == ds::JsonVariantMode::alternativeInSiblingMember) {
      json_object_object_add(json.get(), Trait::alternative,
                             json_object_new_string(std::visit(Trait::nameVisitor, variant)));
      json_object_object_add(json.get(), name,
                             std::visit([](auto&& x) { return toJsonFragment(x); }, variant).release());
    } else {
      static_assert(ds::DependentFalse<Variant>::value, "Unrecognized json variant mode");
    }
  }
};

template <typename T>
struct FromJson<T, std::void_t<ds::reflect_t<T>>> {
  struct Visitor : boost::noncopyable {
    T& x;
    json_object* json;
    Visitor(T& x, json_object* json) : x(x), json(json) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_CONTEXT("/", ds::hideArg(name));
      JsonMember<m>::fromJson(x.*m, json, name);
    }
  };

  struct AsIdArrayVisitor : boost::noncopyable {
    T& x;
    json_object* json;
    AsIdArrayVisitor(T& x, json_object* json) : x(x), json(json) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_CONTEXT("/", ds::hideArg(name));
      auto& field = x.*m;
      using Field = std::decay_t<decltype(field)>;
      static_assert(std::is_same_v<Field, ds::reflect_t<Field>>, "Field must be reflect struct.");
      int size = json_object_array_length(json);
      for (int i = 0; i < size; i++) {
        auto jsonItem = json_object_array_get_idx(json, i);
        auto jsonItemType = json_object_get_type(jsonItem);
        DS_REQUIRE(jsonItemType == json_type_object, "Expecting json object in array.", jsonItemType);
        json_object* jsonItemId{};
        if (json_object_object_get_ex(jsonItem, "id", &jsonItemId)) {
          auto jsonItemIdType = json_object_get_type(jsonItemId);
          DS_REQUIRE(jsonItemIdType == json_type_string, "Expecting json string as id in array.", jsonItemIdType);
          if (std::strcmp(json_object_get_string(jsonItemId), name) == 0) {
            field = fromJson<Field>(jsonItem);
            return;
          }
        }
      }
    }
  };

  static T call(json_object* json) {
    T result{};
    auto jsonType = json_object_get_type(json);
    if constexpr (::DSJsonArrayId<T>::value) {
      DS_REQUIRE(jsonType == json_type_array, "Expecting json array.", jsonType);
      ds::reflector<T>::visit(AsIdArrayVisitor(result, json));
    } else {
      DS_REQUIRE(jsonType == json_type_object, "Expecting json object.", jsonType);
      ds::reflector<T>::visit(Visitor(result, json));
    }
    return result;
  }
};

// implementation details - toJson

typedef std::unique_ptr<json_object, JsonFragmentDeleter> JsonFragment;

template <>
struct ToJsonFragment<bool> {
  static JsonFragment call(bool x) { return JsonFragment{json_object_new_boolean(x)}; }
};

#define DS_INTEGER_CASE(TYPE)                                                           \
  template <>                                                                           \
  struct ToJsonFragment<TYPE> {                                                         \
    static JsonFragment call(TYPE x) { return JsonFragment{json_object_new_int64(x)}; } \
  };
DS_INTEGER_CASE(unsigned char);
DS_INTEGER_CASE(signed char);
DS_INTEGER_CASE(unsigned short);
DS_INTEGER_CASE(short);
DS_INTEGER_CASE(unsigned int);
DS_INTEGER_CASE(int);
#undef DS_INTEGER_CASE

template <>
struct ToJsonFragment<unsigned long long> {
  static JsonFragment call(unsigned long long x) {
    return JsonFragment{json_object_new_string(std::to_string(x).c_str())};
  }
};

template <>
struct ToJsonFragment<long long> {
  static JsonFragment call(long long x) { return JsonFragment{json_object_new_string(std::to_string(x).c_str())}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_integral<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_int64(x)}; }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_floating_point<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_double(x)}; }
};
inline JsonFragment toJsonFragment(gsl::cstring_span x) {
  return toJsonFragment<gsl::cstring_span>(x);
}

template <>
struct ToJsonFragment<const char*> {
  static JsonFragment call(const char* x) { return JsonFragment{json_object_new_string(x)}; }
};

template <>
struct ToJsonFragment<char*> {
  static JsonFragment call(const char* x) { return JsonFragment{json_object_new_string(x)}; }
};

template <>
struct ToJsonFragment<gsl::cstring_span> {
  static JsonFragment call(gsl::cstring_span x) { return JsonFragment{json_object_new_string_len(x.data(), x.size())}; }
};

template <>
struct ToJsonFragment<std::string> {
  static JsonFragment call(const std::string& x) {
    return JsonFragment{json_object_new_string_len(x.data(), x.size())};
  }
};

// support for BOOST_STRONG_TYPEDEF(std::string, T)
template <typename T>
struct ToJsonFragment<T, typename std::enable_if_t<std::is_convertible_v<T, std::string>>> {
  static JsonFragment call(const T& x) { return ToJsonFragment<std::string>::call(static_cast<std::string>(x)); }
};

template <>
struct ToJsonFragment<ds::DateTime> {
  static JsonFragment call(ds::DateTime x) { return JsonFragment{json_object_new_string(ds::toString(x).c_str())}; }
};

template <class Rep, class Period>
struct ToJsonFragment<std::chrono::duration<Rep, Period>> {
  static JsonFragment call(std::chrono::duration<Rep, Period> x) {
    return ToJsonFragment<double>::call(std::chrono::duration_cast<std::chrono::duration<double>>(x).count());
  }
};

template <typename T>
struct ToJsonFragment<T, typename std::enable_if<std::is_enum<T>::value>::type> {
  static JsonFragment call(T x) { return JsonFragment{json_object_new_string(ds::fromEnum<std::string>(x).c_str())}; }
};

template <typename T>
struct ToJsonFragment<std::optional<T>> {
  static JsonFragment call(const std::optional<T>& x) {
    if (x) {
      return toJsonFragment(*x);
    } else {
      return JsonFragment{};
    }
  }
};

template <typename Item>
struct ToJsonFragment<std::vector<Item>> {
  static JsonFragment call(const std::vector<Item>& items) {
    auto result = JsonFragment{json_object_new_array()};

    int i = 0;
    DS_CONTEXT("/", ds::hideArg(i));
    for (const auto& item : items) {
      json_object_array_add(result.get(), toJsonFragment(item).release());
      i++;
    }
    return result;
  }
};

template <typename Container, typename First, typename Second>
struct ToJsonFragmentAssociateContainer {
  static JsonFragment call(const Container& items) {
    auto result = JsonFragment{json_object_new_object()};

    for (const auto& item : items) {
      if (const auto itemFirst = JsonKey<First>::toJson(item.first)) {
        DS_CONTEXT("/", ds::hideArg(itemFirst));
        json_object_object_add(result.get(), itemFirst->c_str(), toJsonFragment(item.second).release());
      } else {
        // skip keys that can not be converted to std::string
      }
    }
    return result;
  }
};

template <typename First, typename Second>
struct ToJsonFragment<std::vector<std::pair<First, Second>>>
    : public ToJsonFragmentAssociateContainer<std::vector<std::pair<First, Second>>, First, Second> {};

template <typename First, typename Second>
struct ToJsonFragment<std::map<First, Second>>
    : public ToJsonFragmentAssociateContainer<std::map<First, Second>, First, Second> {};

template <typename T>
struct ToJsonFragment<std::set<T>> {
  static_assert(std::is_enum<T>::value, "Enum type required as key");
  static JsonFragment call(const std::set<T>& items) {
    auto result = JsonFragment{json_object_new_object()};

    for (const auto& item : items) {
      DS_CONTEXT("/", ds::hideArg(item));
      if (auto key = ds::tryFromEnum<std::string>(item)) {
        json_object_object_add(result.get(), key->c_str(), json_object_new_boolean(true));
      } else {
        // Unknown enum keys are skipped.
      }
    }
    return result;
  }
};

template <>
struct ToJsonFragment<JsonFragment> {
  static JsonFragment call(const JsonFragment& x) { return JsonFragment{json_object_get(x.get())}; }
};

template <typename T>
struct ToJsonFragment<T, std::void_t<ds::reflect_t<T>>> {
  struct Visitor : boost::noncopyable {
    const JsonFragment& json;
    const T& x;
    Visitor(const JsonFragment& json, const T& x) : json(json), x(x) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_CONTEXT("/", ds::hideArg(name));
      if (isEmptyInstance(x.*m)) {
        return;
      }
      JsonMember<m>::toJson(json, x.*m, name);
    }
  };

  struct AsIdArrayVisitor : boost::noncopyable {
    const JsonFragment& json;
    const T& x;
    AsIdArrayVisitor(const JsonFragment& json, const T& x) : json(json), x(x) {}

    template <typename MemberPtr, MemberPtr m>
    void operator()(const char* name) const {
      DS_CONTEXT("/", ds::hideArg(name));
      const auto& field = x.*m;
      using Field = std::decay_t<decltype(field)>;
      static_assert(std::is_same_v<Field, ds::reflect_t<Field>>, "Field must be reflect struct.");
      if (isEmptyInstance(field)) {
        return;
      }
      auto jsonItem = toJsonFragment(field);
      json_object_object_add(jsonItem.get(), "id", json_object_new_string(name));
      json_object_array_add(json.get(), jsonItem.release());
    }
  };

  static JsonFragment call(const T& x) {
    if constexpr (::DSJsonArrayId<T>::value) {
      JsonFragment result{json_object_new_array()};
      ds::reflector<T>::visit(AsIdArrayVisitor(result, x));
      return result;
    } else {
      JsonFragment result{json_object_new_object()};
      ds::reflector<T>::visit(Visitor(result, x));
      return result;
    }
  }
};

} // namespace ds

#define DS_JSON_DECLARE_STRUCT(NAME_SPACES, TYPE, MEMBERS)                                                    \
  DS_REFLECT_DECLARE_STRUCT(NAME_SPACES, TYPE, MEMBERS)                                                       \
  namespace ds {                                                                                              \
  extern template DS_REFLECT_FULL_TYPE(NAME_SPACES,                                                           \
                                       TYPE) fromJson<DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE)>(json_object*); \
  extern template JsonFragment toJsonFragment(const DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE) &);               \
  }

#define DS_JSON_DEFINE_STRUCT(NAME_SPACES, TYPE)                                                                    \
  DS_REFLECT_DEFINE_STRUCT(NAME_SPACES, TYPE)                                                                       \
  namespace ds {                                                                                                    \
  template DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE) fromJson<DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE)>(json_object*); \
  template JsonFragment toJsonFragment(const DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE) &);                            \
  }
