/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "log.h"
#include <ds/catch/catch.h>
#include <thread>

static const char* TAGS = "[dsLog][ds]";

TEST_CASE("dsTrimLogArgName", TAGS) {
  int x = 5;
  SECTION("discards string constants") { CHECK(ds::str(ds::trimArgName("\"str\":")) == ""); }
  SECTION("passes variable names") { CHECK(ds::str(ds::trimArgName("variable:")) == "variable:"); }
  SECTION("discard variables in ds::hideArg") { CHECK(ds::str("" DS_ARG(ds::hideArg(x))) == "5 "); }
  SECTION("discard variables in ::ds::hideArg") { CHECK(ds::str("" DS_ARG(::ds::hideArg(x))) == "5 "); }
}

TEST_CASE("DS_CONTEXT", TAGS) {
  SECTION("is empty at start") { CHECK(ds::str(ds::logContext()) == ""); }

  SECTION("is stacked and scope limited") {
    DS_CONTEXT("a");
    CHECK(ds::str(ds::logContext()) == "a ");
    {
      int x = 7;
      DS_CONTEXT(x);
      CHECK(ds::str(ds::logContext()) == "x:7 a ");
    }
    CHECK(ds::str(ds::logContext()) == "a ");
  }

  SECTION("is thread local") {
    DS_CONTEXT("a");
    std::thread thread([&]() {
      DS_CONTEXT("b");
      CHECK(ds::str(ds::logContext()) == "b ");
    });
    thread.join();
    CHECK(ds::str(ds::logContext()) == "a ");
  }

  SECTION("ds::hideArg hides argument name") {
    int x = 8;
    DS_CONTEXT(ds::hideArg(x));
    CHECK(ds::str(ds::logContext()) == "8 ");
  }
}

TEST_CASE("DS_REQUIRE", TAGS) {
  SECTION("does nothing on success") { DS_REQUIRE(true, ""); }

  SECTION("accepts context arguments on success but does not evaluate them") {
    bool evaluated = false;
    auto evaluate = [&]() {
      evaluated = true;
      return 0;
    };
    DS_CONTEXT(evaluate());
    int x = 5;
    DS_REQUIRE(true, "context string", x, evaluate());
    CHECK(!evaluated);
  }

  SECTION("throws on failure") { CHECK_THROWS_AS(DS_REQUIRE(false, ""), ds::Exception); }

  SECTION("captures context on failure") {
    auto foo = 7;
    DS_CONTEXT(foo);
    int line = 0;
    auto fail = [&] {
      int a = 1;
      line = __LINE__ + 1; // must be one line before DS_REQUIRE macro
      DS_REQUIRE(a == 4, "Variable a was compared to 4.", a);
    };
    CHECK_THROWS_FIND(
        fail(), ds::str("Variable a was compared to 4. a:1 condition:a == 4 ds/log-test.cpp:", line, " foo:7"));
  }
}

TEST_CASE("DS_FAIL_REQUIRE", TAGS) {
  SECTION("throws and captures arguments and context") {
    int line = 0;
    try {
      auto foo = 7;
      DS_CONTEXT(foo);
      int x = 5;
      line = __LINE__ + 1;
      DS_FAIL_REQUIRE("Invalid x.", x);
      CHECK(false);
    } catch (const std::exception& e) {
      CHECK(e.what() == ds::str("Invalid x. x:5 ds/log-test.cpp:", line, " foo:7"));
    }
  }
}

TEST_CASE("DS_FAIL_WITH_USER_MESSAGE", TAGS) {
  SECTION("throws ds::Exception and captures user and technical message") {
    int line = 0;
    try {
      auto foo = 7;
      DS_CONTEXT(foo);
      int x = 5;
      line = __LINE__ + 1;
      DS_FAIL_WITH_USER_MESSAGE("There was a problem.", "Invalid x.", x);
      CHECK(false);
    } catch (const ds::Exception& e) {
      CHECK(e.type() == ds::Exception::Type::FAILED);
      CHECK(e.what() == ds::str("Invalid x. x:5 ds/log-test.cpp:", line, " foo:7"));
      CHECK(e.userMessage() == "There was a problem.");
    }
  }
}

TEST_CASE("DS_ASSERT", TAGS) {
  SECTION("does nothing on success") { DS_ASSERT(true, ""); }

  SECTION("accepts context arguments on success but does not evaluate them") {
    bool evaluated = false;
    auto evaluate = [&]() {
      evaluated = true;
      return 0;
    };
    DS_CONTEXT(evaluate());
    int x = 5;
    DS_ASSERT(true, x, evaluate());
    CHECK(!evaluated);
  }
}

TEST_CASE("DS_FAIL_ASSERT", TAGS) {
  SECTION("compiles with context") {
    auto foo = 7;
    DS_CONTEXT(foo);
    int a = 1;
    if (0) {
      DS_FAIL_ASSERT(a);
    }
  }
}

TEST_CASE("tryParseSeverityChar", TAGS) {
  SECTION("Severity first uppercase letter is parsed") {
    CHECK(ds::tryParseLogSeverityChar("F") == ds::LogSeverity::FATAL);
    CHECK(ds::tryParseLogSeverityChar("E") == ds::LogSeverity::ERROR);
    CHECK(ds::tryParseLogSeverityChar("W") == ds::LogSeverity::WARNING);
    CHECK(ds::tryParseLogSeverityChar("N") == ds::LogSeverity::NOTICE);
    CHECK(ds::tryParseLogSeverityChar("I") == ds::LogSeverity::INFO);
    CHECK(ds::tryParseLogSeverityChar("D") == ds::LogSeverity::DEBUG);
  }
  SECTION("Severity first lowercase letter is not parsed") {
    CHECK(ds::tryParseLogSeverityChar("f") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("e") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("w") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("n") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("i") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("d") == std::nullopt);
  }
  SECTION("Full names, weird strings are not parsed") {
    CHECK(ds::tryParseLogSeverityChar("") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("FATAL") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("ERROR") == std::nullopt);
    CHECK(ds::tryParseLogSeverityChar("?") == std::nullopt);
  }
}

TEST_CASE("severityFromSyslog", TAGS) {
  CHECK(ds::logSeverityFromSyslog(0) == ds::LogSeverity::FATAL);
  CHECK(ds::logSeverityFromSyslog(1) == ds::LogSeverity::FATAL);
  CHECK(ds::logSeverityFromSyslog(2) == ds::LogSeverity::FATAL);
  CHECK(ds::logSeverityFromSyslog(3) == ds::LogSeverity::ERROR);
  CHECK(ds::logSeverityFromSyslog(4) == ds::LogSeverity::WARNING);
  CHECK(ds::logSeverityFromSyslog(5) == ds::LogSeverity::NOTICE);
  CHECK(ds::logSeverityFromSyslog(6) == ds::LogSeverity::INFO);
  CHECK(ds::logSeverityFromSyslog(7) == ds::LogSeverity::DEBUG);
  CHECK(ds::logSeverityFromSyslog(8) == ds::LogSeverity::DEBUG);
  CHECK(ds::logSeverityFromSyslog(100) == ds::LogSeverity::DEBUG);
}

namespace {
class MockedLoggerLogFunction : boost::noncopyable {
public:
  MockedLoggerLogFunction(ds::Logger::LogFunction logFunction) {
    ds::Logger::instance().setLogFunction(std::move(logFunction));
  }

  ~MockedLoggerLogFunction() { ds::Logger::instance().setLogFunction(ds::Logger::defaultLogFunction); }
};

} // namespace

TEST_CASE("dsLogChannel", TAGS) {
  SECTION("Default constructed channel has NOTICE severity.") {
    ds::LogChannel channel("dsLogChannelTest");
    CHECK(channel.name() == std::string("dsLogChannelTest"));
    CHECK(channel.severity() == ds::LogSeverity::NOTICE);
  }

  SECTION("Channel with initial severity") {
    ds::LogChannel channel("dsLogChannelTest", ds::LogSeverity::ERROR);
    CHECK(channel.name() == std::string("dsLogChannelTest"));
    CHECK(channel.severity() == ds::LogSeverity::ERROR);
  }

  SECTION("Severity can be changed") {
    ds::LogChannel channel("dsLogChannelTest");
    channel.setSeverity(ds::LogSeverity::WARNING);
    CHECK(channel.severity() == ds::LogSeverity::WARNING);
  }

  SECTION("shouldLog compares stored severity with passed one") {
    ds::LogChannel channel("dsLogChannelTest");
    CHECK(channel.shouldLog(ds::LogSeverity::FATAL) == true);
    CHECK(channel.shouldLog(ds::LogSeverity::ERROR) == true);
    CHECK(channel.shouldLog(ds::LogSeverity::WARNING) == true);
    CHECK(channel.shouldLog(ds::LogSeverity::NOTICE) == true);
    CHECK(channel.shouldLog(ds::LogSeverity::INFO) == false);
    CHECK(channel.shouldLog(ds::LogSeverity::DEBUG) == false);
  }

  SECTION("NOTICE log() is forwarded to Logger::log()") {
    bool called = false;
    MockedLoggerLogFunction logFunction([&](const char* channelName, ds::LogSeverity severity, std::string message) {
      CHECK(channelName == std::string("dsLogChannelTest"));
      CHECK(severity == ds::LogSeverity::NOTICE);
      CHECK(message == "message");
      called = true;
    });
    ds::LogChannel channel("dsLogChannelTest");
    channel.log(ds::LogSeverity::NOTICE, "message");
    CHECK(called);
  }

  SECTION("INFO log() is not forwarded to Logger::log()") {
    bool called = false;
    MockedLoggerLogFunction logFunction([&](const char*, ds::LogSeverity, std::string) { called = true; });
    ds::LogChannel channel("dsLogChannelTest");
    channel.log(ds::LogSeverity::INFO, "message");
    CHECK(!called);
  }
}

TEST_CASE("dsLogParseRules", TAGS) {
  SECTION("Parse global rules") {
    CHECK(ds::tryParseLogRules("D") == std::vector<ds::LogRule>({ds::LogRule({"", ds::LogSeverity::DEBUG})}));
    CHECK(ds::tryParseLogRules("W") == std::vector<ds::LogRule>({ds::LogRule({"", ds::LogSeverity::WARNING})}));
    CHECK(ds::tryParseLogRules("W,E") == std::vector<ds::LogRule>({ds::LogRule({"", ds::LogSeverity::WARNING}),
                                             ds::LogRule({"", ds::LogSeverity::ERROR})}));
  }

  SECTION("Parse channel rules") {
    CHECK(ds::tryParseLogRules("foo:D") == std::vector<ds::LogRule>({ds::LogRule({"foo", ds::LogSeverity::DEBUG})}));
  }
  SECTION("Parse global and channel rules") {
    CHECK(ds::tryParseLogRules("W,foo:D") == std::vector<ds::LogRule>({ds::LogRule({"", ds::LogSeverity::WARNING}),
                                                 ds::LogRule({"foo", ds::LogSeverity::DEBUG})}));
  }
  SECTION("Channel without severity is not allowed") {
    CHECK(ds::tryParseLogRules("foo:") == std::vector<ds::LogRule>());
  }
  SECTION("Non-ascii charactes in channel name are not allowed") {
    CHECK(ds::tryParseLogRules("foo+:D") == std::vector<ds::LogRule>());
  }
  SECTION("Invalid rules are scipped") {
    CHECK(ds::tryParseLogRules("I,foo+:D,goo:D") == std::vector<ds::LogRule>({ds::LogRule({"", ds::LogSeverity::INFO}),
                                                        ds::LogRule({"goo", ds::LogSeverity::DEBUG})}));
  }
}

TEST_CASE("dsLogDefaultChannel macros", TAGS) {
  auto functionName = ds::str(__PRETTY_FUNCTION__);
  ds::LogChannel dsLogChannel("dsLogDefaultChannelTest");

  CHECK(dsLogChannel.name() == std::string("dsLogDefaultChannelTest"));
  CHECK(dsLogChannel.severity() == ds::LogSeverity::NOTICE);

  ds::LogSeverity lastSeverity;
  std::string lastMessage;
  bool called = false;
  MockedLoggerLogFunction logFunction([&](const char* channelName, ds::LogSeverity severity, std::string message) {
    CHECK(channelName == std::string("dsLogDefaultChannelTest"));
    lastSeverity = severity;
    lastMessage = std::move(message);
    called = true;
  });
#define CASE(severity, message, code)               \
  code;                                             \
  CHECK(called);                                    \
  called = false;                                   \
  CHECK(lastSeverity == ds::LogSeverity::severity); \
  CHECK(lastMessage == ds::str(message, " y:2 ", "test.cpp:", __LINE__));

#define CASE_NOLOG(code) \
  code;                  \
  CHECK(!called);

  int x = 1;
  int y = 2;
  DS_CONTEXT(y);

  dsLogChannel.setSeverity(ds::LogSeverity::DEBUG);
  CASE(DEBUG, "hello x:1", DS_DEBUG("hello", x));
  CASE(DEBUG, ds::str("Enter ", functionName, " x:1"), DS_DEBUG_ENTER(x));
  CASE(DEBUG, ds::str("Leave ", functionName, " x:1"), DS_DEBUG_LEAVE(x));
  CASE(INFO, "hello x:1", DS_INFO("hello", x));
  CASE(INFO, ds::str("Enter ", functionName, " x:1"), DS_INFO_ENTER(x));
  CASE(INFO, ds::str("Leave ", functionName, " x:1"), DS_INFO_LEAVE(x));
  CASE(NOTICE, "hello x:1", DS_NOTICE("hello", x));
  CASE(WARNING, "hello x:1", DS_WARNING("hello", x));
  CASE(ERROR, "hello x:1", DS_ERROR("hello", x));
  CASE(FATAL, "hello x:1", DS_FATAL("hello", x));
  CASE(NOTICE, "hello x:1", DS_LOG(ds::LogSeverity::NOTICE, "hello", x));
  CASE(ERROR, "hello x:1", DS_LOG(ds::LogSeverity::ERROR, "hello", x));

  dsLogChannel.setSeverity(ds::LogSeverity::WARNING);
  CASE_NOLOG(DS_DEBUG("hello", x));
  CASE_NOLOG(DS_INFO("hello", x));
  CASE_NOLOG(DS_NOTICE("hello", x));
  CASE(WARNING, "hello x:1", DS_WARNING("hello", x));
  CASE(ERROR, "hello x:1", DS_ERROR("hello", x));
  CASE(FATAL, "hello x:1", DS_FATAL("hello", x));

#undef CASE
#undef CASE_NOLOG
}

TEST_CASE("DS_PRINT", TAGS) {
  std::string expectedMessage;
  bool called = false;
  MockedLoggerLogFunction logFunction([&](const char* channelName, ds::LogSeverity severity, std::string message) {
    CHECK(channelName == std::string(""));
    CHECK(severity == ds::LogSeverity::PRINT);
    CHECK(message == expectedMessage);
    called = true;
  });
  expectedMessage = ds::str("hello ds/log-test.cpp:", __LINE__ + 1);
  DS_PRINT("hello");
  CHECK(called);
}

TEST_CASE("dsLogStr", TAGS) {
  SECTION("stringifies") {
    CHECK(ds::strArgs() == "");
    CHECK(ds::strArgs("foo", "goo", 8) == "foogoo8");
  }
  SECTION("strips last space") {
    CHECK(ds::strArgs(" ") == "");
    CHECK(ds::strArgs("foo ") == "foo");
  }
}

TEST_CASE("dsTrimFile", TAGS) {
  SECTION("Do not trim files bellow 32 characters without src/") {
    CHECK(ds::trimFile("") == "");
    CHECK(ds::trimFile("a") == "a");
    CHECK(ds::trimFile("foo.h") == "foo.h");
    CHECK(ds::trimFile("foo.cpp") == "foo.cpp");
    CHECK(ds::trimFile("1234567890123456789012345678901") == "1234567890123456789012345678901");
    CHECK(ds::trimFile("12345678901234567890123456789012") == "12345678901234567890123456789012");
  }

  SECTION("Trim up to last src/") {
    CHECK(ds::trimFile("src/foo.cpp") == "foo.cpp");
    CHECK(ds::trimFile("../src/foo.cpp") == "foo.cpp");
    CHECK(ds::trimFile("src/src/foo.cpp") == "foo.cpp");
    CHECK(ds::trimFile("123456789012345678901234567890123/src/foo.cpp") == "foo.cpp");
  }

  SECTION("Trim with dots to 32 characters") {
    CHECK(ds::trimFile("123456789012345678901234567890123") == "...56789012345678901234567890123");
    CHECK(ds::trimFile("123456789012345678901234567890.cpp") == "...6789012345678901234567890.cpp");
    CHECK(ds::trimFile("src/123456789012345678901234567890.cpp") == "...6789012345678901234567890.cpp");
  }
}

TEST_CASE("dsLogger.defaultSeverity", TAGS) {
  auto& logger = ds::Logger::instance();

  SECTION("defaults to none") { CHECK(logger.defaultSeverity() == std::nullopt); }

  SECTION("none does not affect channel severity") {
    ds::LogChannel defaultChannel("dsLogTestDefault");
    ds::LogChannel infoChannel("dsLogTestInfo", ds::LogSeverity::INFO);
    CHECK(defaultChannel.severity() == ds::LogSeverity::NOTICE);
    CHECK(infoChannel.severity() == ds::LogSeverity::INFO);
  }

  SECTION("can be set") {
    auto oldSeverity = logger.defaultSeverity();
    DS_DEFER(logger.setDefaultSeverity(oldSeverity));
    logger.setDefaultSeverity(ds::LogSeverity::DEBUG);

    SECTION("and affects initial channel severity") {
      ds::LogChannel defaultChannel("dsLogTestDefault");
      ds::LogChannel infoChannel("dsLogTestInfo", ds::LogSeverity::INFO);
      CHECK(defaultChannel.severity() == ds::LogSeverity::DEBUG);
      CHECK(infoChannel.severity() == ds::LogSeverity::DEBUG);
    }
  }
}

TEST_CASE("Logger.channelSeverity", TAGS) {
  ds::LogChannel channel("dsLogTest");
  auto& logger = ds::Logger::instance();

  SECTION("can be read") { CHECK(logger.channelSeverity("dsLogTest") == ds::LogSeverity::NOTICE); }

  SECTION("can be set") {
    logger.setChannelSeverity("dsLogTest", ds::LogSeverity::DEBUG);
    CHECK(channel.severity() == ds::LogSeverity::DEBUG);
  }
}

TEST_CASE("dsLogChannel.initialChannelSeverity", TAGS) {
  ds::LogChannel channel("dsLogTest");
  auto& logger = ds::Logger::instance();

  SECTION("Sets channel severity with none default logger severity") {
    channel.setInitialSeverity(ds::LogSeverity::DEBUG);
    CHECK(channel.severity() == ds::LogSeverity::DEBUG);
  }

  SECTION("Does not override non-none default logger severity") {
    auto oldDefaultLoggerSeverity = logger.defaultSeverity();
    DS_DEFER(logger.setDefaultSeverity(oldDefaultLoggerSeverity));
    logger.setDefaultSeverity(ds::LogSeverity::INFO);

    channel.setInitialSeverity(ds::LogSeverity::DEBUG);
    CHECK(channel.severity() == ds::LogSeverity::INFO);
  }
}

TEST_CASE("Logger.channelSeverities", TAGS) {
  ds::LogChannel channel("dsLogLoggerSetDefaultSeverity");
  channel.setSeverity(ds::LogSeverity::DEBUG);
  const auto& logger = ds::Logger::instance();

  SECTION("returns our severity in the list") {
    auto found = false;
    for (const auto& x : logger.channelSeverities()) {
      if (x.name == "dsLogLoggerSetDefaultSeverity") {
        CHECK(x.severity == ds::LogSeverity::DEBUG);
        found = true;
      }
    }
    CHECK(found);
  }
}

TEST_CASE("narrow", TAGS) {
  CHECK(ds::narrow<int>(0) == 0);
  CHECK(ds::narrow<int>(7) == 7);
  CHECK(ds::narrow<int>(-7) == -7);
  CHECK(ds::narrow<unsigned int>(7) == 7);
  CHECK(ds::narrow<signed char>(127) == 127);
  CHECK(ds::narrow<signed char>(-128) == -128);

  CHECK_THROWS(ds::narrow<unsigned char>(777));
  CHECK_THROWS(ds::narrow<unsigned int>(-7));
  CHECK_THROWS(ds::narrow<unsigned char>(-1));
  CHECK_THROWS(ds::narrow<unsigned char>(256));
  CHECK_THROWS(ds::narrow<signed char>(128));
  CHECK_THROWS(ds::narrow<signed char>(-129));
}
