/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "log.h"
#include <execinfo.h>
#include <stdio.h>
#include <sys/syscall.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <boost/algorithm/string/predicate.hpp>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <mutex>
#include "date-time.h"
#include "enum.h"
#include "stack-trace.h"

namespace ds {

static thread_local LogContext* t_logContext;

LogContext::LogContext() : m_next(t_logContext) {
  t_logContext = this;
}

LogContext::~LogContext() {
  DS_ASSERT(t_logContext == this, "");
  t_logContext = m_next;
}

std::ostream& operator<<(std::ostream& stream, LogContext* context) {
  for (; context; context = context->m_next) {
    context->ostream(stream);
  }
  return stream;
}

LogContext* logContext() {
  return t_logContext;
}

ds::Enum<LogSeverity> dsEnum(LogSeverity) {
  static constexpr std::pair<LogSeverity, const char*> x[] = {
      {LogSeverity::PRINT, "print"}, {LogSeverity::FATAL, "fatal"}, {LogSeverity::ERROR, "error"},
      {LogSeverity::WARNING, "warning"}, {LogSeverity::NOTICE, "notice"}, {LogSeverity::INFO, "info"},
      {LogSeverity::DEBUG, "debug"}, {LogSeverity::FATAL, "fatal"},
  };
  return {gsl::make_span(x), "ds::LogSeverity"};
}
std::ostream& operator<<(std::ostream& stream, LogSeverity x) { return ds::ostream(stream, x); }

std::optional<LogSeverity> tryParseLogSeverityChar(const char* x) {
  if (x[0] && !x[1]) {
    switch (x[0]) {
      case 'F':
        return LogSeverity::FATAL;
      case 'E':
        return LogSeverity::ERROR;
      case 'W':
        return LogSeverity::WARNING;
      case 'N':
        return LogSeverity::NOTICE;
      case 'I':
        return LogSeverity::INFO;
      case 'D':
        return LogSeverity::DEBUG;
    }
  }
  return {};
}

LogSeverity logSeverityFromSyslog(int level) {
  switch (level) {
    case 0:
      return LogSeverity::FATAL;
    case 1:
      return LogSeverity::FATAL;
    case 2:
      return LogSeverity::FATAL;
    case 3:
      return LogSeverity::ERROR;
    case 4:
      return LogSeverity::WARNING;
    case 5:
      return LogSeverity::NOTICE;
    case 6:
      return LogSeverity::INFO;
    case 7:
      return LogSeverity::DEBUG;
  }
  return LogSeverity::DEBUG;
}

LogChannel::LogChannel(std::string m_name, LogSeverity defaultSeverity)
    : m_name(std::move(m_name)), m_severity(defaultSeverity) {
  Logger::instance().addChannel(*this);
}

LogChannel::~LogChannel() { Logger::instance().removeChannel(*this); }

void LogChannel::setInitialSeverity(LogSeverity x) {
  m_severity = x;
  Logger::instance().updateChannelSeverity(*this);
}

void LogChannel::log(LogSeverity severity, std::string message) const {
  if (!shouldLog(severity)) {
    return;
  }
  Logger::instance().log(name().c_str(), severity, std::move(message));
}

std::ostream& operator<<(std::ostream& stream, const LogChannel& x) {
  return stream << x.name() << ':' << x.severity();
}

/// Parse rule, return none on failure
std::optional<LogRule> tryParseLogRule(char* rule) {
  auto separator = strchr(rule, ':');
  if (!separator) {
    if (auto severity = tryParseLogSeverityChar(rule)) {
      return LogRule{std::string(), *severity};
    }
    return {};
  }

  auto name = std::string(rule, separator);
  for (const auto& c : name) {
    if (!::isalnum(c)) { // invalid character
      return {};
    }
  }
  if (auto severity = tryParseLogSeverityChar(separator + 1)) {
    return LogRule{std::move(name), *severity};
  }
  return {};
}

std::vector<LogRule> tryParseLogRules(const char* rulesIn) {
  std::vector<LogRule> out;
  std::vector<char> rulesCopy(rulesIn, rulesIn + strlen(rulesIn) + 1);
  char* data = rulesCopy.data();
  while (*data) {
    auto ruleEnd = strchr(data, ',');
    if (ruleEnd) {
      *ruleEnd = 0;
    }

    if (auto rule = tryParseLogRule(data)) {
      out.emplace_back(std::move(*rule));
    }

    if (ruleEnd) {
      data = ruleEnd + 1; // continue with next rule
    } else {
      break;
    }
  }
  return out;
}

struct Logger::Impl {
  std::mutex mutex;
  LogFunction logFunction;
  int fd;         // file descriptor for default logging function
  bool useColors; // use colors in default logging function?
  std::vector<LogChannel*> channels;
  std::optional<LogSeverity> defaultSeverity;
  std::vector<LogRule> rules;
  Impl() : fd(2), useColors() {}
};

static std::optional<std::string> configOption(const char* name) {
  // environment variable has highest priority
  if (const char* x = getenv(name)) {
    if (x[0]) { // empty value means the same as unset value
      return std::string(x);
    }
  }
  // fail back to etc config file
  // DS_SYSCONFDIR is passed from build system - usually "${prefix}/etc"
  auto etcFile = DS_SYSCONFDIR "/libds";
  try {
    if (access(etcFile, F_OK) == 0) {
      gsl::cstring_span nameSpan = name;
      std::ifstream ifstream(etcFile);
      std::string line;
      while (std::getline(ifstream, line)) {
        if (boost::starts_with(line, nameSpan) && line.at(nameSpan.size()) == '=') {
          if (line.size() > nameSpan.size() + 1) { // empty value means the same as unset value
            return line.substr(nameSpan.size() + 1);
          }
        }
      }
    }
  } catch (const std::exception& e) {
    // stick to C interfaces, this code can run very early
    fprintf(stderr, "Failed to parse %s, error:%s\n", etcFile, e.what());
  }
  return {};
}

Logger::Logger() : m_impl(new Impl()) {
  setLogFunction(Logger::defaultLogFunction);
  std::optional<std::string> maybeLog = configOption("DS_LOG");
  std::optional<std::string> maybeLogColors = configOption("DS_LOG_COLORS");

  if (maybeLog) {
    m_impl->rules = tryParseLogRules(maybeLog->c_str());
    for (const auto& rule : m_impl->rules) {
      if (rule.channelNamePrefix.empty()) {
        m_impl->defaultSeverity = rule.severity;
      }
    }
  }
  if (maybeLogColors && *maybeLogColors == "yes") {
    m_impl->useColors = true;
  } else if (maybeLogColors && *maybeLogColors == "no") {
    m_impl->useColors = false;
  } else {
    m_impl->useColors = isatty(m_impl->fd) ? true : false;
  }
}

Logger& Logger::instance() {
  // static pointer that leaks instance to avoid static
  // deinitialization order fiasco
  static Logger* s_instance = new Logger();
  return *s_instance;
}

std::optional<LogSeverity> Logger::defaultSeverity() const {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  return m_impl->defaultSeverity;
}

void Logger::setDefaultSeverity(std::optional<LogSeverity> severity) {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  m_impl->defaultSeverity = severity;
  for (auto&& channel : m_impl->channels) {
    updateChannelSeverity(*channel);
  }
}

LogSeverity Logger::channelSeverity(std::string_view channelName) const {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  DS_CONTEXT(channelName);
  for (auto&& channel : m_impl->channels) {
    if (channel->name() == channelName) {
      return channel->severity();
    }
  }
  DS_FAIL_REQUIRE("Log channel not found.");
}

void Logger::setChannelSeverity(std::string_view channelName, LogSeverity severity) {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  DS_CONTEXT(channelName, severity);
  for (auto&& channel : m_impl->channels) {
    if (channel->name() == channelName) {
      channel->setSeverity(severity);
      return;
    }
  }
  DS_FAIL_REQUIRE("Log channel not found.");
}

std::vector<Logger::ChannelSeverity> Logger::channelSeverities() const {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  std::vector<ChannelSeverity> out;
  out.reserve(m_impl->channels.size());
  for (const auto& channel : m_impl->channels) {
    out.push_back(ChannelSeverity({channel->name(), channel->severity()}));
  }
  return out;
}

void Logger::log(const char* channelName, LogSeverity severity, std::string message) const {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  m_impl->logFunction(channelName, severity, std::move(message));
}

void Logger::setLogFunction(LogFunction logFunction) {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  m_impl->logFunction = std::move(logFunction);
}

void Logger::defaultLogFunction(const char* channelName, LogSeverity severity, std::string message) {
  const auto& impl = Logger::instance().m_impl;
  const auto& fd = impl->fd;
  const auto& useColors = impl->useColors;
  auto timestamp = ds::toLocalMsIsoString(std::chrono::system_clock::now());
  const auto& strSeverity = [&] {
    switch (severity) {
      // Color escape sequences taken from gstreamer gst/gstinfo.c
      case LogSeverity::PRINT:
        return useColors ? "\033[34;01m!!!!!\033[0m" : "!!!!!";
      case LogSeverity::FATAL:
        return useColors ? "\033[31;01mFATAL\033[0m" : "FATAL";
      case LogSeverity::ERROR:
        return useColors ? "\033[31;01mERROR\033[0m" : "ERROR";
      case LogSeverity::WARNING:
        return useColors ? "\033[33;01mWARN \033[0m" : "WARN ";
      case LogSeverity::NOTICE:
        return useColors ? "\033[36mNOTIC\033[0m" : "NOTIC";
      case LogSeverity::INFO:
        return "INFO ";
      case LogSeverity::DEBUG:
        return "DEBUG";
    }
    return "?????";
  }();

  std::array<char, 8> tid;
  snprintf(tid.data(), tid.size(), "%5ld", syscall(SYS_gettid));

  auto str = ds::str('[', timestamp, "] ", tid.data(), " ", strSeverity, " ", channelName, " ", message, "\n");

  // Use raw file descriptor instead of std::cerr. Logger can be created
  // and called
  // from static initialization code and std::cerr may not be initialized
  // at this point.
  auto size = str.size();
  auto data = str.c_str();
  while (size) {
    auto written = ::write(fd, data, size);
    if (written < 0) {
      break;
    }
    size -= written;
    data += written;
  }
  ::fsync(fd);
}

void Logger::updateChannelSeverity(LogChannel& channel) {
  // Locked at higher level
  auto&& channelName = channel.name();
  auto severity = channel.severity();
  if (const auto& defaultSeverity = m_impl->defaultSeverity) {
    severity = *defaultSeverity;
  }
  for (const auto& rule : m_impl->rules) {
    if (boost::starts_with(channelName, rule.channelNamePrefix)) {
      severity = rule.severity;
    }
  }
  channel.setSeverity(severity);
}

void Logger::addChannel(LogChannel& channel) {
  std::lock_guard<std::mutex> lock(m_impl->mutex);

  for (auto&& x : m_impl->channels) {
    DS_REQUIRE(x->name() != channel.name(), "Log channels with duplicate name detected.", channel.name());
  }

  m_impl->channels.emplace_back(&channel);
  updateChannelSeverity(channel);
}

void Logger::removeChannel(LogChannel& channel) {
  std::lock_guard<std::mutex> lock(m_impl->mutex);
  m_impl->channels.erase(
      std::remove(m_impl->channels.begin(), m_impl->channels.end(), &channel), m_impl->channels.end());
}

static std::string trimFileToLastSrc(std::string file) {
  auto src = "src/";
  auto srcLen = ::strlen(src);
  auto pos = file.rfind(src);
  if (pos != std::string::npos) {
    file.erase(0, pos + srcLen);
  }
  return file;
}

static std::string trimFileWithDotsToMaxSize(std::string file) {
  constexpr auto maxSize = 32;
  constexpr auto dotsSize = 3;
  if (file.size() <= maxSize) {
    return file;
  }
  file.erase(0, file.size() - maxSize);
  DS_ASSERT(file.size() == maxSize, "");

  for (int i = 0; i < dotsSize; i++) {
    file[i] = '.';
  }
  return file;
}

std::string trimFile(std::string file) { return trimFileWithDotsToMaxSize(trimFileToLastSrc(std::move(file))); }

std::string trimFile(std::string file, const LogChannel& channel) {
  file = trimFileWithDotsToMaxSize(trimFileToLastSrc(std::move(file)));

  // trim common base with channel name
  const char* filePtr = file.c_str();
  const char* channelNamePtr = channel.name().c_str();
  while (*filePtr) {
    if (*filePtr == '/' || *filePtr == '-' || *filePtr == '_') {
      filePtr++;
      continue;
    }
    if (::tolower(*filePtr) == ::tolower(*channelNamePtr)) {
      filePtr++;
      channelNamePtr++;
      continue;
    }
    break;
  }
  return filePtr;
}

static void assertFailedWrite(gsl::cstring_span x) {
  auto size = x.size();
  auto data = x.data();
  while (size) {
    auto written = ::write(2, data, size);
    if (written < 0) {
      break;
    }
    size -= written;
    data += written;
  }
  ::fsync(2);
}

void assertFailed(char const* condition, char const* message, char const* function, char const* file, long line) {
  assertFailedWrite("Assertion failed. ");
  assertFailedWrite(message);
  assertFailedWrite(" condition:");
  assertFailedWrite(condition);
  assertFailedWrite(" ");
  assertFailedWrite(file);
  assertFailedWrite(":");
  {
    char buffer[33];
    snprintf(buffer, sizeof(buffer), "%ld", line);
    assertFailedWrite(buffer);
  }
  assertFailedWrite(" ");
  assertFailedWrite(function);
  assertFailedWrite("\n");

  const auto stackTrace = ds::StackTrace::capture();
  auto stackTraceSpan = stackTrace.span();
  assertFailedWrite("Stacktrace:");
  backtrace_symbols_fd(stackTraceSpan.data(), stackTraceSpan.size(), 2);
  ::fsync(2);

  assertFailedWrite(ds::str("context:", ::ds::logContext(), "\n"));
  assertFailedWrite("Decoded stacktrace:");
  writeSymbolsToFd(stackTrace, 2);
  ::fsync(2);

  abort();
}

void testAssert() {
  if (configOption("DS_TEST_ASSERT")) {
    DS_FAIL_ASSERT("Test assert.");
  }
}

} // namespace ds

// Boost assertion handlers integrated forwarded to DS_ASSERT.
// Active when compiled with BOOST_ENABLE_ASSERT_HANDLER.
//
// Having these handlers in libds uncoditionally
// with define BOOST_ENABLE_ASSERT_HANDLER
// does not compose with applications
// or other libraries that wants to provide own handlers.
// But it is good default for current ds projects using libds
// as base shared library.
#ifdef BOOST_ENABLE_ASSERT_HANDLER
namespace boost {
void assertion_failed(char const* expr, char const* function, char const* file, long line) {
  ds::assertFailed(expr, "", function, file, line);
}
void assertion_failed_msg(char const* expr, char const* msg, char const* function, char const* file, long line) {
  ds::assertFailed(expr, msg, function, file, line);
}
} // namespace boost
#endif
