#pragma once

#include <functional>
#include <memory>
#include <optional>
#include "promise-fwd.h"

namespace ds {

// Promise of value fullfilled by setting `value` and destroying the `Promise` object.
//
// Similar to std::promise, but std::promise (as of c++17)
// is usable only for thread synchronization.
// This promise is intended for asynchronous notifications.
// The same story for boost::promise.
template <class T>
class Promise {
public:
  /// @args fulfilled Callback called when promise is fulfilled (i.e. destroyed)
  Promise(std::function<void(std::optional<T>&&)>&& fulfilled);
  ~Promise();

  // Promissed value. Nullopt means that promise was cancelled.
  std::optional<T> value{};

private:
  std::function<void(std::optional<T>&&)> m_fulfilled{};
};

template <class T>
std::unique_ptr<Promise<T>> makePromise(std::function<void(std::optional<T>&&)>&& fulfilled) {
  return std::make_unique<Promise<T>>(std::move(fulfilled));
}

// implementation details
template <class T>
Promise<T>::Promise(std::function<void(std::optional<T>&&)>&& fulfilled) : m_fulfilled{std::move(fulfilled)} {};

template <class T>
Promise<T>::~Promise() {
  m_fulfilled(std::move(value));
}

} // namespace ds
