#pragma once

// Origin: https://github.com/bytemaster/boost_reflect reviosion9597e954497db673b952bd6c08515e587723647a
// License: Boost Software License

#include <stdint.h>
#include <array>
#include <boost/function_types/result_type.hpp>
#include <boost/fusion/container/vector.hpp>
#include <boost/preprocessor/facilities/empty.hpp>
#include <boost/preprocessor/seq/enum.hpp>
#include <boost/preprocessor/seq/for_each.hpp>
#include <boost/preprocessor/seq/for_each_i.hpp>
#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/stringize.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/type_traits/remove_const.hpp>
#include <boost/type_traits/remove_pointer.hpp>
#include <boost/type_traits/remove_reference.hpp>
#include <boost/typeof/typeof.hpp>
#include "common.h"
#include "enum-core.h"
#include "str.h"

// Reflects members of a struct and provide operators ==, !=
//
// ````cpp
// namespace nsA {
// namespace nsB {
// struct A {
//   std::string hello;
//   std::string world;
// };
// } //namespace nsB
// } //namespace nsA
//
// // at global scope
// DS_REFLECT_STRUCT( (nsA)(nsB), A, (hello)(world) )
//
//
// template<typename T>
// struct Cout {
//     const T& x;
//     Cout(const T& x) : x(x) {}
//
//     template<typename MemberPtr, MemberPtr m>
//     void operator()(const char* name) const {
//       std::cout << name << ":" << x.*m << "\n";
//     }
// };
// ds::reflector<T>::visit(Cout<T>(a));
//
// [Original documentation](http://bytemaster.github.io/mace/group__mace__reflect__typeinfo.html)

namespace ds {

/// @brief defines reflection for struct T
/// Unless this is specialized, visit() will not be defined for T.
///
/// @tparam T - the type that will be visited.
///
/// The @ref DS_REFLECT(TYPE,MEMBERS) or DS_REFLECT_DERIVED(TYPE,BASES,MEMBERS) macro is used to specialize this
/// class for your type at global scope.
///
template <typename T>
struct reflector;

/// T if T has reflection
template <class T>
using reflect_t = typename ds::reflector<T>::type;

/// constexpr check that all struct fields are reflected
template <typename T>
struct CheckAllFieldsVisitor {
  size_t offset;
  constexpr CheckAllFieldsVisitor() : offset(0) {}

  constexpr void align(size_t x) {
    if (offset % x != 0) {
      offset += x - (offset % x);
    }
  }

  template <typename MemberPtr, MemberPtr m>
  constexpr void operator()(const char* name) {
    align(std::alignment_of<decltype(std::declval<T>().*m)>::value);
    offset += sizeof(std::declval<T>().*m);
  }

  constexpr bool check() {
    align(std::alignment_of<T>::value);
    return offset == sizeof(T);
  }
};

template <typename T>
constexpr bool checkAllFields() {
  CheckAllFieldsVisitor<T> visitor;
  reflector<T>::visit(visitor);
  return visitor.check();
}

#define DS_REFLECT_VISIT_BASE(r, visitor, base) ds::reflector<base>::visit(visitor);

#define DS_REFLECT_VISIT_MEMBER(r, visitor, elem) \
  visitor.template operator()<BOOST_TYPEOF(&type::elem), &type::elem>(BOOST_PP_STRINGIZE(elem));

#define DS_REFLECT_BASE_MEMBER_COUNT(r, OP, elem) OP ds::reflector<elem>::total_member_count

#define DS_REFLECT_DERIVED_IMPL_INLINE(TYPE, INHERITS, MEMBERS) \
  template <typename Visitor>                                   \
  static constexpr void visit(Visitor&& v) {               \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_BASE, v, INHERITS)   \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_MEMBER, v, MEMBERS)  \
  }

#define DS_REFLECT_DERIVED_IMPL_EXT(TYPE, INHERITS, MEMBERS)   \
  template <typename Visitor>                                  \
  void ds::reflector<TYPE>::visit(const Visitor& v) {          \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_BASE, v, INHERITS)  \
    BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_VISIT_MEMBER, v, MEMBERS) \
  }

/**
 *  @def DS_REFLECT_DERIVED(TYPE,INHERITS,MEMBERS)
 *
 *  @brief Specializes ds::reflector for TYPE where
 *         type inherits other reflected classes
 *
 *  @param INHERITS - a sequence of base class names (basea)(baseb)(basec)
 *  @param MEMBERS - a sequence of member names.  (field1)(field2)(field3)
 */
#define DS_REFLECT_DERIVED(TYPE, INHERITS, MEMBERS)                                                            \
  namespace ds {                                                                                               \
  template <>                                                                                                  \
  struct reflector<TYPE> {                                                                                     \
    typedef boost::true_type is_defined;                                                                       \
    typedef TYPE type;                                                                                         \
    enum member_count_enum {                                                                                   \
      local_member_count = BOOST_PP_SEQ_SIZE(MEMBERS),                                                         \
      total_member_count = local_member_count BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_BASE_MEMBER_COUNT, +, INHERITS) \
    };                                                                                                         \
    DS_REFLECT_DERIVED_IMPL_INLINE(TYPE, INHERITS, MEMBERS)                                                    \
  };                                                                                                           \
  } // namespace ds

/**
 *  @def DS_REFLECT(TYPE,MEMBERS)
 *  @brief Specializes ds::reflector for TYPE
 *
 *  @param MEMBERS - a sequence of member names.  (field1)(field2)(field3)
 *
 *  @see DS_REFLECT_DERIVED
 */
#define DS_REFLECT(TYPE, MEMBERS) DS_REFLECT_DERIVED(TYPE, BOOST_PP_SEQ_NIL, MEMBERS)

template <typename T, typename Enable = void>
struct IsEmptyInstance {
  static bool call(const T& x) { return x == T(); }
};

/// Checks whether value is default instance.
/// Usefull in serialization code to skip fields with default value.
template <typename T>
bool isEmptyInstance(T&& x) {
  return IsEmptyInstance<typename std::decay<T>::type>::call(x);
}

template <>
struct IsEmptyInstance<const char*> {
  static bool call(const char* x) { return (x == nullptr) || (*x == 0); }
};

template <>
struct IsEmptyInstance<char*> {
  static bool call(const char* x) { return IsEmptyInstance<const char*>::call(x); }
};

// Specialize for all containers with `empty()` method.
// Operator == in default template expends to more code
// and requires item's `operator == `.
template <typename T>
struct IsEmptyInstance<T,
                       typename std::enable_if<std::is_same<bool, decltype(std::declval<T>().empty())>::value>::type> {
  static bool call(const T& x) { return x.empty(); }
};

template <typename T>
struct ReflectedEqualVisitor {
  bool result;
  const T& a;
  const T& b;
  ReflectedEqualVisitor(const T& a, const T& b) : result(true), a(a), b(b) {}

  template <typename MemberPtr, MemberPtr m>
  void operator()(const char*) {
    result = result && (a.*m == b.*m);
  }
};

// Are two reflected structs equal field by field?
// C++20 will be able to do this automatically with default operator <=>.
template <typename T>
bool reflectedEqual(const T& a, const T& b) {
  ReflectedEqualVisitor visitor(a, b);
  ds::reflector<T>::visit(visitor);
  return visitor.result;
}

template <typename T>
struct ReflectedOstream : boost::noncopyable {
  std::ostream& ostream;
  const T& x;
  mutable bool first;
  ReflectedOstream(std::ostream& ostream, const T& x) : ostream(ostream), x(x), first(true) {}

  template <typename MemberPtr, MemberPtr m>
  void operator()(const char* name) const {
    const auto& value = x.*m;
    typedef typename std::decay<decltype(value)>::type Member;

    if (isEmptyInstance(value)) {
      return;
    }
    if (first) {
      first = false;
    } else {
      ostream << ',';
    }
    ostream << '"' << name << '"' << ':';
    ::ds::detail::Str<Member>::f(ostream, value);
  }
};

template <typename T>
std::ostream& reflectedOstream(std::ostream& ostream, const T& x) {
  ostream << "{";
  ::ds::reflector<T>::visit(ReflectedOstream<T>(ostream, x));
  return ostream << "}";
};

#define DS_REFLECT_NAMESPACE(r, data, elem) \
  ::elem

#define DS_REFLECT_NAMESPACE_STRINGIZE(r, data, elem) "::" BOOST_PP_STRINGIZE(elem)

#define DS_REFLECT_OPEN_NAMESPACE(r, data, elem) \
  namespace elem {
#define DS_REFLECT_OPEN_NAMESPACES(NAME_SPACES) BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_OPEN_NAMESPACE, _, NAME_SPACES)

#define DS_REFLECT_CLOSE_NAMESPACE(r, data, elem) \
  }
#define DS_REFLECT_CLOSE_NAMESPACES(NAME_SPACES) BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_CLOSE_NAMESPACE, _, NAME_SPACES)

#define DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE) BOOST_PP_SEQ_FOR_EACH(DS_REFLECT_NAMESPACE, _, NAME_SPACES)::TYPE

// operators ==, != are good candidates for inlining.
#define DS_REFLECT_DECLARE_STRUCT_OPERATORS(TYPE)                                                              \
  [[maybe_unused]] inline bool operator==(const TYPE& a, const TYPE& b) { return ::ds::reflectedEqual(a, b); } \
  [[maybe_unused]] inline bool operator!=(const TYPE& a, const TYPE& b) { return !(a == b); }                  \
  [[maybe_unused]] std::ostream& operator<<(std::ostream& ostream, const TYPE& x);

#define DS_REFLECT_DEFINE_STRUCT_OPERATORS(TYPE)                                    \
  [[maybe_unused]] std::ostream& operator<<(std::ostream& ostream, const TYPE& x) { \
    return ::ds::reflectedOstream(ostream, x);                                      \
  }

#define DS_REFLECT_INLINE_STRUCT_OPERATORS(TYPE)                                                               \
  [[maybe_unused]] inline bool operator==(const TYPE& a, const TYPE& b) { return ::ds::reflectedEqual(a, b); } \
  [[maybe_unused]] inline bool operator!=(const TYPE& a, const TYPE& b) { return !(a == b); }                  \
  [[maybe_unused]] inline std::ostream& operator<<(std::ostream& ostream, const TYPE& x) {                     \
    return ::ds::reflectedOstream(ostream, x);                                                                 \
  }

#define DS_REFLECT_STRUCT(NAME_SPACES, TYPE, MEMBERS)                                      \
  DS_REFLECT(DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE), MEMBERS)                             \
  DS_REFLECT_OPEN_NAMESPACES(NAME_SPACES)                                                  \
  static_assert(::ds::checkAllFields<TYPE>(), "Some fields are not reflected correctly."); \
  DS_REFLECT_INLINE_STRUCT_OPERATORS(TYPE)                                                 \
  DS_REFLECT_CLOSE_NAMESPACES(NAME_SPACES)

#define DS_REFLECT_DECLARE_STRUCT(NAME_SPACES, TYPE, MEMBERS)  \
  DS_REFLECT(DS_REFLECT_FULL_TYPE(NAME_SPACES, TYPE), MEMBERS) \
  DS_REFLECT_OPEN_NAMESPACES(NAME_SPACES)                      \
  DS_REFLECT_DECLARE_STRUCT_OPERATORS(TYPE)                    \
  DS_REFLECT_CLOSE_NAMESPACES(NAME_SPACES)

#define DS_REFLECT_DEFINE_STRUCT(NAME_SPACES, TYPE)                                        \
  DS_REFLECT_OPEN_NAMESPACES(NAME_SPACES)                                                  \
  static_assert(::ds::checkAllFields<TYPE>(), "Some fields are not reflected correctly."); \
  DS_REFLECT_DEFINE_STRUCT_OPERATORS(TYPE)                                                 \
  DS_REFLECT_CLOSE_NAMESPACES(NAME_SPACES)

#define DS_REFLECT_ANONYMOUS_STRUCT(TYPE, MEMBERS)                                         \
  DS_REFLECT(::TYPE, MEMBERS)                                                              \
  namespace {                                                                              \
  static_assert(::ds::checkAllFields<TYPE>(), "Some fields are not reflected correctly."); \
  DS_REFLECT_INLINE_STRUCT_OPERATORS(::TYPE)                                               \
  }

} // namespace ds
