#include "system.h"
#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>
#include <cerrno>
#include <fstream>
#include <string>
#include "conf.h"
#include "log.h"
#include "memory.h"

static ds::LogChannel dsLogChannel("dsSystem");
namespace ds {

System::~System() {}

class SystemImpl final : public System {
public:
  SystemImpl() : m_dirs(Dirs{DS_SYSCONFDIR, DS_DATADIR}) {}
  SystemImpl(Dirs&& dirs) : m_dirs(std::move(dirs)) {}

  const Dirs& dirs() const override { return m_dirs; }

  std::string readFileImpl(const std::string& filename) const {
    try {
      std::ifstream in(filename, std::ios::in | std::ios::binary);
      in.exceptions(std::ifstream::failbit | std::ifstream::badbit);
      DS_REQUIRE(in, "");
      in.seekg(0, std::ios::end);
      std::string result;
      result.resize(in.tellg());
      in.seekg(0, std::ios::beg);
      in.read(&result[0], result.size());
      in.close();
      DS_DEBUG_LEAVE(filename, result);
      return (result);
    } catch (std::exception &e) {
      DS_FAIL_REQUIRE("Failed to read file", filename, e);
    }
  }

  std::string readFile(const std::string& filename) const override {
    DS_DEBUG_ENTER(filename);
    return readFileImpl(filename);
  }

  std::optional<std::string> readFileIfExists(const std::string& filename) const override {
    DS_DEBUG_ENTER(filename);
    if (boost::filesystem::exists(filename)) {
      return readFileImpl(filename);
    }
    return {};
  }

  ds::DateTime now() const override { return std::chrono::system_clock::now(); }

  ds::Conf& conf() const override { return *m_conf; }

  Release release() const override { return m_release; }

  bool isInternalRelease() const override {
    DS_DEBUG_ENTER(release());
    switch (release()) {
      case Release::fieldtest: [[fallthrough]];
      case Release::production: return false;
      case Release::devel: [[fallthrough]];
      case Release::testing: [[fallthrough]];
      default: return true;
    }
  };

private:
  const Dirs m_dirs;
  const std::unique_ptr<ds::Conf> m_conf{ds::makeConf(*this)};
  const Release m_release{ds::toEnum<Release>(
      boost::trim_copy(readFileIfExists(ds::str(dirs().sysconfdir, "/release-type")).value_or(std::string{})))};
};

System& system() {
  static SystemImpl x;
  return x;
}

std::unique_ptr<System> makeSystem() {
  return std::make_unique<SystemImpl>();
}

std::unique_ptr<System> makeSystem(System::Dirs&& dirs) {
  return std::make_unique<SystemImpl>(std::move(dirs));
}

} // namespace ds
