/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <gsl.h>
#include <optional>
#include "date-time.h"
#include "enum.h"

DS_REFLECTED_ENUM((ds), Release, (devel)(testing)(fieldtest)(production));

namespace ds {

class Conf;

///
/// System provides mocking friendly abstraction for filesystem, ...
///
class System : public boost::noncopyable {
public:
  struct Dirs {
    std::string sysconfdir;
    std::string datadir;
  };
  virtual ~System();

  /// System directories
  virtual const Dirs& dirs() const = 0;

  /// Return file content or fail.
  virtual std::string readFile(const std::string& filename) const = 0;

  /// Return file content if file exists, none otherwise.
  /// Fails if file reading fails.
  virtual std::optional<std::string> readFileIfExists(const std::string& filename) const = 0;

  /// Calls std::chrono::system_clock::now().
  ///
  /// Compared to std::chrono::system_clock::now(), this method can be mocked.
  ///
  /// This does not provided real std::clock concept that can be used e.g. for asio::timer.
  /// But do we really want to have timer against system clock?
  /// Will the behavior be sane when system time jumps?
  virtual ds::DateTime now() const = 0;

  virtual ds::Conf& conf() const = 0;

  virtual Release release() const = 0;

  // is release intended for internal use cases (devel, testing)
  virtual bool isInternalRelease() const = 0;
};

/// returns System singleton
System& system();

/// for unit testing
std::unique_ptr<System> makeSystem(System::Dirs&& dirs);

} // namespace ds
