/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/modelconst.h"
#include "model/modelmaintenance-mock.h"

static ds::LogChannel dsLogChannel("hardwareAbstraction", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][hardwareAbstraction]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid lightId{"0000000000000000000000000000000001"};
static const Dsuid shadeId{"0000000000000000000000000000000002"};
static const Dsuid dummyId{"0000000000000000000000000000000003"};
static const Dsuid jalusieId{"0000000000000000000000000000000004"};
static const Dsuid lightId2{"0000000000000000000000000000000005"};
static const Dsuid jokerId{"0000000000000000000000000000000005"};
static const Dsuid jokerUMRId{"0000000000000000000000000000000206"};
static const Dsuid jokerUMRId2{"0000000000000000000000000000000306"};
static const Dsuid lightIdNonCacheable{"0000000000000000000000000000000006"};

SCENARIO("Scene table access through HAL", TAGS) {
  Apartment apt(nullptr);
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  DeviceSpec spec{};
  spec.FunctionID = 0x1010;
  spec.ProductID = (1 << 10) | 200;
  spec.VendorID = 0;

  boost::shared_ptr<Device> light = apt.allocateDevice(lightId);
  light->setPartiallyFromSpec(spec);
  boost::shared_ptr<Device> shade = apt.allocateDevice(shadeId);
  spec.FunctionID = 0x2000;
  spec.ProductID = (3 << 10) | 200;
  shade->setPartiallyFromSpec(spec);
  boost::shared_ptr<Device> dummy = apt.allocateDevice(dummyId);
  boost::shared_ptr<Device> jalusie = apt.allocateDevice(jalusieId);
  spec.FunctionID = 0x2000;
  spec.ProductID = (3 << 10) | 220;
  jalusie->setPartiallyFromSpec(spec);
  boost::shared_ptr<Device> light2 = apt.allocateDevice(lightId2);
  spec.FunctionID = 0x1010;
  spec.ProductID = (1 << 10) | 200;
  light2->setPartiallyFromSpec(spec);
  boost::shared_ptr<Device> joker = apt.allocateDevice(jokerId);
  spec.FunctionID = 0x8010;
  spec.ProductID = 0x0CC8;
  joker->setPartiallyFromSpec(spec);
  boost::shared_ptr<Device> jokerUMR = apt.allocateDevice(jokerUMRId);
  spec.ProductID = 0x30C8;
  jokerUMR->setPartiallyFromSpec(spec);
  jokerUMR->setHardwareType(HardwareType::SW_UMR200);
  boost::shared_ptr<Device> jokerUMR2 = apt.allocateDevice(jokerUMRId2);
  spec.ProductID = 0x30C8;
  jokerUMR2->setPartiallyFromSpec(spec);
  jokerUMR2->setHardwareType(HardwareType::SW_UMR200);
  boost::shared_ptr<Device> lightNonCacheable = apt.allocateDevice(lightIdNonCacheable);
  spec.ProductID = 0x04E6;
  spec.FunctionID = 0x1002;
  lightNonCacheable->setPartiallyFromSpec(spec);
  lightNonCacheable->setHardwareType(HardwareType::GE_TKM230);

  ds::sqlite3::Database d(":memory:");
  HardwareAbstraction hal(&apt, &d);
  hal.getCache().empty();

  WHEN("A light device hat no outputs") {
    THEN("It is not cacheable") {
      REQUIRE(hal.deviceCacheable(lightIdNonCacheable) == false);
    }
  }

  WHEN("a not implemented device") {
    THEN("hal access fails") {
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/brightness", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/dontCare", true));
      REQUIRE_THROWS(hal.write(dummyId, "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/scene/5/brightness"), 1) == 62.0);
      REQUIRE_THROWS(hal.readBool(dummyId, "/scene/5/dontCare") == false);
      REQUIRE_THROWS(hal.readString(dummyId, "/scene/5/dimTime") == "dimTime1");
    }
  }

  WHEN("not implemented paths") {
    THEN("hal access fails") {
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "//", 50.0));
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", true));
      REQUIRE_THROWS(hal.write(dummyId, "", true));
      REQUIRE_THROWS(hal.write(dummyId, "/", true));
      REQUIRE_THROWS(hal.write(dummyId, "//", true));
      REQUIRE_THROWS(hal.write(dummyId, "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(dummyId, "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(dummyId, "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readBool(dummyId, "/xxx") == false);
      REQUIRE_THROWS(hal.readBool(dummyId, "") == false);
      REQUIRE_THROWS(hal.readBool(dummyId, "/") == false);
      REQUIRE_THROWS(hal.readBool(dummyId, "//") == false);
      REQUIRE_THROWS(hal.readString(dummyId, "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(dummyId, "//") == "ajhasdjka");

      REQUIRE_THROWS(hal.write(lightId, "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(lightId, "", 50.0));
      REQUIRE_THROWS(hal.write(lightId, "/", 50.0));
      REQUIRE_THROWS(hal.write(lightId, "//", 50.0));
      REQUIRE_THROWS(hal.write(lightId, "/xxx", true));
      REQUIRE_THROWS(hal.write(lightId, "", true));
      REQUIRE_THROWS(hal.write(lightId, "/", true));
      REQUIRE_THROWS(hal.write(lightId, "//", true));
      REQUIRE_THROWS(hal.write(lightId, "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(lightId, "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(lightId, "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(lightId, "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(lightId, "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(lightId, ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(lightId, "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(lightId, "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readBool(lightId, "/xxx") == false);
      REQUIRE_THROWS(hal.readBool(lightId, "") == false);
      REQUIRE_THROWS(hal.readBool(lightId, "/") == false);
      REQUIRE_THROWS(hal.readBool(lightId, "//") == false);
      REQUIRE_THROWS(hal.readString(lightId, "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(lightId, "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(lightId, "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(lightId, "//") == "ajhasdjka");

      REQUIRE_THROWS(hal.write(shadeId, "/xxx", 50.0));
      REQUIRE_THROWS(hal.write(shadeId, "", 50.0));
      REQUIRE_THROWS(hal.write(shadeId, "/", 50.0));
      REQUIRE_THROWS(hal.write(shadeId, "//", 50.0));
      REQUIRE_THROWS(hal.write(shadeId, "/xxx", true));
      REQUIRE_THROWS(hal.write(shadeId, "", true));
      REQUIRE_THROWS(hal.write(shadeId, "/", true));
      REQUIRE_THROWS(hal.write(shadeId, "//", true));
      REQUIRE_THROWS(hal.write(shadeId, "/xxx", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shadeId, "", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shadeId, "/", std::string("ajhakjdha")));
      REQUIRE_THROWS(hal.write(shadeId, "//", std::string("ajhakjdha")));
      REQUIRE_THROWS(roundDigits(hal.readDouble(shadeId, "/xxx"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shadeId, ""), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shadeId, "/"), 1) == 62.0);
      REQUIRE_THROWS(roundDigits(hal.readDouble(shadeId, "//"), 1) == 62.0);
      REQUIRE_THROWS(hal.readBool(shadeId, "/xxx") == false);
      REQUIRE_THROWS(hal.readBool(shadeId, "") == false);
      REQUIRE_THROWS(hal.readBool(shadeId, "/") == false);
      REQUIRE_THROWS(hal.readBool(shadeId, "//") == false);
      REQUIRE_THROWS(hal.readString(shadeId, "/xxx") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shadeId, "") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shadeId, "/") == "ajhasdjka");
      REQUIRE_THROWS(hal.readString(shadeId, "//") == "ajhasdjka");
    }
  }

  WHEN("writing a value") {
    THEN("lastChanged is updated") {
      ALLOW_CALL(busInterface.device, setDeviceConfig(_, _, _, _));
      ALLOW_CALL(busInterface.device, getDeviceConfig(_, _, _)).RETURN(0xAF);

      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/brightness", 50.0));
      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/dontCare", true));
      devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/ledConfig", std::string("ledCon1")));
      devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/shadePositionOutside", 50.0));
      devCache = hal.getCache().tryLookup(shadeId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", 34.0));
      devCache = hal.getCache().tryLookup(jalusieId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);
      
      REQUIRE_NOTHROW(hal.write(jokerId, "/scene/5/powerState", 2.0));
      devCache = hal.getCache().tryLookup(jokerId);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);
    }
  }

  WHEN("channel dontCare of terminal blocks") {
    THEN("writing succeeds, but writes 'false'") {
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/brightness/dontCare", true));
      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::brightness].dontCare == false);
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/shadePositionOutside/dontCare", true));
      devCache = hal.getCache().tryLookup(shadeId);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::shadePositionOutside].dontCare == false);
      REQUIRE_NOTHROW(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside/dontCare", true));
      devCache = hal.getCache().tryLookup(jalusieId);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::shadeOpeningAngleOutside].dontCare == false);
      REQUIRE_NOTHROW(hal.write(jokerId, "/scene/5/powerState/dontCare", true));
      devCache = hal.getCache().tryLookup(jokerId);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::powerState].dontCare == false);
    }
    THEN("reading succeeds, but reads 'false'") {
      REQUIRE_NOTHROW(hal.readBool(lightId, "/scene/5/brightness/dontCare") == false);
      REQUIRE_NOTHROW(hal.readBool(shadeId, "/scene/5/shadePositionOutside/dontCare") == false);
      REQUIRE_NOTHROW(hal.readBool(jalusieId, "/scene/5/shadeOpeningAngleOutside/dontCare") == false);
      REQUIRE_NOTHROW(hal.readBool(jokerId, "/scene/5/powerState/dontCare") == false);
    }
  }

  WHEN("Writing a brightness value") {
    THEN("it writes to correct bank/offset") {
      // "/scene/5/brightness" -> 128, 5
      // "/scene/6/dontcare"   -> 128, 134 (=128+6)
      // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 128));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/brightness", 50.0));

      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::brightness].value == 50.0);
    }
  }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/brightness", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 255));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/brightness", 100));
    }
  }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/brightness", 100.1)); }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/brightness", -1)); }
  WHEN("Writing to an unavailable channel fails") {
    hal.getCache().remove(lightId);
    REQUIRE_THROWS(hal.write(lightId, "/scene/5/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(lightId, "/scene/5/unavailable", true));
    REQUIRE_THROWS(hal.write(lightId, "/scene/5/unavailable", std::string("jasdfhasdjklfh")));
    hal.getCache().remove(shadeId);
    REQUIRE_THROWS(hal.write(shadeId, "/scene/5/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(shadeId, "/scene/5/unavailable", true));
    REQUIRE_THROWS(hal.write(shadeId, "/scene/5/unavailable", std::string("jasdfhasdjklfh")));
    REQUIRE_THROWS(hal.write(jokerId, "/scene/5/brightness", 50.0));
  }
  WHEN("Writing to an unavailable property fails") {
    hal.getCache().remove(lightId);
    REQUIRE_THROWS(hal.write(lightId, "/property/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(lightId, "/property/unavailable", true));
    REQUIRE_THROWS(hal.write(lightId, "/property/unavailable", std::string("jasdfhasdjklfh")));
    hal.getCache().remove(shadeId);
    REQUIRE_THROWS(hal.write(shadeId, "/property/unavailable", 50.0));
    REQUIRE_THROWS(hal.write(shadeId, "/property/unavailable", true));
    REQUIRE_THROWS(hal.write(shadeId, "/property/unavailable", std::string("jasdfhasdjklfh")));
  }

  WHEN("Writing flags") {
    THEN("that are not cached yet requires read-modify-write bool") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("that are not cached yet requires read-modify-write string") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/dimTime", std::string("dimTime1")));

      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }
  WHEN("Writing flags with cache") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool(lightId, "/scene/5/dontCare") == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(lightId);
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("flags need to be boolean") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/dontCare", 25.0)); }
    THEN("Write ignoreLocalPrio") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x52));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/ignoreLocalPrio", true));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.ignoreLocalPrio == true);
    }
    THEN("Write specialHandling") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x54));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/specialHandling", true));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.specialHandling == true);
    }
    THEN("Write blink") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x58));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/blink", true));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.blink == true);
    }
    THEN("Write blink as number fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/blink", 3.5)); }
    THEN("Write ledConfig") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x60));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/ledConfig", std::string("ledCon1")));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
    THEN("Write ledConfig as bool fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/ledConfig", true)); }
    THEN("Write ledConfig out of range") {
      REQUIRE_THROWS(hal.write(lightId, "/scene/5/ledConfig", std::string("ledCon3")));
      REQUIRE_THROWS(hal.write(lightId, "/scene/5/ledConfig", std::string("1")));
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(lightId, "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
    THEN("Write dimTime as bool fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/5/dimTime", true)); }
    THEN("Write dimTime out of range") {
      REQUIRE_THROWS(hal.write(lightId, "/scene/5/dimTime", std::string("dimTime3")));
      REQUIRE_THROWS(hal.write(lightId, "/scene/5/dimTime", std::string("1")));
    }
  }
  WHEN("Writing flags with cache to shade device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool(shadeId, "/scene/5/dontCare") == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(shadeId);
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(shadeId).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }
  
  WHEN("Writing flags with cache to joker device") {
    REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
    REQUIRE(hal.readBool(jokerId, "/scene/5/dontCare") == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x51));
      REQUIRE_NOTHROW(hal.write(jokerId, "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(jokerId);
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("Write dimTime") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 128 + 5, 0x90));
      REQUIRE_NOTHROW(hal.write(jokerId, "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE(*hal.getCache().tryLookup(jokerId).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }

  WHEN("reading brightness channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(lightId, "/scene/5/brightness"), 1) == 62.0);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(lightId).scene.sceneTable[5].channels[ChannelType::brightness].value, 1) ==
              62.0);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE(hal.readDouble(lightId, "/scene/5/brightness") == 0);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].channels[ChannelType::brightness].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(255);
      REQUIRE(hal.readDouble(lightId, "/scene/5/brightness") == 100);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].channels[ChannelType::brightness].value == 100);
    }
  }
  WHEN("reading powerState channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(jokerId, "/scene/5/powerState"), 1) == 1.2);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(jokerId).scene.sceneTable[5].channels[ChannelType::powerState].value, 1) ==
              1.2);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE(hal.readDouble(jokerId, "/scene/5/powerState") == 0);
      REQUIRE(*hal.getCache().tryLookup(jokerId).scene.sceneTable[5].channels[ChannelType::powerState].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(255);
      REQUIRE(hal.readDouble(jokerId, "/scene/5/powerState") == 2.0);
      REQUIRE(*hal.getCache().tryLookup(jokerId).scene.sceneTable[5].channels[ChannelType::powerState].value == 2.0);
    }
  }
  WHEN("reading from an unavailable channel fails") {
    REQUIRE_THROWS(hal.readBool(lightId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(lightId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(lightId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readBool(shadeId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(shadeId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(shadeId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readBool(jokerId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(jokerId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readString(jokerId, "/scene/5/unavailable"));
    REQUIRE_THROWS(hal.readDouble(jokerId, "/scene/5/brightness"));
  }
  WHEN("reading from an unavailable property fails") {
    REQUIRE_THROWS(hal.readBool(lightId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(lightId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(lightId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readBool(shadeId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(shadeId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(shadeId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readBool(jokerId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readDouble(jokerId, "/property/unavailable"));
    REQUIRE_THROWS(hal.readString(jokerId, "/property/unavailable"));
  }
  WHEN("reading from an unavailable scene fails") {
    REQUIRE_THROWS(hal.readBool(lightId, "/scene/128/dontCare"));
    REQUIRE_THROWS(hal.readDouble(lightId, "/scene/128/brightness"));
  }

  WHEN("Reading flags") {
    THEN("can be read directly") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE(hal.readBool(lightId, "/scene/5/dontCare") == false);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.dontCare == false);
    }
    THEN("flags need to be boolean") { REQUIRE_THROWS(hal.readDouble(lightId, "/scene/5/dontCare") == 25.0); }
    THEN("Read ignoreLocalPrio") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x52);
      REQUIRE(hal.readBool(lightId, "/scene/5/ignoreLocalPrio") == true);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.ignoreLocalPrio == true);
    }
    THEN("Read specialHandling") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x56);
      REQUIRE(hal.readBool(lightId, "/scene/5/specialHandling") == true);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.specialHandling == true);
    }
    THEN("Read blink") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x5E);
      REQUIRE(hal.readBool(lightId, "/scene/5/blink") == true);
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.blink == true);
    }
    THEN("Read blink as number fails") { REQUIRE_THROWS(hal.readDouble(lightId, "/scene/5/blink") == 3.5); }
    THEN("Read ledConfig") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x6E);
      REQUIRE(hal.readString(lightId, "/scene/5/ledConfig") == "ledCon1");
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
    THEN("Read ledConfig as bool fails") { REQUIRE_THROWS(hal.readBool(lightId, "/scene/5/ledConfig") == true); }
    THEN("Read dimTime") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0xAE);
      REQUIRE(hal.readString(lightId, "/scene/5/dimTime") == "dimTime1");
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
    THEN("Read dimTime as bool fails") { REQUIRE_THROWS(hal.readBool(lightId, "/scene/5/dimTime") == true); }
  }
  WHEN("Reading flags from shade device") {
    THEN("can be read directly") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0x50);
      REQUIRE(hal.readBool(shadeId, "/scene/5/dontCare") == false);
      REQUIRE(*hal.getCache().tryLookup(shadeId).scene.sceneTable[5].flags.dontCare == false);
    }
    THEN("Read dimTime") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 128 + 5)).RETURN(0xAE);
      REQUIRE(hal.readString(shadeId, "/scene/5/dimTime") == "dimTime1");
      REQUIRE(*hal.getCache().tryLookup(shadeId).scene.sceneTable[5].flags.dimTime == DimTime::dimTime1);
    }
  }

  WHEN("Writing a shadePositionOutside value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x02);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0x94));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF2));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/shadePositionOutside", 58.2));

      REQUIRE(
          *hal.getCache().tryLookup(shadeId).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value ==
          58.2);

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0xF8);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0x26));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF2));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/4/shadePositionOutside", 14.9));

      REQUIRE(
          *hal.getCache().tryLookup(shadeId).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value ==
          14.9);
    }
  }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/5/shadePositionOutside", 100.1)); }
  WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/5/shadePositionOutside", -1)); }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x05));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/shadePositionOutside", 0));

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x40));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/4/shadePositionOutside", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 5, 0xFF));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0xF5));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/5/shadePositionOutside", 100));

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x45);
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 128, 4, 0xFF));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 129, 2, 0x4F));
      REQUIRE_NOTHROW(hal.write(shadeId, "/scene/4/shadePositionOutside", 100));
    }
  }

  WHEN("reading shadePositionOutside channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0xE4);
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x08);
      REQUIRE(roundDigits(hal.readDouble(shadeId, "/scene/5/shadePositionOutside"), 1) == 89.1);
      REQUIRE(
          roundDigits(
              *hal.getCache().tryLookup(shadeId).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value,
              1) == 89.1);

      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 4)).RETURN(0x0E);
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x5D);
      REQUIRE(roundDigits(hal.readDouble(shadeId, "/scene/4/shadePositionOutside"), 1) == 5.8);
      REQUIRE(
          roundDigits(
              *hal.getCache().tryLookup(shadeId).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value,
              1) == 5.8);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 5)).RETURN(0);
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0x03);
      REQUIRE(hal.readDouble(shadeId, "/scene/5/shadePositionOutside") == 0);
      REQUIRE(
          *hal.getCache().tryLookup(shadeId).scene.sceneTable[5].channels[ChannelType::shadePositionOutside].value ==
          0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 128, 4)).RETURN(255);
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 129, 2)).RETURN(0xEF);
      REQUIRE(hal.readDouble(shadeId, "/scene/4/shadePositionOutside") == 100);
      REQUIRE(
          *hal.getCache().tryLookup(shadeId).scene.sceneTable[4].channels[ChannelType::shadePositionOutside].value ==
          100);
    }
  }

  WHEN("Writing a shadeOpeningAngleOutside value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0x79));
      REQUIRE_NOTHROW(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", 47.5));

      REQUIRE(*hal.getCache()
                   .tryLookup(jalusieId)
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 47.5);
    }
    THEN("a unsupporting device fails") {
      REQUIRE_THROWS(hal.write(shadeId, "/scene/5/shadeOpeningAngleOutside", 47.5));
    }
  }
  WHEN("Writing a illegal value fails") {
    REQUIRE_THROWS(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", 100.1));
  }
  WHEN("Writing a illegal value fails") {
    REQUIRE_THROWS(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", -1));
  }
  WHEN("Writing extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0x00));
      REQUIRE_NOTHROW(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 130, 5, 0xFF));
      REQUIRE_NOTHROW(hal.write(jalusieId, "/scene/5/shadeOpeningAngleOutside", 100));
    }
  }

  WHEN("reading shadeOpeningAngleOutside channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5)).RETURN(0x8F);
      REQUIRE(roundDigits(hal.readDouble(jalusieId, "/scene/5/shadeOpeningAngleOutside"), 1) == 56.1);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusieId)
                               .scene.sceneTable[5]
                               .channels[ChannelType::shadeOpeningAngleOutside]
                               .value,
                          1) == 56.1);
    }
    THEN("a unsupporting device fails") {
      REQUIRE_THROWS(hal.readDouble(shadeId, "/scene/5/shadeOpeningAngleOutside") == 56.2);
    }
  }
  WHEN("reading extremes correctly") {
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5)).RETURN(0);
      REQUIRE(hal.readDouble(jalusieId, "/scene/5/shadeOpeningAngleOutside") == 0);
      REQUIRE(*hal.getCache()
                   .tryLookup(jalusieId)
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 130, 5)).RETURN(255);
      REQUIRE(hal.readDouble(jalusieId, "/scene/5/shadeOpeningAngleOutside") == 100);
      REQUIRE(*hal.getCache()
                   .tryLookup(jalusieId)
                   .scene.sceneTable[5]
                   .channels[ChannelType::shadeOpeningAngleOutside]
                   .value == 100);
    }
  }
  WHEN("Writing angle to a device that does not support angles") {
    THEN("fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/5/shadeOpeningAngleOutside", 15.6)); }
  }

  WHEN("Updating cache via setDeviceConfig on bank 128 trapped by HAL") {
    THEN("should work for light device scene value") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, 128, 4, 10));
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(lightId).scene.sceneTable[4].channels[ChannelType::brightness].value, 1) ==
              3.9);
    }
    THEN("should work for light device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(lightId).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 128 for light devices") {
    THEN("should work for light device scene settings configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, 128, 4, 10));
      REQUIRE(hal.getCache().tryLookup(lightId).scene.sceneTable[4].channels[ChannelType::brightness].value);

      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(lightId).scene.sceneTable[4].channels[ChannelType::brightness].value, 1) ==
              3.9);
    }
  }
  
  WHEN("Updating cache via setDeviceConfig on bank 128 trapped by HAL for joker device") {
    THEN("should work for joker device scene value") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerId, 128, 4, 10));
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(jokerId).scene.sceneTable[4].channels[ChannelType::powerState].value, 1) ==
                0.1);
    }
    THEN("should work for joker device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerId, 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(jokerId).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 128 for light devices") {
    THEN("should work for light device scene settings configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerId, 128, 4, 10));
      REQUIRE(hal.getCache().tryLookup(jokerId).scene.sceneTable[4].channels[ChannelType::powerState].value);

      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(jokerId).scene.sceneTable[4].channels[ChannelType::powerState].value, 1) ==
              0.1);
    }
  }


  WHEN("Updating cache via device config trap on bank 128/129 for shade devices") {
    THEN("should work only if the whole 16bit are known") {
      /* check "even" scene */
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassScene, 56, 145));
      REQUIRE(
          !hal.getCache().tryLookup(jalusieId).scene.sceneTable[56].channels[ChannelType::shadePositionOutside].value);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassSceneExtention, 28, 14));
      REQUIRE(
          hal.getCache().tryLookup(jalusieId).scene.sceneTable[56].channels[ChannelType::shadePositionOutside].value);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusieId)
                               .scene.sceneTable[56]
                               .channels[ChannelType::shadePositionOutside]
                               .value,
                          1) == 57.0);

      /* check "odd" scene, 2nd register already known */
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassScene, 57, 145));
      REQUIRE(
          hal.getCache().tryLookup(jalusieId).scene.sceneTable[57].channels[ChannelType::shadePositionOutside].value);
      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusieId)
                               .scene.sceneTable[57]
                               .channels[ChannelType::shadePositionOutside]
                               .value,
                          1) == 56.6);
    }
    THEN("should work for shade device scene configuration") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, 128, 128 + 4, 1));
      REQUIRE(*hal.getCache().tryLookup(shadeId).scene.sceneTable[4].flags.dontCare == true);
    }
  }

  WHEN("Updating cache via device config trap on bank 130 for shade devices") {
    THEN("should work") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassSceneAngle, 4, 128));
      REQUIRE(hal.getCache()
                  .tryLookup(jalusieId)
                  .scene.sceneTable[4]
                  .channels[ChannelType::shadeOpeningAngleOutside]
                  .value);

      REQUIRE(roundDigits(*hal.getCache()
                               .tryLookup(jalusieId)
                               .scene.sceneTable[4]
                               .channels[ChannelType::shadeOpeningAngleOutside]
                               .value,
                          1) == 50.2);

      std::optional<ds::DateTime> before = hal.getCache().tryLookup(jalusieId).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassSceneExtention, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(jalusieId).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
  }

  WHEN("Updating cache via device config trap on unsupported bank") {
    REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, 128, 4, 128));
    REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, CfgClassSceneAngle, 4, 128));

    THEN("on light device should do nothing") {
      std::optional<ds::DateTime> before = hal.getCache().tryLookup(lightId).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, 127, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(lightId).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
    THEN("on shade device should do nothing") {
      std::optional<ds::DateTime> before = hal.getCache().tryLookup(jalusieId).lastChanged;
      REQUIRE(before);
      REQUIRE_NOTHROW(hal.updateDeviceConfig(jalusieId, 127, 64, 255));
      std::optional<ds::DateTime> after = hal.getCache().tryLookup(jalusieId).lastChanged;
      REQUIRE(after);
      REQUIRE(*before == *after);
    }
  }

  WHEN("Updating cache via device config update properties for light") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x43},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(lightId, p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(lightId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(lightId).property[p.prop] == 128);
    }
  }

  WHEN("Updating cache via device config update simple properties for shade") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::dimStepNum, 3, 0x3A},
                                        {DeviceProperty::dimStepLen, 3, 0x3B},
                                        {DeviceProperty::dimStepPause, 3, 0x3C}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(shadeId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[p.prop] == 128);
    }
  }
  
  WHEN("Updating cache via device config update properties for joker") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x3C},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerId, p.bank, p.offset, 128));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(jokerId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(jokerId).property[p.prop] == 128);
    }
  }
  
  WHEN("Updating cache via device config update properties for joker UMR primary output") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::outputMode, 3, 0x00},
                                        {DeviceProperty::switchedThreshold, 3, 0x30},
                                        {DeviceProperty::switchedThresholdTwoStep1, 3, 0x33},
                                        {DeviceProperty::switchedThreshold3Step0, 3, 0x34},
                                        {DeviceProperty::dimTime0Up, 3, 0x06},
                                        {DeviceProperty::dimTime0Down, 3, 0x07},
                                        {DeviceProperty::dimTime1Up, 3, 0x08},
                                        {DeviceProperty::dimTime1Down, 3, 0x09},
                                        {DeviceProperty::dimTime2Up, 3, 0x0A},
                                        {DeviceProperty::dimTime2Down, 3, 0x0B},
                                        {DeviceProperty::dimStepUp, 3, 0x0C},
                                        {DeviceProperty::dimStepDown, 3, 0x0D}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerUMRId, p.bank, p.offset, 152));

    THEN("Cache should update") {
      REQUIRE(hal.getCache().tryLookup(jokerUMRId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(jokerUMRId).property[p.prop] == 152);
    }
  }
  
  WHEN("Updating cache via device config update properties for joker UMR secondary output") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset;
    };
    auto p = GENERATE(values<Property>({{DeviceProperty::switchedThresholdTwoStep1, 3, 0x33},
                                        {DeviceProperty::switchedThreshold3Step0, 3, 0x34}}));

    REQUIRE_NOTHROW(hal.updateDeviceConfig(jokerUMRId2, p.bank, p.offset, 152));

    THEN("Cache should not be updated") {
      REQUIRE(hal.getCache().tryLookup(jokerUMRId2).property[p.prop] == std::nullopt);
    }
  }

  WHEN("Updating cache via device config update 2-byte properties for shade") {
    struct Property {
      DeviceProperty prop;
      int bank;
      int offset1;
      int offset2;
    };
    auto p = GENERATE(
        values<Property>({{DeviceProperty::posTimeUp, 3, 0x30, 0x31}, {DeviceProperty::posTimeDown, 3, 0x32, 0x33}}));

    THEN("Cache should update") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, p.bank, p.offset1, 128));
      REQUIRE(hal.getCache().tryLookup(shadeId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[p.prop] == 128);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, p.bank, p.offset2, 128));
      REQUIRE(hal.getCache().tryLookup(shadeId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[p.prop] == ((128 << 8) | 128));
    }

    THEN("Cache should update") {
      {
        auto cache = hal.getCache().tryLookup(shadeId);
        cache.property[p.prop] = 0xfeed;
        hal.getCache().update(shadeId, cache);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, p.bank, p.offset1, 0xad));
      REQUIRE(hal.getCache().tryLookup(shadeId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[p.prop] == 0xfead);

      REQUIRE_NOTHROW(hal.updateDeviceConfig(shadeId, p.bank, p.offset2, 0x55));
      REQUIRE(hal.getCache().tryLookup(shadeId).property[p.prop]);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[p.prop] == ((0x55 << 8) | 0xad));
    }
  }

  WHEN("Writing outputMode") {
    THEN("Write outputMode") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0, 16));
      REQUIRE_NOTHROW(hal.write(lightId, "/property/outputMode", std::string("KM_Switch")));
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::outputMode] ==
              static_cast<int>(OutputMode::KM_Switch));
    }
    THEN("Write outputMode as bool fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/outputMode", true)); }
    THEN("Write outputMode out of range") {
      REQUIRE_THROWS(hal.write(lightId, "/property/outputMode", std::string("KM_Switch45")));
      REQUIRE_THROWS(hal.write(lightId, "/property/outputMode", std::string("16")));
    }
  }

  WHEN("Writing dimTime") {
    THEN("Write dimTime exact") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 6, 0x26));
      REQUIRE_NOTHROW(hal.write(lightId, "/property/dimTime0Up", 0.2875));
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::dimTime0Up] == 0x26);
    }
    THEN("Write dimTime approx") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 6, 0xE3));
      REQUIRE_NOTHROW(hal.write(lightId, "/property/dimTime0Up", 1000));
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::dimTime0Up] == 0xE3);
    }
    THEN("Write outpdimTimeutMode as bool fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/dimTime0Up", true)); }
    THEN("Write dimTime as string fails") {
      REQUIRE_THROWS(hal.write(lightId, "/scene/dimTime0Up", std::string("ajfhadl")));
    }
    THEN("Write dimTime out of range") {
      REQUIRE_THROWS(hal.write(lightId, "/property/outputMode", 0.052));
      REQUIRE_THROWS(hal.write(lightId, "/property/outputMode", 3276.9));
    }
  }

  WHEN("Writing known property but wrong for this device") {
    THEN("fails") { REQUIRE_THROWS(hal.write(lightId, "/scene/dimTime0Up", std::string("ajfhadl"))); }
  }

  WHEN("Writing switchedThreshold") {
    THEN("Write switchedThreshold") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x43, 190));
      REQUIRE_NOTHROW(hal.write(lightId, "/property/switchedThreshold", 74.5));
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Write switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.write(lightId, "/scene/switchedThreshold", true));
    }
    THEN("Write switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.write(lightId, "/scene/switchedThreshold", std::string("KM_Switch45")));
    }
    THEN("Write switchedThreshold out of range") {
      REQUIRE_THROWS(hal.write(lightId, "/property/switchedThreshold", -.2));
      REQUIRE_THROWS(hal.write(lightId, "/property/switchedThreshold", 100.1));
    }
  }
  
  WHEN("Writing switchedThreshold for joker") {
    THEN("Write switchedThreshold") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3C, 190));
      REQUIRE_NOTHROW(hal.write(jokerId, "/property/switchedThreshold", 74.5));
      REQUIRE(*hal.getCache().tryLookup(jokerId).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Write switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.write(jokerId, "/scene/switchedThreshold", true));
    }
    THEN("Write switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.write(jokerId, "/scene/switchedThreshold", std::string("KM_Switch45")));
    }
    THEN("Write switchedThreshold out of range") {
      REQUIRE_THROWS(hal.write(jokerId, "/property/switchedThreshold", -.2));
      REQUIRE_THROWS(hal.write(jokerId, "/property/switchedThreshold", 100.1));
    }
  }

  WHEN("Reading outputMode") {
    THEN("Read outputMode") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0)).RETURN(19);
      REQUIRE(hal.readString(lightId, "/property/outputMode") == "LeadingEdgeDimmer");
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::outputMode] ==
              static_cast<int>(OutputMode::LeadingEdgeDimmer));
    }
    THEN("Read outputMode as bool fails") { REQUIRE_THROWS(hal.readBool(lightId, "/property/outputMode") == true); }
    THEN("Read outputMode as double fails") {
      REQUIRE_THROWS(hal.readDouble(lightId, "/property/outputMode") == 45.54);
    }
  }

  WHEN("Reading dimTime") {
    THEN("Read dimTime exact") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 6)).RETURN(0x26);
      REQUIRE_NOTHROW(hal.readDouble(lightId, "/property/dimTime0Up") == 0.2875);
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::dimTime0Up] == 0x26);
    }
    THEN("Read outpdimTimeutMode as bool fails") { REQUIRE_THROWS(hal.readBool(lightId, "/scene/dimTime0Up") == true); }
    THEN("Read dimTime as string fails") {
      REQUIRE_THROWS(hal.readString(lightId, "/scene/dimTime0Up") == std::string("ajfhadl"));
    }
  }

  WHEN("Reading switchedThreshold") {
    THEN("Read switchedThreshold") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0x43)).RETURN(190);
      REQUIRE_NOTHROW(hal.readDouble(lightId, "/property/switchedThreshold") == 74.5);
      REQUIRE(*hal.getCache().tryLookup(lightId).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Read switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.readBool(lightId, "/scene/switchedThreshold") == true);
    }
    THEN("Read switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.readString(lightId, "/scene/switchedThreshold") == std::string("KM_Switch45"));
    }
  }
  WHEN("Reading switchedThreshold for joker devices") {
    THEN("Read switchedThreshold") {
      REQUIRE_CALL(busInterface.device, getDeviceConfig(_, 3, 0x3C)).RETURN(190);
      REQUIRE_NOTHROW(hal.readDouble(jokerId, "/property/switchedThreshold") == 1.5);
      REQUIRE(*hal.getCache().tryLookup(jokerId).property[DeviceProperty::switchedThreshold] == 190);
    }
    THEN("Read switchedThreshold as bool fails") {
      REQUIRE_THROWS(hal.readBool(jokerId, "/scene/switchedThreshold") == true);
    }
    THEN("Read switchedThreshold as string fails") {
      REQUIRE_THROWS(hal.readString(jokerId, "/scene/switchedThreshold") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing dimStepNum") {
    THEN("Write dimStepNum") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3A, 245));
      REQUIRE_NOTHROW(hal.write(shadeId, "/property/dimStepNum", 245));
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::dimStepNum] == 245);
    }
    THEN("Write dimStepNum as bool fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/dimStepNum", true)); }
    THEN("Write dimStepNum as string fails") {
      REQUIRE_THROWS(hal.write(shadeId, "/scene/dimStepNum", std::string("KM_Switch45")));
    }
    THEN("Write dimStepNum out of range") {
      REQUIRE_THROWS(hal.write(shadeId, "/property/dimStepNum", -1));
      REQUIRE_THROWS(hal.write(shadeId, "/property/dimStepNum", 256));
    }
  }

  WHEN("Reading dimStepNum") {
    THEN("Read dimStepNum") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x3A)).RETURN(0x6730);
      REQUIRE_NOTHROW(hal.readDouble(shadeId, "/property/dimStepNum") == 48);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::dimStepNum] == 48);
    }
    THEN("Read dimStepNum as bool fails") { REQUIRE_THROWS(hal.readBool(shadeId, "/scene/dimStepNum") == true); }
    THEN("Read dimStepNum as string fails") {
      REQUIRE_THROWS(hal.readString(shadeId, "/scene/dimStepNum") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing dimStepLen") {
    THEN("Write dimStepLen") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x3B, 178));
      REQUIRE_NOTHROW(hal.write(shadeId, "/property/dimStepLen", 1.78));
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::dimStepLen] == 178);
    }
    THEN("Write dimStepLen as bool fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/dimStepLen", true)); }
    THEN("Write dimStepLen as string fails") {
      REQUIRE_THROWS(hal.write(shadeId, "/scene/dimStepLen", std::string("KM_Switch45")));
    }
    THEN("Write dimStepLen out of range") {
      REQUIRE_THROWS(hal.write(shadeId, "/property/dimStepLen", -1));
      REQUIRE_THROWS(hal.write(shadeId, "/property/dimStepLen", 256));
    }
  }

  WHEN("Reading dimStepLen") {
    THEN("Read dimStepLen") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x3A)).RETURN(0x6730);
      REQUIRE_NOTHROW(hal.readDouble(shadeId, "/property/dimStepLen") == 1.03);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::dimStepLen] == 103);
    }
    THEN("Read dimStepLen as bool fails") { REQUIRE_THROWS(hal.readBool(shadeId, "/scene/dimStepLen") == true); }
    THEN("Read dimStepLen as string fails") {
      REQUIRE_THROWS(hal.readString(shadeId, "/scene/dimStepLen") == std::string("KM_Switch45"));
    }
  }

  WHEN("Writing posTimeUp") {
    THEN("Write posTimeUp") {
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x30, 45));
      REQUIRE_CALL(busInterface.device, setDeviceConfig(_, 3, 0x31, 201));
      REQUIRE_NOTHROW(hal.write(shadeId, "/property/posTimeUp", 515.01));
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::posTimeUp] == 51501);
    }
    THEN("Write posTimeUp as bool fails") { REQUIRE_THROWS(hal.write(shadeId, "/scene/posTimeUp", true)); }
    THEN("Write posTimeUp as string fails") {
      REQUIRE_THROWS(hal.write(shadeId, "/scene/posTimeUp", std::string("KM_Switch45")));
    }
    THEN("Write posTimeUp out of range") {
      REQUIRE_THROWS(hal.write(shadeId, "/property/posTimeUp", -1));
      REQUIRE_THROWS(hal.write(shadeId, "/property/posTimeUp", 65536));
    }
  }

  WHEN("Reading posTimeUp") {
    THEN("Read posTimeUp") {
      REQUIRE_CALL(busInterface.device, getDeviceConfigWord(_, 3, 0x30)).RETURN(0x4D4A);
      REQUIRE_NOTHROW(hal.readDouble(shadeId, "/property/posTimeUp") == 197.86);
      REQUIRE(*hal.getCache().tryLookup(shadeId).property[DeviceProperty::posTimeUp] == 19786);
    }
    THEN("Read posTimeUp as bool fails") { REQUIRE_THROWS(hal.readBool(shadeId, "/scene/posTimeUp") == true); }
    THEN("Read posTimeUp as string fails") {
      REQUIRE_THROWS(hal.readString(shadeId, "/scene/posTimeUp") == std::string("KM_Switch45"));
    }
  }

  WHEN("Querying area membership") {
    THEN("the answer should be correct") {
      DeviceCache cache = hal.getCache().tryLookup(lightId);
      /* device is only in area 1 */
      cache.scene.sceneTable[SceneA11].flags.dontCare = false;
      cache.scene.sceneTable[SceneA21].flags.dontCare = true;
      cache.scene.sceneTable[SceneA31].flags.dontCare = true;
      cache.scene.sceneTable[SceneA41].flags.dontCare = true;
      hal.getCache().update(lightId, cache);
      std::vector<int> areas = hal.areaMember(lightId);
      REQUIRE(areas.size() == 1);
      REQUIRE(areas.at(0) == SceneA11);
    }
  }

  WHEN("Querying cache status") {
    auto status = hal.cacheStatus(lightId);
    THEN("cache is empty") {
      REQUIRE(status.completed == 0);
      REQUIRE(status.completion == 0);
      REQUIRE(status.total == 149);
    }
  }

  WHEN("Quering cache status of partially populated device") {
    {
      DeviceCache cache = hal.getCache().tryLookup(lightId);
      /* device is only in area 1 */
      cache.scene.sceneTable[SceneA11].flags.dontCare = false;
      cache.scene.sceneTable[SceneA11].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA21].flags.dontCare = true;
      cache.scene.sceneTable[SceneA21].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA31].flags.dontCare = true;
      cache.scene.sceneTable[SceneA31].channels[ChannelType::brightness].value = 50;
      cache.scene.sceneTable[SceneA41].flags.dontCare = true;
      cache.scene.sceneTable[SceneA41].channels[ChannelType::brightness].value = 50;
      hal.getCache().update(lightId, cache);
    }
    auto status = hal.cacheStatus(lightId);
    THEN("cache is partially filled") {
      REQUIRE(status.completed == 8);
      REQUIRE(status.completion == (8.0 / 149.0) * 100);
      REQUIRE(status.total == 149);
    }
  }
}

SCENARIO("Performance testing for cache status", "[dss][hardwareAbstraction][!hide]") {
  Apartment apt(nullptr);
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  DeviceSpec spec;
  spec.FunctionID = 0x1000;
  spec.ProductID = (1 << 10) | 200;

  boost::shared_ptr<Device> light = apt.allocateDevice(lightId);
  light->setPartiallyFromSpec(spec);

//  ds::sqlite3::Database d("./cache.db");
  ds::sqlite3::Database d(":memory:");
  HardwareAbstraction hal(&apt, &d);
  hal.getCache().empty();

  WHEN("Querying cache status speed-test") {
    ds::DateTime start = std::chrono::high_resolution_clock::now();
    for (int i = 0; i < 1000; ++i) {
      auto status = hal.cacheStatus(lightId);
      //      DS_DEBUG(status.completion);
    }
    ds::DateTime end = std::chrono::high_resolution_clock::now();
    REQUIRE(std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count() == 10);
  }
}
} // namespace
