/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <chrono>
#include <cmath>

#include "base.h"
#include "dss.h"
#include "model/apartment.h"
#include "model/device.h"
#include "util.h"

DS_JSON_DEFINE_STRUCT((dss), DeviceInternalState);
DS_JSON_DEFINE_STRUCT((dss), DeviceInternalOutputState);
DS_JSON_DEFINE_STRUCT((dss), DeviceChannelValues);
DS_JSON_DEFINE_STRUCT((dss), DeviceState);
DS_JSON_DEFINE_STRUCT((dss), DeviceCacheState);

namespace dss {

static ds::LogChannel dsLogChannel("HardwareAbstraction");

int sceneFlagsFromCache(const SceneFlags& flags) {
  int deviceFlags = 0;
  if (*flags.dontCare) {
    deviceFlags |= 0x01;
  }
  if (*flags.ignoreLocalPrio) {
    deviceFlags |= 0x02;
  }
  if (*flags.specialHandling) {
    deviceFlags |= 0x04;
  }
  if (*flags.blink) {
    deviceFlags |= 0x08;
  }
  deviceFlags |= (static_cast<int>(*flags.ledConfig) << 4);
  deviceFlags |= (static_cast<int>(*flags.dimTime) << 6);
  return deviceFlags;
}

DeviceSceneSpec_t sceneSpecFromCache(const SceneFlags& flags) {
  DeviceSceneSpec_t spec;
  if (flags.dontCare) {
    spec.dontcare = *flags.dontCare;
  }
  if (flags.ignoreLocalPrio) {
    spec.localprio = *flags.ignoreLocalPrio;
  }
  if (flags.specialHandling) {
    spec.specialmode = *flags.specialHandling;
  }
  if (flags.blink) {
    spec.flashmode = *flags.blink;
  }
  if (flags.ledConfig) {
    spec.ledconIndex = static_cast<int>(*flags.ledConfig);
  }
  spec.dimtimeIndex = 0;
  return spec;
}

void sceneFlagsCacheFromDevice(int deviceFlags, SceneFlags& flags) {
  flags.dontCare = (deviceFlags & 0x01);
  flags.ignoreLocalPrio = (deviceFlags & 0x02);
  flags.specialHandling = (deviceFlags & 0x04);
  flags.blink = (deviceFlags & 0x08);
  flags.ledConfig = static_cast<LedConfig>(((deviceFlags)&0x30) >> 4);
  flags.dimTime = static_cast<DimTime>(((deviceFlags)&0xC0) >> 6);
}

void sceneFlagsCacheSceneSpec(const DeviceSceneSpec_t& spec, SceneFlags& flags) {
  flags.dontCare = spec.dontcare;
  flags.ignoreLocalPrio = spec.localprio;
  flags.specialHandling = spec.specialmode;
  flags.blink = spec.flashmode;
  flags.ledConfig = static_cast<LedConfig>(spec.ledconIndex);
  flags.dimTime = DimTime::notSupported;
}

HardwareAbstraction::HardwareAbstraction(Apartment* apt, ds::sqlite3::Database* db)
    : m_sceneCacheDb(db),
      m_stateCache(db),
      m_apartment(apt),
      m_halLight(m_sceneCacheDb, m_stateCache),
      m_halShade(m_sceneCacheDb, m_stateCache),
      m_halJoker(m_sceneCacheDb, m_stateCache) {}

HardwareAbstraction::~HardwareAbstraction() {
  // TODO Auto-generated destructor stub
}

void HardwareAbstraction::write(const Dsuid& dsuid, const std::string& parameter, double value) const {
  DS_INFO("Writing double to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.write(*device, parameter, value);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

void HardwareAbstraction::write(const Dsuid& dsuid, const std::string& parameter, bool value) const {
  DS_INFO("Writing bool to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.write(*device, parameter, value);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

void HardwareAbstraction::write(const Dsuid& dsuid, const std::string& parameter, const std::string& value) const {
  DS_INFO("Writing string to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.write(*device, parameter, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.write(*device, parameter, value);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

void HardwareAbstraction::write(const Dsuid& dsuid, const std::string& parameter, int value) const {
  return write(dsuid, parameter, static_cast<double>(value));
}

double HardwareAbstraction::readDouble(const Dsuid& dsuid, const std::string& parameter) const {
  DS_INFO("Reading double from device", dsuid, parameter);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.readDouble(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.readDouble(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.readDouble(*device, parameter);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  return 0;
}

bool HardwareAbstraction::readBool(const Dsuid& dsuid, const std::string& parameter) const {
  DS_INFO("Reading bool from device", dsuid, parameter);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.readBool(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.readBool(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.readBool(*device, parameter);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  return false;
}

std::string HardwareAbstraction::readString(const Dsuid& dsuid, const std::string& parameter) const {
  DS_INFO("Reading string from device", dsuid, parameter);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.readString(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.readString(*device, parameter);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.readString(*device, parameter);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  return "";
}

void HardwareAbstraction::updateDeviceConfig(const Dsuid& dsuid, int cls, int index, int value) const {
  DS_INFO("Updating device config", dsuid, cls, index, value);
  boost::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.updateDeviceConfig(*device, cls, index, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.updateDeviceConfig(*device, cls, index, value);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.updateDeviceConfig(*device, cls, index, value);
  }
}

void HardwareAbstraction::populateCache(const Dsuid& dsuid) const {
  DS_INFO("populateCache", dsuid);
  boost::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.populateCache(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.populateCache(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.populateCache(*device);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

DeviceCacheState HardwareAbstraction::cacheStatus(const Dsuid& dsuid) const {
  DS_INFO("cacheStatus", dsuid);
  boost::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.cacheStatus(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.cacheStatus(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.cacheStatus(*device);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

bool HardwareAbstraction::callScene(const Dsuid& dsuid, int sceneId, bool forced, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.callScene(*device, sceneId, forced, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.callScene(*device, sceneId, forced, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.callScene(*device, sceneId, forced, now);
  }
  return false; // no change without implementation
}

bool HardwareAbstraction::undoScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.undoScene(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.undoScene(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.undoScene(*device, sceneId, now);
  }
  return false; // no change without implementation
}

bool HardwareAbstraction::setOutputValue(const Dsuid& dsuid, uint8_t value, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.setOutputValue(*device, value, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.setOutputValue(*device, value, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.setOutputValue(*device, value, now);
  }
  return false; // no change without implementation
}

bool HardwareAbstraction::setOutputValue(const Dsuid& dsuid, ChannelType channel, double value,
                                         ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.setOutputValue(*device, channel, value, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.setOutputValue(*device, channel, value, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.setOutputValue(*device, channel, value, now);
  }
  return false; // no change without implementation
}

bool HardwareAbstraction::setSceneLocalPrio(const Dsuid& dsuid, int sceneId) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.setSceneLocalPrio(*device, sceneId);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.setSceneLocalPrio(*device, sceneId);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.setSceneLocalPrio(*device, sceneId);
  }
  return false;// no change without implementation
}

bool HardwareAbstraction::callSceneMin(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.callSceneMin(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.callSceneMin(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.callSceneMin(*device, sceneId, now);
  }
  return false; // no change without implementation
}

DeviceState HardwareAbstraction::getState(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.getState(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.getState(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.getState(*device);
  }
  return {}; // empty without implementation
}

void HardwareAbstraction::publishToPropertyTree(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.publishToPropertyTree(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.publishToPropertyTree(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.publishToPropertyTree(*device);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  return;
}

void HardwareAbstraction::setOperationLock(const Dsuid& dsuid, bool lock) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.setOperationLock(*device, lock);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.setOperationLock(*device, lock);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.setOperationLock(*device, lock);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
}

std::vector<int> HardwareAbstraction::areaMember(const Dsuid& dsuid) const {
  std::vector<int> ret;
  DeviceCache cacheEntry = m_sceneCacheDb.tryLookup(dsuid);
  /* for area membership we are only interested in dc flags of scenes 6-9 */
  for (int i = 6; i <= 9; i++) {
    if (cacheEntry.scene.sceneTable[i].flags.dontCare && (*cacheEntry.scene.sceneTable[i].flags.dontCare == false)) {
      ret.push_back(i);
    }
  }
  return ret;
}

bool HardwareAbstraction::deviceCacheable(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if ((device->getDeviceClass() == DEVICE_CLASS_GE && device->hasOutput()) ||
      (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) ||
      (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput())) {
    return true;
  }
  return false;
}

bool HardwareAbstraction::stopDimming(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.stopDimming(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.stopDimming(*device, sceneId, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.stopDimming(*device, sceneId, now);
  }
  return {}; // empty without implementation
}

bool HardwareAbstraction::completeMoving(const Dsuid& dsuid, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.completeMoving(*device, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.completeMoving(*device, now);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.completeMoving(*device, now);
  }
  return {}; // empty without implementation
}

std::optional<std::vector<double>> HardwareAbstraction::getDeviceThresholds(const Dsuid& dsuid) {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (device->getDeviceClass() == DEVICE_CLASS_GE) {
    return m_halLight.getDeviceThresholds(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_GR && device->hasExtendendSceneTable()) {
    return m_halShade.getDeviceThresholds(*device);
  } else if (device->getDeviceClass() == DEVICE_CLASS_SW && device->hasOutput()) {
    return m_halJoker.getDeviceThresholds(*device);
  }
  DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  return {};
}

} // namespace dss
