/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once
#include <boost/range/algorithm_ext/push_back.hpp>
#include <boost/range/irange.hpp>
#include <vector>
#include "cache/cache-db.h"
#include "cache/device-cache.h"
#include "ds-model/model.h"
#include "ds/date-time.h"
#include "model/modelconst.h"

DS_REFLECTED_ENUM((dss), DeviceDimmingStatus, (none)(dimming)(stepping));

namespace dss {

class Device;
struct DeviceSceneSpec;
typedef DeviceSceneSpec DeviceSceneSpec_t;
class DSS;
class Apartment;

int sceneFlagsFromCache(const SceneFlags& flags);
DeviceSceneSpec_t sceneSpecFromCache(const SceneFlags& flags);
void sceneFlagsCacheFromDevice(int deviceFlags, SceneFlags& flags);
void sceneFlagsCacheSceneSpec(const DeviceSceneSpec_t& spec, SceneFlags& flags);

struct DeviceInternalOutputState {
  std::optional<double> current;
  std::optional<double> target;
  std::optional<int> level;
};

struct DeviceInternalState {
  ds::DateTime lastChanged;
  std::optional<ds::DateTime> terminatesAt;
  std::map<ChannelType, DeviceInternalOutputState> channels;
  bool localPriority;
  std::optional<int> lastSceneCalled;
  std::optional<int> beforeLastSceneCalled;
  bool operationLocked;
  std::optional<int> lastDimmCommand;
  int fineDimmSteps;
  DeviceDimmingStatus dimmStatus;
  std::optional<ds::DateTime> dimmingTimeout;
};

struct DeviceChannelValues {
  double value;
  double targetValue;
  std::optional<int> level;
};

struct DeviceState {
  ds::DateTime lastChanged;
  std::optional<ds::DateTime> terminatesAt;
  dsModel::OutputStatusStatus status;
  std::map<ChannelType, DeviceChannelValues> outputs;
  bool localPriority;
};

struct DeviceCacheState {
  std::string id{};
  int total{};
  int completed{};
  double completion{};
};

class HardwareAbstraction {
private:
  class HalDevice {
  public:
    HalDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
        : m_sceneCacheDb(sceneCacheDb), m_StateCache(stateCache){};
    virtual ~HalDevice() = default;
    virtual void write(const Device& device, const std::string& parameter, double value) const;
    virtual void write(const Device& device, const std::string& parameter, bool value) const;
    virtual void write(const Device& device, const std::string& parameter, const std::string& value) const;

    virtual double readDouble(const Device& device, const std::string& parameter) const;
    virtual bool readBool(const Device& device, const std::string& parameter) const;
    virtual std::string readString(const Device& device, const std::string& parameter) const;
    virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const;
    void populateCache(const Device& device) const;

    virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const {
      DS_FAIL_REQUIRE("Not supported on this device type.");
    };
    virtual bool undoScene(const Device& device, int sceneId, ds::DateTime now) const;
    virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
      DS_FAIL_REQUIRE("Not supported on this device type.");
    };
    virtual bool setOutputValue(const Device& device, ChannelType channel, double value, ds::DateTime now) const;
    virtual bool setSceneLocalPrio(const Device& device, int sceneId) const;
    virtual bool callSceneMin(const Device& device, int sceneId, ds::DateTime now) const;
    virtual void setOperationLock(const Device& device, bool lock) const;
    virtual bool stopDimming(const Device& device, int sceneId, ds::DateTime now) const;
    virtual bool completeMoving(const Device& device, ds::DateTime now) const;

    virtual DeviceState getState(const Device& device) const;
    virtual void publishToPropertyTree(const Device& device) const;
    virtual DeviceCacheState cacheStatus(const Device& device) const;
    virtual std::optional<std::vector<double>> getDeviceThresholds(const Device& device);

  protected:
    virtual ChannelType getDefaultChannel() const = 0;
    void initializeStatus(const Device& device) const;
    bool updateCurrentValues(const Device& device, const ds::DateTime& now, dss::DeviceInternalState& state) const;
    bool isDimmingUp(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
    bool isDimmingDown(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
    bool isStop(int sceneId, bool forceLocalPrio, DeviceCache& cache) const;
    virtual void updateSwitchStatus(const Device& device, DeviceInternalState& state) const;
    virtual const std::vector<int>& getApplicableScenes() const { return m_applicableScenes; }
    virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const {
      return std::vector<ChannelType>{ChannelType::none};
    }
    virtual const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const { return {}; }
    virtual int getSceneSize() const { return 1; }
    void populateCache(const Device& device, DeviceCache& cache) const;
    virtual std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const;
    virtual std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const;
    double convertThresholdToPercentage(const double& value) const;

    CacheDb<DeviceCache>& m_sceneCacheDb;
    CacheDb<DeviceInternalState>& m_StateCache;
    std::vector<int> m_applicableScenes;
    mutable std::map<DeviceProperty, std::tuple<int, int, int>> m_deviceParameters;
  };

  class HalLightDevice : public HalDevice {
  public:
    using HalDevice::setOutputValue;
    using HalDevice::write;
    HalLightDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
        : HalDevice(sceneCacheDb, stateCache) {
      boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneDimArea));
      boost::push_back(m_applicableScenes, boost::irange(SceneMin, SceneStop));
      boost::push_back(m_applicableScenes, boost::irange(Scene2, SceneDecA1));
      boost::push_back(m_applicableScenes, boost::irange(SceneLocalOff, SceneStopA1));
      boost::push_back(m_applicableScenes, boost::irange(SceneAutoStandBy, SceneApartmentUndefined));

      m_deviceParameters = {
        {DeviceProperty::outputMode, {3, 0x00, 1}}, {DeviceProperty::switchedThreshold, {3, 0x43, 1}},
        {DeviceProperty::dimTime0Up, {3, 0x06, 1}}, {DeviceProperty::dimTime0Down, {3, 0x07, 1}},
        {DeviceProperty::dimTime1Up, {3, 0x08, 1}}, {DeviceProperty::dimTime1Down, {3, 0x09, 1}},
        {DeviceProperty::dimTime2Up, {3, 0x0A, 1}}, {DeviceProperty::dimTime2Down, {3, 0x0B, 1}},
        {DeviceProperty::dimStepUp, {3, 0x0C, 1}},  {DeviceProperty::dimStepDown, {3, 0x0D, 1}},
        {DeviceProperty::minDim, {3, 0x04, 1}}};
    }
    virtual void write(const Device& device, const std::string& parameter, double value) const override;
    virtual void write(const Device& device, const std::string& parameter, const std::string& value) const override;

    virtual double readDouble(const Device& device, const std::string& parameter) const override;
    virtual std::string readString(const Device& device, const std::string& parameter) const override;

    virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;

    virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
    virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const override;
    virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;

  protected:
    std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const override;
    std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const override;
    const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;

  private:
    void invalidateReleatedChannels(SceneEntry& sceneCache, const ChannelType channel) const;
    virtual ChannelType getDefaultChannel() const override { return ChannelType::brightness; }
    bool updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const;
    bool handleDimming(dss::DeviceInternalState& state, dss::DeviceCache& cache, const ds::DateTime& now,
                       const int dimmScene, bool up) const;
  };

  class HalShadeDevice : public HalDevice {
  public:
    using HalDevice::setOutputValue;
    using HalDevice::write;
    HalShadeDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
        : HalDevice(sceneCacheDb, stateCache) {
      boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneDimArea));
      boost::push_back(m_applicableScenes, boost::irange(SceneMin, SceneStop));
      boost::push_back(m_applicableScenes, boost::irange(Scene2, SceneDecA1));
      boost::push_back(m_applicableScenes, boost::irange(SceneLocalOff, SceneStopA1));
      boost::push_back(m_applicableScenes, boost::irange(SceneSunProtection, SceneZoneUndefined));
      boost::push_back(m_applicableScenes, boost::irange(SceneAutoStandBy, SceneApartmentUndefined));
      
      m_deviceParameters = {
        {DeviceProperty::dimStepNum, {3, 0x3A, 1}},
        {DeviceProperty::dimStepLen, {3, 0x3B, 1}},
        {DeviceProperty::dimStepPause, {3, 0x3C, 1}},
        {DeviceProperty::posTimeUp, {3, 0x30, 2}},
        {DeviceProperty::posTimeDown, {3, 0x32, 2}}};
    }
    virtual void write(const Device& device, const std::string& parameter, double value) const override;

    virtual double readDouble(const Device& device, const std::string& parameter) const override;

    virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;

    virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
    virtual bool setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const override;
    virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;
    virtual bool setOutputValue(const Device& device, ChannelType channel, double value,
                                ds::DateTime now) const override;

  private:
    virtual const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;
    void updateSceneConfig(const Device& device, int scene, dss::ChannelType channel, int value,
                           bool firstRegister) const;
    virtual ChannelType getDefaultChannel() const override { return ChannelType::shadePositionOutside; }

    bool handleDimming(int sceneId, DeviceCache& cache, const ds::DateTime& now, DeviceInternalState& state,
                       DeviceInternalOutputState& outputState, bool up) const;
    double calculateRuntime(DeviceCache& cache, DeviceInternalState& state) const;
    int getSceneSize() const override { return 2; }
  };

  class HalJokerDevice : public HalLightDevice {
  public:
    using HalDevice::setOutputValue;
    using HalDevice::write;
    HalJokerDevice(CacheDb<DeviceCache>& sceneCacheDb, CacheDb<DeviceInternalState>& stateCache)
        : HalLightDevice(sceneCacheDb, stateCache) {
      boost::push_back(m_applicableScenes, boost::irange(SceneOff, SceneBurglary));
      
      m_deviceParameters = {
        {DeviceProperty::outputMode, {3, 0x00, 1}}, {DeviceProperty::switchedThreshold, {3, 0x3C, 1}},
        {DeviceProperty::dimTime0Up, {3, 0x06, 1}}, {DeviceProperty::dimTime0Down, {3, 0x07, 1}},
        {DeviceProperty::dimTime1Up, {3, 0x08, 1}}, {DeviceProperty::dimTime1Down, {3, 0x09, 1}},
        {DeviceProperty::dimTime2Up, {3, 0x0A, 1}}, {DeviceProperty::dimTime2Down, {3, 0x0B, 1}},
        {DeviceProperty::dimStepUp, {3, 0x0C, 1}},  {DeviceProperty::dimStepDown, {3, 0x0D, 1}},
        {DeviceProperty::minDim, {3, 0x04, 1}}};
    }
    virtual void write(const Device& device, const std::string& parameter, double value) const override;
    virtual double readDouble(const Device& device, const std::string& parameter) const override;
    virtual bool callScene(const Device& device, int sceneId, bool forced, ds::DateTime now) const override;
    virtual const std::vector<ChannelType> getSceneChannels(const Device& device) const override;
    virtual DeviceState getState(const Device& device) const override;
    virtual void updateDeviceConfig(const Device& device, int bank, int offset, int value) const override;
    virtual std::optional<std::vector<double>> getDeviceThresholds(const Device& device) override;

  protected:
    std::map<DeviceProperty, std::tuple<int, int, int>> getDeviceParameters(const Device& device) const override;
    std::tuple<int, int, int> getDeviceParameter(const Device& device, DeviceProperty property) const override;
    void updateSwitchStatus(const Device& device, DeviceInternalState& state) const override;
    const std::vector<DeviceProperty> getDeviceProperties(const Device& device) const override;

  private:
    void updateSceneConfig(const Device& device, int scene, dss::ChannelType channel, int value,
                           bool firstRegister) const;
    virtual ChannelType getDefaultChannel() const override { return ChannelType::powerState; }

    bool isIgnoredInPowersave(int sceneId) const;
    bool isIgnoredInWipe(int sceneId) const;
    bool updateOutputValue(const Device& device, uint8_t value, ds::DateTime now) const;
  };

public:
  HardwareAbstraction(Apartment* apt, ds::sqlite3::Database* db);
  virtual ~HardwareAbstraction();
  void write(const Dsuid& dsuid, const std::string& parameter, double value) const;
  void write(const Dsuid& dsuid, const std::string& parameter, bool value) const;
  void write(const Dsuid& dsuid, const std::string& parameter, int value) const;
  void write(const Dsuid& dsuid, const std::string& parameter, const std::string& value) const;

  virtual double readDouble(const Dsuid& dsuid, const std::string& parameter) const;
  virtual bool readBool(const Dsuid& dsuid, const std::string& parameter) const;
  virtual std::string readString(const Dsuid& dsuid, const std::string& parameter) const;

  virtual bool callScene(const Dsuid& dsuid, int sceneId, bool forced, ds::DateTime now) const;
  virtual bool undoScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const;
  virtual bool setOutputValue(const Dsuid& dsuid, uint8_t value, ds::DateTime now) const;
  virtual bool setOutputValue(const Dsuid& dsuid, ChannelType channel, double value, ds::DateTime now) const;
  virtual bool setSceneLocalPrio(const Dsuid& dsuid, int sceneId) const;
  virtual bool callSceneMin(const Dsuid& dsuid, int sceneId, ds::DateTime now) const;
  virtual void setOperationLock(const Dsuid& device, bool lock) const;
  virtual bool stopDimming(const Dsuid& dsuid, int sceneId, ds::DateTime now) const;
  virtual bool completeMoving(const Dsuid& dsuid, ds::DateTime now) const;

  virtual DeviceState getState(const Dsuid& dsuid) const;
  void publishToPropertyTree(const Dsuid& dsuid) const;

  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const;
  void populateCache(const Dsuid& dsuid) const;
  virtual DeviceCacheState cacheStatus(const Dsuid& dsuid) const;

  CacheDb<DeviceCache>& getCache() const { return m_sceneCacheDb; }
  CacheDb<DeviceInternalState>& getStateCache() const { return m_stateCache; }

  virtual std::vector<int> areaMember(const Dsuid& dsuid) const;

  bool deviceCacheable(const Dsuid& dsuid) const;

  virtual std::optional<std::vector<double>> getDeviceThresholds(const Dsuid& dsuid);

private:
  mutable CacheDb<DeviceCache> m_sceneCacheDb;
  mutable CacheDb<DeviceInternalState> m_stateCache;
  Apartment* m_apartment;
  HalLightDevice m_halLight;
  HalShadeDevice m_halShade;
  HalJokerDevice m_halJoker;
};

} /* namespace dss */

DS_JSON_DECLARE_STRUCT((dss), DeviceInternalState,
                       (lastChanged)(channels)(terminatesAt)(localPriority)(lastSceneCalled)(beforeLastSceneCalled)(
                           operationLocked)(lastDimmCommand)(fineDimmSteps)(dimmStatus)(dimmingTimeout));
DS_JSON_DECLARE_STRUCT((dss), DeviceInternalOutputState, (current)(target)(level));
DS_JSON_DECLARE_STRUCT((dss), DeviceChannelValues, (value)(targetValue)(level));
DS_JSON_DECLARE_STRUCT((dss), DeviceState, (lastChanged)(terminatesAt)(status)(outputs)(localPriority));
DS_JSON_DECLARE_STRUCT((dss), DeviceCacheState, (id)(total)(completed)(completion));
