/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <iostream>
#include <ostream>
#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/device-cache.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/modelmaintenance-mock.h"
#include "util.h"

static ds::LogChannel dsLogChannel("halJokerStateMachine", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][halJokerStateMachine]";

using namespace dss;
using namespace trompeloeil;
using Catch::Generators::values;
using trompeloeil::_;

namespace {

struct Area {
  int area;
  int inc;
  int dec;
};

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};

SCENARIO("Standard Joker device", TAGS) {
  Apartment apt(nullptr);
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  DeviceSpec spec;
  spec.FunctionID = 0x8010;

  boost::shared_ptr<Device> dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->setHardwareType(HardwareType::SW_KL200);
  dev1->setOutputChannels({ChannelType::powerState});

  ds::sqlite3::Database d(":memory:");
  HardwareAbstraction hal(&apt, &d);
  hal.getCache().empty();

  auto& jokerState = hal.getStateCache();
  jokerState.empty();
  ds::DateTime now = std::chrono::high_resolution_clock::now();

  GIVEN("the cache is empty") {
    WHEN("a scene is called") {
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == false);
      THEN("no action is taken") {}
    }
  }

  GIVEN("the cache is populated") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.2);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.property[DeviceProperty::minDim] = 10;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::powerState].value = 0.5;
      cache.scene.sceneTable[Scene4].channels[ChannelType::powerState].value = 0;

      cache.scene.sceneTable[SceneAutoOff].flags.dontCare = false;
      cache.scene.sceneTable[SceneAutoOff].flags.dimTime = DimTime::dimTime1;
      cache.scene.sceneTable[SceneAutoOff].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneAutoOff].channels[ChannelType::powerState].value = 0.0;
      hal.getCache().update(dsuid1, cache);
    }

    WHEN("No state is known") {
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        DS_DEBUG("Just debugging here");
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(!state.channels[ChannelType::powerState].target);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
        REQUIRE(!state.channels[ChannelType::powerState].target);
      }
      THEN("DEC do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
        REQUIRE(!state.channels[ChannelType::powerState].target);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 0.5;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        jokerState.update(dsuid1, state);
      }
      THEN("after callScene state is known") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.lastChanged > state.lastChanged.min());
      }
      THEN("INC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.77);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.77);
        REQUIRE(!ret.terminatesAt);
        REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(320)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.77);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.77);
        REQUIRE(!ret.terminatesAt);
      }
      THEN("INC does not adjust the current value if the last called scene is unknown") {
        {
          auto state = jokerState.tryLookup(dsuid1);
          state.lastSceneCalled = Scene3;
          jokerState.update(dsuid1, state);
        }
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0.5);
      }
      THEN("INC does not adjust the current value if the last called scene is 0") {
        {
          auto state = jokerState.tryLookup(dsuid1);
          state.lastSceneCalled = Scene4;
          jokerState.update(dsuid1, state);
        }
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0.5);
      }
      THEN("INC will not exceed max value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(7 + 3);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(3 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(4 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(5 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(6 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(7 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(8 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(9 * 250)) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(10 * 250 + 250)) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 2.0);
      }
      THEN("DEC adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneDec, now + std::chrono::milliseconds(320)) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.3);
      }
      THEN("DEC will not reach 0") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(4 + 1);
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneDec, now + std::chrono::milliseconds(320)) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.3);

        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(3 * 250)) == true);
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(4 * 250)) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneDec, now + std::chrono::milliseconds(4 * 250 + 320)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) > 0);
      }
      THEN("callSceneMin does not change the output") {
        REQUIRE(hal.callSceneMin(dsuid1, Scene1, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0.5);
      }
      THEN("after callScene(AUTO_OFF) state is moving") {
        ALLOW_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneAutoOff, false, now) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::moving);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.5);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0);
        REQUIRE(hal.completeMoving(dsuid1, now + std::chrono::milliseconds(60900)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.0);
      }
    }
    WHEN("Current device state is zero") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 0.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        jokerState.update(dsuid1, state);
      }
      THEN("INC does not change the value") {
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0);
        REQUIRE(!state.terminatesAt);
      }
      THEN("DEC not change the value") {
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0);
        REQUIRE(!state.terminatesAt);
      }
      THEN("INC changes the value after callSceneMin") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(2);
        REQUIRE(hal.callSceneMin(dsuid1, Scene1, now) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.27);

        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now + std::chrono::milliseconds(1 * 250)) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(1 * 250 + 320)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.55);
      }
    }
    WHEN("Current device state is max") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 2.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        jokerState.update(dsuid1, state);
      }
      THEN("INC does not change the value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(320)) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 2.0);
      }
      THEN("DEC changes the value") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
        REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(320)) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 1.8);
      }
    }
    WHEN("Area dimming is set") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 0.5;
        state.lastDimmCommand = SceneDecA4;
        jokerState.update(dsuid1, state);
      }
      THEN("A call scene resets the lastDimmCommand") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.lastDimmCommand);
      }
    }
  }

  GIVEN("the cache is populated for undo scene") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(0.2);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1;

      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::powerState].value = 0.5;

      cache.scene.sceneTable[Scene3].flags.dontCare = false;
      cache.scene.sceneTable[Scene3].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene3].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene3].channels[ChannelType::powerState].value = 1.5;
      hal.getCache().update(dsuid1, cache);
    }

    WHEN("No state is known") {
      THEN("after undoScene does not make the state defined") {
        REQUIRE(hal.undoScene(dsuid1, -1, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
      }
      THEN("after undoSceneNr does not make the state defined") {
        REQUIRE(hal.undoScene(dsuid1, Scene1, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 2.0;
        state.lastSceneCalled = Scene2;
        state.beforeLastSceneCalled = Scene3;
        jokerState.update(dsuid1, state);
      }
      THEN("after undoScene previous scene is called") {
        REQUIRE(hal.undoScene(dsuid1, -1, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.5);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with matching sceneId previous scene is called") {
        REQUIRE(hal.undoScene(dsuid1, Scene2, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.5);
        REQUIRE(*state.lastSceneCalled == Scene3);
        REQUIRE(*state.beforeLastSceneCalled == Scene2);
      }
      THEN("after undoSceneNr with not matching sceneId previous scene is not called") {
        REQUIRE(hal.undoScene(dsuid1, Scene44, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 2.0);
        REQUIRE(*state.lastSceneCalled == Scene2);
        REQUIRE(*state.beforeLastSceneCalled == Scene3);
      }
    }
  }

  GIVEN("Area dimming") {
    auto area = GENERATE(values<Area>({{SceneA11, SceneIncA1, SceneDecA1},
                                       {SceneA21, SceneIncA2, SceneDecA2},
                                       {SceneA31, SceneIncA3, SceneDecA3},
                                       {SceneA41, SceneIncA4, SceneDecA4}}));
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::dimStepDown] = 25;
      cache.property[DeviceProperty::dimStepUp] = 35;
      cache.property[DeviceProperty::minDim] = 10;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1.0;

      cache.scene.sceneTable[area.area].flags.dontCare = false;
      cache.scene.sceneTable[area.area].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[area.area].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[area.area].channels[ChannelType::powerState].value = 1.0;
      hal.getCache().update(dsuid1, cache);
    }

    WHEN("No state is known") {
      THEN("INC " << area.inc << " do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, area.inc, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
      }
      THEN("DEC " << area.dec << " do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, area.dec, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
      }
      THEN("CONT do not make the state defined") {
        REQUIRE(hal.callScene(dsuid1, SceneDimArea, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(!state.channels[ChannelType::powerState].current);
      }
    }
    WHEN("Current device state is known") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 0.5;
        state.lastSceneCalled = Scene1;
        jokerState.update(dsuid1, state);
      }
      THEN("INC " << area.inc << " adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(3);
        REQUIRE(hal.callScene(dsuid1, area.inc, false, now) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.77);

        REQUIRE(hal.callScene(dsuid1, SceneDimArea, false, now + std::chrono::milliseconds(1 * 250)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 1.05);

        REQUIRE(hal.callScene(dsuid1, SceneDimArea, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.stopDimming(dsuid1, area.inc, now + std::chrono::milliseconds(2 * 250 + 320)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 1.32);
      }
      THEN("DEC " << area.dec << " adjust the current value") {
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(3);
        REQUIRE(hal.callScene(dsuid1, area.dec, false, now) == true);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.3);

        REQUIRE(hal.callScene(dsuid1, SceneDimArea, false, now + std::chrono::milliseconds(1 * 250)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].targetValue, 2) == 0.11);

        REQUIRE(hal.callScene(dsuid1, SceneDimArea, false, now + std::chrono::milliseconds(2 * 250)) == true);
        REQUIRE(hal.stopDimming(dsuid1, area.dec, now + std::chrono::milliseconds(2 * 250 + 320)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.08);
      }
    }
  }

  GIVEN("Device is not locally prioritized") {
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.5;
      state.localPriority = false;
      jokerState.update(dsuid1, state);
    }
    WHEN("A localOn scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dsuid1);
        cache.scene.sceneTable[SceneLocalOn].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOn].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOn].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOn].flags.specialHandling = true;
        cache.scene.sceneTable[SceneLocalOn].channels[ChannelType::powerState].value = 1.0;
        hal.getCache().update(dsuid1, cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(dsuid1, SceneLocalOn, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.localPriority == true);
      }
    }
  }

  GIVEN("Device is locally prioritized") {
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.5;
      state.localPriority = true;
      state.lastSceneCalled = Scene1;
      jokerState.update(dsuid1, state);
    }
    WHEN("A normal scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dsuid1);
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1.0;
        hal.getCache().update(dsuid1, cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.5);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(dsuid1, Scene1, true, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A dimming scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dsuid1);
        cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1.0;
        hal.getCache().update(dsuid1, cache);
      }
      THEN("Without force, it does not change") {
        REQUIRE(hal.callScene(dsuid1, SceneInc, false, now) == false);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.5);
        REQUIRE(state.localPriority == true);
      }
      THEN("With force, it does change") {
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE(hal.callScene(dsuid1, SceneInc, true, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current > 1.5);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("A scene with ignoreLocalProp is called") {
      {
        auto cache = hal.getCache().tryLookup(dsuid1);
        cache.scene.sceneTable[Scene1].flags.dontCare = false;
        cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = true;
        cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 1.0;
        hal.getCache().update(dsuid1, cache);
      }
      THEN("Without force, it does change") {
        REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.localPriority == false);
      }
      THEN("With force, it does change") {
        REQUIRE(hal.callScene(dsuid1, Scene1, true, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.localPriority == false);
      }
    }
    WHEN("A local off scene is called") {
      {
        auto cache = hal.getCache().tryLookup(dsuid1);
        cache.scene.sceneTable[SceneLocalOff].flags.dontCare = false;
        cache.scene.sceneTable[SceneLocalOff].flags.dimTime = DimTime::noDimTime;
        cache.scene.sceneTable[SceneLocalOff].flags.ignoreLocalPrio = false;
        cache.scene.sceneTable[SceneLocalOff].channels[ChannelType::powerState].value = 1.0;
        hal.getCache().update(dsuid1, cache);
      }
      THEN("Without force, it becomes prioritized") {
        REQUIRE(hal.callScene(dsuid1, SceneLocalOff, false, now) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A scene with a long transition time") {
    DeviceProperty down;
    DeviceProperty up;
    DimTime dim;
    std::tie(down, up, dim) = GENERATE(values<std::tuple<DeviceProperty, DeviceProperty, DimTime>>(
        {{DeviceProperty::dimTime0Down, DeviceProperty::dimTime0Up, DimTime::dimTime0},
         {DeviceProperty::dimTime1Down, DeviceProperty::dimTime1Up, DimTime::dimTime1},
         {DeviceProperty::dimTime2Down, DeviceProperty::dimTime2Up, DimTime::dimTime2}}));
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[down] = convertDimTimeToDevice(10);
      cache.property[up] = convertDimTimeToDevice(10);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = dim;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 2.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.0;
      state.localPriority = false;
      jokerState.update(dsuid1, state);
    }
    WHEN("After the scene is called") {
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
      THEN("termination time is initially set to lastChanged") {
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(state.channels[ChannelType::powerState].target);
        REQUIRE(*state.channels[ChannelType::powerState].target == 2.0);
        REQUIRE(state.lastChanged > state.lastChanged.min());
        REQUIRE(
            std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() ==
            10000);
      }
      THEN("current value changes with subsequent calls to getStatus") {
        auto state = jokerState.tryLookup(dsuid1);
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::moving);
        REQUIRE(ret.outputs[ChannelType::powerState].targetValue == 2.0);
        REQUIRE(ret.outputs[ChannelType::powerState].value == 1.0);

        REQUIRE(hal.completeMoving(dsuid1, now + std::chrono::seconds(10)) == true);
        ret = hal.getState(dsuid1);
        REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
        REQUIRE(ret.outputs[ChannelType::powerState].targetValue == 2.0);
        REQUIRE(ret.outputs[ChannelType::powerState].value == 2.0);
      }
    }
  }

  GIVEN("Stop handling") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::dimTime0Down] = convertDimTimeToDevice(10);
      cache.property[DeviceProperty::dimTime0Up] = convertDimTimeToDevice(10);
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::dimTime0;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 2.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 0.0;
      state.localPriority = false;
      jokerState.update(dsuid1, state);
    }
    WHEN("After a scene is called") {
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
      auto state = jokerState.tryLookup(dsuid1);
      auto origCallTime = state.lastChanged;
      THEN("the stop scene halts the value before it reaches the target") {
        REQUIRE(hal.callScene(dsuid1, SceneStop, false, origCallTime + std::chrono::milliseconds(5000)) == true);
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(!state.channels[ChannelType::powerState].target);
        REQUIRE(
            roundDigits(*state.channels[ChannelType::powerState].current, 2) * 100 ==
            roundDigits(
                std::chrono::duration_cast<std::chrono::milliseconds>(state.lastChanged - origCallTime).count() / 50.0,
                2));
      }
    }
  }

  GIVEN("A scene with dontCare") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.scene.sceneTable[Scene1].flags.dontCare = true;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 100.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.0;
      state.lastSceneCalled = 13;
      state.localPriority = false;
      jokerState.update(dsuid1, state);
    }
    WHEN("After the scene is called") {
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == false);
      THEN("state has not changed") {
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(*state.channels[ChannelType::powerState].current == 1.0);
        REQUIRE(!state.channels[ChannelType::powerState].target);
        REQUIRE(*state.lastSceneCalled == 13);
      }
    }
  }

  GIVEN("A setOutputValue call") {
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.0;
      state.channels[ChannelType::powerState].target = 1.1;
      state.lastSceneCalled = 13;
      state.localPriority = true;
      jokerState.update(dsuid1, state);
    }
    WHEN("After setOutputValue is called") {
      REQUIRE(hal.setOutputValue(dsuid1, 245, now) == true);
      THEN("state is changed") {
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.channels[ChannelType::powerState].current);
        REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 1.92);
        REQUIRE(!state.channels[ChannelType::powerState].target);
        REQUIRE(!state.lastSceneCalled);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("A setLocalPrioCall") {
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.0;
      state.channels[ChannelType::powerState].target = 1.1;
      state.lastSceneCalled = Scene42;
      state.localPriority = false;
      jokerState.update(dsuid1, state);
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(dsuid1, Scene42));
      THEN("state is changed") {
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.localPriority == true);
      }
    }
    WHEN("After setLocalPrioCall is called with matching sceneId") {
      REQUIRE_NOTHROW(hal.setSceneLocalPrio(dsuid1, Scene43));
      THEN("state is changed") {
        auto state = jokerState.tryLookup(dsuid1);
        REQUIRE(state.localPriority == false);
      }
    }
  }

  GIVEN("The device is set to switched mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KM_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 2.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::powerState].value = 0.5;
      hal.getCache().update(dsuid1, cache);
    }
    WHEN("after callScene with value above the threshold") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 0.5;
        state.channels[ChannelType::powerState].level = 0;
        state.lastSceneCalled = 13;
        jokerState.update(dsuid1, state);
      }
      REQUIRE(hal.callScene(dsuid1, Scene1, false, now) == true);
      THEN("the output is on (level1)") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::powerState].value == 2.0);
        REQUIRE(ret.outputs[ChannelType::powerState].level == 1);
      }
    }
    WHEN("after callScene with value below the threshold") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 1.5;
        state.channels[ChannelType::powerState].level = 1;
        state.lastSceneCalled = 13;
        jokerState.update(dsuid1, state);
      }
      REQUIRE(hal.callScene(dsuid1, Scene2, false, now) == true);
      THEN("the output is off (level0)") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::powerState].value == 0.5);
        REQUIRE(ret.outputs[ChannelType::powerState].level == 0);
      }
    }
    WHEN("device is reconfigured to dimmed mode") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 1.5;
        state.channels[ChannelType::powerState].level = 1;
        state.lastSceneCalled = 13;
        jokerState.update(dsuid1, state);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 3, 0, 22));
      THEN("the output level is reset") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(!ret.outputs[ChannelType::powerState].level);
      }
    }
    WHEN("device switch threshold is reconfigured") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 1.5;
        state.channels[ChannelType::powerState].level = 1;
        state.lastSceneCalled = 13;
        jokerState.update(dsuid1, state);
      }
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 3, 0x3C, 205));
      THEN("the output level is reset") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::powerState].level == 0);
      }
    }
    THEN("DEC across the threshold will change level") {
      {
        auto state = jokerState.tryLookup(dsuid1);
        state.channels[ChannelType::powerState].current = 2.0;
        state.channels[ChannelType::powerState].level = 1;
        state.lastSceneCalled = Scene1;
        jokerState.update(dsuid1, state);
      }
      REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(15);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now) == true);
      auto ret = hal.getState(dsuid1);
      REQUIRE(ret.status == dsModel::OutputStatusStatus::dimming);
      REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 1.91);
      REQUIRE(ret.outputs[ChannelType::powerState].level == 1);

      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(1 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(2 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(3 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(4 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(5 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(6 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(7 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(8 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(9 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(10 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(11 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(12 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(13 * 250)) == true);
      REQUIRE(hal.callScene(dsuid1, SceneDec, false, now + std::chrono::milliseconds(14 * 250)) == true);
      REQUIRE(hal.stopDimming(dsuid1, SceneInc, now + std::chrono::milliseconds(14 * 250 + 320)) == true);
      ret = hal.getState(dsuid1);
      REQUIRE(ret.status == dsModel::OutputStatusStatus::ok);
      REQUIRE(roundDigits(ret.outputs[ChannelType::powerState].value, 2) == 0.71);
      REQUIRE(ret.outputs[ChannelType::powerState].level == 0);
    }
  }

  GIVEN("The device is set to dimmed mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::TrailingEdgeDimmerCurve);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[Scene1].flags.dontCare = false;
      cache.scene.sceneTable[Scene1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene1].channels[ChannelType::powerState].value = 2.0;
      cache.scene.sceneTable[Scene2].flags.dontCare = false;
      cache.scene.sceneTable[Scene2].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[Scene2].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[Scene2].channels[ChannelType::powerState].value = 0.5;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 1.5;
      state.lastSceneCalled = 13;
      jokerState.update(dsuid1, state);
    }
    WHEN("device is reconfigured to switched mode") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 3, 0, 16));
      THEN("the output level is correct") {
        auto ret = hal.getState(dsuid1);
        REQUIRE(ret.outputs[ChannelType::powerState].level == 1);
      }
    }
  }

  GIVEN("a joker device is controlled by writing to the runtime bank") {
    THEN("writing the output value will change it") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 64, 0x00, 0x34));
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 0.41);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid1);
    }
  }

  GIVEN("a joker device is switched and controlled by writing to the runtime bank") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KL_Switch);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 0;
      state.channels[ChannelType::powerState].level = 0;
      state.lastSceneCalled = 0;
      jokerState.update(dsuid1, state);
    }
    THEN("writing the output value will change it and the switched state") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid1, 64, 0x00, 0xac));
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 1.35);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid1);
      REQUIRE(ret.outputs[ChannelType::powerState].level == 1);
    }
  }

  GIVEN("a joker device in wipe output mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KL_Passing);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[SceneOff].flags.dontCare = false;
      cache.scene.sceneTable[SceneOff].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[SceneOff].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneOff].channels[ChannelType::powerState].value = 0.0;
      cache.scene.sceneTable[SceneOffE1].flags.dontCare = false;
      cache.scene.sceneTable[SceneOffE1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[SceneOffE1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneOffE1].channels[ChannelType::powerState].value = 0.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 2.0;
      state.lastSceneCalled = 13;
      jokerState.update(dsuid1, state);
    }
    THEN("With scene Off, it does not change") {
      REQUIRE(hal.callScene(dsuid1, SceneOff, true, now) == false);
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(*state.channels[ChannelType::powerState].current == 2.0);
    }
    THEN("With scene Extended Off, it does not change") {
      REQUIRE(hal.callScene(dsuid1, SceneOff, true, now) == false);
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(*state.channels[ChannelType::powerState].current == 2.0);
    }
  }

  GIVEN("a joker device in powersave output mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid1);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::KL_PowerSave);
      cache.property[DeviceProperty::switchedThreshold] = 128;
      cache.scene.sceneTable[SceneOff].flags.dontCare = false;
      cache.scene.sceneTable[SceneOff].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[SceneOff].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneOff].channels[ChannelType::powerState].value = 0.0;
      cache.scene.sceneTable[SceneOffE1].flags.dontCare = false;
      cache.scene.sceneTable[SceneOffE1].flags.dimTime = DimTime::noDimTime;
      cache.scene.sceneTable[SceneOffE1].flags.ignoreLocalPrio = false;
      cache.scene.sceneTable[SceneOffE1].channels[ChannelType::powerState].value = 0.0;
      hal.getCache().update(dsuid1, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid1);
      state.channels[ChannelType::powerState].current = 2.0;
      state.lastSceneCalled = 13;
      jokerState.update(dsuid1, state);
    }
    THEN("With scene Off, it does not change") {
      REQUIRE(hal.callScene(dsuid1, SceneOff, true, now) == false);
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(*state.channels[ChannelType::powerState].current == 2.0);
    }
    THEN("With scene Extended Off, it does not change") {
      REQUIRE(hal.callScene(dsuid1, SceneOff, true, now) == false);
      auto state = jokerState.tryLookup(dsuid1);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(*state.channels[ChannelType::powerState].current == 2.0);
    }
  }
}

SCENARIO("UMR200 Joker device", TAGS) {
  Apartment apt(nullptr);
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  DeviceSpec spec;
  spec.FunctionID = 0x8010;

  boost::shared_ptr<Device> dev1 = apt.allocateDevice(dsuid2);
  dev1->setPartiallyFromSpec(spec);
  dev1->setHardwareType(HardwareType::SW_UMR200);
  dev1->setOutputChannels({ChannelType::powerState});

  ds::sqlite3::Database d(":memory:");
  HardwareAbstraction hal(&apt, &d);
  hal.getCache().empty();

  auto& jokerState = hal.getStateCache();
  jokerState.empty();

  GIVEN("a UMR200 in two stage switched mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid2);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::TwoStageSwitch);
      cache.property[DeviceProperty::switchedThresholdTwoStep0] = 85;
      cache.property[DeviceProperty::switchedThresholdTwoStep1] = 170;
      hal.getCache().update(dsuid2, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid2);
      state.channels[ChannelType::powerState].current = 0;
      state.channels[ChannelType::powerState].level = 0;
      state.lastSceneCalled = 0;
      jokerState.update(dsuid2, state);
    }
    THEN("writing the output value will change it and the switched state") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid2, 64, 0x00, 0xac));
      auto state = jokerState.tryLookup(dsuid2);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 1.35);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid2);
      REQUIRE(ret.outputs[ChannelType::powerState].level == 2);
    }
  }
  GIVEN("a UMR200 in three stage switched mode") {
    {
      auto cache = hal.getCache().tryLookup(dsuid2);
      cache.property[DeviceProperty::outputMode] = static_cast<int>(OutputMode::ThreeStageSwitch);
      cache.property[DeviceProperty::switchedThreshold3Step0] = 64;
      cache.property[DeviceProperty::switchedThreshold3Step1] = 128;
      cache.property[DeviceProperty::switchedThreshold3Step2] = 192;
      hal.getCache().update(dsuid2, cache);
    }
    {
      auto state = jokerState.tryLookup(dsuid2);
      state.channels[ChannelType::powerState].current = 0;
      state.channels[ChannelType::powerState].level = 0;
      state.lastSceneCalled = 0;
      jokerState.update(dsuid2, state);
    }
    THEN("writing the output value will change it and the switched state") {
      REQUIRE_NOTHROW(hal.updateDeviceConfig(dsuid2, 64, 0x00, 0xc3));
      auto state = jokerState.tryLookup(dsuid2);
      REQUIRE(state.channels[ChannelType::powerState].current);
      REQUIRE(roundDigits(*state.channels[ChannelType::powerState].current, 2) == 1.53);
      REQUIRE(!state.terminatesAt);
      auto ret = hal.getState(dsuid2);
      REQUIRE(ret.outputs[ChannelType::powerState].level == 3);
    }
  }
}

} // namespace
