/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <boost/smart_ptr/make_shared.hpp>
#include <chrono>
#include <cmath>

#include "base.h"
#include "ds-api/route-segments.h"
#include "dss.h"
#include "event.h"
#include "model/apartment.h"
#include "model/device.h"
#include "util.h"

namespace dss {

static ds::LogChannel dsLogChannel("shadeDevice");

static const int HalfWavePerSecond = 100;
static const int MillisecondsPerHalfWave = 10;
static constexpr int DefaultPositionTime = 120 * HalfWavePerSecond;
static constexpr int DefaultDimStepLen = 0.2 * HalfWavePerSecond;
static constexpr int DefaultDimStepPause = 0.81 * HalfWavePerSecond;
static constexpr int DefaultDimStepNum = 4;
static constexpr int DefaultIncInterval = 250;

void HardwareAbstraction::HalShadeDevice::write(const Device& device, const std::string& parameter,
                                                double value) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontCare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontCare" -> 128, 135 (=128+7)
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "shadePositionOutside") {
      auto channelSpec = getOutputChannelIdAndSize("shadePositionOutside");
      auto newValue = convertToOutputChannelValue(channelSpec.first, value);
      auto lowValue = device.getDeviceConfig(CfgClassSceneExtention, sceneNr / 2);
      if ((sceneNr % 2) == 0) {
        lowValue &= 0xF0;
        lowValue |= (newValue >> 4) & 0x0F;
      } else {
        lowValue &= 0x0F;
        lowValue |= newValue & 0xF0;
      }
      device.setDeviceConfig(CfgClassScene, sceneNr, (newValue >> 8) & 0xFF, false);
      device.setDeviceConfig(CfgClassSceneExtention, sceneNr / 2, lowValue, false);
      SceneChannel& shadePositionOutside = sceneCache.channels[ChannelType::shadePositionOutside];
      shadePositionOutside.value = value;
      sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else if (param == "shadeOpeningAngleOutside") {
      DeviceFeatures_t features = device.getDeviceFeatures();
      DS_REQUIRE(features.hasOutputAngle, "Device dows not have channel", param);
      auto channelSpec = getOutputChannelIdAndSize("shadeOpeningAngleOutside");
      auto newValue = convertToOutputChannelValue(channelSpec.first, value);
      device.setDeviceConfig(CfgClassSceneAngle, sceneNr, newValue, false);
      SceneChannel& shadeOpeningAngleOutside = sceneCache.channels[ChannelType::shadeOpeningAngleOutside];
      shadeOpeningAngleOutside.value = value;
      sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else {
      DS_REQUIRE(false, "Unknown channel requested", param);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property) {
      unsigned int intValueLow = 0;
      std::optional<unsigned int> intValueHigh;
      if (*property == DeviceProperty::dimStepNum) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 255, "Value too high");
        intValueLow = value;
      } else if (*property == DeviceProperty::dimStepLen || *property == DeviceProperty::dimStepPause) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 2.55, "Value too high");
        intValueLow = value / 0.01;
      } else if (*property == DeviceProperty::posTimeUp || *property == DeviceProperty::posTimeDown) {
        DS_REQUIRE(value >= 0, "Value too low");
        DS_REQUIRE(value <= 655.35, "Value too high");
        unsigned int tempValue = value / 0.01;
        intValueLow = tempValue & 0xFF;
        intValueHigh = (tempValue >> 8) & 0xFF;
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      device.setDeviceConfig(bank, offset, intValueLow, false);
      devCache.property[*property] = intValueLow;
      if (intValueHigh) {
        device.setDeviceConfig(bank, offset + 1, *intValueHigh, false);
        devCache.property[*property] = *intValueHigh << 8 | intValueLow;
      }
      devCache.lastChanged = std::chrono::high_resolution_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
}

double HardwareAbstraction::HalShadeDevice::readDouble(const Device& device, const std::string& parameter) const {
  // "/scene/5/brightness" -> 128, 5
  // "/scene/6/dontcare"   -> 128, 134 (=128+6)
  // "/scene/7/brightness/dontcare" -> 128, 135 (=128+7)
  double returnValue = 0;
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (auto route = dsApi::matchRoute("/scene/{sceneNr}/{channelOrFlag}", parameter)) {
    auto& scene = route.args["sceneNr"];
    auto& param = route.args["channelOrFlag"];
    unsigned int sceneNr = strToUInt(scene);
    DS_REQUIRE(sceneNr <= MaxSceneNumber, "Scene number is out of range.", sceneNr);
    SceneEntry& sceneCache = devCache.scene.sceneTable[sceneNr];
    if (param == "shadePositionOutside") {
      auto value = device.getDeviceConfig(128, sceneNr) << 8;
      auto extValue = device.getDeviceConfig(CfgClassSceneExtention, sceneNr / 2);
      if ((sceneNr % 2) == 1) {
        extValue &= 0xF0; // high bits for odd scene numbers
        extValue |= (extValue >> 4);
      } else {
        extValue &= 0x0F; // lower bits for even scene numbers
        extValue |= (extValue << 4);
      }
      value |= extValue;
      SceneChannel& shadePositionOutside = sceneCache.channels[ChannelType::shadePositionOutside];
      auto channelSpec = getOutputChannelIdAndSize("shadePositionOutside");
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      shadePositionOutside.value = returnValue;
      sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else if (param == "shadeOpeningAngleOutside") {
      DeviceFeatures_t features = device.getDeviceFeatures();
      DS_REQUIRE(features.hasOutputAngle, "Device dows not have channel", param);
      auto value = device.getDeviceConfig(CfgClassSceneAngle, sceneNr);
      auto channelSpec = getOutputChannelIdAndSize("shadeOpeningAngleOutside");
      SceneChannel& shadeOpeningAngleOutside = sceneCache.channels[ChannelType::shadeOpeningAngleOutside];
      returnValue = convertFromOutputChannelValue(channelSpec.first, value);
      shadeOpeningAngleOutside.value = returnValue;
      sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
      devCache.lastChanged = sceneCache.lastChanged;
    } else {
      DS_REQUIRE(false, "Unknown channel requested", param);
    }
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if (auto route = dsApi::matchRoute("/property/{propertyName}", parameter)) {
    auto& propertyName = route.args["propertyName"];
    auto property = ds::tryToEnum<DeviceProperty>(propertyName);
    if (property && (*property == DeviceProperty::dimStepNum || *property == DeviceProperty::dimStepLen ||
                     *property == DeviceProperty::dimStepPause || *property == DeviceProperty::posTimeUp ||
                     *property == DeviceProperty::posTimeDown)) {
      int bank;
      int offset;
      int length;
      std::tie(bank, offset, length) = getDeviceParameter(device, *property);
      auto valueWord = device.getDeviceConfigWord(bank, offset & 0xFE);
      auto valueLow = valueWord & 0x00FF;
      auto valueHigh = (valueWord & 0xFF00) >> 8;
      auto value = (offset % 2 == 0) ? valueLow : valueHigh;
      if (*property == DeviceProperty::dimStepNum) {
        returnValue = value;
        devCache.property[*property] = value;
      } else if (*property == DeviceProperty::dimStepLen || *property == DeviceProperty::dimStepPause) {
        returnValue = value * 0.01;
        devCache.property[*property] = value;
      } else if (*property == DeviceProperty::posTimeUp || *property == DeviceProperty::posTimeDown) {
        returnValue = valueWord * 0.01;
        devCache.property[*property] = valueWord;
      } else {
        DS_FAIL_REQUIRE("No value conversion found for ", propertyName);
      }
      devCache.lastChanged = std::chrono::high_resolution_clock::now();
      m_sceneCacheDb.update(device.getDSID(), devCache);
    } else {
      DS_FAIL_REQUIRE("Unknown property requested", propertyName);
    }
  } else {
    DS_FAIL_REQUIRE("Unknown parameter requested", parameter);
  }
  return returnValue;
}

void HardwareAbstraction::HalShadeDevice::updateSceneConfig(const Device& device, int scene, dss::ChannelType channel,
                                                            int value, bool firstRegister) const {
  DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());

  SceneEntry& sceneCache = devCache.scene.sceneTable[scene];

  SceneChannel& sceneChannel = sceneCache.channels[channel];

  /* Set up the array upon first time access */
  if (sceneChannel.registerValues.empty()) {
    sceneChannel.registerValues.push_back(std::nullopt);
    sceneChannel.registerValues.push_back(std::nullopt);
  }

  if (firstRegister) {
    sceneChannel.registerValues[0] = value;
  } else {
    sceneChannel.registerValues[1] = value;
  }

  /* Save combined logical value only if both registers are valid */
  if (sceneChannel.registerValues.at(0) && sceneChannel.registerValues.at(1)) {
    sceneChannel.value = convertFromOutputChannelValue(
        channel, (*sceneChannel.registerValues[0] << 8) | (*sceneChannel.registerValues[1]));
  }

  sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
  devCache.lastChanged = sceneCache.lastChanged;
  m_sceneCacheDb.update(device.getDSID(), devCache);
}

void HardwareAbstraction::HalShadeDevice::updateDeviceConfig(const Device& device, int bank, int offset,
                                                             int value) const {
  DeviceFeatures_t features = device.getDeviceFeatures();

  /* Shade devices have 16-bit position values which needs to be taken
   * into account upon incoming device config requests that can only
   * write 8-bit at a time.
   *
   * If both values have never been read, we will not update the cache,
   * but save the register values separately. Only if both values have been
   * read at some point in time, then the combined value will be saved in
   * the cache.
   */
  if ((bank == CfgClassScene) && (offset < 128)) {
    updateSceneConfig(device, offset, ChannelType::shadePositionOutside, value, true);
  } else if ((bank == CfgClassScene) && (offset >= 128)) {
    HardwareAbstraction::HalDevice::updateDeviceConfig(device, bank, offset, value);
  } else if ((bank == CfgClassSceneExtention) && (offset < 64)) {
    int scene = offset * 2;
    int scene_odd = scene + 1;
    int value_next = value;

    // even scenes
    value &= 0x0F; // lower bits for even scene numbers
    value |= (value << 4);

    // odd scenes
    value_next &= 0xF0; // high bits for odd scene numbers
    value_next |= (value_next >> 4);

    updateSceneConfig(device, scene, ChannelType::shadePositionOutside, value, false);

    updateSceneConfig(device, scene_odd, ChannelType::shadePositionOutside, value_next, false);
  } else if ((bank == CfgClassSceneAngle) && features.hasOutputAngle && (offset < 128)) {
    DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
    SceneEntry& sceneCache = devCache.scene.sceneTable[offset];
    SceneChannel& shadeOpeningAngleOutside = sceneCache.channels[ChannelType::shadeOpeningAngleOutside];
    shadeOpeningAngleOutside.value = convertFromOutputChannelValue(ChannelType::shadeOpeningAngleOutside, value);
    sceneCache.lastChanged = std::chrono::high_resolution_clock::now();
    devCache.lastChanged = sceneCache.lastChanged;
    m_sceneCacheDb.update(device.getDSID(), devCache);
  } else if ((bank == CfgClassRuntime) && (offset >= 0x02) && (offset <= 0x04)) {
    initializeStatus(device);
    auto now = std::chrono::high_resolution_clock::now();
    auto state = m_StateCache.tryLookup(device.getDSID());
    switch (offset) {
      case 0x02: {
        auto target = convertToOutputChannelValue(
            ChannelType::shadePositionOutside, state.channels[ChannelType::shadePositionOutside].current.value_or(0));
        target &= 0xFF00;
        target |= (value & 0xFF);
        state.channels[ChannelType::shadePositionOutside].target =
            convertFromOutputChannelValue(ChannelType::shadePositionOutside, target);
        break;
      }
      case 0x03: {
        auto target = convertToOutputChannelValue(
            ChannelType::shadePositionOutside, state.channels[ChannelType::shadePositionOutside].current.value_or(0));
        target &= 0x00FF;
        target |= value << 8;
        state.channels[ChannelType::shadePositionOutside].target =
            convertFromOutputChannelValue(ChannelType::shadePositionOutside, target);
        break;
      }
      case 0x04: {
        state.channels[ChannelType::shadeOpeningAngleOutside].target =
            convertFromOutputChannelValue(ChannelType::shadeOpeningAngleOutside, value);
        break;
      }
    }
    if (!state.terminatesAt) {
      state.terminatesAt = now;
    }
    updateCurrentValues(device, now, state);
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
  } else {
    for (const auto& property : getDeviceParameters(device)) {
      int propBank;
      int propOffset;
      int length;
      std::tie(propBank, propOffset, length) = property.second;
      if ((propBank == bank) && (propOffset == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        int prevValue = 0;
        if (devCache.property[property.first]) {
          prevValue = *devCache.property[property.first] & 0xFF00;
        }
        devCache.property[property.first] = prevValue | value;
        devCache.lastChanged = std::chrono::high_resolution_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
      } else if ((length == 2) && (propBank == bank) && ((propOffset + 1) == offset)) {
        DeviceCache devCache = m_sceneCacheDb.tryLookup(device.getDSID());
        int prevValue = 0;
        if (devCache.property[property.first]) {
          prevValue = *devCache.property[property.first] & 0x00FF;
        }
        devCache.property[property.first] = prevValue | (value << 8);
        devCache.lastChanged = std::chrono::high_resolution_clock::now();
        m_sceneCacheDb.update(device.getDSID(), devCache);
      }
    }
  }
}

const std::vector<ChannelType> HardwareAbstraction::HalShadeDevice::getSceneChannels(const Device& device) const {
  DeviceFeatures_t features = device.getDeviceFeatures();
  if (features.hasOutputAngle) {
    return std::vector<ChannelType>({ChannelType::shadePositionOutside, ChannelType::shadeOpeningAngleOutside});
  } else {
    return std::vector<ChannelType>({ChannelType::shadePositionOutside});
  }
}

const std::vector<DeviceProperty> HardwareAbstraction::HalShadeDevice::getDeviceProperties(const Device& device) const {
  return std::vector<DeviceProperty>({DeviceProperty::dimStepNum, DeviceProperty::dimStepLen,
                                      DeviceProperty::dimStepPause, DeviceProperty::posTimeUp,
                                      DeviceProperty::posTimeDown});
}

bool HardwareAbstraction::HalShadeDevice::handleDimming(int sceneId, DeviceCache& cache, const ds::DateTime& now,
                                                        DeviceInternalState& state,
                                                        DeviceInternalOutputState& outputState, bool up) const {
  if (outputState.current && *outputState.current > 0 && state.lastSceneCalled &&
      cache.scene.sceneTable[*state.lastSceneCalled].channels[ChannelType::shadePositionOutside].value.value_or(0) >
          0) {
    if (state.lastDimmCommand != sceneId) {
      state.fineDimmSteps = 0;
      state.dimmStatus = DeviceDimmingStatus::none;
    }
    state.dimmingTimeout = now + std::chrono::milliseconds(2 * DefaultIncInterval);
    state.lastDimmCommand = sceneId;
    if (state.dimmStatus == DeviceDimmingStatus::none ||
        (!state.terminatesAt && (state.dimmStatus == DeviceDimmingStatus::stepping))) {
      // once terminatesAt has expired we can accept the next step
      auto nextStepInMs = (cache.property[DeviceProperty::dimStepLen].value_or(DefaultDimStepLen) +
                           cache.property[DeviceProperty::dimStepPause].value_or(DefaultDimStepPause)) *
                          MillisecondsPerHalfWave;
      state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(nextStepInMs));
      auto posTime = (up ? cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime)
                         : -1 * cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime));
      outputState.target =
          *outputState.current +
          ((double)cache.property[DeviceProperty::dimStepLen].value_or(DefaultDimStepLen) / posTime) * 100;
      if (state.fineDimmSteps < cache.property[DeviceProperty::dimStepNum].value_or(DefaultDimStepNum)) {
        state.fineDimmSteps++;
        state.dimmStatus = DeviceDimmingStatus::stepping;
      } else {
        state.fineDimmSteps = 0;
        state.dimmStatus = DeviceDimmingStatus::dimming;
      }
    }
    if (state.dimmStatus == DeviceDimmingStatus::dimming) {
      state.terminatesAt = now + std::chrono::milliseconds(2 * DefaultIncInterval); // two INC-intervals
      auto posTime = (up ? cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime)
                         : -1 * cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime));
      DS_DEBUG("handleDimming:", state.dimmStatus, *outputState.current, *outputState.target, posTime);
      outputState.target =
          *outputState.current + ((2.0 * DefaultIncInterval) / (posTime * MillisecondsPerHalfWave)) * 100;
    }
    if (!up) {
      outputState.target = std::max(*outputState.target, 0.001);
    } else {
      outputState.target = limitOutputChannelValue(ChannelType::shadePositionOutside, *outputState.target);
    }
    return true;
  }
  return false;
}

double HardwareAbstraction::HalShadeDevice::calculateRuntime(DeviceCache& cache, DeviceInternalState& state) const {
  if (state.channels[ChannelType::shadePositionOutside].current &&
      state.channels[ChannelType::shadePositionOutside].target) {
    int totalRuntime = 0;
    if (*state.channels[ChannelType::shadePositionOutside].current <
        *state.channels[ChannelType::shadePositionOutside].target) {
      totalRuntime = cache.property[DeviceProperty::posTimeUp].value_or(DefaultPositionTime) * MillisecondsPerHalfWave;
    } else {
      totalRuntime =
          cache.property[DeviceProperty::posTimeDown].value_or(DefaultPositionTime) * MillisecondsPerHalfWave;
    }
    return std::abs((*state.channels[ChannelType::shadePositionOutside].target -
                     *state.channels[ChannelType::shadePositionOutside].current) /
                    100.0 * totalRuntime);
  }
  return 0;
}

bool HardwareAbstraction::HalShadeDevice::setOutputValue(const Device& device, ChannelType channel, double value,
                                                         ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  if (channel == ChannelType::shadeOpeningAngleOutside) {
    return HardwareAbstraction::HalDevice::setOutputValue(device, channel, value, now);
  }

  auto channels = getSceneChannels(device);
  if (std::find(channels.begin(), channels.end(), channel) == channels.end()) {
    DS_FAIL_REQUIRE("No such channel", channel);
  }

  updateCurrentValues(device, now, state);
  state.channels[channel].target = value;
  state.lastDimmCommand.reset();
  state.localPriority = false;
  state.lastSceneCalled.reset();
  double dimTime = calculateRuntime(cache, state);
  if (dimTime <= 1000) {
    dimTime = 0;
  }
  if (!state.terminatesAt) {
    state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  }
  state.lastChanged = now;
  updateCurrentValues(device, now, state);
  if (state.terminatesAt &&
      std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() > 1000) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = boost::make_shared<Event>(EventName::ApartmentProxyDeviceTimeout, devRef);
    event->setProperty(
        "time",
        "+" +
            intToString(
                std::chrono::duration_cast<std::chrono::seconds>(*state.terminatesAt - state.lastChanged).count() + 2));
    device.getApartment().pushEvent(event);
  }
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

bool HardwareAbstraction::HalShadeDevice::callScene(const Device& device, int sceneId, bool forced,
                                                    ds::DateTime now) const {
  DS_REQUIRE(sceneId <= MaxSceneNumber, "Scene number is out of range.", sceneId);

  bool needsSaving = false;

  initializeStatus(device);

  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  auto sceneCache = cache.scene.sceneTable[sceneId];

  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  auto state = m_StateCache.tryLookup(device.getDSID());

  if (state.operationLocked) {
    return false;
  }

  needsSaving = updateCurrentValues(device, now, state) || needsSaving;

  if (SceneHelper::isDimScene(sceneId)) {
    auto& channel = state.channels[getDefaultChannel()];
    if (isStop(sceneId, (forced || !state.localPriority), cache)) {
      if (channel.current) {
        state.lastDimmCommand.reset();
        channel.target = *channel.current;
        state.terminatesAt.reset();
        state.dimmingTimeout.reset();
      }
    } else if (isDimmingDown(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(sceneId, cache, now, state, channel, false) || needsSaving;
    } else if (isDimmingUp(sceneId, (forced || !state.localPriority), cache)) {
      needsSaving = handleDimming(sceneId, cache, now, state, channel, true) || needsSaving;
    } else if (sceneId == SceneDimArea) {
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneDecA1 || *state.lastDimmCommand == SceneDecA2 ||
                                    *state.lastDimmCommand == SceneDecA3 || *state.lastDimmCommand == SceneDecA4)) {
        needsSaving = handleDimming(*state.lastDimmCommand, cache, now, state, channel, false) || needsSaving;
      }
      if (state.lastDimmCommand && (*state.lastDimmCommand == SceneIncA1 || *state.lastDimmCommand == SceneIncA2 ||
                                    *state.lastDimmCommand == SceneIncA3 || *state.lastDimmCommand == SceneIncA4)) {
        needsSaving = handleDimming(*state.lastDimmCommand, cache, now, state, channel, true) || needsSaving;
      }
    } else {
      return false;
    }
  } else {
    if (!sceneCache.flags.dontCare) {
      return false;
    }

    if (*sceneCache.flags.dontCare) {
      return false;
    }

    if (forced || *sceneCache.flags.ignoreLocalPrio || !state.localPriority || (sceneId == SceneLocalOff)) {
      bool firstChannel = true;
      for (auto& channel : state.channels) {
        if (sceneCache.channels[channel.first].dontCare && *sceneCache.channels[channel.first].dontCare) {
          continue;
        }
        state.lastDimmCommand.reset();
        state.dimmingTimeout.reset();
        channel.second.target = sceneCache.channels[channel.first].value;
        needsSaving = true;
        if (firstChannel) {
          state.terminatesAt.reset();
          state.beforeLastSceneCalled = state.lastSceneCalled;
          state.lastSceneCalled = sceneId;
        }

        if (channel.second.target) {
          channel.second.target = limitOutputChannelValue(channel.first, *channel.second.target);
        }

        if (firstChannel) {
          firstChannel = false;
        }
      }
      if (sceneId == SceneLocalOn && *sceneCache.flags.specialHandling) {
        state.localPriority = true;
      } else {
        state.localPriority = false;
      }
      needsSaving = true;
    } else {
      return false;
    }
  }
  double dimTime = calculateRuntime(cache, state);
  if (dimTime <= 1000) {
    dimTime = 0;
  }
  if (!state.terminatesAt) {
    state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  }
  state.lastChanged = now;
  needsSaving = updateCurrentValues(device, now, state) || needsSaving;
  if (state.terminatesAt &&
      std::chrono::duration_cast<std::chrono::milliseconds>(*state.terminatesAt - state.lastChanged).count() > 1000) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = boost::make_shared<Event>(EventName::ApartmentProxyDeviceTimeout, devRef);
    event->setProperty(
        "time",
        "+" +
            intToString(
                std::chrono::duration_cast<std::chrono::seconds>(*state.terminatesAt - state.lastChanged).count() + 2));
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  } else if (state.dimmingTimeout) {
    auto devRef = DeviceReference(device.getDSID(), &device.getApartment());
    auto event = boost::make_shared<Event>(EventName::ApartmentProxyDeviceDimmingTimeout, devRef);
    event->setProperty("time", "+5");
    event->setProperty(EventProperty::Unique, "Yes");
    device.getApartment().pushEvent(event);
  }

  if (needsSaving) {
    state.lastChanged = now;
    m_StateCache.update(device.getDSID(), state);
    publishToPropertyTree(device);
    return true;
  }
  return false;
}

bool HardwareAbstraction::HalShadeDevice::setOutputValue(const Device& device, uint8_t value, ds::DateTime now) const {
  initializeStatus(device);
  auto state = m_StateCache.tryLookup(device.getDSID());
  auto cache = m_sceneCacheDb.tryLookup(device.getDSID());
  if (cache.property[DeviceProperty::outputMode] &&
      *cache.property[DeviceProperty::outputMode] == OUTPUT_MODE_DEACTIVATED) {
    m_StateCache.remove(device.getDSID());
    return false;
  }

  state.channels[ChannelType::shadePositionOutside].target =
      convertFromOutputChannelValue(ChannelType::shadePositionOutside, (value << 8 | value));
  double dimTime = calculateRuntime(cache, state);
  state.lastChanged = now;
  state.terminatesAt = now + std::chrono::milliseconds(static_cast<int64_t>(dimTime));
  updateCurrentValues(device, now, state);
  state.lastDimmCommand.reset();
  state.dimmingTimeout.reset();
  state.localPriority = false;
  state.lastSceneCalled.reset();
  m_StateCache.update(device.getDSID(), state);
  publishToPropertyTree(device);
  return true;
}

} /* namespace dss */
