/*
    Copyright (c) 2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "apartment.h"

#include <ds-model/model.h>
#include <ds/log.h>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/make_shared.hpp>
#include <boost/range/algorithm/sort.hpp>
#include <boost/ref.hpp>

#include "businterface.h"
#include "cache/apartment-proxy.h"
#include "dss.h"
#include "event.h"
#include "logger.h"
#include "model/modelconst.h"
#include "propertysystem.h"
#include "propertysystem_common_paths.h"

#include "model/busscanner.h"
#include "model/modelevent.h"
#include "model/modelmaintenance.h"

#include "metering/metering.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "model/floor.h"

namespace dss {

//================================================== Apartment

Apartment::Apartment(DSS* dss, BusInterface* busInterface)
    : m_dss(dss),
      m_pBusInterface(busInterface),
      m_pModelMaintenance(NULL),
      m_pPropertySystem(NULL),
      m_pMetering(NULL),
      m_eventSink{},
      m_proxy(makeNullApartmentProxy()) {
  // create default floor
  boost::shared_ptr<Floor> floor = boost::make_shared<Floor>(APARTMENT_FLOOR_ID, this);
  floor->setName("Apartment");
  m_Floors.push_back(floor);

  // create default (broadcast) zone
  boost::shared_ptr<Zone> zoneZero = allocateZone(0);
  zoneZero->setIsPresent(true);
  zoneZero->setFloorID(FLOOR_ID_NONE);
  boost::shared_ptr<Zone> zoneNewDevices = allocateZone(NEW_DEVICES_ZONE_ID);
  zoneNewDevices->setIsPresent(true);
  zoneNewDevices->setFloorID(FLOOR_ID_NONE);
} // ctor

Apartment::~Apartment() {
  m_Zones.clear();
  m_DSMeters.clear();
  m_Devices.clear();
  m_States.clear();

  m_pPropertyNode.reset();
  m_pBusInterface = NULL;
  m_pModelMaintenance = NULL;
  m_pPropertySystem = NULL;
  m_pMetering = NULL;
} // dtor

std::string Apartment::id() const {
  std::string result{};
  if (m_pPropertySystem) {
    // apartmentId data owner is remote connectivity addon
    result = m_pPropertySystem->getStringValue(pp_websvc_rc_apartmentId);
    if (result.empty()) {
      result = ds::system().conf().readOr<std::string>("apartmentId", {});
    }
  }
  return result;
}

template <class Items, class Id>
static auto ids(const Items& items, const Id& id) {
  auto result = std::vector<decltype(id(items.front()))>{};
  for (const auto& item : items) {
    if (item->isInDsModel()) {
      result.push_back(id(item));
    }
  }
  boost::sort(result);
  return result;
}

std::vector<dsModel::ZoneId> Apartment::zoneIds() const {
  return ids(getZones(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::ClusterId> Apartment::clusterIds() const {
  return ids(getClusters(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::DeviceId> Apartment::deviceIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::SubmoduleId> Apartment::submoduleIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->submoduleId(); });
}

std::vector<dsModel::FunctionBlockId> Apartment::functionBlockIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->functionBlockId(); });
}

void Apartment::pushEvent(const boost::shared_ptr<Event>& event) {
  if (m_eventSink) {
    m_eventSink->pushEvent(event);
  }
}

void Apartment::addDefaultGroupsToZone(boost::shared_ptr<Zone> _zone) {
  boost::shared_ptr<Group> grp = boost::make_shared<Group>(GroupIDBroadcast, _zone);
  grp->setName("broadcast");
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDYellow, _zone));
  grp->setName("yellow");
  grp->setApplicationType(ApplicationType::Lights);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDGray, _zone));
  grp->setName("gray");
  grp->setApplicationType(ApplicationType::Blinds);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDHeating, _zone));
  grp->setName("heating");
  grp->setApplicationType(ApplicationType::Heating);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDCyan, _zone));
  grp->setName("cyan");
  grp->setApplicationType(ApplicationType::Audio);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDViolet, _zone));
  grp->setName("magenta");
  grp->setApplicationType(ApplicationType::Video);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDRed, _zone));
  grp->setName("reserved1");
  grp->setApplicationType(ApplicationType::None);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDGreen, _zone));
  grp->setName("reserved2");
  grp->setApplicationType(ApplicationType::None);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDBlack, _zone));
  grp->setName("black");
  grp->setApplicationType(ApplicationType::Joker);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDCooling, _zone));
  grp->setName("cooling");
  grp->setApplicationType(ApplicationType::Cooling);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDVentilation, _zone));
  grp->setName("ventilation");
  grp->setApplicationType(ApplicationType::Ventilation);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDWindow, _zone));
  grp->setName("window");
  grp->setApplicationType(ApplicationType::Window);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDRecirculation, _zone));
  grp->setName("recirculation");
  grp->setApplicationType(ApplicationType::Recirculation);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDControlTemperature, _zone));
  grp->setName("controltemperature");
  grp->setApplicationType(ApplicationType::ControlTemperature);
  grp->setIsValid(true);
  _zone->addGroup(grp);
} // addDefaultGroupsToZone

boost::shared_ptr<Device> Apartment::tryGetDeviceByDSID(const dsuid_t& dsid) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<Device> dev : m_Devices) {
    if (dev->getDSID() == dsid) {
      return dev;
    }
  }
  return boost::shared_ptr<Device>();
}

boost::shared_ptr<Device> Apartment::getDeviceByDSID(const dsuid_t& dsid) const {
  auto device = tryGetDeviceByDSID(dsid);
  if (!device) {
    throw ItemNotFoundException(dsuid2str(dsid));
  }
  return device;
}

boost::shared_ptr<Device> Apartment::device(const dsModel::DeviceId& deviceId) const {
  auto device = tryGetDeviceByDSID(str2dsuid(deviceId));
  DS_REQUIRE(device, "Device not found", deviceId);
  return device;
}

boost::shared_ptr<Device> Apartment::getDeviceByName(const std::string& _name) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<Device> dev : m_Devices) {
    if (dev->getName() == _name) {
      return dev;
    }
  }
  throw ItemNotFoundException(_name);
} // getDeviceByName

Set Apartment::getDevices() const {
  std::vector<DeviceReference> devs;
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<Device> dev : m_Devices) {
    devs.push_back(DeviceReference(dev, this));
  }

  return Set(devs);
} // getDevices

boost::shared_ptr<Zone> Apartment::getZoneByName(const std::string& _zoneName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<Zone> zone : m_Zones) {
    if (zone->getName() == _zoneName) {
      return zone;
    }
  }
  throw ItemNotFoundException(_zoneName);
} // getZone(name)

boost::weak_ptr<Zone> Apartment::tryGetZone(const int id) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& zone : m_Zones) {
    if (zone->getID() == id) {
      return zone;
    }
  }
  return boost::weak_ptr<Zone>();
}

boost::shared_ptr<Zone> Apartment::getZone(const int id) const {
  if (auto&& zone = tryGetZone(id).lock()) {
    return std::move(zone);
  }
  throw ItemNotFoundException(intToString(id));
}

boost::shared_ptr<Zone> Apartment::zone(const dsModel::ZoneId& id) const {
  if (auto&& zone = tryGetZone(strToInt(id)).lock()) {
    return std::move(zone);
  }
  DS_FAIL_REQUIRE("Zone not found", id);
}

std::vector<boost::shared_ptr<Zone>> Apartment::getZones() const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  return m_Zones;
}

int Apartment::getNextFreeZoneId() const {
  for (int zoneId = 1; zoneId < NEW_DEVICES_ZONE_ID; zoneId++) {
    if (!tryGetZone(zoneId).lock()) {
      return zoneId;
    }
  }
  return -1;
}

boost::weak_ptr<Floor> Apartment::tryGetFloor(const int id) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& floor : m_Floors) {
    if (floor->getID() == id) {
      return floor;
    }
  }
  return boost::weak_ptr<Floor>();
}

boost::shared_ptr<Floor> Apartment::getFloor(const int id) const {
  if (auto&& floor = tryGetFloor(id).lock()) {
    return std::move(floor);
  }
  DS_FAIL_REQUIRE("Floor not found", id);
}

boost::shared_ptr<Floor> Apartment::floor(const dsModel::FloorId& id) const {
  return getFloor(strToInt(id));
}

std::vector<boost::shared_ptr<Floor>> Apartment::getFloors() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<boost::shared_ptr<Floor>> result = m_Floors;
  return result;
}

int Apartment::getNextFreeFloorID() const {
  for (int i = 1; i < INT_MAX; i++) {
    if (!tryGetFloor(i).lock()) {
      return i;
    }
  }
  return FLOOR_ID_NONE;
}

static bool sortFloorsByOrder(const boost::shared_ptr<Floor> &f1, const boost::shared_ptr<const Floor> &f2) {
  return (f1->getOrder() < f2->getOrder());
}

int Apartment::getNextFloorOrderID() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  long long id = -1;
  for (size_t i = 0; i < m_Floors.size(); i++) {
    if (m_Floors.at(i)->getOrder() > id) {
      id = m_Floors.at(i)->getOrder();
    }
  }

  if (id < APARTMENT_FLOOR_ID) {
    id = APARTMENT_FLOOR_ID;
  }

  id++;
  if (id >= INT_MAX) { // about to wrap, time to look for holes
    // first sort the floors vector by order id
    std::sort(m_Floors.begin(), m_Floors.end(), sortFloorsByOrder);

    // the array is now sorted by order id, meaning that we set the orders
    // to be equal with the array index and thus eliminate any "jumps" in
    // ordering which might have been caused by previously removed floors
    for (size_t i = 0; i < m_Floors.size(); i++) {
      m_Floors.at(i)->setOrder(i);
    }

    if ((m_Floors.size() >= INT_MAX) || ((m_Floors.size() + 1) >= INT_MAX)) {
      id = -1;
    } else {
      id = (int)m_Floors.size();
    }

  }

  return (int)id;
}

boost::shared_ptr<DSMeter> Apartment::getDSMeter(const std::string& _modName) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getName() == _modName) {
      return dsMeter;
    }
  }
  throw ItemNotFoundException(_modName);
} // getDSMeter(name)

boost::shared_ptr<DSMeter> Apartment::tryGetDSMeterByDSID(const dsuid_t& _dsid) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getDSID() == _dsid) {
      return dsMeter;
    }
  }
  return boost::shared_ptr<DSMeter>();
}

boost::shared_ptr<DSMeter> Apartment::getDSMeterByDSID(const dsuid_t& _dsid) {
  if (auto x = tryGetDSMeterByDSID(_dsid)) {
    return x;
  }
  throw ItemNotFoundException(dsuid2str(_dsid));
}

boost::shared_ptr<DSMeter> Apartment::dsMeterByDeviceDsuid(const dsuid_t& deviceDsuid) {
  return getDSMeterByDSID(getDeviceByDSID(deviceDsuid)->getDSMeterDSID());
}

std::vector<boost::shared_ptr<DSMeter>> Apartment::getDSMeters() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<boost::shared_ptr<DSMeter>> result;
  for (boost::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    // TODO(someday): remove the filter and move this logic to ui - use HasDevices?
    // we should not return vDSMs as they are not currently filtered in ui
    if (dsMeter->getBusMemberType() != BusMember_vDSM) {
      result.push_back(dsMeter);
    }
  }
  return result;
} // getDSMeters

std::vector<boost::shared_ptr<DSMeter>> Apartment::getDSMetersAll() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<boost::shared_ptr<DSMeter>> result = m_DSMeters;
  return result;
} // getAllDSMeters

// Group queries
boost::shared_ptr<Group> Apartment::getGroup(const std::string& _name) {
  boost::shared_ptr<Group> pResult = getZone(0)->getGroup(_name);
  if (pResult != NULL) {
    return pResult;
  }
  throw ItemNotFoundException(_name);
} // getGroup(name)

boost::shared_ptr<Group> Apartment::getGroup(const int _id) {
  boost::shared_ptr<Group> pResult = getZone(0)->getGroup(_id);
  if (pResult != NULL) {
    return pResult;
  }
  throw ItemNotFoundException(intToString(_id));
} // getGroup(id)

boost::weak_ptr<Cluster> Apartment::tryGetCluster(const int id) {
  if (isAppUserGroup(id)) {
    boost::shared_ptr<Cluster> pResult = boost::dynamic_pointer_cast<Cluster>(getZone(0)->getGroup(id));
    if (pResult != NULL) {
      return pResult;
    }
  }
  return boost::weak_ptr<Cluster>();
}

boost::shared_ptr<Cluster> Apartment::getCluster(const int id) {
  if (auto&& cluster = tryGetCluster(id).lock()) {
    return std::move(cluster);
  }
  throw ItemNotFoundException(intToString(id));
} // getCluster(id)

boost::shared_ptr<Cluster> Apartment::cluster(const dsModel::ClusterId& id) {
  if (auto&& cluster = tryGetCluster(strToInt(id)).lock()) {
    return std::move(cluster);
  }
  DS_FAIL_REQUIRE("Cluster not found", id);
}

std::vector<boost::shared_ptr<Cluster>> Apartment::getClusters() const {
  std::vector<boost::shared_ptr<Cluster>> result;
  for (boost::shared_ptr<Group> pGroup : getZone(0)->getGroups()) {
    if (isCluster(pGroup->getID())) {
      result.push_back(boost::dynamic_pointer_cast<Cluster>(pGroup));
    }
  }
  return result;
}

boost::shared_ptr<Cluster> Apartment::getEmptyCluster() {
  // find a group slot with unassigned state machine id
  for (boost::shared_ptr<Cluster> pCluster : getClusters()) {
    if (pCluster->getApplicationType() == ApplicationType::None) {
      return pCluster;
    }
  }
  return boost::shared_ptr<Cluster>();
}

std::vector<boost::shared_ptr<Group>> Apartment::getGlobalApps() {
  std::vector<boost::shared_ptr<Group>> result;
  for (boost::shared_ptr<Group> pGroup : getZone(0)->getGroups()) {
    if (isGlobalAppGroup(pGroup->getID())) {
      result.push_back(pGroup);
    }
  }
  return result;
}

boost::shared_ptr<Device> Apartment::allocateDevice(const dsuid_t _dsid) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<Device> pResult;
  // search for existing device
  for (boost::shared_ptr<Device> device : m_Devices) {
    if (device->getDSID() == _dsid) {
      pResult = device;
      break;
    }
  }

  if (pResult == NULL) {
    pResult.reset(new Device(_dsid, this));
    pResult->setFirstSeen(DateTime());
    m_Devices.push_back(pResult);
  }
  auto zoneId = pResult->getZoneID();
  DeviceReference devRef(pResult, this);
  getZone(0)->addDevice(devRef);
  if (auto zone = tryGetZone(zoneId).lock()) {
    zone->addDevice(devRef);
  }
  pResult->publishToPropertyTree();
  return pResult;
} // allocateDevice

boost::shared_ptr<DSMeter> Apartment::allocateDSMeter(const dsuid_t _dsid) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (boost::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getDSID() == _dsid) {
      return dsMeter;
    }
  }

  boost::shared_ptr<DSMeter> pResult = boost::make_shared<DSMeter>(_dsid, this);
  m_DSMeters.push_back(pResult);

  return pResult;
} // allocateDSMeter

boost::shared_ptr<Zone> Apartment::allocateZone(int _zoneID) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<Zone> result;
  for (boost::shared_ptr<Zone> zone : m_Zones) {
    if (zone->getID() == _zoneID) {
      result = zone;
      break;
    }
  }

  if (result == NULL) {
    result.reset(new Zone(_zoneID, this));
    result->publishToPropertyTree();
    addDefaultGroupsToZone(result);
    m_Zones.push_back(result);

    if (_zoneID == 0) {
      for (int i = GroupIDAppUserMin; i <= GroupIDAppUserMax; ++i) {
        boost::shared_ptr<Cluster> pCluster = boost::make_shared<Cluster>(i, boost::ref<Apartment>(*this));
        result->addGroup(pCluster);
      }

      // pre-create apartment ventilation group
      boost::shared_ptr<Group> group;
      group.reset(new Group(GroupIDGlobalAppDsVentilation, result));
      group->setName("apartmentVentilation");
      group->setApplicationType(ApplicationType::ApartmentVentilation);
      group->setIsValid(
          true); // TODO(soon): this may not be needed for AV (maybe it is only valid when devices are present)
      result->addGroup(group);

      // pre-create awnings group
      group.reset(new Group(GroupIDGlobalAppDsAwnings, result));
      group->setName("awnings");
      group->setApplicationType(ApplicationType::Awnings);
      group->setIsValid(true);
      result->addGroup(group);

      // TODO(soon): remove again before R1705
      //        group.reset(new Group(GroupIDGlobalAppDsRecirculation, result));
      //        group->setName("apartmentRecirculation");
      //        group->setApplicationType(ApplicationType::ApartmentRecirculation);
      //        group->setIsValid(true);  // TODO(soon): this may not be needed for AV (maybe it is only valid when
      //        devices are present)
      //        result->addGroup(group);
    } else {
      if ((result->getID() != ZONE_ID_BROADCAST) &&
          (result->getID() != NEW_DEVICES_ZONE_ID)) {
        getFloor(0)->addZone(result->getID());
      }
    }
  } else {
    result->publishToPropertyTree();
  }
  return result;
} // allocateZone

void Apartment::removeZone(int _zoneID) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<boost::shared_ptr<Zone>>::iterator ipZone = m_Zones.begin(), e = m_Zones.end(); ipZone != e;
       ++ipZone) {
    boost::shared_ptr<Zone> pZone = *ipZone;
    if (pZone->getID() == _zoneID) {
      pZone->removeFromPropertyTree();
      if (pZone->getFloorID() != FLOOR_ID_NONE) {
          getFloor(pZone->getFloorID())->removeZone(_zoneID);
      }
      m_Zones.erase(ipZone);
      return;
    }
  }
} // removeZone

boost::shared_ptr<Floor> Apartment::allocateFloor(int id) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<Floor> result;

  if (id != FLOOR_ID_NONE) {
    for (boost::shared_ptr<Floor> floor : m_Floors) {
      if (floor->getID() == id) {
        result = floor;
        break;
      }
    }
  }

  if (result == NULL) {
    if (id == FLOOR_ID_NONE) {
      id = getNextFreeFloorID();
      if (id == FLOOR_ID_NONE) {
        throw std::runtime_error("Maximum floor limit reached");
      }
    }
    result.reset(new Floor(id, this));
    if (id != APARTMENT_FLOOR_ID) {
      int order = getNextFloorOrderID();
      if (order == -1) {
        throw std::runtime_error("Maximum order id limit reached");
      }
      result->setOrder(order);
    }
    m_Floors.push_back(result);
  }
  
  result->publishToPropertyTree();
  return result;
} // allocateFloor

void Apartment::removeFloor(int id) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<boost::shared_ptr<Floor>>::iterator ipFloor = m_Floors.begin(), e = m_Floors.end(); ipFloor != e;
       ++ipFloor) {
    boost::shared_ptr<Floor> pFloor = *ipFloor;
    if (pFloor->getID() == id) {
      pFloor->removeFromPropertyTree();
      m_Floors.erase(ipFloor);
      return;
    }
  }
} // removeFloor



void Apartment::removeDevice(dsuid_t _device) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<boost::shared_ptr<Device>>::iterator ipDevice = m_Devices.begin(), e = m_Devices.end();
       ipDevice != e; ++ipDevice) {
    boost::shared_ptr<Device> pDevice = *ipDevice;
    if (pDevice->getDSID() == _device) {
      // Remove from zone
      int zoneID = pDevice->getZoneID();
      DeviceReference devRef = DeviceReference(pDevice, this);

      {
        boost::shared_ptr<DeviceReference> pDevRef = boost::make_shared<DeviceReference>(devRef);
        boost::shared_ptr<Event> mEvent = boost::make_shared<Event>(EventName::DeviceEvent, pDevRef);
        mEvent->setProperty("action", "deleted");
        mEvent->setProperty("dsuid", dsuid2str(pDevice->getDSID()));
        mEvent->setProperty("zoneID", intToString(pDevice->getZoneID()));
        if (DSS::hasInstance()) {
          DSS::getInstance()->getEventQueue().pushEvent(mEvent);
        }
      }

      if (zoneID != 0) {
        getZone(zoneID)->removeDevice(devRef);
      }
      getZone(0)->removeDevice(devRef);

      try {
        // Remove from dsm
        boost::shared_ptr<DSMeter> dsMeter = getDSMeterByDSID(pDevice->getDSMeterDSID());
        dsMeter->removeDevice(devRef);
      } catch (ItemNotFoundException& e) {
        Logger::getInstance()->log(std::string("Apartment::removeDevice: Unknown dSM: ") + e.what(), lsWarning);
      }

      pDevice->clearBinaryInputs(); // calls apartment->removeState from inside
      pDevice->clearStates();       // calls apartment->removeState() from inside

      // Erase
      m_Devices.erase(ipDevice);
      return;
    }
  }
} // removeDevice

void Apartment::removeDSMeter(dsuid_t _dsMeter) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<boost::shared_ptr<DSMeter>>::iterator ipDSMeter = m_DSMeters.begin(), e = m_DSMeters.end();
       ipDSMeter != e; ++ipDSMeter) {
    boost::shared_ptr<DSMeter> pDSMeter = *ipDSMeter;
    if (pDSMeter->getDSID() == _dsMeter) {
      m_DSMeters.erase(ipDSMeter);
      return;
    }
  }
} // removeDSMeter

void Apartment::removeInactiveMeters() {
  std::vector<boost::shared_ptr<DSMeter>> dsMeters = getDSMeters();

  bool dirtyFlag = false;
  for (boost::shared_ptr<DSMeter> dsMeter : dsMeters) {
    if (!dsMeter->isPresent()) {
      removeDSMeter(dsMeter->getDSID());
      dirtyFlag = true;
    }
  }

  if (dirtyFlag) {
    m_pModelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
  }
} // removeInactiveMeters

boost::shared_ptr<State> Apartment::allocateState(const eStateType _type, const std::string& _stateName,
                                                  const std::string& _scriptId) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  for (auto&& state : m_States) {
    if ((state->getName() == _stateName) && (state->getType() == _type) && (state->getProviderService() == _scriptId)) {
      result = state;
      break;
    }
  }
  if (result == NULL) {
    result = boost::make_shared<State>(_type, _stateName, _scriptId);
    m_States.push_back(result);
  }
  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateState(const std::string& _identifier, const std::string& _scriptId,
                                                  boost::shared_ptr<Device> _device, SensorType _sensorType,
                                                  const std::string& _activateCondition,
                                                  const std::string& _deactivateCondition) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  auto name = StateSensor::buildName(_device, _sensorType, _identifier);
  for (auto&& state : m_States) {
    if ((state->getName() == name) && (state->getType() == StateType_SensorDevice) &&
        (state->getProviderService() == _scriptId)) {
      removeState(state);
      break;
    }
  }
  result = boost::make_shared<StateSensor>(_identifier, _scriptId, _device, _sensorType, _activateCondition,
                                           _deactivateCondition);
  m_States.push_back(result);

  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateState(const std::string& _identifier, const std::string& _scriptId,
                                                  boost::shared_ptr<Group> _group, SensorType _sensorType,
                                                  const std::string& _activateCondition,
                                                  const std::string& _deactivateCondition) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  auto name = StateSensor::buildName(_group, _sensorType, _identifier);
  for (auto&& state : m_States) {
    if ((state->getName() == name) && (state->getType() == StateType_SensorZone) &&
        (state->getProviderService() == _scriptId)) {
      removeState(state);
      break;
    }
  }
  result = boost::make_shared<StateSensor>(_identifier, _scriptId, _group, _sensorType, _activateCondition,
                                           _deactivateCondition);
  m_States.push_back(result);

  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateBinaryInputState(boost::shared_ptr<Device> _device, int inputIndex) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  std::string name = BinaryInputState::makeName(_device->getDSID(), inputIndex);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = boost::make_shared<BinaryInputState>(_device, inputIndex);
  m_States.push_back(result);
  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateState(boost::shared_ptr<Device> _device, const std::string& stateName) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  std::string name = State::buildName(_device, stateName);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = boost::make_shared<State>(_device, stateName);
  m_States.push_back(result);
  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateState(boost::shared_ptr<DSMeter> _meter, int _inputIndex) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  std::string name = State::buildName(_meter, _inputIndex);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = boost::make_shared<State>(_meter, _inputIndex);
  m_States.push_back(result);
  return result;
} // allocateState

boost::shared_ptr<State> Apartment::allocateState(boost::shared_ptr<Group> _group, const std::string& name) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  boost::shared_ptr<State> result;
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = boost::make_shared<State>(_group, name);
  m_States.push_back(result);
  return result;
} // allocateState

boost::shared_ptr<State> Apartment::tryGetState(const eStateType _type, const std::string& _name) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() == _type) && (state->getName() == _name)) {
      return state;
    }
  }
  return boost::shared_ptr<State>();
} // tryGetState

boost::shared_ptr<State> Apartment::getState(const eStateType _type, const std::string& _name) const {
  auto state = tryGetState(_type, _name);
  if (!state) {
    throw ItemNotFoundException(_name);
  }
  return state;
} // getState

boost::shared_ptr<State> Apartment::tryGetNonScriptState(const std::string& stateName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() != StateType_Script) && (state->getName() == stateName)) {
      return state;
    }
  }
  return {};
}

boost::shared_ptr<State> Apartment::getNonScriptState(const std::string& stateName) const {
  if (auto state = tryGetNonScriptState(stateName)) {
    return state;
  }
  throw ItemNotFoundException(stateName);
}

boost::shared_ptr<State> Apartment::getState(const eStateType _type, const std::string& _identifier,
                                             const std::string& _stateName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() == _type) && (state->getProviderService() == _identifier) &&
        (state->getName() == _stateName)) {
      return state;
    }
  }
  throw ItemNotFoundException(_stateName);
}

boost::shared_ptr<State> Apartment::getState(const std::string& identifier, const std::string& name) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getProviderService() == identifier) && (state->getName() == name)) {
      return state;
    }
  }
  DS_FAIL_REQUIRE("State not found", name);
}

// throws (dynamic_cast relies on consistent sensor names)
void Apartment::updateSensorStates(const dsuid_t& dsuid, SensorType sensorType, double value, callOrigin_t origin) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  // A single sensor can have multiple states, by setting different on/off threshold
  // TODO(someday) SensorState should have a field to identify the input sensor
  std::string prefix("dev." + dsuid2str(dsuid) + ".type" + intToString(static_cast<int>(sensorType)));

  for (auto&& state : m_States) {
    if (boost::starts_with(state->getName(), prefix)) {
      boost::dynamic_pointer_cast<StateSensor>(state)->newValue(origin, value);
    }
  }
}

// throws (dynamic_cast relies on consistent sensor names)
void Apartment::updateSensorStates(int zoneId, int groupId, SensorType sensorType, double value, callOrigin_t origin) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  // A single sensor can have multiple states, by setting different on/off threshold
  // TODO(someday) SensorState should have a field to identify the input sensor
  std::string prefix("zone.zone" + intToString(zoneId) + ".group" + intToString(groupId) + ".type" +
                     intToString(static_cast<int>(sensorType)));

  for (auto&& state : m_States) {
    if (boost::starts_with(state->getName(), prefix)) {
      boost::dynamic_pointer_cast<StateSensor>(state)->newValue(origin, value);
    }
  }
}

void Apartment::removeState(boost::shared_ptr<State> _state) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  m_States.erase(std::remove(m_States.begin(), m_States.end(), _state));
}

std::vector<boost::shared_ptr<State>> Apartment::getStates() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  return m_States;
}

ActionRequestInterface* Apartment::getActionRequestInterface() {
  if (m_pBusInterface != NULL) {
    return m_pBusInterface->getActionRequestInterface();
  }
  return NULL;
} // getActionRequestInterface

DeviceBusInterface* Apartment::getDeviceBusInterface() {
  if (m_pBusInterface != NULL) {
    return m_pBusInterface->getDeviceBusInterface();
  }
  return NULL;
} // getDeviceBusInterface

void Apartment::setPropertySystem(PropertySystem* _value) {
  m_pPropertySystem = _value;
  if (m_pPropertySystem != NULL) {
    m_pPropertyNode = m_pPropertySystem->createProperty("/apartment");
    for (boost::shared_ptr<Zone> pZone : m_Zones) {
      pZone->publishToPropertyTree();
    }
    for (boost::shared_ptr<Floor> pFloor : m_Floors) {
      pFloor->publishToPropertyTree();
    }
  }
} // setPropertySystem

ApartmentSensorStatus_t Apartment::getSensorStatus() const {
  return m_SensorStatus;
}

void Apartment::setTemperature(double _value, DateTime& _ts) {
  m_SensorStatus.m_TemperatureValue = _value;
  m_SensorStatus.m_TemperatureTS = _ts;
}

void Apartment::setHumidityValue(double _value, DateTime& _ts) {
  m_SensorStatus.m_HumidityValue = _value;
  m_SensorStatus.m_HumidityTS = _ts;
}

void Apartment::setBrightnessValue(double _value, DateTime& _ts) {
  m_SensorStatus.m_BrightnessValue = _value;
  m_SensorStatus.m_BrightnessTS = _ts;
}

void Apartment::setWindSpeed(double _value, DateTime& _ts) {
  m_SensorStatus.m_WindSpeedValue = _value;
  m_SensorStatus.m_WindSpeedTS = _ts;
}

void Apartment::setWindDirection(double _value, DateTime& _ts) {
  m_SensorStatus.m_WindDirectionValue = _value;
  m_SensorStatus.m_WindDirectionTS = _ts;
}

void Apartment::setGustSpeed(double _value, DateTime& _ts) {
  m_SensorStatus.m_GustSpeedValue = _value;
  m_SensorStatus.m_GustSpeedTS = _ts;
}

void Apartment::setGustDirection(double _value, DateTime& _ts) {
  m_SensorStatus.m_GustDirectionValue = _value;
  m_SensorStatus.m_GustDirectionTS = _ts;
}

void Apartment::setPrecipitation(double _value, DateTime& _ts) {
  m_SensorStatus.m_PrecipitationValue = _value;
  m_SensorStatus.m_PrecipitationTS = _ts;
}

void Apartment::setAirPressure(double _value, DateTime& _ts) {
  m_SensorStatus.m_AirPressureValue = _value;
  m_SensorStatus.m_AirPressureTS = _ts;
}

void Apartment::setWeatherInformation(std::string& _iconId, std::string& _conditionId, std::string _serviceId,
                                      DateTime& _ts) {
  m_SensorStatus.m_WeatherIconId = _iconId;
  m_SensorStatus.m_WeatherConditionId = _conditionId;
  m_SensorStatus.m_WeatherServiceId = _serviceId;
  m_SensorStatus.m_WeatherTS = _ts;
}

void Apartment::updateWeatherProtectedClusters(const double _sensorValue) {
  for (boost::shared_ptr<Cluster> cluster : getClusters()) {
    auto protectionClass = cluster->getProtectionClass();
    if (protectionClass <= wpc_none || wpc_last <= protectionClass) {
      continue;
    }
    cluster->checkWindProtection(_sensorValue);
  }
}

std::pair<std::vector<DeviceLock_t>, std::vector<ZoneLock_t>> Apartment::getClusterLocks() {
  std::vector<DeviceLock_t> lockedDevices;
  std::vector<ZoneLock_t> lockedZones;

  std::vector<boost::shared_ptr<Zone>> zones = getZones();
  // loop through all zones
  for (size_t z = 0; z < zones.size(); z++) {
    if (zones.at(z)->getID() == 0) {
      continue;
    }

    Set devices = zones.at(z)->getDevices();

    std::vector<ClassLock_t> classLocks;

    // look for devices of each class in the zone
    for (int c = DEVICE_CLASS_GE; c <= DEVICE_CLASS_WE; c++) {
      if (c == DEVICE_CLASS_SW) {
        // ignore unconfigured SW devices
        continue;
      }

      std::set<int> lockedScenes;

      Set devclass = devices.getByGroup(c);

      // loop through all devices of a certain group
      for (int d = 0; d < devclass.length(); d++) {
        boost::shared_ptr<Device> device = devclass.get(d).getDevice();
        std::vector<int> ls = device->getLockedScenes();
        if (!ls.empty()) {
          DeviceLock_t lockedDevice;
          lockedDevice.dsuid = device->getDSID();
          lockedDevice.lockedScenes = ls;
          lockedDevices.push_back(lockedDevice);
          lockedScenes.insert(ls.begin(), ls.end());
        } else {
          if (device->isInLockedCluster()) {
            DeviceLock_t lockedDevice;
            lockedDevice.dsuid = device->getDSID();
            lockedDevices.push_back(lockedDevice);
          }
        }
      } // devices loop

      if (!lockedScenes.empty()) {
        ClassLock_t cl;
        cl.deviceClass = static_cast<DeviceClasses_t>(c);
        cl.lockedScenes = lockedScenes;
        classLocks.push_back(cl);
      }
    } // device class loop

    if (!classLocks.empty()) {
      ZoneLock_t zl;
      zl.zoneID = zones.at(z)->getID();
      zl.deviceClassLocks = classLocks;
      lockedZones.push_back(zl);
    }
  } // zone loop

  return std::make_pair(lockedDevices, lockedZones);
}

bool Apartment::setDevicesFirstSeen(const DateTime& dateTime) {
  static const DateTime allowedDate = DateTime::parseISO8601("2011-01-01T00:00:00Z");
  if (dateTime < allowedDate) {
    return false;
  }
  for (boost::shared_ptr<Device> dev : m_Devices) {
    const DateTime originDate = dev->getFirstSeen();
    if (originDate < allowedDate) {
      dev->setFirstSeen(dateTime);
    }
  }
  m_pModelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
  return true;
}

void Apartment::setProxy(std::unique_ptr<ApartmentProxy> proxy) {
  m_proxy = std::move(proxy);
}

ApartmentProxy& Apartment::proxy() const {
  return *m_proxy;
}

ItemNotFoundException::ItemNotFoundException(const std::string& name)
    : DSSException(::ds::strArgs("Could not find item. ", ds::hideArg(name), " ", DS_CONTEXT_ARGS)) {}

} // namespace dss
