/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef APARTMENT_H
#define APARTMENT_H

#include <vector>

#include <ds-model/model-fwd.h>
#include <ds/ds-signal.h>
#include <regex.h>
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <set>

#include "datetools.h"
#include "devicecontainer.h"
#include "dssfwd.h"
#include "model/device.h"
#include "model/modelconst.h"
#include "subsystem.h"

namespace dss {
class PropertyNode;
typedef boost::shared_ptr<PropertyNode> PropertyNodePtr;
class DSMeter;
class Cluster;
class ModelMaintenance;
class BusInterface;
class Metering;
class DeviceBusInterface;
class ActionRequestInterface;
class Event;
class EventSink;
class ApartmentProxy;
class Floor;

typedef struct ApartmentSensorStatus {
  ApartmentSensorStatus()
      : m_TemperatureValue(0),
        m_TemperatureTS(0),
        m_HumidityValue(0),
        m_HumidityTS(0),
        m_BrightnessValue(0),
        m_BrightnessTS(0),
        m_WindSpeedValue(0),
        m_WindSpeedTS(0),
        m_WindDirectionValue(0),
        m_WindDirectionTS(0),
        m_GustSpeedValue(0),
        m_GustSpeedTS(0),
        m_GustDirectionValue(0),
        m_GustDirectionTS(0),
        m_PrecipitationValue(0),
        m_PrecipitationTS(0),
        m_AirPressureValue(0),
        m_AirPressureTS(0),
        m_WeatherTS(0) {}
  double m_TemperatureValue;
  DateTime m_TemperatureTS;
  double m_HumidityValue;
  DateTime m_HumidityTS;
  double m_BrightnessValue;
  DateTime m_BrightnessTS;
  double m_WindSpeedValue;
  DateTime m_WindSpeedTS;
  double m_WindDirectionValue;
  DateTime m_WindDirectionTS;
  double m_GustSpeedValue;
  DateTime m_GustSpeedTS;
  double m_GustDirectionValue;
  DateTime m_GustDirectionTS;
  double m_PrecipitationValue;
  DateTime m_PrecipitationTS;
  double m_AirPressureValue;
  DateTime m_AirPressureTS;

  std::string m_WeatherIconId;
  std::string m_WeatherConditionId;
  std::string m_WeatherServiceId;
  DateTime m_WeatherTS;
} ApartmentSensorStatus_t;

typedef struct {
  DeviceClasses_t deviceClass;
  std::set<int> lockedScenes;
} ClassLock_t;

typedef struct {
  int zoneID;
  std::vector<ClassLock_t> deviceClassLocks;
} ZoneLock_t;

typedef struct {
  dsuid_t dsuid;
  std::vector<int> lockedScenes;
} DeviceLock_t;

/** Represents an Apartment
  * This is the root of the datamodel of the dss. The Apartment is responsible for delivering
  * and loading all subitems.
  */
class Apartment : public boost::noncopyable, public DeviceContainer {
private:
  DSS* m_dss;
  std::vector<boost::shared_ptr<Zone> > m_Zones;
  std::vector<boost::shared_ptr<DSMeter> > m_DSMeters;
  std::vector<boost::shared_ptr<Device> > m_Devices;
  std::vector<boost::shared_ptr<State> > m_States;
  std::vector<boost::shared_ptr<Floor> > m_Floors;
  ApartmentSensorStatus_t m_SensorStatus;

  BusInterface* m_pBusInterface;
  PropertyNodePtr m_pPropertyNode;
  ModelMaintenance* m_pModelMaintenance;
  PropertySystem* m_pPropertySystem;
  Metering* m_pMetering;
  EventSink* m_eventSink;
  mutable boost::recursive_mutex m_mutex;
  std::unique_ptr<ApartmentProxy> m_proxy;
  ds::Signal<> m_structureChanged;
  ds::Signal<> m_statusChanged;

  void addDefaultGroupsToZone(boost::shared_ptr<Zone> _zone);

public:
  Apartment(DSS* = nullptr, BusInterface* = nullptr);
  virtual ~Apartment();

  boost::recursive_mutex& getMutex() const { return m_mutex; }
  DSS* getDss() { return m_dss; }

  virtual std::string id() const;
  std::vector<dsModel::ZoneId> zoneIds() const;
  std::vector<dsModel::ClusterId> clusterIds() const;
  std::vector<dsModel::DeviceId> deviceIds() const;
  std::vector<dsModel::SubmoduleId> submoduleIds() const;
  std::vector<dsModel::FunctionBlockId> functionBlockIds() const;

  /// Push event to event queue
  void pushEvent(const boost::shared_ptr<Event>&);

  /** Returns a set containing all devices of the set */
  virtual Set getDevices() const;

  /// Returns a reference to the device with the DSID \a _dsid, nullptr if not found.
  boost::shared_ptr<Device> tryGetDeviceByDSID(const dsuid_t&) const;
  /// Returns a reference to the device with the DSID \a _dsid, throws if not found
  boost::shared_ptr<Device> getDeviceByDSID(const dsuid_t&) const;
  /** Returns a reference to the device with the name \a _name*/
  boost::shared_ptr<Device> getDeviceByName(const std::string& _name);
  std::vector<boost::shared_ptr<Device>> getDevicesVector() const { return m_Devices; }
  boost::shared_ptr<Device> device(const dsModel::DeviceId&) const;

  /** Returns the Zone by name */
  boost::shared_ptr<Zone> getZoneByName(const std::string& _zoneName) const;
  boost::weak_ptr<Zone> tryGetZone(const int id) const;
  /** Returns the Zone by its id */
  boost::shared_ptr<Zone> getZone(const int id) const;
  boost::shared_ptr<Zone> zone(const dsModel::ZoneId&) const;

  /** Returns a vector of all zones */
  std::vector<boost::shared_ptr<Zone>> getZones() const;

  /** Return the next free zoneId. -1 if nothing found */
  int getNextFreeZoneId() const;

  boost::weak_ptr<Floor> tryGetFloor(const int id) const;
  boost::shared_ptr<Floor> getFloor(const int id) const;
  boost::shared_ptr<Floor> floor(const dsModel::FloorId&) const;
  std::vector<boost::shared_ptr<Floor> > getFloors();
  boost::shared_ptr<Floor> allocateFloor(int id = FLOOR_ID_NONE);
  void removeFloor(int id);
  /** Return the next free floorId. -1 if nothing found */
  int getNextFreeFloorID() const;
  int getNextFloorOrderID();


  /** Returns a DSMeter by name */
  boost::shared_ptr<DSMeter> getDSMeter(const std::string& _modName);
  boost::shared_ptr<DSMeter> tryGetDSMeterByDSID(const dsuid_t&);
  boost::shared_ptr<DSMeter> getDSMeterByDSID(const dsuid_t&);
  boost::shared_ptr<DSMeter> dsMeterByDeviceDsuid(const dsuid_t&);
  /** Returns a vector of dsMeters without vDSMs */
  std::vector<boost::shared_ptr<DSMeter> > getDSMeters();
  /** Returns a vector of all dsMeters */
  std::vector<boost::shared_ptr<DSMeter> > getDSMetersAll();

  /** Returns a Group by name */
  boost::shared_ptr<Group> getGroup(const std::string& _name);
  /** Returns a Group by id */
  boost::shared_ptr<Group> getGroup(const int _id);

  /** Returns a Cluster by id */
  boost::weak_ptr<Cluster> tryGetCluster(const int _id);
  boost::shared_ptr<Cluster> getCluster(const int _id);
  boost::shared_ptr<Cluster> cluster(const dsModel::ClusterId& _id);
  /** Returns a vector of Clusters. */
  std::vector<boost::shared_ptr<Cluster> > getClusters() const;
  /** Return an Empty Cluster, if available. */
  boost::shared_ptr<Cluster> getEmptyCluster();

  /** Returns a vector of Global Application Groups */
  std::vector<boost::shared_ptr<Group> > getGlobalApps();

  /// @return nullptr on failure
  boost::shared_ptr<State> tryGetState(const eStateType _type, const std::string& _stateName) const;
  /// @throw ItemNotFoundException on failure
  boost::shared_ptr<State> getState(const eStateType _type, const std::string& _stateName) const;
  boost::shared_ptr<State> tryGetNonScriptState(const std::string& stateName) const;
  boost::shared_ptr<State> getNonScriptState(const std::string& _stateName) const;
  /** Returns the state by type, identifier and name */
  boost::shared_ptr<State> getState(
      const eStateType _type, const std::string& _identifier, const std::string& _stateName) const;

  /// Returns the state of any type, with matching identifier and name
  boost::shared_ptr<State> getState(const std::string& identifier, const std::string& name) const;

  void updateSensorStates(const dsuid_t& dsuid, SensorType sensorType, double value, callOrigin_t origin);
  void updateSensorStates(int zoneId, int groupId, SensorType sensorType, double value, callOrigin_t origin);

  /** Allocates a device and returns a reference to it. */
  boost::shared_ptr<Device> allocateDevice(const dsuid_t _dsid);
  /** Allocates a zone and returns a reference to it. Should a zone with
    * the given _zoneID already exist, a reference to the existing zone will
    * be returned.
    * NOTE: Outside code should never call this function
    */
  boost::shared_ptr<Zone> allocateZone(int _zoneID);
  boost::shared_ptr<DSMeter> allocateDSMeter(const dsuid_t _dsid);
  boost::shared_ptr<State> allocateState(eStateType _type, const std::string& _name, const std::string& _scriptId);
  boost::shared_ptr<State> allocateState(const std::string& _identifier, const std::string& _scriptId,
      boost::shared_ptr<Device> _device, SensorType _sensorType, const std::string& _activateCondition,
      const std::string& _deactivateCondition);
  boost::shared_ptr<State> allocateState(const std::string& _identifier, const std::string& _scriptId,
      boost::shared_ptr<Group> _group, SensorType _sensorType, const std::string& _activateCondition,
      const std::string& _deactivateCondition);
  boost::shared_ptr<State> allocateState(boost::shared_ptr<Device> _device, const std::string& stateName);
  boost::shared_ptr<State> allocateState(boost::shared_ptr<DSMeter> _meter, int _inputIndex);
  boost::shared_ptr<State> allocateState(boost::shared_ptr<Group> _group, const std::string& name);
  boost::shared_ptr<State> allocateBinaryInputState(boost::shared_ptr<Device> _device, int inputIndex);

  void removeZone(int _zoneID);
  void removeDevice(dsuid_t _device);
  void removeDSMeter(dsuid_t _dsMeter);
  void removeInactiveMeters();
  void removeState(boost::shared_ptr<State> _state);
  std::vector<boost::shared_ptr<State>> getStates();

  ApartmentSensorStatus_t getSensorStatus() const;
  void setTemperature(double _value, DateTime& _ts);
  void setHumidityValue(double _value, DateTime& _ts);
  void setBrightnessValue(double _value, DateTime& _ts);
  void setWindSpeed(double _value, DateTime& _ts);
  void setWindDirection(double _value, DateTime& _ts);
  void setGustSpeed(double _value, DateTime& _ts);
  void setGustDirection(double _value, DateTime& _ts);
  void setPrecipitation(double _value, DateTime& _ts);
  void setAirPressure(double _value, DateTime& _ts);
  void setWeatherInformation(std::string& _iconId, std::string& _conditionId, std::string _serviceId, DateTime& _ts);
  void updateWeatherProtectedClusters(const double _sensorValue);
  std::pair<std::vector<DeviceLock_t>, std::vector<ZoneLock_t> > getClusterLocks();

  bool setDevicesFirstSeen(const DateTime& dateTime);

  void setBusInterface(BusInterface* _value) { m_pBusInterface = _value; }
  BusInterface* getBusInterface() { return m_pBusInterface; }
  ActionRequestInterface* getActionRequestInterface();
  DeviceBusInterface* getDeviceBusInterface();
  /** Returns the root-node for the apartment tree */
  PropertyNodePtr getPropertyNode() { return m_pPropertyNode; }
  void setModelMaintenance(ModelMaintenance* _value) { m_pModelMaintenance = _value; }
  ModelMaintenance* getModelMaintenance() { return m_pModelMaintenance; }
  void setMetering(Metering* _value) { m_pMetering = _value; }
  void setPropertySystem(PropertySystem* _value);
  void setEventSink(EventSink& x) { m_eventSink = &x; }
  PropertySystem* getPropertySystem() { return m_pPropertySystem; }

  void setProxy(std::unique_ptr<ApartmentProxy> proxy);
  ApartmentProxy& proxy() const;

  ds::Slots<>& structureChanged() { return m_structureChanged; }
  void emitStructureChanged() { m_structureChanged.emit(); }
  ds::Slots<>& statusChanged() { return m_statusChanged; }
  void emitStatusChanged() { m_statusChanged.emit(); }

}; // Apartment

/** Exception that will be thrown if a given item could not be found */
class ItemNotFoundException : public DSSException {
public:
  ItemNotFoundException(const std::string& name);
};

} // namespace dss

#endif // APARTMENT_H
