/*
    Copyright (c) 2010,2011 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#include "data_types.h"

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "busscanner.h"

#include <type_traits>
#include <vector>
#include <boost/make_shared.hpp>

#include <digitalSTROM/dsuid.h>

#include <boost/locale/generator.hpp>
#include "ds/locale/ds-locale.h"
#include <device-info.h>

#include <ds/log.h>
#include "businterface.h"
#include "model/modelconst.h"
#include "vdc-db.h"
#include "event.h"
#include "dss.h"
#include "ds485types.h"
#include "security/security.h"
#include "ds/str.h"

#include "apartment.h"
#include "cluster.h"
#include "device.h"
#include "ds485-device.h"
#include "ds485/dsbusinterface.h"
#include "dsmeter.h"
#include "group.h"
#include "handler/system_states.h"
#include "model-features.h"
#include "modelmaintenance.h"
#include "state.h"
#include "vdc-connection.h"
#include "vdc-element-reader.h"
#include "zone.h"

#define HEATING_MAX_SENSOR_AGE (60 * 60)
#define SENSOR_MAX_AGE 0xffffffff

static ds::LogChannel dsLogChannel("modelBusScanner");

namespace dss {

  static boost::mutex s_busScannerMutex;

  BusScanner::BusScanner(StructureQueryBusInterface& interface, Apartment& apartment, ModelMaintenance& maintenance)
      : m_Apartment(apartment),
        m_Interface(interface),
        m_Maintenance(maintenance),
        m_structureManipulator(*m_Apartment.getBusInterface()->getStructureModifyingBusInterface(),
            *m_Apartment.getBusInterface()->getStructureQueryBusInterface(), m_Apartment) {}

  bool BusScanner::scanDSMeter(boost::shared_ptr<DSMeter> dsm) {
    dsm->setIsPresent(true);
    dsm->setIsConnected(true);
    dsm->setIsValid(false);

    DSMeterHash_t hash;
    if (!getMeterHash(dsm, hash)) {
      if (!applyMeterSpec(dsm)) {
        // could not get any info about device. We have to retry.
        return false;
      }
      // bus member type can be read. We have to retry for interesting bus devices.
      if (busMemberIsDsmApiDSM(dsm->getBusMemberType())) {
        // for known bus device types retry the readout
        return false;
      }
      // for unknown/uninteresting bus devices
      dsm->setIsValid(true);
      return true;
    }

    uint32_t dsmHash, dsmModCount;
    dsmHash = dsm->getDatamodelHash();
    dsmModCount = dsm->getDatamodelModificationCount();

    log("Scan dS485 device " + dsuid2str(dsm->getDSID()) +
        ": Initializing " + intToString(m_Maintenance.isInitializing()) +
        ", Init Model " + intToString(dsm->isInitialized()) +
        ", Hash Model/dSM " + intToString(dsmHash, true) + "/" + intToString(hash.Hash, true) +
        ", ModCount Model/dSM " + intToString(dsmModCount, true) + "/" + intToString(hash.ModificationCount, true) +
        ", ds Meter Type " + intToString(dsm->getBusMemberType()), lsInfo);

    if (applyMeterSpec(dsm)) {
      if ((dsm->getApiVersion() > 0) && (dsm->getApiVersion() < 0x303)) {
        log("scanDSMeter: dSMeter is incompatible", lsWarning);
        dsm->setDatamodelHash(hash.Hash);
        dsm->setDatamodelModificationcount(hash.ModificationCount);
        dsm->setIsPresent(false);
        dsm->setIsConnected(false);
        dsm->setIsValid(true);
        return true;
      }
    } else {
      log("scanDSMeter: Error getting dSMSpecs", lsWarning);
      return false;
    }

    // update powerline jumble state
    uint8_t state = DSM_STATE_UNKNOWN;
    if (busMemberIsHardwareDSM(dsm->getBusMemberType())) {
      if (getMeterState(dsm, &state)) {
        dsm->setState(state);
      }
    } else if (busMemberIsDsmApiDSM(dsm->getBusMemberType())) {
      dsm->setState(DSM_STATE_IDLE);
    } else if (!dsm->isSynchonized()) {
      log("scanDSMeter: dSMeter is not yet synchronized. Meter: " +
        dsuid2str(dsm->getDSID()), lsWarning);
      return false;
    }

    if (m_Maintenance.isInitializing() ||
        (dsm->isInitialized() == false) ||
        (hash.Hash != dsmHash) ||
        (hash.ModificationCount != dsmModCount)) {

      setMeterCapability(dsm);

      if (dsm->getCapability_HasDevices()) {
        std::vector<int> zoneIDs;
        try {
          zoneIDs = m_Interface.getZones(dsm->getDSID());
        } catch(BusApiError& e) {
          log("scanDSMeter: Error getting ZoneIDs", lsWarning);
          return false;
        }
        for (int zoneID : zoneIDs) {
          log("scanDSMeter:  Found zone with id: " + intToString(zoneID));
          boost::shared_ptr<Zone> zone = m_Apartment.allocateZone(zoneID);
          if (zoneID == 0) {
            try {
              scanStatusOfZone(dsm, zone);
            } catch(BusApiError& e) {
              log("scanDSMeter: Error scanStatusOfZone: " + std::string(e.what()), lsWarning);
            }
            continue;
          }
          zone->addToDSMeter(dsm);
          zone->setIsPresent(true);
          zone->setIsConnected(true);
          if (!scanZone(dsm, zone)) {
            return false;
          }
        }

        // scan groups and status of apartment zone "0", but not devices or temperature control
        boost::shared_ptr<Zone> zone = m_Apartment.allocateZone(0);
        try {
          scanGroupsOfZone(dsm, zone);
          // TODO(someday): read scene history
        } catch(BusApiError& e) {
          log("scanDSMeter: error scanning zone 0: " + std::string(e.what()), lsWarning);
        }

        scanClusters(dsm);
        scanPowerStates(dsm);
      }

      dsm->setIsInitialized(true);
      dsm->setIsUpToDate(true);
      m_Maintenance.addModelEvent(new ModelEvent(ModelEvent::etModelDirty));

    } else {

      log("Doing a quick scan for: " + dsuid2str(dsm->getDSID()), lsInfo);
      if (dsm->getCapability_HasDevices()) {
        std::vector<int> zoneIDs;
        try {
          zoneIDs = m_Interface.getZones(dsm->getDSID());
        } catch(BusApiError& e) {
          log("scanDSMeter: Error getting ZoneIDs", lsWarning);
          return false;
        }
        for (int zoneID : zoneIDs) {
          boost::shared_ptr<Zone> zone = m_Apartment.allocateZone(zoneID);
          zone->addToDSMeter(dsm);
          zone->setIsPresent(true);
          zone->setIsConnected(true);
          scanDevicesOfZoneQuick(dsm, zone);
        }
      }

    }

    dsm->setDatamodelHash(hash.Hash);
    dsm->setDatamodelModificationcount(hash.ModificationCount);
    dsm->setIsValid(true);

    // event counter support first implemented in v2.6
    if ((dsm->getApiVersion() > 0) && (dsm->getApiVersion() < 0x206)) {
      hash.EventCount = 0;
    }
    dsm->setHasPendingEvents(((hash.EventCount == 0) ||
        (dsm->getBinaryInputEventCount() != hash.EventCount)));
    dsm->setBinaryInputEventCount(hash.EventCount);
    dsm->checkUpToDate(); //Check if newer version is available
    return true;
  } // scanDSMeter

  void BusScanner::scanDevicesOfZoneQuick(boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone) {
    std::vector<DeviceSpec> devices;
    try {
      if ((dsm->getApiVersion() > 0) && (dsm->getApiVersion() < 0x200)) {
        log("scanZone: dSMeter " + dsuid2str(dsm->getDSID()) + " is incompatible", lsWarning);
      } else {
        devices = m_Interface.getDevicesInZone(dsm->getDSID(), zone->getID(), false);
        for (DeviceSpec& spec : devices) {
          initializeDeviceFromSpecQuick(dsm, zone, spec);
        }
      }
    } catch(BusApiError& e) {
      log("scanZone: Error getDevicesInZone: " + std::string(e.what()), lsWarning);
    }
  } // scanDevicesOfZone

  bool BusScanner::scanZone(boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone) {
    std::vector<DeviceSpec> devices;
    bool result = true;
    try {
      if ((dsm->getApiVersion() > 0) && (dsm->getApiVersion() < 0x200)) {
        log("scanZone: dSMeter " + dsuid2str(dsm->getDSID()) + " is incompatible", lsWarning);
      } else {
        devices = m_Interface.getDevicesInZone(dsm->getDSID(), zone->getID());
        for (DeviceSpec& spec : devices) {
          initializeDeviceFromSpec(dsm, zone, spec);
        }
      }
    } catch(BusApiError& e) {
      log("scanZone: Error getDevicesInZone: " + std::string(e.what()), lsWarning);
      result = false;
    }

    try {
      result = result && scanGroupsOfZone(dsm, zone);
    } catch(BusApiError& e) {
      log("scanZone: Error scanGroupsOfZone: " + std::string(e.what()), lsWarning);
    }

    try {
      result = result && scanStatusOfZone(dsm, zone);
    } catch(BusApiError& e) {
      log("scanZone: Error scanStatusOfZone: " + std::string(e.what()), lsWarning);
    }

    return result;
  } // scanZone

  bool BusScanner::scanPowerStates(boost::shared_ptr<DSMeter> dsm) {
    std::vector<CircuitPowerStateSpec_t> powerStates;

    if (!dsm->getCapability_HasMetering()) {
      return true;
    }
    try {
      powerStates = m_Interface.getPowerStates(dsm->getDSID());
      dsm->setPowerStates(powerStates);
    } catch(BusApiError& e) {
      log("scanZone: Error scanPowerStates: " + std::string(e.what()), lsWarning);
    }

    return true;
  }

  bool BusScanner::scanClusters(boost::shared_ptr<DSMeter> dsm) {
    std::vector<ClusterSpec_t> clusters;
    try {
      clusters = m_Interface.getClusters(dsm->getDSID());
    } catch (BusApiError& e) {
      log("scanDSMeter: Error getting getClusters", lsWarning);
      return false;
    }

    for (ClusterSpec_t cluster : clusters) {
      // in case this DSM does not provide group configuration ignore it.
      if (dsm->getApiVersion() < 0x303) {
        cluster.applicationConfiguration = 0;
      }

      if (cluster.applicationType != ApplicationType::None) {
        log("scanDSMeter:    Found cluster with id: " + intToString(cluster.GroupID) +
            " and devices: " + intToString(cluster.NumberOfDevices));
      }

      boost::shared_ptr<Cluster> pCluster;

      try {
        DS_ASSERT(isAppUserGroup(cluster.GroupID), "");
        pCluster = m_Apartment.getCluster(cluster.GroupID);
      } catch (ItemNotFoundException&) {
        boost::shared_ptr<Zone> zoneBroadcast = m_Apartment.getZone(0);
        pCluster.reset(new Cluster(cluster.GroupID, m_Apartment));
        zoneBroadcast->addGroup(pCluster);
      }

      /*
       * if the configuration on the dSS and the dSM is different,
       * synchronize the settings back to the dSMs
       */
      if (!pCluster->isConfigEqual(cluster)) {
        pCluster->setIsSynchronized(false);
      }

      /*
       * the dSM is the master source for this data. The dSS will take the data
       * from the first dSM that has a non-zero configuration.
       */
      if ((pCluster->getApplicationType() == ApplicationType::None) ||
          ((pCluster->getApplicationType() != ApplicationType::None) && !pCluster->isReadFromDsm())) {
        pCluster->setFromSpec(cluster);
        pCluster->setReadFromDsm(true);
      }

      try {
        ActionRequestInterface &actionRequest(*m_Apartment.getActionRequestInterface());
        if (actionRequest.isOperationLock(dsm->getDSID(), cluster.GroupID)) {
          // only set operation lock if lock is really active,
          // otherwise a state will be created, which is not wanted
          pCluster->setOperationLock(true, coSystemStartup);
        }
      } catch (BusApiError &e) {
        // leave State_Unknown or take lock value from other dsm's
        log(std::string("OperationLock: ") + e.what(), lsWarning);
      }

      pCluster->setIsPresent(true);
      pCluster->setIsConnected(true);
      pCluster->setLastCalledScene(SceneOff);
      pCluster->setIsValid(true);
    }
    return true;
  }

  bool BusScanner::scanDeviceOnBus(boost::shared_ptr<DSMeter> dsm, devid_t id) {
    if ((dsm->getApiVersion() > 0) && (dsm->getApiVersion() < 0x200)) {
      log("scanDeviceOnBus: dSMeter " + dsuid2str(dsm->getDSID()) + " is incompatible", lsWarning);
      return false;
    }
    DeviceSpec spec = m_Interface.deviceGetSpec(id, dsm->getDSID());
    boost::shared_ptr<Zone> pZone = m_Apartment.allocateZone(spec.ZoneID);
    return initializeDeviceFromSpec(dsm, pZone, spec);
  } // scanDeviceOnBus

  bool BusScanner::initializeDeviceFromSpec(
      boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone, const DeviceSpec& spec) {

    // protect against parallel execution triggered from meterMaintenance (dsMeterReady) and
    // modelMaintenance (DeviceAccessibilityEvent) threads
    boost::lock_guard<boost::mutex> lock(s_busScannerMutex);

    boost::shared_ptr<Device> dev;
    try {
      dev = m_Apartment.getDeviceByDSID(spec.DSID);
      if ((spec.ActiveState == 0) && dev->isPresent()) {
        // ignore this device if there is already an active one with this dSID
        return false;
      }
    } catch(ItemNotFoundException&) {
      dev = m_Apartment.allocateDevice(spec.DSID);
    }
    DS_CONTEXT(dev);
    DS_INFO("Initialize device ", dsm, spec);

    if (dsm && dsm->getBusMemberType() == BusMember_vDC) {
      dev->setVdcDevice(true);
    }

    DeviceReference devRef(dev, &m_Apartment);

    // remove from old dsMeter
    try {
      boost::shared_ptr<DSMeter> oldDSMeter = m_Apartment.getDSMeterByDSID(dev->getLastKnownDSMeterDSID());
      oldDSMeter->removeDevice(devRef);
    } catch(std::runtime_error&) {
    }

    // remove from groups
    dev->resetGroups();

    // remove from old zone
    if(dev->getZoneID() != 0) {
      try {
        boost::shared_ptr<Zone> oldZone = m_Apartment.getZone(dev->getZoneID());
        oldZone->removeDevice(devRef);
        // TODO: check if the zone is empty on the dsMeter and remove it in that case
      } catch(std::runtime_error&) {
      }
    }

    dev->setShortAddress(spec.ShortAddress);
    dev->setDSMeter(dsm);
    dev->setZoneID(zone->getID());
    dev->setPartiallyFromSpec(spec);
    // TODO(someday): move more setters into setSpec

    dev->setButtonActiveGroup(spec.buttonActiveGroup);
    dev->setButtonSetsLocalPriority(spec.buttonSetsLocalPriority);
    dev->setButtonCallsPresent(spec.buttonCallsPresent);
    dev->setButtonID(spec.buttonID);

    uint8_t inputCount = 0;
    uint8_t inputIndex = 0;
    if ((spec.FunctionID & 0xffc0) == 0x1000) {
      // light class + dS-standard R100
      switch(spec.FunctionID & 0x7) {
      case 0: inputCount = 1; break;
      case 1: inputCount = 2; break;
      case 2: inputCount = 4; break;
      case 7: inputCount = 0; break;
      }
    } else if (((spec.FunctionID & 0x0fc0) == 0x0100) || ((spec.FunctionID & 0x0fc0) == 0x01c0)) {
      // dS-standard R105 / with group membership assignment v2
      switch(spec.FunctionID & 0x3) {
      case 0: inputCount = 0; break;
      case 1: inputCount = 1; break;
      case 2: inputCount = 2; break;
      case 3: inputCount = 4; break;
      }
    }
    if(inputCount > 1) {
        if (dev->isVdcDevice()) {
            if (inputCount != 0) {
              inputIndex = spec.DSID.id[DSUID_SIZE - 1] % inputCount;
              dev->setPairedDevices(inputCount);
            }
        } else {
          uint32_t serial;
          if (dsuid_get_serial_number(&spec.DSID, &serial) == 0) {
            inputIndex = (uint8_t)(serial & 0xff) % inputCount;
          }
        }
    }
    dev->setButtonInputMode(spec.LTMode);
    dev->setButtonInputCount(inputCount);
    dev->setButtonInputIndex(inputIndex);

    if(dev->getName().empty()) {
      dev->setName(spec.Name);
    }

    for (int groupID : spec.Groups) {
      if (groupID == 0) {
        continue;
      }
      dev->addToGroup(groupID);
    }

    // set binary input configuration
    if (spec.binaryInputsValid) {
      dev->setBinaryInputs(spec.binaryInputs);
    }
    // set sensor input configuration
    if (spec.sensorInputsValid) {
      dev->setSensors(spec.sensorInputs);
    }

    if ((dev->getBinaryInputCount() > 0) &&
        (!dev->isInGroup(dev->getBinaryInput(0)->m_targetGroupId))) {
      /* group is only added in dSS datamodel, not on the dSM and device */
      dev->addToGroup(dev->getBinaryInput(0)->m_targetGroupId);
    }

    struct StructureQueryBusInterface::VdcInputDesc iDesc;

    try {
      auto db = vdcDb::make(*DSS::getInstance());

      // cannot use getOemEanNumber here, the OEM/Data is not available/valid at this point
      std::string eanString = dev->getVdsdSpec().oemModelGuid;
      if (beginsWith(eanString, "gs1:(01)")) {
        eanString.erase(0, 8);
      }

      if (dev->isVdcDevice()) {
        {
          auto vdsdSpec = VdcHelper::getSpec(dev->getDSMeterDSID(), dev->getDSID());
          // enocean valve hack
          if (dev->isValveDevice() &&
              m_Apartment.getDSMeterByDSID(dev->getDSMeterDSID())->getVdcImplementationId() ==
                  "EnOcean_Bus_Container") {
            if (std::find(vdsdSpec.modelFeatures.begin(), vdsdSpec.modelFeatures.end(),
                    ModelFeatureId::outmodeenoceanvalve) == vdsdSpec.modelFeatures.end()) {
              vdsdSpec.modelFeatures.insert(ModelFeatureId::outmodeenoceanvalve);
            }
          }
          dev->setVdsdSpec(std::move(vdsdSpec));
        }

        const auto& vdsdSpec = dev->getVdsdSpec();
        iDesc = m_Interface.getVdcInputDesc(dev->getDSMeterDSID(), dev->getDSID());

        // climate settings are optional properties: dont fail on non-existing properties
        try {
          VdcHelper::ClimateDesc climaSpec = VdcHelper::getClimateSettings(dev->getDSMeterDSID(), dev->getDSID());
          dev->setActiveCoolingMode(climaSpec.activeCoolingMode);
          dev->setValveType(static_cast<DeviceValveType_t>(climaSpec.heatingSystemType));
        } catch (const std::exception& e) {
          DS_ERROR("Failed to initialize climate settings.", e);
        }

        try {
          ModelFeatures::getInstance()->setFeatures(dev->getDeviceClass(), vdsdSpec.modelUID,
              boost::make_shared<std::vector<ModelFeatureId>>(
                  vdsdSpec.modelFeatures.begin(), vdsdSpec.modelFeatures.end()));
        } catch (const std::exception& e) {
          DS_ERROR("Failed to register model features.", e);
        }

        dev->initStates(db->getStatesLegacy(eanString));       // throws
        dev->setHasActions(db->hasActionInterface(eanString)); // throws

        // #21334 "old" vdc's did not support action templates and reported standard actions as
        // action descriptions. When creating custom actions the DSS needs to distinguish between
        // old and new style and enforce compatibility.
        auto&& vdcActionsMsg = m_Interface.getVdcProperty(
            dev->getDSMeterDSID(), dev->getDSID(), { "deviceActionDescriptions" });
        auto&& vdcActions = VdcElementReader(vdcActionsMsg.vdc_response_get_property().properties())["deviceActionDescriptions"];
        for (auto&& vdcAction : vdcActions) {
          auto name = vdcAction.getName();
          if (beginsWith(name, "std.")) {
            dev->setHasLegacyActions(true);
            break;
          }
        }
      }
      auto&& dss = *DSS::getInstance();
      auto locale = dss.localeGenerator().generate(ds::locale::sanitizeLanguage("en_US"));
      JSONWriter json;
      auto deviceInfo =
          DeviceInfo(DeviceInfo::Prefer(), *m_Apartment.getBusInterface()->getStructureQueryBusInterface(), *db,
                     dev.get(), eanString, locale, json, ds::system());
      dev->setScenarios(deviceInfo.scenarios());
      dev->setStatesStatus(deviceInfo.deviceStatesStatus());
      dev->setStates(deviceInfo.deviceStates());
    } catch (const std::exception& e) {
      DS_ERROR("Failed to initialize vdc device.", e);
      // TODO(someday): device is not correctly initialized.
      // We should throw here and fix the model discovery logic.
      // It may be enough to catch std::runtime_error instead of BusError.
    }

    // set vdc input description properties
    try {
      if (iDesc.m_binaryInputs.size() > 0) {
        auto&& sList = iDesc.m_binaryInputs;
        for (auto it = sList.begin(); it != sList.end(); it ++) {
          dev->setBinaryInputsInfo(it->second.dsIndex, it->first, it->second.inputName);
        }
      }
      if (iDesc.m_sensorInputs.size() > 0) {
        auto&& sList = iDesc.m_sensorInputs;
        for (auto it = sList.begin(); it != sList.end(); it ++) {
          dev->setSensorsInfo(it->second.index, it->first, it->second.sensorName, it->second.sensorUsage);
        }
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to read sensor channel.", e);
    }

    // synchronize output channel configuration
    if (spec.outputChannelsValid) {
      dev->setOutputChannels(spec.outputChannels);
      if (dev->isVdcDevice()) {
        try {
          // read extended channel info
          auto channelList = VdcHelper::getChannelDesc(dev->getDSMeterDSID(), dev->getDSID());
          for (uint8_t i = 0; i < dev->getOutputChannelCount(); i++) {
            for (auto it = channelList.begin(); it != channelList.end(); it ++) {
              if (it->second.index == i) {
                Channel chInfo;
                chInfo.index = i;
                chInfo.id = it->first;
                chInfo.technicalName = it->second.name;
                chInfo.type = it->second.type;
                dev->setOutputChannelInfo(i, chInfo);
                DS_NOTICE("Output channel:", chInfo);
                break;
              }
            }
          }
        } catch (const std::exception& e) {
          DS_ERROR("Failed to read output channel.", e);
        }
      }
    }

    // at this we should have enough information to decide if button usage
    // is supported on a device or not, as specified, we will follow the UI
    // logic here
    if (dev->getButtonInputCount() > 0) {
      if ((dev->getDeviceClass() == DEVICE_CLASS_GR) &&
         (dev->getDeviceType() == DEVICE_TYPE_KL) && // HKL-230
         (dev->getDeviceNumber() == 230)) {
        dev->setButtonUsage(ButtonUsage::manual_unused); // by requirement
      } else if ((dev->getDeviceClass() == DEVICE_CLASS_BL) &&
                 (dev->getDeviceType() == DEVICE_TYPE_KM) &&
                 (dev->getDeviceNumber() == 200)) {
          if (dev->isInGroup(GroupIDControlTemperature)) {
            dev->setButtonUsage(ButtonUsage::manual_unused); // by requirement
          } else {
            dev->setButtonUsage(ButtonUsage::auto_unused);
          }
      } else if ((dev->getDeviceClass() == DEVICE_CLASS_SW) &&
                 (dev->getDeviceType() == DEVICE_TYPE_KL) && // SSL-200
                 (dev->getDeviceNumber() == 213)) {
        dev->setButtonUsage(ButtonUsage::manual_unused); // by requirement
      }

      if (dev->hasPushbutton() &&
          ((dev->getRevisionID() >= 0x0314) || (dev->isVdcDevice()))) {
        // do not overwrite button usage value set by apartment.xml
        if (dev->getButtonUsage() == ButtonUsage::unsupported) {
          // if the device is configured as binary input, then usage stays
          // unsupported
          if (!dev->isBinaryInput()) {
            dev->setButtonUsage(ButtonUsage::auto_unused);
          }
        }
      }
    }

    zone->addToDSMeter(dsm);
    zone->addDevice(devRef);
    dsm->addDevice(devRef);
    dev->setIsPresent(spec.ActiveState == 1);
    dev->setIsConnected(true);
    dev->setIsValid(true);

    if ((dev->getRevisionID() >= TBVersion_OemConfigLock) &&
        (dev->getOemInfoState() == DEVICE_OEM_UNKNOWN)){
        // will be reset in OEM Readout
        dev->setConfigLock(true);
    }


    {
      boost::shared_ptr<Event> readyEvent = boost::make_shared<Event>(EventName::DeviceEvent, dev->reference());
      readyEvent->setProperty("action", "ready");
      if(DSS::hasInstance()) {
        DSS::getInstance()->getEventQueue().pushEvent(readyEvent);
      }
    }

    if (((dev->getDeviceType() == DEVICE_TYPE_AKM) ||
         (dev->getDeviceType() == DEVICE_TYPE_TKM)) &&
        (dev->getDeviceClass() == DEVICE_CLASS_SW)) {
      if (dev->getDeviceNumber() == 200) {
        dev->setPairedDevices(4);
      } else if (dev->getDeviceNumber() == 210) {
        dev->setPairedDevices(2);
      }
    } else if (dev->getDeviceType() == DEVICE_TYPE_SDS) {
      dev->setPairedDevices(2);
    } else if ((dev->getDeviceType() == DEVICE_TYPE_SK) &&
               (dev->getDeviceClass() == DEVICE_CLASS_BL) &&
               (dev->getDeviceNumber() == 204)) {
      dev->setPairedDevices(2);
    }

    scheduleDeviceReadout(dev);

    m_Maintenance.addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
    return true;
  } // scanDeviceOnBus

  bool BusScanner::initializeDeviceFromSpecQuick(
      boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone, const DeviceSpec& spec) {
    log("UpdateDevice: DSUID:" + dsuid2str(spec.DSID) + "active:" + intToString(spec.ActiveState), lsInfo);

    boost::shared_ptr<Device> dev;
    try {
      dev = m_Apartment.getDeviceByDSID(spec.DSID);
      if ((spec.ActiveState == 0) && dev->isPresent()) {
        // ignore this device if there is already an active one with this dSID
        return false;
      }
    } catch(ItemNotFoundException&) {
      return false;
    }

    boost::shared_ptr<DSMeter> oldDSMeter;
    try {
      oldDSMeter = m_Apartment.getDSMeterByDSID(dev->getLastKnownDSMeterDSID());
    } catch(ItemNotFoundException&) {
      return false;
    }

    // compare ds meter. if attached on wrong dsmeter
    // -> reattach to correct one given by dsMeter
    // reference temporary
    if (oldDSMeter->getDSID() != dsm->getDSID()) {
      return initializeDeviceFromSpec(dsm, zone, spec);
    }

    dev->setIsPresent(spec.ActiveState == 1);
    dev->setIsConnected(true);
    dev->setIsValid(true);

    m_Maintenance.addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
    return true;
  } // initializeDeviceFromSpecQuick

  void BusScanner::scheduleDeviceReadout(const boost::shared_ptr<Device> device) {
    if (device->isPresent() && (device->getOemInfoState() == DEVICE_OEM_UNKNOWN)) {
      if (device->getRevisionID() >= TBVersion_OemEanConfig) {
        log("scheduleOEMReadout: schedule EAN readout for: " +
            dsuid2str(device->getDSID()));
        boost::shared_ptr<DSDeviceBusInterface::OEMDataReader> task;
        std::string connURI = m_Apartment.getBusInterface()->getConnectionURI();
        task = boost::make_shared<DSDeviceBusInterface::OEMDataReader>(connURI);
        task->setup(device);
        m_Apartment.getModelMaintenance()->scheduleDeviceReadout(device->getDSMeterDSID(), task);
        device->setOemInfoState(DEVICE_OEM_LOADING);
      } else {
        device->setOemInfoState(DEVICE_OEM_NONE);
        device->setConfigLock(false);
      }
    } else if (device->isPresent() &&
              (device->getOemInfoState() != DEVICE_OEM_UNKNOWN) &&
              (device->getOemInfoState() != DEVICE_OEM_LOADING) &&
              ((device->getDeviceType() == DEVICE_TYPE_TNY) ||
               ((device->getDeviceType() == DEVICE_TYPE_SK) &&
               (device->getDeviceNumber() == 204)))) {
      // this is an optimization for TNY and SK-204, the visibility flag
      // at bank 1 / 0x1f is also retrieved by the OEM reader, so when the
      // OEM data is being read, we will get the visibility from there.
      // However, OEM data is only read once per device, so on all
      // subsequent scans we need to read out the visibility explicitly
      boost::shared_ptr<DSDeviceBusInterface::TNYConfigReader> task;
      std::string connURI = m_Apartment.getBusInterface()->getConnectionURI();
      task = boost::make_shared<DSDeviceBusInterface::TNYConfigReader>(connURI);
      task->setup(device);
      m_Apartment.getModelMaintenance()->scheduleDeviceReadout(device->getDSMeterDSID(), task);
    } else if (device->isPresent() &&
                (device->getOemInfoState() == DEVICE_OEM_VALID) &&
                ((device->getOemProductInfoState() != DEVICE_OEM_VALID) &&
                 (device->getOemProductInfoState() != DEVICE_OEM_LOADING))) {
      TaskProcessor &tp(m_Apartment.getModelMaintenance()->getTaskProcessor());
      tp.addEvent(boost::make_shared<ModelMaintenance::OEMWebQuery>(WebserviceConnection::getInstanceMsHub(), device, device->getOemProductInfoState()));
      device->setOemProductInfoState(DEVICE_OEM_LOADING);
    }

    // read properties of virtual devices connected to a VDC
    boost::shared_ptr<DSMeter> pMeter;
    try {
      pMeter = m_Apartment.getDSMeterByDSID(device->getDSMeterDSID());
    } catch (ItemNotFoundException& e) {
    }
    if (pMeter && pMeter->getBusMemberType() == BusMember_vDC) {
      TaskProcessor &tp(m_Apartment.getModelMaintenance()->getTaskProcessor());
      tp.addEvent(boost::make_shared<ModelMaintenance::VdcDataQuery>(device));
    }
  }

  void BusScanner::setMeterCapability(boost::shared_ptr<DSMeter> dsm) {
    switch (dsm->getBusMemberType()) {
      case BusMember_dSM11:
        {
          dsm->setCapability_HasDevices(true);
          dsm->setCapability_HasMetering(true);
          dsm->setCapability_HasTemperatureControl(false);
          dsm->setCapability_HasBlinking(true);
        }
        break;
      case BusMember_dSM12:
      case BusMember_dSM20:
      case BusMember_dSM25:
        {
          dsm->setCapability_HasDevices(true);
          dsm->setCapability_HasMetering(true);
          dsm->setCapability_HasTemperatureControl(true);
          dsm->setCapability_HasBlinking(true);
        }
        break;
      case BusMember_vDSM:
        {
          dsm->setCapability_HasDevices(false);
          dsm->setCapability_HasMetering(false);
          dsm->setCapability_HasTemperatureControl(true);
          dsm->setCapability_HasBlinking(false);
        }
        break;
      case BusMember_vDC:
        {
          boost::shared_ptr<VdcSpec> props;
          try {
            props = VdcHelper::getVdcSpec(dsm->getDSID());
            if (props) {
              dsm->setCapability_HasMetering(props->capabilities.metering);
              dsm->setCapability_HasBlinking(props->capabilities.identification);
              dsm->setCapability_HasDynamicDefinitions(props->capabilities.dynamicDefinitions);
              dsm->setHardwareVersion(props->hardwareVersion);
              dsm->setHardwareName(props->model);
              dsm->setSoftwareVersion(props->modelVersion);
              dsm->setVdcModelUID(props->modelUID);
              dsm->setVdcConfigURL(props->configURL);
              dsm->setVdcHardwareGuid(props->hardwareGuid);
              dsm->setVdcHardwareModelGuid(props->hardwareModelGuid);
              dsm->setVdcImplementationId(props->implementationId);
              dsm->setVdcVendorGuid(props->vendorGuid);
              dsm->setVdcOemGuid(props->oemGuid);
              dsm->setVdcOemModelGuid(props->oemModelGuid);
              if (dsm->getName().empty()) {
                dsm->setName(props->name);
              }
            }
          } catch (const std::exception& e) {
            DS_ERROR(e);
          }
          dsm->setCapability_HasDevices(true);
          dsm->setCapability_HasTemperatureControl(true); // transparently trapped by the vdSM
        }
        break;
      case BusMember_backendVdc:
        // bypasses busscanner
        break;
      case BusMember_Unknown:
      case BusMember_ds485d:
      case BusMember_ds485p:
      case BusMember_dSF25:
      case BusMember_dSS:
      case BusMember_dSA:
      case BusMember_dSG:
      default:
        {
          dsm->setCapability_HasDevices(false);
          dsm->setCapability_HasMetering(false);
          dsm->setCapability_HasTemperatureControl(false);
          dsm->setCapability_HasBlinking(false);
        }
        break;
    }
  }

  bool BusScanner::applyMeterSpec(boost::shared_ptr<DSMeter> dsm) {
    try {
      DSMeterSpec_t spec;
      spec = m_Interface.getDSMeterSpec(dsm->getDSID());
      synchronizeDSMeterData(dsm, spec);
    } catch(BusApiError& e) {
      log("applyMeterSpec: Error getting dSMSpecs", lsWarning);
      return false;
    }
    return true;
  }

  bool BusScanner::getMeterHash(boost::shared_ptr<DSMeter> dsm, DSMeterHash_t& hash) {
    try {
      hash = m_Interface.getDSMeterHash(dsm->getDSID());
    } catch(BusApiError& e) {
      log("getMeterHash " + dsuid2str(dsm->getDSID()) +
          ": getDSMeterHash Error: " + e.what(), lsWarning);
      return false;
    }
    return true;
  }

  bool BusScanner::getMeterState(boost::shared_ptr<DSMeter> dsm, uint8_t *state) {
    try {
      m_Interface.getDSMeterState(dsm->getDSID(), state);
    } catch(BusApiError& e) {
      log("getMeterHash " + dsuid2str(dsm->getDSID()) +
          ": getDSMeterHash Error: " + e.what(), lsWarning);
      return false;
    }
    return true;
  }

  bool BusScanner::scanGroupsOfZone(boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone) {
    std::vector<GroupSpec_t> groups;
    try {
      groups = m_Interface.getGroups(dsm->getDSID(), zone->getID());
    } catch (BusApiError& e) {
      log("scanDSMeter: Error getting getGroups", lsWarning);
      return false;
    }

    for (GroupSpec_t& group : groups) {
      if ((group.GroupID == 0) || isAppUserGroup(group.GroupID)) {
        // ignore broadcast group and clusters
        continue;
      }

      // in case this DSM does not provide group configuration it is invalid and should be ignored
      if (dsm->getApiVersion() < 0x303) {
        group.applicationConfiguration = 0;
      }

      log("scanDSMeter:    Found group with id: " + intToString(group.GroupID) +
          " and devices: " + intToString(group.NumberOfDevices));

      // apartment-wide unique standard- and user-groups published in zone<0>
      boost::shared_ptr<Zone> zoneBroadcast = m_Apartment.getZone(0);

      if (isDefaultGroup(group.GroupID)) {
        boost::shared_ptr<Group> zoneGroup = zone->tryGetGroup(group.GroupID).lock();

        // TODO(someday): re-implement the configuration synchronization logic for default groups
        if (zoneGroup == NULL) {
          // note: This should never happen as default groups are created during zone allocation
          log(" scanDSMeter:    Adding new group to zone");
          zoneGroup = Group::make(group, zone);
          zone->addGroup(zoneGroup);
        }
        zoneGroup->setIsPresent(true);
        zoneGroup->setIsConnected(true);
        zoneGroup->setLastCalledScene(SceneOff);
        zoneGroup->setIsValid(true);

        boost::shared_ptr<Group> broadcastGroup = zoneBroadcast->tryGetGroup(group.GroupID).lock();

        if (broadcastGroup == NULL) {
          // note: This should never happen as default groups are created during zone allocation
          log(" scanDSMeter:    Adding new group to broadcast zone");
          broadcastGroup = Group::make(group, zoneBroadcast);
          zoneBroadcast->addGroup(broadcastGroup);
        }
        broadcastGroup->setIsPresent(true);
        broadcastGroup->setIsConnected(true);
        broadcastGroup->setLastCalledScene(SceneOff);
        broadcastGroup->setIsValid(true);
      } else if (isGlobalAppGroup(group.GroupID)) {
        boost::shared_ptr<Group> globalAppGroup = zoneBroadcast->tryGetGroup(group.GroupID).lock();

        if (globalAppGroup == NULL) {
          const auto& applicationType = group.applicationType;
          if (ds::isValidEnum(applicationType)) {
            log(" scanDSMeter:    Adding new apartment application group");
            globalAppGroup = Group::make(group, zoneBroadcast);
            globalAppGroup->setReadFromDsm(true);
            zoneBroadcast->addGroup(globalAppGroup);
          } else {
            // Not logged as error because it happens regularly
            /// and this error recovery is correct.
            // Shall be fixed in dsm.
            DS_NOTICE("Ignoring group with unknown application type.", applicationType);
            continue;
          }
        } else if (!globalAppGroup->isConfigEqual(group)) {
          if (!globalAppGroup->isReadFromDsm()) {
            // First DSM is owner of Apartment Application group
            globalAppGroup->setFromSpec(group);
            globalAppGroup->setReadFromDsm(true);
          } else {
            // All other DSMs need to be overwritten in synchronizeGroups() function
            globalAppGroup->setIsSynchronized(false);
          }
        }

        globalAppGroup->setIsPresent(true);
        globalAppGroup->setIsConnected(true);
        globalAppGroup->setLastCalledScene(SceneOff);
        globalAppGroup->setIsValid(true);
      } else {
        boost::shared_ptr<Group> userGroup = zone->tryGetGroup(group.GroupID).lock();

        if (userGroup == NULL) {
          log(" scanDSMeter:    Adding new group to zone");
          userGroup = Group::make(group, zone);
          zone->addGroup(userGroup);
        } else {
          // DSS is the owner of user groups configuration we will overwrite in DSM
          if (!userGroup->isConfigEqual(group)) {
            userGroup->setIsSynchronized(false);
          }
        }
        userGroup->setIsPresent(true);
        userGroup->setIsConnected(true);
        userGroup->setLastCalledScene(SceneOff);
        userGroup->setIsValid(true);
      }
    }
    return true;
  } // scanGroupsOfZone

  bool BusScanner::scanStatusOfZone(boost::shared_ptr<DSMeter> dsm, boost::shared_ptr<Zone> zone) {
    std::vector<std::pair<int, int> > sceneHistory = m_Interface.getLastCalledScenes(dsm->getDSID(), zone->getID());
    std::vector<std::pair<int, int> >::iterator it;
    for (it = sceneHistory.begin(); it < sceneHistory.end(); it++) {
      int gid = it->first & 0x3f;
      bool isUndo = (it->first & 0x80) > 0;
      bool isForced = (it->first & 0x40) > 0;
      log("Scene History in Zone " + intToString(zone->getID()) +
          ": group " + intToString(gid) + " - " +
          (isUndo? "undo-scene" : (isForced ? "forced-scene" : "scene")) + " " + intToString(it->second));
      if (gid == 0) {
        std::vector<boost::shared_ptr<Group> > pGroups = zone->getGroups();
        for (boost::shared_ptr<Group> pGroup : pGroups) {
          if (isUndo) {
            pGroup->setLastButOneCalledScene(it->second);
          } else {
            pGroup->setLastCalledScene(it->second);
          }
        }
      } else {
        boost::shared_ptr<Group> pGroup = zone->getGroup(gid);
        if (pGroup) {
          switch (gid) {
            case GroupIDControlTemperature:
              if (isUndo) {
                pGroup->setLastButOneCalledScene(it->second);
              } else {
                pGroup->setLastCalledScene(it->second);
                zone->setHeatingOperationMode(it->second);
              }
              break;
            default:
              if (isUndo) {
                pGroup->setLastButOneCalledScene(it->second);
              } else {
                pGroup->setLastCalledScene(it->second);
              }
              break;
          }
        }
      }
    }

    std::bitset<7> states = m_Interface.getZoneStates(dsm->getDSID(), zone->getID());
    // only for light
    boost::shared_ptr<Group> pGroup = zone->getGroup(GroupIDYellow);
    // light group will automatically generate the appropriate state
    if ((pGroup) && (pGroup->getState() == State_Unknown)) {
      pGroup->setOnState(coUnknown, states.test(GroupIDYellow-1));
    }

    SensorType idList[] = {
        SensorType::TemperatureIndoors,
        SensorType::HumidityIndoors,
        SensorType::CO2Concentration,
        SensorType::BrightnessIndoors
    };

    for (uint8_t i=0; i < std::extent<decltype(idList)>::value; i++) {
      dsuid_t sensorDevice;
      try {
        sensorDevice = m_Interface.getZoneSensor(dsm->getDSID(), zone->getID(), idList[i]);
        DeviceReference devRef = zone->getDevices().getByDSID(sensorDevice);
        boost::shared_ptr<Device> pDev = devRef.getDevice();

        // check if zone sensor already assigned
        boost::shared_ptr<Device> oldDev = zone->getAssignedSensorDevice(idList[i]);
        if (oldDev && (oldDev->getDSID() != sensorDevice)) {
          log("Duplicate sensor type " + sensorTypeName(idList[i]) +
              " registration on zone " + intToString(zone->getID()) +
              ": dsuid " + dsuid2str(oldDev->getDSID()) +
              " is zone reference, and dsuid " + dsuid2str(sensorDevice) +
              " is additionally registered on dSM " + dsuid2str(dsm->getDSID()), lsWarning);
        } else {
          zone->setSensor(*pDev, idList[i]);
        }
      } catch (ItemNotFoundException& e) {
        log("Sensor on zone " + intToString(zone->getID()) +
            " with id " + dsuid2str(sensorDevice) +
            " is not present but assigned as zone reference on dSM " +
            dsuid2str(dsm->getDSID()), lsWarning);
      } catch (BusApiError& e) {
        // not fatal, catch exception here and avoid endless readout
        break;
      } catch (std::runtime_error& e) {
        log("Sensor with id " + dsuid2str(sensorDevice) +
            " cannot be assigned as zone reference: " + e.what(), lsWarning);
      }
    }

    if (dsm->getCapability_HasTemperatureControl()) {
      if (!scanTemperatureControl(dsm, zone)) {
        return false;
      }
    }
    return true;
  }

  bool BusScanner::scanTemperatureControl(boost::shared_ptr<DSMeter> meter, boost::shared_ptr<Zone> zone) {
    ZoneHeatingConfigSpec_t config = {};
    ZoneHeatingOperationModeSpec_t operationModes = {};
    auto&& modifyingInterface = m_Apartment.getBusInterface()->getStructureModifyingBusInterface();

    try {
      config = m_Interface.getZoneHeatingConfig(meter->getDSID(), zone->getID());
      operationModes = m_Interface.getZoneHeatingOperationModes(meter->getDSID(), zone->getID());

      log(ds::str("Heating config zone:", zone->getID(), " dsm:", meter->getDSID(), " mode:", config.mode), lsNotice);
    } catch (std::exception& e) {
      log(ds::str("Error getting heating config. dsm:", meter->getDSID(), +" what:", e.what()), lsError);
    }

    // There is no way to distiguish whether mode:OFF means
    // the dsm is selected controller with mode::OFF
    // or whether the dsm has disabled controller.
    // We consider it disabled as that covers more important use cases.

    if (!zone->isTemperatureConfigValid()) {
      // Dss has no configuration for this zone, take the first valid configuration from dsms
      //
      // Zone is configured to OFF mode and dss loses its configuration,
      // zone stays becomes unconfigured instead of configured in OFF mode.
      if (config.mode != HeatingControlMode::OFF) {
        log(ds::str("Taking temperature control config. zone:", zone->getID(), " dsm:", meter->getDSID(), " mode:",
                config.mode),
            lsNotice);
        zone->setHeatingConfigSpec(config);
        zone->setHeatingOperationMode(operationModes);
      }
    }

    if (zone->isTemperatureConfigValid()) {
      auto controller = zone->temperatureController();
      // No controlling dsm yet? Pick the first non-OFF one
      if (!controller && config.mode != HeatingControlMode::OFF) {
        DS_NOTICE("Selecting first temperature controller.", *zone, *meter, config.mode);
        zone->setTemperatureController(meter);
        controller = meter;
      }

      if (controller == meter) {
        if (!zone->temperatureConfig().isEqual(config, operationModes)) {
          DS_NOTICE("Temperature controller config mismatch, synchronize DSM.", *zone, *meter, config, operationModes,
              zone->temperatureConfig());
          m_structureManipulator.setDeviceZoneHeatingConfig(meter, *zone);
        } else {
          DS_NOTICE("Temperature controller config matches.", *zone, *meter, config.mode);
        }
      } else {
        if (config.mode != HeatingControlMode::OFF) {
          DS_NOTICE("Temperature controller shall not be active, disable.", *zone, *meter);
          modifyingInterface->disableZoneHeatingConfig(meter->getDSID(), zone->getID());
        }
      }
    }

    // sync zone sensors only if they are not valid
    uint16_t sensorValue;
    uint32_t sensorAge;
    DateTime now;
    ZoneHeatingStatus zValues = zone->getHeatingStatus();
    ZoneSensorStatus zSensors = zone->getSensorStatus();

    // TODO (soon): adapt this code to logic found in the new synchronization between meters
    // get the temperature from this meter if we do not have already a valid one
    if (zSensors.temperatureValueTS == 0) {
      try {
        m_Interface.getZoneSensorValue(meter->getDSID(), zone->getID(), SensorType::TemperatureIndoors,
            &sensorValue, &sensorAge);
        if (sensorAge <= HEATING_MAX_SENSOR_AGE) {
          DateTime age = now.addSeconds(-1 * sensorAge);
          zone->setTemperature(
              sensorValueToDouble(SensorType::TemperatureIndoors, sensorValue), age);
        }
      } catch (BusApiError& e) {
        log("Error getting heating temperature value on zone " + intToString(zone->getID()) +
            " from " + dsuid2str(meter->getDSID()) + ": " + e.what(), lsWarning);
      }
    }

    // get the nominal value from this meter if we do not have already a valid one
    if (zValues.nominalValueTS == 0) {
      try {
        m_Interface.getZoneSensorValue(meter->getDSID(), zone->getID(), SensorType::RoomTemperatureSetpoint,
            &sensorValue, &sensorAge);
        // set the nominal value only when it is valid
        if (sensorAge <= SENSOR_MAX_AGE) {
          DateTime age = now.addSeconds(-1 * sensorAge);
          zone->setNominalValue(
              sensorValueToDouble(SensorType::RoomTemperatureSetpoint, sensorValue), age);
        }
      } catch (BusApiError& e) {
        log("Error reading heating nominal temperature value on zone " + intToString(zone->getID()) +
            " from " + dsuid2str(meter->getDSID()) + ": " + e.what(), lsWarning);
      }
    }

    // get the control value from this meter if we do not have already a valid one
    if (zValues.controlValueTS == 0) {
      try {
        m_Interface.getZoneSensorValue(meter->getDSID(), zone->getID(), SensorType::RoomTemperatureControlVariable,
            &sensorValue, &sensorAge);
        if (sensorAge <= HEATING_MAX_SENSOR_AGE) {
          DateTime age = now.addSeconds(-1 * sensorAge);
          zone->setControlValue(
              sensorValueToDouble(SensorType::RoomTemperatureControlVariable, sensorValue), age);
        }
      } catch (BusApiError& e) {
        log("Error reading heating control value on zone " + intToString(zone->getID()) +
            " from " + dsuid2str(meter->getDSID()) + ": " + e.what(), lsWarning);
      }
    }

    return true;
  }

  void BusScanner::syncBinaryInputStates(boost::shared_ptr<DSMeter> dsm, boost::shared_ptr <Device> device) {
    boost::shared_ptr<BinaryInputScanner> task;
    std::string connURI = m_Apartment.getBusInterface()->getConnectionURI();
    task = boost::make_shared<BinaryInputScanner>(&m_Apartment, connURI);
    task->setup(dsm, device);
    TaskProcessor &tp(m_Apartment.getModelMaintenance()->getTaskProcessorMaySleep());
    tp.addEvent(task);
  } // syncBinaryInputStates

  void BusScanner::synchronizeDSMeterData(boost::shared_ptr<DSMeter> dsm, DSMeterSpec_t &spec)
  {
    dsm->setArmSoftwareVersion(spec.SoftwareRevisionARM);
    dsm->setDspSoftwareVersion(spec.SoftwareRevisionDSP);
    dsm->setHardwareVersion(spec.HardwareVersion);
    dsm->setApiVersion(spec.APIVersion);
    dsm->setPropertyFlags(spec.flags);

    log("synchronizeDSMeterData Meter: " + dsuid2str(dsm->getDSID())
        + " BusMemberType: " + intToString(spec.DeviceType), lsDebug);

    if ( busMemberTypeIsValid(spec.DeviceType)) {
      dsm->setBusMemberType(spec.DeviceType);
    } else if ( !busMemberTypeIsValid(dsm->getBusMemberType()) &&
                DsmApiIsdSM(dsm->getDSID())){
      // recover old dSM11. (firmware < 1.8.3.) REFS #10417 - git 124e88e5c
      dsm->setBusMemberType(BusMember_dSM11);
    }

    dsm->setApartmentState(spec.ApartmentState);
    if (dsm->getName().empty()) {
      dsm->setName(spec.Name);
    }

    dsm->setSynchronized();

    auto fireMuteEnabled = m_Apartment.getNonScriptState(StateName::FireMuteEnabled)->getState() == 1;
    try {
      if (spec.flags[5] != fireMuteEnabled) {
        m_Apartment.getBusInterface()->getStructureModifyingBusInterface()->setFireMuteEnabled(
            dsm->getDSID(), fireMuteEnabled);
      }
    } catch (const std::exception& e) {
      DS_ERROR("Failed to set fire mute enabled", dsm, fireMuteEnabled, spec.flags);
    }
  } // synchronizeDSMeterData

  void BusScanner::log(const std::string& line, aLogSeverity severity) {
    Logger::getInstance()->log(line, severity);
  } // log


  BinaryInputScanner::BinaryInputScanner(Apartment* _papartment, const std::string& _busConnection)
    : m_pApartment(_papartment),
      m_busConnection(_busConnection),
      m_dsmApiHandle(NULL)
  {
    m_device.reset();
    m_dsm.reset();
  }

  BinaryInputScanner::~BinaryInputScanner()
  {
    if (m_dsmApiHandle != NULL) {
      DsmApiClose(m_dsmApiHandle);
      DsmApiCleanup(m_dsmApiHandle);
      m_dsmApiHandle = NULL;
    }
  }

  uint16_t BinaryInputScanner::getDeviceSensorValue(const dsuid_t& _dsm,
                            dev_t _device,
                            uint8_t _sensorIndex) const
  {
    if (m_dsmApiHandle == NULL) {
      throw std::runtime_error("Invalid libdsm api handle");
    }

    uint16_t retVal;
    int ret = DeviceSensor_get_value_sync(m_dsmApiHandle, _dsm,
        _device,
        _sensorIndex,
        30,
        &retVal);
    DSBusInterface::checkResultCode(ret);
    return retVal;
  }

  uint8_t BinaryInputScanner::getDeviceConfig(const dsuid_t& _dsm,
      dev_t _device,
      uint8_t _configClass,
      uint8_t _configIndex) const
  {
    if (m_dsmApiHandle == NULL) {
      throw std::runtime_error("Invalid libdsm api handle");
    }

    uint8_t retVal;
    int ret = DeviceConfig_get_sync_8(m_dsmApiHandle, _dsm,
        _device,
        _configClass,
        _configIndex,
        30,
        &retVal);
    DSBusInterface::checkResultCode(ret);
    return retVal;
  }

  void BinaryInputScanner::run() {
    DSS* dss = nullptr;
    if (DSS::hasInstance()) {
      dss = DSS::getInstance();
      dss->getSecurity().loginAsSystemUser("BinaryInputScanner needs system-rights");
    }

    std::vector<boost::shared_ptr<Device> > devices;
    boost::shared_ptr<Device> device = m_device.lock();
    boost::shared_ptr<DSMeter> dsm = m_dsm.lock();

    if (device) {
      // synchronize a single device
      if (device->isPresent() && device->getBinaryInputCount() > 0) {
        devices.push_back(device);
      }
    } else if (dsm) {
      Set D = dsm->getDevices();
      BinaryInputOrStateDeviceFilter filter;
      D.perform(filter);
      devices = filter.getDeviceList();
    } else {
      Set D = m_pApartment->getDevices();
      BinaryInputOrStateDeviceFilter filter;
      D.perform(filter);
      devices = filter.getDeviceList();
    }

    Logger::getInstance()->log("BinaryInputScanner run:"
        " dsm = " + (dsm ? dsuid2str(dsm->getDSID()) : "NULL") +
        ", dev = " + (device ? dsuid2str(device->getDSID()) : "NULL"), lsDebug);

    m_dsmApiHandle = DsmApiInitialize();
    if (!m_dsmApiHandle) {
      throw std::runtime_error("BinaryInputScanner: Unable to init dsmapi handle");
    }

    int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
    if (result < 0) {
      throw std::runtime_error(std::string("BinaryInputScanner: Unable to open connection to: ") + m_busConnection);
    }
    if (dss) {
      dss->setDsmApiBlockingCallback(m_dsmApiHandle);
    }

    // send a request to each binary input device to resend its current status
    for (boost::shared_ptr<Device> dev : devices) {
      uint16_t value = 0;
      boost::shared_ptr<DSMeter> dsm;
      try {
        dsm = m_pApartment->getDSMeterByDSID(dev->getDSMeterDSID());
      } catch (ItemNotFoundException& ex) {
        Logger::getInstance()->log("BinaryInputScanner: dsm " + dsuid2str(dev->getDSMeterDSID())
            + " is unknown", lsWarning);
        continue;
      }

      while (dsm->isPresent() && (dsm->getState() == DSM_STATE_REGISTRATION)) {
        sleep(5);
      }
      if (!dsm->isPresent()) {
        // ignore non-present dsm's, they will be discovered later with a new event
        Logger::getInstance()->log("BinaryInputScanner: dsm " + dsuid2str(dev->getDSMeterDSID())
            + " is inactive", lsWarning);
        continue;
      }

      if (! dev->isVdcDevice()) {
        // Powerline Device
        try {
          value = getDeviceSensorValue(dev->getDSMeterDSID(), dev->getShortAddress(), 0x20);
          Logger::getInstance()->log("BinaryInputScanner: device " +
              dsuid2str(dev->getDSID()) + ", state = " + intToString(value), lsDebug);
          for (int index = 0; index < dev->getBinaryInputCount(); index++) {
            dev->handleBinaryInputEvent(index,
                ((value >> index) & 1) ? BinaryInputStateValue::Active : BinaryInputStateValue::Inactive);
          }
          // avoid powerline bus monopolization
          sleep(5);
        } catch (BusApiError& ex) {
          Logger::getInstance()->log("BinaryInputScanner: device " + dsuid2str(dev->getDSID())
              + " did not respond to input state query", lsWarning);
          for (int index = 0; index < dev->getBinaryInputCount(); index++) {
            dev->handleBinaryInputEvent(index, BinaryInputStateValue::Unknown);
          }
        }

      } else {
        // IP Device
        try {
          VdcHelper::State state = VdcHelper::getState(dsm->getDSID(), dev->getDSID());
          const auto& sInput = state.binaryInputStates;
          Logger::getInstance()->log("BinaryInputScanner: device " +
              dsuid2str(dev->getDSID()) + ", state response fields = " + intToString(sInput.size()), lsDebug);
          for (auto it = sInput.begin(); it != sInput.end(); it++) {
            dev->handleBinaryInputEvent(it->first, static_cast<BinaryInputStateValue> (it->second));
          }
          dev->setStateValues(state.deviceStates);
        } catch (std::runtime_error& e) {
          Logger::getInstance()->log("BinaryInputScanner: VdcDevice " + dsuid2str(dev->getDSID()) +
              " failure: " + e.what(), lsWarning);
          for (int index = 0; index < dev->getBinaryInputCount(); index++) {
            dev->handleBinaryInputEvent(index, BinaryInputStateValue::Unknown);
          }
        }
      }
    }
  }

  void BinaryInputScanner::setup(boost::shared_ptr<DSMeter> _dsMeter, boost::shared_ptr<Device> _device)
  {
    m_dsm = _dsMeter;
    m_device = _device;
  }

} // namespace dss
