/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "device.h"

#include <digitalSTROM/dsm-api-v2/dsm-api.h>
#include <digitalSTROM/dsuid.h>
#include <ds-model/scenarios.h>
#include <ds/locale/ds-locale.h>
#include <ds/log.h>
#include <math.h>
#include <unistd.h>
#include <boost/container/flat_map.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/shared_ptr.hpp>
#include <cmath>
#include <limits>
#include "businterface.h"
#include "cache/apartment-proxy.h"
#include "device-data.h"
#include "ds-scenarios.h"
#include "ds485types.h"
#include "event.h"
#include "event/event_create.h"
#include "hal/hardware-abstraction.h"
#include "messages/vdc-messages.pb.h"
#include "model-features.h"
#include "model/action.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/devicereference.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelevent.h"
#include "model/modelmaintenance.h"
#include "model/scenehelper.h"
#include "model/state.h"
#include "propertysystem.h"
#include "protobufjson.h"
#include "status-field.h"
#include "status.h"
#include "util.h"
#include "vdc-connection.h"
#include "vdc-db.h"
#include "vdc-element-reader.h"

static ds::LogChannel dsLogChannel("modelDevice");

#define BLINK_DELAY_SCALE 33.333333 // value specced by Christian Theiss

namespace dss {

  //================================================== DeviceBinaryInput

  __DEFINE_LOG_CHANNEL__(DeviceBinaryInput, lsNotice);

  // RAII class to handling stream to StatusField
  // We capture only weak reference to StatusField to avaid crashes
  // if this object is not notified soon enough that StatusField was deleted.
  class DeviceBinaryInput::StatusFieldHandle: boost::noncopyable {
  public:
    StatusFieldHandle(DeviceBinaryInput& parent, StatusField& statusField)
        : m_parent(parent),
        m_weakPtr(boost::shared_ptr<StatusField>(statusField.getStatus().getGroup().sharedFromThis(), &statusField)) {
      statusField.addSubState(*m_parent.m_state);
    }

    ~StatusFieldHandle() {
      if (auto&& ptr = m_weakPtr.lock()) {
        ptr->removeSubState(*m_parent.m_state);
      }
    }

    StatusField* getPtr() {
      if (auto&& ptr = m_weakPtr.lock()) {
        return ptr.get();
      }
      return nullptr;
    }

    void update() {
      log(std::string("StatusFieldHandle::update this:") + m_parent.m_name, lsDebug);
      if (auto&& ptr = m_weakPtr.lock()) {
        ptr->updateSubState(*m_parent.m_state);
      } else {
        // We should have been deleted at this point
        log(ds::str("Referenced StatusField instance was deleted"), lsError);
      }
    }
  private:
    DeviceBinaryInput& m_parent;
    boost::weak_ptr<StatusField> m_weakPtr;
  };

  DeviceBinaryInput::DeviceBinaryInput(Device& device, const DeviceBinaryInputSpec_t& spec, int index)
      : m_inputIndex(index)
      , m_inputType(spec.InputType)
      , m_inputId(spec.InputID)
      , m_targetGroupId(spec.TargetGroup)
      , m_independent(spec.Independent)
      , m_device(device)
      , m_name(m_device.getName() + "/" + intToString(index)) {
    log(std::string("DeviceBinaryInput this:") + m_name
        + " inputType:" + intToString(static_cast<int>(m_inputType))
        + " inputId:" + intToString(static_cast<int>(m_inputId))
        + " targetGroupId:" + intToString(m_targetGroupId), lsInfo);
    m_state = device.getApartment().allocateBinaryInputState(device.sharedFromThis(), index);

    // assignCustomBinaryInputValues
    if (m_inputType == BinaryInputType::WindowTilt) {
      State::ValueRange_t windowTiltValues;
      windowTiltValues.push_back("invalid");
      windowTiltValues.push_back("closed");
      windowTiltValues.push_back("open");
      windowTiltValues.push_back("tilted");
      windowTiltValues.push_back("unknown");
      m_state->setValueRange(windowTiltValues);
    }

    updateStatusFieldHandle();
  }

  DeviceBinaryInput::~DeviceBinaryInput() {
    log(std::string("~DeviceBinaryInput this:") + m_name, lsInfo);
    try {
      m_device.getApartment().removeState(m_state);
    } catch (const std::exception& e) {
      log(std::string("~DeviceBinaryInput: remove state failed:") + e.what(), lsWarning);
    }
  }

  void DeviceBinaryInput::setTargetId(uint8_t targetGroupId) {
    if (m_targetGroupId == targetGroupId) {
      return;
    }
    log(std::string("setTarget this:") + m_name
        + " targetGroupId:" + intToString(targetGroupId), lsInfo);
    m_targetGroupId = targetGroupId;
    updateStatusFieldHandle();
  }

  void DeviceBinaryInput::setInputId(BinaryInputId inputId) {
    if (m_inputId == inputId) {
      return;
    }
    log(std::string("setInputId this:") + m_name
        + " inputId:" + intToString(static_cast<int>(inputId)), lsInfo);
    m_inputId = inputId;
  }

  void DeviceBinaryInput::setInputType(BinaryInputType inputType) {
    if (m_inputType == inputType) {
      return;
    }
    log(std::string("setInputType this:") + m_name
        + " inputType:" + intToString(static_cast<int>(inputType)), lsInfo);
    m_inputType = inputType;
    updateStatusFieldHandle();
  }

  void DeviceBinaryInput::handleEvent(BinaryInputStateValue inputState) {
    log(std::string("handleEvent this:") + m_name
        + " m_inputType:" + intToString(static_cast<int>(m_inputType))
        + " inputState:" + intToString(static_cast<int>(inputState)), lsInfo);
    try {
      if (m_inputType == BinaryInputType::WindowTilt) {
        m_state->setState(coSystem,
            [&]() {
              switch (static_cast<BinaryInputWindowHandleState>(inputState)) {
                case BinaryInputWindowHandleState::Closed: return StateWH_Closed;
                case BinaryInputWindowHandleState::Open: return StateWH_Open;
                case BinaryInputWindowHandleState::Tilted: return StateWH_Tilted;
                case BinaryInputWindowHandleState::Unknown: return StateWH_Unknown;
              };
              return StateWH_Unknown;
            }());
      } else {
        m_state->setState(coSystem,
            [&]() {
              switch (inputState) {
                case BinaryInputStateValue::Inactive: return State_Inactive;
                case BinaryInputStateValue::Active: return State_Active;
                case BinaryInputStateValue::Unknown: return State_Unknown;
              };
              return State_Unknown;
            }());
      }
    } catch (const std::exception& e) {
      log(std::string("handleEvent: what:")+ e.what(), lsWarning);
    }
    if (m_statusFieldHandle) {
      m_statusFieldHandle->update();
    }
  }

  void DeviceBinaryInput::updateStatusFieldHandle() {
    // find status object within target group
    StatusField* statusField = nullptr;
    if (m_targetGroupId != 0) {
      if (auto type = statusFieldTypeForBinaryInputType(m_inputType)) {
        // TODO(someday): status sensor fixed to apartment groups for now --> zone 0
        const std::array<int, 2> zoneIDs{{0, m_device.getZoneID()}};
        for (int zoneID : zoneIDs) {
          const auto& targetGroup = m_device.getApartment().getZone(zoneID)->tryGetGroup(m_targetGroupId).lock();
          if ( !targetGroup ){
            continue;
          }
          if (const auto& status = targetGroup->tryGetStatus()) {
            statusField = &status->getField(*type);
          } else {
            log(ds::str("updateStatusFieldHandle Group does not support status. this:", m_name,
                        " m_inputType:", m_inputType, " m_targetGroupId:", m_targetGroupId),
                lsWarning);
            // some groups do not support status
          }
        }
      }
    }
    auto oldStatusField = m_statusFieldHandle ? m_statusFieldHandle->getPtr() : (StatusField*)nullptr;
    if (oldStatusField == statusField) {
      return;
    }

    m_statusFieldHandle.reset();
    if (!statusField) {
      log(ds::str("updateStatusFieldHandle this:", m_name,
          " m_inputType:", m_inputType, " m_targetGroupId:", m_targetGroupId,
          " statusField:nullptr"), lsInfo);
      return;
    }
    log(ds::str("updateStatusFieldHandle this:", m_name,
        " m_inputType:", m_inputType, " m_targetGroupId:", m_targetGroupId,
        " type:",  statusField->getType(), " name:", statusField->getName()), lsInfo);
    m_statusFieldHandle.reset(new StatusFieldHandle(*this, *statusField));
  }

  static std::vector<ChannelType> hardwareOutputChannels(HardwareType hardwareType, const Device& device) {
    switch (hardwareType) {
      case HardwareType::GE_KM200:
      case HardwareType::GE_TKM210:
      case HardwareType::GE_KM300:
      case HardwareType::GE_TKM300:
      case HardwareType::GE_SDM200:
      case HardwareType::GE_SDM200JS:
      case HardwareType::GE_SDM200CS:
      case HardwareType::GE_SDS200CW:
      case HardwareType::GE_SDS200CS:
      case HardwareType::GE_SDS220CT:
      case HardwareType::GE_KL200:
      case HardwareType::GE_UMV200:
        return {ChannelType::brightness};

      case HardwareType::GR_KL200:
      case HardwareType::GR_KL210:
        return {ChannelType::shadePositionOutside};

      case HardwareType::GR_KL220:
      case HardwareType::GR_HKL230:
        return {ChannelType::shadePositionOutside, ChannelType::shadeOpeningAngleOutside};

      // Configurable joker devices
      case HardwareType::SW_KL200:
      case HardwareType::SW_ZWS200:
      case HardwareType::SW_ZWS200F:
      case HardwareType::SW_ZWS200EF:
      case HardwareType::SW_ZWS205:
      case HardwareType::SW_SSL200JS:
      case HardwareType::SW_SSL200FS:
        if (device.getActiveGroup() == ColorIDYellow) {
          return {ChannelType::brightness};
        }
        break;

      // TODO(someday):
      case HardwareType::GN_TKM200:
      case HardwareType::GN_KM200:
      case HardwareType::GN_TKM210:
      case HardwareType::RT_TKM200:
      case HardwareType::RT_SDM200:
      case HardwareType::SW_UMR200:
      case HardwareType::BL_KM200:
      case HardwareType::SW_TNY200:

      // Input only devices
      case HardwareType::GE_TKM220:
      case HardwareType::GE_TKM230:
      case HardwareType::GE_UMV210:
      case HardwareType::GR_TKM200:
      case HardwareType::GR_TKM210:
      case HardwareType::SW_TKM200:
      case HardwareType::SW_TKM210:
      case HardwareType::SW_AKM200:
      case HardwareType::SW_AKM210:
      case HardwareType::SW_AKM220:
      case HardwareType::SW_SK200:
      case HardwareType::SW_SK201:
      case HardwareType::SW_SK202:
      case HardwareType::SW_SDS200CW:
      case HardwareType::SW_SDS200CS:
      case HardwareType::SW_SDS220CT:
      case HardwareType::BL_SDS200:
      case HardwareType::BL_SK204:

      case HardwareType::UNKNOWN:
        break;
    }
    return {};
  }

  //================================================== Device
  __DEFINE_LOG_CHANNEL__(Device, lsNotice);

  Device::Device(dsuid_t _dsid, Apartment* _pApartment)
      : AddressableModelItem(_pApartment),
        m_DSID(_dsid),
        m_ShortAddress(ShortAddressStaleDevice),
        m_LastKnownShortAddress(ShortAddressStaleDevice),
        m_ZoneID(NEW_DEVICES_ZONE_ID),
        m_LastKnownZoneID(0),
        m_DSMeterDSID(DSUID_NULL),
        m_LastKnownMeterDSID(DSUID_NULL),
        m_DSMeterDSIDstr(),
        m_LastKnownMeterDSIDstr(),
        m_ActiveGroup(0),
        m_DefaultGroup(0),
        m_FunctionID(0),
        m_ProductID(0),
        m_VendorID(0),
        m_RevisionID(0),
        m_hardwareType(),
        m_LastCalledScene(SceneOff),
        m_LastButOneCalledScene(SceneOff),
        m_Consumption(0),
        m_Energymeter(0),
        m_LastDiscovered(DateTime::NullDate),
        m_FirstSeen(DateTime::NullDate),
        m_IsValid(false),
        m_IsLockedInDSM(false),
        m_IgnoreOperationLock(false),
        m_OutputMode(0),
        m_ButtonInputMode(ButtonInputMode::STANDARD),
        m_ButtonInputIndex(0),
        m_ButtonInputCount(0),
        m_ButtonSetsLocalPriority(false),
        m_ButtonCallsPresent(true),
        m_ButtonActiveGroup(255),
        m_ButtonID(0xff),
        m_OemEanNumber(0),
        m_OemSerialNumber(0),
        m_OemPartNumber(0),
        m_OemInetState(DEVICE_OEM_EAN_NO_EAN_CONFIGURED),
        m_OemState(DEVICE_OEM_UNKNOWN),
        m_OemIsIndependent(true),
        m_ActiveCoolingMode(0),
        m_HWInfo(),
        m_iconPath("unknown.png"),
        m_OemProductInfoState(DEVICE_OEM_UNKNOWN),
        m_OemProductName(),
        m_OemProductIcon(),
        m_OemProductURL(),
        m_OemConfigLink(),
        m_isVdcDevice(false),
        m_hasActions(false),
        m_hasLegacyActions(false),
        m_data{new DeviceData{}},
        m_ValveType(DEVICE_VALVE_UNKNOWN),
        m_IsConfigLocked(false),
        m_outputChannelCount(0),
        m_AKMInputProperty(),
        m_cardinalDirection(cd_none),
        m_windProtectionClass(wpc_none),
        m_floor(0),
        m_pairedDevices(0),
        m_visible(true),
        m_maxPosTimeSeconds(0),
        m_buttonUsage(ButtonUsage::unsupported) {
    m_DSMeterDSUIDstr = dsuid2str(m_DSMeterDSID);
    m_LastKnownMeterDSUIDstr = dsuid2str(m_LastKnownMeterDSID);

    dsid_t dsid;
    if (dsuid_to_dsid(m_DSMeterDSID, &dsid)) {
      m_DSMeterDSIDstr = dsid2str(dsid);
    }
    if (dsuid_to_dsid(m_LastKnownMeterDSID, &dsid)) {
      m_LastKnownMeterDSIDstr = dsid2str(dsid);
    }
  } // ctor

  Device::~Device() {
    removeFromPropertyTree();
  }

  dsModel::DeviceId Device::id() const {
    DS_REQUIRE(isInDsModel(), "Accessing device that does not exist in ds model.", *this);
    return dsModel::DeviceId{ds::str(getDSID())};
  }

  dsModel::SubmoduleId Device::submoduleId() const {
    DS_REQUIRE(isInDsModel(), "Accessing device that does not exist in ds model.", *this);
    return dsModel::SubmoduleId{id()};
  }

  dsModel::FunctionBlockId Device::functionBlockId() const {
    DS_REQUIRE(isInDsModel(), "Accessing device that does not exist in ds model.", *this);
    return dsModel::FunctionBlockId{id()};
  }

  bool Device::isInDsModel() const {
    return !is2WaySlave() && isVisible();
  }

  int Device::getGroupZoneID(int groupID) const {
    return (isAppUserGroup(groupID) || isGlobalAppGroup(groupID)) ? 0 : (m_ZoneID > 0 ? m_ZoneID : m_LastKnownZoneID);
  }

  boost::weak_ptr<Group> Device::tryGetGroup(int groupId) const {
    if (auto&& zone = m_pApartment->tryGetZone(getGroupZoneID(groupId)).lock()) {
      return zone->tryGetGroup(groupId);
    }
    return boost::weak_ptr<Group>();
  }

  std::vector<boost::shared_ptr<Cluster>> Device::getClusters() const {
    std::vector<boost::shared_ptr<Cluster>> clusters;
    for (auto&& groupId : m_groupIds) {
      if (isCluster(groupId)) {
        if (auto&& group = tryGetGroup(groupId).lock()) {
          if (auto&& cluster = boost::dynamic_pointer_cast<Cluster>(group)) {
            clusters.push_back(std::move(cluster));
          } else {
            DS_ERROR("Failed to cast group", groupId, this);
          }
        } else {
          DS_ERROR("Failed to get group", groupId, this);
        }
      }
    }
    return clusters;
  }

  void Device::removeFromPropertyTree() {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (m_pPropertyNode != NULL) {
      try {
        if ((m_DSMeterDSID != DSUID_NULL) && (m_DSID != DSUID_NULL)) {
          std::string devicePath = "devices/" + dsuid2str(m_DSID);
          PropertyNodePtr dev = m_pApartment->getDSMeterByDSID(m_DSMeterDSID)->getPropertyNode()->getProperty(devicePath);
          dev->alias(PropertyNodePtr());
          dev->getParentNode()->removeChild(dev);
        }
      } catch (std::runtime_error &err) {
        log(err.what(), lsError);
      }

      if (m_pApartment->getPropertyNode() != NULL) {
        for (auto&& g : m_groupIds) {
          std::string gPath = "zones/zone" + intToString(getGroupZoneID(g)) +
              "/groups/group" + intToString(g) + "/devices/" +
              dsuid2str(m_DSID);
          PropertyNodePtr gnode = m_pApartment->getPropertyNode()->getProperty(gPath);
          if (gnode != NULL) {
            gnode->alias(PropertyNodePtr());
            gnode->getParentNode()->removeChild(gnode);
          }
          PropertyNodePtr gsubnode = m_pPropertyNode->getProperty("groups/group" + intToString(g));
          if (gsubnode != NULL) {
            gsubnode->getParentNode()->removeChild(gsubnode);
          }
        }
      }

      m_pPropertyNode->unlinkProxy(true);
      PropertyNode *parent = m_pPropertyNode->getParentNode();
      if (parent != NULL) {
        parent->removeChild(m_pPropertyNode);
      }
      m_pPropertyNode.reset();
    }
    if (m_pAliasNode != NULL) {
      m_pAliasNode->alias(PropertyNodePtr());
      PropertyNode *parent = m_pAliasNode->getParentNode();
      if (parent != NULL) {
        parent->removeChild(m_pAliasNode);
      }
      m_pAliasNode.reset();
    }
  }

  void Device::publishVdcToPropertyTree() {
    if (m_isVdcDevice && (m_pPropertyNode != NULL)) {
      PropertyNodePtr propNode = m_pPropertyNode->createProperty("properties");

      propNode->createProperty("DisplayId")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.displayId, false));
      propNode->createProperty("HardwareModelGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.hardwareModelGuid, false));
      propNode->createProperty("ModelUID")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.modelUID, false));
      propNode->createProperty("ModelVersion")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.modelVersion, false));
      propNode->createProperty("VendorGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.vendorGuid, false));
      propNode->createProperty("OemGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.oemGuid, false));
      propNode->createProperty("OemModelGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.oemModelGuid, false));
      propNode->createProperty("ConfigURL")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.configURL, false));
      propNode->createProperty("HardwareGuid")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.hardwareGuid, false));
      propNode->createProperty("HardwareInfo")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.model, false));
      propNode->createProperty("HardwareVersion")
        ->linkToProxy(PropertyProxyReference<std::string>(m_vdsdSpec.hardwareVersion, false));
    }
  }

  void Device::publishToPropertyTree() {
    if (m_pPropertyNode != NULL) {
      // already published
      return;
    }

    if (m_pApartment->getPropertyNode() == NULL) {
      // never happens
      return;
    }

    if (!m_visible) {
      return;
    }

    m_pPropertyNode = m_pApartment->getPropertyNode()->createProperty("zones/zone0/devices/" + dsuid2str(m_DSID));

    dsid_t dsid;
    if (dsuid_to_dsid(m_DSID, &dsid)) {
      m_pPropertyNode->createProperty("dSID")->setStringValue(dsid2str(dsid));
    } else {
      m_pPropertyNode->createProperty("dSID")->setStringValue("");
    }
    m_pPropertyNode->createProperty("DisplayID")
          ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getDisplayID));

    m_pPropertyNode->createProperty("dSUID")->setStringValue(dsuid2str(m_DSID));
    m_pPropertyNode->createProperty("present")
      ->linkToProxy(PropertyProxyMemberFunction<Device, bool>(*this, &Device::isPresent));
    m_pPropertyNode->createProperty("name")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string>(*this, &Device::getName, &Device::setName));
    if (!m_DSMeterDSIDstr.empty()) {
      m_pPropertyNode->createProperty("DSMeterDSID")
        ->linkToProxy(PropertyProxyReference<std::string>(m_DSMeterDSIDstr, false));
    } else {
      m_pPropertyNode->createProperty("DSMeterDSID")->setStringValue("");
    }
    m_pPropertyNode->createProperty("DSMeterDSUID")
      ->linkToProxy(PropertyProxyReference<std::string>(m_DSMeterDSUIDstr, false));

    m_pPropertyNode->createProperty("ZoneID")->linkToProxy(PropertyProxyReference<int>(m_ZoneID, false));
    m_pPropertyNode->createProperty("functionID")
      ->linkToProxy(PropertyProxyReference<int>(m_FunctionID, false));
    m_pPropertyNode->createProperty("revisionID")
      ->linkToProxy(PropertyProxyReference<int>(m_RevisionID, false));
    m_pPropertyNode->createProperty("productID")
      ->linkToProxy(PropertyProxyReference<int>(m_ProductID, false));
    m_pPropertyNode->createProperty("vendorID")
      ->linkToProxy(PropertyProxyReference<int>(m_VendorID, false));
    m_pPropertyNode->createProperty("HWInfo")
      ->linkToProxy(PropertyProxyReference<std::string>(m_HWInfo, false));
    m_pPropertyNode->createProperty("GTIN")
      ->linkToProxy(PropertyProxyReference<std::string>(m_GTIN, false));

    PropertyNodePtr oemNode = m_pPropertyNode->createProperty("productInfo");
    oemNode->createProperty("ProductState")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getOemProductInfoStateAsString));
    oemNode->createProperty("ProductName")
      ->linkToProxy(PropertyProxyReference<std::string>(m_OemProductName, false));
    oemNode->createProperty("ProductIcon")
      ->linkToProxy(PropertyProxyReference<std::string>(m_OemProductIcon, false));
    oemNode->createProperty("ProductURL")
      ->linkToProxy(PropertyProxyReference<std::string>(m_OemProductURL, false));
    oemNode->createProperty("ConfigLink")
      ->linkToProxy(PropertyProxyReference<std::string>(m_OemConfigLink, false));
    oemNode->createProperty("State")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getOemStateAsString));
    oemNode->createProperty("EAN")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getOemEanAsString));
    oemNode->createProperty("SerialNumber")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(m_OemSerialNumber, false));
    oemNode->createProperty("PartNumber")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(m_OemPartNumber, false));
    oemNode->createProperty("isIndependent")
      ->linkToProxy(PropertyProxyReference<bool>(m_OemIsIndependent, false));
    oemNode->createProperty("InternetState")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getOemInetStateAsString));
    oemNode->createProperty("configurationLocked")
      ->linkToProxy(PropertyProxyReference<bool>(m_IsConfigLocked, false));

    m_pPropertyNode->createProperty("isVdcDevice")
      ->linkToProxy(PropertyProxyReference<bool>(m_isVdcDevice, false));
    m_pPropertyNode->createProperty("hasActions")
      ->linkToProxy(PropertyProxyReference<bool>(m_hasActions, false));
    publishVdcToPropertyTree();

    m_pPropertyNode->createProperty("lastKnownZoneID")
      ->linkToProxy(PropertyProxyReference<int>(m_LastKnownZoneID, false));
    m_pPropertyNode->createProperty("shortAddress")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(m_ShortAddress, false));
    m_pPropertyNode->createProperty("lastKnownShortAddress")
      ->linkToProxy(PropertyProxyReference<int, uint16_t>(m_LastKnownShortAddress, false));
    if (!m_LastKnownMeterDSIDstr.empty()) {
      m_pPropertyNode->createProperty("lastKnownMeterDSID")
        ->linkToProxy(PropertyProxyReference<std::string>(m_LastKnownMeterDSIDstr, false));
    } else {
      m_pPropertyNode->createProperty("lastKnownMeterDSID")->setStringValue("");
    }
    m_pPropertyNode->createProperty("lastKnownMeterDSUID")
      ->linkToProxy(PropertyProxyReference<std::string>(m_LastKnownMeterDSUIDstr, false));
    m_pPropertyNode->createProperty("firstSeen")
      ->linkToProxy(PropertyProxyMemberFunction<DateTime, std::string, false>(m_FirstSeen, &DateTime::toString));
    m_pPropertyNode->createProperty("lastDiscovered")
      ->linkToProxy(PropertyProxyMemberFunction<DateTime, std::string, false>(m_LastDiscovered, &DateTime::toString));
    m_pPropertyNode->createProperty("inactiveSince")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getInactiveSinceStr));
    m_pPropertyNode->createProperty("locked")
      ->linkToProxy(PropertyProxyReference<bool>(m_IsLockedInDSM, false));
    m_pPropertyNode->createProperty("ignoreOperationLock")
      ->linkToProxy(PropertyProxyReference<bool>(m_IgnoreOperationLock, false));
    m_pPropertyNode->createProperty("outputMode")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(m_OutputMode, false));
    m_pPropertyNode->createProperty("button/id")
      ->linkToProxy(PropertyProxyReference<int>(m_ButtonID, false));
    m_pPropertyNode->createProperty("button/inputMode")
      ->linkToProxy(PropertyProxyReference<int, ButtonInputMode>(m_ButtonInputMode, false));
    m_pPropertyNode->createProperty("button/inputIndex")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(m_ButtonInputIndex, false));
    m_pPropertyNode->createProperty("button/inputCount")
      ->linkToProxy(PropertyProxyReference<int, uint8_t>(m_ButtonInputCount, false));
    m_pPropertyNode->createProperty("button/activeGroup")
      ->linkToProxy(PropertyProxyReference<int>(m_ButtonActiveGroup, false));
    m_pPropertyNode
        ->createProperty("button/groupMembership") // deprecated
        ->linkToProxy(PropertyProxyReference<int>(m_ButtonActiveGroup, false));
    m_pPropertyNode->createProperty("button/setsLocalPriority")
      ->linkToProxy(PropertyProxyReference<bool>(m_ButtonSetsLocalPriority));
    m_pPropertyNode->createProperty("button/callsPresent")
      ->linkToProxy(PropertyProxyReference<bool>(m_ButtonCallsPresent));
    m_pPropertyNode->createProperty("button/usage")
      ->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getButtonUsageAsString));
    if (!m_pPropertyNode->getProperty("sensorEvents")) {
      PropertyNodePtr sensorNode = m_pPropertyNode->createProperty("sensorEvents");
    }
    PropertyNodePtr binaryInputNode = m_pPropertyNode->createProperty("binaryInputs");
    PropertyNodePtr sensorInputNode = m_pPropertyNode->createProperty("sensorInputs");
    PropertyNodePtr outputChannelNode = m_pPropertyNode->createProperty("outputChannels");

    m_pPropertyNode->createProperty("isValveType")
      ->linkToProxy(PropertyProxyMemberFunction<Device, bool>(*this, &Device::isValveDevice));

    m_pPropertyNode->createProperty("CardinalDirection")
      ->linkToProxy(PropertyProxyToString<CardinalDirection_t>(m_cardinalDirection));
    m_pPropertyNode->createProperty("WindProtectionClass")
      ->linkToProxy(PropertyProxyReference<int,
                    WindProtectionClass_t>(m_windProtectionClass));
    m_pPropertyNode->createProperty("Floor")
      ->linkToProxy(PropertyProxyReference<int>(m_floor));

    m_pPropertyNode->createProperty("supportedBasicScenes")->linkToProxy(PropertyProxyFunction<std::string>([this] {
      return serializeBasicScenes(m_supportedBasicScenes);
    }));

    publishValveTypeToPropertyTree();

    m_TagsNode = m_pPropertyNode->createProperty("tags");
    m_TagsNode->setFlag(PropertyNode::Archive, true);

    if (m_ZoneID != 0) {
      std::string basePath = "zones/zone" + intToString(m_ZoneID) + "/devices";
      if (m_pAliasNode == NULL) {
        PropertyNodePtr node = m_pApartment->getPropertyNode()->getProperty(basePath + "/" + dsuid2str(m_DSID));
        if ((node == NULL) || ((node != NULL) && (node->size() == 0))) {
          m_pAliasNode = m_pApartment->getPropertyNode()->createProperty(basePath + "/" + dsuid2str(m_DSID));
          m_pAliasNode->alias(m_pPropertyNode);
        }
      } else {
        PropertyNodePtr base = m_pApartment->getPropertyNode()->getProperty(basePath);
        if (base != NULL) {
          base->addChild(m_pAliasNode);
        }
      }
    }

    publishAliasInZoneGroups();
    for (auto&& g : m_groupIds) {
      PropertyNodePtr gsubnode = m_pPropertyNode->createProperty("groups/group" + intToString(g));
      gsubnode->createProperty("id")->setIntegerValue(g);
    }

    if (m_DSMeterDSID != DSUID_NULL) {
      setDSMeter(m_pApartment->getDSMeterByDSID(m_DSMeterDSID));
    }

    republishModelFeaturesToPropertyTree();
    if (m_pApartment->getDss() && m_pApartment->getDss()->getDeviceCacheDatabase()) {
      HardwareAbstraction hal(m_pApartment, m_pApartment->getDss()->getDeviceCacheDatabase());
      try {
        hal.publishToPropertyTree(getDSID());
      } catch (const ds::Exception& e) {
      }
    }
  } // publishToPropertyTree

  void Device::publishAliasInZoneGroups() const {
    for (auto&& g : m_groupIds) {
      std::string gPath = "zones/zone" + intToString(getGroupZoneID(g)) +
                          "/groups/group" + intToString(g) + "/devices/" +
                          dsuid2str(m_DSID);
      PropertyNodePtr gnode = m_pApartment->getPropertyNode()->createProperty(gPath);
      if (gnode) {
        gnode->alias(m_pPropertyNode);
      }
    }
  }

  boost::shared_ptr<DeviceReference> Device::makeSharedReference() const {
    return boost::make_shared<DeviceReference>(m_DSID, m_pApartment);
  }

  bool Device::isOn() const {
    return (m_LastCalledScene != SceneOff) &&
           (m_LastCalledScene != SceneMin) &&
           (m_LastCalledScene != SceneDeepOff) &&
           (m_LastCalledScene != SceneStandBy);
  } // isOn

  void Device::setPartiallyFromSpec(const DeviceSpec& spec) {
    m_FunctionID = spec.FunctionID;
    m_ProductID = spec.ProductID;
    m_VendorID = spec.VendorID;
    m_RevisionID = spec.revisionId;
    m_IsLockedInDSM = spec.Locked;
    m_OutputMode = spec.OutputMode;
    m_ActiveGroup = spec.activeGroup;
    m_DefaultGroup = spec.defaultGroup;

    if ((m_FunctionID != 0) && (m_ProductID != 0) && (m_VendorID != 0)) {
      calculateHWInfo();
    }
    updateModelFeatures();
    updateIconPath();
    updateAKMNode();
    publishValveTypeToPropertyTree();

    int subclass = (m_FunctionID & 0xfc0) >> 6;
    if (!isVdcDevice() && subclass == 9) {
      setActiveCoolingMode(true);
    }
  }

  int Device::getFunctionID() const {
    return m_FunctionID;
  } // getFunctionID

  int Device::getProductID() const {
    return m_ProductID;
  } // getProductID

  void Device::updateAKMNode() {
    if ((m_pPropertyNode != NULL) && ((getDeviceType() == DEVICE_TYPE_AKM) ||
                                      (getDeviceType() == DEVICE_TYPE_UMR) ||
                                      (getDeviceType() == DEVICE_TYPE_TKM && getDeviceNumber() >= 300))) {
      m_pPropertyNode->createProperty("AKMInputProperty")
          ->linkToProxy(PropertyProxyReference<std::string>(m_AKMInputProperty, false));
    }

  }

  int Device::getRevisionID() const {
    if (isVdcDevice()) {
      return 0;
    }
    return m_RevisionID;
  }

  std::string Device::getRevisionAsString() const {
    if (isVdcDevice()) {
      return "";
    }
    return ds::str((m_RevisionID & 0xff00) >> 8, '.', (m_RevisionID & 0xf0) >> 4, '.', m_RevisionID & 0xf);
  }

  int Device::getVendorID() const { return m_VendorID; }

  std::string Device::vendorTitle() const {
    switch (m_VendorID) {
      case 1:
        return "aizo GmbH";
      case 2:
        return "digitalSTROM AG";
      case 3:
        return "ONE SmartControl";
    }
    return "";
  }

  void Device::fillSensorTable(std::vector<DeviceSensorSpec_t>& _slist) {
    DeviceSensorSpec_t sensorInputReserved1 = {SensorType::Reserved1, 0, 0, 0};
    DeviceSensorSpec_t sensorInputReserved2 = {SensorType::Reserved2, 0, 0, 0};
    DeviceSensorSpec_t sensorInput04 = {SensorType::ActivePower, 0, 0, 0};
    DeviceSensorSpec_t sensorInput05 = {SensorType::OutputCurrent, 0, 0, 0};
    DeviceSensorSpec_t sensorInput06 = {SensorType::ElectricMeter, 0, 0, 0};
    DeviceSensorSpec_t sensorInput64 = {SensorType::OutputCurrent16A, 0, 0, 0};
    DeviceSensorSpec_t sensorInput65 = {SensorType::ActivePowerVA, 0, 0, 0};
    DeviceClasses_t deviceClass = getDeviceClass();
    int devType = (deviceClass << 16) | m_ProductID;

    if (isVdcDevice()) {
        // vdc devices have no implicit sensors, but 1st
        // generation vdc devices mocked themselves as Klemmen,
        // so they accidently got an implicit sensor attached
        // need to exit early here
        // http://redmine.digitalstrom.org/issues/8294
        return;
    }

    /* common */
    _slist.push_back(sensorInputReserved1);
    _slist.push_back(sensorInputReserved2);
    _slist.push_back(sensorInput04);
    _slist.push_back(sensorInput05);
    _slist.push_back(sensorInput06);

    switch (devType) {
      /* KM, SDM, TKM type with output */
      case 0x100c8:
      case 0x104d2:
      case 0x108c8:
      case 0x300c8:
      case 0x600c8:
      case 0x608c8:
      case 0x700c8:
      case 0x704c8:
      case 0x704d2:
        break;
      /* SDS200 with add. PowerVA sensor */
      case 0x118c8:
      case 0x118c9:
      case 0x118ca:
        _slist.push_back(sensorInput65);
        break;
      /* KL types */
      case 0x10cc8:
      case 0x20cc8:
      case 0x20cd2:
      case 0x20cdc:
      case 0x80cc8:
      case 0x814c8:
      case 0x814c9:
      case 0x814ca:
        _slist.push_back(sensorInput64);
        _slist.push_back(sensorInput65);
        break;
      /* all others - like TKM without output */
      default:
        _slist.clear();
        break;
    }
  } // fillSensorTable

  void Device::setDeviceConfig(uint8_t _configClass, uint8_t _configIndex, uint8_t _value, bool sendEvents) const {
    if (m_pApartment->getDeviceBusInterface() == NULL) {
      throw std::runtime_error("DeviceBusInterface missing");
    }

    m_pApartment->getDeviceBusInterface()->setDeviceConfig(
        *this, _configClass, _configIndex, _value);

    if (!sendEvents) {
      return;
    }
    if (auto modelMaintenance = m_pApartment->getModelMaintenance()) {
      ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceConfigChanged, m_DSMeterDSID);
      pEvent->addParameter(m_ShortAddress);
      pEvent->addParameter(_configClass);
      pEvent->addParameter(_configIndex);
      pEvent->addParameter(_value);
      modelMaintenance->addModelEvent(pEvent);
    }
  } // setDeviceConfig

  void Device::setDeviceConfig16(uint8_t _configClass, uint8_t _configIndex, uint16_t _value, bool sendEvents) const
  {
    /*
     * webroot/js/dss/dss-setup-interface/dSS/util/Util.js: dSS.util.decode16
     * -- used to configure lamella time(uint16_t) for KLEMME-GR using
     *  2x setDeviceConfig
     */
    if (m_pApartment->getDeviceBusInterface() == NULL) {
      throw std::runtime_error("DeviceBusInterface missing");
    }
    if (m_pApartment->getModelMaintenance() == NULL) {
      throw std::runtime_error("ModelMaintenance missing");
    }
    DeviceBusInterface *shorty = m_pApartment->getDeviceBusInterface();

    uint8_t low = (_value & 0x000000ff);
    uint8_t high = (_value & 0x0000ff00) >> 8;

    shorty->setDeviceConfig(*this, _configClass, _configIndex, low);
    shorty->setDeviceConfig(*this, _configClass, _configIndex + 1, high);

    if (!sendEvents) {
      return;
    }
    if (auto modelMaintenance = m_pApartment->getModelMaintenance()) {
      ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceConfigChanged, m_DSMeterDSID);
      pEvent->addParameter(m_ShortAddress);
      pEvent->addParameter(_configClass);
      pEvent->addParameter(_configIndex);
      pEvent->addParameter(_value);
      modelMaintenance->addModelEvent(pEvent);
    }
  }

  void Device::setDeviceButtonID(uint8_t _buttonId) {
    setButtonID(_buttonId);

    uint8_t value = m_ButtonID & 0x0f;
    if (m_ButtonActiveGroup < 16) {
      value |= m_ButtonActiveGroup << 4;
    }
    setDeviceConfig(CfgClassFunction, CfgFunction_ButtonMode, value);
  }

  void Device::setDeviceButtonActiveGroup(uint8_t _buttonActiveGroup) {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->setDeviceButtonActiveGroup(*this, _buttonActiveGroup);
      m_ButtonActiveGroup = _buttonActiveGroup;
      if (_buttonActiveGroup == BUTTON_ACTIVE_GROUP_RESET) {
        std::vector<boost::shared_ptr<Cluster>> oldClusters = getClusters();
        for (auto it = oldClusters.begin(); it != oldClusters.end(); ++it) {
          removeFromGroup((*it)->getID());
        }
        if (oldClusters.size() > 0) {
          int value = oldClusters.at(0)->getColor();
      
          DS_NOTICE("Restoring joker group due to cluster reset.", *this, value);
          m_ButtonActiveGroup = value;
          uint8_t buttonValue = m_ButtonID & 0x0f;
          if (m_ButtonActiveGroup < 16) {
            buttonValue |= m_ButtonActiveGroup << 4;
          }
          setDeviceConfig(CfgClassFunction, CfgFunction_ButtonMode, buttonValue);
        }
      }
      /* re-configure area or device button mode for groups other then lights and shades */
      bool isAreaButton =
          ((m_ButtonID >= ButtonId_Area1) && (m_ButtonID <= ButtonId_Area4)) ||
          ((m_ButtonID >= ButtonId_Area1_Extended) && (m_ButtonID <= ButtonId_Area4_Extended));
      if (isAreaButton && (_buttonActiveGroup != BUTTON_ACTIVE_GROUP_RESET) &&
          ((_buttonActiveGroup < GroupIDYellow) || (_buttonActiveGroup > GroupIDGray))) {
        setDeviceButtonID(ButtonId_Zone);
      }
    }
  }

  void Device::setDeviceJokerGroup(uint8_t groupId) {
    if (!isDefaultGroup(groupId) && !isGlobalAppDsGroup(groupId)) {
      throw std::runtime_error("Invalid joker group value");
    }
    int oldGroupId = m_ActiveGroup;

    // assign device to new group
    addToGroup(groupId);

    // check button target group
    if (getButtonInputCount() > 0 && groupId != m_ButtonActiveGroup) {

      // change from joker to a color -> reset button input to standard (for UMR???)
      if ((oldGroupId == GroupIDBlack) && (groupId != GroupIDBlack) &&
          hasInput() && (getButtonInputMode() != ButtonInputMode::STANDARD)
          // Do not override slave device button input mode
          // http://redmine.digitalstrom.org/issues/16805
          && !is2WaySlave()) {
        setDeviceButtonInputMode(ButtonInputMode::STANDARD);
      }

      setDeviceButtonActiveGroup(groupId);
    }

    // force refresh device information (-> busscanner readout) to update active group
    if ((m_pApartment != NULL) && (m_pApartment->getModelMaintenance() != NULL)) {
      ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceChanged, m_DSMeterDSID);
      pEvent->addParameter(m_ShortAddress);
      m_pApartment->getModelMaintenance()->addModelEvent(pEvent);
      sleep(3); // #8900: make sure all settings were really saved
    }

    updateIconPath();
    updateLedGroupColor();
    updateOutputMode();
  } // setDeviceJokerGroup

  void Device::setDeviceOutputMode(uint8_t _modeId) {
    setDeviceConfig(CfgClassFunction, CfgFunction_Mode, _modeId);
    deviceChanged();
  }

  void Device::setDeviceButtonInputMode(ButtonInputMode mode) {
    setDeviceConfig(CfgClassFunction, CfgFunction_LTMode, static_cast<uint8_t>(mode));
    setButtonInputMode(mode);
  } // setDeviceButtonInputMode

  void Device::setProgMode(uint8_t _modeId) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->setDeviceProgMode(*this, _modeId);
    }
  } // setProgMode

  boost::shared_ptr<Device> Device::tryGetPartnerDevice() const {
    dsuid_t next;
    dsuid_get_next_dsuid(getDSID(), &next);
    return getApartment().tryGetDeviceByDSID(next);
  }

  boost::shared_ptr<Device> Device::getPartnerDevice() const {
    auto device = tryGetPartnerDevice();
    DS_REQUIRE(device, "Failed to find partner device.", m_DSID);
    return device;
  }

 void Device::increaseDeviceOutputChannelValue(uint8_t _channel) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->increaseDeviceOutputChannelValue(*this, _channel);
    }
  } // increaseDeviceOutputChannelValue

  void Device::decreaseDeviceOutputChannelValue(uint8_t _channel) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->decreaseDeviceOutputChannelValue(*this, _channel);
    }
  } // decreaseDeviceOutputChannelValue

  void Device::stopDeviceOutputChannelValue(uint8_t _channel) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->stopDeviceOutputChannelValue(*this, _channel);
    }
  } // decreaseDeviceOutputChannelValue

  uint16_t Device::getDeviceOutputChannelValue(uint8_t _channel) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getDeviceOutputChannelValue(*this, _channel);
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceOutputChannelValue

  void Device::setDeviceOutputChannelValue(uint8_t channel, uint8_t size, uint16_t value, bool applyNow) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      auto type = getOutputChannel(channel)->type;
      if (m_emulatedHardwareOutputs) {
        // doc: https://confluence.digitalstrom.org/display/DSD/Output+channels
        if (channel == 0 && type == ChannelType::brightness) {
          return setDeviceValue(value);
        }
        if (channel == 0 && type == ChannelType::shadePositionOutside) {
          return setDeviceOutputValue(2, value);
        }
        if (channel == 1 && type == ChannelType::shadeOpeningAngleOutside) {
          return setDeviceConfig(64, 4, value);
        }
        DS_FAIL_REQUIRE("Set output channel value not supported yet for this device.", *this, hardwareType(), type,
                        channel, size, value);
      }
      m_pApartment->getDeviceBusInterface()->setDeviceOutputChannelValue(*this, channel, size, value, applyNow);
      m_pApartment->proxy().setOutputValue(getDSID(), type, convertFromOutputChannelValue(type, value), std::chrono::high_resolution_clock::now());
    }
  }

  uint16_t Device::getDeviceOutputChannelSceneValue(uint8_t _channel,
                                                    uint8_t _scene) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getDeviceOutputChannelSceneValue(*this, _channel, _scene);
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceOutputChannelSceneValue

  void Device::setDeviceOutputChannelSceneValue(uint8_t _channel,
                                                uint8_t _size,
                                                uint8_t _scene,
                                                uint16_t _value) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->setDeviceOutputChannelSceneValue(*this, _channel, _size, _scene, _value);
    }
  }

  void Device::getDeviceOutputChannelSceneConfig(uint8_t _scene,
                                                 DeviceSceneSpec_t& _config) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      uint16_t mode = m_pApartment->getDeviceBusInterface()->getDeviceOutputChannelSceneConfig(*this, _scene);
      _config.dontcare = (mode & 1) > 0;
      _config.localprio = (mode & 2) > 0;
      _config.specialmode = (mode & 4) > 0;
      _config.ledconIndex = (mode >> 4) & 7;
      _config.dimtimeIndex = 0; // no dim time index
      _config.flashmode = (mode & (1 << 8)) > 0; // effect 1
      return;
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceOutputChannelSceneConfig

  void Device::setDeviceOutputChannelSceneConfig(uint8_t _scene,
                                                 DeviceSceneSpec_t _config) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      uint16_t mode = _config.dontcare ? 1 : 0;
      mode |= _config.localprio ? 2 : 0;
      mode |= _config.specialmode ? 4 : 0;
      mode |= ((_config.ledconIndex & 7) << 4);
      mode |= _config.flashmode ? (1 << 8) : 0;

      m_pApartment->getDeviceBusInterface()->setDeviceOutputChannelSceneConfig(*this, _scene, mode);
    }
  } // setDeviceOutputChannelSceneConfig

  void Device::setDeviceSceneModeStandard(uint8_t _sceneId, DeviceSceneSpec_t _config) const {
    uint8_t mode = _config.dontcare ? 1 : 0;
    mode |= _config.localprio ? 2 : 0;
    mode |= _config.specialmode ? 4 : 0;
    mode |= _config.flashmode ? 8 : 0;
    mode |= ((_config.ledconIndex & 3) << 4);
    mode |= ((_config.dimtimeIndex & 3) << 6);
    setDeviceConfig(CfgClassScene, 128 + _sceneId, mode);
  } // setDeviceSceneMode

  void Device::setDeviceSceneMode(uint8_t sceneId, DeviceSceneSpec_t config) const {
    if (getProductID() == ProductID_UMV_210) {
      setDeviceOutputChannelSceneConfig(sceneId, config);
    } else {
      setDeviceSceneModeStandard(sceneId, config);
    }
  }

  void Device::getDeviceSceneModeStandard(uint8_t _sceneId, DeviceSceneSpec_t& _config) const {
    uint8_t mode = getDeviceConfig(CfgClassScene, 128 + _sceneId);
    _config.dontcare = (mode & 1) > 0;
    _config.localprio = (mode & 2) > 0;
    _config.specialmode = (mode & 4) > 0;
    _config.flashmode = (mode & 8) > 0;
    _config.ledconIndex = (mode >> 4) & 3;
    _config.dimtimeIndex = (mode >> 6) & 3;
  } // getDeviceSceneMode

  void Device::getDeviceSceneMode(uint8_t sceneId, DeviceSceneSpec_t &config) const {
    if (getProductID() == ProductID_UMV_210) {
      getDeviceOutputChannelSceneConfig(sceneId, config);
    } else {
      getDeviceSceneModeStandard(sceneId, config);
    }
  }

  void Device::setDeviceOutputChannelDontCareFlags(uint8_t _scene,
                                                   uint16_t _value) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->setDeviceOutputChannelDontCareFlags(*this, _scene, _value);
    }
  }

  uint16_t Device::getDeviceOutputChannelDontCareFlags(uint8_t _scene) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getDeviceOutputChannelDontCareFlags(*this, _scene);
    }
    throw std::runtime_error("Bus interface not available");
  }

  int Device::getSceneAngle(const int _scene) const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (features.hasOutputAngle) {
      return getDeviceConfig(CfgClassSceneAngle, _scene);
    } else {
      throw std::runtime_error("Device does not support output angle setting");
    }
  }

  void Device::setSceneAngle(const int _scene, const int _angle) const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.hasOutputAngle) {
      throw std::runtime_error("Device does not support output angle setting");
    }
    if ((_angle < 0) || (_angle > 255)) {
      throw std::runtime_error("Device angle value out of bounds");
    }

    setDeviceConfig(CfgClassSceneAngle, _scene, _angle);
  }

  void Device::setDeviceLedMode(uint8_t _ledconIndex, DeviceLedSpec_t _config) const {
    if (_ledconIndex > 2) {
      throw DSSException("Device::setDeviceLedMode: index out of range");
    }
    uint8_t mode = _config.colorSelect & 7;
    mode |= ((_config.modeSelect & 3) << 3);
    mode |= _config.dimMode ? 32 : 0;
    mode |= _config.rgbMode ? 64 : 0;
    mode |= _config.groupColorMode ? 128 : 0;
    setDeviceConfig(CfgClassFunction, CfgFunction_LedConfig0 + _ledconIndex, mode);
  }  // setDeviceLedMode

  void Device::getDeviceLedMode(uint8_t _ledconIndex, DeviceLedSpec_t& _config) const {
    if (_ledconIndex > 2) {
      throw DSSException("Device::getDeviceLedMode: index out of range");
    }
    uint8_t mode = getDeviceConfig(CfgClassFunction, CfgFunction_LedConfig0 + _ledconIndex);
    _config.colorSelect = mode & 7;
    _config.modeSelect = (mode >> 3) & 3;
    _config.dimMode = (mode & 32) > 0;
    _config.rgbMode = (mode & 64) > 0;
    _config.groupColorMode = (mode & 128) > 0;
  } // getDeviceLedMode

  int Device::transitionVal2Time(uint8_t _value) const {
    int hval = _value / 16;
    int lval = _value % 16;
    return 100 * (1 << hval) - (100 * (1 << hval) / 32) * (15 - lval);
  }

  uint8_t Device::transitionTimeEval(int timems) const {
    uint8_t val;
    int thigh, tlow;
    for (val = 0, tlow = thigh = 0; val < 255 && thigh < timems; ) {
      val ++;
      tlow = thigh;
      thigh = transitionVal2Time(val);
    }
    if ((thigh - timems) > (timems - tlow)) {
      val --;
    }
    return val;
  }

  void Device::setDeviceTransitionTime(uint8_t _dimtimeIndex, int up, int down) const {
    if (_dimtimeIndex > 2) {
      throw DSSException("Device::setDeviceTransitionTime: index out of range");
    }
    if (up >= 0) {
      uint8_t vup = transitionTimeEval(up);
      setDeviceConfig(CfgClassFunction, CfgFunction_DimTime0 + _dimtimeIndex*2, vup);
    }

    if (down >= 0) {
      uint8_t vdown  = transitionTimeEval(down);
      setDeviceConfig(CfgClassFunction, CfgFunction_DimTime0 + _dimtimeIndex*2 + 1, vdown);
    }
  } // setDeviceTransitionTime

  void Device::getDeviceTransitionTime(uint8_t _dimtimeIndex, int& up, int& down) const {
    if (_dimtimeIndex > 2) {
      throw DSSException("Device::getDeviceTransitionTime: index out of range");
    }
    uint16_t value = getDeviceConfigWord(CfgClassFunction, CfgFunction_DimTime0 + _dimtimeIndex*2);
    uint8_t vup = value & 0xff;
    uint8_t vdown = (value >> 8) & 0xff;
    up = transitionVal2Time(vup);
    down = transitionVal2Time(vdown);
  } // getDeviceTransitionTime

  /** Configure climate actuator */
  void Device::setDeviceValveTimer(DeviceValveTimerSpec_t _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    if (isBLKM300()) {
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_EmergencyValue, _config.emergencyControlValue);
      setDeviceConfig16(CfgClassFunction, CfgFunction_BL_KM300_Valve_EmergencyTimer, _config.emergencyTimer);
      setDeviceConfig16(CfgClassFunction, CfgFunction_BL_KM300_Valve_ProtectionTimer, _config.protectionTimer);
    } else {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_EmergencyValue, _config.emergencyControlValue);
      setDeviceConfig16(CfgClassFunction, CfgFunction_Valve_EmergencyTimer, _config.emergencyTimer);
      setDeviceConfig16(CfgClassFunction, CfgFunction_Valve_ProtectionTimer, _config.protectionTimer);
    }
  } // setDeviceValveTimer

  void Device::getDeviceValveTimer(DeviceValveTimerSpec_t& _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }

    if (isBLKM300()) {
      _config.emergencyControlValue = getDeviceConfig(CfgClassFunction,
              CfgFunction_BL_KM300_Valve_EmergencyValue);
      _config.emergencyTimer = getDeviceConfigWord(CfgClassFunction,
              CfgFunction_BL_KM300_Valve_EmergencyTimer);
      _config.protectionTimer = getDeviceConfigWord(CfgClassFunction,
              CfgFunction_BL_KM300_Valve_ProtectionTimer);
    } else {
      _config.emergencyControlValue = getDeviceConfig(CfgClassFunction,
              CfgFunction_Valve_EmergencyValue);
      _config.emergencyTimer = getDeviceConfigWord(CfgClassFunction,
              CfgFunction_Valve_EmergencyTimer);
      _config.protectionTimer = getDeviceConfigWord(CfgClassFunction,
              CfgFunction_Valve_ProtectionTimer);
    }
  } // getDeviceValveTimer

  void Device::setDeviceValvePwm(DeviceValvePwmSpec_t _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    if (isBLKM300()) {
      setDeviceConfig16(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmPeriod, _config.pwmPeriod);
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmMinX, _config.pwmMinX);
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmMaxX, _config.pwmMaxX);
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmMinY, _config.pwmMinY);
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmMaxY, _config.pwmMaxY);
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmOffset, static_cast<uint8_t>(_config.pwmOffset));
    } else {
      setDeviceConfig16(CfgClassFunction, CfgFunction_Valve_PwmPeriod, _config.pwmPeriod);
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMinX, _config.pwmMinX);
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMaxX, _config.pwmMaxX);
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMinY, _config.pwmMinY);
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMaxY, _config.pwmMaxY);
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmOffset, static_cast<uint8_t>(_config.pwmOffset));
    }
  } // setDeviceValvePwm

  void Device::getDeviceValvePwm(DeviceValvePwmSpec_t& _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    uint16_t value;
    if (isBLKM300()) {
      _config.pwmPeriod = getDeviceConfigWord(CfgClassFunction,
                  CfgFunction_BL_KM300_Valve_PwmPeriod);
      value = getDeviceConfigWord(CfgClassFunction,
                  CfgFunction_BL_KM300_Valve_PwmMinX);
    } else {
      _config.pwmPeriod = getDeviceConfigWord(CfgClassFunction,
                  CfgFunction_Valve_PwmPeriod);
      value = getDeviceConfigWord(CfgClassFunction, CfgFunction_Valve_PwmMinX);
    }
    _config.pwmMinX = value & 0xff; // CfgFunction_Valve_PwmMinX
    _config.pwmMaxX = (value >> 8) & 0xff; // CfgFunction_Valve_PwmMaxX
    if (isBLKM300()) {
      value = getDeviceConfigWord(CfgClassFunction,
                  CfgFunction_BL_KM300_Valve_PwmMinY);
    } else {
      value = getDeviceConfigWord(CfgClassFunction, CfgFunction_Valve_PwmMinY);
    }
    _config.pwmMinY = value & 0xff; // CfgFunction_Valve_PwmMinY
    _config.pwmMaxY = (value >> 8) & 0xff; // CfgFunction_Valve_PwmMaxY
    if (isBLKM300()) {
      _config.pwmOffset = static_cast<int8_t>(getDeviceConfig(CfgClassFunction,
                  CfgFunction_BL_KM300_Valve_PwmOffset));
    } else {
      _config.pwmOffset = static_cast<int8_t>(getDeviceConfig(CfgClassFunction,
                  CfgFunction_Valve_PwmOffset));
    }
  } // getDeviceValvePwm

  void Device::getDeviceValvePwmState(DeviceValvePwmStateSpec_t& _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    uint16_t value = getDeviceConfigWord(CfgClassRuntime, CfgRuntime_Valve_PwmValue);
    _config.pwmValue = value & 0xff;
    _config.pwmPriorityMode = (value >> 8) & 0xff;
  } // getDeviceValvePwmState

  void Device::setDeviceValveControl(DeviceValveControlSpec_t _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    uint8_t value = _config.ctrlRawValue;
    value &= 0xf0; // clear bits 0..3, reserved bits 4..7
    value |= _config.ctrlClipMinZero ? 1 : 0;
    value |= _config.ctrlClipMinLower? 2 : 0;
    value |= _config.ctrlClipMaxHigher ? 4 : 0;
    value |= _config.ctrlNONC ? 8 : 0;
    if (isBLKM300()) {
      setDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmConfig, value);
    } else {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmConfig, value);
    }
  }

  void Device::getDeviceValveControl(DeviceValveControlSpec_t& _config) const {
    if (!isValveDevice()) {
      throw DSSException("Not a climate device");
    }
    uint8_t value;
    if (isBLKM300()) {
      value = getDeviceConfig(CfgClassFunction, CfgFunction_BL_KM300_Valve_PwmConfig);
    } else {
      value = getDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmConfig);
    }
    _config.ctrlClipMinZero = ((value & 0x01) == 0x01);
    _config.ctrlClipMinLower = ((value & 0x02) == 0x02);
    _config.ctrlClipMaxHigher = ((value & 0x04) == 0x04);
    _config.ctrlNONC = ((value & 0x08) == 0x08);
    _config.ctrlRawValue = value;
  } // getDeviceValveControl

  void Device::setDeviceValveSwitchedThresholdMin(uint8_t value) const {
    if (isBLKM300()) {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_SwitchedMin, value);
    } else {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMinY, value);
    }
  }

  void Device::getDeviceValveSwitchedThresholdMin(uint8_t *threshold) const {
    if (threshold == NULL) {
      throw DSSException("invalid parameter pointer");
    }

    if (isBLKM300()) {
      *threshold = getDeviceConfig(CfgClassFunction,
                                   CfgFunction_Valve_SwitchedMin);
    } else {
      *threshold = getDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMinY);
    }
  }

  void Device::setDeviceValveSwitchedThresholdMax(uint8_t value) const {
    if (isBLKM300()) {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_SwitchedMax, value);
    } else {
      setDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMaxX, value);
    }
  }

  void Device::getDeviceValveSwitchedThresholdMax(uint8_t *threshold) const {
    if (threshold == NULL) {
      throw DSSException("invalid parameter pointer");
    }
    if (isBLKM300()) {
      *threshold = getDeviceConfig(CfgClassFunction,
                                   CfgFunction_Valve_SwitchedMax);
    } else {
      *threshold = getDeviceConfig(CfgClassFunction, CfgFunction_Valve_PwmMaxX);
    }
  }

  uint8_t Device::getDeviceConfig(uint8_t _configClass, uint8_t _configIndex) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getDeviceConfig(*this,
                                                                  _configClass,
                                                                  _configIndex);
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceConfig

  uint16_t Device::getDeviceConfigWord(uint8_t _configClass, uint8_t _configIndex) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getDeviceConfigWord(*this,
                                                                  _configClass,
                                                                  _configIndex);
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceConfigWord

  std::pair<uint8_t, uint16_t> Device::getDeviceTransmissionQuality() const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      return m_pApartment->getDeviceBusInterface()->getTransmissionQuality(*this);
    }
    throw std::runtime_error("Bus interface not available");
  } // getDeviceTransmissionQuality

  void Device::setDeviceValue(uint8_t _value) const {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->setValue(*this, _value);
    }
  } // setDeviceValue (8)

  void Device::setDeviceOutputValue(uint8_t _offset, uint16_t _value) const {
    if (_offset & 1) {
      setDeviceConfig(CfgClassRuntime, _offset, _value & 0xff);
    } else {
      setDeviceConfig(CfgClassRuntime, _offset+1, (_value >> 8) & 0xff, false);
      setDeviceConfig(CfgClassRuntime, _offset, _value & 0xff, false);
      auto type = _offset == 2 ? ChannelType::shadePositionOutside : ChannelType::shadeOpeningAngleOutside;
      m_pApartment->proxy().setOutputValue(getDSID(), type, convertFromOutputChannelValue(type, _value), std::chrono::high_resolution_clock::now());
    }
  } // setDeviceOutputValue (offset)

  uint16_t Device::getDeviceOutputValue(uint8_t _offset) const {
    uint16_t result = getDeviceConfigWord(CfgClassRuntime, _offset);
    if (_offset == 0) result &= 0xff;   // fix offset 0 value which is 8-bit actually
    return result;
  } // getDeviceOutputValue (offset)

  uint16_t Device::getDeviceSensorValue(const int _sensorIndex) const {
    return m_pApartment->getDeviceBusInterface()->getSensorValue(*this, _sensorIndex);
  } // getDeviceSensorValue

  DeviceSensorValue_t Device::getDeviceSensorValueEx(const int _sensorIndex) const {
    DeviceSensorValue_t result;
    if (m_isVdcDevice) {
      result = m_pApartment->getDeviceBusInterface()->getSensorValueEx(*this, _sensorIndex);
    } else {
      // TODO(someday): support age and double value on power line
      result.sensorIndex = _sensorIndex;
      result.sensorType = getSensor(_sensorIndex).m_sensorType;
      result.value = static_cast<double> (m_pApartment->getDeviceBusInterface()->getSensorValue(*this, _sensorIndex));
      result.timestamp = DateTime();
    }
    return result;
  } // getDeviceSensorValueEx

  unsigned long Device::getPowerConsumption() {
    return m_Consumption;
  } // getPowerConsumption

  unsigned long Device::getEnergymeterValue() {
    return m_Energymeter;
  } // getEnergymeterValue

  void Device::nextScene(const callOrigin_t _origin, const SceneAccessCategory _category) {
    callScene(_origin, _category, SceneHelper::getNextScene(m_LastCalledScene),  "", false);
  } // nextScene

  void Device::previousScene(const callOrigin_t _origin, const SceneAccessCategory _category) {
    callScene(_origin, _category, SceneHelper::getNextScene(m_LastCalledScene), "", false);
  } // previousScene

  const std::string& Device::getName() const {
    return m_Name;
  } // getName

  void Device::setName(const std::string& _name) {
    if (m_Name != _name) {
      m_Name = _name;
      dirty();
    }
  } // setName

  void Device::dirty() {
    if ((m_pApartment != NULL) && (m_pApartment->getModelMaintenance() != NULL)) {
      m_pApartment->getModelMaintenance()->addModelEvent(
          new ModelEvent(ModelEvent::etModelDirty)
      );
    }
  } // dirty

  void Device::checkAutoCluster() const {
    if ((m_pApartment != NULL) && (m_pApartment->getModelMaintenance() != NULL)) {
      ModelEventWithDSID* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceDirty, getDSID());
      m_pApartment->getModelMaintenance()->addModelEvent(pEvent);
    }
  }

  bool Device::operator==(const Device& _other) const {
    return (_other.m_DSID == m_DSID);
  } // operator==

  void Device::setShortAddress(const devid_t _shortAddress) {
    m_ShortAddress = _shortAddress;
    m_LastKnownShortAddress = _shortAddress;
    publishToPropertyTree();
    m_LastDiscovered = DateTime();
  } // setShortAddress

  devid_t Device::getLastKnownShortAddress() const {
    return m_LastKnownShortAddress;
  } // getLastKnownShortAddress

  void Device::setLastKnownShortAddress(const devid_t _shortAddress) {
    m_LastKnownShortAddress = _shortAddress;
  } // setLastKnownShortAddress

  dsuid_t Device::getDSID() const {
    return m_DSID;
  } // getDSID;

  dsuid_t Device::getDSMeterDSID() const {
    return m_DSMeterDSID;
  } // getDSMeterID

  void Device::setLastKnownDSMeterDSID(const dsuid_t& _value) {
    m_LastKnownMeterDSID = _value;
  } // setLastKnownDSMeterDSID

  const dsuid_t& Device::getLastKnownDSMeterDSID() const {
    return m_LastKnownMeterDSID;
  } // getLastKnownDSMeterDSID

  void Device::setDSMeter(boost::shared_ptr<DSMeter> _dsMeter) {
    PropertyNodePtr alias;
    std::string devicePath = "devices/" + dsuid2str(m_DSID);
    if ((m_pPropertyNode != NULL) && (m_DSMeterDSID != DSUID_NULL)) {
      alias = m_pApartment->getDSMeterByDSID(m_DSMeterDSID)->getPropertyNode()->getProperty(devicePath);
    }
    m_DSMeterDSID = _dsMeter->getDSID();
    m_LastKnownMeterDSID = _dsMeter->getDSID();
    m_DSMeterDSUIDstr = dsuid2str(_dsMeter->getDSID());
    m_LastKnownMeterDSUIDstr = dsuid2str(_dsMeter->getDSID());

    dsid_t dsid;
    if (dsuid_to_dsid(m_DSMeterDSID, &dsid)) {
      m_DSMeterDSIDstr = dsid2str(dsid);
    } else {
      m_DSMeterDSIDstr = "";
    }
    if (dsuid_to_dsid(m_LastKnownMeterDSID, &dsid)) {
      m_LastKnownMeterDSIDstr = dsid2str(dsid);
    } else {
      m_LastKnownMeterDSIDstr = "";
    }

    if (m_pPropertyNode != NULL) {
      PropertyNodePtr target = _dsMeter->getPropertyNode()->createProperty("devices");
      if (alias != NULL) {
        target->addChild(alias);
      } else {
        alias = target->createProperty(dsuid2str(m_DSID));
        alias->alias(m_pPropertyNode);
      }
    }
  } // setDSMeter

  int Device::getZoneID() const {
    return m_ZoneID;
  } // getZoneID

  void Device::setZoneID(const int _value) {
    if (_value == m_ZoneID) {
      // nothing to do
      return;
    }

    if (_value != 0) {
      m_LastKnownZoneID = _value;
    }

    m_ZoneID = _value;
    if ((m_pPropertyNode != NULL) && (m_pApartment->getPropertyNode() != NULL)) {
      std::string basePath = "zones/zone" + intToString(m_ZoneID) + "/devices";
      if (m_pAliasNode == NULL) {
        PropertyNodePtr node = m_pApartment->getPropertyNode()->getProperty(basePath + "/" + dsuid2str(m_DSID));
        if (node != NULL) {
          log("Device::setZoneID: Target node for device " + dsuid2str(m_DSID) + " already exists", lsError);
          if (node->size() > 0) {
            log("Device::setZoneID: Target node for device " + dsuid2str(m_DSID) + " has children", lsFatal);
            return;
          }
        }
        m_pAliasNode = m_pApartment->getPropertyNode()->createProperty(basePath + "/" + dsuid2str(m_DSID));
        m_pAliasNode->alias(m_pPropertyNode);
      } else {
        PropertyNodePtr base = m_pApartment->getPropertyNode()->getProperty(basePath);
        if (base == NULL) {
          throw std::runtime_error("PropertyNode of the new zone does not exist");
        }
        base->addChild(m_pAliasNode);
      }

      publishAliasInZoneGroups();
    }
    // Binary input status bit handles depend on zoneId by call to `tryGetGroup`.
    // TODO(someday): refactor to some kind of observer pattern?
    for (auto&& binaryInput : m_binaryInputs) {
      binaryInput->updateStatusFieldHandle();
    }
  } // setZoneID

  int Device::getLastKnownZoneID() const {
    return m_LastKnownZoneID;
  } // getLastKnownZoneID

  void Device::setLastKnownZoneID(const int _value) {
    m_LastKnownZoneID = _value;
  } // setLastKnownZoneID

  void Device::setIgnoreOperationLock(const bool value) {
    m_IgnoreOperationLock = value;
  } // setIgnoreOperationLock

  void Device::setDeviceIgnoreOperationLock(const bool value) {
    if (m_RevisionID >= 0x365) {
      setIgnoreOperationLock(value);
      if (m_pApartment->getDeviceBusInterface() != NULL) {
        //Bank 5 - Scene action table; Offset 10
        //lock enabled: 120, disabled: 0
        m_pApartment->getDeviceBusInterface()->setDeviceConfig(*this, 0x5, 0x0a, (value ? 0 : 120));
        dirty();
      } else {
        log("Bus interface not available", lsError);
      }
    } else if (!isVdcDevice()) {
      log("Device " + dsuid2str(m_DSID) + ": Could not set ignore operation lock: FW version " + intToString(m_RevisionID) + " is lower than 0x365", lsWarning);
    }
  } // setDeviceIgnoreOperationLock

  int Device::getGroupIdByIndex(const int index) const {
    return m_groupIds.at(index);
  } // getGroupIdByIndex

  boost::shared_ptr<Group> Device::getGroupByIndex(const int _index) {
    return m_pApartment->getGroup(getGroupIdByIndex(_index));
  } // getGroupByIndex

  void Device::setBinaryStateTargetGroup(const int groupID) {
    if (isDefaultGroup(groupID) || isGlobalAppDsGroup(groupID)) {
      std::vector<int>::iterator it = m_groupIds.begin();
      while (it != m_groupIds.end()) {
        if ((isDefaultGroup(*it) && *it != GroupIDBlack) || isGlobalAppDsGroup(*it)) {
          int index = it - m_groupIds.begin();
          removeFromGroup(*it);
          int newIndex = index - 1;
          if (newIndex < 0) {
            newIndex = 0;
          }
          it = m_groupIds.begin() + newIndex;
        } else {
          ++it;
        }
      }
      addToGroup(groupID);
    } else {
      log("Device " + dsuid2str(m_DSID) + ": Invalid Group " + intToString(groupID) + " could not be assigned for Binary Input Target Group", lsWarning);
    }
  } // setBinaryStateTargetGroup

  void Device::addToGroup(const int _groupID) {
    if (isValidGroup(_groupID)) {
      if (find(m_groupIds.begin(), m_groupIds.end(), _groupID) == m_groupIds.end()) {
        m_groupIds.push_back(_groupID);
        updateIconPath();
        if ((m_pPropertyNode != NULL) && (m_pApartment->getPropertyNode() != NULL)) {
          // create alias in group list
          std::string gPath = "zones/zone" + intToString(getGroupZoneID(_groupID)) + "/groups/group" + intToString(_groupID) + "/devices/"  +  dsuid2str(m_DSID);
          PropertyNodePtr gnode = m_pApartment->getPropertyNode()->createProperty(gPath);
          if (gnode) {
            gnode->alias(m_pPropertyNode);
          }
          // create group-n property
          PropertyNodePtr gsubnode = m_pPropertyNode->createProperty("groups/group" + intToString(_groupID));
          gsubnode->createProperty("id")->setIntegerValue(_groupID);
          log("Device " + dsuid2str(m_DSID) + " (bus: " + intToString(m_ShortAddress) + ", zone: " + intToString(m_ZoneID) + ") was internally added to group " + intToString(_groupID), lsDebug);
        }
      } else {
        log("Device " + dsuid2str(m_DSID) + " (bus: " + intToString(m_ShortAddress) + ", zone: " + intToString(m_ZoneID) + ") is already in group " + intToString(_groupID), lsDebug);
      }
    } else {
      log("Device::addToGroup: Group ID out of bounds: " + intToString(_groupID), lsInfo);
    }
  } // addToGroup

  void Device::removeFromGroup(const int _groupID) {
    if (isValidGroup(_groupID)) {
      std::vector<int>::iterator it = find(m_groupIds.begin(), m_groupIds.end(), _groupID);
      if (it != m_groupIds.end()) {
        m_groupIds.erase(it);
        updateIconPath();
        if ((m_pPropertyNode != NULL) && (m_pApartment->getPropertyNode() != NULL)) {
          // remove alias in group list
          std::string gPath = "zones/zone" + intToString(getGroupZoneID(_groupID)) + "/groups/group" + intToString(_groupID) + "/devices/"  +  dsuid2str(m_DSID);
          PropertyNodePtr gnode = m_pApartment->getPropertyNode()->getProperty(gPath);
          if (gnode) {
            gnode->getParentNode()->removeChild(gnode);
          }
          // remove property in device group list
          PropertyNodePtr gsubnode = m_pPropertyNode->getProperty("groups/group" + intToString(_groupID));
          if (gsubnode) {
            gsubnode->getParentNode()->removeChild(gsubnode);
          }
        }
      }
    } else {
      log("Device::removeFromGroup: Group ID out of bounds: " + intToString(_groupID), lsInfo);
    }
  } // removeFromGroup

  void Device::resetGroups() {
    while (m_groupIds.size() > 0) {
      int g = m_groupIds.front();
      removeFromGroup(g);
    }
  } // resetGroups

  int Device::getGroupsCount() const {
    return m_groupIds.size();
  } // getGroupsCount

  bool Device::isInGroup(const int _groupID) const {
    bool result = false;
    if (_groupID == 0) {
      result = true;
    } else if (!isValidGroup(_groupID)) {
      result = false;
    } else {
      auto it = find(m_groupIds.begin(), m_groupIds.end(), _groupID);
      result = (it != m_groupIds.end());
    }
    return result;
  } // isInGroup

  void Device::setIsLockedInDSM(const bool _value) {
    m_IsLockedInDSM = _value;
  } // setIsLockedInDSM

  void Device::lock() {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->lockOrUnlockDevice(*this, true);
      m_IsLockedInDSM = true;
    }
  } // lock

  void Device::unlock() {
    if (m_pApartment->getDeviceBusInterface() != NULL) {
      m_pApartment->getDeviceBusInterface()->lockOrUnlockDevice(*this, false);
      m_IsLockedInDSM = false;
    }
  } // unlock

  bool Device::hasTag(const std::string& _tagName) const {
    if (m_TagsNode != NULL) {
      return m_TagsNode->getPropertyByName(_tagName) != NULL;
    }
    return false;
  } // hasTag

  void Device::addTag(const std::string& _tagName) {
    if (!hasTag(_tagName)) {
      if (m_TagsNode != NULL) {
        PropertyNodePtr pNode = m_TagsNode->createProperty(_tagName);
        pNode->setFlag(PropertyNode::Archive, true);
        dirty();
      }
    }
  } // addTag

  void Device::removeTag(const std::string& _tagName) {
    if (hasTag(_tagName)) {
      if (m_TagsNode != NULL) {
        m_TagsNode->removeChild(m_TagsNode->getPropertyByName(_tagName));
        dirty();
      }
    }
  } // removeTag

  std::vector<std::string> Device::getTags() {
    std::vector<std::string> result;
    if (m_TagsNode != NULL) {
      int count = m_TagsNode->getChildCount();
      for(int iNode = 0; iNode < count; iNode++) {
        result.push_back(m_TagsNode->getChild(iNode)->getName());
      }
    }
    return result;
  } // getTags

  std::string Device::getSensorEventName(const int _eventIndex) const {
    if (m_pPropertyNode != NULL) {
      PropertyNodePtr nameNode = m_pPropertyNode->getProperty("sensorEvents/event" + intToString(_eventIndex));
      if (nameNode) {
        return nameNode->getStringValue();
      }
    }
    return "";
  }

  void Device::setSensorEventName(const int _eventIndex, const std::string& _name) {
    if (m_pPropertyNode != NULL) {
      PropertyNodePtr nameNode = m_pPropertyNode->createProperty("sensorEvents/event" + intToString(_eventIndex));
      if (_name.empty()) {
        m_pPropertyNode->getProperty("sensorEvents")->removeChild(nameNode);
      } else {
        nameNode->setStringValue(_name);
        nameNode->setFlag(PropertyNode::Archive, true);
      }
      dirty();
    }
  }

  void Device::getSensorEventTableEntryZws205(const int row, DeviceSensorEventSpec_t& entry) {
    DS_REQUIRE(row < 2, "ZWS205 only supports on/off consumption event");

    entry = DeviceSensorEventSpec_t();
    entry.name = getSensorEventName(row);

    // read 16bit per call
    auto value0_1 = getDeviceConfigWord(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 0);
    auto value2_3 = getDeviceConfigWord(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 2);

    entry.action = (value0_1 & 0x0003);
    entry.test = (value0_1 & 0x000C) >> 2;
    entry.sensorIndex = (value0_1 & 0x00F0) >> 4;

    entry.value = ((value0_1 & 0xFF00) | (value2_3 & 0x00F0)) >> 4;
    entry.hysteresis = 0;

    auto minimalDurationOffset = (row == 0) ? 0x30 : 0x32;
    entry.minimalDuration = getDeviceConfig(CfgClassSensorEvent, minimalDurationOffset);
  }

  void Device::getSensorEventEntry(const int _eventIndex, DeviceSensorEventSpec_t& _entry) {
    DS_REQUIRE(_eventIndex <= 7, "row index exceeds table size");
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      getSensorEventTableEntryZws205(_eventIndex, _entry);
      return;
    }

    _entry.name = getSensorEventName(_eventIndex);
    uint16_t value = getDeviceConfigWord(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 0);
    _entry.sensorIndex = (value & 0x00F0) >> 4;
    _entry.test = (value & 0x000C) >> 2;
    _entry.action = (value & 0x0003);
    _entry.value = (value & 0xFF00) >> 4;
    value = getDeviceConfigWord(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 2);
    _entry.value |= (value & 0x00F0) >> 4;
    _entry.hysteresis = ((value & 0x000F) << 8) | ((value & 0xFF00) >> 8);
    value = getDeviceConfigWord(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 4);
    _entry.sceneDeviceMode = (value & 0x000C) >> 2;
    _entry.validity = (value & 0x0003);
    _entry.buttonNumber = (value & 0xF000) >> 12;
    _entry.clickType = (value & 0x0F00) >> 8;
    _entry.sceneID = (value & 0x7F00) >> 8;
  }

  void Device::setSensorEventTableEntryZws205(const int row, const DeviceSensorEventSpec_t entry) {
    // see ds-basics: Appendix C.4 "Sensor Event Table"
    DS_REQUIRE(row < 2, "ZWS205 only supports on/off consumption event ", row);
    DS_REQUIRE(entry.sensorIndex <= 0xff, "sensor index exceeded");
    DS_REQUIRE(entry.test != 0x3, "invalid comparison operator");
    DS_REQUIRE(entry.action == 0 || entry.action == 0x3, "on/off action only");
    uint8_t offset0 = (entry.sensorIndex << 4 | entry.test << 2 | entry.action);

    DS_REQUIRE(entry.value >= 5, "ZWS205 has 5W lower bound limit");
    uint8_t offset1 = (entry.value & 0xFF0) >> 4;
    uint8_t offset2 = (entry.value & 0x00F) << 4;

    if (entry.hysteresis != 0) {
      DS_NOTICE("ZWS205, does not support hysteresis, ignored");
    }

    setDeviceConfig(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 0, offset0);
    setDeviceConfig(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 1, offset1);
    setDeviceConfig(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 2, offset2);
    // clear hysteresis
    setDeviceConfig(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 3, 0);
    // execute always, independent of output value
    setDeviceConfig(CfgClassSensorEvent, row * CfgFSensorEvent_TableSize + 4, 0);

    setSensorEventName(row, entry.name);

    // offset outside of sensor event table
    auto minimalDurationOffset = (row == 0) ? 0x30 : 0x32;
    setDeviceConfig(CfgClassSensorEvent, minimalDurationOffset, entry.minimalDuration);
  }

  void Device::setSensorEventEntry(const int _eventIndex, DeviceSensorEventSpec_t _entry) {
    DS_REQUIRE(_eventIndex <= 7, "row index exceeds table size");
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      setSensorEventTableEntryZws205(_eventIndex, _entry);
      return;
    } else if (getRevisionID() < 0x0328) {
      /* older devices have a bug, where the hysteresis setting leads to strange behavior */
      _entry.hysteresis = 0;
    }
    uint8_t value;
    value = (_entry.sensorIndex << 4) | (_entry.test << 2) | (_entry.action);
    setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 0, value);
    value = (_entry.value & 0xFF0) >> 4;
    setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 1, value);
    value = ((_entry.value & 0x00F) << 4) | ((_entry.hysteresis & 0xF00) >> 8);
    setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 2, value);
    value = (_entry.hysteresis & 0x0FF);
    setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 3, value);
    value = ((_entry.sceneDeviceMode & 0x03) << 2) | (_entry.validity & 0x03);
    setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 4, value);
    if (_entry.action == 2) {
      value = (_entry.buttonNumber << 4) | (_entry.clickType);
      setDeviceConfig(CfgClassSensorEvent, CfgFSensorEvent_TableSize * _eventIndex + 5, value);
    }
    setSensorEventName(_eventIndex, _entry.name);
  }

  bool Device::hasExtendendSceneTable() const {
    if ((((getFunctionID() & 0xf000) >> 12) == GroupIDGray) &&
        (((getProductID() >> 10) & 0x3f) == DEVICE_TYPE_KL)) {
      return true;
    }
    return false;
  }

  bool Device::evaluateGRKL387Workaround(const int _scene, const int _value, const int _angle) {
    int realValue = _value;
    if (getDeviceClass() == DEVICE_CLASS_GR && getRevisionID() == 0x387) {
      constexpr std::array<int, 16> possibleValues = {
          0x0000, 0x1111, 0x2222, 0x3333, 0x4444, 0x5555, 0x6666, 0x7777,
          0x8888, 0x9999, 0xAAAA, 0xBBBB, 0xCCCC, 0xDDDD, 0xEEEE, 0xFFFF,
      };
      auto lower = std::lower_bound(possibleValues.begin(), possibleValues.end(), _value);
      if (lower != possibleValues.begin() && _value != *lower) {
        lower--;
      }
      auto upper = std::upper_bound(possibleValues.begin(), possibleValues.end(), _value);
      if (upper == possibleValues.end()) {
        upper--;
      }
      if (std::abs(_value - *lower) < std::abs(_value - *upper)) {
        realValue = *lower;
      } else {
        realValue = *upper;
      }
      DS_DEBUG(*lower, _value, *upper, realValue);
      //Save current position and angle
      int pos = getDeviceOutputValue(2);
      int angle = getDeviceOutputValue(4);
      //Move to desired position and save scene on device
      setDeviceOutputValue(2, (realValue));
      setDeviceOutputValue(4, _angle);
      saveScene(coSystem, _scene, "");
      //Restore posion,angle
      setDeviceOutputValue(2, (pos));
      setDeviceOutputValue(4, angle & 0xff);
      int extValue = getDeviceConfig(CfgClassSceneExtention, _scene / 2);
      if ((_scene % 2) == 0) {
        extValue &= 0xF0;
        extValue |= (realValue >> 4) & 0x0F;
      } else {
        extValue &= 0x0F;
        extValue |= realValue & 0xF0;
      }
      setDeviceConfig(CfgClassSceneExtention, _scene / 2, extValue);
      return true;
    }
    return false;
  }


  void Device::setSceneValue(const int _scene, const int _value) const {
    int realValue = _value;
    if (hasExtendendSceneTable()) {
      int extValue = getDeviceConfig(CfgClassSceneExtention, _scene / 2);
      if ((_scene % 2) == 0) {
        extValue &= 0xF0;
        extValue |= (realValue >> 4) & 0x0F;
      } else {
        extValue &= 0x0F;
        extValue |= realValue & 0xF0;
      }
      setDeviceConfig(CfgClassScene, _scene, (realValue >> 8) & 0xFF);
      setDeviceConfig(CfgClassSceneExtention, _scene / 2, extValue);
    } else {
      setDeviceConfig(CfgClassScene, _scene, realValue);
    }
  }

  int Device::getSceneValue(const int _scene) const {
    int value = getDeviceConfig(CfgClassScene, _scene);
    if (hasExtendendSceneTable()) {
      value <<= 8;
      int extValue = getDeviceConfig(CfgClassSceneExtention, _scene / 2);
      if ((_scene % 2) == 1) {
        extValue &= 0xF0; // high bits for odd scene numbers
        extValue |= (extValue >> 4);
      } else {
        extValue &= 0x0F; // lower bits for even scene numbers
        extValue |= (extValue << 4);
      }
      value |= extValue;
    }
    return value;
  }

  void Device::setSupportedBasicScenes(const std::vector<BasicScene>& x) {
    if (m_supportedBasicScenes == x) {
      return;
    }
    DS_NOTICE("Set supported basic scenes.", x);
    m_supportedBasicScenes = x;
    dirty();
  }

  void Device::configureAreaMembership(const int _areaScene, const bool _addToArea) {
    if (!SceneHelper::isAreaScene(_areaScene)) {
      throw DSSException("Device::configureAreaMembership: scene number does not indicate an area");
    }
    int areaOnScene = (_areaScene >= SceneA11) ? _areaScene : (_areaScene - SceneOffA1 + SceneA11);
    uint8_t outValue = 0;

    if (_addToArea) {
      outValue = 0xFF;
    }
    setDeviceValue(outValue);
    saveScene(coSystem, areaOnScene, "");
  }

  void Device::setButtonInputMode(ButtonInputMode mode) {
    bool wasSlave = is2WaySlave();

    m_ButtonInputMode = mode;
    m_AKMInputProperty = getAKMButtonInputString(mode);
    if (is2WaySlave() && !wasSlave) {
      removeFromPropertyTree();
    } else if (wasSlave && !is2WaySlave()) {
      publishToPropertyTree();
      updateAKMNode();
    }

    if (hasPushbutton() &&
        ((getRevisionID() >= 0x0314) || (isVdcDevice()))) {
      // button switched to sensor mode
      if (isBinaryInput()) {
        if (getButtonUsage() != ButtonUsage::unsupported) {
          setButtonUsage(ButtonUsage::unsupported);
        }
      } else {
        if (getButtonUsage() == ButtonUsage::unsupported) {
          setButtonUsage(ButtonUsage::auto_unused);
        }
      }
    }
  }

  void Device::setOutputMode(const uint8_t _value) {
    bool wasSlave = is2WaySlave();

    m_OutputMode = _value;
    if ((getDeviceType() == DEVICE_TYPE_UMR) && (multiDeviceIndex() == 3)) {
      if (is2WaySlave()) {
        removeFromPropertyTree();
      } else if (wasSlave) {
        publishToPropertyTree();
        updateAKMNode();
      }
    }
  }

  std::string Device::getAKMInputProperty() const {
    return m_AKMInputProperty;
  }

  bool Device::is2WayMaster() const {

    if (!getDeviceFeatures().pairing) {
      return false;
    }

    bool ret = false;
    try {
      if ((multiDeviceIndex() == 2) && (getDeviceType() == DEVICE_TYPE_UMR)) {
        ret = ((m_OutputMode == OUTPUT_MODE_TWO_STAGE_SWITCH) ||
               (m_OutputMode == OUTPUT_MODE_BIPOLAR_SWITCH)   ||
               (m_OutputMode == OUTPUT_MODE_THREE_STAGE_SWITCH) ||
               (m_OutputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_2STEPS) ||
               (m_OutputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_3STEPS) ||
               (m_OutputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_PARALLEL)) &&
               IsEvenDsuid(m_DSID); // even dSID
      } else if (hasInput()) {
        ret = ((m_ButtonInputMode == ButtonInputMode::TWO_WAY_DW_WITH_INPUT2) ||
               (m_ButtonInputMode == ButtonInputMode::TWO_WAY_DW_WITH_INPUT4) ||
               (m_ButtonInputMode == ButtonInputMode::TWO_WAY_UP_WITH_INPUT2) ||
               (m_ButtonInputMode == ButtonInputMode::TWO_WAY_UP_WITH_INPUT4) ||
               ((!isVdcDevice() && ((m_ButtonInputMode == ButtonInputMode::TWO_WAY) ||
                 (m_ButtonInputMode == ButtonInputMode::ONE_WAY))) &&
                IsEvenDsuid(m_DSID))); // even dSID
      }
    } catch (std::runtime_error &err) {
      log(err.what(), lsError);
    }
    return ret;
  }

  bool Device::isBinaryInput() const {
    switch (m_ButtonInputMode) {
      case ButtonInputMode::AKM_STANDARD:
      case ButtonInputMode::AKM_INVERTED:
      case ButtonInputMode::AKM_ON_RISING_EDGE:
      case ButtonInputMode::AKM_ON_FALLING_EDGE:
      case ButtonInputMode::AKM_OFF_RISING_EDGE:
      case ButtonInputMode::AKM_OFF_FALLING_EDGE:
      case ButtonInputMode::AKM_RISING_EDGE:
      case ButtonInputMode::AKM_FALLING_EDGE:
        return true;
      case ButtonInputMode::STANDARD:
      case ButtonInputMode::TURBO:
      case ButtonInputMode::SWITCHED:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY:
      case ButtonInputMode::ONE_WAY:
      case ButtonInputMode::HEATING_PUSHBUTTON:
      case ButtonInputMode::DEACTIVATED:
        break;
    }
    return false;
  }

  bool Device::is2WaySlave() const {

    if (!(multiDeviceIndex() == 3) && (getDeviceType() == DEVICE_TYPE_UMR)) {
      if (!hasInput()) {
        return false;
      }
    }

    bool ret = false;
    try {
      if ((getDeviceType() == DEVICE_TYPE_UMR) && (multiDeviceIndex() == 3)) {
        // if our device is set to "deactivated" for the paired modes, check
        // our parent partner
        if (m_OutputMode == OUTPUT_MODE_DEACTIVATED) {
          dsuid_t previous;
          dsuid_get_previous_dsuid(getDSID(), &previous);
          boost::shared_ptr<Device> dev = getApartment().tryGetDeviceByDSID(previous);
          if (dev) {
            uint8_t outputMode = dev->getOutputMode();
            ret = ((outputMode == OUTPUT_MODE_TWO_STAGE_SWITCH) ||
                   (outputMode == OUTPUT_MODE_BIPOLAR_SWITCH)   ||
                   (outputMode == OUTPUT_MODE_THREE_STAGE_SWITCH) ||
                   (outputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_2STEPS) ||
                   (outputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_3STEPS) ||
                   (outputMode == OUTPUT_MODE_TEMPCONTROL_2OUT_PARALLEL));
          } else {
            // the partner device *has* to be there, if its not there is no
            // recovery, if it happens we assume we are not a slave and show
            // ourselves
            DS_ERROR("Partner device for ", getDSID(), " not found.");
          }
        }
      } else if (hasInput()) {
        // Only devices of type SDS, TKM, UMR and vDCs can be paired.
        if ((getDeviceType() == DEVICE_TYPE_SDS) ||
            (getDeviceType() == DEVICE_TYPE_TKM) ||
            (getDeviceType() == DEVICE_TYPE_UMR) ||
            (isVdcDevice() && (getPairedDevices() > 0))) {
          ret = ((m_ButtonInputMode == ButtonInputMode::TWO_WAY_DW_WITH_INPUT1) ||
                 (m_ButtonInputMode == ButtonInputMode::TWO_WAY_DW_WITH_INPUT3) ||
                 (m_ButtonInputMode == ButtonInputMode::TWO_WAY_UP_WITH_INPUT1) ||
                 (m_ButtonInputMode == ButtonInputMode::TWO_WAY_UP_WITH_INPUT3) ||
                 ((m_ButtonInputMode == ButtonInputMode::DEACTIVATED) &&
                 !IsEvenDsuid(m_DSID))); // odd dSID
        }
      }
    } catch (std::runtime_error &err) {
      log(err.what(), lsError);
    }

    return ret;
  }

  bool Device::hasMultibuttons() const {
    int subclass = (m_FunctionID & 0xfc0) >> 6;
    int funcmodule = (m_FunctionID & 0x3f);

    if ((subclass == 0x04) && ((funcmodule & 0x3) > 1)) {
      return true;
    }
    if ((subclass == 0x07) && ((funcmodule & 0x3) > 1)) {
      return true;
    }
    if ((subclass == 0x08) && ((funcmodule & 0x3) > 1)) {
      return true;
    }
    if ((subclass == 0x09) && ((funcmodule & 0x3) > 1)) {
      return true;
    }
    return false;
  }

  bool Device::hasInput() const {
    int subclass = (m_FunctionID & 0xfc0) >> 6;
    int funcmodule = (m_FunctionID & 0x3f);

    if ((subclass == 0x04) && ((funcmodule & 0x3) > 0)) {
      return true;
    }
    if ((subclass == 0x07) && ((funcmodule & 0x3) > 0)) {
      return true;
    }
    if ((subclass == 0x08) && ((funcmodule & 0x3) > 0)) {
      return true;
    }
    if ((subclass == 0x09) && ((funcmodule & 0x3) > 0)) {
      return true;
    }
    return false;
  }

  bool Device::hasOutput() const {
    int funcmodule = (m_FunctionID & 0x3f);
    return (((funcmodule & 0x10) == 0x10) || (m_FunctionID == 0x6001));
  }

  bool Device::isInternallyControlled() const {
    int subclass = (m_FunctionID & 0xfc0) >> 6;
    return (subclass == 0x08) || (subclass == 0x09);
  }

  DeviceTypes_t Device::getDeviceType() const {
    if (m_ProductID == 0) {
      return DEVICE_TYPE_INVALID;
    }
    int t = (m_ProductID >> 10) & 0x3f;
    switch (t) {
      case 0:
        return DEVICE_TYPE_KM;
      case 1:
        return DEVICE_TYPE_TKM;
      case 2:
        return DEVICE_TYPE_SDM;
      case 3:
        return DEVICE_TYPE_KL;
      case 4:
        return DEVICE_TYPE_TUP;
      case 5:
        return DEVICE_TYPE_ZWS;
      case 6:
        return DEVICE_TYPE_SDS;
      case 7:
        return DEVICE_TYPE_SK;
      case 8:
        return DEVICE_TYPE_AKM;
      case 10:
        return DEVICE_TYPE_TNY;
      case 11:
        return DEVICE_TYPE_UMV;
      case 12:
        return DEVICE_TYPE_UMR;
      default:
        return DEVICE_TYPE_INVALID;
    }
  }

  const std::string Device::getDeviceTypeString(const DeviceTypes_t _type) {
    switch (_type) {
      case DEVICE_TYPE_KM:
        return "KM";
      case DEVICE_TYPE_TKM:
        return "TKM";
      case DEVICE_TYPE_SDM:
        return "SDM";
      case DEVICE_TYPE_KL:
        return "KL";
      case DEVICE_TYPE_TUP:
        return "TUP";
      case DEVICE_TYPE_ZWS:
        return "ZWS";
      case DEVICE_TYPE_SDS:
        return "SDS";
      case DEVICE_TYPE_SK:
        return "SK";
      case DEVICE_TYPE_AKM:
        return "AKM";
      case DEVICE_TYPE_TNY:
        return "TNY";
      case DEVICE_TYPE_UMV:
        return "UMV";
      case DEVICE_TYPE_UMR:
        return "UMR";
      case DEVICE_TYPE_INVALID:
      default:
        return "";
    }
  }

  int Device::getDeviceNumber() const {
    return m_ProductID & 0x3ff;
  }

  DeviceClasses_t Device::getDeviceClass() const {
    if (m_FunctionID == 0) {
      return DEVICE_CLASS_INVALID;
    }

    int c = (m_FunctionID & 0xf000) >> 12;
    switch (c) {
      case 1:
        return DEVICE_CLASS_GE;
      case 2:
        return DEVICE_CLASS_GR;
      case 3:
        return DEVICE_CLASS_BL;
      case 4:
        return DEVICE_CLASS_TK;
      case 5:
        return DEVICE_CLASS_MG;
      case 6:
        return DEVICE_CLASS_RT;
      case 7:
        return DEVICE_CLASS_GN;
      case 8:
        return DEVICE_CLASS_SW;
      case 9:
        return DEVICE_CLASS_WE;
      default:
        return DEVICE_CLASS_INVALID;
    }
  }

  void Device::calculateHWInfo()
  {
    if (!m_HWInfo.empty()) {
      m_HWInfo.clear();
    }

    int classId = (m_FunctionID & 0xf000) >> 12; // getDeviceClass() would work just DEVICE_CLASS_INVALID==-1
    const char* displayName = NULL;
    const char* hwInfo = NULL;
    const char* devGTIN = NULL;

    DsmApiGetDeviceDescription(m_VendorID, m_ProductID, classId, m_RevisionID, &displayName, &hwInfo, &devGTIN);

    // HWInfo - Priorities: 1. OEM Data, 2. Device Product Data, 3. Device EEPROM Data (Vendor independent)
    if ((m_OemProductInfoState == DEVICE_OEM_VALID) && !m_OemProductName.empty()) {
      m_HWInfo = m_OemProductName;
    } else if (!m_vdsdSpec.model.empty()) {
      m_HWInfo = m_vdsdSpec.model;
    } else if (displayName != NULL) {
      m_HWInfo = displayName;
    } else {
      DeviceClasses_t deviceClass = getDeviceClass();
      m_HWInfo += getDeviceClassString(deviceClass);
      m_HWInfo += "-";

      DeviceTypes_t deviceType = getDeviceType();
      m_HWInfo += getDeviceTypeString(deviceType);

      int deviceNumber = getDeviceNumber();
      m_HWInfo += intToString(deviceNumber);
    }

    // if GTIN is known ...
    if (devGTIN) {
      m_GTIN = devGTIN;
    }

    m_hardwareType = [&]() -> HardwareType {
      // Generated form productlist_dsd.cvs colums: {{Color ID, Produkt ID}, Produkt Name}
      static const boost::container::flat_map<std::pair<int, int>, HardwareType> hardwareTypes = {
          {{1, 0x00C8}, HardwareType::GE_KM200},    //
          {{1, 0x04D2}, HardwareType::GE_TKM210},   //
          {{1, 0x08C8}, HardwareType::GE_SDM200},   //
          {{1, 0x08C9}, HardwareType::GE_SDM200JS}, //
          {{1, 0x08CA}, HardwareType::GE_SDM200CS}, //
          {{1, 0x18C8}, HardwareType::GE_SDS200CW}, //
          {{1, 0x18C9}, HardwareType::GE_SDS200CS}, //
          {{1, 0x18DC}, HardwareType::GE_SDS220CT}, //
          {{1, 0x04DC}, HardwareType::GE_TKM220},   //
          {{1, 0x04E6}, HardwareType::GE_TKM230},   //
          {{1, 0x0CC8}, HardwareType::GE_KL200},    //
          {{1, 0x2CC8}, HardwareType::GE_UMV200},   //
          {{1, 0x2CD2}, HardwareType::GE_UMV210},   //
          {{1, 0x012C}, HardwareType::GE_KM300},    //
          {{1, 0x052C}, HardwareType::GE_TKM300},   //
          {{7, 0x00C8}, HardwareType::GN_KM200},    //
          {{7, 0x04C8}, HardwareType::GN_TKM200},   //
          {{7, 0x04D2}, HardwareType::GN_TKM210},   //
          {{2, 0x04C8}, HardwareType::GR_TKM200},   //
          {{2, 0x04D2}, HardwareType::GR_TKM210},   //
          {{2, 0x0CC8}, HardwareType::GR_KL200},    //
          {{2, 0x0CD2}, HardwareType::GR_KL210},    //
          {{2, 0x0CDC}, HardwareType::GR_KL220},    //
          {{2, 0x0CE6}, HardwareType::GR_HKL230},   //
          {{3, 0x00C8}, HardwareType::BL_KM200},    //
          {{3, 0x18C8}, HardwareType::BL_SDS200},   //
          {{6, 0x04C8}, HardwareType::RT_TKM200},   //
          {{6, 0x08C8}, HardwareType::RT_SDM200},   //
          {{8, 0x04C8}, HardwareType::SW_TKM200},   //
          {{8, 0x04D2}, HardwareType::SW_TKM210},   //
          {{8, 0x20C8}, HardwareType::SW_AKM200},   //
          {{8, 0x20D2}, HardwareType::SW_AKM210},   //
          {{8, 0x20DC}, HardwareType::SW_AKM220},   //
          {{8, 0x0CC8}, HardwareType::SW_KL200},    //
          {{8, 0x14C8}, HardwareType::SW_ZWS200},   //
          {{8, 0x14C9}, HardwareType::SW_ZWS200F},  //
          {{8, 0x14CA}, HardwareType::SW_ZWS200EF}, //
          {{8, 0x0CD5}, HardwareType::SW_SSL200JS}, //
          {{8, 0x0CD6}, HardwareType::SW_SSL200FS}, //
          {{8, 0x18C8}, HardwareType::SW_SDS200CW}, //
          {{8, 0x18C9}, HardwareType::SW_SDS200CS}, //
          {{8, 0x18DC}, HardwareType::SW_SDS220CT}, //
          {{8, 0x28C8}, HardwareType::SW_TNY200},   //
          {{8, 0x30C8}, HardwareType::SW_UMR200},   //
          {{8, 0x1CC8}, HardwareType::SW_SK200},    //
          {{8, 0x1CC9}, HardwareType::SW_SK201},    //
          {{8, 0x1CCA}, HardwareType::SW_SK202},    //
          {{3, 0x1CCC}, HardwareType::BL_SK204},    //
          {{8, 0x14CD}, HardwareType::SW_ZWS205},   //
      };

      auto it = hardwareTypes.find({classId, m_ProductID});
      if (it != hardwareTypes.end()) {
        return it->second;
      }
      return HardwareType::UNKNOWN;
    }();
  }

  void Device::updateModelFeatures() {
    decltype(m_modelFeatures) modelFeatures;
    auto&& subclass = (m_FunctionID >> 6) & 0x3F;
    if (subclass == 0x07 || subclass == 0x08 || subclass == 0x09) {
      modelFeatures[ModelFeatureId::apartmentapplication] = true;
    }
    if ((getDeviceType() == DEVICE_TYPE_UMR) && (getDeviceNumber() == 200) &&
        (getRevisionID() >= 0x0383) && (multiDeviceIndex() <= 1)) {
      modelFeatures[ModelFeatureId::setumr200config] = true;
    }
    // #20143: enable the consumption message for the UMR on output 1
    if ((getDeviceType() == DEVICE_TYPE_UMR) && (getDeviceNumber() == 200) &&
        (multiDeviceIndex() == 2)) {
      modelFeatures[ModelFeatureId::consumption] = true;
    }
    if (getDeviceType() == DEVICE_TYPE_KL && getRevisionID() >= 0x365) {
      modelFeatures[ModelFeatureId::operationlock] = true;
    }
    if ((getDeviceType() == DEVICE_TYPE_KL) && (getRevisionID() == 0x387) &&
        ((getDeviceNumber() == 200) || (getDeviceNumber() == 210) ||
         (getDeviceNumber() == 220) || (getDeviceNumber() == 230))) {
      // dsd-2590
      modelFeatures[ModelFeatureId::grkl387workaround] = true;
    }

    if (m_modelFeatures == modelFeatures) {
      return;
    }
    m_modelFeatures.swap(modelFeatures);
    republishModelFeaturesToPropertyTree();
  }

  void Device::republishModelFeaturesToPropertyTree() {
    if (!m_pPropertyNode) {
      return;
    }
    // remove old nodes
    {
      auto&& node = m_pPropertyNode->getProperty("modelFeatures");
      if (node) {
        m_pPropertyNode->removeChild(node);
      }
    }
    // add new nodes
    auto&& modelFeaturesNode = m_pPropertyNode->createProperty("modelFeatures");
    for (auto&& modelFeaturePair : m_modelFeatures) {
      const auto& id = modelFeaturePair.first;
      if (const auto& name = ds::tryFromEnum<const char*>(id)) {
        auto&& node = modelFeaturesNode->createProperty(*name);
        node->setValue(modelFeaturePair.second);
      } else {
        log(ds::str("Cannot publish model feature ", id), lsWarning);
      }
    }
  }

  void Device::updateIconPath() {
    if (!m_iconPath.empty()) {
      m_iconPath.clear();
    }

    if ((m_OemProductInfoState == DEVICE_OEM_VALID) && !m_OemProductIcon.empty()) {
      m_iconPath = m_OemProductIcon;
    } else if (!m_VdcIconPath.empty()) {
      m_iconPath = m_VdcIconPath;
    } else {
      DeviceClasses_t deviceClass = getDeviceClass();
      DeviceTypes_t deviceType = getDeviceType();
      if (deviceClass == DEVICE_CLASS_INVALID) {
        m_iconPath = "unknown.png";
        return;
      }
      if ((deviceType == DEVICE_TYPE_KL) && ((getDeviceNumber() == 213) || (getDeviceNumber() == 214))) {
        m_iconPath = "ssl";
      } else if ((deviceType == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
        m_iconPath = "zws205";
      } else if ((deviceType == DEVICE_TYPE_SDM) && ((getDeviceNumber() == 201) || (getDeviceNumber() == 202))) {
        m_iconPath = "sdm_plug";
      } else if ((deviceType == DEVICE_TYPE_KM) && (getDeviceNumber() == 300)) {
        m_iconPath = "km300";
      } else if ((deviceType == DEVICE_TYPE_TKM) && (getDeviceNumber() == 300)) {
        m_iconPath = "tkm300";
      } else if ((deviceType == DEVICE_TYPE_TNY) && (getDeviceNumber() == 206)) {
        m_iconPath = "tny206";
      } else {
        m_iconPath = getDeviceTypeString(deviceType);
        std::transform(m_iconPath.begin(), m_iconPath.end(), m_iconPath.begin(), ::tolower);
      }
      if (m_iconPath.empty()) {
        m_iconPath = "star";
      }

      m_iconPath += "_" + getColorString(deviceClass);

      if (getDeviceClass() == DEVICE_CLASS_SW) {
        // static_cast to ApplicationType works for all currently valid (zone, apartment) groupIds
        auto&& jokerApplicationType = static_cast<ApplicationType>(m_ActiveGroup);
        auto&& jokerColor = getApplicationTypeColor(jokerApplicationType);
        m_iconPath += "_" + getColorString(jokerColor);
      }

      m_iconPath += ".png";
    }
  }

  const std::string Device::getColorString(const int _class) {
    switch (_class) {
      case DEVICE_CLASS_GE:
        return "yellow";
      case DEVICE_CLASS_GR:
        return "grey";
      case DEVICE_CLASS_BL:
        return "blue";
      case DEVICE_CLASS_TK:
        return "cyan";
      case DEVICE_CLASS_MG:
        return "magenta";
      case DEVICE_CLASS_RT:
        return "red";
      case DEVICE_CLASS_GN:
        return "green";
      case DEVICE_CLASS_SW:
        return "black";
      case DEVICE_CLASS_WE:
        return "white";
      case DEVICE_CLASS_INVALID: //no break
      default:
        return "";
    }
  }

  const std::string Device::getDeviceClassString(const DeviceClasses_t _class) {
    switch (_class) {
      case DEVICE_CLASS_GE:
        return "GE";
      case DEVICE_CLASS_GR:
        return "GR";
      case DEVICE_CLASS_BL:
        return "BL";
      case DEVICE_CLASS_TK:
        return "TK";
      case DEVICE_CLASS_MG:
        return "MG";
      case DEVICE_CLASS_RT:
        return "RT";
      case DEVICE_CLASS_GN:
        return "GN";
      case DEVICE_CLASS_SW:
        return "SW";
      case DEVICE_CLASS_WE:
        return "WE";
      case DEVICE_CLASS_INVALID: //no break
      default:
        return "";
    }
  }

  const DeviceFeatures_t Device::getDeviceFeatures() const {
    DeviceFeatures_t features;
    features.pairing = false;
    features.syncButtonID = false;
    features.hasOutputAngle = false;
    features.posTimeMax = false;
    features.posTime = false;

    if (m_FunctionID == 0) {
        return features;
    }

    DeviceTypes_t devType = this->getDeviceType();
    int devNumber = this->getDeviceNumber();
    DeviceClasses_t devCls = this->getDeviceClass();

    if ((devCls == DEVICE_CLASS_SW) && (devType == DEVICE_TYPE_TKM) &&
       ((devNumber == 200) || (devNumber == 210)) &&
       this->hasMultibuttons() &&
       ((m_ButtonInputIndex == 0) || (m_ButtonInputIndex == 2))) {
      features.pairing = true;
      features.syncButtonID = true;
    } else if ((devCls == DEVICE_CLASS_GE) && // GE-SDS20x, GE-SDS22x
               (devType == DEVICE_TYPE_SDS) &&
               ((devNumber / 10 == 20) || (devNumber / 10 == 22)) &&
               this->hasMultibuttons() &&
               (m_ButtonInputIndex == 0)) {
      features.pairing = true;
    } else if ((devCls == DEVICE_CLASS_SW) && (devType == DEVICE_TYPE_UMR) &&
               (devNumber == 200)) {
      if (multiDeviceIndex() == 0) {
        features.pairing = true;
        features.syncButtonID = true;
      } else if (multiDeviceIndex() == 2) {
        features.pairing = true;
      }
    }

    if ((devCls == DEVICE_CLASS_GR) && (devType == DEVICE_TYPE_KL)) {
      features.posTime = true;

      if (getRevisionID() >= 0x360) {
        features.posTimeMax = true;
      }

      if ((devNumber == 220) || (devNumber == 230)) {
        features.hasOutputAngle = true;
      }
    }

    if (isVdcDevice()) {
      if ((hasMultibuttons() && ((m_ButtonInputIndex == 0) || (m_ButtonInputIndex == 2))) && (getPairedDevices() > 0)) {
        features.pairing = true;
        features.syncButtonID = true;
      }
    }
    return features;
  }

  void Device::setOemInfo(const unsigned long long _eanNumber,
          const uint16_t _serialNumber, const uint8_t _partNumber,
          const DeviceOEMInetState_t _iNetState,
          bool _isIndependent)
  {
    m_OemEanNumber = _eanNumber;
    m_OemSerialNumber = _serialNumber;
    m_OemPartNumber = _partNumber;
    m_OemInetState = _iNetState;
    m_OemIsIndependent = _isIndependent;

    dirty();
  }

  void Device::setOemInfoState(const DeviceOEMState_t _state)
  {
    m_OemState = _state;

    if ((m_OemState == DEVICE_OEM_NONE) || (m_OemState == DEVICE_OEM_VALID)) {
      dirty();
    }
  }

  std::string Device::oemStateToString(const DeviceOEMState_t _state)
  {
    switch (_state) {
    case DEVICE_OEM_UNKNOWN:
      return "Unknown";
    case DEVICE_OEM_NONE:
      return "None";
    case DEVICE_OEM_LOADING:
      return "Loading";
    case DEVICE_OEM_VALID:
      return "Valid";
    }
    return "";
  }

  DeviceOEMInetState_t Device::getOemInetStateFromString(const char* _string) const
  {
    if (strcmp(_string, "No EAN") == 0) {
      return DEVICE_OEM_EAN_NO_EAN_CONFIGURED;
    } else if (strcmp(_string, "No Internet") == 0) {
      return DEVICE_OEM_EAN_NO_INTERNET_ACCESS;
    } else if (strcmp(_string, "Internet optional") == 0) {
      return DEVICE_OEM_EAN_INTERNET_ACCESS_OPTIONAL;
    } else if (strcmp(_string, "Internet mandatory") == 0) {
      return DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY;
    } else {
      return DEVICE_OEM_EAN_NO_EAN_CONFIGURED;
    }
  }

  std::string Device::oemInetStateToString(const DeviceOEMInetState_t _state)
  {
    switch (_state) {
    case DEVICE_OEM_EAN_NO_EAN_CONFIGURED:
      return "No EAN";
    case DEVICE_OEM_EAN_NO_INTERNET_ACCESS:
      return "No Internet";
    case DEVICE_OEM_EAN_INTERNET_ACCESS_OPTIONAL:
      return "Internet optional";
    case DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY:
      return "Internet mandatory";
    }
    return "";
  }

  DeviceOEMState_t Device::getOemStateFromString(const char* _string) const
  {
    if (strcmp(_string, "None") == 0) {
      return DEVICE_OEM_NONE;
    } else if (strcmp(_string, "Valid") == 0) {
      return DEVICE_OEM_VALID;
    } else {
      return DEVICE_OEM_UNKNOWN;
    }
  }

  void Device::setOemProductInfo(const std::string& _productName, const std::string& _iconPath, const std::string& _productURL, const std::string& _configLink)
  {
    m_OemProductName = _productName;
    m_OemProductIcon = _iconPath;
    m_OemProductURL = _productURL;
    m_OemConfigLink = _configLink;
  }

  void Device::setOemProductInfoState(const DeviceOEMState_t _state)
  {
    m_OemProductInfoState = _state;

    calculateHWInfo();
    updateIconPath();

    if ((m_OemProductInfoState == DEVICE_OEM_NONE) || (m_OemProductInfoState == DEVICE_OEM_VALID)) {
      dirty();
    }
  }

  std::string Device::getValveTypeAsString() const
  {
    return valveTypeToString(m_ValveType);
  }

  bool Device::setDeviceValveTypeAsString(const std::string& _value)
  {
    if (isValveDevice() || isInternallyControlled()) {
      DeviceValveType_t type;
      if (getValveTypeFromString(_value.c_str(), type)) {
        setDeviceValveType(type);
        return true;
      }
    }
    return false;
  }

  std::string Device::valveTypeToString(const DeviceValveType_t _type) {
    switch (_type) {
    case DEVICE_VALVE_UNKNOWN:
      return "Unknown";
    case DEVICE_VALVE_FLOOR:
      return "Floor";
    case DEVICE_VALVE_RADIATOR:
      return "Radiator";
    case DEVICE_VALVE_WALL:
      return "Wall";
    case DEVICE_CONVECTOR_PASSIVE:
      return "ConvectorPassive";
    case DEVICE_CONVECTOR_ACTIVE:
      return "ConvectorActive";
    case DEVICE_VALVE_FLOOR_LOW_ENERGY:
      return "FloorLowEnergy";

    }
    return "";
  }

  bool Device::getValveTypeFromString(const char* _string, DeviceValveType_t &deviceType) {
    bool assigned = false;
    if (strcmp(_string, "Floor") == 0) {
      deviceType = DEVICE_VALVE_FLOOR;
      assigned = true;
    } else if (strcmp(_string, "Radiator") == 0) {
      deviceType = DEVICE_VALVE_RADIATOR;
      assigned = true;
    } else if (strcmp(_string, "Wall") == 0) {
      deviceType = DEVICE_VALVE_WALL;
      assigned = true;
    } else if (strcmp(_string, "Unknown") == 0) {
      deviceType = DEVICE_VALVE_UNKNOWN;
      assigned = true;
    } else if (strcmp(_string, "ConvectorPassive") == 0) {
      deviceType = DEVICE_CONVECTOR_PASSIVE;
      assigned = true;
    } else if (strcmp(_string, "ConvectorActive") == 0) {
      deviceType = DEVICE_CONVECTOR_ACTIVE;
      assigned = true;
    } else if (strcmp(_string, "FloorLowEnergy") == 0) {
      deviceType = DEVICE_VALVE_FLOOR_LOW_ENERGY;
      assigned = true;
    } else {
      log(std::string("Invalid valve type: ") + _string,
                                 lsWarning);
    }

    return assigned;
  }

  void Device::publishValveTypeToPropertyTree() {
    if ((m_pPropertyNode != NULL)) {
      if (isValveDevice()){
        PropertyNodePtr valveType = m_pPropertyNode->createProperty("ValveType");
        valveType->linkToProxy(PropertyProxyMemberFunction<Device, std::string, false>(*this, &Device::getValveTypeAsString));
      }
      PropertyNodePtr activeCoolingMode = m_pPropertyNode->createProperty("ActiveCoolingMode");
      activeCoolingMode->linkToProxy(PropertyProxyMemberFunction<Device, bool, false>(*this, &Device::getActiveCoolingMode));
    }
  }

  bool Device::isValveDevice() const {
    return (hasOutput() && ((!isInternallyControlled() && (getDeviceClass() == DEVICE_CLASS_BL)) ||
                            ((getDeviceType() == DEVICE_TYPE_UMR) && (getRevisionID() >= 0x0383)) ||
                            ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205))));
  }

  DeviceValveType_t Device::getValveType () const {
    return m_ValveType;
  }

  void Device::setValveType(DeviceValveType_t _type) {
    m_ValveType = _type;
    if (DSS::getInstance()) {
      boost::shared_ptr<Event> pEvent;
      pEvent.reset(new Event(EventName::DeviceHeatingTypeChanged));
      pEvent->setProperty("zoneID", intToString(getZoneID()));
      if (!DSS::getInstance()->getModelMaintenance().isInitializing()) {
        DSS::getInstance()->getEventQueue().pushEvent(pEvent);
      }
    }
    dirty();
  }

  void Device::setDeviceValveType(DeviceValveType_t _type) {
    if (m_isVdcDevice) {
      google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> query;
      vdcapi::PropertyElement* e1 = query.Add();
      e1->set_name("outputSettings");
      vdcapi::PropertyElement* e2 = e1->add_elements();
      e2->set_name("heatingSystemType");
      e2->mutable_value()->set_v_int64(_type);
      try {
          setVdcProperty(query);
      } catch(std::runtime_error& e) {
          log(std::string("Device::setValveType: ") + e.what(), lsWarning);
      }
    } else if ((getDeviceClass() == DEVICE_CLASS_BL) &&
          (getDeviceType() == DEVICE_TYPE_KL) &&
          (getDeviceNumber() == 200)) {
      setDeviceConfig(CfgClassFunction, 0x41, static_cast<uint8_t>(_type));
    } else if ((getDeviceClass() == DEVICE_CLASS_SW) &&
          (getDeviceType() == DEVICE_TYPE_UMR) &&
          (getDeviceNumber() == 200) &&
          (getRevisionID() >= 0x383)) {
      setDeviceConfig(CfgClassFunction, 0x43, static_cast<uint8_t>(_type));
    } else if ((getDeviceClass() == DEVICE_CLASS_SW) &&
          (getDeviceType() == DEVICE_TYPE_ZWS) &&
          ((getDeviceNumber() == 205) || (getDeviceNumber() == 215))) {
      setDeviceConfig(CfgClassFunction, 0x41, static_cast<uint8_t>(_type));
    } else {
      log(std::string("setDeviceValveType: not of valve type: ") + dsuid2str(getDSID()), lsWarning);
    }
    setValveType(_type);
  }

  DeviceValveType_t Device::getDeviceValveType() {
    DeviceValveType_t type = DEVICE_VALVE_UNKNOWN;
    if (m_isVdcDevice) {
      google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> query;
      vdcapi::PropertyElement* e1 = query.Add();
      e1->set_name("outputSettings");
      vdcapi::PropertyElement* e2 = e1->add_elements();
      e2->set_name("heatingSystemType");
      vdcapi::Message message = getVdcProperty(query);
      VdcElementReader reader(message.vdc_response_get_property().properties());
      type = static_cast<DeviceValveType_t> (reader["outputSettings"]["heatingSystemType"].getValueAsInt());
    } else if ((getDeviceClass() == DEVICE_CLASS_BL) &&
        (getDeviceType() == DEVICE_TYPE_KL) &&
        (getDeviceNumber() == 200)) {
      type = static_cast<DeviceValveType_t> (getDeviceConfig(CfgClassFunction, 0x41));
    } else if ((getDeviceClass() == DEVICE_CLASS_SW) &&
        (getDeviceType() == DEVICE_TYPE_UMR) &&
        (getDeviceNumber() == 200) &&
        (getRevisionID() >= 0x383)) {
      type = static_cast<DeviceValveType_t> (getDeviceConfig(CfgClassFunction, 0x43));
    } else if ((getDeviceClass() == DEVICE_CLASS_SW) &&
        (getDeviceType() == DEVICE_TYPE_ZWS) &&
        ((getDeviceNumber() == 205) || (getDeviceNumber() == 215))) {
      type = static_cast<DeviceValveType_t> (getDeviceConfig(CfgClassFunction, 0x41));
    }
    return type;
  }

  void Device::setDeviceBinaryInputType(uint8_t _inputIndex, BinaryInputType inputType) const {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_inputIndex >= m_binaryInputs.size()) {
      throw ItemNotFoundException("Invalid binary input index");
    }
    lock.unlock();
    setDeviceConfig(CfgClassDevice, 0x40 + 3 * _inputIndex + 1, static_cast<int>(inputType));
  }

  void Device::setDeviceBinaryInputTargetId(uint8_t _inputIndex, uint8_t _targetGroup)
  {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_inputIndex >= m_binaryInputs.size()) {
      throw ItemNotFoundException("Invalid binary input index");
    }
    setDeviceConfig(CfgClassDevice, 0x40 + 3 * _inputIndex + 0, _targetGroup);
    setBinaryStateTargetGroup (_targetGroup);
  }

  BinaryInputType Device::getDeviceBinaryInputType(uint8_t _inputIndex) const {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_inputIndex >= m_binaryInputs.size()) {
      throw ItemNotFoundException("Invalid binary input index");
    }
    return m_binaryInputs[_inputIndex]->m_inputType;
  }

  void Device::setDeviceBinaryInputId(uint8_t _inputIndex, BinaryInputId id) const {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_inputIndex >= m_binaryInputs.size()) {
      throw ItemNotFoundException("Invalid binary input index");
    }
    uint8_t val = (static_cast<int>(id) & 0xf) << 4;
    if (_inputIndex == m_binaryInputs.size() - 1) {
      val |= 0x80;
    }
    setDeviceConfig(CfgClassDevice, 0x40 + 3 * _inputIndex + 2, val);
  }

  void Device::setDeviceAKMInputTimeouts(int _onDelay, int _offDelay) const {
    if (_onDelay >= 0) {
      uint16_t onDelay = _onDelay / 100;
      setDeviceConfig(CfgClassFunction, CfgFunction_LTTimeoutOn, onDelay & 0xff);
      setDeviceConfig(CfgClassFunction, CfgFunction_LTTimeoutOn + 1, onDelay >> 8);
    }
    if (_offDelay >= 0) {
      uint16_t offDelay = _offDelay / 100;
      setDeviceConfig(CfgClassFunction, CfgFunction_LTTimeoutOff, offDelay & 0xff);
      setDeviceConfig(CfgClassFunction, CfgFunction_LTTimeoutOff + 1, offDelay >> 8);
    }
  }

  void Device::getDeviceAKMInputTimeouts(int& _onDelay, int& _offDelay) const {
    _onDelay = getDeviceConfigWord(CfgClassFunction, CfgFunction_LTTimeoutOn) * 100;
    _offDelay = getDeviceConfigWord(CfgClassFunction, CfgFunction_LTTimeoutOff) * 100;
  }

  std::string Device::getAKMButtonInputString(ButtonInputMode mode) const {
    switch (mode) {
      case ButtonInputMode::AKM_STANDARD:
        return BUTTONINPUT_AKM_STANDARD;
      case ButtonInputMode::AKM_INVERTED:
        return BUTTONINPUT_AKM_INVERTED;
      case ButtonInputMode::AKM_ON_RISING_EDGE:
        return BUTTONINPUT_AKM_ON_RISING_EDGE;
      case ButtonInputMode::AKM_ON_FALLING_EDGE:
        return BUTTONINPUT_AKM_ON_FALLING_EDGE;
      case ButtonInputMode::AKM_OFF_RISING_EDGE:
        return BUTTONINPUT_AKM_OFF_RISING_EDGE;
      case ButtonInputMode::AKM_OFF_FALLING_EDGE:
        return BUTTONINPUT_AKM_OFF_FALLING_EDGE;
      case ButtonInputMode::AKM_RISING_EDGE:
        return BUTTONINPUT_AKM_RISING_EDGE;
      case ButtonInputMode::AKM_FALLING_EDGE:
        return BUTTONINPUT_AKM_FALLING_EDGE;
      case ButtonInputMode::STANDARD:
      case ButtonInputMode::TURBO:
      case ButtonInputMode::SWITCHED:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_DW_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT1:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT2:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT3:
      case ButtonInputMode::TWO_WAY_UP_WITH_INPUT4:
      case ButtonInputMode::TWO_WAY:
      case ButtonInputMode::ONE_WAY:
      case ButtonInputMode::HEATING_PUSHBUTTON:
      case ButtonInputMode::DEACTIVATED:
      default:
        break;
    }
    return "";
  }

  uint8_t Device::getBinaryInputCount() const {
    return (uint8_t) m_binaryInputs.size();
  }

  void Device::setBinaryInputTargetId(uint8_t index, uint8_t targetGroupId) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    getBinaryInput(index)->setTargetId(targetGroupId);
  }

  void Device::setBinaryInputId(uint8_t index, BinaryInputId inputId) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    getBinaryInput(index)->setInputId(inputId);
  }

  void Device::setBinaryInputType(uint8_t index, BinaryInputType inputType) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    getBinaryInput(index)->setInputType(inputType);
  }

  void Device::clearBinaryInputs() {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    m_binaryInputs.clear();
  }

  void Device::initStates(const std::vector<DeviceStateSpec_t>& stateSpecs) {
    if (stateSpecs.empty()) {
      return;
    }
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    PropertyNodePtr node;
    if (m_pPropertyNode != NULL) {
      node = m_pPropertyNode->getPropertyByName("states");
      if (node != NULL) {
        node->getParentNode()->removeChild(node);
      }
      node = m_pPropertyNode->createProperty("states");
    }
    m_states.clear();

    for (const DeviceStateSpec_t& stateSpec : stateSpecs) {
      const std::string& stateName = stateSpec.Name;
      try {
        boost::shared_ptr<State> state = m_pApartment->allocateState(sharedFromThis(), stateName);
        std::vector<std::string> values;
        values.push_back(State::INVALID);
        values.insert(values.end(), stateSpec.Values.begin(), stateSpec.Values.end());
        state->setValueRange(values);
        m_states[stateName] = state;

        if (m_pPropertyNode != NULL) {
          PropertyNodePtr entry = node->createProperty(stateName);
          PropertyNodePtr stateValueNode = state->getPropertyNode()->getProperty("value");
          if (stateValueNode != NULL) {
            PropertyNodePtr stateValueAlias = entry->createProperty("stateValue");
            stateValueAlias->alias(stateValueNode);
          }
        }
      } catch (std::runtime_error& ex) {
          log("Device::initStates:" + dsuid2str(m_DSID)
              + " state:" + stateName + " what:" + ex.what(), lsError);
          throw;
      }
    }
  }

  void Device::clearStates() {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    for (const States::value_type& state : m_states) {
      try {
        m_pApartment->removeState(state.second);
      } catch (ItemNotFoundException& e) {
        log(std::string("Apartment::removeDevice: Unknown state: ") + e.what(), lsWarning);
      }
    }
    m_states.clear();
  }

  void Device::setStateValue(const std::string& name, const std::string& value) {
    log("Device::setStateValue name:" + name + " value:" + value, lsDebug);
    try {
      for (const States::value_type& state : m_states) {
        if (state.first == name) {
          state.second->setState(coDsmApi, value);
        }
      }
    } catch(std::runtime_error& e) {
      log("Device::setStateValue name:" + name
          + " value:" + value + " what:" + e.what(), lsWarning);
    }
  }

  void Device::setStateValues(const std::vector<std::pair<std::string, std::string> >& values) {
    for (const States::value_type& statePair : m_states) {
      const std::string* newValue = &State::INVALID;
      const std::string& stateName = statePair.first;
      try {
        typedef std::pair<std::string, std::string> StringStringPair;
        for (const StringStringPair& value : values) {
          if (stateName == value.first) {
            newValue = &(value.second);
            break;
          }
        }
        log("Device::setStateValues name:" + stateName + " value:" + *newValue, lsDebug);
        statePair.second->setState(coDsmApi, *newValue);
      } catch(std::runtime_error& e) {
        log("Device::setStateValues name:" + stateName
            + " value:" + *newValue + " what:" + e.what(), lsWarning);
      }
    }
  }

  void Device::handleBinaryInputEvent(const int index, BinaryInputStateValue inputState) {
    getBinaryInput(index)->handleEvent(inputState);
  }

  void Device::handleBinaryInputEvent(const std::string& inputId, BinaryInputStateValue inputState) {
    boost::shared_ptr<DeviceBinaryInput> binput = getBinaryInput(inputId);
    binput->handleEvent(inputState);
  }

  bool Device::supportsApartmentApplications() const {
    auto&& supportsApartmentApplicationsIt = m_modelFeatures.find(ModelFeatureId::apartmentapplication);
    return supportsApartmentApplicationsIt != m_modelFeatures.end() ? supportsApartmentApplicationsIt->second : false;
  }

  void Device::setBinaryInputs(const std::vector<DeviceBinaryInputSpec_t>& _binaryInputs) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    PropertyNodePtr binaryInputNode;
    if (m_pPropertyNode != NULL) {
      binaryInputNode = m_pPropertyNode->getPropertyByName("binaryInputs");
      if (binaryInputNode != NULL) {
        binaryInputNode->getParentNode()->removeChild(binaryInputNode);
      }
      binaryInputNode = m_pPropertyNode->createProperty("binaryInputs");
    }
    m_binaryInputs.clear();

    for (std::vector<DeviceBinaryInputSpec_t>::const_iterator it = _binaryInputs.begin();
        it != _binaryInputs.end();
        ++it) {
      auto binaryInputIndex = m_binaryInputs.size();
      m_binaryInputs.push_back(boost::make_shared<DeviceBinaryInput>(*this, *it, binaryInputIndex));
      auto& binaryInput = m_binaryInputs.back();

      if (m_pPropertyNode != NULL) {
        std::string bpath = std::string("binaryInput") + intToString(binaryInputIndex);
        PropertyNodePtr entry = binaryInputNode->createProperty(bpath);
        entry->createProperty("targetGroupId")
                ->linkToProxy(PropertyProxyReference<int>(binaryInput->m_targetGroupId));
        entry->createProperty("inputType")
                ->linkToProxy(PropertyProxyReference<int, BinaryInputType>(binaryInput->m_inputType));
        entry->createProperty("inputId")
                ->linkToProxy(PropertyProxyReference<int, BinaryInputId>(binaryInput->m_inputId));
        entry->createProperty("inputIndex")
                ->linkToProxy(PropertyProxyReference<int>(binaryInput->m_inputIndex));
        entry->createProperty("independent")
                ->linkToProxy(PropertyProxyReference<bool>(binaryInput->m_independent));
        if (auto stateNode = binaryInput->getState().getPropertyNode()) {
          if (auto stateValueNode = stateNode->getProperty("value")) {
            PropertyNodePtr stateValueAlias = entry->createProperty("stateValue");
            stateValueAlias->alias(stateValueNode);
          }
        }
      }
    }
  }

  const std::vector<boost::shared_ptr<DeviceBinaryInput> >& Device::getBinaryInputs() const {
    return m_binaryInputs;
  }

  const boost::shared_ptr<DeviceBinaryInput> Device::getBinaryInput(uint8_t _inputIndex) const {
    if (_inputIndex >= getBinaryInputCount()) {
      throw ItemNotFoundException(std::string("Device::getBinaryInput: index out of bounds"));
    }
    return m_binaryInputs[_inputIndex];
  }

  const boost::shared_ptr<DeviceBinaryInput> Device::getBinaryInput(const std::string& inputId) const {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    for (const auto& binaryInput : m_binaryInputs) {
      if (binaryInput->m_sInputId == inputId) {
        return binaryInput;
      }
    }
    throw ItemNotFoundException(std::string("Device::getBinaryInput: inputId not found"));
  }

  uint8_t Device::getSensorCount() const {
    // uint8_t seems hard limit of the underlying ds devices
    // TODO(someday) move this check closer to the lowest layer
    if (m_sensorInputs.size() > std::numeric_limits<uint8_t>::max()) {
      throw std::runtime_error("SensorInput count exceeds datatype");
    }
    return static_cast<uint8_t>(m_sensorInputs.size());
  }

  void Device::setSensors(const std::vector<DeviceSensorSpec_t>& _sensorInputs) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    PropertyNodePtr sensorInputNode;
    if (m_pPropertyNode != NULL) {
      sensorInputNode = m_pPropertyNode->getPropertyByName("sensorInputs");
      if (sensorInputNode != NULL) {
        sensorInputNode->getParentNode()->removeChild(sensorInputNode);
      }
      sensorInputNode = m_pPropertyNode->createProperty("sensorInputs");
    }
    m_sensorInputs.clear();

    // fill in "standard" sensors if sensor table has not been read from device or is empty
    std::vector<DeviceSensorSpec_t> _slist = _sensorInputs;
    if (_slist.empty()) {
      fillSensorTable(_slist);
    }

    int sensorInputCount = 0;
    for (std::vector<DeviceSensorSpec_t>::const_iterator it = _slist.begin();
        it != _slist.end();
        ++it) {
      m_sensorInputs.emplace_back(new DeviceSensor_t());
      m_sensorInputs.back()->m_sensorIndex = sensorInputCount;
      m_sensorInputs.back()->m_sensorType = it->sensorType;
      m_sensorInputs.back()->m_sensorUsage = 0;
      m_sensorInputs.back()->m_sensorPollInterval = it->SensorPollInterval;
      m_sensorInputs.back()->m_sensorBroadcastFlag = it->SensorBroadcastFlag;
      m_sensorInputs.back()->m_sensorPushConversionFlag = it->SensorConversionFlag;
      m_sensorInputs.back()->m_sensorValue = 0;
      m_sensorInputs.back()->m_sensorValueFloat = 0;
      m_sensorInputs.back()->m_sensorPrevValueFloat = 0;
      m_sensorInputs.back()->m_sensorValueTS = DateTime::NullDate;
      m_sensorInputs.back()->m_sensorValueValidity = false;

      if (m_pPropertyNode != NULL) {
        std::string bpath = std::string("sensorInput") + intToString(sensorInputCount);
        PropertyNodePtr entry = sensorInputNode->createProperty(bpath);
        entry->createProperty("type")->linkToProxy(
            PropertyProxyReference<int, SensorType>(m_sensorInputs[sensorInputCount]->m_sensorType));
        entry->createProperty("index")->linkToProxy(
            PropertyProxyReference<int>(m_sensorInputs[sensorInputCount]->m_sensorIndex));
        entry->createProperty("usage")->linkToProxy(
            PropertyProxyReference<int, uint8_t>(m_sensorInputs[sensorInputCount]->m_sensorUsage));
        entry->createProperty("name")->linkToProxy(
            PropertyProxyReference<std::string>(m_sensorInputs[sensorInputCount]->m_sensorName));
        entry->createProperty("valid")->linkToProxy(
            PropertyProxyReference<bool>(m_sensorInputs[sensorInputCount]->m_sensorValueValidity));
        entry->createProperty("value")->linkToProxy(
            PropertyProxyReference<double>(m_sensorInputs[sensorInputCount]->m_sensorValueFloat));
        entry->createProperty("valueDS")->linkToProxy(
            PropertyProxyReference<int, unsigned int>(m_sensorInputs[sensorInputCount]->m_sensorValue));
        entry->createProperty("timestamp")
            ->linkToProxy(PropertyProxyMemberFunction<DateTime, std::string, false>(
                m_sensorInputs[sensorInputCount]->m_sensorValueTS, &DateTime::toString));
        entry->createProperty("timeISO8601")
            ->linkToProxy(PropertyProxyMemberFunction<DateTime, std::string, false>(
                m_sensorInputs[sensorInputCount]->m_sensorValueTS, &DateTime::toISO8601_ms));
      }

      sensorInputCount++;
    }
  }

  void Device::setSensorsInfo(const uint8_t _sensorIndex, const std::string&  _sensorId,
      const std::string& _sensorName, const int _sensorUsage) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_sensorIndex >= m_sensorInputs.size()) {
      throw ItemNotFoundException(std::string("Device::setSensorsInfo: index out of bounds"));
    }
    auto& binput = m_sensorInputs[_sensorIndex];
    binput->m_sensorId = _sensorId;
    binput->m_sensorName = _sensorName;
    binput->m_sensorUsage = _sensorUsage;
  }

  void Device::setBinaryInputsInfo(const uint8_t _dsIndex, const std::string&  _inputId,
      const std::string& _inputName) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_dsIndex >= m_binaryInputs.size()) {
      throw ItemNotFoundException(std::string("Device::setBinaryInputsInfo: index out of bounds"));
    }
    auto& binput = m_binaryInputs[_dsIndex];
    binput->m_sInputId = _inputId;
    binput->m_sInputName = _inputName;
  }

  void Device::setOutputChannels(const std::vector<ChannelType>& typesIn) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    PropertyNodePtr outputChannelNode;
    if (m_pPropertyNode != NULL) {
      outputChannelNode = m_pPropertyNode->getPropertyByName("outputChannels");
      if (outputChannelNode != NULL) {
        outputChannelNode->getParentNode()->removeChild(outputChannelNode);
      }
      outputChannelNode = m_pPropertyNode->createProperty("outputChannels");
    }
    m_outputChannelCount = 0;
    m_outputChannels.clear();

    auto types = typesIn;
    if (types.empty()) {
      types = hardwareOutputChannels(hardwareType(), *this);
      if (!types.empty()) {
        m_emulatedHardwareOutputs = true;
      }
    }
    for (const auto& type : types) {
      auto output = boost::make_shared<Channel>();
      m_outputChannels.push_back(output);
      output->index = m_outputChannelCount;
      output->type = type;
      output->id = ds::str(type);
      output->technicalName = output->id;

      if (m_pPropertyNode != NULL) {
        std::string bpath = std::string("outputChannel") + intToString(m_outputChannelCount);
        PropertyNodePtr entry = outputChannelNode->createProperty(bpath);
        entry->createProperty("channelID")
            ->linkToProxy(PropertyProxyReference<int, ChannelType>(m_outputChannels[m_outputChannelCount]->type));
        entry->createProperty("name")->linkToProxy(
            PropertyProxyReference<std::string>(m_outputChannels[m_outputChannelCount]->technicalName));
        entry->createProperty("id")->linkToProxy(
            PropertyProxyReference<std::string>(m_outputChannels[m_outputChannelCount]->id));
        entry->createProperty("index")->linkToProxy(
            PropertyProxyReference<int>(m_outputChannels[m_outputChannelCount]->index));
      }
      m_outputChannelCount++;
    }
  }

  void Device::setOutputChannelInfo(const uint8_t _channelIndex, const Channel& _channelInfo) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (_channelIndex >= m_outputChannelCount) {
      throw ItemNotFoundException(std::string("Device::setOutputChannelInfo: index out of bounds"));
    }
    m_outputChannels[_channelIndex]->id = _channelInfo.id;
    m_outputChannels[_channelIndex]->technicalName = _channelInfo.technicalName;
    m_outputChannels[_channelIndex]->type = _channelInfo.type;
  }

  int Device::getOutputChannelIndex(ChannelType _channelType) const {
    for (int i = 0; i < m_outputChannelCount; i++) {
      if (m_outputChannels[i]->type == _channelType) {
        return m_outputChannels[i]->index;
      }
    }
    return -1;
  }

  boost::shared_ptr<Channel> Device::getOutputChannel(uint8_t index) const {
    if (index >= m_outputChannelCount) {
      throw ItemNotFoundException(ds::str("deviceOutputIndex:", index));
    }
    return m_outputChannels[index];
  }

  boost::shared_ptr<Channel> Device::getOutputChannelById(const std::string& id) const {
    for (int i = 0; i < m_outputChannelCount; i++) {
      if (m_outputChannels[i]->id == id) {
        return m_outputChannels[i];
      }
    }
    throw ItemNotFoundException(ds::str("deviceOutputId:", id));
  }

  const std::vector<boost::shared_ptr<Channel>>& Device::getOutputChannels() const {
    return m_outputChannels;
  }

  int Device::getOutputChannelCount() const {
    return m_outputChannelCount;
  }

  const std::vector<std::unique_ptr<DeviceSensor_t>>& Device::getSensors() const { return m_sensorInputs; }

  const DeviceSensor_t& Device::getSensor(uint8_t _sensorIndex) const {
    if (_sensorIndex >= getSensorCount()) {
      throw ItemNotFoundException(std::string("Device::getSensor: index out of bounds"));
    }
    return *m_sensorInputs[_sensorIndex];
  }

  const DeviceSensor_t& Device::getSensorByType(SensorType _sensorType) const {
    for (const auto& sensor : m_sensorInputs) {
      if (sensor->m_sensorType == _sensorType) {
        return *sensor;
      }
    }

    throw ItemNotFoundException(std::string("Device::getSensor: no sensor with given type found"));
  }

  void Device::setSensorValue(int _sensorIndex, unsigned int _sensorValue) {
    if (_sensorIndex >= getSensorCount()) {
      throw ItemNotFoundException(std::string("Device::setSensorValue: index out of bounds"));
    }
    DateTime now;
    m_sensorInputs[_sensorIndex]->m_sensorValue = _sensorValue;
    m_sensorInputs[_sensorIndex]->m_sensorPrevValueFloat = m_sensorInputs[_sensorIndex]->m_sensorValueFloat;
    m_sensorInputs[_sensorIndex]->m_sensorValueFloat =
        sensorValueToDouble(m_sensorInputs[_sensorIndex]->m_sensorType, _sensorValue);
    m_sensorInputs[_sensorIndex]->m_sensorValueTS = now;
    m_sensorInputs[_sensorIndex]->m_sensorValueValidity = true;
  }

  void Device::setSensorValue(int _sensorIndex, double _sensorValue, uint32_t _age) {
    if (_sensorIndex >= getSensorCount()) {
      throw ItemNotFoundException(std::string("Device::setSensorValue: index out of bounds"));
    }
    DateTime now;
    if (_age > 0) {
      now = now.addMilliSeconds((int) _age * (-1));
    }
    m_sensorInputs[_sensorIndex]->m_sensorPrevValueFloat = m_sensorInputs[_sensorIndex]->m_sensorValueFloat;
    m_sensorInputs[_sensorIndex]->m_sensorValueFloat = _sensorValue;
    m_sensorInputs[_sensorIndex]->m_sensorValue = 0;
    m_sensorInputs[_sensorIndex]->m_sensorValueTS = now;
    m_sensorInputs[_sensorIndex]->m_sensorValueValidity = true;
  }

  void Device::setSensorDataValidity(int _sensorIndex, bool _valid) {
    if (_sensorIndex >= getSensorCount()) {
      throw ItemNotFoundException(std::string("Device::setSensorValue: index out of bounds"));
    }
    m_sensorInputs[_sensorIndex]->m_sensorValueValidity = _valid;
  }

  bool Device::isSensorDataValid(int _sensorIndex) const {
    if (_sensorIndex >= getSensorCount()) {
      throw ItemNotFoundException(std::string("Device::setSensorValue: index out of bounds"));
    }

    if (!this->isPresent() || !this->isValid()) {
      return true;
    }

    if (m_sensorInputs[_sensorIndex]->m_sensorPollInterval == 0) {
      return true;
    }

    return m_sensorInputs[_sensorIndex]->m_sensorValueValidity;
  };

  bool Device::isOemCoupledWith(boost::shared_ptr<Device> _otherDev)
  {
    return ((m_OemState == DEVICE_OEM_VALID) &&
            !m_OemIsIndependent &&
            (m_OemSerialNumber > 0) &&
            isPresent() &&
            _otherDev->isPresent() &&
            !_otherDev->getOemIsIndependent() &&
            (_otherDev->getOemInfoState() == DEVICE_OEM_VALID) &&
            (_otherDev->getDSID() != m_DSID) &&
            (_otherDev->getDSMeterDSID() != m_DSMeterDSID) &&
            (_otherDev->getOemEan() == m_OemEanNumber) &&
            (_otherDev->getOemSerialNumber() == m_OemSerialNumber));
  }

  void Device::setConfigLock(bool _lockConfig) {
    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    m_IsConfigLocked = _lockConfig;
  }

  void Device::setVdcDevice(bool _isVdcDevice)
  {
    m_isVdcDevice = _isVdcDevice;
    publishVdcToPropertyTree();
  }

  std::string Device::getDisplayID() const
  {
    dsid_t dsid;
    std::string displayID;
    if (dsuid_to_dsid(m_DSID, &dsid)) {
      displayID = dsid2str(dsid);
      displayID = displayID.substr(displayID.size() - 8);
    } else if (!m_vdsdSpec.displayId.empty()) {
      displayID = m_vdsdSpec.displayId;
    } else if (!m_vdsdSpec.hardwareGuid.empty()) {
      const auto hardwareGuid = m_vdsdSpec.hardwareGuid;
      displayID = hardwareGuid.substr(hardwareGuid.find(":") + 1);
    } else {
      displayID = dsuid2str(m_DSID).substr(0, 8) + "\u2026";
    }
    if (m_DSID.id[16] != 0) {
      displayID += "-" + intToString(m_DSID.id[16]);
    }
    return displayID;
  }

  uint8_t Device::getDeviceUMVRelayValue()
  {
    if ((getDeviceType() != DEVICE_TYPE_UMV) || (getDeviceNumber() != 200)) {
      throw std::runtime_error("unsupported configuration for this device");
    }
    return getDeviceConfig(CfgClassFunction, CfgFunction_UMV_Relay_Config);
  }

  void Device::setDeviceUMVRelayValue(uint8_t _value)
  {
    if ((getDeviceType() != DEVICE_TYPE_UMV) || (getDeviceNumber() != 200)) {
      throw std::runtime_error("unsupported configuration for this device");
    }

    setDeviceConfig(CfgClassFunction, CfgFunction_UMV_Relay_Config, _value);
  }

  bool Device::hasBlinkSettings() const {
    if ((getDeviceType() == DEVICE_TYPE_UMR) ||
        (getDeviceClass() == DEVICE_CLASS_GE) ||
        (((getDeviceType() == DEVICE_TYPE_KL) ||
          (getDeviceType() == DEVICE_TYPE_ZWS)) &&
         (getDeviceClass() == DEVICE_CLASS_SW))) {
            return true;
    }

    return false;
  }

  void Device::setDeviceBlinkRepetitions(uint8_t count) const {
    DS_REQUIRE(hasBlinkSettings(), "blink configuration not supported by this device");
    setDeviceConfig(CfgClassFunction, CfgFunction_FCount1, count);
  }

  void Device::setDeviceBlinkOnDelay(double delay) const {
    DS_REQUIRE(hasBlinkSettings(), "blink configuration not supported by this device");

    auto lowerBound = [&] {
      if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
        return 0.5;
      }
      return 0.0; // TODO(someday) probably wrong but the historically limit
    }();

    // upper limit with 8bit registers
    // 255 * BLINK_DELAY_SCALE / 1000.0
    auto upperBound = 8.4;

    DS_REQUIRE(delay >= lowerBound, "delay value too small", delay);
    DS_REQUIRE(delay <= upperBound, "delay value too big", delay);

    auto value = round(1000.0 * delay / BLINK_DELAY_SCALE);
    DS_REQUIRE(value >= 0 && value <= UCHAR_MAX, "delay exceeds data type", delay);
    setDeviceConfig(CfgClassFunction, CfgFunction_FOnTime1, static_cast<uint8_t>(value));
  }

  void Device::setDeviceBlinkOffDelay(double delay) const {
    DS_REQUIRE(hasBlinkSettings(), "blink configuration not supported by this device");

    auto lowerBound = [&] {
      if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
        return 0.5;
      }
      return 0.0; // TODO(someday) probably wrong but the historically limit
    }();

    // upper limit with 8bit registers
    // 255 * BLINK_DELAY_SCALE / 1000.0
    auto upperBound = 8.4;

    DS_REQUIRE(delay >= lowerBound, "delay value too small", delay);
    DS_REQUIRE(delay <= upperBound, "delay value too big", delay);

    auto value = round(1000.0 * delay / BLINK_DELAY_SCALE);
    DS_REQUIRE(value >= 0 && value <= UCHAR_MAX, "delay exceeds data type", delay);
    setDeviceConfig(CfgClassFunction, CfgFunction_FOffTime1, static_cast<uint8_t>(value));
  }

  void Device::getDeviceBlinkSettings(double *ondelay, double *offdelay, uint8_t  *count) const {
    DS_REQUIRE(hasBlinkSettings(), "blink configuration not supported by this device");

    uint16_t value = getDeviceConfigWord(CfgClassFunction, CfgFunction_FOnTime1);
    *ondelay = (double)((value & 0xff) * BLINK_DELAY_SCALE) / 1000.0;
    *count = (uint8_t)(value >> 8) & 0xff;

    uint8_t value2 = getDeviceConfig(CfgClassFunction, CfgFunction_FOffTime1);
    *offdelay = (value2 * BLINK_DELAY_SCALE) / 1000.0; // convert to seconds
  }

  std::vector<int> Device::getLockedScenes() const {
    std::vector<int> ls;

    for (auto&& gid : m_groupIds) {
      if (isAppUserGroup(gid)) {
        boost::shared_ptr<Cluster> cluster = m_pApartment->getCluster(gid);
        if (cluster->isConfigurationLocked()) {
            const std::vector<int>& locks = cluster->getLockedScenes();
            ls.insert(ls.end(), locks.begin(), locks.end());
        }
      }
    }

    return ls;
  }

  bool Device::isInLockedCluster() const {
    for (auto&& gid : m_groupIds) {
      if (isAppUserGroup(gid)) {
        boost::shared_ptr<Cluster> cluster = m_pApartment->getCluster(gid);
        if (cluster->isConfigurationLocked()) {
          return true;
        }
      }
    }

    return false;
  }

  int Device::multiDeviceIndex() const {
    if (isVdcDevice()) {
      return (m_DSID.id[DSUID_SIZE-1]);
    }

    uint8_t deviceCount = 1;
    uint8_t deviceIndex = 0;
    if ((getDeviceType() == DEVICE_TYPE_UMV) && (getDeviceNumber() == 200) &&
        (getDeviceClass() == DEVICE_CLASS_GE)) {
      deviceCount = 4;
    } else if ((getDeviceType() == DEVICE_TYPE_UMR) &&
               (getDeviceNumber() == 200) &&
               (getDeviceClass() == DEVICE_CLASS_SW)) {
      deviceCount = 4;
    } else if ((getDeviceType() == DEVICE_TYPE_SK) &&
               (getDeviceNumber() == 204)) {
      deviceCount = 2;
    } else if ((m_FunctionID & 0xffc0) == 0x1000) {
      // Legacy R100 stuff, subclass 0
      switch (m_FunctionID & 0x7) {
        case 0: deviceCount = 1; break;
        case 1: deviceCount = 2; break;
        case 2: deviceCount = 4; break;
        case 7: deviceCount = 1; break;
      }
    } else if (((m_FunctionID & 0x0fc0) == 0x0100) || ((m_FunctionID & 0x0fc0) == 0x01c0) || ((m_FunctionID & 0x0fc0) == 0x0200) || ((m_FunctionID & 0x0fc0) == 0x0240)) {
      // subclasses 4, 7, 8 or 9
      switch (m_FunctionID & 0x3) {
        case 0: deviceCount = 1; break;
        case 1: deviceCount = 1; break;
        case 2: deviceCount = 2; break;
        case 3: deviceCount = 4; break;
      }
    }

    uint32_t serial;
    if (dsuid_get_serial_number(&m_DSID, &serial) == 0) {
      deviceIndex = (uint8_t)(serial & 0xff) % deviceCount;
    }
    return deviceIndex;
  }

  void Device::setCardinalDirection(CardinalDirection_t _direction, bool _initial) {
    DS_ASSERT(valid(_direction), "");

    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);
    if (m_cardinalDirection != _direction) {
      m_cardinalDirection = _direction;
      if (!_initial) {
        checkAutoCluster();
      }
    }
  }

  void Device::setWindProtectionClass(WindProtectionClass_t _klass, bool _initial) {
    DS_ASSERT(valid(_klass), "");

    boost::recursive_mutex::scoped_lock lock(m_deviceMutex);

    if (!_initial) {
      if (getDeviceClass() != DEVICE_CLASS_GR) {
        throw std::runtime_error("Wind protection class can only be set on grey devices");
      }
      if ((getProductID() == ProductID_KL_210) && !(_klass <= wpc_awning_class_3)) {
        throw std::runtime_error("Wind protection class " + intToString(_klass) + "not applicable for awning devices");
      }
      if ((getProductID() != ProductID_KL_210) && !((_klass == wpc_none) || (_klass >= wpc_blind_class_1))) {
        throw std::runtime_error("Wind protection class " + intToString(_klass) + "not applicable for blinds devices");
      }
    }
    if (m_windProtectionClass != _klass) {
      m_windProtectionClass = _klass;
      if (!_initial) {
        checkAutoCluster();
      }
    }
  }

  uint16_t Device::getDeviceMaxMotionTime() const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTimeMax) {
      throw std::runtime_error("Maximum motion time setting not supported"
                               "by this device");
    }

    uint16_t hw = getDeviceConfigWord(CfgClassFunction, CfgFunction_Shade_PosTimeMax);
    // 100 halfwaves = 1 second
    return (hw / 100);
  }

  void Device::setDeviceMaxMotionTime(uint16_t seconds) const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTimeMax) {
      throw std::runtime_error("Maximum motion time setting not supported"
                               "by this device");
    }

    if ((seconds * 100) > USHRT_MAX) {
      throw std::runtime_error("Seconds value unsupported: too large");
    }

    setDeviceConfig16(CfgClassFunction, CfgFunction_Shade_PosTimeMax, seconds * 100);
  }

  double Device::getDevicePosTimeUp() const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTime) {
      throw std::runtime_error("Up and down position timings are not suported "
                               "by this device");
    }

    // 100 halfwaves = 1 second
    return (getDeviceConfigWord(CfgClassFunction, CfgFunction_Shade_PosTimeUp) * 10 / 1000.0);
  }

  void Device::setDevicePosTimeUp(double seconds) const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTime) {
      throw std::runtime_error("Up and down position timings are not suported "
                               "by this device");
    }

    seconds = floor(seconds * 1000 / 10);
    if ((seconds < 0) || (seconds > UINT16_MAX)) {
      throw std::runtime_error("Seconds value unsupported: negative or too large");
    }

    setDeviceConfig16(CfgClassFunction, CfgFunction_Shade_PosTimeUp,
                      (uint16_t)seconds);
  }

  double Device::getDevicePosTimeDown() const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTime) {
      throw std::runtime_error("Up and down position timings are not suported "
                               "by this device");
    }

    // 100 halfwaves = 1 second
    return (getDeviceConfigWord(CfgClassFunction, CfgFunction_Shade_PosTimeDown) * 10 / 1000.0);
  }

  void Device::setDevicePosTimeDown(double seconds) const {
    DeviceFeatures_t features = getDeviceFeatures();
    if (!features.posTime) {
      throw std::runtime_error("Up and down position timings are not suported "
                               "by this device");
    }

    seconds = floor(seconds * 1000 / 10);
    if ((seconds < 0) || (seconds > UINT16_MAX)) {
      throw std::runtime_error("Seconds value unsupported: negative or too large");
    }

    setDeviceConfig16(CfgClassFunction, CfgFunction_Shade_PosTimeDown,
                      (uint16_t)seconds);
  }

  uint8_t Device::getMaxPosTime() const {
    return m_maxPosTimeSeconds;
  }

  void Device::setMaxPosTime(uint8_t seconds) {
    bool changes = (seconds != m_maxPosTimeSeconds);
    m_maxPosTimeSeconds = seconds;
    if (changes) {
      dirty();
    }
  }

  uint8_t Device::getSWThresholdAddress() const {
    switch (getDeviceType()) {
      case DEVICE_TYPE_KM:
      case DEVICE_TYPE_SDM:
      case DEVICE_TYPE_SDS:
      case DEVICE_TYPE_TKM:
        return CfgFunction_KM_SWThreshold;
      case DEVICE_TYPE_KL:
      case DEVICE_TYPE_ZWS:
        switch (getDeviceClass()) {
          case DEVICE_CLASS_GE:
          case DEVICE_CLASS_SW:
            return CfgFunction_KL_SWThreshold;
          case DEVICE_CLASS_GR:
          case DEVICE_CLASS_BL:
          case DEVICE_CLASS_TK:
          case DEVICE_CLASS_MG:
          case DEVICE_CLASS_RT:
          case DEVICE_CLASS_GN:
          case DEVICE_CLASS_WE:
          case DEVICE_CLASS_INVALID:
          default:
            break;
        }
        break;
      case DEVICE_TYPE_UMV:
        if (getDeviceClass() == DEVICE_CLASS_GE) {
          return CfgFunction_UMV_SWThreshold;
        }
        break;
      case DEVICE_TYPE_TNY:
        if (getDeviceClass() == DEVICE_CLASS_SW && isMainDevice()) {
          return CfgFunction_Tiny_SWThreshold;
        }
        break;
      case DEVICE_TYPE_UMR:
        if (getDeviceClass() == DEVICE_CLASS_SW) {
          return CfgFunction_UMR_SWThreshold;
        }
        break;
      case DEVICE_TYPE_INVALID:
      case DEVICE_TYPE_TUP:
      case DEVICE_TYPE_SK:
      case DEVICE_TYPE_AKM:
      default:
        // TODO(someday) we should limit this to a DEVICE_TYPE
        if (getDeviceClass() == DEVICE_CLASS_BL) {
          return CfgFunction_Valve_SWThreshold;
        }
    }
    DS_FAIL_REQUIRE("Device does not support changing the switching threshold");
  }

  void Device::setSwitchThreshold(uint8_t _threshold) const {
    if (_threshold == 0 || _threshold == 255) {
      throw std::runtime_error("Threshold must not be 0 or 255");
    }

    setDeviceConfig(CfgClassFunction, getSWThresholdAddress(), _threshold);
  }

  uint8_t Device::getSwitchThreshold() const {
    return getDeviceConfig(CfgClassFunction, getSWThresholdAddress());
  }

  void Device::setTemperatureOffset(int8_t _offset) const {
    if (getDeviceType() != DEVICE_TYPE_SK) {
      throw std::runtime_error("Temperature offset setting is not supported by this device");
    }
    setDeviceConfig(CfgClassFunction, CfgFunction_SK_TempOffsetExt, _offset);
  }

  int8_t Device::getTemperatureOffset() const {
    if (getDeviceType() != DEVICE_TYPE_SK) {
      throw std::runtime_error("Temperature offset setting is not supported by this device");
    }
    return getDeviceConfig(CfgClassFunction, CfgFunction_SK_TempOffsetExt);
  }

  void Device::setPairedDevices(int _num) {
    bool update = (m_pairedDevices != _num);

    m_pairedDevices = _num;

    if (isMainDevice() && update && (m_pPropertyNode != NULL)) {
      PropertyNodePtr paired = m_pPropertyNode->getPropertyByName("pairedDevices");
      if (paired != NULL) {
        m_pPropertyNode->removeChild(paired);
      }

      if (_num > 0) {
        dsuid_t current = getDSID();
        for (int i = 0; i < _num - 1; i++) {
          PropertyNodePtr sub = m_pPropertyNode->
              createProperty("pairedDevices/device" + intToString(i));
          dsuid_t next;
          dsuid_get_next_dsuid(current, &next);
          sub->createProperty("dSUID")->setStringValue(dsuid2str(next));
          current = next;
        }
      }
    }
  }

  int Device::getPairedDevices() const {
    return m_pairedDevices;
  }

  void Device::setVisibility(bool _isVisible) {
    bool wasVisible = m_visible;
    m_visible = _isVisible;
    if (wasVisible && !m_visible) {
      removeFromPropertyTree();
    } else if (!wasVisible && m_visible) {
      publishToPropertyTree();
    }
  }

  void Device::setDeviceVisibility(bool _isVisible) {
    if ((getDeviceType() == DEVICE_TYPE_TNY) ||
        ((getDeviceType() == DEVICE_TYPE_SK) && (getDeviceNumber() == 204))) {
      if (isMainDevice()) {
        throw std::runtime_error("Visibility setting not allowed on main device");
      }

      uint8_t val = getDeviceConfig(CfgClassDevice, CfgFunction_DeviceActive);
      uint8_t bit4 = ((val & (1 << 4)) == 0); // inverted logic, 0 is "active"
      if (bit4 != _isVisible) {
        val ^= 1 << 4;
        setDeviceConfig(CfgClassDevice, CfgFunction_DeviceActive, val);
        setVisibility(_isVisible);
      }
    } else {
      throw std::runtime_error("Visibility setting is not supported by this device");
    }
  }

  bool Device::isVisible() const {
    if ((getDeviceType() == DEVICE_TYPE_TNY) ||
        ((getDeviceType() == DEVICE_TYPE_SK) && (getDeviceNumber() == 204))) {
      return m_visible;
    }

    return true;
  }

  bool Device::isMainDevice() const {
    if (m_pairedDevices == 0) {
      return true;
    }

    if (isVdcDevice()) {
      return (m_DSID.id[DSUID_SIZE - 1] == 0);
    } else {
      uint32_t serial;
      if (dsuid_get_serial_number(&m_DSID, &serial) == 0) {
        if ((serial % m_pairedDevices) != 0) {
          return false;
        }
      }
    }
    return true;
  }

  dsuid_t Device::getMainDeviceDSUID() const {
    dsuid_t main_dsuid;
    if (!isMainDevice()) {
      if (isVdcDevice()) {
        main_dsuid = getDSID();
        // submodule zero is the main device
        main_dsuid.id[DSUID_SIZE - 1] = 0;
        return main_dsuid;
      } else {
        if (dsuid_get_main_dsuid(&m_DSID, m_pairedDevices, &main_dsuid) == 0) {
          return main_dsuid;
        }
      }
    }

    return getDSID();
  }

  void Device::callAction(callOrigin_t origin, SceneAccessCategory category, const std::string& actionId,
      const vdcapi::PropertyElement& params, const std::locale& locale) {
    DS_CONTEXT(actionId);
    if (!m_isVdcDevice) {
      const auto& actions = this->actions();
      if (const auto& actionFound = boost::find_if<boost::return_found_end>(
              actions, [&](const Action<Device>& x) { return x.id == actionId; })) {
        const auto& action = actionFound.front();
        action.invoke(*this, origin, category);
      } else {
        DS_FAIL_REQUIRE("Failed to call unknown action.");
      }

    } else { // m_isVdcDevice
      DeviceBusInterface* deviceBusInterface = m_pApartment->getDeviceBusInterface();
      if (!deviceBusInterface) {
        throw std::runtime_error("Bus interface not available");
      }

      google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> actionCall;
      vdcapi::PropertyElement* param0 = actionCall.Add();
      param0->set_name("id");
      param0->mutable_value()->set_v_string(actionId);

      if (params.elements_size()) {
        vdcapi::PropertyElement* param1 = actionCall.Add();
        param1->set_name("params");
        *param1->mutable_elements() = params.elements();
      }
      deviceBusInterface->genericRequest(*this, "invokeDeviceAction", actionCall, locale);
    }

    //action finished with success -> raise event
    m_pApartment->pushEvent(createDeviceActionEvent(makeSharedReference(), actionId, params));
  }

  void Device::setProperty(const vdcapi::PropertyElement& propertyElement) {
    if (!m_isVdcDevice) {
      throw std::runtime_error("CallAction can be called only on vdc devices.");
    }
    DeviceBusInterface* deviceBusInterface = m_pApartment->getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> setPropertyParams;
    vdcapi::PropertyElement* devicePropertiesElement = setPropertyParams.Add();
    devicePropertiesElement->set_name("deviceProperties");
    *devicePropertiesElement->add_elements() = propertyElement;
    deviceBusInterface->setProperty(*this, setPropertyParams);
  }

  void Device::deviceChanged() {
    ModelEvent* pEvent = new ModelEventWithDSID(ModelEvent::etDeviceChanged, m_DSMeterDSID);
    pEvent->addParameter(m_ShortAddress);
    m_pApartment->getModelMaintenance()->addModelEvent(pEvent);
  }

  void Device::setCustomAction(const std::string& id, const std::string& title, const std::string& action,
                               const vdcapi::PropertyElement& actionParams) {
    if (!m_isVdcDevice) {
      throw std::runtime_error("CallAction can be called only on vdc devices.");
    }
    DeviceBusInterface* deviceBusInterface = m_pApartment->getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    if (!beginsWith(id, "custom.")) {
      throw std::runtime_error("Custom action id must start with 'custom.'");
    }

    std::string devAction(action);
    if (beginsWith(action, "std.") && !getHasLegacyActions()) {
      // old web ui added prefix as workaround, strip off for updated vdc's
      devAction = action.substr(4);
    }

    google::protobuf::RepeatedPtrField<vdcapi::PropertyElement> setPropertyParams;
    vdcapi::PropertyElement* customActionsElement = setPropertyParams.Add();
    customActionsElement->set_name("customActions");
    vdcapi::PropertyElement* actionElement = customActionsElement->add_elements();
    actionElement->set_name(id);
    if (!action.empty()) { // empty action means remove
      {
        vdcapi::PropertyElement* element = actionElement->add_elements();
        element->set_name("title");
        element->mutable_value()->set_v_string(title);
      }
      {
        vdcapi::PropertyElement* element = actionElement->add_elements();
        element->set_name("action");
        element->mutable_value()->set_v_string(devAction);
      }
      {
        vdcapi::PropertyElement* element = actionElement->add_elements();
        element->set_name("params");
        *element->mutable_elements() = actionParams.elements();
      }
    }
    deviceBusInterface->setProperty(*this, setPropertyParams);

    deviceChanged();
    m_pApartment->pushEvent(
        createDeviceCustomActionChangedEvent(makeSharedReference(), id, devAction, title, actionParams));
    dirty();
  }

  vdcapi::Message Device::getVdcProperty(
      const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& query) {
    if (!m_isVdcDevice) {
      throw std::runtime_error("getVdcProperty can be called only on vdc devices.");
    }
    DeviceBusInterface* deviceBusInterface = m_pApartment->getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    return deviceBusInterface->getProperty(*this, query);
  }

  void Device::setVdcProperty(
      const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& query) {
    if (!m_isVdcDevice) {
      throw std::runtime_error("getVdcProperty can be called only on vdc devices.");
    }
    DeviceBusInterface* deviceBusInterface = m_pApartment->getDeviceBusInterface();
    if (!deviceBusInterface) {
      throw std::runtime_error("Bus interface not available");
    }
    deviceBusInterface->setProperty(*this, query);
  }

  void Device::setVdsdSpec(VdsdSpec_t &&x) {
    m_vdsdSpec = std::move(x);

    // trigged by new x.hardwareInfo
    calculateHWInfo();
  }

  std::string Device::vdcTrDomain() const { return ds::str("vdc-common,vdc-gtin-", m_OemEanNumber); }

  void Device::updateZws205GroupColor() {
    auto bits = 040 | static_cast<uint8_t>(getApplicationTypeRgbBitmask(static_cast<ApplicationType>(m_ActiveGroup)));
    setDeviceConfig(CfgClassFunction, CfgFunction_LedConfig0, bits);
  }

  void Device::updateLedGroupColor() {
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      updateZws205GroupColor();
    }
  }

  void Device::updateOutputMode() {
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      updateZws205OutputMode();
    }
  }

  void Device::updateZws205OutputMode() {
    auto mode = [&] {
      switch (static_cast<ApplicationType>(m_ActiveGroup)) {
        case ApplicationType::Lights:
        case ApplicationType::Blinds:
        case ApplicationType::Awnings:
        case ApplicationType::Heating:
        case ApplicationType::Ventilation:
        case ApplicationType::Recirculation:
        case ApplicationType::ApartmentVentilation:
        case ApplicationType::ApartmentRecirculation:
        case ApplicationType::Audio:
        case ApplicationType::Video:
        case ApplicationType::Joker: return OUTPUT_MODE_SIMPLE_SWITCH;
        case ApplicationType::ControlTemperature:
          return OUTPUT_MODE_TEMPCONTROL_SWITCHED;
        case ApplicationType::Cooling:
        case ApplicationType::Window:
        case ApplicationType::None:
        case ApplicationType::Security:
        case ApplicationType::Access:
          break;
      }
      DS_ERROR("ZWS205: no output mode for application type", m_ActiveGroup);
      return OUTPUT_MODE_SIMPLE_SWITCH; // go with the 90%
    }();
    setDeviceOutputMode(mode);
  }

  std::ostream& operator<<(std::ostream& stream, const Device& x) {
    return stream << x.getName() << "(" << x.getDSID() << ")";
  }

  void Device::setConsumptionVisualizationEnabledZws205(bool enabled) const {
    uint8_t bits = enabled ? 0x2 : 0xff;
    setDeviceConfig(CfgClassFunction, CfgFunction_ZWS_LedVisuSensId, bits);
  }

  bool Device::getConsumptionVisualizationEnabledZws205() const {
    auto bits = getDeviceConfig(CfgClassFunction, CfgFunction_ZWS_LedVisuSensId);
    switch (bits) {
      case 0x2:
        return true;
      case 0xff:
        return false;
    }
    DS_ERROR("unknown led visualisation value", bits);
    return false;
  }

  void Device::setConsumptionVisualizationEnabled(bool enabled) {
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      return setConsumptionVisualizationEnabledZws205(enabled);
    }
    DS_FAIL_REQUIRE("Consumption visualization not support by device");
  }

  bool Device::getConsumptionVisualizationEnabled() {
    if ((getDeviceType() == DEVICE_TYPE_ZWS) && (getDeviceNumber() == 205)) {
      return getConsumptionVisualizationEnabledZws205();
    }
    DS_FAIL_REQUIRE("Consumption visualization not support by device");
  }

  bool Device::isBLKM300() const {
    return ((getDeviceClass() == DEVICE_CLASS_BL) &&
            (getDeviceType() == DEVICE_TYPE_KM) &&
            (getDeviceNumber() == 300));
  }

  std::vector<Action<Device>> Device::actions() const {
    auto callSceneMin = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneMin, "");
    };
    auto callSceneMax = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneMax, "");
    };
    auto callSceneImpulse = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneImpulse, "");
    };
    auto callSceneAutoOff = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneAutoOff, "");
    };
    auto callSceneBell = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneBell, "");
    };
    auto callSceneLocalOff = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneLocalOff, "");
    };
    auto callSceneLocalOn = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneLocalOn, "");
    };
    auto callSceneStop = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.callScene(origin, category, SceneStop, "");
    };
    auto setValue33Percent = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.setValue(origin, category, 33 * 255 / 100, "");
    };
    auto setValue50Percent = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.setValue(origin, category, 50 * 255 / 100, "");
    };
    auto setValue66Percent = [](Device& device, callOrigin_t origin, SceneAccessCategory category) {
      device.setValue(origin, category, 66 * 255 / 100, "");
    };

    static Action<Device> turnOff{"std.turnOff", dsApiTranslation("dss-mainline", "deviceAction", "Turn off"),
                                  callSceneMin, ApplicationType::None};
    static Action<Device> turnOn{"std.turnOn", dsApiTranslation("dss-mainline", "deviceAction", "Turn on"),
                                 callSceneMax, ApplicationType::None};
    static Action<Device> slowOff{"std.turnOffSlowly",
                                  dsApiTranslation("dss-mainline", "deviceAction", "Turn off slowly"), callSceneAutoOff,
                                  ApplicationType::None};
    static Action<Device> ring{"std.ring", dsApiTranslation("dss-mainline", "deviceAction", "Ring"), callSceneBell,
                               ApplicationType::None};
    static Action<Device> impulse{"std.impulse", dsApiTranslation("dss-mainline", "deviceAction", "Impulse"),
                                  callSceneImpulse, ApplicationType::None};
    static Action<Device> close{"std.close", dsApiTranslation("dss-mainline", "deviceAction", "Close"), callSceneMin,
                                ApplicationType::None};
    static Action<Device> open{"std.open", dsApiTranslation("dss-mainline", "deviceAction", "Open"), callSceneMax,
                               ApplicationType::None};
    static Action<Device> fold{"std.fold", dsApiTranslation("dss-mainline", "deviceAction", "Fold"), callSceneMin,
                               ApplicationType::None};
    static Action<Device> extend{"std.extends", dsApiTranslation("dss-mainline", "deviceAction", "Extend"),
                                 callSceneMax, ApplicationType::None};
    static Action<Device> stop{"std.stop", dsApiTranslation("dss-mainline", "deviceAction", "Stop"), callSceneStop,
                               ApplicationType::None};
    static Action<Device> forceOpen{"std.forceOpen", dsApiTranslation("dss-mainline", "deviceAction", "Force open"),
                                    callSceneLocalOn, ApplicationType::None};
    static Action<Device> forceClose{"std.forceClose", dsApiTranslation("dss-mainline", "deviceAction", "Force close"),
                                     callSceneLocalOff, ApplicationType::None};
    static Action<Device> resetForce{"std.resetForce", dsApiTranslation("dss-mainline", "deviceAction", "Reset force"),
                                     callSceneStop, ApplicationType::None};
    static auto step1Message = dsApiTranslation("dss-mainline", "deviceAction", "Step 1");
    static auto step2Message = dsApiTranslation("dss-mainline", "deviceAction", "Step 2");
    static auto step3Message = dsApiTranslation("dss-mainline", "deviceAction", "Step 3");
    static auto step4Message = dsApiTranslation("dss-mainline", "deviceAction", "Step 4");

    switch (hardwareType()) {
      case HardwareType::GE_KM200:
      case HardwareType::GE_TKM210:
      case HardwareType::GE_KM300:
      case HardwareType::GE_TKM300:
      case HardwareType::GE_SDM200:
      case HardwareType::GE_SDM200JS:
      case HardwareType::GE_SDM200CS:
      case HardwareType::GE_SDS200CW:
      case HardwareType::GE_SDS200CS:
      case HardwareType::GE_SDS220CT:
      case HardwareType::GE_KL200:
      case HardwareType::GE_UMV200:
      case HardwareType::GN_TKM200:
        return {turnOff, turnOn, slowOff};

      case HardwareType::GN_KM200:
      case HardwareType::GN_TKM210:
        return {ring};

      case HardwareType::GR_KL200:
      case HardwareType::GR_KL220:
      case HardwareType::GR_HKL230:
        return {close, open, stop};

      case HardwareType::GR_KL210:
        return {fold, extend, stop};

      case HardwareType::RT_TKM200:
      case HardwareType::RT_SDM200:
      case HardwareType::SW_KL200:
      case HardwareType::SW_ZWS200:
      case HardwareType::SW_ZWS200F:
      case HardwareType::SW_ZWS200EF:
      case HardwareType::SW_SSL200JS:
      case HardwareType::SW_SSL200FS:
      case HardwareType::SW_ZWS205:
        return {turnOff, turnOn};

      case HardwareType::SW_UMR200: {
        switch (multiDeviceIndex()) {
          case 2: // output 0
          case 3: // output 1
            switch (getOutputMode()) {
              case OUTPUT_MODE_DEACTIVATED:
                return {};
              case OUTPUT_MODE_SIMPLE_SWITCH:
              case OUTPUT_MODE_BIPOLAR_SWITCH:
              case OUTPUT_MODE_TEMPCONTROL_2OUT_PARALLEL:
                return {
                    //
                    turnOff,                                                          //
                    turnOn,                                                           //
                    impulse,                                                          //
                    {"std.step1", step1Message, callSceneMin, ApplicationType::None}, //
                    {"std.step2", step2Message, callSceneMax, ApplicationType::None}, //
                };
              case OUTPUT_MODE_TWO_STAGE_SWITCH:
              case OUTPUT_MODE_TEMPCONTROL_2OUT_2STEPS:
                return {
                    //
                    turnOff,                                                               //
                    turnOn,                                                                //
                    {"std.step1", step1Message, callSceneMin, ApplicationType::None},      //
                    {"std.step2", step2Message, setValue50Percent, ApplicationType::None}, //
                    {"std.step3", step3Message, callSceneMax, ApplicationType::None},      //
                };
              case OUTPUT_MODE_THREE_STAGE_SWITCH:
              case OUTPUT_MODE_TEMPCONTROL_2OUT_3STEPS:
                return {
                    //
                    turnOff,                                                               //
                    turnOn,                                                                //
                    {"std.step1", step1Message, callSceneMin, ApplicationType::None},      //
                    {"std.step2", step2Message, setValue33Percent, ApplicationType::None}, //
                    {"std.step3", step3Message, setValue66Percent, ApplicationType::None}, //
                    {"std.step4", step4Message, callSceneMax, ApplicationType::None},      //
                };
            };
            break;
        }
        break;
      }

      case HardwareType::BL_KM200:;
        if (m_ActiveGroup == GroupIDControlTemperature) {
          return {forceOpen, forceClose, resetForce};
        }
        return {open, close};

      // Input only devices without actions
      case HardwareType::GE_TKM220:
      case HardwareType::GE_TKM230:
      case HardwareType::GE_UMV210:
      case HardwareType::GR_TKM200:
      case HardwareType::GR_TKM210:
      case HardwareType::SW_TKM200:
      case HardwareType::SW_TKM210:
      case HardwareType::SW_AKM200:
      case HardwareType::SW_AKM210:
      case HardwareType::SW_AKM220:
      case HardwareType::SW_SK200:
      case HardwareType::SW_SK201:
      case HardwareType::SW_SK202:
      case HardwareType::SW_SDS200CW:
      case HardwareType::SW_SDS200CS:
      case HardwareType::SW_SDS220CT:
      case HardwareType::BL_SDS200:
      case HardwareType::BL_SK204:
        break;

      // Ignore for now
      case HardwareType::SW_TNY200:
        break;

      case HardwareType::UNKNOWN:
        break;
    }
    return {};
  }

  void BulkConfig::run() {
    size_t deviceCount = 0;
    bool lastEventSent = false;
    for (auto it = m_configVector.begin(); it != m_configVector.end(); ++it) {
      auto entry = *it;
      try {
        dsuid_t dsuid = str2dsuid(entry.dsuid);
        auto device = m_pApartment->getDeviceByDSID(dsuid);
        device->setDeviceConfig(entry.bank, entry.offset, entry.value);
        auto event = boost::make_shared<Event>(EventName::DeviceBulkSetConfig, DeviceReference(device, m_pApartment));
        event->setProperty("bank", entry.bank);
        event->setProperty("offset", entry.offset);
        event->setProperty("value", entry.value);
        event->setProperty("count", ++deviceCount);
        event->setProperty("total", m_configVector.size());
        event->setProperty("taskId", m_id);
        m_pApartment->pushEvent(event);
        if (deviceCount != m_configVector.size()) {
          sleep(1);
        } else {
          lastEventSent = true;
        }
      } catch (std::exception& e) {
        ++deviceCount;
        DS_WARNING("skip bulk write", entry.dsuid, entry.bank, entry.offset, e);
      }
    }
    if (!lastEventSent) {
      auto event = boost::make_shared<Event>(EventName::DeviceBulkSetConfig);
      event->setProperty("count", deviceCount);
      event->setProperty("total", m_configVector.size());
      event->setProperty("taskId", m_id);
      m_pApartment->pushEvent(event);
    }
  }

  void Device::reregister() const {
    if (m_pApartment->getDeviceBusInterface() == NULL) {
      throw std::runtime_error("DeviceBusInterface missing");
    }

    m_pApartment->getDeviceBusInterface()->reregister(*this);
  } // setDeviceConfig

  void Device::setButtonUsage(ButtonUsage usage) {
    m_buttonUsage = usage;
    dirty();
  }

  ButtonUsage Device::getButtonUsage() const {
    return m_buttonUsage;
  }

  std::string Device::getButtonUsageAsString() const {
    return ds::str(m_buttonUsage);
  }

  bool Device::hasPushbutton() const {
    if (getButtonInputCount() == 0) {
      return false;
    }
    if ((getButtonInputCount() > 0) &&
        (((getDeviceClass() == DEVICE_CLASS_BL) &&
          (getDeviceType() == DEVICE_TYPE_KM) &&
          (getDeviceNumber() == 200)) ||
          (getDeviceType() == DEVICE_TYPE_AKM) ||
          (getButtonInputMode() == ButtonInputMode::DEACTIVATED))) {
      return false;
    }

    return true;
  }

  std::vector<dsModel::ScenarioId> Device::scenarioIds() const {
    std::vector<dsModel::ScenarioId> result{};
    for (auto const& deviceAction : data().scenarios) {
      result.push_back(deviceAction.id);
    }
    return result;
  }

  const DeviceData& Device::data() const {
    return *m_data;
  }

  void Device::setData(DeviceData&& data) {
    if (*m_data == data) {
      return;
    }
    *m_data = std::move(data);
    dirty();
  }

  const std::vector<dsModel::Scenario>& Device::scenarios() const {
    return data().scenarios;
  }

  void Device::setScenarios(std::vector<dsModel::Scenario>&& scenarios) {
    if (m_data->scenarios == scenarios) {
      return;
    }
    m_data->scenarios = std::move(scenarios);
    dirty();
  }

  const std::vector<dsModel::DeviceStateStatus>& Device::statesStatus() const {
    return data().statesStatus;
  }

  void Device::setStatesStatus(std::vector<dsModel::DeviceStateStatus>&& statesStatus) {
    if (m_data->statesStatus == statesStatus) {
      return;
    }
    m_data->statesStatus = std::move(statesStatus);
  }

  const std::vector<dsModel::DeviceState>& Device::states() const {
    return data().states;
  }

  void Device::setStates(std::vector<dsModel::DeviceState>&& states) {
    if (m_data->states == states) {
      return;
    }
    m_data->states = std::move(states);
  }
} // namespace dss

