/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DEVICE_H
#define DEVICE_H

#include <iosfwd>
#include <bitset>
#include <utility>

#include <boost/locale/message.hpp>
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include <ds-model/model-fwd.h>
#include <ds-model/model-ids.h>
#include <ds-model/scenarios-fwd.h>
#include <ds/enum.h>

#include "taskprocessor.h"
#include "addressablemodelitem.h"
#include "businterface.h"
#include "datetools.h"
#include "device-info.h"
#include "ds485types.h"
#include "model/action.h"
#include "model/data_types.h"
#include "model/devicereference.h"
#include "modelevent.h"
#include "vdc-connection.h"

#define BUTTON_ACTIVE_GROUP_RESET    0xff

// pairing JSON strings
#define BUTTONINPUT_1WAY            "1way"
#define BUTTONINPUT_2WAY_DOWN       "2way_down"
#define BUTTONINPUT_2WAY_UP         "2way_up"
#define BUTTONINPUT_2WAY            "2way"
#define BUTTONINPUT_1WAY_COMBINED   "1way_combined"

// AKM JSON strings
#define BUTTONINPUT_AKM_STANDARD            "standard"
#define BUTTONINPUT_AKM_INVERTED            "inverted"
#define BUTTONINPUT_AKM_ON_RISING_EDGE      "on_rising_edge"
#define BUTTONINPUT_AKM_ON_FALLING_EDGE     "on_falling_edge"
#define BUTTONINPUT_AKM_OFF_RISING_EDGE     "off_rising_edge"
#define BUTTONINPUT_AKM_OFF_FALLING_EDGE    "off_falling_edge"
#define BUTTONINPUT_AKM_RISING_EDGE         "rising_edge"
#define BUTTONINPUT_AKM_FALLING_EDGE        "falling_edge"

// UMR output modes for pairing, import from Config.js in UI
#define OUTPUT_MODE_DEACTIVATED             0
#define OUTPUT_MODE_TWO_STAGE_SWITCH        34
#define OUTPUT_MODE_SIMPLE_SWITCH 35
#define OUTPUT_MODE_THREE_STAGE_SWITCH      38
#define OUTPUT_MODE_SWITCHED_AUTO           39
#define OUTPUT_MODE_WIPE                    40
#define OUTPUT_MODE_POWERSAVE               41
#define OUTPUT_MODE_BIPOLAR_SWITCH          43

#define OUTPUT_MODE_TEMPCONTROL_SWITCHED    65
#define OUTPUT_MODE_TEMPCONTROL_2OUT_2STEPS 67
#define OUTPUT_MODE_TEMPCONTROL_2OUT_3STEPS 68
#define OUTPUT_MODE_TEMPCONTROL_2OUT_PARALLEL 69

namespace vdcapi {
  class PropertyElement;
}

DS_REFLECTED_ENUM((dss), ButtonUsage, (unsupported)(used)(auto_unused)(manual_unused));

namespace dss {

  class Group;
  class Cluster;
  class State;
  class DSMeter;
  class VdcElementReader;
  struct VdsdSpec_t;
  struct DeviceSpec;

  typedef struct DeviceSceneSpec {
    bool dontcare;
    bool localprio;
    bool specialmode;
    bool flashmode;
    uint8_t ledconIndex;
    uint8_t dimtimeIndex;
  } DeviceSceneSpec_t;

  typedef struct {
    uint8_t bank;
    uint8_t offset;
    uint8_t value;
    uint8_t mask;
  } DeviceConfig_t;

  typedef struct {
    uint8_t colorSelect;
    uint8_t modeSelect;
    bool dimMode;
    bool rgbMode;
    bool groupColorMode;
  } DeviceLedSpec_t;

  typedef struct {
    std::string name;
    uint8_t sensorIndex;
    uint8_t test;
    uint8_t action;
    uint16_t value;
    uint16_t hysteresis;
    uint8_t sceneDeviceMode;
    uint8_t validity;
    uint8_t buttonNumber;
    uint8_t clickType;
    uint8_t sceneID;
    uint8_t minimalDuration; //< condition must last at least for so many seconds
  } DeviceSensorEventSpec_t;

  typedef struct {
    uint16_t protectionTimer;
    uint16_t emergencyTimer;
    uint8_t emergencyControlValue;
  } DeviceValveTimerSpec_t;

  typedef struct {
    uint16_t pwmPeriod;
    uint8_t pwmMinX;
    uint8_t pwmMaxX;
    uint8_t pwmMinY;
    uint8_t pwmMaxY;
    int8_t pwmOffset;
    uint8_t switchedMin; // BL-KM300 only, for BL-KM200 these values equal
    uint8_t switchedMax; // pwmMinY and PwmMaxX
  } DeviceValvePwmSpec_t;

  typedef struct {
    uint8_t pwmValue;
    uint8_t pwmPriorityMode;
  } DeviceValvePwmStateSpec_t;

  typedef struct {
    bool ctrlClipMinZero;
    bool ctrlClipMinLower;
    bool ctrlClipMaxHigher;
    bool ctrlNONC;
    uint8_t ctrlRawValue;
  } DeviceValveControlSpec_t;

  typedef struct {
    bool pairing;       // device supports pairing
    bool syncButtonID;  // sync button ID setting of slave with master
    bool hasOutputAngle;
    bool posTimeMax; // device supports maximum positioning time
    bool posTime; // device supports up/down positioning time settings
  } DeviceFeatures_t;

  class Device;
  struct DeviceBinaryInput : boost::noncopyable {
    int m_inputIndex;        // input line index
    BinaryInputType m_inputType; // type of input signal
    BinaryInputId m_inputId;           // target Id, like ButtonId
    int m_targetGroupId;     // index of target group, 0..63
    bool m_independent;

    std::string m_sInputName;  // read-only name, vdc-api
    std::string m_sInputId;    // unique id (vdc-api)

    DeviceBinaryInput(Device& device, const DeviceBinaryInputSpec_t& spec, int index);
    ~DeviceBinaryInput();

    const State& getState() const { return *m_state; }
    void setTargetId(uint8_t group);
    void setInputId(BinaryInputId inputId);
    void setInputType(BinaryInputType inputType);
    void handleEvent(BinaryInputStateValue inputState);

    void updateStatusFieldHandle();
  private:
    __DECL_LOG_CHANNEL__;
    Device& m_device;
    std::string m_name;
    boost::shared_ptr<State> m_state;
    class StatusFieldHandle;
    std::unique_ptr<StatusFieldHandle> m_statusFieldHandle;
  };

  typedef struct {
    int m_sensorIndex;       // sensor index
    std::string m_sensorId;
    std::string m_sensorName;
    SensorType m_sensorType;
    uint8_t m_sensorUsage;
    uint32_t m_sensorPollInterval;
    bool m_sensorBroadcastFlag;
    bool m_sensorPushConversionFlag;
    bool m_sensorValueValidity;
    unsigned int m_sensorValue;
    double m_sensorPrevValueFloat;
    double m_sensorValueFloat;
    DateTime m_sensorValueTS;
  } DeviceSensor_t;

  typedef struct {
    int index;
    std::string id;
    std::string technicalName;
    ChannelType type;
  } Channel;

  typedef enum {
    DEVICE_TYPE_INVALID = -1,
    DEVICE_TYPE_KM      = 0,
    DEVICE_TYPE_TKM     = 1,
    DEVICE_TYPE_SDM     = 2,
    DEVICE_TYPE_KL      = 3,
    DEVICE_TYPE_TUP     = 4,
    DEVICE_TYPE_ZWS     = 5,
    DEVICE_TYPE_SDS     = 6,
    DEVICE_TYPE_SK      = 7,
    DEVICE_TYPE_AKM     = 8,
    DEVICE_TYPE_TNY     = 10,
    DEVICE_TYPE_UMV     = 11,
    DEVICE_TYPE_UMR     = 12,
  } DeviceTypes_t;

  typedef enum {
    DEVICE_CLASS_INVALID    = -1,
    DEVICE_CLASS_GE         = 1,
    DEVICE_CLASS_GR         = 2,
    DEVICE_CLASS_BL         = 3,
    DEVICE_CLASS_TK         = 4,
    DEVICE_CLASS_MG         = 5,
    DEVICE_CLASS_RT         = 6,
    DEVICE_CLASS_GN         = 7,
    DEVICE_CLASS_SW         = 8,
    DEVICE_CLASS_WE         = 9
  } DeviceClasses_t;

  typedef enum {
    DEVICE_OEM_UNKNOWN,
    DEVICE_OEM_NONE,
    DEVICE_OEM_LOADING,
    DEVICE_OEM_VALID,
  } DeviceOEMState_t;

  typedef enum {
    DEVICE_OEM_EAN_NO_EAN_CONFIGURED = 0,
    DEVICE_OEM_EAN_NO_INTERNET_ACCESS = 1,
    DEVICE_OEM_EAN_INTERNET_ACCESS_OPTIONAL = 2,
    DEVICE_OEM_EAN_INTERNET_ACCESS_MANDATORY = 3,
  } DeviceOEMInetState_t;

  typedef enum {
    DEVICE_VALVE_UNKNOWN = 0,
    DEVICE_VALVE_FLOOR = 1,
    DEVICE_VALVE_RADIATOR = 2,
    DEVICE_VALVE_WALL = 3,
    DEVICE_CONVECTOR_PASSIVE = 4,
    DEVICE_CONVECTOR_ACTIVE = 5,
    DEVICE_VALVE_FLOOR_LOW_ENERGY = 6
  } DeviceValveType_t;

  struct DeviceConfigBulk {
    std::string dsuid;
    int bank;
    int offset;
    int value;
  };
  typedef std::vector<struct DeviceConfigBulk> DeviceConfigBulkVector;

  class BulkConfig : public Task {
  public:
    BulkConfig(Apartment* pApartment, DeviceConfigBulkVector&& configVector)
    : m_pApartment(pApartment),
      m_configVector(configVector) {
      m_id = "BulkConfig" + DateTime().toISO8601_ms_local();
    }
    void run() override;
    const std::string& getId () const { return m_id; }
  private:
    Apartment* m_pApartment;
    DeviceConfigBulkVector m_configVector;
    std::string m_id;
  };

  struct DeviceData;

  /** Represents a dsID */
  class Device : public AddressableModelItem,
                 public boost::noncopyable {
  private:
    __DECL_LOG_CHANNEL__;
    std::string m_Name;
    dsuid_t m_DSID;
    devid_t m_ShortAddress;
    devid_t m_LastKnownShortAddress;
    int m_ZoneID;
    int m_LastKnownZoneID;
    dsuid_t m_DSMeterDSID;
    dsuid_t m_LastKnownMeterDSID;
    std::string m_DSMeterDSIDstr; // for proptree publishing
    std::string m_DSMeterDSUIDstr; // for proptree publishing
    std::string m_LastKnownMeterDSIDstr; // for proptree publishing
    std::string m_LastKnownMeterDSUIDstr; // for proptree publishing
    std::vector<int> m_groupIds;
    int m_ActiveGroup;
    int m_DefaultGroup;
    int m_FunctionID;
    int m_ProductID;
    int m_VendorID;
    int m_RevisionID;
    HardwareType m_hardwareType;
    std::string m_GTIN;
    int m_LastCalledScene;
    int m_LastButOneCalledScene;
    unsigned long m_Consumption;
    unsigned long m_Energymeter;
    DateTime m_LastDiscovered;
    DateTime m_FirstSeen;
    bool m_IsValid;
    bool m_IsLockedInDSM;
    bool m_IgnoreOperationLock;

    uint8_t m_OutputMode;
    ButtonInputMode m_ButtonInputMode;
    uint8_t m_ButtonInputIndex;
    uint8_t m_ButtonInputCount;
    bool m_ButtonSetsLocalPriority;
    bool m_ButtonCallsPresent;
    int m_ButtonActiveGroup;
    int m_ButtonID;
    unsigned long long m_OemEanNumber;
    uint16_t m_OemSerialNumber;
    uint8_t m_OemPartNumber;
    DeviceOEMInetState_t m_OemInetState;
    DeviceOEMState_t m_OemState;
    bool m_OemIsIndependent;
    bool m_ActiveCoolingMode;

    PropertyNodePtr m_pAliasNode;
    PropertyNodePtr m_TagsNode;
    std::string m_HWInfo;
    std::string m_iconPath;
    DeviceOEMState_t m_OemProductInfoState;
    std::string m_OemProductName;
    std::string m_OemProductIcon;
    std::string m_OemProductURL;
    std::string m_OemConfigLink;

    VdsdSpec_t m_vdsdSpec;
    bool m_isVdcDevice;
    std::string m_VdcIconPath;
    // TODO(someday) merge with VdsdSpec_t
    bool m_hasActions;
    // TODO(someday) merge with VdsdSpec_t
    bool m_hasLegacyActions; // legacy actionDescriptions with std. prefix

    const std::unique_ptr<DeviceData> m_data;
    std::vector<BasicScene> m_supportedBasicScenes;

    std::map<ModelFeatureId, bool> m_modelFeatures;

    DeviceValveType_t m_ValveType;

    bool m_IsConfigLocked;

    std::vector<boost::shared_ptr<DeviceBinaryInput> > m_binaryInputs;

    typedef std::map<std::string, boost::shared_ptr<State> > States;
    States m_states;

    std::vector<std::unique_ptr<DeviceSensor_t>> m_sensorInputs;
    uint8_t m_outputChannelCount;
    std::vector<boost::shared_ptr<Channel>> m_outputChannels;

    std::string m_AKMInputProperty;

    mutable boost::recursive_mutex m_deviceMutex;

    CardinalDirection_t m_cardinalDirection; //<  wind protection of blinds
    WindProtectionClass_t m_windProtectionClass;
    int m_floor;
    int m_pairedDevices;
    bool m_visible; // currently relevant for TNY only
    uint8_t m_maxPosTimeSeconds;
    bool m_emulatedHardwareOutputs{}; // true -> emulate setDeviceOutputChannelValue calls over legacy outputs api
    ButtonUsage m_buttonUsage;

    void updateModelFeatures();
    void republishModelFeaturesToPropertyTree();

    friend class BusScanner;
    friend class ModelMaintenance;
    friend class BackendVdcsImpl;

    void setSensorEventTableEntryZws205(const int row, const DeviceSensorEventSpec_t entry);
    void getSensorEventTableEntryZws205(const int row, DeviceSensorEventSpec_t& entry);

  protected:
    /** Sends the application a note that something has changed.
     * This will cause the \c apartment.xml to be updated. */
    void dirty();

    /** Sends the application a note that the cluster for the given device
     *  has to be checked and reassigned. */
    void checkAutoCluster() const;

    void fillSensorTable(std::vector<DeviceSensorSpec_t>& _slist);
    void calculateHWInfo();
    void updateIconPath();
    std::string getAKMButtonInputString(ButtonInputMode mode) const;
    bool hasBlinkSettings() const;
    void publishAliasInZoneGroups() const;

    // use get/setDeviceSceneMode instaed of these functions, only internal
    void setDeviceSceneModeStandard(uint8_t _sceneId, DeviceSceneSpec_t _config) const;
    void getDeviceSceneModeStandard(uint8_t _sceneId, DeviceSceneSpec_t& _config) const;

    void setConsumptionVisualizationEnabledZws205(bool enabled) const;
    bool getConsumptionVisualizationEnabledZws205() const;
    bool isBLKM300() const;

  public:
    void getDeviceOutputChannelSceneConfig(uint8_t scene, DeviceSceneSpec_t& config) const;
    void setDeviceOutputChannelSceneConfig(uint8_t scene, DeviceSceneSpec_t config) const;
    /** Creates and initializes a device. */
    Device(const dsuid_t _dsid, Apartment* _pApartment);
    virtual ~Device();

    dsModel::DeviceId id() const;
    dsModel::SubmoduleId submoduleId() const;
    dsModel::FunctionBlockId functionBlockId() const;

    bool isInDsModel() const;

    boost::shared_ptr<Device> sharedFromThis() { return boost::static_pointer_cast<Device>(shared_from_this()); }
    DeviceReference reference() const { return DeviceReference{m_DSID, m_pApartment}; }
    boost::shared_ptr<DeviceReference> makeSharedReference() const;

    /** @copydoc DeviceReference::isOn() */
    virtual bool isOn() const;

    /** Set device configuration values */
    void setDeviceConfig(uint8_t _configClass, uint8_t _configIndex, uint8_t _value, bool sendEvents = true) const;
    void setDeviceConfig16(uint8_t _configClass, uint8_t _configIndex, uint16_t _value, bool sendEvents = true) const;
    void setDeviceButtonID(uint8_t _buttonId);
    void setDeviceButtonActiveGroup(uint8_t _buttonActiveGroup);
    void setDeviceJokerGroup(uint8_t _groupId);
    void setDeviceOutputMode(uint8_t _modeId);
    void setDeviceButtonInputMode(ButtonInputMode mode);
    void setProgMode(uint8_t _modeId) const;
    void setButtonInputMode(ButtonInputMode mode);

    bool hasExtendendSceneTable() const;

    boost::shared_ptr<Device> tryGetPartnerDevice() const;
    boost::shared_ptr<Device> getPartnerDevice() const;

    void increaseDeviceOutputChannelValue(uint8_t _channel) const;
    void decreaseDeviceOutputChannelValue(uint8_t _channel) const;
    void stopDeviceOutputChannelValue(uint8_t _channel) const;
    uint16_t getDeviceOutputChannelValue(uint8_t _channel) const;
    void setDeviceOutputChannelValue(uint8_t _channel, uint8_t _size,
                                     uint16_t _value, bool _applyNow = true) const;
    uint16_t getDeviceOutputChannelSceneValue(uint8_t _channel, uint8_t _scene) const;
    void setDeviceOutputChannelSceneValue(uint8_t _channel, uint8_t _size,
                                          uint8_t _scene, uint16_t _value) const;
    void setDeviceOutputChannelDontCareFlags(uint8_t _scene, uint16_t _value) const;
    uint16_t getDeviceOutputChannelDontCareFlags(uint8_t _scene) const;

    std::string getDisplayID() const;

    bool is2WayMaster() const;
    bool is2WaySlave() const;
    bool hasMultibuttons() const;
    bool hasInput() const;
    bool isBinaryInput() const;
    bool hasOutput() const;
    bool isInternallyControlled() const;
    DeviceTypes_t getDeviceType() const;
    int getDeviceNumber() const;
    DeviceClasses_t getDeviceClass() const;
    const DeviceFeatures_t getDeviceFeatures() const;
    std::string getAKMInputProperty() const;
    int multiDeviceIndex() const;

    /** Configure scene configuration */
    void setDeviceSceneMode(uint8_t sceneId, DeviceSceneSpec_t config) const;
    void getDeviceSceneMode(uint8_t sceneId, DeviceSceneSpec_t& config) const;

    void setDeviceLedMode(uint8_t _ledconIndex, DeviceLedSpec_t _config) const;
    void getDeviceLedMode(uint8_t _ledconIndex, DeviceLedSpec_t& _config) const;
    void setDeviceTransitionTime(uint8_t _dimtimeIndex, int up, int down) const;
    void getDeviceTransitionTime(uint8_t _dimtimeIndex, int& up, int& down) const;
    int transitionVal2Time(uint8_t _value) const;
    uint8_t transitionTimeEval(int timems) const;
    bool evaluateGRKL387Workaround(const int _scene, const int _value, const int _angle);
    void setSceneValue(const int _scene, const int _value) const;
    int getSceneValue(const int _scene) const;
    void setSceneAngle(const int _scene, const int _angle) const;
    int getSceneAngle(const int _scene) const;
    void configureAreaMembership(const int _areaScene, const bool _addToArea);

    /** Configure climate actuator */
    void setDeviceValveTimer(DeviceValveTimerSpec_t _config) const;
    void getDeviceValveTimer(DeviceValveTimerSpec_t& _config) const;
    void setDeviceValvePwm(DeviceValvePwmSpec_t _config) const;
    void getDeviceValvePwm(DeviceValvePwmSpec_t& _config) const;
    void getDeviceValvePwmState(DeviceValvePwmStateSpec_t& _config) const;
    void setDeviceValveControl(DeviceValveControlSpec_t _config) const;
    void getDeviceValveControl(DeviceValveControlSpec_t& _config) const;
    void setDeviceValveSwitchedThresholdMin(uint8_t value) const;
    void getDeviceValveSwitchedThresholdMin(uint8_t *threshold) const;
    void setDeviceValveSwitchedThresholdMax(uint8_t value) const;
    void getDeviceValveSwitchedThresholdMax(uint8_t *threshold) const;


    /** Binary input devices */
    void setDeviceBinaryInputId(uint8_t _inputIndex, BinaryInputId _targetId) const;
    void setDeviceBinaryInputTargetId(uint8_t _inputIndex, uint8_t _targetGroup);
    void setDeviceBinaryInputType(uint8_t _inputIndex, BinaryInputType _inputType) const;
    BinaryInputType getDeviceBinaryInputType(uint8_t _inputIndex) const;
    /** AKM2xx timeout settings */
    void setDeviceAKMInputTimeouts(int _onDelay, int _offDelay) const;
    void getDeviceAKMInputTimeouts(int& _onDelay, int& _offDelay) const;

    /** Returns device configuration value */
    uint8_t getDeviceConfig(uint8_t _configIndex, uint8_t _value) const;
    uint16_t getDeviceConfigWord(uint8_t _configIndex, uint8_t _value) const;

    /** Verify communication path */
    std::pair<uint8_t, uint16_t> getDeviceTransmissionQuality() const;

    /** Set device output value */
    void setDeviceValue(uint8_t _value) const;
    void setDeviceOutputValue(uint8_t _offset, uint16_t _value) const;
    uint16_t getDeviceOutputValue(uint8_t _offset) const;

    /** query device sensor values */
    uint16_t getDeviceSensorValue(const int _sensorIndex) const;
    DeviceSensorValue_t getDeviceSensorValueEx(const int _sensorIndex) const;

    /** query cached last known values */
    virtual unsigned long getPowerConsumption();
    virtual unsigned long getEnergymeterValue();

    /** convenience methods for simulated devices */
    virtual void nextScene(const callOrigin_t _origin, const SceneAccessCategory _category);
    virtual void previousScene(const callOrigin_t _origin, const SceneAccessCategory _category);

    void setPartiallyFromSpec(const DeviceSpec& spec);

    /** Returns the function ID of the device.
     * A function ID specifies a certain subset of functionality that
     * a device implements.
     */
    int getFunctionID() const;

    /** Returns the Product ID of the device.
     * The Product ID identifies the manufacturer and type of device.
     */
    int getProductID() const;

    /** Returns the Revision ID of the device.
     * The revision identifies the device hardware revision.
     */
    int getRevisionID() const;

    // Revision id serialized as "major.mid.minor"
    // It is called "Firmware Version" in UI
    std::string getRevisionAsString() const;

    /** Returns the Vendor ID of the device.
     * The ID identifies the device hardware vendor.
     */
    int getVendorID() const;
    std::string vendorTitle() const;

    /** Returns the name of the device. */
    const std::string& getName() const;
    /** Sets the name of the device.
     * @note This will cause the apartment to store
     * \c apartment.xml
     */
    void setName(const std::string& _name);

    const std::string& getHWInfo() const { return m_HWInfo; }
    const std::string& getGTIN() const { return m_GTIN; }
    const std::string& getIconPath() const { return m_iconPath; }

    static const std::string getDeviceTypeString(const DeviceTypes_t _type);
    static const std::string getDeviceClassString(const DeviceClasses_t _class);
    static const std::string getColorString(const int _class);

    /** Returns wheter the device is in group \a _groupID or not. */
    bool isInGroup(const int _groupID) const;
    /** Adds the device to group \a _groupID. */
    void addToGroup(const int _groupID);
    /** Removes the device from group \a _groupID */
    void removeFromGroup(const int _groupID);
    /** Sets Binray Target Group \a groupID for Binary Input */
    void setBinaryStateTargetGroup(const int groupID);

    /** Returns the group id of the \a _index'th group */
    int getGroupIdByIndex(const int _index) const;
    /** Returns \a _index'th group of the device */
    boost::shared_ptr<Group> getGroupByIndex(const int _index);
    /** Returns the number of groups the device is a member of */
    int getGroupsCount() const;
    /** Returns the vector of groups the device is in */
    const std::vector<int>& getGroupIds() const { return m_groupIds; }

    /** Returns the zoneID that this device group is in. */
    int getGroupZoneID(int groupID) const;
    /// Returns zone local group or global group
    boost::weak_ptr<Group> tryGetGroup(int groupId) const;

    std::vector<boost::shared_ptr<Cluster>> getClusters() const;

    /** Removes the device from all group.
     * The device will remain in the broadcastgroup though.
     */
    void resetGroups();

    /** Returns the last called scene.
     * At the moment this information is used to determine whether a device is
     * turned on or not. */
    int getLastCalledScene() const { return m_LastCalledScene; }
    /** Sets the last called scene. */
    void setLastCalledScene(const int _value) {
      if (_value != m_LastCalledScene) {
        m_LastButOneCalledScene = m_LastCalledScene;
        m_LastCalledScene = _value;
      }
    }
    /** If state hasn't changed undo the last called scene command and restore state. */
    void setLastButOneCalledScene(const int _value) {
      if (_value == m_LastCalledScene) {
        m_LastCalledScene = m_LastButOneCalledScene;
      }
    }
    /** Undo the last called scene command and restore state. */
    void setLastButOneCalledScene() {
      m_LastCalledScene = m_LastButOneCalledScene;
    }

    /** Returns the short address of the device. This is the address
     * the device got from the dSM. */
    const devid_t& getShortAddress() const { return m_ShortAddress; }
    /** Sets the short-address of the device. */
    void setShortAddress(const devid_t _shortAddress);

    devid_t getLastKnownShortAddress() const;
    void setLastKnownShortAddress(const devid_t _shortAddress);
    /** Returns the DSID of the device */
    dsuid_t getDSID() const;
    /** Returns the DSID of the dsMeter the device is connected to */
    dsuid_t getDSMeterDSID() const;

    const dsuid_t& getLastKnownDSMeterDSID() const;
    void setLastKnownDSMeterDSID(const dsuid_t& _value);
    void setDSMeter(boost::shared_ptr<DSMeter> _dsMeter);

    /** Returns the zone ID the device resides in. */
    int getZoneID() const;
    /** Sets the zone ID of the device. */
    void setZoneID(const int _value);
    int getLastKnownZoneID() const;
    void setLastKnownZoneID(const int _value);
    /** Returns the apartment the device resides in. */
    Apartment& getApartment() const { return *m_pApartment; }

    const DateTime& getLastDiscovered() const { return m_LastDiscovered; }
    const DateTime& getFirstSeen() const { return m_FirstSeen; }
    void setFirstSeen(const DateTime& _value) { m_FirstSeen = _value; }

    const PropertyNodePtr& getPropertyNode() const { return m_pPropertyNode; }
    PropertyNodePtr getPropertyNode() { return m_pPropertyNode; }

    bool getIgnoreOperationLock() const { return m_IgnoreOperationLock; }
    void setIgnoreOperationLock(const bool value);
    void setDeviceIgnoreOperationLock(const bool value);

    void updateAKMNode();
    /** Publishes the device to the property tree.
     * @see DSS::getPropertySystem */
    void publishToPropertyTree();
    /** Removes the device from the propertytree. */
    void removeFromPropertyTree();

    /** Returns wheter the dSM has been told to never forget
        this device (cached value) */
    bool getIsLockedInDSM() const { return m_IsLockedInDSM; }
    void setIsLockedInDSM(const bool _value);

    /** Returns whether the device data is completely synchronized or
        the synchronization is pending. */
    bool isValid() const { return m_IsValid; }
    void setIsValid(const bool _value) { m_IsValid = _value; }

    /** Tells the dSM to never forget a device. */
    void lock();
    /** Tells the dSM that it may forget a device if it's not present. */
    void unlock();

    /** Device level active and default group used for Global Applications. */
    void setActiveGroup(const int _value) { m_ActiveGroup = _value; }
    void setActiveGroup(ApplicationType x) { setActiveGroup(static_cast<int>(x)); }
    int getActiveGroup() const { return m_ActiveGroup; }
    ApplicationType applicationType() const { return static_cast<ApplicationType>(m_ActiveGroup); }
    int getDefaultGroup() const { return m_DefaultGroup; }

    void setButtonSetsLocalPriority(const bool _value) { m_ButtonSetsLocalPriority = _value; }
    bool getButtonSetsLocalPriority() const { return m_ButtonSetsLocalPriority; }
    void setButtonCallsPresent(const bool _value) { m_ButtonCallsPresent = _value; }
    bool getButtonCallsPresent() const { return m_ButtonCallsPresent; }
    void setButtonActiveGroup(int value) { m_ButtonActiveGroup = value; }
    int getButtonActiveGroup() const { return m_ButtonActiveGroup; }
    void setButtonID(const int _value) { m_ButtonID = _value; }
    int getButtonID() const { return m_ButtonID; }
    uint8_t getOutputMode() const { return m_OutputMode; }
    void setOutputMode(const uint8_t _value);
    ButtonInputMode getButtonInputMode() const { return m_ButtonInputMode; }
    void setButtonInputIndex(const uint8_t _value) { m_ButtonInputIndex = _value; }
    uint8_t getButtonInputIndex() const { return m_ButtonInputIndex; }
    void setButtonInputCount(const uint8_t _value) { m_ButtonInputCount = _value; }
    uint8_t getButtonInputCount() const { return m_ButtonInputCount; }

    bool hasTag(const std::string& _tagName) const;
    void addTag(const std::string& _tagName);
    void removeTag(const std::string& _tagName);
    std::vector<std::string> getTags();

    /** Returns wheter two devices are equal.
     * Devices are considered equal if their DSID are a match.*/
    bool operator==(const Device& _other) const;

    std::string getSensorEventName(const int _eventIndex) const;
    void setSensorEventName(const int _eventIndex, const std::string& _name);
    void getSensorEventEntry(const int _eventIndex, DeviceSensorEventSpec_t& _entry);
    void setSensorEventEntry(const int _eventIndex, DeviceSensorEventSpec_t _entry);

    // valve configuration
    bool isValveDevice() const;
    DeviceValveType_t getValveType() const;
    void setValveType(DeviceValveType_t _type);
    void setDeviceValveType(DeviceValveType_t _type);
    DeviceValveType_t getDeviceValveType();

    std::string getValveTypeAsString() const;
    bool setDeviceValveTypeAsString(const std::string &_value);

    static std::string valveTypeToString(const DeviceValveType_t _type);
    static bool getValveTypeFromString(const char* _string, DeviceValveType_t &deviceType);

    void publishValveTypeToPropertyTree();

    void setOemInfo(const unsigned long long _eanNumber,
        const uint16_t _serialNumber, const uint8_t _partNumber,
        const DeviceOEMInetState_t _iNetState,
        const bool _isIndependent);
    void setOemInfoState(const DeviceOEMState_t _state);

    unsigned long long getOemEan() const { return m_OemEanNumber; }
    std::string getOemEanAsString() const { return unsignedLongIntToString(m_OemEanNumber); }
    DeviceOEMState_t getOemInfoState() const { return m_OemState; }
    std::string getOemStateAsString() const { return oemStateToString(m_OemState); }
    static std::string oemStateToString(const DeviceOEMState_t _state);
    uint16_t getOemSerialNumber() const { return m_OemSerialNumber; }
    uint8_t getOemPartNumber() const { return m_OemPartNumber; }
    bool getOemIsIndependent() const { return m_OemIsIndependent; }
    DeviceOEMInetState_t getOemInetState() const { return m_OemInetState; }
    std::string getOemInetStateAsString() const { return oemInetStateToString(m_OemInetState); }
    static std::string oemInetStateToString(const DeviceOEMInetState_t _state);

    DeviceOEMState_t getOemStateFromString(const char* _string) const;
    DeviceOEMInetState_t getOemInetStateFromString(const char* _string) const;

    void setSupportedBasicScenes(const std::vector<BasicScene>& scenes);
    std::vector<BasicScene> getSupportedBasicScenes() const { return m_supportedBasicScenes; }

    void setOemProductInfo(const std::string& _productName, const std::string& _iconPath, const std::string& _productURL, const std::string& _configLink);
    void setOemProductInfoState(const DeviceOEMState_t _state);
    std::string getOemProductInfoStateAsString() const { return oemStateToString(m_OemProductInfoState); }
    DeviceOEMState_t getOemProductInfoState() const { return m_OemProductInfoState; }
    const std::string& getOemProductName() const { return m_OemProductName; }
    const std::string& getOemProductIcon() const { return m_OemProductIcon; }
    const std::string& getOemProductURL() const { return m_OemProductURL; }
    const std::string& getOemConfigLink() const { return m_OemConfigLink; }
    bool isOemCoupledWith(boost::shared_ptr<Device> _otherDev);
    void setConfigLock(bool _lockConfig);
    bool isConfigLocked() const { return m_IsConfigLocked; }

    void publishVdcToPropertyTree();
    void setVdcDevice(bool _isVdcDevice);
    bool isVdcDevice() const { return m_isVdcDevice; }
    const VdsdSpec_t& getVdsdSpec() const { return m_vdsdSpec; }
    void setVdsdSpec(VdsdSpec_t &&x);
    /// Domain for translating user visible strings originating from this vdc device
    std::string vdcTrDomain() const;
    void setActiveCoolingMode(bool activeCoolingMode) { m_ActiveCoolingMode = activeCoolingMode; }
    bool getActiveCoolingMode() const { return m_ActiveCoolingMode; }

  public:
    void setVdcIconPath(const std::string& _value) { m_VdcIconPath = _value; updateIconPath(); }
    const std::string& getVdcIconPath() const { return m_VdcIconPath; }

    void setHasActions(bool x) { m_hasActions = x; }
    bool getHasActions() const { return m_hasActions; }
    void setHasLegacyActions(bool x) { m_hasLegacyActions = x; }
    bool getHasLegacyActions() const { return m_hasLegacyActions; }

    const std::map<ModelFeatureId, bool>& getModelFeatures() const { return m_modelFeatures; }
    bool supportsApartmentApplications() const;

    void setBinaryInputs(const std::vector<DeviceBinaryInputSpec_t>& _binaryInput);
    void setBinaryInputsInfo(const uint8_t dsIndex, const std::string& _descriptionId, const std::string& _inputName);
    uint8_t getBinaryInputCount() const;
    const std::vector<boost::shared_ptr<DeviceBinaryInput> >& getBinaryInputs() const;
    const boost::shared_ptr<DeviceBinaryInput> getBinaryInput(uint8_t _inputIndex) const;
    const boost::shared_ptr<DeviceBinaryInput> getBinaryInput(const std::string& _inputId) const;
    void setBinaryInputTargetId(uint8_t _index, uint8_t targetGroup);
    void setBinaryInputId(uint8_t _index, BinaryInputId _inputId);
    void setBinaryInputType(uint8_t _index, BinaryInputType _inputType);
    void clearBinaryInputs();

    void initStates(const std::vector<DeviceStateSpec_t>& specs);
    const std::map<std::string, boost::shared_ptr<State> >& getStates() const { return m_states; }
    void clearStates();
    void setStateValue(const std::string& name, const std::string& value);
    void setStateValues(const std::vector<std::pair<std::string, std::string> >& values);

    void setSensors(const std::vector<DeviceSensorSpec_t>& _binaryInput);
    void setSensorsInfo(const uint8_t sensorIndex, const std::string& _sensorId, const std::string& _sensorName, const int _sensorUsage);
    uint8_t getSensorCount() const;
    const std::vector<std::unique_ptr<DeviceSensor_t>>& getSensors() const;
    const DeviceSensor_t& getSensor(uint8_t _inputIndex) const;
    const DeviceSensor_t& getSensorByType(SensorType _sensorType) const;
    void setSensorValue(int _sensorIndex, unsigned int _sensorValue);
    void setSensorValue(int _sensorIndex, double _sensorValue, uint32_t _age = 0);
    void setSensorDataValidity(int _sensorIndex, bool _valid);
    bool isSensorDataValid(int _sensorIndex) const;

    void setOutputChannels(const std::vector<ChannelType>&);
    void setOutputChannelInfo(const uint8_t index, const Channel&);
    int getOutputChannelCount() const;
    int getOutputChannelIndex(ChannelType) const;
    boost::shared_ptr<Channel> getOutputChannel(uint8_t _index) const;
    boost::shared_ptr<Channel> getOutputChannelById(const std::string& _channelId) const;
    const std::vector<boost::shared_ptr<Channel>>& getOutputChannels() const;

    uint8_t getDeviceUMVRelayValue();
    void setDeviceUMVRelayValue(uint8_t _value);

    void setDeviceBlinkRepetitions(uint8_t count) const;
    void setDeviceBlinkOnDelay(double delay) const;
    void setDeviceBlinkOffDelay(double delay) const;
    void getDeviceBlinkSettings(double *ondelay, double *offdelay, uint8_t *count) const;

    void setCardinalDirection(CardinalDirection_t _direction, bool _initial = false);
    CardinalDirection_t getCardinalDirection() const {
      return m_cardinalDirection;
    }

    void setWindProtectionClass(WindProtectionClass_t _klass, bool _initial = false);
    WindProtectionClass_t getWindProtectionClass() const {
      return m_windProtectionClass;
    }


    uint8_t getSWThresholdAddress() const;
    void setSwitchThreshold(uint8_t _threshold) const;
    uint8_t getSwitchThreshold() const;

    void setTemperatureOffset(int8_t _offset) const;
    int8_t getTemperatureOffset() const;

    void setFloor(int _floor) {
      if (m_floor != _floor) {
        m_floor = _floor;
        dirty();
      }
    }
    int getFloor() const {
      return m_floor;
    }

    std::vector<int> getLockedScenes() const;
    bool isInLockedCluster() const;

    uint16_t getDeviceMaxMotionTime() const;
    void setDeviceMaxMotionTime(uint16_t seconds) const;

    double getDevicePosTimeUp() const;
    void setDevicePosTimeUp(double seconds) const;

    double getDevicePosTimeDown() const;
    void setDevicePosTimeDown(double seconds) const;

    uint8_t getMaxPosTime() const;
    void setMaxPosTime(uint8_t seconds);

    // below functions are currently only supported by TNYs
    void setPairedDevices(int _num);
    int getPairedDevices() const;
    void setVisibility(bool _isVisible);
    void setDeviceVisibility(bool _isVisible);
    bool isVisible() const;
    bool isMainDevice() const;
    dsuid_t getMainDeviceDSUID() const;

    void handleBinaryInputEvent(const int index, BinaryInputStateValue state);
    void handleBinaryInputEvent(const std::string& inputId, BinaryInputStateValue inputState);

    /// Calls (invokes) device specific action.
    void callAction(callOrigin_t, SceneAccessCategory, const std::string& actionId,
        const vdcapi::PropertyElement& params, const std::locale&);

    void setProperty(const vdcapi::PropertyElement& element); // element.name() is property id
    void deviceChanged();
    void setCustomAction(const std::string& id, const std::string& title, const std::string& action,
                         const vdcapi::PropertyElement& params);
    vdcapi::Message getVdcProperty(const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& query);
    void setVdcProperty(const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement >& query);

    // internal data stored persistently in apartment.xml
    const DeviceData& data() const;
    void setData(DeviceData&&);

    std::vector<dsModel::ScenarioId> scenarioIds() const;
    const std::vector<dsModel::Scenario>& scenarios() const;
    void setScenarios(std::vector<dsModel::Scenario>&&);

    const std::vector<dsModel::DeviceStateStatus>& statesStatus() const;
    void setStatesStatus(std::vector<dsModel::DeviceStateStatus>&&);
    const std::vector<dsModel::DeviceState>& states() const;
    void setStates(std::vector<dsModel::DeviceState>&&);

    /** LED indicates group membership of devices
     *
     * Some device indicate group membership with an LED, but the update is not
     * done by the device firmware, but needs to be handled by dss
     */
    void updateLedGroupColor();
    void updateZws205GroupColor();

    /** Somes device have application specififc default output mode */
    void updateOutputMode();
    void updateZws205OutputMode();

    void setConsumptionVisualizationEnabled(bool enabled);
    bool getConsumptionVisualizationEnabled();

    HardwareType hardwareType() const { return m_hardwareType; }
    void setHardwareType(HardwareType x) { m_hardwareType = x; } // For mocking purposes

    std::vector<Action<Device>> actions() const;

    // Check whether device is a legacy Awnings:
    // GR_KL210 with ApplicationType::Blinds
    // https://redmine.digitalstrom.org/issues/23064
    bool isLegacyAwning() const {
      return m_ActiveGroup == static_cast<int>(ApplicationType::Blinds) && m_hardwareType == HardwareType::GR_KL210;
    }
    bool isShadeDevice() const {
      return m_ActiveGroup == static_cast<int>(ApplicationType::Blinds) && m_hardwareType != HardwareType::GR_KL210;
    }

    void reregister() const;

    bool hasEmulatedHardwareOutputs() const { return m_emulatedHardwareOutputs; }
    void setButtonUsage(ButtonUsage usage);
    ButtonUsage getButtonUsage() const;
    std::string getButtonUsageAsString() const; // for JSON API
    bool hasPushbutton() const; // device type and number must be known for this
                                // to work correctly
  }; // Device

  std::ostream& operator<<(std::ostream& out, const Device&);

} // namespace dss

DS_REFLECT_STRUCT((dss), DeviceConfigBulk, (dsuid)(bank)(offset)(value));
DS_REFLECT_STRUCT((dss), Channel, (index)(id)(technicalName)(type));

#endif // DEVICE_H
