/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include "modelconst.h"

#include <ds/json.h>
#include <ds/log.h>
#include <ds/str.h>
#include <math.h>
#include <boost/range/algorithm/sort.hpp>

#include "base.h"
#include "exception.h"
#include "model-features.h"

namespace dss {
  double sensorValueToDouble(SensorType _sensorType, int _sensorValue) {
    switch (_sensorType) {
    case SensorType::Status:
    case SensorType::ActivePower:
    case SensorType::OutputCurrent:
    case SensorType::ActivePowerVA:
    case SensorType::WaterQuantity:
    case SensorType::WaterFlowRate:
      return static_cast<double>(_sensorValue);
    case SensorType::ElectricMeter:
      return roundDigits(_sensorValue * 0.01, 2);
    case SensorType::OutputCurrent16A:
      return static_cast<double>(_sensorValue * 4);
    case SensorType::TemperatureIndoors:
    case SensorType::TemperatureOutdoors:
    case SensorType::RoomTemperatureSetpoint:
    case SensorType::Temperature:
      return roundDigits(_sensorValue * 0.1 / 4 - 273.15 + 230.0, 3);
    case SensorType::BrightnessIndoors:
    case SensorType::BrightnessOutdoors:
    case SensorType::Brightness:
    case SensorType::CO2Concentration:
    case SensorType::COConcentration:
      return roundDigits(pow(10, _sensorValue / 800.0), 4);
    case SensorType::HumidityIndoors:
    case SensorType::HumidityOutdoors:
    case SensorType::Humidity:
      return roundDigits(_sensorValue * 0.1 / 4 , 3);
    case SensorType::WindSpeed:
    case SensorType::GustSpeed:
      return roundDigits(_sensorValue * 0.1 / 4, 3);
    case SensorType::SoundPressureLevel:
      return roundDigits(_sensorValue * 0.25 / 4, 3);
    case SensorType::RoomTemperatureControlVariable:
      return static_cast<double>(_sensorValue - 100);
    case SensorType::WindDirection:
    case SensorType::GustDirection:
      return roundDigits(_sensorValue * 0.5 / 4, 3);
    case SensorType::Precipitation:
      return roundDigits(_sensorValue * 0.1 / 4, 3);
    case SensorType::AirPressure:
      return roundDigits(_sensorValue / 4.0 + 200, 3);
    case SensorType::Reserved1:
    case SensorType::Reserved2:
    case SensorType::NotUsed:
    case SensorType::UnknownType:
      return static_cast<double>(_sensorValue);
    }
    return static_cast<double>(_sensorValue);
  } // sensorValueToDouble

  int doubleToSensorValue(SensorType _sensorType, double _sensorValue) {
    switch (_sensorType) {
    case SensorType::Status:
    case SensorType::ActivePower:
    case SensorType::OutputCurrent:
    case SensorType::ActivePowerVA:
    case SensorType::WaterQuantity:
    case SensorType::WaterFlowRate:
      if (_sensorValue < 0 || _sensorValue > 4095) {
        throw DSSException("Value must be in range [0..4095]");
      }
      return static_cast<int>(_sensorValue + 0.5);
    case SensorType::ElectricMeter:
      if (_sensorValue < 0 || _sensorValue > 40.95) {
        throw DSSException("Value must be in range [0..40.95]");
      }
      return static_cast<int>((_sensorValue + 0.005) / 0.01);
    case SensorType::OutputCurrent16A:
      if (_sensorValue < 0 || _sensorValue > 16380) {
        throw DSSException("Value must be in range [0..16380]");
      }
      return static_cast<int>((_sensorValue + 2) / 4);
    case SensorType::TemperatureIndoors:
    case SensorType::TemperatureOutdoors:
    case SensorType::RoomTemperatureSetpoint:
    case SensorType::Temperature:
      if (_sensorValue < -43.15 || _sensorValue > 59.225) {
        throw DSSException("Value must be in range [-43.15..59.225]");
      }
      return static_cast<int>((_sensorValue + 0.0125 + 273.15 - 230.0) * 4 / 0.1);
    case SensorType::BrightnessIndoors:
    case SensorType::BrightnessOutdoors:
    case SensorType::Brightness:
    case SensorType::CO2Concentration:
    case SensorType::COConcentration:
      if (_sensorValue < 1 || _sensorValue > 131446.795) {
        throw DSSException("Value must be in range [1..131446.795]");
      }
      return static_cast<int>(800 * log10(_sensorValue));
    case SensorType::HumidityIndoors:
    case SensorType::HumidityOutdoors:
    case SensorType::Humidity:
      if (_sensorValue < 0 || _sensorValue > 102.375) {
        throw DSSException("Value must be in range [0..102.375]");
      }
      return static_cast<int>((_sensorValue + 0.0125) * 4 / 0.1);
    case SensorType::WindSpeed:
    case SensorType::GustSpeed:
      if (_sensorValue < 0 || _sensorValue > 102.375) {
        throw DSSException("Value must be in range [0..102.375]");
      }
      return static_cast<int>((_sensorValue + 0.0125) * 4 / 0.1);
    case SensorType::SoundPressureLevel:
      if (_sensorValue < 0 || _sensorValue > 255.938) {
        throw DSSException("Value must be in range [0..255.938]");
      }
      return static_cast<int>((_sensorValue + 0.0125) * 4 / 0.25);
    case SensorType::RoomTemperatureControlVariable:
      if (_sensorValue < -100 || _sensorValue > 100) {
        throw DSSException("Value must be in range [-100..100]");
      }
      return static_cast<int>(_sensorValue + 0.5 + 100);
    case SensorType::WindDirection:
    case SensorType::GustDirection:
      if (_sensorValue < 0 || _sensorValue > 511.875) {
        throw DSSException("Value must be in range [0..511.875]");
      }
      return static_cast<int>((_sensorValue + 0.0625) * 8);
    case SensorType::Precipitation:
      if (_sensorValue < 0 || _sensorValue > 102.375) {
        throw DSSException("Value must be in range [0..102.375]");
      }
      return static_cast<int>((_sensorValue + 0.0125) * 4 / 0.1);
    case SensorType::AirPressure:
      if (_sensorValue < 200 || _sensorValue > 1223.75) {
        throw DSSException("Value must be in range [200..1223.75]");
      }
      return static_cast<int>((_sensorValue + 0.125 - 200) * 4);
    case SensorType::Reserved1:
    case SensorType::Reserved2:
    case SensorType::NotUsed:
    case SensorType::UnknownType:
      return static_cast<int>(_sensorValue + 0.5);
    }
    return static_cast<int>(_sensorValue + 0.5);
  } // doubleToSensorValue

  uint8_t sensorTypeToPrecision(SensorType _sensorType) {
    switch (_sensorType) {
      case SensorType::Status:
      case SensorType::RoomTemperatureControlVariable:
        return 1;
      case SensorType::ActivePower:
      case SensorType::OutputCurrent:
      case SensorType::ElectricMeter:
      case SensorType::TemperatureIndoors:
      case SensorType::TemperatureOutdoors:
      case SensorType::BrightnessIndoors:
      case SensorType::BrightnessOutdoors:
      case SensorType::HumidityIndoors:
      case SensorType::HumidityOutdoors:
      case SensorType::AirPressure:
      case SensorType::GustSpeed:
      case SensorType::GustDirection:
      case SensorType::WindSpeed:
      case SensorType::WindDirection:
      case SensorType::Precipitation:
      case SensorType::CO2Concentration:
      case SensorType::COConcentration:
      case SensorType::SoundPressureLevel:
      case SensorType::RoomTemperatureSetpoint:
      case SensorType::Reserved1:
      case SensorType::Reserved2:
      case SensorType::OutputCurrent16A:
      case SensorType::ActivePowerVA:
      case SensorType::Temperature: //no break
      case SensorType::Brightness: //no break
      case SensorType::Humidity: //no break
      case SensorType::WaterQuantity: //no break
      case SensorType::WaterFlowRate: //no break
      case SensorType::NotUsed:
      case SensorType::UnknownType:
        return 0;
    }
    return 0;
  }

  std::string sensorTypeName(SensorType _sensorType) {
    switch(_sensorType) {
      case SensorType::Status:
        return "Status";
      case SensorType::ActivePower:
        return "Active Power";
      case SensorType::OutputCurrent:
        return "Output Current";
      case SensorType::ElectricMeter:
        return "Electric Meter";
      case SensorType::OutputCurrent16A:
        return "Output Current Ex";
      case SensorType::ActivePowerVA:
        return "Active Power Ex";
      case SensorType::TemperatureIndoors:
        return "Temperature Indoors";
      case SensorType::TemperatureOutdoors:
        return "Temperature Outdoors";
      case SensorType::BrightnessIndoors:
        return "Brightness Indoors";
      case SensorType::BrightnessOutdoors:
        return "Brightness Outdoors";
      case SensorType::HumidityIndoors:
        return "Humidity Indoors";
      case SensorType::HumidityOutdoors:
        return "Humidity Outdoors";
      case SensorType::AirPressure:
        return "Air Pressure";
      case SensorType::GustSpeed:
        return "Gust Speed";
      case SensorType::GustDirection:
        return "Gust Direction";
      case SensorType::WindSpeed:
        return "Wind Speed";
      case SensorType::WindDirection:
        return "Wind Direction";
      case SensorType::SoundPressureLevel:
        return "Sound Pressure Level";
      case SensorType::Precipitation:
        return "Precipitation";
      case SensorType::RoomTemperatureSetpoint:
        return "Nominal Temperature";
      case SensorType::RoomTemperatureControlVariable:
        return "Temperature Control Value";
      case SensorType::CO2Concentration:
        return "Carbon Dioxide Concentration";
      case SensorType::COConcentration:
        return "Carbon Monoxide Concentration";
      case SensorType::Temperature:
        return "Tempeerature";
      case SensorType::Brightness:
        return "Brightness";
      case SensorType::Humidity:
        return "Humidity";
      case SensorType::WaterQuantity:
        return "Water Quantity";
      case SensorType::WaterFlowRate:
        return "Water Flow Rate";
      case SensorType::UnknownType:
        return "Unknown Type";
      case SensorType::Reserved1:
        return "Reserved1";
      case SensorType::Reserved2:
        return "Reserved2";
      case SensorType::NotUsed:
        return "NotUsed";
    }
    return "Invalid Type: " +  intToString(static_cast<int>(_sensorType));
  } // sensorTypeName

  std::ostream& operator<<(std::ostream& stream, SensorType type) {
    stream << dss::sensorTypeName(type);
    return stream;
  }

  std::optional<const char*> statusFieldTypeName(StatusFieldType x) {
    switch (x) {
      case StatusFieldType::MALFUNCTION:
        return "malfunction";
      case StatusFieldType::SERVICE:
        return "service";
    }
    return {};
  }

  std::optional<StatusFieldType> statusFieldTypeFromName(const std::string& x) {
    if (x == "malfunction") {
      return StatusFieldType::MALFUNCTION;
    } else if (x == "service") {
      return StatusFieldType::SERVICE;
    }
    return {};
  }

  std::ostream& operator<<(std::ostream& stream, StatusFieldType x) {
    return [&]()-> std::ostream& {
      if (auto&& name = statusFieldTypeName(x)) {
        return stream << *name;
      } else {
        return stream << "unknown";
      }
    }() << '(' << static_cast<int>(x) << ')';
  }

  std::optional<const char*> statusFieldValueName(StatusFieldValue x) {
    switch (x) {
      case StatusFieldValue::INACTIVE:
        return "inactive";
      case StatusFieldValue::ACTIVE:
        return "active";
    }
    return {};
  }

  std::optional<StatusFieldValue> statusFieldValueFromName(const std::string& x) {
    if (x == "inactive") {
      return StatusFieldValue::INACTIVE;
    } else if (x == "active") {
      return StatusFieldValue::ACTIVE;
    }
    return {};
  }

  std::ostream& operator<<(std::ostream& stream, StatusFieldValue x) {
    return [&]()-> std::ostream& {
      if (auto&& name = statusFieldValueName(x)) {
        return stream << *name;
      } else {
        return stream << "unknown";
      }
    }() << '(' << static_cast<int>(x) << ')';
  }

  std::ostream& operator<<(std::ostream& stream, BinaryInputType x) {
    return [&]()-> std::ostream& {
      switch (x) {
        case BinaryInputType::AppMode: return stream << "appMode";
        case BinaryInputType::Presence: return stream << "presence";
        case BinaryInputType::RoomBrightness: return stream << "roomBrightness";
        case BinaryInputType::PresenceInDarkness: return stream << "presenceInDarkness";
        case BinaryInputType::TwilightExternal: return stream << "twilightExternal";
        case BinaryInputType::Movement: return stream << "movement";
        case BinaryInputType::MovementInDarkness: return stream << "movementInDarkness";
        case BinaryInputType::SmokeDetector: return stream << "smokeDetector";
        case BinaryInputType::WindDetector: return stream << "windDetector";
        case BinaryInputType::RainDetector: return stream << "rainDetector";
        case BinaryInputType::SunRadiation: return stream << "sunRadiation";
        case BinaryInputType::RoomThermostat: return stream << "roomThermostat";
        case BinaryInputType::BatteryLow: return stream << "batteryLow";
        case BinaryInputType::WindowContact: return stream << "windowContact";
        case BinaryInputType::DoorContact: return stream << "doorContact";
        case BinaryInputType::WindowTilt: return stream << "windowTilt";
        case BinaryInputType::GarageDoorContact: return stream << "garageDoorContact";
        case BinaryInputType::SunProtection: return stream << "sunProtection";
        case BinaryInputType::FrostDetector: return stream << "frostDetector";
        case BinaryInputType::HeatingSystem: return stream << "heatingSystem";
        case BinaryInputType::HeatingSystemMode: return stream << "heatingSystemMode";
        case BinaryInputType::PowerUp: return stream << "powerUp";
        case BinaryInputType::Malfunction: return stream << "malfunction";
        case BinaryInputType::Service: return stream << "service";
      }
      return stream << "unknown";
    }() << '(' << static_cast<int>(x) << ')';
  }

  std::optional<StatusFieldType> statusFieldTypeForBinaryInputType(BinaryInputType x) {
    switch (x) {
      case BinaryInputType::AppMode:
      case BinaryInputType::Presence:
      case BinaryInputType::RoomBrightness:
      case BinaryInputType::PresenceInDarkness:
      case BinaryInputType::TwilightExternal:
      case BinaryInputType::Movement:
      case BinaryInputType::MovementInDarkness:
      case BinaryInputType::SmokeDetector:
      case BinaryInputType::WindDetector:
      case BinaryInputType::RainDetector:
      case BinaryInputType::SunRadiation:
      case BinaryInputType::RoomThermostat:
      case BinaryInputType::BatteryLow:
      case BinaryInputType::WindowContact:
      case BinaryInputType::DoorContact:
      case BinaryInputType::WindowTilt:
      case BinaryInputType::GarageDoorContact:
      case BinaryInputType::SunProtection:
      case BinaryInputType::FrostDetector:
      case BinaryInputType::HeatingSystem:
      case BinaryInputType::HeatingSystemMode:
      case BinaryInputType::PowerUp:
        return {};
      case BinaryInputType::Malfunction:
        return StatusFieldType::MALFUNCTION;
      case BinaryInputType::Service:
        return StatusFieldType::SERVICE;
    }
    return {};
  }

  ds::Enum<ApplicationType> dsEnum(ApplicationType) {
    static constexpr std::pair<ApplicationType, const char*> x[] = {
        {ApplicationType::None, "none"}, {ApplicationType::Lights, "lights"}, {ApplicationType::Blinds, "blinds"},
        {ApplicationType::Heating, "heating"}, {ApplicationType::Audio, "audio"}, {ApplicationType::Video, "video"},
        {ApplicationType::Security, "security"}, {ApplicationType::Access, "access"}, {ApplicationType::Joker, "joker"},
        {ApplicationType::Cooling, "cooling"}, {ApplicationType::Ventilation, "ventilation"},
        {ApplicationType::Window, "window"}, {ApplicationType::Recirculation, "recirculation"},
        {ApplicationType::ControlTemperature, "ControlTemperature"},
        {ApplicationType::ApartmentVentilation, "apartmentVentilation"},
        {ApplicationType::ApartmentRecirculation, "apartmentRecirculation"},
    };
    return {gsl::make_span(x), "dss::ApplicationType"};
  }
  std::ostream& operator<<(std::ostream& stream, ApplicationType x) { return ds::ostream(stream, x); }

  int getApplicationTypeColor(ApplicationType x) {
    switch (x) {
      case ApplicationType::Lights:
        return ColorIDYellow;
      case ApplicationType::Blinds:
      case ApplicationType::Awnings:
        return ColorIDGray;
      case ApplicationType::Heating:
      case ApplicationType::Cooling:
      case ApplicationType::Ventilation:
      case ApplicationType::Recirculation:
      case ApplicationType::Window:
      case ApplicationType::ControlTemperature:
      case ApplicationType::ApartmentVentilation:
      case ApplicationType::ApartmentRecirculation:
        return ColorIDBlue;
      case ApplicationType::Audio:
        return ColorIDCyan;
      case ApplicationType::Video:
        return ColorIDViolet;
      case ApplicationType::Joker:
        return ColorIDBlack;
      case ApplicationType::Security:
        return ColorIDRed;
      case ApplicationType::Access:
        return ColorIDGreen;
      case ApplicationType::None:
        return 0;
    }
    return 0;
  }

  RgbBitmask getApplicationTypeRgbBitmask(ApplicationType type) {
    switch (getApplicationTypeColor(type)) {
      case ColorIDYellow:
        return RgbBitmask::yellow;
      case ColorIDGray:
        return RgbBitmask::gray;
      case ColorIDBlue:
        return RgbBitmask::blue;
      case ColorIDCyan:
        return RgbBitmask::cyan;
      case ColorIDViolet:
        return RgbBitmask::magenta;
      case ColorIDRed:
        return RgbBitmask::red;
      case ColorIDGreen:
        return RgbBitmask::green;
      case ColorIDBlack:
        return RgbBitmask::gray;
      case ColorIDWhite:
        return RgbBitmask::gray;
    }
    return RgbBitmask::gray;
  }

  ds::Enum<BinaryInputId> dsEnum(BinaryInputId) {
    static constexpr std::pair<BinaryInputId, const char*> x[] = {
            {BinaryInputId::APP_MODE, "appMode"}
    };
    return {gsl::make_span(x), "dss::BinaryInputId"};
  }
  std::ostream& operator<<(std::ostream& stream, BinaryInputId x) { return ds::ostream(stream, x); }

  std::optional<const char*> heatingControlModeName(HeatingControlMode x) {
    switch (x) {
      case HeatingControlMode::OFF:
        return "off";
      case HeatingControlMode::PID:
        return "control";
      case HeatingControlMode::ZONE_FOLLOWER:
        return "zoneFollower";
      case HeatingControlMode::FIXED:
        return "fixed";
      case HeatingControlMode::MANUAL:
        return "manual";
    }
    return {};
  }

  std::optional<HeatingControlMode> heatingControlModeFromName(const std::string& x) {
    if (x == "off") {
      return HeatingControlMode::OFF;
    } else if (x == "control") {
      return HeatingControlMode::PID;
    } else if (x == "zoneFollower") {
      return HeatingControlMode::ZONE_FOLLOWER;
    } else if (x == "fixed") {
      return HeatingControlMode::FIXED;
    } else if (x == "manual") {
      return HeatingControlMode::MANUAL;
    }

    return {};
  }

  std::ostream& operator<<(std::ostream& stream, HeatingControlMode x) {
    return [&]()-> std::ostream& {
      if (auto&& name = heatingControlModeName(x)) {
        return stream << *name;
      } else {
        return stream << "unknown";
      }
    }() << '(' << static_cast<int>(x) << ')';
  }

  ds::Enum<ButtonInputMode> dsEnum(ButtonInputMode) {
    static constexpr std::pair<ButtonInputMode, const char*> x[] = {
        {ButtonInputMode::STANDARD, "standard"},                         //
        {ButtonInputMode::TURBO, "turbo"},                               //
        {ButtonInputMode::SWITCHED, "switched"},                         //
        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT1, "twoWayDwWithInput1"}, //
        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT2, "twoWayDwWithInput2"}, //
        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT3, "twoWayDwWithInput3"}, //
        {ButtonInputMode::TWO_WAY_DW_WITH_INPUT4, "twoWayDwWithInput4"}, //
        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT1, "twoWayUpWithInput1"}, //
        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT2, "twoWayUpWithInput2"}, //
        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT3, "twoWayUpWithInput3"}, //
        {ButtonInputMode::TWO_WAY_UP_WITH_INPUT4, "twoWayUpWithInput4"}, //
        {ButtonInputMode::TWO_WAY, "twoWay"},                            //
        {ButtonInputMode::ONE_WAY, "oneWay"},                            //
        {ButtonInputMode::AKM_STANDARD, "akmStandard"},                  //
        {ButtonInputMode::AKM_INVERTED, "akmInverted"},                  //
        {ButtonInputMode::AKM_ON_RISING_EDGE, "akmOnRisingEdge"},        //
        {ButtonInputMode::AKM_ON_FALLING_EDGE, "akmOnFallingEdge"},      //
        {ButtonInputMode::AKM_OFF_RISING_EDGE, "akmOffRisingEdge"},      //
        {ButtonInputMode::AKM_OFF_FALLING_EDGE, "akmOffFallingEdge"},    //
        {ButtonInputMode::AKM_RISING_EDGE, "akmRisingEdge"},             //
        {ButtonInputMode::AKM_FALLING_EDGE, "akmFallingEdge"},           //
        {ButtonInputMode::HEATING_PUSHBUTTON, "heatingPushbutton"},      //
        {ButtonInputMode::DEACTIVATED, "deactivated"},                   //
    };
    return {gsl::make_span(x), "dss::ButtonInputMode"};
  }

  std::ostream& operator<<(std::ostream& stream, ButtonInputMode x) { return ds::ostream(stream, x); }

  ds::Enum<ModelFeatureId> dsEnum(ModelFeatureId) {
    static constexpr std::pair<ModelFeatureId, const char*> x[] = {
        {ModelFeatureId::dontcare, "dontcare"},                                               //
        {ModelFeatureId::blink, "blink"},                                                     //
        {ModelFeatureId::identification, "identification"},                                   //
        {ModelFeatureId::ledauto, "ledauto"},                                                 //
        {ModelFeatureId::leddark, "leddark"},                                                 //
        {ModelFeatureId::transt, "transt"},                                                   //
        {ModelFeatureId::outmode, "outmode"},                                                 //
        {ModelFeatureId::outmodeswitch, "outmodeswitch"},                                     //
        {ModelFeatureId::outvalue8, "outvalue8"},                                             //
        {ModelFeatureId::pushbutton, "pushbutton"},                                           //
        {ModelFeatureId::pushbdevice, "pushbdevice"},                                         //
        {ModelFeatureId::pushbarea, "pushbarea"},                                             //
        {ModelFeatureId::pushbadvanced, "pushbadvanced"},                                     //
        {ModelFeatureId::pushbcombined, "pushbcombined"},                                     //
        {ModelFeatureId::shadeprops, "shadeprops"},                                           //
        {ModelFeatureId::shadeposition, "shadeposition"},                                     //
        {ModelFeatureId::motiontimefins, "motiontimefins"},                                   //
        {ModelFeatureId::optypeconfig, "optypeconfig"},                                       //
        {ModelFeatureId::shadebladeang, "shadebladeang"},                                     //
        {ModelFeatureId::highlevel, "highlevel"},                                             //
        {ModelFeatureId::consumption, "consumption"},                                         //
        {ModelFeatureId::jokerconfig, "jokerconfig"},                                         //
        {ModelFeatureId::akmsensor, "akmsensor"},                                             //
        {ModelFeatureId::akminput, "akminput"},                                               //
        {ModelFeatureId::akmdelay, "akmdelay"},                                               //
        {ModelFeatureId::twowayconfig, "twowayconfig"},                                       //
        {ModelFeatureId::outputchannels, "outputchannels"},                                   //
        {ModelFeatureId::heatinggroup, "heatinggroup"},                                       //
        {ModelFeatureId::heatingoutmode, "heatingoutmode"},                                   //
        {ModelFeatureId::heatingprops, "heatingprops"},                                       //
        {ModelFeatureId::pwmvalue, "pwmvalue"},                                               //
        {ModelFeatureId::valvetype, "valvetype"},                                             //
        {ModelFeatureId::extradimmer, "extradimmer"},                                         //
        {ModelFeatureId::umvrelay, "umvrelay"},                                               //
        {ModelFeatureId::blinkconfig, "blinkconfig"},                                         //
        {ModelFeatureId::impulseconfig, "impulseconfig"},                                     //
        {ModelFeatureId::umroutmode, "umroutmode"},                                           //
        {ModelFeatureId::pushbsensor, "pushbsensor"},                                         //
        {ModelFeatureId::locationconfig, "locationconfig"},                                   //
        {ModelFeatureId::windprotectionconfigawning, "windprotectionconfigawning"},           //
        {ModelFeatureId::windprotectionconfigblind, "windprotectionconfigblind"},             //
        {ModelFeatureId::outmodegeneric, "outmodegeneric"},                                   //
        {ModelFeatureId::outconfigswitch, "outconfigswitch"},                                 //
        {ModelFeatureId::temperatureoffset, "temperatureoffset"},                             //
        {ModelFeatureId::apartmentapplication, "apartmentapplication"},                       //
        {ModelFeatureId::ftwtempcontrolventilationselect, "ftwtempcontrolventilationselect"}, //
        {ModelFeatureId::ftwdisplaysettings, "ftwdisplaysettings"},                           //
        {ModelFeatureId::ftwbacklighttimeout, "ftwbacklighttimeout"},                         //
        {ModelFeatureId::ventconfig, "ventconfig"},                                           //
        {ModelFeatureId::fcu, "fcu"},                                                         //
        {ModelFeatureId::pushbdisabled, "pushbdisabled"},                                     //
        {ModelFeatureId::consumptioneventled, "consumptioneventled"},                         //
        {ModelFeatureId::consumptiontimer, "consumptiontimer"},                               //
        {ModelFeatureId::jokertempcontrol, "jokertempcontrol"},                               //
        {ModelFeatureId::dimtimeconfig, "dimtimeconfig"},                                     //
        {ModelFeatureId::outmodeauto, "outmodeauto"},                                         //
        {ModelFeatureId::dimmodeconfig, "dimmodeconfig"},                                     //
        {ModelFeatureId::setumr200config, "setumr200config"},                                 //
        {ModelFeatureId::extendedvalvetypes, "extendedvalvetypes"},                           //
        {ModelFeatureId::customtransitiontime, "customtransitiontime"},                       //
        {ModelFeatureId::outmodetempcontrol, "outmodetempcontrol"},                           //
        {ModelFeatureId::outmodeenoceanvalve, "outmodeenoceanvalve"},                         //
        {ModelFeatureId::operationlock, "operationlock"},                                     //
        {ModelFeatureId::grkl387workaround, "grkl387workaround"}
    };
    return {gsl::make_span(x), "dss::ModelFeatureId"};
  }
  std::ostream& operator<<(std::ostream& stream, ModelFeatureId x) { return ds::ostream(stream, x); }

  ds::Enum<BasicScene> dsEnum(BasicScene) {
    static constexpr std::pair<BasicScene, const char*> x[] = {
        {BasicScene::OFF, "off"}, {BasicScene::LEVEL_1, "level1"}, {BasicScene::LEVEL_2, "level2"},
        {BasicScene::LEVEL_3, "level3"}, {BasicScene::LEVEL_4, "level4"}, {BasicScene::BOOST, "boost"},
    };
    return {gsl::make_span(x), "BasicScene"};
  }
  std::ostream& operator<<(std::ostream& stream, BasicScene x) { return ds::ostream(stream, x); }

  std::string serializeBasicScenes(const std::vector<BasicScene>& basicScenes) {
    std::string result;
    for (const auto& basicScene : basicScenes) {
      if (!result.empty()) {
        result.append(",");
      }
      result.append(ds::str(static_cast<int>(basicScene)));
    }
    return result;
  }

  std::vector<BasicScene> parseBasicScenesJson(const std::string& json) {
    std::vector<BasicScene> result;
    for (const auto& number : ds::fromJson<std::vector<int>>(json)) {
      result.push_back(ds::requireEnum<BasicScene>(number));
    }
    // Normalize vector to set.
    // Use std::set / boost::container::flat_set to unsure this invariant?
    boost::range::sort(result, [](BasicScene a, BasicScene b) { return static_cast<int>(a) < static_cast<int>(b); });
    result.erase(std::unique(result.begin(), result.end()), result.end());
    return result;
  }

  std::vector<BasicScene> parseBasicScenes(const std::string& x) { return parseBasicScenesJson(ds::str("[", x, "]")); }

  ds::Enum<OutputMode> dsEnum(OutputMode) {
    static constexpr std::pair<OutputMode, const char*> x[] = {
        {OutputMode::Disabled, "Disabled"},
        {OutputMode::Diagnostic, "Diagnostic"},
        {OutputMode::KM_Switch, "KM_Switch"},
        {OutputMode::RMSDimmer, "RMSDimmer"},
        {OutputMode::RMSDimmerCurve, "RMSDimmerCurve"},
        {OutputMode::LeadingEdgeDimmer, "LeadingEdgeDimmer"},
        {OutputMode::LeadingEdgeDimmerCurve, "LeadingEdgeDimmerCurve"},
        {OutputMode::TrailingEdgeDimmer, "TrailingEdgeDimmer"},
        {OutputMode::TrailingEdgeDimmerCurve, "TrailingEdgeDimmerCurve"},
        {OutputMode::KM_PWM, "KM_PWM"},
        {OutputMode::KM_PWMCurve, "KM_PWMCurve"},
        {OutputMode::KM_PassingMake, "KM_PassingMake"},
        {OutputMode::KM_PassingBreak, "KM_PassingBreak"},
        {OutputMode::Auto, "Auto"},
        {OutputMode::Umsteuerrelais, "Umsteuerrelais"},
        {OutputMode::Position, "Position"},
        {OutputMode::TwoStageSwitch, "TwoStageSwitch"},
        {OutputMode::Switch2Pol, "Switch2Pol"},
        {OutputMode::PassingBreak2Pol, "PassingBreak2Pol"},
        {OutputMode::PassingMake2Pol, "PassingMake2Pol"},
        {OutputMode::ThreeStageSwitch, "ThreeStageSwitch"},
        {OutputMode::KL_Switch, "KL_Switch"},
        {OutputMode::KL_Passing, "KL_Passing"},
        {OutputMode::KL_PowerSave, "KL_PowerSave"},
        {OutputMode::PositionUncalibrated, "PositionUncalibrated"},
        {OutputMode::SwitchBipolar, "SwitchBipolar"},
        {OutputMode::UMV_ZeroTen, "UMV_ZeroTen"},
        {OutputMode::UMV_ZeroTenCurve, "UMV_ZeroTenCurve"},
        {OutputMode::UMV_OneTen, "UMV_OneTen"},
        {OutputMode::UMV_OneTenCurve, "UMV_OneTenCurve"},
        {OutputMode::RegulationPWM, "RegulationPWM"},
        {OutputMode::RegulationSwitch, "RegulationSwitch"},
        {OutputMode::UMR_TwoStageSwitch, "UMR_TwoStageSwitch"},
        {OutputMode::UMR_ThreeStageSwitch, "UMR_ThreeStageSwitch"},
        {OutputMode::UMR_ParallelSwitch, "UMR_ParallelSwitch"},
        {OutputMode::GenericActive, "GenericActive"},
    };
    return {gsl::make_span(x), "dss::OutputMode"};
  }
  std::ostream& operator<<(std::ostream& stream, OutputMode x) {
    return ds::ostream(stream, x);
  }

} // namespace dss
