/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include <vector>

#include <ds-model/model-fwd.h>
#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>

#include "devicecontainer.h"
#include "ds-model.h"
#include "nonaddressablemodelitem.h"
#include "physicalmodelitem.h"
#include "businterface.h"

namespace dss {
  class Apartment;
  class DSMeter;
  class Group;
  class Set;
  class DeviceReference;
  class PropertyNode;
  typedef boost::shared_ptr<PropertyNode> PropertyNodePtr;
  class DateTime;

  struct TemperatureConfig {
    static const std::array<double, 16> DEFAULT_TARGET_TEMPERATURES;
    static const std::array<double, 16> DEFAULT_FIXED_VALUES;

    TemperatureConfig();
    bool isEqual(const ZoneHeatingConfigSpec_t& config, const ZoneHeatingOperationModeSpec_t& operationMode) const;

    void parseTargetTemperatures(const std::string& jsonObject);
    void parseFixedValues(const std::string& jsonObject);
    void parseControlMode(const std::string& jsonObject);
    void parseFollowerMode(const std::string& jsonObject);
    void parseManualMode(const std::string& jsonObject);
    void parseFields(const std::string& jsonObject);

    bool m_setPointsValid;
    bool m_propertiesValid;
    HeatingControlMode m_mode;
    uint16_t m_Kp;
    uint8_t  m_Ts;
    uint16_t m_Ti;
    uint16_t m_Kd;
    int16_t  m_Imin;
    int16_t  m_Imax;
    uint8_t  m_Ymin;
    uint8_t  m_Ymax;
    uint8_t  m_AntiWindUp;
    uint8_t  m_KeepFloorWarm;
    int m_HeatingControlState;     // Control state: 0=internal; 1=external; 2=exbackup; 3=emergency
    int m_HeatingMasterZone;       // only used for mode 2
    int m_CtrlOffset;              // only used for mode 2
    uint8_t  m_EmergencyValue;
    int m_ManualValue;             // only used for mode 4
    std::array<double, 16> targetTemperatures; // temperature set points used in mode 1
    std::array<double, 16> fixedValues;        // control values used in mode 3

    void setSpec(const ZoneHeatingConfigSpec spec);
    ZoneHeatingConfigSpec spec() const;
  };

  struct ZoneHeatingStatus {
    std::optional<double> nominalValue; // only used for mode 1
    DateTime nominalValueTS{DateTime::NullDate};
    std::optional<double> controlValue;
    DateTime controlValueTS{DateTime::NullDate};
  };

  struct ZoneSensorStatus {
    std::optional<double> temperatureValue;
    DateTime temperatureValueTS{DateTime::NullDate};
    std::optional<double> humidityValue;
    DateTime humidityValueTS{DateTime::NullDate};
    std::optional<double> brightnessValue;
    DateTime brightnessValueTS{DateTime::NullDate};
    std::optional<double> co2ConcentrationValue;
    DateTime co2ConcentrationValueTS{DateTime::NullDate};
  };

  typedef struct {
    dsuid_t m_DSUID;
    SensorType m_sensorType;
    int m_sensorIndex;
  } MainZoneSensor_t;
    /** Represents a Zone.
   * A Zone houses multiple devices. It can span over multiple dsMeters.
   */
  class Zone : public DeviceContainer,
               public NonAddressableModelItem,
               public PhysicalModelItem,
               public boost::noncopyable {
  private:
    int m_ZoneID;
    std::vector<DeviceReference> m_Devices;
    std::vector<boost::weak_ptr<const DSMeter> > m_DSMeters;
    std::vector<boost::shared_ptr<Group> > m_Groups;
    std::vector<MainZoneSensor_t> m_MainSensors;
    TemperatureConfig m_temperatureConfig;
    ZoneHeatingStatus m_HeatingStatus;
    int m_HeatingOperationMode;
    ZoneSensorStatus m_SensorStatus;
    Apartment* m_pApartment;
    PropertyNodePtr m_pPropertyNode;
    boost::weak_ptr<const DSMeter> m_temperatureController;
    DateTime m_temperatureControllerTs;
    int m_floorID;

  public:
    Zone(const int _id, Apartment* _pApartment);
    ~Zone() override;

    bool isInDsModel() const;
    dsModel::ZoneId id() const;
    std::vector<dsModel::SubmoduleId> submoduleIds() const;

    Set getDevices() const override;
    std::vector<DeviceReference> devicesReferences() const { return m_Devices; }

    Apartment& getApartment() { return *m_pApartment; }
    ModelMaintenance* tryGetModelMaintenance();

    void setName(const std::string& name) override;

    /** Adds the Zone to a dsMeter. */
    void addToDSMeter(boost::shared_ptr<DSMeter> _dsMeter);
    /** Removes the Zone from a dsMeter. */
    void removeFromDSMeter(boost::shared_ptr<DSMeter> _dsMeter);

    /** Adds a device to the zone.
     * This will permanently add the device to the zone.
     */
    void addDevice(DeviceReference device);

    /** Removes a device from the zone.
     * This will permanently remove the device from the zone.
     */
    void removeDevice(DeviceReference device);

    /** Returns the group with the name \a _name */
    boost::shared_ptr<Group> getGroup(const std::string& _name) const;
    boost::weak_ptr<Group> tryGetGroup(const int id) const;
    /// WARNING: returns nullptr if group does not exist.
    boost::shared_ptr<Group> getGroup(const int id) const;

    /** Adds a group to the zone */
    void addGroup(boost::shared_ptr<Group> _group);
    /** Removes a group from the zone */
    void removeGroup(boost::shared_ptr<Group> _group);

    /** Returns the zones id */
    int getID() const;
    /** Sets the zones id */
    void setZoneID(const int _value);

    bool registeredOnDSMeter(boost::shared_ptr<const DSMeter> _dsMeter) const;
    bool isRegisteredOnAnyMeter() const;
    std::vector<boost::weak_ptr<const DSMeter> > getDSMeters() { return m_DSMeters; }

    void nextScene(const callOrigin_t _origin, const SceneAccessCategory _category) override;
    void previousScene(const callOrigin_t _origin, const SceneAccessCategory _category) override;

    void publishToPropertyTree();
    void removeFromPropertyTree();
    PropertyNodePtr getPropertyNode() {  return m_pPropertyNode; }

    unsigned long getPowerConsumption() override;

    /** Returns a vector of groups present on the zone. */
    std::vector<boost::shared_ptr<Group> > getGroups() { return m_Groups; }

    /** Returns the heating  and sensor properties */
    const TemperatureConfig& temperatureConfig() const;
    const ZoneHeatingStatus& getHeatingStatus() const;
    const ZoneSensorStatus& getSensorStatus() const;

    /** Set heating properties and runtime values */
    void setHeatingConfigSpec(const ZoneHeatingConfigSpec&);
    ZoneHeatingConfigSpec getHeatingConfigSpec();
    void setHeatingControlState(int _ctrlState);
    void setHeatingControlOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues);
    void setHeatingFixedOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues);
    void setHeatingOperationMode(const ZoneHeatingOperationModeSpec_t& operationModeValues);
    ZoneHeatingOperationModeSpec_t getHeatingControlOperationModeValues() const;
    ZoneHeatingOperationModeSpec_t getHeatingFixedOperationModeValues() const;
    ZoneHeatingOperationModeSpec_t getHeatingOperationModeValues() const;
    void setHeatingOperationMode(int _operationMode);
    int getHeatingOperationMode() const;
    void setTemperature(double value, const DateTime&);
    void resetTemperature();
    void setNominalValue(double value, const DateTime&);
    void resetNominalValue();
    void setControlValue(double value, const DateTime&);
    void resetControlValue();
    void setHumidityValue(double value, const DateTime&);
    void resetHumidityValue();
    void setBrightnessValue(double value, const DateTime&);
    void resetBrightnessValue();
    void setCO2ConcentrationValue(double value, const DateTime&);
    void resetCO2ConcentrationValue();

    void setSensor(const Device &_device, SensorType _sensorType);
    void setSensor(const MainZoneSensor_t &_mainZoneSensor);
    void resetSensor(SensorType _sensorType);
    bool isSensorAssigned(SensorType _sensorType) const;
    const std::vector<MainZoneSensor_t>& getAssignedSensors() const { return m_MainSensors; }

    std::vector<SensorType> getUnassignedSensorTypes() const;
    std::vector<SensorType> getAssignedSensorTypes(const Device &_device) const;

    boost::shared_ptr<Device> getAssignedSensorDevice(SensorType _sensorType) const;
    bool isZoneSensor(const Device &_device, SensorType _sensorType) const;

    bool isDeviceZoneMember(const DeviceReference& _device) const;
    void removeInvalidZoneSensors();

    /** Set heating properties */
    void setTemperatureConfig(TemperatureConfig config);
    bool isTemperatureConfigValid() const;
    bool isHeatingSetpointsValid() const;
    void setHeatingSetpoints(const TemperatureConfig& config);
    void setHeatingFixValues(const TemperatureConfig& config);

    boost::shared_ptr<const DSMeter> temperatureController() const { return m_temperatureController.lock(); }
    DateTime temperatureControllerTs() const { return m_temperatureControllerTs; }
    void setTemperatureController(const boost::shared_ptr<const DSMeter>& meter);

    /// applications present in this zone
    std::vector<dsModel::ZoneApplicationType> applications() const;

    int getFloorID() const;
    void setFloorID(int id);

  protected:
    std::vector<boost::shared_ptr<AddressableModelItem> > splitIntoAddressableItems() override;
    bool isAllowedSensorType(SensorType _sensorType);
    void raiseOperationModeChanged();

  private:
    void sanitizeHeatingProperties();
  }; // Zone

  std::ostream& operator<<(std::ostream&, const Zone&);

} // namespace dss

DS_REFLECT_STRUCT((dss), TemperatureConfig,
                  (m_mode)(m_Kp)(m_Ts)(m_Ti)(m_Kd)(m_Imin)(m_Imax)(m_Ymin)(m_Ymax)(m_AntiWindUp)(m_KeepFloorWarm)(
                      m_HeatingControlState)(m_HeatingMasterZone)(m_CtrlOffset)(m_EmergencyValue)(m_ManualValue)(
                      targetTemperatures)(fixedValues));
DS_REFLECT_STRUCT((dss), ZoneHeatingStatus, (nominalValue)(nominalValueTS)(controlValue)(controlValueTS));
