/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include <ds/catch/catch.h>
#include <ds/log.h>
#include "businterface-mock.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/modelmaintenance-mock.h"
#include "scripting/jshandler.h"
#include "scripting/jsmodel.h"
#include "taskprocessor-mock.h"

static ds::LogChannel dsLogChannel("modelJsTest");
static const char* TAGS = "[dss][dssModelJs]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid dsuid1{"000000000000000000000000000000000a"};
static const Dsuid dsuid2{"0000000000000000000000000000000001"};
static const Dsuid dsuid3{"0000000000000000000000000000000002"};

TEST_CASE("bulkDeviceConfig", TAGS) {
  Apartment apt(nullptr);
  BusInterfaceMock busInterface;
  auto& deviceBusInterface = busInterface.device;
  EventSinkMock eventSink;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);

  boost::shared_ptr<Device> dev1 = apt.allocateDevice(dsuid1);
  boost::shared_ptr<Device> dev2 = apt.allocateDevice(dsuid2);
  boost::shared_ptr<Device> dev3 = apt.allocateDevice(dsuid3);
  boost::scoped_ptr<ScriptEnvironment> env(new ScriptEnvironment());
  env->initialize();
  ModelScriptContextExtension* ext = new ModelScriptContextExtension(apt);
  env->addExtension(ext);
  boost::scoped_ptr<ScriptContext> ctx(env->getContext());
  ModelMaintenanceMock modelMaintenance;
  apt.setModelMaintenance(&modelMaintenance);
  TaskProcessorMock taskProcessor;
  ALLOW_CALL(modelMaintenance, getTaskProcessor()).LR_RETURN(taskProcessor);
  ALLOW_CALL(taskProcessor, addEvent(_)).SIDE_EFFECT(_1->run());

  SECTION("works for two devices") {
    sequence seq1;
    Event event1;
    Event event2;
    REQUIRE_CALL(deviceBusInterface, setDeviceConfig(_, 42, 4, 231)).IN_SEQUENCE(seq1);
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event1 = *_1).IN_SEQUENCE(seq1);
    REQUIRE_CALL(deviceBusInterface, setDeviceConfig(_, 43, 4, 231)).IN_SEQUENCE(seq1);
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event2 = *_1).IN_SEQUENCE(seq1);
    std::string result = ctx->evaluate<std::string>(
        R"(setDeviceConfigBulk('[{"dsuid":"000000000000000000000000000000000a", "bank": 42, "offset": 4, "value": 231}, \
                                 {"dsuid":"0000000000000000000000000000000001", "bank": 43, "offset": 4, "value": 231}]');)");
    CHECK(result == event1.getPropertyByName("taskId"));
    CHECK(event1.getRaisedAtDevice());
    CHECK(event1.getRaisedAtDevice()->getDSID() == dsuid1);
    CHECK(strToInt(event1.getPropertyByName("bank")) == 42);
    CHECK(strToInt(event1.getPropertyByName("offset")) == 4);
    CHECK(strToInt(event1.getPropertyByName("value")) == 231);
    CHECK(strToInt(event1.getPropertyByName("count")) == 1);
    CHECK(strToInt(event1.getPropertyByName("total")) == 2);
    CHECK(result == event2.getPropertyByName("taskId"));
    CHECK(event2.getRaisedAtDevice());
    CHECK(event2.getRaisedAtDevice()->getDSID() == dsuid2);
    CHECK(strToInt(event2.getPropertyByName("bank")) == 43);
    CHECK(strToInt(event2.getPropertyByName("offset")) == 4);
    CHECK(strToInt(event2.getPropertyByName("value")) == 231);
    CHECK(strToInt(event2.getPropertyByName("count")) == 2);
    CHECK(strToInt(event2.getPropertyByName("total")) == 2);
  }

  SECTION("works when one device is skipped") {
    sequence seq1;
    Event event1;
    Event event2;
    REQUIRE_CALL(deviceBusInterface, setDeviceConfig(_, 42, 4, 231)).IN_SEQUENCE(seq1);
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event1 = *_1).IN_SEQUENCE(seq1);
    REQUIRE_CALL(deviceBusInterface, setDeviceConfig(_, 43, 4, 231)).IN_SEQUENCE(seq1);
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event2 = *_1).IN_SEQUENCE(seq1);
    std::string result = ctx->evaluate<std::string>(
        R"(setDeviceConfigBulk('[{"dsuid":"000000000000000000000000000000000a", "bank": 42, "offset": 4, "value": 231}, \
                                 {"dsuid":"0000000000000000000000000000000003", "bank": 43, "offset": 4, "value": 231}, \
                                 {"dsuid":"0000000000000000000000000000000001", "bank": 43, "offset": 4, "value": 231}]');)");
    CHECK(result == event1.getPropertyByName("taskId"));
    CHECK(event1.getRaisedAtDevice());
    CHECK(event1.getRaisedAtDevice()->getDSID() == dsuid1);
    CHECK(strToInt(event1.getPropertyByName("bank")) == 42);
    CHECK(strToInt(event1.getPropertyByName("offset")) == 4);
    CHECK(strToInt(event1.getPropertyByName("value")) == 231);
    CHECK(strToInt(event1.getPropertyByName("count")) == 1);
    CHECK(strToInt(event1.getPropertyByName("total")) == 3);
    CHECK(result == event2.getPropertyByName("taskId"));
    CHECK(event2.getRaisedAtDevice());
    CHECK(event2.getRaisedAtDevice()->getDSID() == dsuid2);
    CHECK(strToInt(event2.getPropertyByName("bank")) == 43);
    CHECK(strToInt(event2.getPropertyByName("offset")) == 4);
    CHECK(strToInt(event2.getPropertyByName("value")) == 231);
    CHECK(strToInt(event2.getPropertyByName("count")) == 3);
    CHECK(strToInt(event2.getPropertyByName("total")) == 3);
  }

  SECTION("still calls event if all devices are skipped") {
    Event event1;
    FORBID_CALL(deviceBusInterface, setDeviceConfig(_, _, _, _));
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event1 = *_1);
    std::string result = ctx->evaluate<std::string>(
        R"(setDeviceConfigBulk('[{"dsuid":"0000000000000000000000000000000003", "bank": 43, "offset": 4, "value": 231}]');)");
    CHECK(result == event1.getPropertyByName("taskId"));
    CHECK(!event1.getRaisedAtDevice());
    CHECK(event1.getPropertyByName("bank") == "");
    CHECK(event1.getPropertyByName("offset") == "");
    CHECK(event1.getPropertyByName("value") == "");
    CHECK(strToInt(event1.getPropertyByName("count")) == 1);
    CHECK(strToInt(event1.getPropertyByName("total")) == 1);
  }

  SECTION("calls event on empty list") {
    Event event1;
    FORBID_CALL(deviceBusInterface, setDeviceConfig(_, _, _, _));
    REQUIRE_CALL(eventSink, pushEvent(_)).LR_SIDE_EFFECT(event1 = *_1);
    std::string result = ctx->evaluate<std::string>(
        R"(setDeviceConfigBulk('[]');)");
    CHECK(result == event1.getPropertyByName("taskId"));
    CHECK(!event1.getRaisedAtDevice());
    CHECK(event1.getPropertyByName("bank") == "");
    CHECK(event1.getPropertyByName("offset") == "");
    CHECK(event1.getPropertyByName("value") == "");
    CHECK(strToInt(event1.getPropertyByName("count")) == 0);
    CHECK(strToInt(event1.getPropertyByName("total")) == 0);
  }
}

} // namespace
