## Style guide

## C++ best practices

Refer to [C++ Core Guidelines](https://github.com/isocpp/CppCoreGuidelines/blob/master/CppCoreGuidelines.md)
for a set of vendor neutral guidelines for using C++ well.

Most rules are possible to enforce and even fix by `clang-tidy`.

Benefits:

* Avoid bikeshedding
* You can hurt yourself on this; do it this way to be safe

If we disagree with some rules, we shall document those rules we disagree with here.

Refs:

* [CppCon 2017: Kate Gregory “10 Core Guidelines You Need to Start Using Now](https://www.youtube.com/watch?v=XkDEzfpdcSg)

## Design philosophy

### Value Types vs. Resource Types

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#value-types-vs-resource-types)

There are two kinds of types: values and resources.
Value types are simple data structures;
they serve no purpose except to represent pure data.
Resource types represent live objects with state and behavior,
and often represent resources external to the program.

* Value types make sense to copy (though they don't necessarily have copy constructors). Resource types are not copyable.
* Value types always have move constructors (and sometimes copy constructors). Resource types are not movable; if ownership transfer is needed, the resource must be allocated on the heap.
* Value types almost always have implicit destructors. Resource types may have an explicit destructor.
* Value types should only be compared by value, not identity. Resource types can only be compared by identity.
* Value types make sense to serialize. Resource types fundamentally cannot be serialized.
* Value types rarely use inheritance and never have virtual methods. Resource types commonly do.
* Value types generally use templates for polymorphism. Resource types generally use virtual methods / abstract interfaces.
* You might even say that value types are used in functional programming style while resource types are used in object-oriented style.

### RAII (Resource Acquisition Is Initialization)

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#raii-resource-acquisition-is-initialization)

Whenever it is the case that "this block of code cannot exit cleanly without performing operation X",
then X must be performed in a destructor,
so that X will happen regardless of how the block is exited (including by exception).

Use the macros DS_DEFER,
to easily specify some code that must be executed on exit from the current scope,
without the need to define a whole class with a destructor.

Be careful when writing complicated destructors.
If a destructor performs multiple cleanup actions,
you generally need to make sure that the latter actions occur even if the former ones throw an exception.
For this reason, a destructor should generally perform no more than one cleanup action.
If you need to clean up multiple things,
have you class contain multiple members representing the different things that need cleanup,
each with its own destructor.
This way, if one member's destructor throws, the others still run.

### No Singletons

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#no-singletons)

A "singleton" is any mutable object or value that is globally accessible.
"Globally accessible" means that the object is declared as a global variable or static member variable,
or that the object can be found by following pointers from such variables.

Never use singletons.
Singletons cause invisible and unexpected dependencies between components of your software that appear unrelated.
Worse, the assumption that "there should only be one of this object per process" is almost always wrong,
but its wrongness only becomes apparent after so much code uses the singleton that it is infeasible to change.
Singleton interfaces ofter turn into unusable monstrosities in an attempt to work around the fact
that they should never have been a singleton in the first place.

See ["Singletons Considered Harmful"](http://www.object-oriented-security.org/lets-argue/singletons) for a complete discussion.

### Threads vs. Event Loops

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#threads-vs-event-loops)

Threads are hard, and synchronization between threads is slow.
Even "lock-free" data structures usually require atomic operations,
which are costly, and such algorithms are notoriously difficult to get right.
Fine-grained synchronization will therefore be expensive at best and highly unstable at worst.

We prefer event loop concurrency.
In this model, each event callback is effectively a transaction;
it does not need to worry about concurrent modification within the body of the function.

From [Thinking Outside the Synchronisation Quadrant - Kevlin Henney](://www.youtube.com/watch?v=2yXtZ8x7TXw)

Ask programmers what comes to mind when you say concurrency and most are likely to say threads.
Ask what comes to mind when you say threads and most are likely to say locks or synchronisation.
These assumptions are so deeply held that they define and constrain
how programmers are taught and think about concurrency:
thread safety is almost synonymous with the avoidance of race conditions
and the guarded protection of mutable state.
But this is only one quadrant of four possibilities,
a quadrant diagram partitioned by mutable–immutable along one axis
and shared–unshared along another.

#### Prefer asio (based) library as event loop for asynchronous IO

Asio will become part of `c++` standard.
All asio based libraries easily share one event loop thread.

Consider asio based coroutines as good compromise how to avoid threads and callback hell.

### Lazy input validation

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#lazy-input-validation)

As we all know, you should always validate your input.

But, when should you validate it? There are two plausible answers:

* Upfront, on receipt.
* Lazily, on use.

Upfront validation occasionally makes sense for the purpose of easier debugging of problems: if an error is reported earlier, it's easier to find where it came from.

However, upfront validation has some big problems.

* It encourages people to skip validation at time of use, on the assumption that it was already validated earlier.
   This is dangerous, as it entails a non-local assumption.
   E.g. are you really sure that there is no way to insert data into your database without having validated it?
   Are you really sure that the data hasn't been corrupted?
   Are you really sure that your code will never be called in a new situation where validation hasn't happened?
   Are you sure the data cannot have been modified between validation and use?
   In practice, you should be validating your input at time of use
   _even if_ you know it has already been checked previously.

* The biggest problem: Upfront validation tends not to match actual usage,
   because the validation site is far away from the usage site.
   Over time, as the usage code changes, the validator can easily get out-of-sync.
   Note that this could mean the code itself is out-of-sync,
   or it could be that running servers are out-of-sync, because they have different update schedules.
   Or, the validator may be written with incorrect assumptions in the first place.
   The consequences of this can be severe.
   Protocol Buffers' concept of "required fields" is essentially an upfront validation check
   that [has been responsible for outages of Google Search, GMail, and others](https://capnproto.org/faq.html#how-do-i-make-a-field-required-like-in-protocol-buffers).

We recommend, therefore, that validation occur at time of use.
Code should be written to be tolerant of validation failures.
For example, most code dealing with UTF-8 text should treat it as a blob of bytes,
not worrying about invalid byte sequences.
When you actually need to decode the code points
-- such as to display them --
you should do something reasonable with invalid sequences
-- such as display the Unicode replacement character.

With that said, when storing data in a database long-term,
it can make sense to perform an additional validation check at time of storage,
in order to more directly notify the caller that their input was invalid.
This validation should be considered optional,
since the data will be validated again when it is read from storage and used.


### Exceptions

From [capnproto style guide](https://github.com/capnproto/capnproto/blob/master/style-guide.md#exceptions)

An exception represents something that "should never happen", assuming everything is working as expected.
Of course, things that "should never happen" in fact happen all the time.
But, a program should never be written in such a way that it _expects_ an exception under normal circumstances.

Put another way, exceptions are a way to achieve _fault tolerance_.
Throwing an exception is a less-disruptive alternative to aborting the process.
Exceptions are a _logistical_ construct, as opposed to a semantic one: an exception should never be part of your "business logic".

For example, exceptions may indicate conditions like:

* Logistics of software development:
  * There is a bug in the code.
  * The requested method is not implemented.
* Logistics of software usage:
  * There is an error in the program's configuration.
  * The input is invalid.
* Logistics of distributed systems:
  * A network connection was reset.
  * An optimistic transaction was aborted due to concurrent modification.
* Logistics of physical computation:
  * The system's resources are exhausted (e.g. out of memory, out of disk space).
  * The system is overloaded and must reject some requests to avoid long queues.

#### Business logic should never catch

If you find that callers of your interface need to catch
and handle certain kinds of exceptions in order to operate correctly,
then you must change your interface (or overload it)
such that those conditions can be handled without an exception ever being thrown.
For example, if you have a method `unique_ptr<File> open(string_span name)` that opens a file,
you may also want to offer `optional<unique_ptr<File>> openIfExists(string_span name)`
that returns null rather than throwing an exception if the file is not found.
But you should probably keep `open()` as well,
for the convenience of the common case where the caller will just throw an exception anyway.

Note that with this exception philosophy,
Java-style "checked exceptions"
(exceptions which are explicitly declared to be thrown by an interface)
make no sense.

#### How to handle an exception

In framework and logistical code, you may catch exceptions and try to handle them.
Given the nature of exceptions, though,
there are only a few things that are reasonable to do when receiving an exception:

* On network disconnect or transaction failures,
    back up and start over from the beginning (restore connections and state, redo operations).
* On resources exhausted / overloaded, retry again later, with exponential back-off.
* On unimplemented methods, retry with a different implementation strategy, if there is one.
* When no better option is available, report the problem to a human (the user and/or the developer).


## C++

#### Headers

Use `#pragma once` to prevent multiple inclusion. It is supported on all
[relevant compilers](https://en.wikipedia.org/wiki/Pragma_once#Portability)
and it is less error prone.

Include the header as first `#include` in corresponding `.cpp` file to ensure
that the header is independent.

#### Unit tests

Unit tests are placed in the same directory as the code being tested in "-test.cpp".
Unit tests placed in seperate directory force a lot of jumping around directory structure.

#### Enums

Enum variables are effectively just ints and may hold value that is not matching
any declared enum option. Always handle these unknown values
(Lazy Input Validation).

````c++
    Enum Vehicle { CAR, TRAIN };

    std::string vehicleName(Vehicle x) {
        switch (x) {
            case Vehicle::CAR:
                return "car";
            case Vehicle::TRAIN:
                return "train";
        }
        return ds::str("unknown(", static_cast<int>(x), ")");
    }
````

## Formating rules

## Naming

* Type names: `TitleCase`
* Variable, member, function, and method names: `camelCase`
* Enumerant names: `camelCase` - in line with json serialization
* Constant names: `camelCase` - they are variables, just non-mutable
* Macros: `CAPITAL_WITH_UNDERSCORES`
* Files: module-name.cpp, module-name.h, module-name-test.cpp
* Libraries: ds, ds-sqlite3
* Folders: prefer single words, follow rules for files
* Private member class variables are prefixed with `m_`
* Public bember struct  are not prefixed with `m_`.
* Static globally visible variables are prefixed with `g_` (but avoided as much as possible)
* Static member and static local variables are prefixed with `s_`
* Setters start with `set`, getters don't use `get` prefix.
* Avoid abbreviation, do not delete letters within word
* Acronyms: all letters follow the case of the first letter. I.e. dsDevice, DSDevice, sslSocket, openSSLSocket
* Trivial functions should be inlined.

### Spacing and bracing

Do not waste time fixing and reviewing white space. Use `clang-format`.

TODO(soon): As `clang-format` produces different output on different versions,
we need to align on one version.

* Indents are 2 spaces
* Never use tabs
* Namespaces do not indent
* Maximum line length is 120 characters
* Indent a continuation line 2x indent = 4 spaces
* Place a space between a keyword and an open parenthesis, e.g.: `if (foo)`
* Do not place a space between a function name and an open parenthesis, e.g.: `foo(bar)`
* Place an opening brace at the end of the statement which initiates the block,
    not on its own line.
* Place a closing brace on a new line indented the same as the parent block.
    If there is post-brace code related to the block (e.g. `else` or `while`),
    place it on the same line as the closing brace.

````c++
    if (s == "a") {
        return "A";
    } else if (s == "B") {
        return "B";
    } else {
        return "";
    }
````

* Always place braces around a block.

````c++
    if (done) {
        return;
    }
````

* `case` statements are indented within the switch, and their following
    blocks are _further_ indented (so the actual statements in a case
    are 2x more indented than the `switch`).

````c++
    switch (x) {
        case 0:
            return "a";
        case 1: {
            std::string s = "a";
            return s + s;
        }
        default: {
            std::string s = "b";
            return b + b;
        }
    }
````

* `public:`, `private:`, and `protected:` are reverse-indented by one stop.
* Set your editor to strip trailing whitespace on save, otherwise other people
    who use this setting will see spurious diffs when they edit a file after you.

### Comments

Always use line comments (`//`, `///`, `///<`). Never use block comments (`/**/`).
Use one space after `//`. Use one space before `//` or align with comments around.

````cpp
    return true; // Success
````

TODO comments are of the form `// TODO(type): description`, where `type` is one of:
* now: Do before next git push.
* soon: Do before next release.
* someday: A feature that might be nice to have some day, but no urgency.
* perf: Possible performance enhancement.
* memory: Possible memory overhead enhancement.
* cleanup: An improvement to maintainability with no user-visible effects.
* test: Something that needs better testing.
* upstream: Code upstreamable to more generic place.
* others: Additional TODO types may be defined for use in certain contexts.
