/*
    Copyright (c) 2013 digitalSTROM.org, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <unistd.h>
#include <stdlib.h>

#include "watchdog.h"
#include "datetools.h"
#include "dss.h"
#include "logger.h"
#include "event.h"

#define WATCHDOG_SLEEP_TIME 360 // in seconds, 3 minutes
namespace dss {

Watchdog::Watchdog(DSS* _pDSS) : ThreadedSubsystem(_pDSS, "Watchdog") {
}

void Watchdog::initialize() {
  Subsystem::initialize();
}

void Watchdog::doStart() {
  run();
}

void Watchdog::dumpLastEvent(void) {
  if (!DSS::hasInstance()) {
    log(std::string(__func__) + " no instance", lsNotice);
    return;
  }

  PropertyNodePtr evtMonitor = getDSS().getPropertySystem().getProperty("/system/EventInterpreter");
  if (!evtMonitor) {
    log(std::string(__func__) + " no monitor", lsNotice);
    return;
  }

  PropertyNodePtr time = evtMonitor->getProperty("running/time");
  PropertyNodePtr event = evtMonitor->getProperty("running/event");
  PropertyNodePtr handler = evtMonitor->getProperty("running/handler");

  std::string timeStr = "<unknown>";
  std::string eventStr = "<unknown>";
  std::string handlerStr = "<unknown>";

  if (time) {
    timeStr = time->getAsString();
  }

  if (event) {
    eventStr = event->getAsString();
  }

  if (handler) {
    handlerStr = handler->getAsString();
  }

  log("Watchdog last event: " + eventStr + " runtime: " + timeStr + " handler: " + handlerStr, lsFatal);
}

void Watchdog::execute() {
  while (!m_Terminated) {
    int sleepTimeSec = WATCHDOG_SLEEP_TIME;
    int eventCount = getDSS().getEventInterpreter().getEventsProcessed();
    while (!m_Terminated && (sleepTimeSec > 0)) {
       const int kMinSleepTimeSec = 10; // 10 seconds
       sleepSeconds(std::min(sleepTimeSec, kMinSleepTimeSec));
       sleepTimeSec -= kMinSleepTimeSec;
    }
    if (getDSS().getEventInterpreter().getEventsProcessed() == eventCount) {
      log("Watchdog noticed event interpreter lockup, event count stuck at " +
              intToString(eventCount), lsFatal);
      dumpLastEvent();
      abort();
    }
  }
}

} // namespace dss

