/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <algorithm>

#include "apartmentrequesthandler.h"


#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <digitalSTROM/dsuid.h>
#include <ds/elapsed.h>
#include <ds/locale/ds-locale.h>
#include <ds/log.h>
#include <external/civetweb/civetweb.h>
#include <boost/locale/generator.hpp>
#include <boost/make_shared.hpp>
#include <boost/range/algorithm_ext.hpp>
#include "includerapidjson.h"

#include "backend-vdcs.h"
#include "comm-channel.h"
#include "controller-statuses.h"
#include "device-info.h"
#include "dsa-http-client.h"
#include "event.h"
#include "handler/system_states.h"
#include "http_client.h"
#include "jsonhelper.h"
#include "model-features.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/deviceinterface.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "hal/hardware-abstraction.h"
#include "util.h"
#include "vdc-db.h"

#include "zonerequesthandler.h"

static ds::LogChannel dsLogChannel("jsonApiApartment");

namespace dss {

//=========================================== ApartmentRequestHandler

ApartmentRequestHandler::ApartmentRequestHandler(Apartment& apartment, ModelMaintenance& _modelMaintenance)
    : DeviceInterfaceRequestHandler(apartment), m_Apartment(apartment), m_ModelMaintenance(_modelMaintenance) {}

std::string ApartmentRequestHandler::getReachableGroups(const RestfulRequest& _request) {
  JSONWriter json;
  json.startArray("zones");

  std::vector<boost::shared_ptr<Zone> > zones = DSS::getInstance()->getApartment().getZones();
  for (size_t i = 0; i < zones.size(); i++) {
    boost::shared_ptr<Zone> zone = zones.at(i);
    if (zone == NULL) {
      continue;
    }

    json.startObject();
    json.add("zoneID", zone->getID());
    json.add("name", zone->getName());
    json.startArray("groups");

    std::vector<boost::shared_ptr<Group> > groups = zone->getGroups();
    for (size_t g = 0; g < groups.size(); g++) {
      boost::shared_ptr<Group> group = groups.at(g);
      if (group->getID() == GroupIDBroadcast) {
        continue;
      }
      if (zone->getID() == 0) {
        if (group->getID() < GroupIDGlobalAppMin) {
          continue;
        }
      } else {
        if ((group->getID() == GroupIDBlack) || (group->getID() == GroupIDRed) || (group->getID() == GroupIDGreen)) {
          continue;
        }
      }

      bool hasRelevantDevices = false;
      Set devices = group->getDevices();
      for (int d = 0; d < devices.length(); d++) {
        boost::shared_ptr<Device> device = devices.get(d).getDevice();
        if ((device->hasOutput() && (device->getOutputMode() != 0)) || device->getHasActions()) {
          hasRelevantDevices = true;
          break;
        }
      } // devices loop

      if (hasRelevantDevices || group->hasConnectedDevices()) {
        json.add(group->getID());
      }
    } // groups loop
    json.endArray();
    json.endObject();
  } // zones loop
  json.endArray();

  return json.successJSON();
}

std::string ApartmentRequestHandler::removeMeter(const RestfulRequest& _request) {
  std::string dsidStr = _request.getParameter("dsid");
  std::string dsuidStr = _request.getParameter("dsuid");
  if (dsidStr.empty() && dsuidStr.empty()) {
    return JSONWriter::failure("Missing parameter 'dsuid'");
  }

  dsuid_t meterID = dsidOrDsuid2dsuid(dsidStr, dsuidStr);

  boost::shared_ptr<DSMeter> meter = DSS::getInstance()->getApartment().getDSMeterByDSID(meterID);

  if (meter->isPresent()) {
    // ATTENTION: this is string is translated by the web UI
    return JSONWriter::failure("Cannot remove present meter");
  }

  DS_NOTICE("Remove meter.", meterID);

  m_Apartment.removeDSMeter(meterID);
  m_ModelMaintenance.addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
  return JSONWriter::success();
}

WebServerResponse ApartmentRequestHandler::jsonHandleRequest(const RestfulRequest& _request,
                                                             boost::shared_ptr<Session> _session,
                                                             const struct mg_connection* _connection) {
  std::string errorMessage;
  if (_request.getMethod() == "getConsumption") {
    int accumulatedConsumption = 0;
    for (boost::shared_ptr<DSMeter> pDSMeter : m_Apartment.getDSMeters()) {
      accumulatedConsumption += pDSMeter->getPowerConsumption();
    }
    JSONWriter json;
    json.add("consumption", accumulatedConsumption);
    return json.successJSON();
  } else if (isDeviceInterfaceCall(_request)) {
    boost::shared_ptr<IDeviceInterface> interface;
    std::string groupName = _request.getParameter("groupName");
    std::string groupIDString = _request.getParameter("groupID");
    if (!groupName.empty()) {
      try {
        interface = m_Apartment.getGroup(groupName);
      } catch (std::runtime_error& e) {
        return JSONWriter::failure("Could not find group with name '" + groupName + "'");
      }
    } else if (!groupIDString.empty()) {
      try {
        int groupID = strToIntDef(groupIDString, -1);
        if (groupID != -1) {
          interface = m_Apartment.getGroup(groupID);
        } else {
          return JSONWriter::failure("Could not parse group ID '" + groupIDString + "'");
        }
      } catch (std::runtime_error& e) {
        return JSONWriter::failure("Could not find group with ID '" + groupIDString + "'");
      }
    }
    if (interface == NULL) {
      interface = m_Apartment.getGroup(GroupIDBroadcast);
    }

    return handleDeviceInterfaceRequest(_request, interface, _session);

  } else {
    if (_request.getMethod() == "getStructure") {
      JSONWriter json;
      toJSON(m_Apartment, json);
      return json.successJSON();
    } else if (_request.getMethod() == "getDevices") {
      Set devices;
      if (_request.getParameter("unassigned").empty()) {
        devices = m_Apartment.getDevices();
      } else {
        devices = getUnassignedDevices();
      }

      bool showHidden = false;
      _request.getParameter("showHidden", showHidden);

      JSONWriter json(JSONWriter::jsonArrayResult);
      toJSON(devices, json, showHidden);
      return json.successJSON();
    } else if (_request.getMethod() == "getCircuits") {
      JSONWriter json;

      json.startArray("circuits");
      std::vector<boost::shared_ptr<DSMeter> > dsMeters = m_Apartment.getDSMeters();
      for (boost::shared_ptr<DSMeter> dsMeter : dsMeters) {
        toJSON(dsMeter, json);
      }
      json.endArray();
      return json.successJSON();
    } else if (_request.getMethod() == "getVdcs") {
      JSONWriter json;

      std::string implementationId = _request.getParameter("implementationId");
      if (implementationId.empty()) {
        return JSONWriter::failure("Missing parameter 'implementationId'");
      }

      json.startArray("vdcs");
      std::vector<boost::shared_ptr<DSMeter> > dsMeters = m_Apartment.getDSMeters();
      for (boost::shared_ptr<DSMeter> dsMeter : dsMeters) {
        if (dsMeter->getVdcImplementationId() != implementationId) {
          continue;
        }
        toJSON(dsMeter, json);
      }
      json.endArray();
      return json.successJSON();
    } else if (_request.getMethod() == "getName") {
      JSONWriter json;
      json.add("name", m_Apartment.getName());
      return json.successJSON();
    } else if (_request.getMethod() == "setName") {
      const auto& name = escapeHTML(_request.getRequiredParameter("newName"));

      DS_NOTICE("Set apartment name.", name);

      m_Apartment.setName(name);
      DSS::getInstance()->getBonjourHandler().restart();
      return JSONWriter::success();
    } else if (_request.getMethod() == "rescan") {
      DS_NOTICE("Rescan.");

      std::vector<boost::shared_ptr<DSMeter> > mods = m_Apartment.getDSMeters();
      for (boost::shared_ptr<DSMeter> pDSMeter : mods) {
        pDSMeter->setIsValid(false);
      }
      return JSONWriter::success();
    } else if (_request.getMethod() == "muteFire") {
      m_Apartment.getModelMaintenance()->muteFire(coJSON);
      return JSONWriter::success();
    } else if (_request.getMethod() == "removeMeter") {
      return removeMeter(_request);
    } else if (_request.getMethod() == "removeInactiveMeters") {
      DS_NOTICE("Remove inactive meter.");

      m_Apartment.removeInactiveMeters();
      return JSONWriter::success();
    } else if (_request.getMethod() == "getReachableGroups") {
      return getReachableGroups(_request);
    } else if (_request.getMethod() == "getLockedScenes") {
      std::list<uint32_t> scenes = CommChannel::getInstance()->getLockedScenes();
      JSONWriter json;
      json.startArray("lockedScenes");
      while (!scenes.empty()) {
        json.add((int)scenes.front());
        scenes.pop_front();
      }
      json.endArray();
      return json.successJSON();

    } else if (_request.getMethod() == "getTemperatureControlStatus") {
      JSONWriter json;

      json.startArray("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        if (pZone->getID() == ZONE_ID_BROADCAST || pZone->getID() == NEW_DEVICES_ZONE_ID) {
          continue;
        }
        json.startObject();
        json.add("id", pZone->getID());
        json.add("name", pZone->getName());
        ZoneRequestHandler::addTemperatureControlStatus(json, pZone);
        json.endObject();
      }
      json.endArray();
      return json.successJSON();

    } else if (_request.getMethod() == "getTemperatureControlConfig") {
      JSONWriter json;

      json.startArray("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        if (pZone->getID() == ZONE_ID_BROADCAST || pZone->getID() == NEW_DEVICES_ZONE_ID) {
          continue;
        }
        json.startObject();
        json.add("id", pZone->getID());
        json.add("name", pZone->getName());
        ZoneRequestHandler::addTemperatureControlConfig(json, pZone);
        json.endObject();
      }
      json.endArray();
      return json.successJSON();

    } else if (_request.getMethod() == "getTemperatureControlConfig2") {
      JSONWriter json;

      json.startObject("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        if (pZone->getID() == ZONE_ID_BROADCAST || pZone->getID() == NEW_DEVICES_ZONE_ID) {
          continue;
        }
        json.startObject(ds::str(pZone->getID()));
        json.add("name", pZone->getName());

        json.startObject("config");
        ZoneRequestHandler::addTemperatureControlConfig2(json, pZone);
        json.endObject();

        json.endObject();
      }
      json.endObject();
      return json.successJSON();

    } else if (_request.getMethod() == "getTemperatureControlValues") {
      JSONWriter json;

      json.startArray("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        if (pZone->getID() == ZONE_ID_BROADCAST || pZone->getID() == NEW_DEVICES_ZONE_ID) {
          continue;
        }
        json.startObject();
        json.add("id", pZone->getID());
        json.add("name", pZone->getName());
        ZoneRequestHandler::addTemperatureControlValues(json, pZone);
        json.endObject();
      }
      json.endArray();
      return json.successJSON();

    } else if (_request.getMethod() == "getAssignedSensors") {
      JSONWriter json;

      json.startArray("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        json.startObject();
        json.add("id", pZone->getID());
        json.add("name", pZone->getName());

        json.startArray("sensors");

        for (auto&& devSensor : pZone->getAssignedSensors()) {
          json.startObject();
          json.add("sensorType", devSensor.m_sensorType);
          json.add("dsuid", devSensor.m_DSUID);
          json.endObject();
        }
        json.endArray();
        json.endObject();
      }
      json.endArray();
      return json.successJSON();

    } else if (_request.getMethod() == "getSensorValues") {
      ApartmentSensorStatus_t aSensorStatus = m_Apartment.getSensorStatus();
      JSONWriter json;

      json.startObject("weather");
      if (aSensorStatus.m_WeatherTS != DateTime::NullDate) {
        json.add("WeatherIconId", aSensorStatus.m_WeatherIconId);
        json.add("WeatherConditionId", aSensorStatus.m_WeatherConditionId);
        json.add("WeatherServiceId", aSensorStatus.m_WeatherServiceId);
        json.add("WeatherServiceTime", aSensorStatus.m_WeatherTS.toISO8601_ms());
      }
      json.endObject();

      json.startObject("outdoor");
      if (aSensorStatus.m_TemperatureTS != DateTime::NullDate) {
        json.startObject("temperature");
        json.add("value", aSensorStatus.m_TemperatureValue);
        json.add("time", aSensorStatus.m_TemperatureTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_HumidityTS != DateTime::NullDate) {
        json.startObject("humidity");
        json.add("value", aSensorStatus.m_HumidityValue);
        json.add("time", aSensorStatus.m_HumidityTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_BrightnessTS != DateTime::NullDate) {
        json.startObject("brightness");
        json.add("value", aSensorStatus.m_BrightnessValue);
        json.add("time", aSensorStatus.m_BrightnessTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_WindSpeedTS != DateTime::NullDate) {
        json.startObject("windspeed");
        json.add("value", aSensorStatus.m_WindSpeedValue);
        json.add("time", aSensorStatus.m_WindSpeedTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_WindDirectionTS != DateTime::NullDate) {
        json.startObject("winddirection");
        json.add("value", aSensorStatus.m_WindDirectionValue);
        json.add("time", aSensorStatus.m_WindDirectionTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_GustSpeedTS != DateTime::NullDate) {
        json.startObject("gustspeed");
        json.add("value", aSensorStatus.m_GustSpeedValue);
        json.add("time", aSensorStatus.m_GustSpeedTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_GustDirectionTS != DateTime::NullDate) {
        json.startObject("gustdirection");
        json.add("value", aSensorStatus.m_GustDirectionValue);
        json.add("time", aSensorStatus.m_GustDirectionTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_PrecipitationTS != DateTime::NullDate) {
        json.startObject("precipitation");
        json.add("value", aSensorStatus.m_PrecipitationValue);
        json.add("time", aSensorStatus.m_PrecipitationTS.toISO8601_ms());
        json.endObject();
      }
      if (aSensorStatus.m_AirPressureTS != DateTime::NullDate) {
        json.startObject("airpressure");
        json.add("value", aSensorStatus.m_AirPressureValue);
        json.add("time", aSensorStatus.m_AirPressureTS.toISO8601_ms());
        json.endObject();
      }
      json.endObject();

      json.startArray("zones");
      std::vector<boost::shared_ptr<Zone> > zoneList = m_Apartment.getZones();
      for (boost::shared_ptr<Zone> pZone : zoneList) {
        if (pZone->getID() == 0) {
          continue;
        }
        ZoneSensorStatus sensorStatus = pZone->getSensorStatus();
        json.startObject();
        json.add("id", pZone->getID());
        json.add("name", pZone->getName());

        json.startArray("values");

        if (sensorStatus.temperatureValueTS != DateTime::NullDate && sensorStatus.temperatureValue) {
          json.startObject();
          json.add("TemperatureValue", *sensorStatus.temperatureValue);
          json.add("TemperatureValueTime", sensorStatus.temperatureValueTS.toISO8601_ms());
          json.endObject();
        }
        if (sensorStatus.humidityValueTS != DateTime::NullDate && sensorStatus.humidityValue) {
          json.startObject();
          json.add("HumidityValue", *sensorStatus.humidityValue);
          json.add("HumidityValueTime", sensorStatus.humidityValueTS.toISO8601_ms());
          json.endObject();
        }
        if (sensorStatus.co2ConcentrationValueTS != DateTime::NullDate && sensorStatus.co2ConcentrationValue) {
          json.startObject();
          json.add("CO2concentrationValue", *sensorStatus.co2ConcentrationValue);
          json.add("CO2concentrationValueTime", sensorStatus.co2ConcentrationValueTS.toISO8601_ms());
          json.endObject();
        }
        if (sensorStatus.brightnessValueTS != DateTime::NullDate && sensorStatus.brightnessValue) {
          json.startObject();
          json.add("BrightnessValue", *sensorStatus.brightnessValue);
          json.add("BrightnessValueTime", sensorStatus.brightnessValueTS.toISO8601_ms());
          json.endObject();
        }
        json.endArray();
        json.endObject();
      }
      json.endArray();
      return json.successJSON();
    } else if (_request.getMethod() == "getModelFeatures") {
      JSONWriter json;

      json.startObject("modelFeatures");
      for (int colorID = ColorIDYellow; colorID <= ColorIDWhite; colorID++) {
        auto&& f = ModelFeatures::getInstance()->getFeatures(colorID);
        json.startObject(ModelFeatures::getInstance()->colorToString(colorID));

        for (size_t i = 0; i < f.size(); i++) {
          auto&& model = f.at(i);
          json.startObject(model->first);

          for (auto&& feature : *model->second) {
            if (auto&& name = ds::tryFromEnum<const char*>(feature)) {
              json.add(*name, true);
            }
          }
          json.endObject();
        }
        json.endObject();
      }
      json.endObject();

      json.startObject("reference");
      for (const auto& feature : ModelFeatures::getInstance()->getAvailableFeatures()) {
        if (const auto& name = ds::tryFromEnum<const char*>(feature)) {
          json.add(*name, false);
        }
      }
      json.endObject();

      return json.successJSON();
    } else if (_request.getMethod() == "getClusterLocks") {
      JSONWriter json;

      std::pair<std::vector<DeviceLock_t>, std::vector<ZoneLock_t> > locks;
      std::vector<DeviceLock_t> lockedDevices;
      std::vector<ZoneLock_t> lockedZones;

      locks = m_Apartment.getClusterLocks();
      lockedDevices = locks.first;
      lockedZones = locks.second;

      json.startArray("lockedDevices");

      for (size_t i = 0; i < lockedDevices.size(); i++) {
        json.startObject();
        json.add("dsuid", lockedDevices.at(i).dsuid);

        json.startArray("lockedScenes");
        for (size_t j = 0; j < lockedDevices.at(i).lockedScenes.size(); j++) {
          json.add(lockedDevices.at(i).lockedScenes.at(j));
        }
        json.endArray();

        json.endObject();
      }
      json.endArray();

      // lockedZones: [ { zoneID: xx, deviceClasses: [ { deviceClass: 3, lockedScenes: [ 1, 2, 3 ] } ] }, ... ]
      json.startArray("lockedZones");
      for (size_t i = 0; i < lockedZones.size(); i++) {
        ZoneLock_t zl = lockedZones.at(i);
        std::set<int> lockedApartmentScenes;
        json.startObject();
        json.add("zoneID", zl.zoneID);

        json.startArray("deviceClasses");
        for (size_t k = 0; k < zl.deviceClassLocks.size(); k++) {
          ClassLock_t cl = zl.deviceClassLocks.at(k);
          json.startObject();
          json.add("deviceClass", cl.deviceClass);
          json.startArray("lockedScenes");
          for (std::set<int>::iterator it = cl.lockedScenes.begin(); it != cl.lockedScenes.end(); it++) {
            int scene = *it;
            json.add(scene);
            // collect apartment scenes
            if (scene >= SceneAutoStandBy) {
              lockedApartmentScenes.insert(scene);
            }
          }
          json.endArray();
          json.endObject();
        }
        json.endArray();

        json.startArray("lockedApartmentScenes");

        for (std::set<int>::iterator it = lockedApartmentScenes.begin(); it != lockedApartmentScenes.end(); it++) {
          json.add(*it);
        }

        json.endArray();

        json.endObject();
      }
      json.endArray();

      return json.successJSON();
    } else if (_request.getMethod() == "setDevicesFirstSeen") {
      DateTime time = DateTime::parseISO8601(_request.getRequiredParameter("time"));

      std::string xDsTrackingId;
      if (const char* data = mg_get_header(_connection, "X-DS-TrackingID")) {
        std::string xDsTrackingId = data;
      }

      DS_NOTICE("Set devices first seen.", time, xDsTrackingId);

      if (!m_Apartment.setDevicesFirstSeen(time)) {
        return JSONWriter::failure("can not set date. Date too old.");
      }

      // log X-DS-TrackingID
      if (!xDsTrackingId.empty()) {
        // create event, log entry ends in system-event.log
        boost::shared_ptr<Event> evtDevicesFirstSeen = boost::make_shared<Event>(EventName::DevicesFirstSeen);
        evtDevicesFirstSeen->setProperty("dateTime", time.toISO8601());
        evtDevicesFirstSeen->setProperty("X-DS-TrackingID", xDsTrackingId);
        if (DSS::hasInstance()) {
          DSS::getInstance()->getEventQueue().pushEvent(evtDevicesFirstSeen);
        }
      }

      return JSONWriter::success();
    } else if (_request.getMethod() == "getDeviceBinaryInputs") {
      JSONWriter json;

      Apartment& apt = DSS::getInstance()->getApartment();
      Set devices = apt.getZone(0)->getDevices();

      json.startArray("devices");

      for (int d = 0; d < devices.length(); d++) {
        boost::shared_ptr<Device> device = devices.get(d).getDevice();
        if (device->isPresent() == true) {
          auto&& binaryInputs = device->getBinaryInputs();

          if (!binaryInputs.empty()) {
            json.startObject();
            json.add("dsuid", device->getDSID());

            json.startArray("binaryInputs");

            for (size_t i = 0; i < binaryInputs.size(); i++) {
              auto&& input = binaryInputs.at(i);
              json.startObject();
              json.add("targetGroup", input->m_targetGroupId);
              json.add("inputType", input->m_inputType);
              json.add("inputId", input->m_inputId);
              json.add("state", input->getState().getState());
              json.endObject();
            }

            json.endArray();
            json.endObject();
          }
        }
      }

      json.endArray();
      return json.successJSON();
    } else if (_request.getMethod() == "getDeviceInfo") {
      auto&& dss = *DSS::getInstance();
      std::string filterParam;
      int onlyActive = 0;
      std::string activeStr;

      // "filter" can be a comma separated combination of:
      // spec
      // stateDesc
      // propertyDesc
      // actionDesc
      // standardActions
      // customActions
      _request.getParameter("filter", filterParam);
      _request.getParameter("onlyActive", activeStr);
      onlyActive = strToIntDef(activeStr, 0);

      auto locale = _request.getLocale(dss.localeGenerator());
      auto prefer =
          ds::tryToEnum<DeviceInfo::Prefer>(_request.tryGetParameter("prefer")).value_or(DeviceInfo::Prefer());

      auto filter = DeviceInfo::parseFilter(filterParam);

      auto db = vdcDb::make(dss);
      JSONWriter json;

      Apartment& apt = dss.getApartment();
      auto query = apt.getBusInterface()->getStructureQueryBusInterface();

      // get all present devices
      Set devices = apt.getZone(0)->getDevices();
      if (onlyActive) {
        devices = devices.getByPresence(true);
      }

      json.startArray("devices");

      for (int d = 0; d < devices.length(); d++) {
        const boost::shared_ptr<Device> device = devices.get(d).getDevice();

        json.startObject();
        json.add("dSUID", device->getDSID());
        // do not fail the whole set if one devices messes up
        try {
          DeviceInfo(prefer, *query, *db, device.get(), "", locale, json, ds::system()).addByFilter(filter);
        } catch (std::exception& e) {
          DS_ERROR("DeviceInfo failed.", device, e);
        }
        json.endObject();
      }

      json.endArray();
      return json.successJSON();
    } else if (_request.getMethod() == "getControllerStatuses") {
      auto&& dss = *DSS::getInstance();
      auto locale = _request.getLocale(dss.localeGenerator());
      JSONWriter json;
      auto dsa = makeDsaHttpClient();
      json.addReflected("controllerStatuses", controllerStatuses(locale, *dsa, dss.backendVdcs()));
      return json.successJSON();
    } else {
      throw std::runtime_error("Unhandled function");
    }
  }
} // handleApartmentCall

Set ApartmentRequestHandler::getUnassignedDevices() {
  Apartment& apt = DSS::getInstance()->getApartment();
  Set devices = apt.getZone(0)->getDevices();

  std::vector<boost::shared_ptr<Zone> > zones = apt.getZones();
  for (std::vector<boost::shared_ptr<Zone> >::iterator ipZone = zones.begin(), e = zones.end(); ipZone != e; ++ipZone) {
    boost::shared_ptr<Zone> pZone = *ipZone;
    if (pZone->getID() == 0) {
      // zone 0 holds all devices, so we're going to skip it
      continue;
    }

    devices = devices.remove(pZone->getDevices());
  }

  return devices;
} // getUnassignedDevices

} // namespace dss
