/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "model/apartment.h"
#include "model/modelmaintenance-mock.h"
#include "structurerequesthandler.h"

static ds::LogChannel dsLogChannel("structureRequestHandler");

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const char* TAGS = "[web][handler][structure]";

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};
static const Dsuid dsuid4{"0000000000000000000000000000000004"};
static const Dsuid dsuid5{"0000000000000000000000000000000005"};
static const Dsuid dsuid6{"0000000000000000000000000000000006"};
static const Dsuid dsuid7{"0000000000000000000000000000000007"};

TEST_CASE("webHandlerStructureRequest", TAGS) {
  Apartment apartment(nullptr);
  auto busInterface = BusInterfaceMock{};
  auto modelMaintenance = ModelMaintenanceMock{};

  StructureRequestHandler structHandler(apartment, modelMaintenance, busInterface.structureModify,
                                        busInterface.structureQuery);

  /*
   * we still don't have defines for this, really?
   */
  uint16_t outputMode = 0x10;
  uint16_t extraEquipment = 0x4; // includes any type of sensor

  auto functionId = [&](DeviceClasses_t color, unsigned int baseConfig, int nrKeys) {
    uint16_t ret = color << 12;
    ret |= 0x4 << 6;    // R105 subclass
    ret |= (baseConfig & 0x3c); // bitmask guard

    switch (nrKeys) {
      case 4: ret |= 0x3; break;
      case 2: ret |= 0x2; break;
      case 1: ret |= 0x1; break;
      case 0: break;
    }
    return ret;
  };

  // int ret = ::dsuid_get_next_dsuid(&dsuid, out);
  auto createDevice = [&](auto dsuid, HardwareType type, int subindex) {
    auto dev = apartment.allocateDevice(dsuid);
    switch (type) {
      case HardwareType::GE_KM200:
        dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_GE, outputMode, 1),
                                             .ProductID = DEVICE_TYPE_KM << 10 | 200});
        break;
      case HardwareType::GE_TKM210:
        dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_GE, outputMode, 1),
                                             .ProductID = DEVICE_TYPE_TKM << 10 | 210});
        break;
      case HardwareType::GE_TKM230:
        // 2 inputs, no output
        dev->setPartiallyFromSpec(
            DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_GE, 0x0u, 2), .ProductID = DEVICE_TYPE_TKM << 10 | 230});
        break;
      case HardwareType::SW_AKM200:
        // 4 inputs, sensor only
        dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_SW, extraEquipment, 4),
                                             .ProductID = DEVICE_TYPE_AKM << 10 | 200});
        dev->setButtonInputMode(ButtonInputMode::AKM_STANDARD);
        break;
      case HardwareType::SW_UMR200:
        // 2 t|akm  2 inputs, configure as binary/taster + 2 relais output
        // reliable transport, in the back channel != regular TKM/AKM
        switch (subindex) {
          case 0:
            // input device, configure as sensor
            dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_SW, extraEquipment, 4),
                                                 .ProductID = DEVICE_TYPE_UMR << 10 | 200});
            dev->setButtonInputMode(ButtonInputMode::AKM_STANDARD);
            break;
          case 1:
            // input device, configure as button
            dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_SW, extraEquipment, 4),
                                                 .ProductID = DEVICE_TYPE_UMR << 10 | 200});
            break;
          case 2:
          case 3:
          default:
            // output device
            dev->setPartiallyFromSpec(DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_SW, outputMode, 0),
                                                 .ProductID = DEVICE_TYPE_UMR << 10 | 200});
            break;
        }
        break;
      case HardwareType::SW_TKM210:
        dev->setPartiallyFromSpec(
            DeviceSpec{.FunctionID = functionId(DEVICE_CLASS_SW, 0x0u, 4), .ProductID = DEVICE_TYPE_TKM << 10 | 210});
        break;
    }
    return dev;
  };

  auto devKM200 = createDevice(dsuid1, HardwareType::GE_KM200, 0);
  auto devTKM210 = createDevice(dsuid2, HardwareType::GE_TKM210, 0);
  auto devTKM230 = createDevice(dsuid3, HardwareType::GE_TKM230, 0);
  auto devAKM200 = createDevice(dsuid4, HardwareType::SW_AKM200, 0);
  auto devUMR200_input = createDevice(dsuid5, HardwareType::SW_UMR200, 1);
  auto devUMR200_sensor = createDevice(dsuid6, HardwareType::SW_UMR200, 0);
  auto devSW_TKM210 = createDevice(dsuid7, HardwareType::SW_TKM210, 0);

  SECTION("Multi user group membership permitted") {
    CHECK(devKM200->getDeviceClass() == DEVICE_CLASS_GE);
    CHECK(devKM200->getDeviceType() == DEVICE_TYPE_KM);
    CHECK(devKM200->hasOutput() == true);
    CHECK(structHandler.onlyOneGroupMembership(*devKM200) == false);

    CHECK(devTKM210->getDeviceClass() == DEVICE_CLASS_GE);
    CHECK(devTKM210->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(devTKM210->hasOutput() == true);
    CHECK(structHandler.onlyOneGroupMembership(*devTKM210) == false);

    CHECK(devTKM230->getDeviceClass() == DEVICE_CLASS_GE);
    CHECK(devTKM230->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(devTKM230->hasOutput() == false);
    CHECK(devTKM230->hasInput() == true);
    CHECK(structHandler.onlyOneGroupMembership(*devTKM230) == true);

    // sensors can be in multiple groups
    CHECK(devAKM200->getDeviceClass() == DEVICE_CLASS_SW);
    CHECK(devAKM200->getDeviceType() == DEVICE_TYPE_AKM);
    CHECK(devAKM200->hasOutput() == false);
    CHECK(devAKM200->hasInput() == true);
    CHECK(devAKM200->isBinaryInput() == true);
    CHECK(structHandler.onlyOneGroupMembership(*devAKM200) == false);

    CHECK(devUMR200_input->getDeviceClass() == DEVICE_CLASS_SW);
    CHECK(devUMR200_input->getDeviceType() == DEVICE_TYPE_UMR);
    CHECK(devUMR200_input->hasOutput() == false);
    CHECK(devUMR200_input->hasInput() == true);
    CHECK(devUMR200_input->isBinaryInput() == false);
    CHECK(structHandler.onlyOneGroupMembership(*devUMR200_input) == true);

    CHECK(devUMR200_sensor->getDeviceClass() == DEVICE_CLASS_SW);
    CHECK(devUMR200_sensor->getDeviceType() == DEVICE_TYPE_UMR);
    CHECK(devUMR200_sensor->hasOutput() == false);
    CHECK(devUMR200_sensor->hasInput() == true);
    CHECK(devUMR200_sensor->isBinaryInput() == true);
    CHECK(structHandler.onlyOneGroupMembership(*devUMR200_sensor) == false);

    CHECK(devSW_TKM210->getDeviceClass() == DEVICE_CLASS_SW);
    CHECK(devSW_TKM210->getDeviceType() == DEVICE_TYPE_TKM);
    CHECK(devSW_TKM210->hasOutput() == false);
    CHECK(devSW_TKM210->hasInput() == true);
    CHECK(devSW_TKM210->isBinaryInput() == false);
    CHECK(structHandler.onlyOneGroupMembership(*devSW_TKM210) == true);
  }
}

} // namespace
