/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif


#include "webrequests.h"

#include <ds/enum.h>
#include <boost/shared_ptr.hpp>

#include <external/civetweb/civetweb.h>

namespace dss {


  //================================================== WebServerRequestHandlerJSON

  JSONWriter::JSONWriter(jsonResult_t _responseType) : m_writer(m_buffer), m_resultType(_responseType) {
    startObject();
    switch (m_resultType) {
    case jsonObjectResult:
      startObject("result");
      break;
    case jsonArrayResult:
      startArray("result");
      break;
    case jsonNoneResult:
      break;
    }
  }

  std::string JSONWriter::successJSON() {
    switch (m_resultType) {
    case jsonObjectResult:
      endObject();
      add("ok", true);
      break;
    case jsonArrayResult:
      endArray();
      add("ok", true);
      break;
    case jsonNoneResult:
      break;
    }
    endObject();
    return m_buffer.GetString();
  }
  void JSONWriter::add(const std::string &value) {
    add(value.c_str());
  }
  void JSONWriter::add(const char* _value) {
    m_writer.String(_value);
  }
  void JSONWriter::add(int _value) {
    m_writer.Int(_value);
  }
  void JSONWriter::add(unsigned _value) {
    m_writer.Uint(_value);
  }
  void JSONWriter::add(long long int _value) {
    m_writer.Int64(_value);
  }
  void JSONWriter::add(unsigned long long int _value) {
    m_writer.Uint64(_value);
  }
  void JSONWriter::add(bool _value) {
    m_writer.Bool(_value);
  }
  void JSONWriter::add(double _value) {
    m_writer.Double(_value);
  }
  bool JSONWriter::addKey(ModelFeatureId x) {
    if (const auto& name = ds::tryFromEnum<const char*>(x)) {
      add(*name);
      return true;
    } else {
      // Do not expose unknown model features to public api.
      // We cannot give them real name.
      return false;
    }
  }

  void JSONWriter::addNull() {
    m_writer.Null();
  }
  void JSONWriter::startArray(std::string _name) {
    m_writer.String(_name);
    m_writer.StartArray();
  }
  void JSONWriter::startArray() {
    m_writer.StartArray();
  }
  void JSONWriter::endArray() {
    m_writer.EndArray();
  }
  void JSONWriter::startObject(std::string _name) {
    m_writer.String(_name);
    m_writer.StartObject();
  }
  void JSONWriter::startObject() {
    m_writer.StartObject();
  }
  void JSONWriter::endObject() {
    m_writer.EndObject();
  }
  std::string JSONWriter::success() {
    return success("");
  }
  std::string JSONWriter::success(std::string _message) {
    StringBuffer s;
    Writer<StringBuffer> writer(s);
    writer.StartObject();
    writer.String("ok"); writer.Bool(true);
    if (!_message.empty()) {
      writer.String("message");
      writer.String(_message);
    }
    writer.EndObject();
    return s.GetString();
  }
  std::string JSONWriter::failure(std::string _message) {
    StringBuffer s;
    Writer<StringBuffer> writer(s);
    s.Clear();
    writer.Reset(s);
    writer.StartObject();
    writer.String("ok"); writer.Bool(false);
    if (!_message.empty()) {
      writer.String("message");
      writer.String(_message);
    }
    writer.EndObject();
    return s.GetString();
  }

  std::string JSONWriter::failure(const std::exception& e) {
    StringBuffer s;
    Writer<StringBuffer> writer(s);
    s.Clear();
    writer.Reset(s);
    writer.StartObject();
    writer.String("ok");
    writer.Bool(false);
    writer.String("message");
    writer.String(e.what());
    if (auto dsException = dynamic_cast<const ds::Exception*>(&e)) {
      const auto& userMessage = dsException->userMessage();
      if (!userMessage.empty()) {
        writer.String("userMessage");
        writer.String(userMessage);
      }
      const auto& type = dsException->type();
      if (type != ds::Exception::Type::FAILED) {
        writer.String("type");
        writer.String(ds::fromEnum<std::string>(type));
      }
    }
    writer.EndObject();
    return s.GetString();
  }

  std::string JSONWriter::raw() {
    return m_buffer.GetString();
  }
} // namespace dss
