/*
    Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "connection.h"

#include <boost/algorithm/cxx11/any_of.hpp>
#include <boost/algorithm/cxx11/none_of.hpp>
#include <boost/range/adaptors.hpp>
#include <boost/range/algorithm/find_if.hpp>
#include <boost/range/algorithm_ext/erase.hpp>
#include <boost/thread.hpp>
#include <boost/thread/locks.hpp>

#include "event.h"

namespace dss {

namespace websocket {

namespace {

ds::LogChannel dsLogChannel("webWebsocketConnection");

} // namespace

void Connection::addMsgToSend(const std::string& msg) {
  boost::unique_lock<boost::mutex> lock(m_msgQueueMtx);
  if (m_msgQueue.size() > m_maxMessagesNo) {
    return;
  }
  m_msgQueue.push_back(msg);
  m_msgQueueCond.notify_all();
}

void Connection::startMsgSendingThread() {
  boost::thread t{sendMsgWorker, shared_from_this()};
  t.detach();
}

void Connection::sendMsgWorker(const boost::shared_ptr<Connection> connection) {
  /// take own copy of shared_ptr to avoid dangling reference

  DS_DEBUG_ENTER("Start new thread: ", reinterpret_cast<size_t>(connection.get()));
  try {
    while (connection->m_state != State::ws_disconnected) {
      boost::unique_lock<boost::mutex> lock(connection->m_msgQueueMtx);
      if (connection->m_msgQueue.empty()) {
        connection->m_msgQueueCond.wait(lock);
      }

      while (!connection->m_msgQueue.empty()) {
        auto msg = std::move(connection->m_msgQueue.front());
        connection->m_msgQueue.pop_front();
        lock.unlock();
        connection->sendMsg(msg);
        lock.lock();
      }
    }
  } catch (const std::exception& e) {
    DS_WARNING(e)
  }
  DS_DEBUG_LEAVE("Stop thread: ", reinterpret_cast<size_t>(connection.get()));
}

void Connection::sendMsg(const std::string& msg) {
  if (m_state == State::ws_ready) {
    mg_context* ctx = mg_get_context(m_connection); // Could this be dangling ptr?
    mg_lock_context(ctx);
    const std::unique_ptr<mg_context, decltype(&mg_unlock_context)> unlockContext{ctx, mg_unlock_context};
    mg_websocket_write(m_connection, MG_WEBSOCKET_OPCODE_TEXT, msg.c_str(), msg.size());
  }
}

void Connection::setState(State s) {
  m_state = s;
  if (s == State::ws_disconnected) {
    m_msgQueueCond.notify_all();
  }
}

void JsonApiWebSocketConnection::addObserver(const std::string& id, std::unique_ptr<EventObserver>&& o) {
  if (!o) {
    return;
  }
  boost::unique_lock<boost::mutex> lock(m_observerMtx);
  m_observers[id] = std::move(o);
}

void JsonApiWebSocketConnection::removeObserver(const std::string& id) {
  boost::unique_lock<boost::mutex> lock(m_observerMtx);
  m_observers.erase(id);
}

bool JsonApiWebSocketConnection::hasObserverFor(const Event& e) const {
  boost::unique_lock<boost::mutex> lock(m_observerMtx);
  for (const auto& o : m_observers) {
    if (o.second->checkIfForMe(e)) {
      return true;
    }
  }
  return false;
}

std::vector<JsonApiWebSocketConnection::MsgId> JsonApiWebSocketConnection::observersLokingFor(const Event& e) const {
  boost::unique_lock<boost::mutex> lock(m_observerMtx);
  std::vector<JsonApiWebSocketConnection::MsgId> ids;
  for (const auto& o : m_observers) {
    if (o.second->checkIfForMe(e)) {
      ids.push_back(o.first);
    }
  }
  return ids;
}

ds::JsonFragment JsonApiWebSocketConnection::tryCreateJsonFragmentByObserver(
    const JsonApiWebSocketConnection::MsgId& id, const Event& e) const {
  boost::unique_lock<boost::mutex> lock(m_observerMtx);
  const auto& found = m_observers.find(id);
  return found == m_observers.end() ? ds::JsonFragment() : found->second->toJson(e);
}

} // namespace websocket

} // namespace dss
