/*
    Copyright (c) 2017 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <boost/container/map.hpp>
#include <deque>
#include <map>
#include <string>
#include <vector>

#include <boost/atomic.hpp>
#include <boost/enable_shared_from_this.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/condition_variable.hpp>
#include <boost/thread/mutex.hpp>

#include <external/civetweb/civetweb.h>

#include "web/websocket/eventobserver.h"
#include "web/websocket/message.h"

#include "dssfwd.h"

namespace dss {

namespace websocket {

enum class State { ws_disconnected = 0, ws_connected, ws_ready };

///
/// \brief The Connection struct is responsible for single WebSocket connection.
///
/// Each Connection have to start thread by _startThread_ to send data to client.
///
class Connection : public boost::enable_shared_from_this<Connection> {
public:
  enum { DEFALUT_MAX_MESSAGES_NO = 64 };
  Connection(mg_connection* connection, std::deque<std::string>::size_type maxMessagesNo = DEFALUT_MAX_MESSAGES_NO)
      : m_connection{connection}, m_maxMessagesNo{maxMessagesNo} {}

  virtual ~Connection() {}

  void addMsgToSend(const std::string& msg);

  void startMsgSendingThread();

  static void sendMsgWorker(const boost::shared_ptr<Connection> connection);

  void sendMsg(const std::string& msg);

  void setState(State s);

  bool isReady() const { return m_state == State::ws_ready; }

private:
  mg_connection* const m_connection;
  boost::atomic<State> m_state;

  mutable boost::condition_variable m_msgQueueCond;
  mutable boost::mutex m_msgQueueMtx;
  std::deque<std::string> m_msgQueue;
  std::deque<std::string>::size_type m_maxMessagesNo;
};

class JsonApiWebSocketConnection : public Connection {
public:
  typedef std::string MsgId;

  JsonApiWebSocketConnection(
      mg_connection* connection, std::deque<std::string>::size_type maxMessagesNo = DEFALUT_MAX_MESSAGES_NO)
      : Connection{connection, maxMessagesNo} {}

  void addObserver(const std::string& id, std::unique_ptr<EventObserver>&& o);

  void removeObserver(const std::string& id);

  bool hasObserverFor(const Event& e) const;

  std::vector<MsgId> observersLokingFor(const Event& e) const;

  ds::JsonFragment tryCreateJsonFragmentByObserver(const MsgId& id, const Event& e) const;

private:
  mutable boost::mutex m_observerMtx;
  boost::container::map<MsgId, std::unique_ptr<EventObserver>> m_observers;
};

} // namespace websocket

} // namespace dss
